/*
 * Copyright (c) 1998-2008 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioClip.cpp
 *
 * @brief	整数型←→浮動小数点型の変換．
 *			デバイスは整数型データを扱い、Mac OS X の Audio HAL は浮動小数点型データを扱う．
 *			このファイルはＣ言語で書かれており、ライブラリ（libDJM-V10FloatLib.a）
 *			としてコンパイルされたのち、kext にリンクされる．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/

/*------------------------------------------*/
/*				インクルード				*/
/*------------------------------------------*/

#include "PdjUsbAudioCommon.h"

extern "C" {

/*------------------------------------------*/
/*				定数						*/
/*------------------------------------------*/

#define kMaxClipSInt8		0.9921875
#define kFloat32ToSInt8		((Float32)0x80)

#define kMaxClipSInt16		0.9999694824219
#define kFloat32ToSInt16	((Float32)0x8000)

#define kMaxClipSInt24		0.9999998807907

#define kMaxClipSInt32		0.9999999995343387		// <rdar://7138492>
#define kFloat32ToSInt32	((Float64)0x80000000)

static const float		kOneOverMaxSInt8Value	= 1.0 / 128.0f;
static const float		kOneOverMaxSInt16Value	= 1.0 / 32768.0f;
// static const float	kOneOverMaxSInt24Value	= 1.0 / 8388608.0f;
static const float		kOneOverMaxSInt24Value	= 0.00000011920928955078125f;
static const float		kOneOverMaxSInt32Value	= 1.0 / 2147483648.0f;

#pragma mark - Private Functions -

/****************************************************************************/
/**
 * @brief	32bit 浮動小数点数を、
 *			符号付き 8bit 整数型に変換できる上限および下限の値で飽和させる．
 *
 * @param[in]		inSample		入力データ（32bit 浮動小数点数）．
 *
 * @return			飽和処理後のデータ（32bit 浮動小数点数）．
 */
inline static Float32
ClipFloat32ForSInt8(
	Float32			inSample
)
{
	// Float32		maxClip = kMaxSampleSInt8 / (kMaxSampleSInt8 + 1.0);

	if (inSample > kMaxClipSInt8)
	{
		return kMaxClipSInt8;
	}

	if (inSample < -1.0)
	{
		return -1.0;
	}

	return inSample;
}

/****************************************************************************/
/**
 * @brief	32bit 浮動小数点数を、
 *			符号付き 16bit 整数型に変換できる上限および下限の値で飽和させる．
 *
 * @param[in]		inSample		入力データ（32bit 浮動小数点数）．
 *
 * @return			飽和処理後のデータ（32bit 浮動小数点数）．
 */
inline static Float32
ClipFloat32ForSInt16(
	Float32			inSample
)
{
	// Float32		maxClip = kMaxSampleSInt16 / (kMaxSampleSInt16 + 1.0);

	if (inSample > kMaxClipSInt16)
	{
		return kMaxClipSInt16;
	}

	if (inSample < -1.0)
	{
		return -1.0;
	}

	return inSample;
}

/****************************************************************************/
/**
 * @brief	32bit 浮動小数点数を、
 *			符号付き 24bit 整数型に変換できる上限および下限の値で飽和させる．
 *
 * @param[in]		inSample		入力データ（32bit 浮動小数点数）．
 *
 * @return			飽和処理後のデータ（32bit 浮動小数点数）．
 */
inline static Float32
ClipFloat32ForSInt24(
	Float32			inSample
)
{
	// Float32		maxClip = kMaxSampleSInt24 / (kMaxSampleSInt24 + 1.0);

	if (inSample > kMaxClipSInt24)
	{
		return kMaxClipSInt24;
	}

	if (inSample < -1.0)
	{
		return -1.0;
	}

	return inSample;
}

/****************************************************************************/
/**
 * @brief	64bit 浮動小数点数を、
 *			符号付き 32bit 整数型に変換できる上限および下限の値で飽和させる．
 *
 * @param[in]		inSample		入力データ（64bit 浮動小数点数）．
 *
 * @return			飽和処理後のデータ（64bit 浮動小数点数）．
 */
// <rdar://7138492>
inline static Float64
ClipFloat32ForSInt32(
	Float64			inSample
)
{
	// Float64		maxClip64 = kMaxSampleSInt32 / (kMaxSampleSInt32 + 1.0);
	// Float32		maxClip = maxClip64;

	if (inSample > kMaxClipSInt32)
	{
		return kMaxClipSInt32;
	}

	if (inSample < -1.0)
	{
		return -1.0;
	}

	return inSample;
}

/****************************************************************************/
/**
 * @brief	符号付き 16bit 整数をリトルエンディアンに変換．
 *
 * @param[in]		inValue			ネイティブのエンディアンで表された符号付き 16bit 整数．
 *
 * @return			リトルエンディアンに変換された符号付き 16bit 整数．
 */
inline static SInt16
SInt16NativeToLittleEndian(
	SInt16			inValue
)
{
	return inValue;
}

/****************************************************************************/
/**
 * @brief	符号付き 32bit 整数をリトルエンディアンに変換．
 *
 * @param[in]		inValue			ネイティブのエンディアンで表された符号付き 32bit 整数．
 *
 * @return			リトルエンディアンに変換された符号付き 32bit 整数．
 */
inline static SInt32
SInt32NativeToLittleEndian(
	SInt32			inValue
)
{
	return inValue;
}

/****************************************************************************/
/**
 * @brief	32bit 浮動小数点数を符号付き 8bit 整数型に変換．
 *
 * @param[in]		*inInputBuffer		入力データバッファ（32bit 浮動小数点型）．
 * @param[out]		*outOutputBuffer	出力データバッファ（符号付き 8bit 整数型）．
 * @param[in]		inNumberSamples		変換するサンプル数．
 */
// Float32 -> SInt8
static void
ClipFloat32ToSInt8_4(
	const Float32	*inInputBuffer,
	SInt8			*outOutputBuffer,
	UInt32			inNumberSamples
)
{
	register UInt32		theLeftOvers = inNumberSamples % 4;

	while (inNumberSamples > theLeftOvers)
	{
		register Float32	theFloat32Value1 = *(inInputBuffer + 0);
		register Float32	theFloat32Value2 = *(inInputBuffer + 1);
		register Float32	theFloat32Value3 = *(inInputBuffer + 2);
		register Float32	theFloat32Value4 = *(inInputBuffer + 3);

		inInputBuffer += 4;

		theFloat32Value1 = ClipFloat32ForSInt8(theFloat32Value1);
		theFloat32Value2 = ClipFloat32ForSInt8(theFloat32Value2);
		theFloat32Value3 = ClipFloat32ForSInt8(theFloat32Value3);
		theFloat32Value4 = ClipFloat32ForSInt8(theFloat32Value4);

		*(outOutputBuffer + 0) = (SInt8)(theFloat32Value1 * kFloat32ToSInt8);
		*(outOutputBuffer + 1) = (SInt8)(theFloat32Value2 * kFloat32ToSInt8);
		*(outOutputBuffer + 2) = (SInt8)(theFloat32Value3 * kFloat32ToSInt8);
		*(outOutputBuffer + 3) = (SInt8)(theFloat32Value4 * kFloat32ToSInt8);

		outOutputBuffer += 4;
		inNumberSamples -= 4;
	}

	while (inNumberSamples > 0)
	{
		register Float32	theFloat32Value = *inInputBuffer;

		++inInputBuffer;

		theFloat32Value = ClipFloat32ForSInt8(theFloat32Value);

		*outOutputBuffer = (SInt8)(theFloat32Value * kFloat32ToSInt8);

		++outOutputBuffer;
		--inNumberSamples;
	}
}

/****************************************************************************/
/**
 * @brief	32bit 浮動小数点数を符号付き 16bit 整数型に変換．
 *
 * @param[in]		*inInputBuffer		入力データバッファ（32bit 浮動小数点型）．
 * @param[out]		*outOutputBuffer	出力データバッファ（符号付き 16bit 整数型）．
 * @param[in]		inNumberSamples		変換するサンプル数．
 */
// Float32 -> SInt16
static void
ClipFloat32ToSInt16LE_4(
	const Float32	*inInputBuffer,
	SInt16			*outOutputBuffer,
	UInt32			inNumberSamples
)
{
	register UInt32		theLeftOvers = inNumberSamples % 4;

	while (inNumberSamples > theLeftOvers)
	{
		register Float32	theFloat32Value1 = *(inInputBuffer + 0);
		register Float32	theFloat32Value2 = *(inInputBuffer + 1);
		register Float32	theFloat32Value3 = *(inInputBuffer + 2);
		register Float32	theFloat32Value4 = *(inInputBuffer + 3);

		inInputBuffer += 4;

		theFloat32Value1 = ClipFloat32ForSInt16(theFloat32Value1);
		theFloat32Value2 = ClipFloat32ForSInt16(theFloat32Value2);
		theFloat32Value3 = ClipFloat32ForSInt16(theFloat32Value3);
		theFloat32Value4 = ClipFloat32ForSInt16(theFloat32Value4);

		*(outOutputBuffer + 0) = SInt16NativeToLittleEndian((SInt16)(theFloat32Value1 * kFloat32ToSInt16));
		*(outOutputBuffer + 1) = SInt16NativeToLittleEndian((SInt16)(theFloat32Value2 * kFloat32ToSInt16));
		*(outOutputBuffer + 2) = SInt16NativeToLittleEndian((SInt16)(theFloat32Value3 * kFloat32ToSInt16));
		*(outOutputBuffer + 3) = SInt16NativeToLittleEndian((SInt16)(theFloat32Value4 * kFloat32ToSInt16));

		outOutputBuffer += 4;
		inNumberSamples -= 4;
	}

	while (inNumberSamples > 0)
	{
		register Float32	theFloat32Value = *inInputBuffer;

		++inInputBuffer;

		theFloat32Value = ClipFloat32ForSInt16(theFloat32Value);

		*outOutputBuffer = SInt16NativeToLittleEndian((SInt16)(theFloat32Value * kFloat32ToSInt16));

		++outOutputBuffer;
		--inNumberSamples;
	}
}

/****************************************************************************/
/**
 * @brief	32bit 浮動小数点数を符号付き 24bit 整数型に変換．
 *
 * @param[in]		*inInputBuffer		入力データバッファ（32bit 浮動小数点型）．
 * @param[out]		*outOutputBuffer	出力データバッファ（符号付き 24bit 整数型）．
 * @param[in]		inNumberSamples		変換するサンプル数．
 */
// Float32 -> SInt24
// we use the MaxSInt32 value because of how we munge the data
static void
ClipFloat32ToSInt24LE_4(
	const Float32	*inInputBuffer,
	SInt32			*outOutputBuffer,
	UInt32			inNumberSamples
)
{
	register UInt32		theLeftOvers = inNumberSamples % 4;

	while (inNumberSamples > theLeftOvers)
	{
		register Float32	theFloat32Value1 = *(inInputBuffer + 0);
		register Float32	theFloat32Value2 = *(inInputBuffer + 1);
		register Float32	theFloat32Value3 = *(inInputBuffer + 2);
		register Float32	theFloat32Value4 = *(inInputBuffer + 3);

		inInputBuffer += 4;

		theFloat32Value1 = ClipFloat32ForSInt24(theFloat32Value1);
		theFloat32Value2 = ClipFloat32ForSInt24(theFloat32Value2);
		theFloat32Value3 = ClipFloat32ForSInt24(theFloat32Value3);
		theFloat32Value4 = ClipFloat32ForSInt24(theFloat32Value4);

		// Multiply by kFloat32ToSInt32 instead of kFloat32toSInt24 to make the binary operations below work properly.
		register UInt32		a = (UInt32)(SInt32)(theFloat32Value1 * kFloat32ToSInt32);
		register UInt32		b = (UInt32)(SInt32)(theFloat32Value2 * kFloat32ToSInt32);
		register UInt32		c = (UInt32)(SInt32)(theFloat32Value3 * kFloat32ToSInt32);
		register UInt32		d = (UInt32)(SInt32)(theFloat32Value4 * kFloat32ToSInt32);

		//						a    b    c    d					a    b    c    d
		//	IN REGISTER:		123X 456X 789X ABCX					abc0 def0 ghi0 jkl0
		//	OUT REGISTERS:		6123 8945 ABC7						fabc hide jklg
		//	OUT MEMORY:			3216 5498 7CBA

		register SInt32		theOutputValue1 = (
													((b << 16) & 0xFF000000)
											  |
													((a >>  8) & 0x00FFFFFF)
											  );

		register SInt32		theOutputValue2 = (
													((c <<  8) & 0xFFFF0000)
											  |
													((b >> 16) & 0x0000FFFF)
											  );

		register SInt32		theOutputValue3 = (
													( d        & 0xFFFFFF00)
											  |
													((c >> 24) & 0x000000FF)
											  );

		//	store everything back to memory
		*(outOutputBuffer + 0) = theOutputValue1;
		*(outOutputBuffer + 1) = theOutputValue2;
		*(outOutputBuffer + 2) = theOutputValue3;

		outOutputBuffer += 3;
		inNumberSamples -= 4;
	}

	SInt8		*theOutputBuffer = (SInt8 *)outOutputBuffer;

	while (inNumberSamples > 0)
	{
		register Float32	theFloat32Value = *inInputBuffer;

		++inInputBuffer;

		theFloat32Value = ClipFloat32ForSInt24(theFloat32Value);

		// Multiply by kFloat32ToSInt32 instead of kFloat32toSInt24 to make the binary operations below work properly.
		register SInt32		theSInt32Value = (SInt32)(theFloat32Value * kFloat32ToSInt32);

		// Byte swapping will be handled automatically by the CPU if necessary.
		*(theOutputBuffer + 0) = (SInt8)((((UInt32)theSInt32Value) >>  8) & 0x000000FF);
		*(theOutputBuffer + 1) = (SInt8)((((UInt32)theSInt32Value) >> 16) & 0x000000FF);
		*(theOutputBuffer + 2) = (SInt8)((((UInt32)theSInt32Value) >> 24) & 0x000000FF);

		theOutputBuffer += 3;
		--inNumberSamples;
	}
}

/****************************************************************************/
/**
 * @brief	32bit 浮動小数点数を符号付き 32bit 整数型に変換．
 *
 * @param[in]		*inInputBuffer		入力データバッファ（32bit 浮動小数点型）．
 * @param[out]		*outOutputBuffer	出力データバッファ（符号付き 32bit 整数型）．
 * @param[in]		inNumberSamples		変換するサンプル数．
 */
// Float32 -> SInt32
static void
ClipFloat32ToSInt32LE_4(
	const Float32	*inInputBuffer,
	SInt32			*outOutputBuffer,
	UInt32			inNumberSamples
)
{
	register UInt32		theLeftOvers = inNumberSamples % 4;

	while (inNumberSamples > theLeftOvers)
	{
		register Float64	theFloat32Value1 = *(inInputBuffer + 0);	// <rdar://7138492>
		register Float64	theFloat32Value2 = *(inInputBuffer + 1);	// <rdar://7138492>
		register Float64	theFloat32Value3 = *(inInputBuffer + 2);	// <rdar://7138492>
		register Float64	theFloat32Value4 = *(inInputBuffer + 3);	// <rdar://7138492>

		inInputBuffer += 4;

		theFloat32Value1 = ClipFloat32ForSInt32(theFloat32Value1);
		theFloat32Value2 = ClipFloat32ForSInt32(theFloat32Value2);
		theFloat32Value3 = ClipFloat32ForSInt32(theFloat32Value3);
		theFloat32Value4 = ClipFloat32ForSInt32(theFloat32Value4);

		*(outOutputBuffer + 0) = SInt32NativeToLittleEndian((SInt32)(theFloat32Value1 * kFloat32ToSInt32));
		*(outOutputBuffer + 1) = SInt32NativeToLittleEndian((SInt32)(theFloat32Value2 * kFloat32ToSInt32));
		*(outOutputBuffer + 2) = SInt32NativeToLittleEndian((SInt32)(theFloat32Value3 * kFloat32ToSInt32));
		*(outOutputBuffer + 3) = SInt32NativeToLittleEndian((SInt32)(theFloat32Value4 * kFloat32ToSInt32));

		outOutputBuffer += 4;
		inNumberSamples -= 4;
	}

	while (inNumberSamples > 0)
	{
		register Float64	theFloat32Value = *inInputBuffer;	// <rdar://7138492>

		++inInputBuffer;

		theFloat32Value = ClipFloat32ForSInt32(theFloat32Value);

		*outOutputBuffer = SInt32NativeToLittleEndian((SInt32)(theFloat32Value * kFloat32ToSInt32));

		++outOutputBuffer;
		--inNumberSamples;
	}
}

#pragma mark - Public Functions -

/****************************************************************************/
/**
 * @brief	32bit 浮動小数点数を整数型に変換．
 *			オーディオ出力時に必要な処理．
 *
 * @param[in]		*mixBuf				入力データバッファ（32bit 浮動小数点型）．
 * @param[out]		*sampleBuf			出力データバッファ（符号付き整数型）．
 * @param[in]		firstSampleFrame	変換開始位置（入力バッファ先頭からのサンプルフレーム数）．
 * @param[in]		numSampleFrames		変換するサンプルフレーム数．
 * @param[in]		*streamFormat		ストリームのフォーマット．
 *
 * @return			IOReturn コード．
 */
IOReturn
clipPdjUsbAudioToOutputStream(
	const void					*mixBuf,
	void						*sampleBuf,
	UInt32						firstSampleFrame,
	UInt32						numSampleFrames,
	const IOAudioStreamFormat	*streamFormat
)
{
	if (!streamFormat)
	{
		return kIOReturnBadArgument;
	}

	UInt32		theNumberSamples = numSampleFrames * streamFormat->fNumChannels;
	UInt32		theFirstSample	 = firstSampleFrame * streamFormat->fNumChannels;
	Float32		*theMixBuffer	 = ((Float32 *)mixBuf) + theFirstSample;

	// aml, added optimized routines [3034710]
	switch (streamFormat->fBitWidth)
	{
		case 8:
			{
				SInt8	*theOutputBufferSInt8 = ((SInt8 *)sampleBuf) + theFirstSample;

				ClipFloat32ToSInt8_4(theMixBuffer, theOutputBufferSInt8, theNumberSamples);
			}
			break;

		case 16:
			{
				SInt16	*theOutputBufferSInt16 = ((SInt16 *)sampleBuf) + theFirstSample;

				ClipFloat32ToSInt16LE_4(theMixBuffer, theOutputBufferSInt16, theNumberSamples);
			}
			break;

		case 20:
		case 24:
			{
				SInt32	*theOutputBufferSInt24 = (SInt32 *)(((UInt8 *)sampleBuf) + (theFirstSample * 3));

				ClipFloat32ToSInt24LE_4(theMixBuffer, theOutputBufferSInt24, theNumberSamples);
			}
			break;

		case 32:
			{
				SInt32	*theOutputBufferSInt32 = ((SInt32 *)sampleBuf) + theFirstSample;

				ClipFloat32ToSInt32LE_4(theMixBuffer, theOutputBufferSInt32, theNumberSamples);
			}
			break;
	};

	return kIOReturnSuccess;
}

/****************************************************************************/
/**
 * @brief	整数を 32bit 浮動小数点型に変換．
 *			オーディオ入力時に必要な処理．
 *
 * @param[in]		*sampleBuf			入力データバッファ（符号付き整数型）．
 * @param[out]		*destBuf			出力データバッファ（32bit 浮動小数点型）．
 * @param[in]		firstSampleFrame	変換開始位置（入力バッファ先頭からのサンプルフレーム数）．
 * @param[in]		numSampleFrames		変換するサンプルフレーム数．
 * @param[in]		*streamFormat		ストリームのフォーマット．
 *
 * @return			IOReturn コード．
 */
IOReturn
convertFromPdjUsbAudioInputStream_NoWrap(
	const void					*sampleBuf,
	void						*destBuf,
	UInt32						firstSampleFrame,
	UInt32						numSampleFrames,
	const IOAudioStreamFormat	*streamFormat
)
{
	UInt32			numSamplesLeft;
	float			*floatDestBuf;

	floatDestBuf = (float *)destBuf;

	numSamplesLeft = numSampleFrames * streamFormat->fNumChannels;

	// DP(DBG_INFO, "destBuf = %p, firstSampleFrame = %u, numSampleFrames = %u", destBuf, firstSampleFrame, numSampleFrames);

	switch (streamFormat->fBitWidth)
	{
		case 8:
			SInt8	*inputBuf8;

			inputBuf8 = &(((SInt8 *)sampleBuf)[firstSampleFrame * streamFormat->fNumChannels]);

			while ((numSamplesLeft--) > 0)
			{
				*(floatDestBuf++) = (float)(*(inputBuf8++)) * kOneOverMaxSInt8Value;
			}

			break;

		case 16:
			SInt16	*inputBuf16;

			inputBuf16 = &(((SInt16 *)sampleBuf)[firstSampleFrame * streamFormat->fNumChannels]);

			while ((numSamplesLeft--) > 0)
			{
				*(floatDestBuf++) = (float)(*(inputBuf16++)) * kOneOverMaxSInt16Value;
			}

			break;

		case 20:
		case 24:
			register SInt8		*inputBuf24;

			// Multiply by 3 because 20 and 24 bit samples are packed into only three bytes, so we have to index bytes, not shorts or longs
			inputBuf24 = &(((SInt8 *)sampleBuf)[firstSampleFrame * streamFormat->fNumChannels * 3]);

			register SInt32		inputSample;

			// [rdar://4311684] - Fixed 24-bit input convert routine. /thw
			while ((numSamplesLeft--) > 1)
			{
				inputSample = (* (UInt32 *)inputBuf24) & 0x00FFFFFF;

				// Sign extend if necessary
				if (inputSample > 0x7FFFFF)
				{
					inputSample |= 0xFF000000;
				}

				inputBuf24 += 3;

				*(floatDestBuf++) = (float)inputSample * kOneOverMaxSInt24Value;
			}

			// Convert last sample. The following line does the same work as above without going over the edge of the buffer.
			inputSample = SInt32(
								(UInt32(*(UInt16 *)inputBuf24) & 0x0000FFFF)
						  |
								(SInt32(*(inputBuf24 + 2)) << 16)
						  );

			*(floatDestBuf++) = (float)inputSample * kOneOverMaxSInt24Value;

			break;

		case 32:
			register SInt32		*inputBuf32;

			inputBuf32 = &(((SInt32 *)sampleBuf)[firstSampleFrame * streamFormat->fNumChannels]);

			while ((numSamplesLeft--) > 0)
			{
				*(floatDestBuf++) = (float)(*(inputBuf32++)) * kOneOverMaxSInt32Value;
			}

			break;
	}

	return kIOReturnSuccess;
}

} // extern "C"
