/*
 * Copyright (c) 1998-2009 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioStream.h
 *
 * @brief	IOAudioStream クラスの派生クラス．
 *			オーディオストリーミング用のリングバッファを表現する．
 *			PdjUsbAudioEngine::initHardware() においてインスタンス化される．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/
#ifndef CC_PDJ_USB_AUDIO_STREAM_H_FILE_
#define CC_PDJ_USB_AUDIO_STREAM_H_FILE_

/*------------------------------------------*/
/*				型定義						*/
/*------------------------------------------*/

// <rdar://problem/6954295>
typedef struct {
	UInt32	whole;
	UInt32	fraction;		// This fraction is stored x 1000 to preserve precision
} IOAudioSamplesPerFrame;

/*------------------------------------------*/
/*				定数						*/
/*------------------------------------------*/

#pragma mark - Constants -

#define kFixedPoint10_14ByteSize				3
#define kFixedPoint16_16ByteSize				4

#define kMinimumFrameOffset						6

#define RECORD_NUM_USB_FRAME_LISTS				128
#define RECORD_NUM_USB_FRAMES_PER_LIST			2
#define RECORD_NUM_USB_FRAME_LISTS_TO_QUEUE		64

#define PLAY_NUM_USB_FRAME_LISTS				4
#define PLAY_NUM_USB_FRAMES_PER_LIST			64
#define PLAY_NUM_USB_FRAME_LISTS_TO_QUEUE		2
#define PLAY_NUM_USB_FRAMES_PER_LIST_SYNC		32

// [rdar://5623096] Make note of the slowest polling interval in ms for feedback endpoints
#define kMaxFeedbackPollingInterval				512

#define kSampleFractionAccumulatorRollover		65536 * 1000	// <rdar://problem/6954295> Fractional part of mSamplesPerPacket stored x 1000

// <rdar://6411577> Overruns threshold in packets (about 2ms at 48kHz, close to the safety offset value)
#define kOverrunsThreshold						100

// kMinimumSyncRefreshInterval is the smallest log base 2 amount of time in ms PdjUsbAudio will honor for a sync endppoint
#define kMinimumSyncRefreshInterval				1

#define kMaxFilterSize							33				// <rdar://problem/7378275>
#define kFilterScale							1024			// <rdar://problem/7378275>

#ifdef TARGET_PDJ
 #define kMaxErrorCount							50
#endif

// kUSBInputRecoveryTimeFraction represents the denominator (with 1 as the numerator) of the fractional multiple of a USB frame we must wait
// before data read via isoc in is available for converting. This must be greater than zero.
#define kUSBInputRecoveryTimeFraction			4

// The following definitions are for lock delay units
enum
{
	kLockDelayUnitMilliseconds					= 1,
	kLockDelayUnitsDecodedPCMSamples			= 2
};

/*------------------------------------------*/
/*			PdjUsbAudioStream クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbAudioStream -

class PdjUsbAudioStream : public IOAudioStream
{
	OSDeclareDefaultStructors(PdjUsbAudioStream)

public:

	typedef IOAudioStream				super;

#pragma mark - Public Methods -

	bool
	initWithAudioEngine(
		PdjUsbAudioDevice					*device,
		PdjUsbAudioEngine					*engine,
#ifdef TARGET_PDJ
 		UInt16								interfaceNumber,
#else
		IOUSBHostInterface					*streamInterface,
#endif
		IOAudioSampleRate					sampleRate,
		const char							*streamDescription = NULL,
		OSDictionary						*properties = NULL
	);

	void
	free(
		void
	);

	void							// <rdar://7295322>
	registerService(
		IOOptionBits						options = 0
	);

	bool
	matchPropertyTable(
		OSDictionary						*table,
		SInt32								*score
	);

	void
	stop(
		IOService							*provider
	);

	bool
	requestTerminate(
		IOService							*provider,
		IOOptionBits						options
	);

	bool
	willTerminate(
		IOService							*provider,
		IOOptionBits						options
	);

	bool
	terminate(
		IOOptionBits						options = 0
	);

	bool
	openStreamInterface(
		void
	);

	void
	closeStreamInterface(
		void
	);

	bool
	configureAudioStream(
		IOAudioSampleRate					sampleRate
	);

	void
	updateSampleOffsetAndLatency(
		void
	);

	IOReturn
	addAvailableFormats(
		PdjUsbConfigurationDictionary		*configDictionary,
		IOAudioSampleRate					sampleRate
	);

	IOReturn						// <rdar://7259238>
	setFormat(
		const IOAudioStreamFormat			*streamFormat,
		bool								callDriver = true
	);

	IOReturn						// <rdar://7259238>
	setFormat(
		const IOAudioStreamFormat			*streamFormat,
		const IOAudioStreamFormatExtension	*formatExtension,
		OSDictionary						*formatDict,
		bool								callDriver = true
	);

	IOReturn
	controlledFormatChange(
		const IOAudioStreamFormat			*newFormat,
		const IOAudioSampleRate				*newSampleRate
	);

	IOReturn
	prepareUSBStream(
		void
	);

	IOReturn
	startUSBStream(
		UInt64								currentUSBFrame,
		UInt32								usbFramesToDelay
	);

	IOReturn
	stopUSBStream(
		void
	);

	void
	queueOutputFrames(
		void
	);

	void
	queueInputFrames(
		void
	);

	IOReturn
	coalesceInputSamples(
		UInt32								numBytesToCoalesce,
		IOUSBHostIsochronousFrame			*pFrames
	);

	void
	takeTimeStamp(
		bool								incrementLoopCount = true,
		AbsoluteTime						*timestamp = NULL
	);

#ifdef TARGET_PDJ
 	void
 	logTimeStamp(
 		bool								incrementLoopCount = true,
 		AbsoluteTime						*timestamp = NULL
 	);
#endif

	UInt32
	getCurrentSampleFrame(
		void
	);

#if DEBUGLATENCY
	UInt64
	getQueuedFrameForSample(
		UInt32								sampleFrame
	);
#endif

	UInt8
	getSyncType(
		void
	);

	UInt32
	getLockDelayFrames(
		void
	);

	UInt16
	getAlternateFrameSize(
		void
	)
	{
		return mAlternateFrameSize;
	}

	void
	setMasterStreamMode(
		bool								masterMode
	)
	{
		mMasterMode = masterMode;
	}

	void
	compensateForSynchronization(
		bool								syncCompensation
	)
	{
		mSyncCompensation = syncCompensation;
	}

	void
	registerPlugin(
		PdjUsbAudioPlugin					*thePlugin
	);

	void
	pluginSetConfigurationApp(
		const char							*bundleID
	);

	IOReturn
	pluginDeviceRequest(
		StandardUSB::DeviceRequest			*request,
		IOMemoryDescriptor					*dataBuffer,
		IOUSBHostCompletion					*completion
	);

protected:

#pragma mark - Protected Methods -

	static void
	pluginLoaded(
		PdjUsbAudioStream					*usbAudioStreamObject
	);

	IOReturn						// added for rdar://3866513
	getDefaultSettings(
		UInt8								*altSettingID,
		IOAudioSampleRate					*sampleRate
	);

	IOReturn
	checkForFeedbackEndpoint(
		PdjUsbConfigurationDictionary		*configDictionary
	);

	IOReturn
	setSampleRateControl(
		UInt8								address,
		UInt32								sampleRate
	);

	// for IODMACommand
	IOBufferMemoryDescriptor *
	allocateBufferDescriptor(
		IOOptionBits						options,
		vm_size_t							capacity,
		vm_offset_t							alignment = 1
	);

	void							// <rdar://7568547>
	initializeUSBFrameList(
		IOUSBHostIsochronousFrame			*usbIsocFrames,
		UInt32								numFrames
	);

	IOReturn
	prepareWriteFrameList(
		UInt32								frameListNum
	);

	IOReturn
	writeFrameList(
		UInt32								frameListNum
	);

	IOReturn
	readFrameList(
		UInt32								frameListNum
	);

#if PRIMEISOCINPUT
	void
	primeInputPipe(
		IOUSBHostPipe						*pipeToPrime,
		UInt32								bytesPerUSBFrame,
		UInt32								usbFramesToDelay
	);
#endif

	static void
	writeHandler(
		void								*object,
		void								*parameter,
		IOReturn							result,
		IOUSBHostIsochronousFrame			*pFrames
	);

#ifdef TARGET_PDJ
#else
	static void
	writeHandlerForUHCI(
		void								*object,
		void								*parameter,
		IOReturn							result,
		IOUSBHostIsochronousFrame			*pFrames
	);
#endif

	static void
	readHandler(
		void								*object,
		void								*parameter,
		IOReturn							result,
		IOUSBHostIsochronousFrame			*pFrames
	);

#if PRIMEISOCINPUT
	static void
	primeInputPipeHandler(
		void								*object,
		void								*parameter,
		IOReturn							result,
		IOUSBHostIsochronousFrame			*pFrames
	);
#endif

	static void
	sampleRateHandler(
		void								*target,
		void								*parameter,
		IOReturn							result,
		IOUSBHostIsochronousFrame			*pFrames
	);

	AbsoluteTime
	generateTimeStamp(
		SInt32								usbFrameIndex,
		UInt32								preWrapBytes,
		UInt32								byteCount
	);

	UInt64							// <rdar://problem/7378275>
	jitterFilter(
		UInt64								curr,
		UInt32								nIter
	);

	IOReturn						// <rdar://problem/7378275>
	copyAnchor(
		UInt64								anchorFrame,
		AbsoluteTime						*anchorTime,
		UInt64								*usbCycleTime
	);

	void
	calculateSamplesPerPacket(
		UInt32								sampleRate,
		UInt16								*averageFrameSize,
		UInt16								*additionalSampleFrameFreq
	);

	UInt32							// <rdar://problem/6954295>
	getRateFromSamplesPerPacket(
		IOAudioSamplesPerFrame				samplesPerPacket
	);

#ifdef TARGET_PDJ
 	IOReturn
 	processOutputSamples(
 		IOAudioClientBuffer					*clientBuffer,
 		UInt32								firstSampleFrame,
 		UInt32								loopCount,
 		bool								samplesAvailable
 	);
 
 	void
 	clipIfNecessary(
 		void
 	);
#endif

public:

	PdjUsbAudioPlugin					*mPlugin;
	IOUSBHostInterface					*mStreamInterface;
	UInt8								mInterfaceNumber;
	UInt8								mAlternateSettingID;
	OSArray								*mActiveClockPath;
	IOAudioStreamFormat					mDefaultAudioStreamFormat;
	UInt16								mSampleSize;
	UInt32								mSampleBufferSize;
#ifdef TARGET_PDJ
 	UInt64								mNumSamplesInBuffer;
#endif
	UInt32								mBufferOffset;
	UInt32								mNumUSBFrameListsToQueue;
	UInt32								mReadUSBFrameListSize;
	IORecursiveLock						*mCoalescenceMutex;						// <rdar://problem/7378275>

protected:

	thread_call_t						mPluginInitThread;

	PdjUsbAudioDevice					*mUSBAudioDevice;
	PdjUsbAudioEngine					*mUSBAudioEngine;
	IOUSBHostPipe						*mPipe;
	IOUSBHostPipe						*mAssociatedPipe;

	UInt8								mDirection;

	UInt8								mTransactionsPerUSBFrame;
	UInt8								mRefreshInterval;
	UInt8								mFeedbackPacketSize;

	IOUSBHostIsochronousCompletion		*mUSBCompletion;
	IOUSBHostIsochronousFrame			*mUSBIsocFrames;

#if PRIMEISOCINPUT
	IOUSBHostIsochronousCompletion		mPrimeInputCompletion;
	IOUSBHostIsochronousFrame			*mPrimeInputIsocFrames;
#endif

	IOUSBHostIsochronousCompletion		mExtraUSBCompletion;

	IOUSBHostIsochronousCompletion		mSampleRateCompletion;
	IOUSBHostIsochronousFrame			mSampleRateFrame;

	IOBufferMemoryDescriptor			*mUSBBufferDescriptor;
	void								*mReadBuffer;

	IOBufferMemoryDescriptor			*mSampleBufferMemoryDescriptor;
	IOSubMemoryDescriptor				**mSampleBufferDescriptors;

	IOMultiMemoryDescriptor				*mWrapRangeDescriptor;
	IOSubMemoryDescriptor				*mWrapDescriptors[2];

	IOBufferMemoryDescriptor			*mAssociatedEndpointMemoryDescriptor;	// <rdar://7000283>
	UInt32								*mAverageSampleRateBuffer;				// needs to be 4 bytes for a 10.14 or 16.16 value

	UInt32								mSafeErasePoint;
	UInt32								mNextSafeErasePoint;
	UInt32								mLastPreparedBufferOffset;

	UInt64								*mFrameQueuedForList;
	UInt32								mNumUSBFrameLists;
	volatile UInt32						mCurrentFrameList;
	UInt32								mNumUSBFramesPerList;
	UInt32								mFractionalSamplesLeft;
	UInt16								mNumFramesInFirstList;
	UInt32								mNumTransactionsPerList;
	UInt64								mUSBFrameToQueue;
	UInt64								mNextSyncReadFrame;

#if DEBUGLATENCY
	UInt32								mLastFrameListSize;
	UInt32								mThisFrameListSize;
#endif

	UInt16								mReadUSBFrameSize;
	UInt16								mAverageFrameSize;						// These sizes are in bytes.
	UInt16								mAlternateFrameSize;

#ifdef TARGET_PDJ
#else
	bool								mUHCISupport;
	UInt32								mSampleBufferSizeExtended;
#endif

	IOAudioSampleRate					mCurSampleRate;
	IOAudioSamplesPerFrame				mSamplesPerPacket;						// store this as a 16.16 value <rdar://problem/6954295>

	// <rdar://problem/7378275>
	bool								mInitStampDifference;
	UInt32								mNumTimeStamp;
	UInt64								mFilterData[kMaxFilterSize];
	UInt32								mFilterWritePointer;
	UInt64								mLastRawTimeStamp_nanos;
	UInt64								mLastFilteredTimeStamp_nanos;
	UInt64								mLastWrapFrame;
	UInt64								mLastFilteredStampDifference;
#if DEBUGTIMESTAMPS
	SInt64								mStampDrift;
#endif

	volatile UInt32						mShouldStop;
	bool								mTerminatingDriver;
	bool								mUSBStreamRunning;
	bool								mMasterMode;
	bool								mSyncCompensation;
	bool								mSplitTransactions;
	bool								mHaveTakenFirstTimeStamp;
	bool								mNeedTimeStamps;
	bool								mInCompletion;

	// [rdar://5417631] Keep track of devices that generate overruns on input
	bool								mGeneratesOverruns;
	UInt32								mOverrunsCount;							// <rdar://6902105>
	UInt32								mOverrunsThreshold;						// <rdar://6411577>

#ifdef TARGET_PDJ
 	UInt32								mNumFramesInClientBuffer;
 
 	UInt32								mErrorCount;
 	bool								mShouldRestartStreaming;
 
 	UInt64								mLastTimeStamp_nanos;
#endif
};

#endif /* CC_PDJ_USB_AUDIO_STREAM_H_FILE_ */
