/*
 * Copyright (c) 1998-2009 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioCommon.h
 *
 * @brief	ドライバ共通ヘッダファイル．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/
#ifndef PDJ_USB_AUDIO_COMMON_H_FILE_
#define PDJ_USB_AUDIO_COMMON_H_FILE_

// TRUE・FALSE の定義
#include <mach/boolean.h>

/*------------------------------------------*/
/*			ターゲットデバイス指定			*/
/*------------------------------------------*/

// ターゲット固有の処理
#define TARGET_PDJ

// プロバイダは IOUSBHostDevice（未定義ならプロバイダは IOUSBHostInterface）
//#define PROVIDER_USB_DEVICE

// ベースにしている AppleUSBAudio 自体のバグ
#define BUG_FIX

/*------------------------------------------*/
/*				デバッグ					*/
/*------------------------------------------*/

#define DEBUGLOGGING


#ifdef DEBUGLOGGING
 #define DP(level, format, args...)		\
 {										\
 	if ((level) & DBG_LEVEL)			\
 	{									\
 		IOLog(format "\n", ##args);		\
 	/*	IOSleep(100);	*/				\
 	}									\
 }
#else
 #define DP(level, format, args...)
#endif


#define DBG_ALL			(0xFFFFFFFF)
#define DBG_ERROR		(0x00000001)
#define DBG_WARNING		(0x00000002)
#define DBG_INFO		(0x00000004)
#define DBG_TRACE		(0x00000008)
#define DBG_READ		(0x00000010)
#define DBG_WRITE		(0x00000020)
#define DBG_TIMESTAMP	(0x00000040)
#define DBG_TIMER		(0x00000080)
#define DBG_FEEDBACK	(0x00000100)
#define DBG_UHCI		(0x00000200)
#define DBG_VERBOSE		(0x80000000)
#define DBG_LEVEL		(DBG_ERROR | DBG_WARNING)


#define SoundAssertionMessage(cond, file, line, handler)										\
	"Sound assertion \"" #cond "\" failed in " #file " at line " #line " goto " #handler ""

#define SoundAssertionFailed(cond, file, line, handler)											\
	DP(DBG_ERROR, SoundAssertionMessage(cond, file, line, handler))

#define FailIf(cond, handler)										\
{																	\
	if (cond)														\
	{																\
		SoundAssertionFailed(cond, __FILE__, __LINE__, handler);	\
		goto handler;												\
	}																\
}

#define FailWithAction(cond, action, handler)						\
{																	\
	if (cond)														\
	{																\
		SoundAssertionFailed(cond, __FILE__, __LINE__, handler);	\
		{ action; }													\
		goto handler;												\
	}																\
}

#define FailMessage(cond)											\
{																	\
	if (cond)														\
	{																\
		SoundAssertionFailed(cond, __FILE__, __LINE__, handler);	\
	}																\
}

// Following are a list of precompiler variables that affect the way that PdjUsbAudio executes, logs, and compiles.

// LOCALIZABLE
#define LOCALIZABLE					FALSE

// PRIMEISOCINPUT queues kNumIsocFramesToPrime USB frames to be read and disregarded before attempting to stream samples to CoreAudio
#define PRIMEISOCINPUT				TRUE

#if PRIMEISOCINPUT
 #define kNumUSBFramesToPrime		12
#else
 #define kNumUSBFramesToPrime		0
#endif

// <rdar://5811247> Poll for the clock status.
#define POLLCLOCKSTATUS				TRUE

// RESETAFTERSLEEP causes a device reset to be issued after waking from sleep for all devices.
#define RESETAFTERSLEEP				TRUE


#ifdef DEBUGLOGGING

 // DEBUGLATENCY enables methods that allow the tracking of how long clipped samples stay in the output buffer prior to DMA
 #define DEBUGLATENCY				FALSE

 // DEBUGANCHORS prints out the last kAnchorsToAccumulate anchors whenever the list fills; used to check anchor accuracy.
 #define DEBUGANCHORS				FALSE
 #if DEBUGANCHORS
  #define kAnchorsToAccumulate		10
 #endif

 // DEBUGTIMESTAMPS
 #define DEBUGTIMESTAMPS			FALSE

 // LOGTIMESTAMPS prints the timestamp in nanoseconds whenever takeTimeStamp it is called
 #define LOGTIMESTAMPS				FALSE

 // DEBUGINPUT
 #define DEBUGINPUT					FALSE

 // [rdar://5600254] Log the data cadence for debug purposes.
 #define SHOWCADENCE				FALSE

 // STAGGERINTERFACES delays even-numbered streaming interfaces' initHardware by a fixed value (for readability)
 #define STAGGERINTERFACES			FALSE

#else

 #define DEBUGLATENCY				FALSE
 #define DEBUGANCHORS				FALSE
 #define DEBUGTIMESTAMPS			FALSE
 #define LOGTIMESTAMPS				FALSE
 #define DEBUGINPUT					FALSE
 #define SHOWCADENCE				FALSE
 #define STAGGERINTERFACES			FALSE

#endif

/*------------------------------------------*/
/*				インクルード				*/
/*------------------------------------------*/

#include <AvailabilityMacros.h>

#ifdef __LP64__
 #include <IOKit/IOSubMemoryDescriptor.h>
#endif

#include <IOKit/IOMultiMemoryDescriptor.h>
#include <IOKit/IOKitKeys.h>
#include <IOKit/IOTimerEventSource.h>
#include <IOKit/pwr_mgt/RootDomain.h>
#include <IOKit/usb/USB.h>
#include <IOKit/usb/IOUSBHostInterface.h>
#include <IOKit/audio/IOAudioDefines.h>
#include <IOKit/audio/IOAudioTypes.h>
#include <IOKit/audio/IOAudioDevice.h>
#include <IOKit/audio/IOAudioLevelControl.h>
#include <IOKit/audio/IOAudioToggleControl.h>
#include <IOKit/audio/IOAudioSelectorControl.h>
#include <IOKit/audio/IOAudioEngineUserClient.h>

#define PdjUsbAudioDevice					com_pioneerdj_DRIVERCLASSPREFIX_UsbAudioDevice
#define LOGSTRAUDIODEVICE					"DJM-V10AudioDevice"
class	PdjUsbAudioDevice;

#define PdjUsbAudioTopology					com_pioneerdj_DRIVERCLASSPREFIX_UsbAudioTopology
#define LOGSTRAUDIOTOPOLOGY					"DJM-V10AudioTopology"
class	PdjUsbAudioTopology;

#define PdjUsbAudioEngine					com_pioneerdj_DRIVERCLASSPREFIX_UsbAudioEngine
#define LOGSTRAUDIOENGINE					"DJM-V10AudioEngine"
class	PdjUsbAudioEngine;

#define PdjUsbAudioStream					com_pioneerdj_DRIVERCLASSPREFIX_UsbAudioStream
#define LOGSTRAUDIOSTREAM					"DJM-V10AudioStream"
class	PdjUsbAudioStream;

#define PdjUsbAudioPlugin					com_pioneerdj_DRIVERCLASSPREFIX_UsbAudioPlugin
#define LOGSTRAUDIOPLUGIN					"DJM-V10AudioPlugin"
class	PdjUsbAudioPlugin;

#define PdjUsbAudioDictionary				com_pioneerdj_DRIVERCLASSPREFIX_UsbAudioDictionary
#define LOGSTRAUDIODICTIONARY				"DJM-V10AudioDictionary"
class	PdjUsbAudioDictionary;

#define PdjUsbConfigurationDictionary		com_pioneerdj_DRIVERCLASSPREFIX_ConfigurationDictionary
#define LOGSTRCONFIGURATIONDICTIONARY		"DJM-V10ConfigurationDictionary"
class	PdjUsbConfigurationDictionary;

#define PdjUsbControlDictionary				com_pioneerdj_DRIVERCLASSPREFIX_ControlDictionary
#define LOGSTRCONTROLDICTIONARY				"DJM-V10ControlDictionary"
class	PdjUsbControlDictionary;

#define PdjUsbStreamDictionary				com_pioneerdj_DRIVERCLASSPREFIX_StreamDictionary
#define LOGSTRSTREAMDICTIONARY				"DJM-V10StreamDictionary"
class	PdjUsbStreamDictionary;

#define PdjUsbEndpointDictionary			com_pioneerdj_DRIVERCLASSPREFIX_EndpointDictionary
#define LOGSTRENDPOINTDICTIONARY			"DJM-V10EndpointDictionary"
class	PdjUsbEndpointDictionary;

#define PdjUsbASEndpointDictionary			com_pioneerdj_DRIVERCLASSPREFIX_ASEndpointDictionary
#define LOGSTRASENDPOINTDICTIONARY			"DJM-V10ASEndpointDictionary"
class	PdjUsbASEndpointDictionary;

#define PdjUsbUnitDictionary				com_pioneerdj_DRIVERCLASSPREFIX_UnitDictionary
#define LOGSTRUNITDICTIONARY				"DJM-V10UnitDictionary"
class	PdjUsbUnitDictionary;

#define PdjUsbInputTerminalDictionary		com_pioneerdj_DRIVERCLASSPREFIX_InputTerminalDictionary
#define LOGSTRINPUTTERMINALDICTIONARY		"DJM-V10InputTerminalDictionary"
class	PdjUsbInputTerminalDictionary;

#define PdjUsbOutputTerminalDictionary		com_pioneerdj_DRIVERCLASSPREFIX_OutputTerminalDictionary
#define LOGSTROUTPUTTERMINALDICTIONARY		"DJM-V10OutputTerminalDictionary"
class	PdjUsbOutputTerminalDictionary;

#define PdjUsbMixerUnitDictionary			com_pioneerdj_DRIVERCLASSPREFIX_MixerUnitDictionary
#define LOGSTRMIXERUNITDICTIONARY			"DJM-V10MixerUnitDictionary"
class	PdjUsbMixerUnitDictionary;

#define PdjUsbSelectorUnitDictionary		com_pioneerdj_DRIVERCLASSPREFIX_SelectorUnitDictionary
#define LOGSTRSELECTORUNITDICTIONARY		"DJM-V10SelectorUnitDictionary"
class	PdjUsbSelectorUnitDictionary;

#define PdjUsbEffectUnitDictionary			com_pioneerdj_DRIVERCLASSPREFIX_EffectUnitDictionary
#define LOGSTREFFECTUNITDICTIONARY			"DJM-V10EffectUnitDictionary"
class	PdjUsbEffectUnitDictionary;

#define PdjUsbProcessingUnitDictionary		com_pioneerdj_DRIVERCLASSPREFIX_ProcessingUnitDictionary
#define LOGSTRPROCESSINGUNITDICTIONARY		"DJM-V10ProcessingUnitDictionary"
class	PdjUsbProcessingUnitDictionary;

#define PdjUsbExtensionUnitDictionary		com_pioneerdj_DRIVERCLASSPREFIX_ExtensionUnitDictionary
#define LOGSTREXTENSIONUNITDICTIONARY		"DJM-V10ExtensionUnitDictionary"
class	PdjUsbExtensionUnitDictionary;

#define PdjUsbFeatureUnitDictionary			com_pioneerdj_DRIVERCLASSPREFIX_FeatureUnitDictionary
#define LOGSTRFEATUREUNITDICTIONARY			"DJM-V10FeatureUnitDictionary"
class	PdjUsbFeatureUnitDictionary;

#define PdjUsbClockSourceDictionary			com_pioneerdj_DRIVERCLASSPREFIX_ClockSourceDictionary
#define LOGSTRCLOCKSOURCEDICTIONARY			"DJM-V10ClockSourceDictionary"
class	PdjUsbClockSourceDictionary;

#define PdjUsbClockSelectorDictionary		com_pioneerdj_DRIVERCLASSPREFIX_ClockSelectorDictionary
#define LOGSTRCLOCKSELECTORDICTIONARY		"DJM-V10ClockSelectorDictionary"
class	PdjUsbClockSelectorDictionary;

#define PdjUsbClockMultiplierDictionary		com_pioneerdj_DRIVERCLASSPREFIX_ClockMultiplierDictionary
#define LOGSTRCLOCKMULTIPLIERDICTIONARY		"DJM-V10ClockMultiplierDictionary"
class	PdjUsbClockMultiplierDictionary;

#include "PdjUsbAudioBigNum.h"
#include "PdjUsbAudioClip.h"
#include "PdjUsbAudioDictionary.h"
#include "PdjUsbAudioDevice.h"
#include "PdjUsbAudioTopology.h"
#include "PdjUsbAudioEngine.h"
#include "PdjUsbAudioStream.h"
#include "PdjUsbAudioPlugin.h"
#include "CC_PdjUsbAudioClip.h"
#include "CC_PdjUsbAudioStream.h"

/*------------------------------------------*/
/*				マクロ						*/
/*------------------------------------------*/

// Sizes in bits for OSNumbers
#define BITSTOBYTES			8
#define SIZEINBITS(x)		(BITSTOBYTES * sizeof(x))


#ifndef ABSOLUTETIME_SCALAR_TYPE

 #define clock_get_uptime(abstime)												\
	clock_get_uptime(__OSAbsoluteTimePtr(abstime))

 #define absolutetime_to_nanoseconds(abstime, nanoseconds)						\
	absolutetime_to_nanoseconds(__OSAbsoluteTime(abstime), (nanoseconds))

 #define nanoseconds_to_absolutetime(nanoseconds, abstime)						\
	nanoseconds_to_absolutetime((nanoseconds), __OSAbsoluteTimePtr(abstime))

 #define thread_call_enter_delayed(call, deadline)								\
	thread_call_enter_delayed((call), __OSAbsoluteTime(deadline))

#endif

#ifdef KERNEL_PRIVATE

OS_INLINE uint64_t
__OSAbsoluteTime(
	AbsoluteTime	abstime
)
{
	return (*(uint64_t *)&abstime);
}

OS_INLINE
uint64_t *
__OSAbsoluteTimePtr(
	AbsoluteTime	*abstime
)
{
	return ((uint64_t *)abstime);
}

#define AbsoluteTime_to_scalar(x)	(*(uint64_t *)(x))

/* t1 < = > t2 */
#define CMP_ABSOLUTETIME(t1, t2)				\
	(AbsoluteTime_to_scalar(t1) >				\
		AbsoluteTime_to_scalar(t2)? (int)+1 :	\
	 (AbsoluteTime_to_scalar(t1) <				\
		AbsoluteTime_to_scalar(t2)? (int)-1 : 0))

/* t1 += t2 */
#define ADD_ABSOLUTETIME(t1, t2)				\
	(AbsoluteTime_to_scalar(t1) +=				\
				AbsoluteTime_to_scalar(t2))

/* t1 -= t2 */
#define SUB_ABSOLUTETIME(t1, t2)				\
	(AbsoluteTime_to_scalar(t1) -=				\
				AbsoluteTime_to_scalar(t2))

#define ADD_ABSOLUTETIME_TICKS(t1, ticks)		\
	(AbsoluteTime_to_scalar(t1) +=				\
						(int32_t)(ticks))

#endif	/* KERNEL_PRIVATE */

#endif /* PDJ_USB_AUDIO_COMMON_H_FILE_ */
