/*
 * Copyright (c) 1998-2009 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioDevice.cpp
 *
 * @brief	IOAudioDevice クラスの派生クラス．
 *			ドライバ全体のルートクラス．
 *			１個のデバイスの全体を表現する．
 *			AppleUSBAudioDevice クラスを２つに分離して
 *			PdjUsbAudioDevice クラスおよび PdjUsbAudioTopology クラスとした．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/

/*------------------------------------------*/
/*				インクルード				*/
/*------------------------------------------*/

#include "PdjUsbAudioCommon.h"

/*------------------------------------------*/
/*			PdjUsbAudioDevice クラス		*/
/*------------------------------------------*/

OSDefineMetaClassAndStructors(PdjUsbAudioDevice, super)

#pragma mark - Public Methods -

/****************************************************************************/
/**
 * @brief	参照カウントアップ．
 *			OSObject::retain() をオーバーライド．
 */
#ifdef DEBUGLOGGING
void
PdjUsbAudioDevice::retain(
	void
) const
{
	DP(DBG_VERBOSE, "? " LOGSTRAUDIODEVICE "(%p)::retain() - rc = %d", this, this->getRetainCount());
	super::retain();
}
#endif

/****************************************************************************/
/**
 * @brief	参照カウントダウン．
 *			OSObject::release() をオーバーライド．
 */
#ifdef DEBUGLOGGING
void
PdjUsbAudioDevice::release(
	void
) const
{
	DP(DBG_VERBOSE, "? " LOGSTRAUDIODEVICE "(%p)::release() - rc = %d", this, this->getRetainCount());
	super::release();
}
#endif

/****************************************************************************/
/**
 * @brief	インスタンス初期化．
 *			IOAudioDevice::init() をオーバーライド．
 *
 * @param[in]		*properties		デバイスプロパティ．
 *
 * @retval			true			初期化成功．
 * @retval			false			初期化失敗．
 */
bool
PdjUsbAudioDevice::init(
	OSDictionary	*properties
)
{
	DP(DBG_WARNING, "+ " LOGSTRAUDIODEVICE "[%p]::init()", this);

	bool	result = super::init(properties);

	DP(DBG_WARNING, "- " LOGSTRAUDIODEVICE "[%p]::init() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	インスタンス解放．
 *			IOAudioDevice::free() をオーバーライド．
 *			参照カウントがゼロになった時呼び出される．
 */
void
PdjUsbAudioDevice::free(
	void
)
{
	DP(DBG_WARNING, "+ " LOGSTRAUDIODEVICE "[%p]::free()", this);

	if (NULL != mUpdateTimer)
	{
		mUpdateTimer->cancelTimeout();
		mUpdateTimer->release();
		mUpdateTimer = NULL;
	}

	if (mStatusInterruptBuffer)
	{
		// <rdar://7000283>
		// Pointer was obtained from mInterruptEndpointMemoryDescriptor, so no need to free it explicitly here.
		mStatusInterruptBuffer = NULL;
	}

	if (mInterruptEndpointMemoryDescriptor)
	{
		mInterruptEndpointMemoryDescriptor->release();
		mInterruptEndpointMemoryDescriptor = NULL;
	}

	// <rdar://problem/7378275>
	if (mTimeLock)
	{
		IOLockFree (mTimeLock);
		mTimeLock = NULL;
	}

	if (mInterfaceLock)
	{
		IORecursiveLockFree(mInterfaceLock);
		mInterfaceLock = NULL;
	}

	// <rdar://problem/6369110>
	if (mRegisteredEnginesMutex)
	{
		IORecursiveLockFree(mRegisteredEnginesMutex);
		mRegisteredEnginesMutex = NULL;
	}

	// <rdar://problem/6420832>
	if (mRegisteredStreamsMutex)
	{
		IORecursiveLockFree(mRegisteredStreamsMutex);
		mRegisteredStreamsMutex = NULL;
	}

	if (mConfigDictionary)
	{
		mConfigDictionary->release();
		mConfigDictionary = NULL;
	}

	if (mRegisteredEngines)
	{
		mRegisteredEngines->release();
		mRegisteredEngines = NULL;
	}

	// <rdar://problem/6420832>
	if (mRegisteredStreams)
	{
		mRegisteredStreams->release();
		mRegisteredStreams = NULL;
	}

	if (NULL != mEngineArray)
	{
		for (UInt32 engineIndex = 0; engineIndex < mEngineArray->getCount(); engineIndex++)
		{
			PdjUsbAudioEngine	*engine = OSDynamicCast(PdjUsbAudioEngine, mEngineArray->getObject(engineIndex));
			if (NULL != engine)
			{
				engine->release();
			}
		}

		mEngineArray->release();
		mEngineArray = NULL;
	}

	if (mTopology)
	{
		mTopology->release();
		mTopology = NULL;
	}

	super::free();

	DP(DBG_WARNING, "- " LOGSTRAUDIODEVICE "[%p]::free()", this);
}

/****************************************************************************/
/**
 * @brief	サービスマッチング．
 *			IOService::matchPropertyTable() をオーバーライド．
 *			ドライバ固有のマッチング条件があるときはここで判定を行う．
 *
 * @param[in]		*table			マッチングの対象となるプロパティ．
 * @param[in, out]	*score			プローブスコア．必要に応じて書き換えることができる．
 *
 * @retval			true			マッチング成功．
 * @retval			false			マッチング失敗．
 */
// Implemented for rdar://3993906. Allows matching based on a custom dictionary.
bool
PdjUsbAudioDevice::matchPropertyTable(
	OSDictionary	*table,
	SInt32			*score
)
{
	DP(DBG_VERBOSE, "+ " LOGSTRAUDIODEVICE "[%p]::matchPropertyTable(%p, %p)", this, table, score);

	bool			returnValue = false;
	OSObject		*deviceName;

	deviceName = table->getObject(kIOAudioDeviceNameKey);

	if (deviceName)
	{
		// This custom dictionary wants the device to have a name.
		if (IORegistryEntry::getProperty(kIOAudioDeviceNameKey))
		{
			// No need to match name; it exists, and that's all that matters
			returnValue = true;
		}
		else
		{
			// Device doesn't have a name yet, so don't match
			returnValue = false;
		}
	}
	else
	{
		// This is our standard matchPropertyTable implementation
		returnValue = IOService::matchPropertyTable(table, score);
	}

	DP(DBG_VERBOSE, "- " LOGSTRAUDIODEVICE "[%p]::matchPropertyTable(%p, %p) = %d", this, table, score, returnValue);
	return returnValue;
}

/****************************************************************************/
/**
 * @brief	サービス開始．
 *			IOAudioDevice::start() をオーバーライド．
 *			super::start() 内でハードウェアの初期化とサービスの登録が行われる．
 *
 * @param[in]		*provider		ハードウェアへのアクセスを提供するサービスプロバイダ nub．
 *									PROVIDER_USB_DEVICE が定義されていれば IOUSBHostDevice，
 *									未定義なら IOUSBHostInterface．
 *
 * @retval			true			開始成功．
 * @retval			false			開始失敗．デバイスは認識されない．
 */
bool
PdjUsbAudioDevice::start(
	IOService		*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::start(%p)", this, provider);

	bool								result = false;

#ifdef PROVIDER_USB_DEVICE
	const StandardUSB::ConfigurationDescriptor	*desc;
	IOUSBFindInterfaceRequest					req;

	mUSBDevice = OSDynamicCast(IOUSBHostDevice, provider);
	FailIf(NULL == mUSBDevice, Exit);

	desc = mUSBDevice->getConfigurationDescriptor(0);
	FailIf(NULL == desc, Exit);

	FailIf(!mUSBDevice->open(this), Exit);

	FailIf(kIOReturnSuccess != mUSBDevice->setConfiguration(desc->bConfigurationValue), Exit);

	mUSBDevice->close(this);

	req.bInterfaceClass		= AUDIO;
	req.bInterfaceSubClass	= AUDIOCONTROL;

	req.bInterfaceProtocol	= kIOUSBFindInterfaceDontCare;
	req.bAlternateSetting	= kIOUSBFindInterfaceDontCare;

	mControlInterface = this->findInterface(&req);
	FailIf(NULL == mControlInterface, Exit);

	mControlInterface->retain();

#else // #ifdef PROVIDER_USB_DEVICE

	mControlInterface = OSDynamicCast(IOUSBHostInterface, provider);
	FailIf(NULL == mControlInterface, Exit);	// <rdar://7085810>
#endif

	if (!mControlInterface->open(this))
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "::start() - mControlInterface->open() failed! Re-enumerating...");

		IOUSBHostDevice		*usbDevice = mControlInterface->getDevice();
		FailIf(NULL == usbDevice, Exit);

		usbDevice->reset();

		goto Exit;
	}

	// <rdar://problem/6686515>
	mInitHardwareThread = ::thread_call_allocate(
								(thread_call_func_t)PdjUsbAudioDevice::initHardwareThread,
								(thread_call_param_t)this
						  );
	FailIf(NULL == mInitHardwareThread, Exit);

	result = super::start(provider);	// Causes our initHardware routine to be called.

Exit:
	// <rdar://problems/6576824&6674310> PdjUsbAudio fails to unload properly
	if (!result)
	{
		this->stop(provider);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::start(%p) = %d", this, provider, result);

	return result;
}

/****************************************************************************/
/**
 * @brief	サービス停止．
 *			IOAudioDevice::stop() をオーバーライド．
 *
 * @param[in]		*provider		ハードウェアへのアクセスを提供するサービスプロバイダ nub．
 *									PROVIDER_USB_DEVICE が定義されていれば IOUSBHostDevice，
 *									未定義なら IOUSBHostInterface．
 */
void
PdjUsbAudioDevice::stop(
	IOService		*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::stop(%p) - audioEngines = %p, rc = %d",
		this, provider, audioEngines, this->getRetainCount());

#ifdef TARGET_PDJ
 	if (mSleepWakeNotifier)
 	{
 		mSleepWakeNotifier->remove();
 
 		mSleepWakeNotifier = NULL;
 	}
#endif

	// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
	if (NULL != mInterruptPipe)
	{
		mInterruptPipe->abort();
		mInterruptPipe->release();
		mInterruptPipe = NULL;
	}

	if (mProcessStatusInterruptThread)
	{
		::thread_call_cancel(mProcessStatusInterruptThread);
		::thread_call_free(mProcessStatusInterruptThread);
		mProcessStatusInterruptThread = NULL;
	}

	if (mUpdateTimer)
	{
		// Stop the rate calculation timer
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::stop() - Cancelling time stamp rate timer ...", this);

		mUpdateTimer->cancelTimeout();
		mUpdateTimer->disable();

#ifdef BUG_FIX
		workLoop->removeEventSource(mUpdateTimer);
#endif
	}

	// <rdar://7295322>
	if (mRegisteredEngines)
	{
		mRegisteredEngines->release();
		mRegisteredEngines = NULL;
	}

	// <rdar://7295322>
	if (mRegisteredStreams)
	{
		mRegisteredStreams->release();
		mRegisteredStreams = NULL;
	}

	if (NULL != mEngineArray)
	{
		for (UInt32 engineIndex = 0; engineIndex < mEngineArray->getCount(); engineIndex++)
		{
			PdjUsbAudioEngine	*engine = OSDynamicCast(PdjUsbAudioEngine, mEngineArray->getObject(engineIndex));
			if (NULL != engine)
			{
				engine->release();
			}
		}

		mEngineArray->release();
		mEngineArray = NULL;
	}

	// <rdar://problem/6686515>
	if (mInitHardwareThread)
	{
		::thread_call_cancel(mInitHardwareThread);
		::thread_call_free(mInitHardwareThread);
		mInitHardwareThread = NULL;
	}

	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::stop() - mControlInterface now closing ...", this);
	if (mControlInterface)
	{
		mControlInterface->close(this);

#ifdef PROVIDER_USB_DEVICE
		mControlInterface->release();
#endif

		mControlInterface = NULL;
	}

	// call the IOAudioDevice generic stop routine
	super::stop(provider);	// <rdar://problem/7800198> Move this call to the end of the method, after closing interface

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::stop()", this);
}

/****************************************************************************/
/**
 * @brief	メッセージハンドラ．
 *			IOService::message() をオーバーライド．
 *			サービスプロバイダからメッセージを受け取った時に呼び出される．
 *
 * @param[in]		type			メッセージの種類．IOMessage.h で定義されている．
 * @param[in]		*provider		ハードウェアへのアクセスを提供するサービスプロバイダ nub．
 *									PROVIDER_USB_DEVICE が定義されていれば IOUSBHostDevice，
 *									未定義なら IOUSBHostInterface．
 * @param[in]		*arg			メッセージの引数．未使用．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDevice::message(
	UInt32			type,
	IOService		*provider,
	void			*arg
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::message(0x%x, %p) - rc = %d", this, type, provider, this->getRetainCount());

	switch (type)
	{
		case kIOMessageServiceIsTerminated:
		case kIOMessageServiceIsRequestingClose:
#ifdef PROVIDER_USB_DEVICE
			if ((mControlInterface != NULL) && (mUSBDevice == provider))
			{
				mControlInterface->close(this);
				mControlInterface->release();
				mControlInterface = NULL;
			}
#else
			if ((mControlInterface != NULL) && (mControlInterface == provider))
			{
				mControlInterface->close(this);
				mControlInterface = NULL;
			}
#endif
			break;

		case kIOUSBMessagePortHasBeenReset:
			// If the device has been reset, we must take steps to ensure that streaming can be resumed.
			// Flush controls to the device in case the device reset changed their states.
			DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::message() - Flushing controls to the device.", this);

			super::flushAudioControls();

			FailIf(NULL == mRegisteredEngines, Exit);

			DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::message() - Resetting engines.", this);

			for (UInt8 engineIndex = 0; engineIndex < mRegisteredEngines->getCount(); engineIndex++)
			{
				PdjUsbAudioEngine	*currentEngine = NULL;
				OSDictionary		*currentEngineInfo = NULL;

				currentEngineInfo = OSDynamicCast(OSDictionary, mRegisteredEngines->getObject(engineIndex));
				FailIf(NULL == currentEngineInfo, Exit);

				currentEngine = OSDynamicCast(PdjUsbAudioEngine, currentEngineInfo->getObject(kEngine));
				FailIf(NULL == currentEngine, Exit);

				DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::message() - Resetting engine %p...", this, currentEngine);

				currentEngine->pauseAudioEngine();

				// <rdar://6277511> Make sure that mStreamInterface is not null.
				if (currentEngine)
				{
					// Close the stream interface just to be safe.
					currentEngine->closeStreamInterfaces();

#ifdef TARGET_PDJ
 					// (mitani) 2010/12/16 追加でスリープから復帰するようになる
 					mControlInterface->close(this);
 					mControlInterface->open(this);
#endif

					// Reopen the stream interface
					currentEngine->openStreamInterfaces();
				}

				currentEngine->resumeAudioEngine();
			}
			break;

		default:
			break;
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::message(0x%x, %p) - rc = %d", this, type, provider, this->getRetainCount());
	return kIOReturnSuccess;
}

/****************************************************************************/
/**
 * @brief	プロバイダ終了通知．
 *			IOAudioDevice::willTerminate() をオーバーライド．
 *			サービスプロバイダがサービスを終了した時に呼び出される．
 *
 * @param[in]		*provider		サービスを終了したプロバイダ nub．
 *									PROVIDER_USB_DEVICE が定義されていれば IOUSBHostDevice，
 *									未定義なら IOUSBHostInterface．
 * @param[in]		options			オプション指定．
 *
 * @retval			true			プロバイダの終了を受け入れる．
 * @retval			false			プロバイダの終了に対処できない．
 */
bool
PdjUsbAudioDevice::willTerminate(
	IOService		*provider,
	IOOptionBits	options
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::willTerminate(%p)", this, provider);

#ifdef PROVIDER_USB_DEVICE
	if (mUSBDevice == provider)
#else
	if (mControlInterface == provider)
#endif
	{
		mTerminatingDriver = true;
	}

	// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
	if (NULL != mInterruptPipe)
	{
		mInterruptPipe->abort();
		mInterruptPipe->release();
		mInterruptPipe = NULL;
	}

	if (mProcessStatusInterruptThread)
	{
		::thread_call_cancel(mProcessStatusInterruptThread);
		::thread_call_free(mProcessStatusInterruptThread);
		mProcessStatusInterruptThread = NULL;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::willTerminate()", this);

	return super::willTerminate(provider, options);
}

/****************************************************************************/
/**
 * @brief	デバイスの電源状態変更．
 *			IOAudioDevice::performPowerStateChange() をオーバーライド．
 *
 * @param[in]		oldPowerState				変更前の電源状態．
 * @param[in]		newPowerState				変更後の電源状態．
 * @param[in, out]	*microSecsUntilComplete		この関数から return 後、電源状態変更への対応が
 *												終わるまでに要する時間 [mSec]．
 *												非同期の処理が必要な場合は、この値を書き換えることができる．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDevice::performPowerStateChange(
	IOAudioDevicePowerState		oldPowerState,
	IOAudioDevicePowerState		newPowerState,
	UInt32						*microSecsUntilComplete
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::performPowerStateChange(%d, %d, %p)",
		this, oldPowerState, newPowerState, microSecsUntilComplete);

	IOReturn					result;
	bool						performDeviceResetOnWake = false;
	IOUSBHostDevice				*usbDevice = NULL;

	result = super::performPowerStateChange(oldPowerState, newPowerState, microSecsUntilComplete);

	// We need to stop the time stamp rate timer now
	if ((mUpdateTimer) && (kIOAudioDeviceSleep == newPowerState))
	{
		// Stop the timer and reset the anchor.
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::performPowerStateChange() - Going to sleep - stopping the rate timer.", this);
		mUpdateTimer->cancelTimeout();

		// The frame/time correlation isn't preserved across sleep/wake
		mLastUSBFrame		= 0ull;
		mLastWallTime_nanos	= 0ull;
	}

	if (oldPowerState == kIOAudioDeviceSleep)
	{
		// A new anchor should be taken at the first opportunity. The timer action will handle this with the following instruction.
		// [rdar://4380545] We need to reset the wall time per USB cycle because the frame number could become invalid entering sleep.
		this->resetRateTimer();

		// We should get a new anchor immediately.
		this->updateUSBCycleTime();		// <rdar://problem/7378275>, <rdar://problem/7666699>

#if RESETAFTERSLEEP
		// [rdar://4234453] Reset the device after waking from sleep just to be safe.
		performDeviceResetOnWake = true;
#endif

		FailIf(NULL == mControlInterface, Exit);

		// <rdar://7499125> Check if the getDevice() returns a valid pointer.
		usbDevice = mControlInterface->getDevice();
		FailIf(NULL == usbDevice, Exit);

		// <rdar://problem/6392504> Make sure that the device is connected before resetting it
		if (kIOReturnSuccess == usbDevice->message(kIOUSBMessageHubIsDeviceConnected, NULL, 0))	// <rdar://7499125>
		{
			if (performDeviceResetOnWake)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::performPowerStateChange() - Resetting port after wake from sleep ...", this);
//				usbDevice->ResetDevice();	// <rdar://7499125>
				IOSleep(10);
			}
		}

		// We need to restart the time stamp rate timer now
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::performPowerStateChange() - Waking from sleep - restarting the rate timer.", this);
		this->timerAction(this, mUpdateTimer);

		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::performPowerStateChange() - Flushing controls to the device ...", this);
		super::flushAudioControls();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::performPowerStateChange(%d, %d, %p)",
		this, oldPowerState, newPowerState, microSecsUntilComplete);
	return result;
}

/****************************************************************************/
/**
 * @brief	デバイス設定用の外部アプリを登録．
 *			プラグインからの要求により、必要に応じて外部アプリを関連付ける．
 *
 * @param[in]		*bundleID		外部アプリのバンドル ID 文字列．
 */
void
PdjUsbAudioDevice::setConfigurationApp(
	const char		*bundleID
)
{
	super::setConfigurationApplicationBundle(bundleID);
}

/****************************************************************************/
/**
 * @brief	USB インターフェイス取得．
 *
 * @param[in]		interfaceNumber		USB interface 番号．
 *
 * @return			IOUSBHostInterface クラスのインスタンス．
 */
IOUSBHostInterface *
PdjUsbAudioDevice::getUSBInterface(
	UInt8			interfaceNumber
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "::getUSBInterface()");

	IOUSBHostInterface	*interface	= NULL;
	OSIterator			*iterator	= NULL;

	FailIf(NULL == mControlInterface, Exit);

	iterator = mControlInterface->getDevice()->getChildIterator(gIOServicePlane);
	FailIf(NULL == iterator, Exit);

	for (OSObject *object = iterator->getNextObject(); NULL != object; object = iterator->getNextObject())
	{
		interface = OSDynamicCast(IOUSBHostInterface, object);

		if (NULL != interface)
		{
			if (interface->getInterfaceDescriptor()->bInterfaceNumber == interfaceNumber)
			{
				break;
			}
#ifdef BUG_FIX
			else
			{
				interface = NULL;
			}
#endif
		}
	}

	iterator->release();	// <rdar://7362494>

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "::getUSBInterface()");
	return interface;
}

/****************************************************************************/
/**
 * @brief	USB ストリングディスクリプタ取得．
 *
 * @param[in]		index			ストリングディスクリプタインデックス．
 * @param[out]		*buf			取得した文字列．UNICODE は UTF-8 に変換される．
 * @param[in]		maxLen			バッファサイズ．終端の NULL 文字分を含む．
 * @param[in]		lang			言語 ID（wLANGID）．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDevice::getStringDescriptor(
	UInt8			index,
	char			*buf,
	int				maxLen,
	UInt16			lang
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "::getStringDescriptor()");

	IOReturn				result = kIOReturnError;
	IOUSBHostDevice			*usbDevice;
	const StringDescriptor	*stringDescriptor;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	stringDescriptor = usbDevice->getStringDescriptor(index);

	if (stringDescriptor && (stringDescriptor->bLength > StandardUSB::kDescriptorSize))
	{
		size_t	stringLength = maxLen;

		StandardUSB::stringDescriptorToUTF8(stringDescriptor, buf, stringLength);

		result = kIOReturnSuccess;
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "::getStringDescriptor() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	製造者名のストリングディスクリプタのインデックスを取得．
 *
 * @return			製造者名のストリングディスクリプタのインデックス．
 */
UInt8
PdjUsbAudioDevice::getManufacturerStringIndex(
	void
)
{
	IOUSBHostDevice		*usbDevice;
	UInt8				stringIndex = 0;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	stringIndex = usbDevice->getDeviceDescriptor()->iManufacturer;

Exit:
	return stringIndex;
}

/****************************************************************************/
/**
 * @brief	製品名のストリングディスクリプタのインデックスを取得．
 *
 * @return			製品名のストリングディスクリプタのインデックス．
 */
UInt8
PdjUsbAudioDevice::getProductStringIndex(
	void
)
{
	IOUSBHostDevice		*usbDevice;
	UInt8				stringIndex = 0;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	stringIndex = usbDevice->getDeviceDescriptor()->iProduct;

Exit:
	return stringIndex;
}

/****************************************************************************/
/**
 * @brief	シリアル番号のストリングディスクリプタのインデックスを取得．
 *
 * @return			シリアル番号のストリングディスクリプタのインデックス．
 */
UInt8
PdjUsbAudioDevice::getSerialNumberStringIndex(
	void
)
{
	IOUSBHostDevice		*usbDevice;
	UInt8				stringIndex = 0;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	stringIndex = usbDevice->getDeviceDescriptor()->iSerialNumber;

Exit:
	return stringIndex;
}

/****************************************************************************/
/**
 * @brief	USB Vendor ID を取得．
 *
 * @return			Vendor ID．
 */
UInt16
PdjUsbAudioDevice::getVendorID(
	void
)
{
	IOUSBHostDevice		*usbDevice;
	UInt16				vendorID = 0;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	vendorID = USBToHost16(usbDevice->getDeviceDescriptor()->idVendor);

Exit:
	return vendorID;
}

/****************************************************************************/
/**
 * @brief	USB Product ID を取得．
 *
 * @return			Product ID．
 */
UInt16
PdjUsbAudioDevice::getProductID(
	void
)
{
	IOUSBHostDevice		*usbDevice;
	UInt16				productID = 0;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	productID = USBToHost16(usbDevice->getDeviceDescriptor()->idProduct);

Exit:
	return productID;
}

/****************************************************************************/
/**
 * @brief	USB ロケーション ID を取得．
 *			ロケーション ID は、デバイスが接続されている USB ポートに応じて
 *			IOUSBFamily が付与する ID である．
 *
 * @return			ロケーション ID．
 */
OSNumber *
PdjUsbAudioDevice::getLocationID(
	void
)
{
	IOUSBHostDevice		*usbDevice;
	OSNumber			*usbLocation = NULL;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();

	usbLocation = OSDynamicCast(OSNumber, usbDevice->getProperty(kUSBDevicePropertyLocationID));

Exit:
	return usbLocation;
}

/****************************************************************************/
/**
 * @brief	USB フレーム番号取得．
 *			IOUSBFamily はフレーム番号のオーバーフローを監視しており、
 *			64bit まで有効なフレーム番号を保持している．
 *			マイクロフレーム番号は含まれない．
 *
 * @return			フレーム番号（64bit まで有効）．
 */
UInt64
PdjUsbAudioDevice::getUSBFrameNumber(
	void
)
{
	IOUSBHostDevice		*usbDevice;
	UInt64				frameNumber = 0;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	frameNumber = usbDevice->getFrameNumber();

Exit:
	return frameNumber;
}

/****************************************************************************/
/**
 * @brief	USB バススピード取得．
 *
 * @retval			kUSBDeviceSpeedLow		Low   Speed．
 * @retval			kUSBDeviceSpeedFull		Full  Speed．
 * @retval			kUSBDeviceSpeedHigh		High  Speed．
 * @retval			kUSBDeviceSpeedSuper	Super Speed．
 */
// added for rdar://4168019 . Returns the device speed (high, full, or low).
UInt8
PdjUsbAudioDevice::getDeviceSpeed(
	void
)
{
	IOUSBHostDevice		*usbDevice;
	UInt8				speed = 0;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();

	speed = usbDevice->getSpeed();

#ifdef DEBUGLOGGING
	switch (speed)
	{
	case kUSBDeviceSpeedLow:
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::getDeviceSpeed() = kUSBDeviceSpeedLow", this);
		break;
	case kUSBDeviceSpeedFull:
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::getDeviceSpeed() = kUSBDeviceSpeedFull", this);
		break;
	case kUSBDeviceSpeedHigh:
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::getDeviceSpeed() = kUSBDeviceSpeedHigh", this);
		break;
	case kUSBDeviceSpeedSuper:
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::getDeviceSpeed() = kUSBDeviceSpeedSuper", this);
		break;
	default:
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::getDeviceSpeed() = %d (UNKNOWN)", this, speed);
	}
#endif

Exit:
	return speed;
}

/****************************************************************************/
/**
 * @brief	スプリットトランザクションの有無を調べる．
 *			このデバイスが Full Speed で、ハブが High Speed であるとき、true を返す．
 *
 * @retval			true			このデバイスは Full Speed であり、ハブは High Speed である．
 * @retval			false			上記以外．
 */
// added for rdar://3959606 . Detects whether this is a full speed device plugged into a high speed hub.
bool
PdjUsbAudioDevice::detectSplitTransactions(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "::detectSplitTransactions()");

	IOUSBHostDevice			*usbDevice;
	const IORegistryPlane	*usbPlane = IORegistryEntry::getPlane(kIOUSBPlane);
	IORegistryEntry			*currentEntry;
	UInt8					speed;
	bool					canStop = false;
	bool					splitTransactions = false;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	currentEntry = OSDynamicCast(IORegistryEntry, usbDevice);

	if (kUSBDeviceSpeedHigh == usbDevice->getSpeed())
	{
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "::detectSplitTransactions() - This is a high speed device, so there are no split transactions.");
		splitTransactions = false;
		canStop = true;
	}

	while ((!canStop) && (currentEntry) && (usbDevice))
	{
		// Searching for a high speed hub. Since it is not possible to run at high speed with a full speed
		// hub closer to the root hub in the chain, we can stop when we find the first high speed hub.

		speed = usbDevice->getSpeed();

		if (kUSBDeviceSpeedHigh == speed)
		{
			// Must be connected via USB 2.0 hub
			DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "::detectSplitTransactions() = true");
			splitTransactions = true;
			canStop = true;
		}
		else
		{
			// Get parent in USB plane
			currentEntry = OSDynamicCast(IORegistryEntry, currentEntry->getParentEntry(usbPlane));

			// If the current registry entry is not a device, this will make usbDevice NULL and exit the loop
			usbDevice = OSDynamicCast(IOUSBHostDevice, currentEntry);
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "::detectSplitTransactions() = %d", splitTransactions);
	return splitTransactions;
}

#ifdef TARGET_PDJ
 /****************************************************************************/
 /**
  * @brief	USB ホストコントローラが xHCI であるかどうか調べる．
  *
  * @retval			true			ホストコントローラは xHCI である．
  * @retval			false			ホストコントローラは xHCI でない．
  */
 bool
 PdjUsbAudioDevice::checkForXHCI(
 	void
 )
 {
 	return true;
 }
#else
/****************************************************************************/
/**
 * @brief	USB ホストコントローラが UHCI であるかどうか調べる．
 *
 * @retval			true			ホストコントローラは UHCI である．
 * @retval			false			ホストコントローラは UHCI でない．
 */
bool
PdjUsbAudioDevice::checkForUHCI(
	void
)
{
	bool					result = false;
	IOUSBHostDevice			*usbDevice;
	const IORegistryPlane	*servicePlane = IORegistryEntry::getPlane(kIOServicePlane);
	IOService				*currentEntry;
	IORegistryEntry			*parentEntry;
	char					serviceName[20];

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	parentEntry = usbDevice->getParentEntry(servicePlane);
	FailIf(NULL == parentEntry, Exit);

	currentEntry = OSDynamicCast(IOService, parentEntry);
	FailIf(NULL == currentEntry, Exit);

	::strncpy(serviceName, currentEntry->getName(servicePlane), 20);

	while (
				(currentEntry)
		  &&
				::strcmp(serviceName, "AppleUSBUHCI")
		  &&
				::strcmp(serviceName, "AppleUSBOHCI")
		  &&
				::strcmp(serviceName, "AppleUSBEHCI")
		  &&
				::strcmp(serviceName, "AppleUSBXHCI")
		  )
	{
		// Searching for the root hub type. We only expect one of three types, so we can stop when we have found any of them.
		// The types are AppleUSBUHCI, AppleUSBOHCI, and AppleUSBEHCI.

		// Get parent in IOService plane
		parentEntry = currentEntry->getParentEntry(servicePlane);
		FailIf(NULL == parentEntry, Exit);

		currentEntry = OSDynamicCast(IOService, parentEntry);
		if (currentEntry)
		{
			::strncpy(serviceName, currentEntry->getName(servicePlane), 20);
		}
	}

	FailIf(NULL == currentEntry, Exit);

	if (0 == ::strcmp(serviceName, "AppleUSBUHCI"))
	{
		// We are on a UHCI connection
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "::checkForUHCI() - UHCI connection detected!\n");
		result = true;
	}

Exit:
	return result;
}
#endif

/****************************************************************************/
/**
 * @brief	USB デバイスリクエスト発行．
 *
 * @param[in]		*request			デバイスリクエスト．
 * @param[in, out]	*dataBuffer			デバイスリクエストのデータ．
 * @param[out]		*bytesTransferred	転送バイト数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDevice::deviceRequest(
	StandardUSB::DeviceRequest	*request,
	void						*dataBuffer,
	UInt32						*bytesTransferred
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::deviceRequest()", this);

	IOReturn	result = kIOReturnError;
	UInt32		numBytes = 0;

	FailIf(NULL == mControlInterface, Exit);

	FailIf(NULL == mInterfaceLock, Exit);
	IORecursiveLockLock(mInterfaceLock);

	if (!mTerminatingDriver)
	{
		result = mControlInterface->deviceRequest(*request, dataBuffer, numBytes);
	}

	IORecursiveLockUnlock(mInterfaceLock);

	if (bytesTransferred)
	{
		*bytesTransferred = numBytes;
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::deviceRequest() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	USB デバイスリクエスト発行．
 *
 * @param[in]		*request		デバイスリクエスト．
 * @param[in, out]	*dataBuffer		デバイスリクエストのデータ．
 * @param[in]		*completion		完了ルーチン．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDevice::deviceRequest(
	StandardUSB::DeviceRequest	*request,
	IOMemoryDescriptor			*dataBuffer,
	IOUSBHostCompletion			*completion
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::deviceRequest(%p, %p)", this, request, completion);

	IOReturn				result = kIOReturnSuccess;
	bool					done;
	UInt32					timeout;

	FailIf(NULL == mInterfaceLock, Exit);
	IORecursiveLockLock(mInterfaceLock);

	if (!mTerminatingDriver)
	{
		done	= false;
		timeout	= 5;

		while (!done && timeout && mControlInterface)
		{
			result = mControlInterface->deviceRequest(
											*request,
											dataBuffer,
											completion,
											kUSBHostStandardRequestCompletionTimeout
										);

			if (result != kIOReturnSuccess)
			{
				timeout--;
				IOSleep(1);
			}
			else
			{
				done = true;
			}
		}
	}

	IORecursiveLockUnlock(mInterfaceLock);

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::deviceRequest(%p, %p) = %d", this, request, completion, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ターミナルタイプを表す文字列．
 *
 * @param[in]		terminalType	入出力ターミナルディスクリプタの wTerminalType フィールド．
 *
 * @return			ターミナルタイプを表す文字列．
 */
char *
PdjUsbAudioDevice::terminalTypeString(
	UInt16			terminalType
)
{
	char			*strTerminalType;

	switch (terminalType)
	{
	case USB_STREAMING:									strTerminalType = (char *)"USB streaming";									break;
#if LOCALIZABLE
	case INPUT_UNDEFINED:								strTerminalType = (char *)"StringInputUndefined";							break;
	case INPUT_MICROPHONE:								strTerminalType = (char *)"StringMicrophone";								break;
	case INPUT_DESKTOP_MICROPHONE:						strTerminalType = (char *)"StringDesktopMicrophone";						break;
	case INPUT_PERSONAL_MICROPHONE:						strTerminalType = (char *)"StringPersonalMicrophone";						break;
	case INPUT_OMNIDIRECTIONAL_MICROPHONE:				strTerminalType = (char *)"StringOmnidirectionalMicrophone";				break;
	case INPUT_MICROPHONE_ARRAY:						strTerminalType = (char *)"StringMicrophoneArray";							break;
	case INPUT_PROCESSING_MICROPHONE_ARRAY:				strTerminalType = (char *)"StringProcessingMicrophoneArray";				break;
	case INPUT_MODEM_AUDIO:								strTerminalType = (char *)"StringModemAudio";								break;
	case OUTPUT_UNDEFINED:								strTerminalType = (char *)"StringOutputUndefined";							break;
	case OUTPUT_SPEAKER:								strTerminalType = (char *)"StringSpeaker";									break;
	case OUTPUT_HEADPHONES:								strTerminalType = (char *)"StringHeadphones";								break;
	case OUTPUT_HEAD_MOUNTED_DISPLAY_AUDIO:				strTerminalType = (char *)"StringHeadMountedDisplayAudio";					break;
	case OUTPUT_DESKTOP_SPEAKER:						strTerminalType = (char *)"StringDesktopSpeaker";							break;
	case OUTPUT_ROOM_SPEAKER:							strTerminalType = (char *)"StringRoomSpeaker";								break;
	case OUTPUT_COMMUNICATION_SPEAKER:					strTerminalType = (char *)"StringCommunicationSpeaker";						break;
	case OUTPUT_LOW_FREQUENCY_EFFECTS_SPEAKER:			strTerminalType = (char *)"StringLowFrequencyEffectsSpeaker";				break;
	case BIDIRECTIONAL_UNDEFINED:						strTerminalType = (char *)"StringBidirectionalUndefined";					break;
	case BIDIRECTIONAL_HANDSET:							strTerminalType = (char *)"StringBidirectionalHandset";						break;
	case BIDIRECTIONAL_HEADSET:							strTerminalType = (char *)"StringBidirectionalHeadset";						break;
	case BIDIRECTIONAL_SPEAKERPHONE_NO_ECHO_REDX:		strTerminalType = (char *)"StringBidirectionalSpeakerphoneNoEchoRedx";		break;
	case BIDIRECTIONAL_ECHO_SUPPRESSING_SPEAKERPHONE:	strTerminalType = (char *)"StringBidirectionalEchoSuppressingSpeakerphone";	break;
	case BIDIRECTIONAL_ECHO_CANCELING_SPEAKERPHONE:		strTerminalType = (char *)"StringBidirectionalEchoCancelingSpeakerphone";	break;
	case TELEPHONY_UNDEFINED:							strTerminalType = (char *)"StringTelephoneUndefined";						break;
	case TELEPHONY_PHONE_LINE:							strTerminalType = (char *)"StringTelephoneLine";							break;
	case TELEPHONY_TELEPHONE:							strTerminalType = (char *)"StringTelephone";								break;
	case TELEPHONY_DOWN_LINE_PHONE:						strTerminalType = (char *)"StringDownLinePhone";							break;
	case EXTERNAL_UNDEFINED:							strTerminalType = (char *)"StringExternalUndefined";						break;
	case EXTERNAL_ANALOG_CONNECTOR:						strTerminalType = (char *)"StringExternalAnalogConnector";					break;
	case EXTERNAL_DIGITAL_AUDIO_INTERFACE:				strTerminalType = (char *)"StringExternalDigitalAudioInterface";			break;
	case EXTERNAL_LINE_CONNECTOR:						strTerminalType = (char *)"StringExternalLineConnector";					break;
	case EXTERNAL_LEGACY_AUDIO_CONNECTOR:				strTerminalType = (char *)"StringExternalLegacyAudioConnector";				break;
	case EXTERNAL_SPDIF_INTERFACE:						strTerminalType = (char *)"StringExternalSPDIFInterface";					break;
	case EXTERNAL_1394_DA_STREAM:						strTerminalType = (char *)"StringExternal1394DAStream";						break;
	case EXTERNAL_1394_DV_STREAM_SOUNDTRACK:			strTerminalType = (char *)"StringExternal1394DVStreamSoundtrack";			break;
	case EMBEDDED_UNDEFINED:							strTerminalType = (char *)"StringEmbeddedUndefined";						break;
	case EMBEDDED_LEVEL_CALIBRATION_NOISE_SOURCE:		strTerminalType = (char *)"StringEmbeddedLevelCalibrationNoiseSource";		break;
	case EMBEDDED_EQUALIZATION_NOISE:					strTerminalType = (char *)"StringEmbeddedEqualizationNoise";				break;
	case EMBEDDED_CD_PLAYER:							strTerminalType = (char *)"StringEmbeddedCDPlayer";							break;
	case EMBEDDED_DAT:									strTerminalType = (char *)"StringEmbeddedDAT";								break;
	case EMBEDDED_DCC:									strTerminalType = (char *)"StringEmbeddedDCC";								break;
	case EMBEDDED_MINIDISK:								strTerminalType = (char *)"StringEmbeddedMiniDisc";							break;
	case EMBEDDED_ANALOG_TAPE:							strTerminalType = (char *)"StringEmbeddedAnalogTape";						break;
	case EMBEDDED_PHONOGRAPH:							strTerminalType = (char *)"StringEmbeddedPhonograph";						break;
	case EMBEDDED_VCR_AUDIO:							strTerminalType = (char *)"StringEmbeddedVCRAudio";							break;
	case EMBEDDED_VIDEO_DISC_AUDIO:						strTerminalType = (char *)"StringEmbeddedVideoDiscAudio";					break;
	case EMBEDDED_DVD_AUDIO:							strTerminalType = (char *)"StringEmbeddedDVDAudio";							break;
	case EMBEDDED_TV_TUNER_AUDIO:						strTerminalType = (char *)"StringEmbeddedTVTunerAudio";						break;
	case EMBEDDED_SATELLITE_RECEIVER_AUDIO:				strTerminalType = (char *)"StringEmbeddedSatelliteReceiverAudio";			break;
	case EMBEDDED_CABLE_TUNER_AUDIO:					strTerminalType = (char *)"StringEmbeddedCableTunerAudio";					break;
	case EMBEDDED_DSS_AUDIO:							strTerminalType = (char *)"StringEmbeddedDSSAudio";							break;
	case EMBEDDED_RADIO_RECEIVER:						strTerminalType = (char *)"StringEmbeddedRadioReceiver";					break;
	case EMBEDDED_RADIO_TRANSMITTER:					strTerminalType = (char *)"StringEmbeddedRadioTransmitter";					break;
	case EMBEDDED_MULTITRACK_RECORDER:					strTerminalType = (char *)"StringEmbeddedMultitrackRecorder";				break;
	case EMBEDDED_SYNTHESIZER:							strTerminalType = (char *)"StringEmbeddedSynthesizer";						break;
	default:											strTerminalType = (char *)"StringUnknown";									break;
#else
	case INPUT_UNDEFINED:								strTerminalType = (char *)"InputUndefined";									break;
	case INPUT_MICROPHONE:								strTerminalType = (char *)"Microphone";										break;
	case INPUT_DESKTOP_MICROPHONE:						strTerminalType = (char *)"Desktop Microphone";								break;
	case INPUT_PERSONAL_MICROPHONE:						strTerminalType = (char *)"Personal Microphone";							break;
	case INPUT_OMNIDIRECTIONAL_MICROPHONE:				strTerminalType = (char *)"Omnidirectional Microphone";						break;
	case INPUT_MICROPHONE_ARRAY:						strTerminalType = (char *)"Microphone Array";								break;
	case INPUT_PROCESSING_MICROPHONE_ARRAY:				strTerminalType = (char *)"Processing Microphone Array";					break;
	case INPUT_MODEM_AUDIO:								strTerminalType = (char *)"Modem Audio";									break;
	case OUTPUT_UNDEFINED:								strTerminalType = (char *)"Output Undefined";								break;
	case OUTPUT_SPEAKER:								strTerminalType = (char *)"Speaker";										break;
	case OUTPUT_HEADPHONES:								strTerminalType = (char *)"Headphones";										break;
	case OUTPUT_HEAD_MOUNTED_DISPLAY_AUDIO:				strTerminalType = (char *)"Head Mounted Display Audio";						break;
	case OUTPUT_DESKTOP_SPEAKER:						strTerminalType = (char *)"Desktop Speaker";								break;
	case OUTPUT_ROOM_SPEAKER:							strTerminalType = (char *)"Room Speaker";									break;
	case OUTPUT_COMMUNICATION_SPEAKER:					strTerminalType = (char *)"Communication Speaker";							break;
	case OUTPUT_LOW_FREQUENCY_EFFECTS_SPEAKER:			strTerminalType = (char *)"Low Frequency Effects Speaker";					break;
	case BIDIRECTIONAL_UNDEFINED:						strTerminalType = (char *)"Bidirectional Undefined";						break;
	case BIDIRECTIONAL_HANDSET:							strTerminalType = (char *)"Bidirectional Handset";							break;
	case BIDIRECTIONAL_HEADSET:							strTerminalType = (char *)"Bidirectional Headset";							break;
	case BIDIRECTIONAL_SPEAKERPHONE_NO_ECHO_REDX:		strTerminalType = (char *)"Bidirectional Speakerphone No Echo Redx";		break;
	case BIDIRECTIONAL_ECHO_SUPPRESSING_SPEAKERPHONE:	strTerminalType = (char *)"Bidirectional Echo Suppressing Speakerphone";	break;
	case BIDIRECTIONAL_ECHO_CANCELING_SPEAKERPHONE:		strTerminalType = (char *)"Bidirectional Echo Canceling Speakerphone";		break;
	case TELEPHONY_UNDEFINED:							strTerminalType = (char *)"Telephone Undefined";							break;
	case TELEPHONY_PHONE_LINE:							strTerminalType = (char *)"Telephone Line";									break;
	case TELEPHONY_TELEPHONE:							strTerminalType = (char *)"Telephone";										break;
	case TELEPHONY_DOWN_LINE_PHONE:						strTerminalType = (char *)"Down Line Phone";								break;
	case EXTERNAL_UNDEFINED:							strTerminalType = (char *)"External Undefined";								break;
	case EXTERNAL_ANALOG_CONNECTOR:						strTerminalType = (char *)"External Analog Connector";						break;
	case EXTERNAL_DIGITAL_AUDIO_INTERFACE:				strTerminalType = (char *)"External Digital Audio Interface";				break;
	case EXTERNAL_LINE_CONNECTOR:						strTerminalType = (char *)"External Line Connector";						break;
	case EXTERNAL_LEGACY_AUDIO_CONNECTOR:				strTerminalType = (char *)"External Legacy Audio Connector";				break;
	case EXTERNAL_SPDIF_INTERFACE:						strTerminalType = (char *)"External SPDIF Interface";						break;
	case EXTERNAL_1394_DA_STREAM:						strTerminalType = (char *)"External 1394 DA Stream";						break;
	case EXTERNAL_1394_DV_STREAM_SOUNDTRACK:			strTerminalType = (char *)"External 1394 DV Stream Soundtrack";				break;
	case EMBEDDED_UNDEFINED:							strTerminalType = (char *)"Embedded Undefined";								break;
	case EMBEDDED_LEVEL_CALIBRATION_NOISE_SOURCE:		strTerminalType = (char *)"Embedded Level Calibration Noise Source";		break;
	case EMBEDDED_EQUALIZATION_NOISE:					strTerminalType = (char *)"Embedded Equalization Noise";					break;
	case EMBEDDED_CD_PLAYER:							strTerminalType = (char *)"Embedded CD Player";								break;
	case EMBEDDED_DAT:									strTerminalType = (char *)"Embedded DAT";									break;
	case EMBEDDED_DCC:									strTerminalType = (char *)"Embedded DCC";									break;
	case EMBEDDED_MINIDISK:								strTerminalType = (char *)"Embedded Mini Disc";								break;
	case EMBEDDED_ANALOG_TAPE:							strTerminalType = (char *)"Embedded Analog Tape";							break;
	case EMBEDDED_PHONOGRAPH:							strTerminalType = (char *)"Embedded Phonograph";							break;
	case EMBEDDED_VCR_AUDIO:							strTerminalType = (char *)"Embedded VCR Audio";								break;
	case EMBEDDED_VIDEO_DISC_AUDIO:						strTerminalType = (char *)"Embedded Video Disc Audio";						break;
	case EMBEDDED_DVD_AUDIO:							strTerminalType = (char *)"Embedded DVD Audio";								break;
	case EMBEDDED_TV_TUNER_AUDIO:						strTerminalType = (char *)"Embedded TV Tuner Audio";						break;
	case EMBEDDED_SATELLITE_RECEIVER_AUDIO:				strTerminalType = (char *)"Embedded Satellite Receiver Audio";				break;
	case EMBEDDED_CABLE_TUNER_AUDIO:					strTerminalType = (char *)"Embedded Cable Tuner Audio";						break;
	case EMBEDDED_DSS_AUDIO:							strTerminalType = (char *)"Embedded DSS Audio";								break;
	case EMBEDDED_RADIO_RECEIVER:						strTerminalType = (char *)"Embedded Radio Receiver";						break;
	case EMBEDDED_RADIO_TRANSMITTER:					strTerminalType = (char *)"Embedded Radio Transmitter";						break;
	case EMBEDDED_MULTITRACK_RECORDER:					strTerminalType = (char *)"Embedded Multitrack Recorder";					break;
	case EMBEDDED_SYNTHESIZER:							strTerminalType = (char *)"Embedded Synthesizer";							break;
	default:											strTerminalType = (char *)"Unknown";										break;
#endif
	}

	return strTerminalType;
}

/****************************************************************************/
/**
 * @brief	クロックタイプを表す文字列．
 *
 * @param[in]		clockType		クロックソースディスクリプタの bmAttributes フィールド
 *									で指定されたクロックタイプ．
 *
 * @return			クロックタイプを表す文字列．
 */
// <rdar://5811247>
char *
PdjUsbAudioDevice::clockTypeString(
	UInt8			clockType
)
{
	char			*strClockType;

	switch (clockType) {
#if LOCALIZABLE
	case USBAUDIO_0200::CLOCK_TYPE_EXTERNAL:				strClockType = (char *)"StringExternalClock";				break;
	case USBAUDIO_0200::CLOCK_TYPE_INTERNAL_FIXED:			strClockType = (char *)"StringInternalFixedClock";			break;
	case USBAUDIO_0200::CLOCK_TYPE_INTERNAL_VARIABLE:		strClockType = (char *)"StringInternalVariableClock";		break;
	case USBAUDIO_0200::CLOCK_TYPE_INTERNAL_PROGRAMMABLE:	strClockType = (char *)"StringInternalProgrammableClock";	break;
	default:												strClockType = (char *)"StringUnknown";						break;
#else
	case USBAUDIO_0200::CLOCK_TYPE_EXTERNAL:				strClockType = (char *)"External Clock";					break;
	case USBAUDIO_0200::CLOCK_TYPE_INTERNAL_FIXED:			strClockType = (char *)"Internal Fixed Clock";				break;
	case USBAUDIO_0200::CLOCK_TYPE_INTERNAL_VARIABLE:		strClockType = (char *)"Internal Variable Clock";			break;
	case USBAUDIO_0200::CLOCK_TYPE_INTERNAL_PROGRAMMABLE:	strClockType = (char *)"Internal Programmable Clock";		break;
	default:												strClockType = (char *)"Unknown";							break;
#endif
	}

	return strClockType;
}

/****************************************************************************/
/**
 * @brief	基準時刻が大きくずれた場合はオフセット補正を行なう．
 */
// <rdar://problem/7666699> Check to see if system time has discontinuity by calculating offset to be applied to filtered cycle time
void
PdjUsbAudioDevice::calculateOffset(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "::calculateOffset(%p)", this);

	if (this->allEnginesStopped())
	{
		mRampUpdateCounter = 0;	// increase cycle time update frequency

		if (mTimeLock && (mAnchorTime.n >= MIN_ENTRIES_APPLY_OFFSET))
		{
			this->applyOffsetAmountToFilter();
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "::calculateOffset(%p)", this);
}

/****************************************************************************/
/**
 * @brief	USB フレーム番号から時刻を計算．
 *
 * @param[in]		frameNumber		USB フレーム番号．
 *
 * @return			算出された時刻．
 */
// This function should only be called if anchorTime->n > 1
UInt64
PdjUsbAudioDevice::getTimeForFrameNumber(
	UInt64			frameNumber
)
{
	DP(DBG_TIMESTAMP, "+ " LOGSTRAUDIODEVICE "::getTimeForFrameNumber()");

	UInt64			result = 0;

	if (mAnchorTime.n > 1)
	{
		// y = ( P * ( n * x - sumX ) + QSumY ) / Qn;
		// <rdar://problem/7711404> Split the calculation up to delay the subtraction to the last moment to avoid underflow.
		U128	nx		= ::mul64(mAnchorTime.n, frameNumber);

#if (MAX_ANCHOR_ENTRIES <= 1024)
		U256	Pnx		= ::mul128(mAnchorTime.P, nx.lo);
		U256	PsumX	= ::mul128(mAnchorTime.P, mAnchorTime.sumX);
		U256	temp	= ::add256(Pnx, mAnchorTime.QSumY);

		temp = ::sub256(temp, PsumX);
#else
		U512	Pnx		= ::mul256(mAnchorTime.P, nx.lo);
		U512	PsumX	= ::mul256(mAnchorTime.P, mAnchorTime.sumX);
		U256	temp	= ::add256(Pnx.lo, mAnchorTime.QSumY);

		temp = ::sub256(temp, PsumX.lo);
#endif

		temp = ::div256(temp, mAnchorTime.Qn);

		result = temp.lo.lo;
	}

	DP(DBG_TIMESTAMP, "- " LOGSTRAUDIODEVICE "::getTimeForFrameNumber()");
	return result;
}

/****************************************************************************/
/**
 * @brief	USB のフレーム間隔を再計算．
 */
void
PdjUsbAudioDevice::updateUSBCycleTime(
	void
)
{
	DP(DBG_TIMER, "+ " LOGSTRAUDIODEVICE "::updateUSBCycleTime()");

	AbsoluteTime	timeStamp;
	UInt64			frameNumber;

	if ((kIOReturnSuccess == this->getAnchorFrameAndTimeStamp(&frameNumber, &timeStamp)) && mTimeLock)
	{
		UInt64			timeStamp_nanos;
		AbsoluteTime	refTime;

		::absolutetime_to_nanoseconds(timeStamp, &timeStamp_nanos);
		DP(DBG_TIMESTAMP, "   frameNumber = %llu, timeStamp = %llu\n", frameNumber, timeStamp);

		IOLockLock(mTimeLock);

		DP(DBG_TIMESTAMP, "? " LOGSTRAUDIODEVICE "::updateUSBCycleTime() mWallTimePerUSBCycle = %llu frames elapsed: %llu\n",
			mWallTimePerUSBCycle, frameNumber - lastAnchorFrame());

		this->updateAnchorTime(&mAnchorTime, frameNumber, timeStamp_nanos);
		DP(DBG_TIMESTAMP, "   mAnchorTime.n  = %u\n", mAnchorTime.n);

		if (mAnchorTime.n > 1)
		{
			UInt64	refTime_nanos = this->getTimeForFrameNumber(frameNumber);

			::nanoseconds_to_absolutetime(refTime_nanos, &refTime);
			DP(DBG_TIMESTAMP, "   frameNumber = %llu, refTime_nanos = %llu\n", frameNumber, refTime_nanos);

			mWallTimePerUSBCycle = this->getUSBCycleTime(&mAnchorTime);
		}
		else
		{
			::nanoseconds_to_absolutetime(0, &refTime);

			mWallTimePerUSBCycle = 1000000ull * kWallTimeExtraPrecision;
		}

		IOLockUnlock(mTimeLock);

		DP(DBG_TIMESTAMP, "   New anchor! frameNumber = %llu, refTime = %llu\n", frameNumber, refTime);
		DP(DBG_TIMESTAMP, "   mWallTimePerUSBCycle = %llu\n", mWallTimePerUSBCycle);
	}

	DP(DBG_TIMER, "- " LOGSTRAUDIODEVICE "::updateUSBCycleTime()");
}

/****************************************************************************/
/**
 * @brief	オーディオコントロール生成．
 *			PdjUsbAudioEngine::controlledFormatChange() から呼び出される．
 *			interface 内の各種オーディオコントロールを生成する．
 *
 * @param[in]		*audioEngine		PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		interfaceNum		USB interface 番号．
 * @param[in]		altSettingNum		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDevice::createControlsForInterface(
	IOAudioEngine	*audioEngine,
	UInt8			interfaceNum,
	UInt8			altSettingNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::createControlsForInterface() - Interface %d alternate setting %d",
		this, interfaceNum, altSettingNum);

	IOReturn	result = kIOReturnError;

	mTerminatingDriver = false;

	PdjUsbAudioEngine	*usbAudioEngine = OSDynamicCast(PdjUsbAudioEngine, audioEngine);
	FailIf(NULL == usbAudioEngine, Exit);

	mTopology->doControlStuff(audioEngine, interfaceNum, altSettingNum);

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::createControlsForInterface()", this);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオフォーマット変更．
 *			PdjUsbAudioEngine::performFormatChange() から呼び出される．
 *
 * @param[in]		*audioEngine		PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		*audioStream		フォーマット変更の対象となるストリーム．
 * @param[in]		*newFormat			変更後のフォーマット．
 * @param[in]		*newSampleRate		変更後のサンプリング周波数．
 *
 * @return			IOReturn コード．
 */
UInt32
PdjUsbAudioDevice::formatChangeController(
	PdjUsbAudioEngine			*audioEngine,
	IOAudioStream				*audioStream,
	const IOAudioStreamFormat	*newFormat,
	const IOAudioSampleRate		*newSampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::formatChangeController(%p, %p, %p, %p)",
		this, audioEngine, audioStream, newFormat, newSampleRate);

	PdjUsbAudioEngine			*thisAudioEngine	= audioEngine;
	PdjUsbAudioEngine			*otherAudioEngine	= NULL;
	PdjUsbAudioStream			*thisStream			= NULL;
	PdjUsbAudioStream			*otherStream		= NULL;
	const IOAudioStreamFormat	*thisFormat;
	const IOAudioStreamFormat	*otherFormat;
	const IOAudioSampleRate		*thisSampleRate;
	const IOAudioSampleRate		*otherSampleRate;
	const IOAudioStreamFormat	*thisDefaultAudioStreamFormat;
	const IOAudioStreamFormat	*otherDefaultAudioStreamFormat;
	UInt32						result = kPdjUsbFormatChangeError;
	IOReturn					formatChangeReturnCode = kIOReturnError;
	UInt8						altSetting;
	bool						mustMatchFormats;
	bool						enginesPaused = false;
	bool						otherEngineNeedSampleRateChange = false;

	FailIf(NULL == mControlInterface, Exit);

	// [rdar://4801012]
	if (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	{
		if (newFormat && newSampleRate)
		{
			mTopology->getOptimalClockPath(
							thisAudioEngine,
							(UInt8)(newFormat->fDriverTag >> 16),
							(UInt8)(newFormat->fDriverTag),
							newSampleRate->whole,
							&otherEngineNeedSampleRateChange
					   );
		}
	}

	mustMatchFormats = (
							(
								(mRegisteredEngines)
							&&
								(2 == mRegisteredEngines->getCount())
							&&
								(
									(mSingleSampleRateDevice)
								||
									(NULL == audioStream)
								)
							)
					   ||
							(otherEngineNeedSampleRateChange)
					   );

	if (mustMatchFormats)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - Attempting to match this format with the format for the other stream interface.", this);
		result = kPdjUsbFormatChangeForceFailure;

		if (NULL == thisAudioEngine)
		{
			// Get both engines.
			FailIf(kIOReturnSuccess != mTopology->getBothEngines(&thisAudioEngine, &otherAudioEngine), Exit);
		}
		else
		{
			// Just get the other engine.
			otherAudioEngine = mTopology->otherEngine(thisAudioEngine);
		}

		FailIf(NULL == thisAudioEngine, Exit);
		FailIf(NULL == otherAudioEngine, Exit);

		thisAudioEngine->pauseAudioEngine();
		otherAudioEngine->pauseAudioEngine();
		enginesPaused = true;

		// <rdar://7298196> Lock all audio streams for I/O since the format changes deallocate the sample buffer &
		// related structure. This is to prevent a race condition between the format changes and performClientIO()/USB
		// callback completion routines.
		thisAudioEngine->lockAllStreams();
		otherAudioEngine->lockAllStreams();

		// Get formats and other sample rate.
		thisStream  = thisAudioEngine->mMainOutputStream  ? thisAudioEngine->mMainOutputStream  : thisAudioEngine->mMainInputStream;
		otherStream = otherAudioEngine->mMainOutputStream ? otherAudioEngine->mMainOutputStream : otherAudioEngine->mMainInputStream;
		FailIf(NULL == thisStream,  Exit);
		FailIf(NULL == otherStream, Exit);

		thisFormat  = thisStream->getFormat();
		otherFormat = otherStream->getFormat();
		FailIf(NULL == thisFormat,  Exit);
		FailIf(NULL == otherFormat, Exit);

		thisSampleRate  = thisAudioEngine->getSampleRate();
		otherSampleRate = otherAudioEngine->getSampleRate();
		FailIf(NULL == thisSampleRate,  Exit);
		FailIf(NULL == otherSampleRate, Exit);

		thisDefaultAudioStreamFormat  = &thisStream->mDefaultAudioStreamFormat;
		otherDefaultAudioStreamFormat = &otherStream->mDefaultAudioStreamFormat;
		FailIf(NULL == thisDefaultAudioStreamFormat,  Exit);
		FailIf(NULL == otherDefaultAudioStreamFormat, Exit);

		// Log what we have so far.
		DP(DBG_INFO, "\n");
		DP(DBG_INFO, "-------------------- BEFORE --------------------");
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - engine %p (interface %d, alternateSetting %d) info:",
			this, thisAudioEngine, thisStream->mInterfaceNumber, thisStream->mAlternateSettingID);
		DP(DBG_INFO, "    thisFormat = %p", thisFormat);
		DP(DBG_INFO, "        fNumChannels = %u", thisFormat->fNumChannels);
		DP(DBG_INFO, "        fBitDepth = %d", thisFormat->fBitDepth);
		DP(DBG_INFO, "        fDriverTag = 0x%x", thisFormat->fDriverTag);
		DP(DBG_INFO, "    thisSampleRate->whole = %u", thisSampleRate->whole);
		DP(DBG_INFO, "\n");
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - engine %p (interface %d, alternateSetting %d) info:",
			this, otherAudioEngine, otherStream->mInterfaceNumber, otherStream->mAlternateSettingID);
		DP(DBG_INFO, "    otherFormat = %p", otherFormat);
		DP(DBG_INFO, "        fNumChannels = %u", otherFormat->fNumChannels);
		DP(DBG_INFO, "        fBitDepth = %d", otherFormat->fBitDepth);
		DP(DBG_INFO, "        fDriverTag = 0x%x", otherFormat->fDriverTag);
		DP(DBG_INFO, "    otherSampleRate->whole = %u", otherSampleRate->whole);
		DP(DBG_INFO, "\n");
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - newFormat = %p", this, newFormat);

		if (newFormat)
		{
			DP(DBG_INFO, "        fNumChannels = %u", newFormat->fNumChannels);
			DP(DBG_INFO, "        fBitDepth = %d", newFormat->fBitDepth);
			DP(DBG_INFO, "        fDriverTag = 0x%x", newFormat->fDriverTag);
		}
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - newSampleRate = %p", this, newSampleRate);

		if (newSampleRate)
		{
			DP(DBG_INFO, "        whole = %u", newSampleRate->whole);
		}

		DP(DBG_INFO, "------------------------------------------------");
		DP(DBG_INFO, "\n");

		if ((!mSingleSampleRateDevice) && (!otherEngineNeedSampleRateChange))
		{
			// This is an emergency format change. We need to determine which engine(s) need(!s) to get a format change. Our order of preferability is:
			//		1. This engine should change if the sample rate can be matched at the current bit depth and channel count. If not,
			//		2. The other engine should change if this engine's reported sample rate can be matched at the other engine's current bit depth and channel count. If not,
			//		3. Both engines should return to their default settings, which had better be compatible.

			if (
						kIOReturnSuccess == mConfigDictionary->getAltSettingWithSettings(
																	&altSetting,
																	thisStream->mInterfaceNumber,
																	thisFormat->fNumChannels,
																	thisFormat->fBitDepth,
																	otherSampleRate->whole
															   )
			   )
			{
				// We'll change this engine only.
				formatChangeReturnCode = thisAudioEngine->controlledFormatChange(NULL, NULL, otherSampleRate);
				if (kIOReturnSuccess == formatChangeReturnCode)
				{
					DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - This engine (%p) sample rate changed successfully to %u.",
						this, thisAudioEngine, otherSampleRate->whole);

					result = kPdjUsbFormatChangeForced;

					thisAudioEngine->hardwareSampleRateChanged(otherSampleRate);
				}
			}
			else if (
						kIOReturnSuccess == mConfigDictionary->getAltSettingWithSettings(
																	&altSetting,
																	otherStream->mInterfaceNumber,
																	otherFormat->fNumChannels,
																	otherFormat->fBitDepth,
																	thisSampleRate->whole
															   )
					)
			{
				// We'll change the other engine only.
				formatChangeReturnCode = otherAudioEngine->controlledFormatChange(NULL, NULL, thisSampleRate);
				if (kIOReturnSuccess == formatChangeReturnCode)
				{
					DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - Other engine (%p) sample rate changed succsesfully to %u.",
						this, otherAudioEngine, thisSampleRate->whole);

					result = kPdjUsbFormatChangeForced;

					otherAudioEngine->hardwareSampleRateChanged(thisSampleRate);
				}
			}
			else
			{
				// We'll restore both engines to their default settings.
				DP(DBG_WARNING, "! " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - Restoring both engines to their default settings.", this);

				thisStream->setFormat(thisDefaultAudioStreamFormat, false);

				formatChangeReturnCode = thisAudioEngine->controlledFormatChange(
																thisStream,
																thisDefaultAudioStreamFormat,
																&(thisAudioEngine->mDefaultAudioSampleRate)
														  );
				if (kIOReturnSuccess == formatChangeReturnCode)
				{
					DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - This engine (%p) restored to default settings succsesfully.",
						this, thisAudioEngine);

					result = kPdjUsbFormatChangeForced;

					thisAudioEngine->hardwareSampleRateChanged(&(thisAudioEngine->mDefaultAudioSampleRate));
				}

				otherStream->setFormat(otherDefaultAudioStreamFormat, false);

				formatChangeReturnCode = otherAudioEngine->controlledFormatChange(
																otherStream,
																otherDefaultAudioStreamFormat,
																&(otherAudioEngine->mDefaultAudioSampleRate)
														   );
				if (kIOReturnSuccess == formatChangeReturnCode)
				{
					DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - Other engine (%p) restored to default settings succsesfully.",
						this, otherAudioEngine);

					otherAudioEngine->hardwareSampleRateChanged(&(otherAudioEngine->mDefaultAudioSampleRate));
				}
				else
				{
					result = kPdjUsbFormatChangeForceFailure;
				}
			}
		} // Emergency format change case
		else
		{
			// This device is already known to be a single-sample rate device. If the sample rate is changing, change it for both engines or change it for neither.
			if ((newSampleRate) && (newSampleRate->whole != otherSampleRate->whole))
			{
				// See if we can the change the other sample rate at the current format.
				if (
						kIOReturnSuccess == mConfigDictionary->getAltSettingWithSettings(
																	&altSetting,
																	otherStream->mInterfaceNumber,
																	otherFormat->fNumChannels,
																	otherFormat->fBitDepth,
																	newSampleRate->whole
															   )
				   )
				{
					// Issue format changes to both engines.
					thisStream->setFormat(newFormat, false);

					formatChangeReturnCode = thisAudioEngine->controlledFormatChange(thisStream, newFormat, newSampleRate);

					if (kIOReturnSuccess == formatChangeReturnCode)
					{
						result = kPdjUsbFormatChangeForced;

						thisAudioEngine->hardwareSampleRateChanged(newSampleRate);

						formatChangeReturnCode = otherAudioEngine->controlledFormatChange(otherStream, otherFormat, newSampleRate);

						if (kIOReturnSuccess != formatChangeReturnCode)
						{
							result = kPdjUsbFormatChangeForceFailure;
						}
						else
						{
							otherAudioEngine->hardwareSampleRateChanged(newSampleRate);
						}
					}
				}
				else
				{
					// Fail this request.
					DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - Other audio engine (%p) does not support sample rate %u at %u bit %u channel(s). Failing.",
						this, otherAudioEngine, newSampleRate->whole, otherFormat->fBitDepth, otherFormat->fNumChannels);
					result = kPdjUsbFormatChangeForceFailure;
				}
			}
			else
			{
				// The sample rate isn't changing, so process this request normally.
				result = kPdjUsbFormatChangeNormal;

				thisStream->setFormat(newFormat, false);

				formatChangeReturnCode = thisAudioEngine->controlledFormatChange(audioStream, newFormat, newSampleRate);
				// [rdar://5204813] Sending a null pointer here can cause a kernel panic. This should be unnecessary since the sample rate didn't change.
				// thisAudioEngine->hardwareSampleRateChanged(newSampleRate);
			}
		}

		FailIf(NULL == thisAudioEngine,  Exit);
		FailIf(NULL == otherAudioEngine, Exit);

		thisStream  = thisAudioEngine->mMainOutputStream  ? thisAudioEngine->mMainOutputStream  : thisAudioEngine->mMainInputStream;
		otherStream = otherAudioEngine->mMainOutputStream ? otherAudioEngine->mMainOutputStream : otherAudioEngine->mMainInputStream;
		FailIf(NULL == thisStream,  Exit);
		FailIf(NULL == otherStream, Exit);

		thisFormat  = thisStream->getFormat();
		otherFormat = otherStream->getFormat();
		FailIf(NULL == thisFormat,  Exit);
		FailIf(NULL == otherFormat, Exit);

		// Log the new values.
		DP(DBG_INFO, "\n");
		DP(DBG_INFO, "-------------------- AFTER --------------------");
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - engine %p (interface %d, alternateSetting %d) info:",
			this, thisAudioEngine, thisStream->mInterfaceNumber, thisStream->mAlternateSettingID);
		DP(DBG_INFO, "    thisFormat = %p", thisFormat);
		DP(DBG_INFO, "        fNumChannels = %u", thisFormat->fNumChannels);
		DP(DBG_INFO, "        fBitDepth = %d", thisFormat->fBitDepth);
		DP(DBG_INFO, "        fDriverTag = 0x%x", thisFormat->fDriverTag);
		DP(DBG_INFO, "    thisSampleRate->whole = %u", thisSampleRate->whole);
		DP(DBG_INFO, "\n");
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - engine %p (interface %d, alternateSetting %d) info:",
			this, otherAudioEngine, otherStream->mInterfaceNumber, otherStream->mAlternateSettingID);
		DP(DBG_INFO, "    otherFormat = %p", otherFormat);
		DP(DBG_INFO, "        fNumChannels = %u", otherFormat->fNumChannels);
		DP(DBG_INFO, "        fBitDepth = %d", otherFormat->fBitDepth);
		DP(DBG_INFO, "        fDriverTag = 0x%x", otherFormat->fDriverTag);
		DP(DBG_INFO, "    otherSampleRate->whole = %u", otherSampleRate->whole);
		DP(DBG_INFO, "-----------------------------------------------");
		DP(DBG_INFO, "\n");
	}
	else
	{
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - Attempting normal format change request.", this);

		thisAudioEngine->pauseAudioEngine();	// <rdar://7535236>

		// <rdar://7298196> Lock all audio streams for I/O since the format changes deallocate the sample buffer &
		// related structure. This is to prevent a race condition between the format changes and performClientIO()/USB
		// callback completion routines.
		thisAudioEngine->lockAllStreams();

		result = kPdjUsbFormatChangeNormal;

		// Issue this format change request.
		formatChangeReturnCode = thisAudioEngine->controlledFormatChange(audioStream, newFormat, newSampleRate);

		// <rdar://7298196> Unlock the audio streams for I/O.
		thisAudioEngine->unlockAllStreams();

		thisAudioEngine->resumeAudioEngine();	// <rdar://7535236>
	}

	if (kIOReturnSuccess != formatChangeReturnCode)
	{
		result = kPdjUsbFormatChangeError;

		DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - This format change failed with error 0x%x.", this, result);
	}
	else
	{
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - This format change was successful.", this);

		if ((kPdjUsbFormatChangeNormal != result) && (kPdjUsbFormatChangeForced != result))
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::formatChangeController() - Forced format change failed.", this);
			result = kPdjUsbFormatChangeForceFailure;
		}
	}

Exit:
	if (enginesPaused)
	{
		// <rdar://7298196> Unlock the audio streams for I/O.
		otherAudioEngine->unlockAllStreams();
		thisAudioEngine->unlockAllStreams();

		thisAudioEngine->resumeAudioEngine();
		otherAudioEngine->resumeAudioEngine();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::formatChangeController(%p, %p, %p, %p) = 0x%x",
		this, audioEngine, audioStream, newFormat, newSampleRate, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	USB Interface を検索．
 *
 * @param[in]		*request		検索条件．
 *
 * @return			USB interface．
 */
#ifdef PROVIDER_USB_DEVICE
IOUSBHostInterface *
PdjUsbAudioDevice::findInterface(
	IOUSBFindInterfaceRequest	*request
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "::findInterface()");

	OSIterator			*iterator		= mUSBDevice->getChildIterator(gIOServicePlane);
	OSObject			*candidate		= NULL;
	IOUSBHostInterface	*foundInterface	= NULL;

	while (iterator && ((candidate = iterator->getNextObject()) != NULL))
	{
		IOUSBHostInterface	*interfaceCandidate = OSDynamicCast(IOUSBHostInterface, candidate);

		if (   (interfaceCandidate != NULL)
			&& (interfaceCandidate->getInterfaceDescriptor()->bInterfaceClass    == request->bInterfaceClass   )
			&& (interfaceCandidate->getInterfaceDescriptor()->bInterfaceSubClass == request->bInterfaceSubClass))
		{
			foundInterface = interfaceCandidate;
			break;
		}
	}

	OSSafeReleaseNULL(iterator);

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "::findInterface()");
	return foundInterface;
}
#endif

/****************************************************************************/
/**
 * @brief	USB Pipe を検索．
 *
 * @param[in]		*interface		検索対象の USB interface．
 * @param[in]		*request		検索条件．
 *
 * @return			USB pipe．
 */
IOUSBHostPipe *
PdjUsbAudioDevice::findPipe(
	IOUSBHostInterface			*interface,
	IOUSBFindEndpointRequest	*request
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "::findPipe()");

	const StandardUSB::EndpointDescriptor	*endpointCandidate	= NULL;
	IOUSBHostPipe							*foundPipe			= NULL;

	while (
		(
			endpointCandidate = (const StandardUSB::EndpointDescriptor *)
								StandardUSB::getNextAssociatedDescriptorWithType(
													interface->getConfigurationDescriptor(),
													interface->getInterfaceDescriptor(),
													endpointCandidate,
													kDescriptorTypeEndpoint
											 )
		) != NULL
	)
	{
		if (
				(
					((endpointCandidate->bmAttributes & kEndpointDescriptorTransferType) >> kIOUSBEndpointDescriptorTransferTypePhase)
					== request->type
				)
			&&
				(
					((endpointCandidate->bEndpointAddress & kEndpointDescriptorDirection) >> kIOUSBEndpointDescriptorDirectionPhase)
					== request->direction
				)
		   )
		{
			foundPipe = interface->copyPipe(StandardUSB::getEndpointAddress(endpointCandidate));
			break;
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "::findPipe()");
	return foundPipe;
}

/****************************************************************************/
/**
 * @brief	ハードウェア初期化．
 *			IOAudioDevice::initHardware() をオーバーライド．
 *			super::start() から呼び出される．
 *			この関数から return 時には、サービスを提供する準備ができていなければならない．
 *
 * @param[in]		*provider		ハードウェアへのアクセスを提供するサービスプロバイダ nub．
 *									PROVIDER_USB_DEVICE が定義されていれば IOUSBHostDevice，
 *									未定義なら IOUSBHostInterface．
 *
 * @retval			true			初期化成功．
 * @retval			false			初期化失敗．デバイスは認識されない．
 */
// <rdar://problem/6686515>
bool
PdjUsbAudioDevice::initHardware(
	IOService		*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::initHardware(%p)", this, provider);

	bool	result = false;

	FailIf(NULL == mInitHardwareThread, Exit);

	// <rdar://6223517> Retain a reference in case the PdjUsbAudioDevice object is release while it is trying to execute the thread.
	this->retain();

	if (::thread_call_enter1(mInitHardwareThread, (void *)provider))
	{
		// <rdar://6223517> Release the reference if the thread is already scheduled.
		this->release();
	}

	result = true;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::initHardware(%p) = %d", this, provider, result);
	return result;
}

#pragma mark - Protected Methods -

/****************************************************************************/
/**
 * @brief	ハードウェア初期化［スレッド］．
 *
 * @param[in]		*audioDevice	PdjUsbAudioDevice クラスのインスタンス．
 * @param[in]		*provider		ハードウェアへのアクセスを提供するサービスプロバイダ nub．
 *									PROVIDER_USB_DEVICE が定義されていれば IOUSBHostDevice，
 *									未定義なら IOUSBHostInterface．
 */
// <rdar://problem/6686515>
void
PdjUsbAudioDevice::initHardwareThread(
	PdjUsbAudioDevice		*audioDevice,
	IOService				*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "::initHardwareThread(%p, %p)", audioDevice, provider);

	IOCommandGate	*cg;

	FailIf(NULL == audioDevice, Exit);
//	FailIf(audioDevice->mTerminating, Exit);

	cg = audioDevice->getCommandGate();
	if (cg)
	{
		if (kIOReturnSuccess != cg->runAction(audioDevice->initHardwareThreadAction, (void *)provider))
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "::initHardwareThread(%p, %p) - Error on IOCommandGate::runAction().", audioDevice, provider);
		}
	}

	audioDevice->release();	// <rdar://6223517> Release the reference that were retained before thread_call_enter1().

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "::initHardwareThread(%p, %p)", audioDevice, provider);
	return;
}

/****************************************************************************/
/**
 * @brief	ハードウェア初期化［コマンドゲート］．
 *
 * @param[in]		*owner			PdjUsbAudioDevice クラスのインスタンス．
 * @param[in]		*provider		ハードウェアへのアクセスを提供するサービスプロバイダ nub．
 *									PROVIDER_USB_DEVICE が定義されていれば IOUSBHostDevice，
 *									未定義なら IOUSBHostInterface．
 * @param[in]		*arg2			未使用．
 * @param[in]		*arg3			未使用．
 * @param[in]		*arg4			未使用．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDevice::initHardwareThreadAction(
	OSObject		*owner,
	void			*provider,
	void			*arg2,
	void			*arg3,
	void			*arg4
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "::initHardwareThreadAction(%p, %p)", owner, provider);

	IOReturn				result		 = kIOReturnError;
	PdjUsbAudioDevice		*audioDevice = (PdjUsbAudioDevice *)owner;

	FailIf(NULL == audioDevice, Exit);

	result = audioDevice->protectedInitHardware((IOService *)provider);

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "::initHardwareThreadAction(%p, %p) = %d", owner, provider, result);

	return result;
}

/****************************************************************************/
/**
 * @brief	ハードウェア初期化［処理本体］．
 *
 * @param[in]		*provider		ハードウェアへのアクセスを提供するサービスプロバイダ nub．
 *									PROVIDER_USB_DEVICE が定義されていれば IOUSBHostDevice，
 *									未定義なら IOUSBHostInterface．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDevice::protectedInitHardware(
	IOService		*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware(%p)", this, provider);

	IOReturn		err;
	bool			resultCode = false;
#ifdef TARGET_PDJ
#else
	OSArray			*streamNumberArray;
	UInt8			numStreamInterfaces;
	UInt8			numStreams;
#endif
	char			string[kStringBufferSize];
	UInt8			stringIndex;
	OSObject		*nameObject = NULL;
	OSString		*nameString = NULL;
	OSString		*localizedBundle = NULL;
	OSArray			*controlGraph;
#ifdef TARGET_PDJ
 	UInt8			speed;
#endif

	FailIf(NULL == mControlInterface, Exit);

#ifdef TARGET_PDJ
 	speed = this->getDeviceSpeed();
 	FailIf((speed != kUSBDeviceSpeedHigh) && (speed != kUSBDeviceSpeedFull), Exit);
#endif

	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - %d configuration(s) on this device. This control interface number is %d",
		this, mControlInterface->getDevice()->getDeviceDescriptor()->bNumConfigurations, mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);

	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - Attempting to create configuration dictionary...", this);
	mConfigDictionary = PdjUsbConfigurationDictionary::create(this->getConfigurationDescriptor(), mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);	// rdar://5495653
	FailIf(NULL == mConfigDictionary, Exit);
	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - Successfully created configuration dictionary.", this);

#ifdef TARGET_PDJ
#else
	if (!mConfigDictionary->hasAudioStreamingInterfaces())
	{
		// Bail since here is nothing related to audio streaming.
		DP(DBG_WARNING, "! " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - No audio streaming interfaces in configuration dictionary.", this);
		goto Exit;
	}
#endif

	mTopology = PdjUsbAudioTopology::create(this);
	FailIf(NULL == mTopology, Exit);

	// <rdar://problem/7378275>
	mTimeLock = IOLockAlloc();
	FailIf(NULL == mTimeLock, Exit);

	mInterfaceLock = IORecursiveLockAlloc();
	FailIf(NULL == mInterfaceLock, Exit);

	// <rdar://problem/6369110> Occassional kernel panic when multiple threads access mRegisteredEngines array simultaneously
	mRegisteredEnginesMutex = IORecursiveLockAlloc();
	FailIf(NULL == mRegisteredEnginesMutex, Exit);

	// <rdar://problem/6420832> Serialize access to mRegisteredStreams
	mRegisteredStreamsMutex = IORecursiveLockAlloc();
	FailIf(NULL == mRegisteredStreamsMutex, Exit);

	controlGraph = mTopology->buildConnectionGraph(mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
	FailIf(NULL == controlGraph, Exit);
#ifdef TARGET_PDJ
#else
	FailIf(0 == controlGraph->getCount(), Exit);
#endif

	// [rdar://4801032]
	if (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	{
		FailIf(NULL == mTopology->buildClockGraph(mControlInterface->getInterfaceDescriptor()->bInterfaceNumber), Exit);
		// From this moment forward, we may assume that a device is attempting to be USB 2.0 audio class-compliant by the presence of mClockGraph
		// Since supported sample rates are no longer listed explicitly in the USB 2.0 audio specification, we must discover them through device request inquiries.
		FailIf(kIOReturnSuccess != mTopology->addSampleRatesFromClockSpace(), Exit);
	}

#ifdef TARGET_PDJ
#else
	// Check to make sure that the control interface we loaded against has audio streaming interfaces and not just MIDI.
	FailIf(kIOReturnSuccess != mConfigDictionary->getControlledStreamNumbers(&streamNumberArray, &numStreams), Exit);
	FailIf(kIOReturnSuccess != mConfigDictionary->getNumStreamInterfaces(&numStreamInterfaces), Exit);
	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - %d controlled stream(s). %d stream interface(s).", this, numStreams, numStreamInterfaces);
	FailIf(0 == numStreams, Exit);
#endif

	// [rdar://problem/5797761]
	// Try to get the name to give this control. We will attempt to name it by the following order:
	// 1. "USB Interface Name" for this control interface
	// 2. Get the "InterfaceStringIndex" and GetStringDescriptor and use that string
	// 3. Get the device's "USB Product Name" and use that
	err = kIOReturnError;

	string[0] = '\0';

	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - Trying to retrieve the USB Interface Name... ", this);

	nameObject = mControlInterface->getProperty("USB Interface Name");
	if (nameObject)
	{
		if (NULL != (nameString = OSDynamicCast(OSString, nameObject)))
		{
			DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - Retrieved product name %s from registry", this, nameString->getCStringNoCopy());
			::strncpy(string, nameString->getCStringNoCopy(), kStringBufferSize);
			err = kIOReturnSuccess;
		}
	}
	else
	{
		stringIndex = mControlInterface->getInterfaceDescriptor()->iInterface;
		if (0 != stringIndex)
		{
			err = this->getStringDescriptor(stringIndex, string, kStringBufferSize);
			// [rdar://3886272] - This could fail on some devices, so we should retry
			if (kIOReturnSuccess != err)
			{
				// Try a regular retry once
				DP(DBG_WARNING, "! " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - couldn't get string descriptor. Retrying ...", this);
				err = this->getStringDescriptor(stringIndex, string, kStringBufferSize);
			}
			if (kIOReturnSuccess != err)
			{
				// Reset the device and try one last time
				DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - Still couldn't get string descriptor. Resetting device ...", this);
//				mControlInterface->getDevice()->ResetDevice();	// Doesn't matter if this fails
				IOSleep(50);									// Give the device 50 ms to get ready
				DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - Last retry ...", this);
				err = this->getStringDescriptor(stringIndex, string, kStringBufferSize);
			}
		}
		else
		{
			// [rdar://5159683] This device may not have either string, so read it from the IOUSBHostDevice.
			nameString = NULL;
			DP(DBG_INFO, "! " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - Trying to retrieve the product name from the IOUSBHostDevice ... ", this);
			nameObject = mControlInterface->getDevice()->getProperty("USB Product Name");
			if (nameObject)
			{
				if (NULL != (nameString = OSDynamicCast(OSString, nameObject)))
				{
					DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - Retrieved product name %s from registry", this, nameString->getCStringNoCopy());
					::strncpy(string, nameString->getCStringNoCopy(), kStringBufferSize);
					err = kIOReturnSuccess;
				}
			}
			else
			{
				// If there is no USB Product Name, there is no ProductString index, so just return an error.
				DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - There was no USB Product Name", this);
				err = kIOReturnBadArgument;
			}
		}
	}

	if ('\0' == string[0] || kIOReturnSuccess != err)
	{
		::strncpy(string, "Unknown USB Audio Device", kStringBufferSize);
	}

	super::setDeviceName(string);

	// <rdar://problem/6398888> Attempt to retrieve the manufacturer name from registry before reading it from the device
	err = kIOReturnError;
	string[0] = '\0';

	nameObject = mControlInterface->getDevice()->getProperty("USB Vendor Name");
	if (nameObject)
	{
		if (NULL != (nameString = OSDynamicCast(OSString, nameObject)))
		{
			DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - Retrieved vendor name %s from registry", this, nameString->getCStringNoCopy());
			::strncpy(string, nameString->getCStringNoCopy(), kStringBufferSize);
			err = kIOReturnSuccess;
		}
	}
	else
	{
		stringIndex = mControlInterface->getDevice()->getDeviceDescriptor()->iManufacturer;
		if (0 != stringIndex)
		{
			err = this->getStringDescriptor(stringIndex, string, kStringBufferSize);
		}
	}

	if (0 == string[0] || kIOReturnSuccess != err)
	{
		::strncpy(string, "Unknown Manufacturer", kStringBufferSize);
	}

	super::setManufacturerName(string);

	super::setDeviceTransportType(kIOAudioDeviceTransportTypeUSB);

	// <rdar://problem/5797761> <rdar://problem/5797877>
	// If our provider has a kIOAudioDeviceLocalizedBundleKey, copy it to our entry.
	// This allows us to localize a string by using a codeless kext to specify the bundle
	// Try to get the localized kext bundle. We will attempt to name it by the following order:
	// 1. kIOAudioDeviceLocalizedBundleKey for this control interface
	// 2. Get the device's kIOAudioDeviceLocalizedBundleKey and use that
	localizedBundle = OSDynamicCast(OSString, mControlInterface->getProperty(kIOAudioDeviceLocalizedBundleKey));
	if (NULL == localizedBundle)
	{
		localizedBundle = OSDynamicCast(OSString, mControlInterface->getDevice()->getProperty(kIOAudioDeviceLocalizedBundleKey));
	}

	if (NULL != localizedBundle)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - setting kIOAudioDeviceLocalizedBundleKey property to %s",
			this, localizedBundle->getCStringNoCopy());

		IORegistryEntry::setProperty(kIOAudioDeviceLocalizedBundleKey, localizedBundle);
	}
#if LOCALIZABLE
	else
	{
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - setting kIOAudioDeviceLocalizedBundleKey property to DJM-V10 USBAudio.kext", this);
		IORegistryEntry::setProperty(kIOAudioDeviceLocalizedBundleKey, "DJM-V10 USBAudio.kext");
	}
#endif

	// Create the audio engines
	resultCode = this->createAudioEngines();
	FailIf(!resultCode, Exit);	// <rdar://problem/6576824> PdjUsbAudio fails to unload properly

	resultCode = this->activateAudioEngines();
	FailIf(!resultCode, Exit);	// <rdar://problem/6576824> PdjUsbAudio fails to unload properly

	resultCode = super::initHardware(provider);
	FailIf(!resultCode, Exit);	// <rdar://problem/6576824> PdjUsbAudio fails to unload properly

	// Start the anchored time stamp timer if necessary

	// Initialize time stamp member variables
	mLastUSBFrame			= 0ull;
	mLastWallTime_nanos		= 0ull;

	// Initialize mWallTimePerUSBCycle
	this->resetRateTimer();

	// We should get a new anchor immediately.
	this->updateUSBCycleTime();	// <rdar://problem/7378275>, <rdar://problem/7666699>

	// [rdar://] This member variable keeps track an engine that has had a catastrophic failure that requires an emergency format change.
	mFailingAudioEngine = NULL;

	// Register and start update timer <rdar://problem/7378275>
	mTimerCallCount		= 0;
	mRampUpdateCounter	= 0;	// <rdar://problem/7666699>

	mUpdateTimer = IOTimerEventSource::timerEventSource(this, timerAction);
	FailIf(NULL == mUpdateTimer, Exit);

	workLoop->addEventSource(mUpdateTimer);

	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware() - starting rate timer", this);
	this->timerAction(this, mUpdateTimer);

	// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
	mProcessStatusInterruptThread = ::thread_call_allocate((thread_call_func_t)processStatusInterrupt, (thread_call_param_t)this);
	FailIf(NULL == mProcessStatusInterruptThread, Exit);

	this->checkForStatusInterruptEndpoint();

#ifdef TARGET_PDJ
 	//mSleepWakeNotifier = ::registerPrioritySleepWakeInterest(sleepWakeHandler, this, 0);
#endif

	// Added for rdar://3993906 . This forces matchPropertyTable() to run again.
	// <rdar://7295322> Asynchronous to prevent deadlock if the device or interface is terminated while
	// registerService() is performing matching.
	IOService::registerService(kIOServiceAsynchronous);

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::protectedInitHardware(%p) = %d", this, provider, (resultCode ? kIOReturnSuccess : kIOReturnError));
	return (resultCode ? kIOReturnSuccess : kIOReturnError);	// <rdar://problem/6576824>
}

/****************************************************************************/
/**
 * @brief	コンフィグレーションディスクリプタ取得．
 *
 * @return			コンフィグレーションディスクリプタ．
 */
// added for rdar://5495653. Returns the current configuration descriptor.
const StandardUSB::ConfigurationDescriptor *
PdjUsbAudioDevice::getConfigurationDescriptor(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::getConfigurationDescriptor()", this);

	IOUSBHostDevice								*usbDevice;
	const StandardUSB::ConfigurationDescriptor	*configDescriptor = NULL;
	UInt8										numConfigs = 0;

	FailIf(NULL == mControlInterface, Exit);

	usbDevice = mControlInterface->getDevice();
	FailIf(NULL == usbDevice, Exit);

	numConfigs = usbDevice->getDeviceDescriptor()->bNumConfigurations;
	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::getConfigurationDescriptor() - numConfigs = %d", this, numConfigs);

	if (1 == numConfigs)
	{
		configDescriptor = usbDevice->getConfigurationDescriptor(0);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::getConfigurationDescriptor() - configDescriptor = %p", this, configDescriptor);
	return configDescriptor;
}

/****************************************************************************/
/**
 * @brief	すべてのオーディオエンジンを生成．
 *
 * @retval			true			成功．
 * @retval			false			失敗．
 */
bool
PdjUsbAudioDevice::createAudioEngines(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::createAudioEngines()", this);

	bool			result = false;

#ifdef TARGET_PDJ
 	{
 		unsigned long long	interfaceNumber;
 		OSNumber			*streamInterfaceNumber;
 		OSArray				*streamInterfaceNumberArray = OSArray::withCapacity(1);
 
 		FailIf(NULL == streamInterfaceNumberArray, Exit);
 
 		interfaceNumber = 1;
 		streamInterfaceNumber = OSNumber::withNumber(interfaceNumber, 8);
 		streamInterfaceNumberArray->setObject(streamInterfaceNumber);
 		streamInterfaceNumber->release();
 
 		interfaceNumber = 2;
 		streamInterfaceNumber = OSNumber::withNumber(interfaceNumber, 8);
 		streamInterfaceNumberArray->setObject(streamInterfaceNumber);
 		streamInterfaceNumber->release();
 
 		result = this->createAudioEngine(streamInterfaceNumberArray);
 		FailWithAction(!result, streamInterfaceNumberArray->release(), Exit);
 		streamInterfaceNumberArray->release();
 	}
#else
	OSArray			*streamNumberArray;
	UInt8			numStreams;
	OSNumber		*streamInterfaceNumber;
	UInt8			numStreamInterfaces;
	OSBoolean		*useSingleAudioEngine = NULL;

	// Find USB interfaces.
	FailIf(NULL == mControlInterface, Exit);
	FailIf(kIOReturnSuccess != mConfigDictionary->getControlledStreamNumbers(&streamNumberArray, &numStreams), Exit);
	FailIf(kIOReturnSuccess != mConfigDictionary->getNumStreamInterfaces(&numStreamInterfaces), Exit);

	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::createAudioEngines() - streamNumberArray = %p, numStreams = %u, numStreamInterfaces = %u",
		this, streamNumberArray, numStreams, numStreamInterfaces);

	// The override kext property to turn on/off single audio engine behavior. This takes priority over other mechanism to
	// determine if the device can support single audio engine.
	useSingleAudioEngine = OSDynamicCast(OSBoolean, mControlInterface->getProperty("UseSingleAudioEngine"));

	if (NULL != useSingleAudioEngine)
	{
		if (useSingleAudioEngine->isTrue())
		{
			DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::createAudioEngines() - Forced (via override kext) single audio engine", this);
			FailMessage(!(result = this->createAudioEngine(streamNumberArray)));
		}
		else
		{
			DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::createAudioEngines() - Forced (via override kext) separate audio engines", this);

			for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < numStreamInterfaces; streamInterfaceIndex++)
			{
				FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, streamNumberArray->getObject(streamInterfaceIndex))), Exit);

				OSArray		*streamInterfaceNumberArray = OSArray::withCapacity(1);
				FailIf(NULL == streamInterfaceNumberArray, Exit);

				streamInterfaceNumberArray->setObject(streamInterfaceNumber);

				result = this->createAudioEngine(streamInterfaceNumberArray);
				FailWithAction(!result, streamInterfaceNumberArray->release(), Exit);

				streamInterfaceNumberArray->release();
			}
		}
	}
	else
	{
		// If this device is a known single sample rate device, then we should use a single audio engine.
		if (this->getSingleSampleRateDevice())
		{
			DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::createAudioEngines() - Known single sample rate device", this);

			FailMessage(!(result = this->createAudioEngine(streamNumberArray)));
		}
		else
		{
			OSArray		*commonSampleRatesStreamList = NULL;

			OSArray		*availableStreamsList = OSArray::withArray(streamNumberArray);
			FailIf(NULL == availableStreamsList, Exit);

			// Find all streams with common sample rates.
			while (NULL != (commonSampleRatesStreamList = mTopology->findStreamsWithCommonSampleRates(availableStreamsList)))
			{
				OSArray		*compatibleEndpointsStreamList = NULL;

				// Find all streams with compatible endpoints.
				while (NULL != (compatibleEndpointsStreamList = mTopology->findStreamsWithCompatibleEndpoints(commonSampleRatesStreamList)))
				{
					// If it is USB Audio 2.0 devices, then find all streams with common clock.
					if (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
					{
						OSArray		*commonClockStreamList = NULL;

						// Find all streams with common clock.
						while (NULL != (commonClockStreamList = mTopology->findStreamsWithCommonClocks(compatibleEndpointsStreamList)))
						{
							DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::createAudioEngines() - USB Audio 2.0 device", this);

							FailMessage(!(result = this->createAudioEngine(commonClockStreamList)));
							commonClockStreamList->release();

							if (!result)
							{
								break;
							}
						}
					}
					else
					{
						DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::createAudioEngines() - USB Audio 1.0 device", this);

						FailMessage(!(result = this->createAudioEngine(compatibleEndpointsStreamList)));
					}

					compatibleEndpointsStreamList->release();
					if (!result)
					{
						break;
					}
				}

				commonSampleRatesStreamList->release();
				if (!result)
				{
					break;
				}
			}

			availableStreamsList->release();
		}
	}
#endif

Exit:
	// <rdar://problem/6892754> 10.5.7 Regression: Devices with unsupported formats stopped working
	// If all the streams have unsupported formats and no engines are created, then return an error.
	if (0 == mEngineArray)
	{
		result = false;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::createAudioEngines() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	１つのオーディオエンジンを生成．
 *
 * @param[in]		*streamInterfaceNumberArray		AudioStreaming Interface の interface 番号の OSArray．
 *
 * @retval			true			成功．
 * @retval			false			失敗．
 */
bool
PdjUsbAudioDevice::createAudioEngine(
	OSArray			*streamInterfaceNumberArray
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::createAudioEngine()", this);

	bool					result = false;
	PdjUsbAudioEngine		*engine = NULL;
	OSArray					*newStreamInterfaceArray = NULL;	// <rdar://problem/6892754>
#ifdef TARGET_PDJ
#else
	bool					formatSupported;					// <rdar://problem/6892754>
#endif

	FailIf(NULL == streamInterfaceNumberArray, Exit);

	newStreamInterfaceArray = OSArray::withCapacity(streamInterfaceNumberArray->getCount());
	FailIf(NULL == newStreamInterfaceArray, Exit);

	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::createAudioEngine() - create audio engine with stream interfaces:", this);

	// <rdar://problem/6892754> 10.5.7 Regression: Devices with unsupported formats stopped working
	for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < streamInterfaceNumberArray->getCount(); streamInterfaceIndex++)
	{
		OSNumber	*streamInterfaceNumber = OSDynamicCast(OSNumber, streamInterfaceNumberArray->getObject(streamInterfaceIndex));

		if (NULL != streamInterfaceNumber)
		{
			DP(DBG_INFO, "--> #%u", streamInterfaceNumber->unsigned8BitValue());

#ifdef TARGET_PDJ
#else
			// <rdar://problem/6892754>
			// In the case that anything goes wrong while checking the formats,
			// maintain the original behavior and add the interface to the list regardless of the format check.
			if (
					(kIOReturnSuccess != this->interfaceContainsSupportedFormat(streamInterfaceNumber->unsigned8BitValue(), &formatSupported))
			   ||
					formatSupported
			   )
#endif
			{
				newStreamInterfaceArray->setObject(streamInterfaceNumber);
			}
		}
		else
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::createAudioEngine() - could not get streamInterfaceNumber", this);
		}
	}

	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::createAudioEngine() - final new array count: %d", this, newStreamInterfaceArray->getCount());

	// <rdar://problem/6892754>
	// Create an engine with the new stream interface array that contains interfaces with supported format types
	if (newStreamInterfaceArray->getCount() > 0)
	{
		if (NULL == mEngineArray)
		{
			mEngineArray = OSArray::withCapacity(1);
			FailIf(NULL == mEngineArray, Exit);
		}

		engine = OSTypeAlloc(PdjUsbAudioEngine);
		FailIf(NULL == engine, Exit);

		if (engine->init(newStreamInterfaceArray))
		{
			// <rdar://problem/7349398>
			OSNumber	*displayRoutingRead = NULL;

			if (NULL != mControlInterface)
			{
				displayRoutingRead = OSDynamicCast(OSNumber, mControlInterface->getDevice()->getProperty(kDisplayRoutingPropertyKey));
			}

			if (NULL != displayRoutingRead)
			{
				UInt32		displayRoutingValue	 = 0;
				OSNumber	*displayRoutingWrite = NULL;

				displayRoutingValue = displayRoutingRead->unsigned32BitValue();

				if (0 == displayRoutingValue)	// 0 is an invalid value
				{
					displayRoutingValue = 1;
				}

				displayRoutingWrite = OSNumber::withNumber(displayRoutingValue, 32);
				if (NULL != displayRoutingWrite)
				{
					engine->setProperty(kDisplayRoutingPropertyKey, displayRoutingWrite);
					displayRoutingWrite->release();
				}
			}

			if (!mEngineArray->setObject(engine))
			{
				engine->release();
			}
		}
		else
		{
			engine->release();
		}
	}
	// <rdar://problem/6892754>
	// else there are no streams with supported format types:
	// in this case, don't init the engine but still return true as this is not an error

	result = true;

Exit:
	if (newStreamInterfaceArray)
	{
		newStreamInterfaceArray->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::createAudioEngine()", this);
	return result;
}

/****************************************************************************/
/**
 * @brief	すべてのオーディオエンジンをアクティベート．
 *
 * @retval			true			成功．
 * @retval			false			失敗．
 */
bool
PdjUsbAudioDevice::activateAudioEngines(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::activateAudioEngines()", this);

	bool			result = true;

	if (NULL != mEngineArray)
	{
		for (UInt32 engineIndex = 0; engineIndex < mEngineArray->getCount(); engineIndex++)
		{
			PdjUsbAudioEngine	*engine = OSDynamicCast(PdjUsbAudioEngine, mEngineArray->getObject(engineIndex));
			FailWithAction(NULL == engine, result = false, Exit);

			FailWithAction(kIOReturnSuccess != super::activateAudioEngine(engine), result = false, Exit);
		}
	}

Exit:
	// <rdar://problems/6576824&6674310> PdjUsbAudio fails to unload properly
	if (!result)
	{
		super::deactivateAllAudioEngines();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::activateAudioEngines() - %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface に利用可能なオーディオフォーマットが含まれているか調べる．
 *
 * @param[in]		streamInterface		AudioStreaming Interface の interface 番号．
 * @param[out]		*formatSupported	true なら、利用可能なオーディオフォーマットが存在する．
 *										false なら、利用可能なオーディオフォーマットは存在しない．
 *
 * @return			IOReturn コード．
 */
// <rdar://problem/6892754>
// Look at all of alternate settings and make sure that we support at least one of the formats (Type II AC-3 format not supported)
IOReturn
PdjUsbAudioDevice::interfaceContainsSupportedFormat(
	UInt8			streamInterface,
	bool			*formatSupported
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::interfaceContainsSupportedFormat()", this);

	IOReturn		returnCode = kIOReturnSuccess;
	UInt8			numAltSettings;
	UInt16			format;
	bool			startAtZero = true;

	*formatSupported = false;

	FailWithAction(NULL == mConfigDictionary, returnCode = kIOReturnError, Exit);

	startAtZero = mConfigDictionary->alternateSettingZeroCanStream(streamInterface);

	FailIf(kIOReturnSuccess != (returnCode = mConfigDictionary->getNumAltSettings(&numAltSettings, streamInterface)), Exit);

	for (UInt8 altSettingIndex = (startAtZero ? 0 : 1); altSettingIndex < numAltSettings; altSettingIndex++)
	{
		FailIf(kIOReturnSuccess != (returnCode = mConfigDictionary->getFormat(&format, streamInterface, altSettingIndex)), Exit);

		if (AC3 != format)
		{
			*formatSupported = true;
		}
		else
		{
			DP(DBG_WARNING, "WARNING: UsbAudio has detected an unsupported format type: Type II AC-3");
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::interfaceContainsSupportedFormat() streamInterface: %u returnCode: %d formatSupported: %u",
		this, streamInterface, returnCode, *formatSupported);
	return returnCode;
}

/****************************************************************************/
/**
 * @brief	タイマハンドラ．
 *
 * @param[in]		*owner			PdjUsbAudioDevice クラスのインスタンス．
 * @param[in]		*sender			タイマイベントソース．
 */
void
PdjUsbAudioDevice::timerAction(
	OSObject			*owner,
	IOTimerEventSource	*sender
)
{
	DP(DBG_TIMER, "+ " LOGSTRAUDIODEVICE "::timerAction(%p, %p)", owner, sender);

	PdjUsbAudioDevice	*audioDevice = (PdjUsbAudioDevice *)owner;

	FailIf(NULL == audioDevice, Exit);

	audioDevice->doTimerAction(sender);

Exit:
	DP(DBG_TIMER, "- " LOGSTRAUDIODEVICE "::timerAction()");
	return;
}

/****************************************************************************/
/**
 * @brief	タイマハンドラ．
 *
 * @param[in]		*timer		タイマイベントソース．
 */
void
PdjUsbAudioDevice::doTimerAction(
	IOTimerEventSource	*timer
)
{
	DP(DBG_TIMER, "+ " LOGSTRAUDIODEVICE "::doTimerAction(%p)", timer);

	// This method updates our running wall time per USB cycle every kRefreshInterval ms.
	// This timer thread is also used to perform routine watchdog-type events.

	UInt32	curRefreshInterval = kRefreshInterval;	// <rdar://problem/7378275>

	FailIf(NULL == timer, Exit);

	// <rdar://problem/7666699>
	if (mAnchorTime.deviceStart || (!this->allEnginesStopped()))
	{
		this->updateUSBCycleTime();	// <rdar://problem/7378275>
	}

	// <rdar://problem/7378275>, <rdar://problem/7666699> Determine the next timer firing time. When the device first
	// enumerates and when streaming first starts, we sample anchor times at a high frequency until the least squares
	// data array fills up. Once it is full, we reduce to approximately 8 times per second.
	if (mRampUpdateCounter < MAX_ANCHOR_ENTRIES)
	{
		curRefreshInterval = kAnchorSamplingFreq1;
	}
	else
	{
		curRefreshInterval = kRefreshInterval;
	}

	mRampUpdateCounter++;	// <rdar://problem/7666699>

	if (++mTimerCallCount >= (kRefreshInterval / curRefreshInterval))	// Always perform these items at kRefreshInterval
	{
		// Perform any watchdog-type events here.
		if (mFailingAudioEngine)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::doTimerAction() - Detected failing audio engine (%p)! Performing emergency format change.",
				this, mFailingAudioEngine);

			// Attempt to synchronize the input and output sample rates.
			this->formatChangeController(mFailingAudioEngine, NULL, NULL, NULL);

			mFailingAudioEngine = NULL;

			this->setSingleSampleRateDevice(true);
		}
		else if (
					(mEngineToRestart)
				&&
					(mEngineToRestart->mUSBStreamRunning)
				)
		{
			// [rdar://5355808] If PdjUsbAudioStream::coalesceInputSamples() gets more than a USB frame's worth of samples behind,
			// we need to restart the engine.
			DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::doTimerAction() - Restarting engine %p", this, mEngineToRestart);

			mEngineToRestart->pauseAudioEngine();
			mEngineToRestart->resumeAudioEngine();

			mEngineToRestart = NULL;
		}

		if (mShouldAttemptDeviceRecovery)
		{
			this->attemptDeviceRecovery();

			mShouldAttemptDeviceRecovery = false;
		}

		// <rdar://5811247> Run the polled task on the engine.
		if (0 != mEngineArray)
		{
			for (UInt32 engineIndex = 0; engineIndex < mEngineArray->getCount(); engineIndex++)
			{
				PdjUsbAudioEngine	*engine = OSDynamicCast(PdjUsbAudioEngine, mEngineArray->getObject(engineIndex));
				if (engine)
				{
					engine->runPolledTask();
				}
			}
		}

		mTimerCallCount = 0;
	}

	// Schedule the next anchor frame and time update
	if (timer)
	{
		timer->setTimeoutMS(curRefreshInterval);
	}

Exit:
	DP(DBG_TIMER, "- " LOGSTRAUDIODEVICE "::doTimerAction()");
	return;
}

/****************************************************************************/
/**
 * @brief	基準時刻と基準 USB フレーム番号を取得．
 *			タイマハンドラから呼び出され、タイムスタンプ作成の基準として使われる．
 *			USB フレーム番号が変化するまでビジーループで待つので、最大 1mSec ブロックする．
 *
 * @param[out]		*frame			この関数から return したときの USB フレーム番号．
 * @param[out]		*time			この関数から return したときの現在時刻．
 *
 * @return			IOReturn コード．
 */
// <rdar://problem/7378275> Improved timestamp generation accuracy
IOReturn
PdjUsbAudioDevice::getAnchorFrameAndTimeStamp(
	UInt64			*frame,
	AbsoluteTime	*time
)
{
	DP(DBG_TIMER, "+ " LOGSTRAUDIODEVICE "[%p]::getAnchorFrameAndTimeStamp()", this);

	IOReturn	result = this->tryGetAnchorFrameAndTimeStamp(frame, time);
	if (result)
	{
		DP(DBG_TIMER, "! " LOGSTRAUDIODEVICE "::getAnchorFrameAndTimeStamp() - Retry!");

		result = this->tryGetAnchorFrameAndTimeStamp(frame, time);
		if (result)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "::getAnchorFrameAndTimeStamp() - Retry NG!");
		}
		else
		{
			DP(DBG_TIMER, "! " LOGSTRAUDIODEVICE "::getAnchorFrameAndTimeStamp() - Retry OK!");
		}
	}

	DP(DBG_TIMER, "- " LOGSTRAUDIODEVICE "[%p]::getAnchorFrameAndTimeStamp() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	基準時刻と基準 USB フレーム番号を取得．
 *			タイマハンドラから呼び出され、タイムスタンプ作成の基準として使われる．
 *			USB フレーム番号が変化するまでビジーループで待つので、最大 1mSec ブロックする．
 *
 * @param[out]		*frame			この関数から return したときの USB フレーム番号．
 * @param[out]		*time			この関数から return したときの現在時刻．
 *
 * @return			IOReturn コード．
 */
// <rdar://problem/7378275> Improved timestamp generation accuracy
IOReturn
PdjUsbAudioDevice::tryGetAnchorFrameAndTimeStamp(
	UInt64			*frame,
	AbsoluteTime	*time
)
{
	DP(DBG_TIMER, "+ " LOGSTRAUDIODEVICE "[%p]::tryGetAnchorFrameAndTimeStamp()", this);

	AbsoluteTime	finishTime;
	AbsoluteTime	offset;
	AbsoluteTime	curTime;
	AbsoluteTime	prevTime;
	AbsoluteTime	thisTime;
	UInt64			thisFrame;
	AbsoluteTime	diffAbs;
	UInt64			diffNanos;
	IOReturn		result = kIOReturnError;

	FailIf(NULL == mControlInterface, Exit);

	::nanoseconds_to_absolutetime(1100000, &offset);
	::clock_get_uptime(&finishTime);
	ADD_ABSOLUTETIME(&finishTime, &offset);		// finishTime is when we timeout

	::clock_get_uptime(&curTime);

	FailIf(NULL == mControlInterface->getDevice(), Exit);

	thisFrame = mControlInterface->getDevice()->getFrameNumber();

	// spin until the frame changes
	do
	{
		prevTime = curTime;
		::clock_get_uptime(&curTime);

	} while (
				(!mTerminatingDriver)	// <rdar://problem/7800198>
			&&
				(mControlInterface)
			&&
				(thisFrame == mControlInterface->getDevice()->getFrameNumber())
			&&
				(CMP_ABSOLUTETIME(&finishTime, &curTime) > 0)
			);

	::clock_get_uptime(&thisTime);

	if (CMP_ABSOLUTETIME(&finishTime, &curTime) < 0)	// if we timed out
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "::tryGetAnchorFrameAndTimeStamp() - Time out!!");
		goto Exit;
	}

	diffAbs = curTime;
	SUB_ABSOLUTETIME(&diffAbs, &prevTime);
	::absolutetime_to_nanoseconds(diffAbs, &diffNanos);

	if (diffNanos > kMaxTimeStampJitter)	// this timestamp is an outlier, throw it out
	{
		DP(DBG_TIMER, "! " LOGSTRAUDIODEVICE "::tryGetAnchorFrameAndTimeStamp() - An outlier, throw it out.");
		goto Exit;
	}

	diffAbs = thisTime;
	SUB_ABSOLUTETIME(&diffAbs, &curTime);
	::absolutetime_to_nanoseconds(diffAbs, &diffNanos);

	if (diffNanos > kMaxTimeStampJitter)	// this timestamp is an outlier, throw it out
	{
		DP(DBG_TIMER, "! " LOGSTRAUDIODEVICE "::tryGetAnchorFrameAndTimeStamp() - An outlier, throw it out.");
		goto Exit;
	}

	ADD_ABSOLUTETIME(&thisTime, &curTime);
	AbsoluteTime_to_scalar(&thisTime) /= 2;

	*frame	= ++thisFrame;
	*time	= thisTime;

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TIMER, "- " LOGSTRAUDIODEVICE "[%p]::tryGetAnchorFrameAndTimeStamp() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	基準時刻を更新．
 *
 * @param[in, out]	*anchorTime		基準時刻．
 * @param[in]		X				USB フレーム番号．
 * @param[in]		Y				時刻．
 */
// From http://people.hofstra.edu/stefan_waner/realworld/calctopic1/regression.html:
// The best fit line associated with the n points (x1, y1), (x2, y2), . . . , (xn, yn) has the form
//    y = m *x + b
// slope:
//    m = ( n * sumXY - sumX * sumY ) / ( n * sumXX - sumX * sumX )
//      = P / Q
//    where P = n * sumXY - sumX * sumY
//          Q = n * sumXX - sumX * sumX
// intercept:
//    b = ( sumY - m * sumX ) / n
//
// y = m * x + b
//   = m * x + ( ( sumY - m * sumX ) / n )
//   = ( m * n * x + sumY - m * sumX ) / n
//   = ( m * ( n * x - sumX ) + sumY ) / n
//   = ( ( P / Q ) * ( n * x - sumX ) ) + sumY ) / n
//   = ( P * ( n * x - sumX ) + Q * sumY ) / ( Q * n )
//
// <rdar://problem/7378275> Improved timestamp generation accuracy
void
PdjUsbAudioDevice::updateAnchorTime(
	ANCHORTIME		*anchorTime,
	UInt64			X,
	UInt64			Y
)
{
	DP(DBG_TIMER, "+ " LOGSTRAUDIODEVICE "::updateAnchorTime()");

	UInt32			index = anchorTime->index;

#if DEBUGANCHORS
	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "::updateAnchorTime() - index: %u X: %llu Y: %llu", anchorTime->index, X, Y);
#endif

	if (anchorTime->n < MAX_ANCHOR_ENTRIES)
	{
		anchorTime->X[index]	= X;
		anchorTime->Y[index]	= Y;
		anchorTime->XX[index]	= ::mul64(X, X);
		anchorTime->XY[index]	= ::mul64(X, Y);

		anchorTime->sumX += X;
		anchorTime->sumY += Y;
		anchorTime->sumXX = ::add128(anchorTime->sumXX, anchorTime->XX[index]);
		anchorTime->sumXY = ::add128(anchorTime->sumXY, anchorTime->XY[index]);
		anchorTime->n++;
	}
	else
	{
		anchorTime->sumX		-= anchorTime->X[index];
		anchorTime->X[index]	 = X;
		anchorTime->sumX		+= X;

		anchorTime->sumY		-= anchorTime->Y[index];
		anchorTime->Y[index]	 = Y;
		anchorTime->sumY		+= Y;

		anchorTime->sumXX		 = ::sub128(anchorTime->sumXX, anchorTime->XX[index]);
		anchorTime->XX[index]	 = ::mul64(X, X);
		anchorTime->sumXX		 = ::add128(anchorTime->sumXX, anchorTime->XX[index]);

		anchorTime->sumXY		 = ::sub128(anchorTime->sumXY, anchorTime->XY[index]);
		anchorTime->XY[index]	 = ::mul64(X, Y);
		anchorTime->sumXY		 = ::add128(anchorTime->sumXY, anchorTime->XY[index]);
	}

	if (anchorTime->n > 1)
	{
		U256	nSumXY	 = ::mul128(anchorTime->n, anchorTime->sumXY);
		U128	sumXSumY = ::mul64 (anchorTime->sumX, anchorTime->sumY);

#if (MAX_ANCHOR_ENTRIES <= 1024)
		anchorTime->P = ::sub128(nSumXY.lo, sumXSumY);
#else
		anchorTime->P = ::sub256(nSumXY, sumXSumY);
#endif

		U256	nSumXX	 = ::mul128(anchorTime->n, anchorTime->sumXX);
		U128	sumXSumX = ::mul64 (anchorTime->sumX, anchorTime->sumX);

#if (MAX_ANCHOR_ENTRIES <= 1024)
		anchorTime->Q = ::sub128(nSumXX.lo, sumXSumX);
#else
		anchorTime->Q = ::sub256(nSumXX, sumXSumX);
#endif

#if (MAX_ANCHOR_ENTRIES <= 1024)
		anchorTime->QSumY			= ::mul128(anchorTime->Q, anchorTime->sumY);
		anchorTime->Qn				= ::mul128(anchorTime->Q, anchorTime->n);
		anchorTime->mExtraPrecision	= ::div256(::mul128(anchorTime->P, kWallTimeExtraPrecision), anchorTime->Q).lo;
#else
		anchorTime->QSumY			= ::mul256(anchorTime->Q, anchorTime->sumY).lo;
		anchorTime->Qn				= ::mul256(anchorTime->Q, anchorTime->n).lo;
		anchorTime->mExtraPrecision	= ::div256(::mul256(anchorTime->P, kWallTimeExtraPrecision).lo, anchorTime->Q).lo;
#endif
	}
	else
	{
#if (MAX_ANCHOR_ENTRIES <= 1024)
		anchorTime->P.hi = 0; anchorTime->P.lo = 0;
		anchorTime->Q.hi = 0; anchorTime->Q.lo = 1;
#else
		anchorTime->P.hi.hi = 0; anchorTime->P.hi.lo = 0; anchorTime->P.lo.hi = 0; anchorTime->P.lo.lo = 0;
		anchorTime->Q.hi.hi = 0; anchorTime->Q.hi.lo = 0; anchorTime->Q.lo.hi = 0; anchorTime->Q.lo.lo = 1;
#endif
		anchorTime->mExtraPrecision.hi = anchorTime->mExtraPrecision.lo = 0;
	}

	anchorTime->index++;
	if (anchorTime->index >= MAX_ANCHOR_ENTRIES)
	{
		anchorTime->index = 0;
	}

	DP(DBG_TIMER, "- " LOGSTRAUDIODEVICE "::updateAnchorTime()");
}

/****************************************************************************/
/**
 * @brief	基準時刻が大きくずれた場合のオフセット補正．
 */
// <rdar://problem/7666699>
void
PdjUsbAudioDevice::applyOffsetAmountToFilter(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "::applyOffsetAmountToFilter()");

	UInt64			timeOffset = 0;
	AbsoluteTime	timeStamp;
	UInt64			currentFrame;

	if ((kIOReturnSuccess == this->getAnchorFrameAndTimeStamp(&currentFrame, &timeStamp)) && mTimeLock)
	{
		UInt64		actualTime;

		::absolutetime_to_nanoseconds(timeStamp, &actualTime);

		IOLockLock(mTimeLock);

		// offset = predicted - actual
		// predicted time = obtain from filtered data

		// don't apply offset unless a minimum number of frames have elapsed and there is data in the filter
		if (((currentFrame - this->lastAnchorFrame()) > MIN_FRAMES_APPLY_OFFSET) && (mAnchorTime.n > 0))
		{
			int		numFilterPoints;
			bool	isPositive		= true;
			UInt64	predictedTime	= this->getTimeForFrameNumber(currentFrame);

			if (predictedTime >= actualTime)
			{
				// new offset is positive
				timeOffset = predictedTime - actualTime;
			}
			else
			{
				// new offset is negative
				timeOffset = actualTime - predictedTime;
				isPositive = false;
			}

#if DEBUGANCHORS
			DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "::applyOffsetAmountToFilter timeOffset: %llu framesElapsed: %llu currentFrame: %llu predictedTime: %llu actualTime: %llu",
						timeOffset, currentFrame - lastAnchorFrame(), currentFrame, predictedTime, actualTime);
#endif
			numFilterPoints = mAnchorTime.n;

			// copy old values
			for (int index = 0; index < numFilterPoints; index++)
			{
				Xcopy[index] = mAnchorTime.X[index];
				Ycopy[index] = mAnchorTime.Y[index];
			}

			// clear out all data from filter
			this->resetRateTimer();

			// repopulate filter with old timestamp data with offset applied
			for (int index = 0; index < numFilterPoints; index++)
			{
				this->updateAnchorTime(&mAnchorTime, Xcopy[index], (isPositive ? (Ycopy[index] - timeOffset) : (Ycopy[index] + timeOffset)));
			}
		}

		// add the anchor time obtained to calculate offset to the filter
		this->updateAnchorTime(&mAnchorTime, currentFrame, actualTime);

		IOLockUnlock(mTimeLock);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "::applyOffsetAmountToFilter()");
}

/****************************************************************************/
/**
 * @brief	タイムスタンプ用係数をリセット．
 *			この係数は PC のクロックと USB のクロックの間の調整係数である．
 */
// [rdar://5165798] At initHardware() time and if anything goes wrong, this is how we reset the timer code.
void
PdjUsbAudioDevice::resetRateTimer(
	void
)
{
	mWallTimePerUSBCycle = 1000000ull * kWallTimeExtraPrecision;

	::bzero(&mAnchorTime, sizeof(ANCHORTIME));		// <rdar://problem/7378275>

	mAnchorTime.deviceStart = true;					// <rdar://problem/7666699>
}

/****************************************************************************/
/**
 * @brief	１つ前の基準 USB フレーム番号を取得．
 *
 * @return			１つ前の基準 USB フレーム番号．
 */
// <rdar://problem/7666699>
UInt64
PdjUsbAudioDevice::lastAnchorFrame(
	void
)
{
	return (mAnchorTime.index ? mAnchorTime.X[mAnchorTime.index - 1] : mAnchorTime.X[MAX_ANCHOR_ENTRIES - 1]);
}

#if DEBUGANCHORS
/****************************************************************************/
/**
 * @brief	基準時刻と基準 USB フレーム番号をデバッグログに出力．
 *			タイムスタンプ作成用の基準時刻と基準 USB フレーム番号を
 *			記録しておき、一定数溜まったらデバッグログに出力する．
 *
 * @param[in]		anchorFrame		基準 USB フレーム番号．
 * @param[in]		timeStamp		基準時刻．
 */
void
PdjUsbAudioDevice::accumulateAnchor(
	UInt64			anchorFrame,
	AbsoluteTime	timeStamp
)
{
	if (mAnchorFrames[kAnchorsToAccumulate - 1] == 0ull)
	{
		// Find the last empty space and store this anchor frame and time stamp
		for (UInt32 i = 0; i < kAnchorsToAccumulate; i++)
		{
			if (0 == mAnchorFrames[i])
			{
				mAnchorFrames[i] = anchorFrame;
				mAnchorTimes[i]	 = timeStamp;
				break;
			}
		}
	}

	if (mAnchorFrames[kAnchorsToAccumulate - 1] != 0ull)
	{
		UInt64		time_nanos;

		// Maximum number of anchors has already been accumulated.
		DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "::accumulateAnchor() - Frame # %d accumulated.", kAnchorsToAccumulate);

		for (UInt32 i = 0; i < kAnchorsToAccumulate; i++)
		{
			::absolutetime_to_nanoseconds(mAnchorTimes[i], &time_nanos);

			DP(DBG_INFO, "  - %llu \t %llu", mAnchorFrames[i], time_nanos);
		}

		// Reset data
		for (UInt32 i = 0; i < kAnchorsToAccumulate; i++)
		{
			mAnchorFrames[i] = 0ull;
		}
	}
}
#endif

/****************************************************************************/
/**
 * @brief	すべてのオーディオエンジンが停止しているか確認．
 *
 * @retval			true			すべてのオーディオエンジンは停止している．
 * @retval			false			動作中のオーディオエンジンがある．
 */
// <rdar://problem/7666699>
bool
PdjUsbAudioDevice::allEnginesStopped(
	void
)
{
	DP(DBG_VERBOSE, "+ " LOGSTRAUDIODEVICE "::allEnginesStopped()");

	bool	result = true;

	if (NULL != mEngineArray)
	{
		for (UInt32 engineIndex = 0; engineIndex < mEngineArray->getCount(); engineIndex++)
		{
			PdjUsbAudioEngine	*engine = OSDynamicCast(PdjUsbAudioEngine, mEngineArray->getObject(engineIndex));
			if (NULL != engine)
			{
				if (engine->mUSBStreamRunning)
				{
					result = false;
					break;
				}
			}
		}
	}

	DP(DBG_VERBOSE, "- " LOGSTRAUDIODEVICE "::allEnginesStopped() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	デバイスリセット．
 *			処理内容がコメントアウトされているので何もしない．
 */
void
PdjUsbAudioDevice::attemptDeviceRecovery(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::attemptDeviceRecovery()", this);

	FailIf(NULL == mControlInterface, Exit);
/*
	DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::attemptDeviceRecovery() - Issuing device reset.", this);
	mControlInterface->getDevice()->ResetDevice();
	IOSleep(10);
*/

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::attemptDeviceRecovery()", this);
}

/****************************************************************************/
/**
 * @brief	Interrupt IN 転送起動．
 *			AudioControl Interface に Status Interrupt Endpoint がある場合は IN 転送を起動する．
 */
// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
void
PdjUsbAudioDevice::checkForStatusInterruptEndpoint(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::checkForStatusInterruptEndpoint()", this);

	IOUSBFindEndpointRequest	interruptEndpoint;
	UInt8						endpointAddress;
	UInt8						controlInterfaceNum;
	UInt32						messageLength;

	FailIf(NULL == mControlInterface, Exit);	// <rdar://7085810>

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	// check for interrupt endpoint
	if (
			(mConfigDictionary->hasInterruptEndpoint(controlInterfaceNum, 0))
	   &&
			(kIOReturnSuccess == mConfigDictionary->getInterruptEndpointAddress(&endpointAddress, controlInterfaceNum, 0))
	   )
	{
		// setup pipe
		if (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
		{
			// USB Audio 2.0
			mStatusInterruptBufferType = kInterruptDataMessageFormat;
			messageLength = sizeof(USBAUDIO_0200::InterruptDataMessageFormat);
		}
		else
		{
			// USB Audio 1.0
			mStatusInterruptBufferType = kAudioStatusWordFormat;
			messageLength = sizeof(AudioStatusWordFormat);
		}

		interruptEndpoint.type			= kUSBInterrupt;
		interruptEndpoint.direction		= kUSBIn;	// The associated endpoint always goes "in"
		interruptEndpoint.maxPacketSize	= messageLength;
		interruptEndpoint.interval		= 0xFF;

		mInterruptPipe = PdjUsbAudioDevice::findPipe(mControlInterface, &interruptEndpoint);
		FailIf(NULL == mInterruptPipe, Exit);

		if (NULL == mInterruptEndpointMemoryDescriptor)
		{
			// <rdar://7000283> Remove the use of deprecated IOMallocContiguous(). Use IOBufferMemoryDescriptor to allocate memory instead.
			mInterruptEndpointMemoryDescriptor = IOBufferMemoryDescriptor::withOptions(kIODirectionInOut, messageLength, 8);
			FailIf(NULL == mInterruptEndpointMemoryDescriptor, Exit);

			mStatusInterruptBuffer = mInterruptEndpointMemoryDescriptor->getBytesNoCopy();
			FailIf(NULL == mStatusInterruptBuffer, Exit);

			::bzero(mStatusInterruptBuffer, messageLength);
		}

		mStatusInterruptCompletion.owner	 = (void *)this;
		mStatusInterruptCompletion.action	 = statusInterruptHandler;
		mStatusInterruptCompletion.parameter = NULL;

		mInterruptPipe->retain();

		FailMessage(
			kIOReturnSuccess != mInterruptPipe->io(
													mInterruptEndpointMemoryDescriptor,
													(UInt32)mInterruptEndpointMemoryDescriptor->getLength(),
													&mStatusInterruptCompletion
												)
		);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::checkForStatusInterruptEndpoint()", this);
}

/****************************************************************************/
/**
 * @brief	Interrupt IN 転送完了ルーチン．
 *			AudioControl Interface に Status Interrupt Endpoint がある場合の転送完了ルーチン．
 *
 * @param[in]		*target					IN 転送を起動した PdjUsbAudioDevice のインスタンス．
 * @param[in]		*parameter				IN 転送起動時に設定したパラメータ．未使用．
 * @param[in]		status					IN 転送の完了ステータス．
 * @param[in]		bufferSizeRemaining		転送要求サイズから実際の転送サイズを引いた残り．
 */
void
PdjUsbAudioDevice::statusInterruptHandler(
	void			*target,
	void			*parameter,
	IOReturn		status,
	UInt32			bufferSizeRemaining
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::statusInterruptHandler()", target);

	PdjUsbAudioDevice	*self = (PdjUsbAudioDevice *)target;
	FailIf(NULL == self, Exit);

	// spawn thread to process interrupt
	// <rdar://problem/6622487> Clear stall if detected

	if (kIOReturnAborted == status || self->isInactive())
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::statusInterruptHandler() error from USB: 0x%X or IOService inactive: %u, NOT reposting read to interrupt pipe",
			self, status, self->isInactive());
	}
	else
	{
		if (kIOUSBPipeStalled == status)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::statusInterruptHandler() error from USB: kIOUSBPipeStalled, reposting read to interrupt pipe", self);
			// <rdar://7865729> Defer the clearStall() to our own thread.
			self->mInterruptPipeStalled = true;
		}
		else if (kIOReturnSuccess != status)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::statusInterruptHandler() error from USB: 0x%X, reposting read to interrupt pipe", self, status);
		}

		self->retain();

		if (::thread_call_enter(self->mProcessStatusInterruptThread))
		{
			self->release();
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::statusInterruptHandler()", self);
	return;
}

/****************************************************************************/
/**
 * @brief	Status Interrupt Endpoint からの受信処理［スレッド］．
 *			statusInterruptHandler() から起動されるスレッドである．
 *
 * @param[in]		*arg			スレッドを起動した PdjUsbAudioDevice のインスタンス．
 */
void
PdjUsbAudioDevice::processStatusInterrupt(
	void			*arg
)
{
	PdjUsbAudioDevice		*self;
	IOCommandGate			*cg;

	FailIf(NULL == arg, Exit);

	self = (PdjUsbAudioDevice *)arg;

	cg = self->getCommandGate();
	if (cg)
	{
		cg->runAction(self->runStatusInterruptTask, (void *)0, (void *)0);
	}

	self->release();

Exit:
	return;
}

/****************************************************************************/
/**
 * @brief	Status Interrupt Endpoint からの受信処理［コマンドゲート］．
 *
 * @param[in]		*target			PdjUsbAudioDevice クラスのインスタンス．
 * @param[in]		*arg0			未使用．
 * @param[in]		*arg1			未使用．
 * @param[in]		*arg2			未使用．
 * @param[in]		*arg3			未使用．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDevice::runStatusInterruptTask(
	OSObject		*target,
	void			*arg0,
	void			*arg1,
	void			*arg2,
	void			*arg3
)
{
	PdjUsbAudioDevice	*self;

	FailIf(NULL == target, Exit);

	self = (PdjUsbAudioDevice *)target;

	self->handleStatusInterrupt();

Exit:
	return kIOReturnSuccess;
}

/****************************************************************************/
/**
 * @brief	Status Interrupt Endpoint からの受信処理［処理本体］．
 */
void
PdjUsbAudioDevice::handleStatusInterrupt(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::handleStatusInterrupt()", this);

	AudioStatusWordFormatPtr						audioStatusWord;
	USBAUDIO_0200::InterruptDataMessageFormatPtr	interruptDataMessage;
	UInt8											subType;
	PdjUsbAudioEngine								*currentEngine = NULL;
	OSDictionary									*currentEngineInfo = NULL;
	StandardUSB::DeviceRequest						devReq;
	OSSet											*defaultAudioControls;
	bool											interruptPending = false;
	bool											originatedFromACInterface = false;
	UInt8											bOriginator;

	FailIf(NULL == mControlInterface, Exit);

	// <rdar://7865729> Clear the pipe stall here instead of the statusInterruptHandler().
	// clearStall() cannot be made at that point because it will never be able to complete the correctly
	// as that is the only thread available for processing interrupts.
	if (mInterruptPipeStalled)
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIODEVICE "[%p]::handleStatusInterrupt() clearing pipe stall", this);

		if (mInterruptPipe)
		{
			mInterruptPipe->clearStall(true);
		}

		mInterruptPipeStalled = false;
	}

	if (kAudioStatusWordFormat == mStatusInterruptBufferType)
	{
		audioStatusWord				= (AudioStatusWordFormatPtr)mStatusInterruptBuffer;
		interruptPending			= (0 != (audioStatusWord->bStatusType & 0x80));
		originatedFromACInterface	= (0 == (audioStatusWord->bStatusType & 0x0F));
		bOriginator					= audioStatusWord->bOriginator;
	}
	else // if ( kInterruptDataMessageFormat == mStatusInterruptBufferType )
	{
		interruptDataMessage		= (USBAUDIO_0200::InterruptDataMessageFormatPtr)mStatusInterruptBuffer;
		interruptPending			= (0 == (interruptDataMessage->bInfo & 0x1)); // Class-specific interrupt.

		originatedFromACInterface	= (
											(0 == (interruptDataMessage->bInfo & 0x02))
									  &&
											((interruptDataMessage->wIndex & 0xFF) == mControlInterface->getInterfaceDescriptor()->bInterfaceNumber)
									  );

		bOriginator					= (UInt8)((interruptDataMessage->wIndex >> 8) & 0xFF);
	}

	if (interruptPending)
	{
		if (originatedFromACInterface)	// AudioControl interface
		{
			mConfigDictionary->getSubType(&subType, mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, bOriginator);

			// if this is a feature unit, dispatch to engines
			if ((FEATURE_UNIT == subType) || (SELECTOR_UNIT == subType))	// <rdar://6523676>
			{
				FailIf(NULL == mRegisteredEngines, Exit);

				for (UInt8 engineIndex = 0; engineIndex < mRegisteredEngines->getCount(); engineIndex++)
				{
					currentEngineInfo = OSDynamicCast(OSDictionary, mRegisteredEngines->getObject(engineIndex));
					FailIf(NULL == currentEngineInfo, Exit);

					currentEngine = OSDynamicCast(PdjUsbAudioEngine, currentEngineInfo->getObject(kEngine));
					FailIf(NULL == currentEngine, Exit);

					defaultAudioControls = currentEngine->copyDefaultAudioControls();
					if (NULL != defaultAudioControls) // <rdar://7695055>
					{
						mTopology->controlHasChangedOnDevice(bOriginator, defaultAudioControls);

						defaultAudioControls->release();
						defaultAudioControls = NULL;
					}
				}
			}
			else if (USBAUDIO_0200::CLOCK_SOURCE == subType)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIODEVICE "[%p]::handleStatusInterrupt() - CLOCK_SOURCE : %d", this, bOriginator);

				FailIf(NULL == mRegisteredEngines, Exit);

				for (UInt8 engineIndex = 0; engineIndex < mRegisteredEngines->getCount(); engineIndex++)
				{
					currentEngineInfo = OSDynamicCast(OSDictionary, mRegisteredEngines->getObject(engineIndex));
					FailIf(NULL == currentEngineInfo, Exit);

					currentEngine = OSDynamicCast(PdjUsbAudioEngine, currentEngineInfo->getObject(kEngine));
					FailIf(NULL == currentEngine, Exit);

					currentEngine->updateClockStatus(bOriginator);
				}
			}
		}
	}

	if (kAudioStatusWordFormat == mStatusInterruptBufferType)
	{
		FailIf(NULL == mControlInterface, Exit);

		// Clear status interrupt
		devReq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBClass, kUSBInterface);
		devReq.bRequest		 = GET_STAT;
		devReq.wValue		 = 0;
		devReq.wIndex		 = (0xFF00 & (bOriginator << 8)) | (0x00FF & mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
		devReq.wLength		 = 0;

		FailIf(kIOReturnSuccess != this->deviceRequest(&devReq, NULL), Exit);
	}
	else
	{
		// For USB Audio 2.0, interrupts are considered 'edge-triggered' type, meaning that an interrupt is generated whenever
		// an event occurs, but there is no specific action required from the Host to clear the interrupt condition.
	}

	// Clear the buffer before next read
	::bzero(
		mStatusInterruptBuffer,
		(kAudioStatusWordFormat == mStatusInterruptBufferType)
			? sizeof(AudioStatusWordFormat)
			: sizeof(USBAUDIO_0200::InterruptDataMessageFormat)
	);

	// Queue next read
	if (mInterruptPipe)
	{
		FailMessage(
						kIOReturnSuccess != mInterruptPipe->io(
																mInterruptEndpointMemoryDescriptor,
																(UInt32)mInterruptEndpointMemoryDescriptor->getLength(),
																&mStatusInterruptCompletion
															)
				   );
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::handleStatusInterrupt()", this);
	return;
}

#ifdef TARGET_PDJ
 /****************************************************************************/
 /**
  * @brief	システムの電源状態遷移通知．
  *
  * @param[in]		*target				メッセージの送信先．PdjUsbAudioDevice クラスのインスタンス．
  * @param[in]		*refCon				registerPrioritySleepWakeInterest() で指定したパラメータ．
  * @param[in]		messageType			メッセージの種類．IOMessage.h で定義されている．
  * @param[in]		*provider			メッセージの送信元．
  * @param[in]		*messageArgument	メッセージの引数．
  * @param[in]		argSize				メッセージの引数のサイズ．
  *
  * @return			IOReturn コード．
  */
 IOReturn
 PdjUsbAudioDevice::sleepWakeHandler(
 	void		*target,
 	void		*refCon,
 	UInt32		messageType,
 	IOService	*provider,
 	void		*messageArgument,
 	vm_size_t	argSize
 )
 {
 	PdjUsbAudioDevice	*self;
 
 	FailIf(NULL == target, Exit);
 
 	self = (PdjUsbAudioDevice *)target;
 
 	self->sleepWakeHandlerAction(messageType, messageArgument);
 
 Exit:
 	return kIOReturnSuccess;
 }
 
 /****************************************************************************/
 /**
  * @brief	システムの電源状態遷移通知．
  *
  * @param[in]		messageType			メッセージの種類．IOMessage.h で定義されている．
  * @param[in]		*messageArgument	メッセージの引数．
  */
 void
 PdjUsbAudioDevice::sleepWakeHandlerAction(
 	UInt32		messageType,
 	void		*messageArgument
 )
 {
 	DP(DBG_TRACE, "+ " LOGSTRAUDIODEVICE "[%p]::sleepWakeHandlerAction(0x%08X, %p)",
 		this, messageType, messageArgument);
 
 	if (messageType == kIOMessageSystemCapabilityChange)
 	{
 		IOPMSystemCapabilityChangeParameters	*param;
 
 		param = (IOPMSystemCapabilityChangeParameters *)messageArgument;
 		if (param)
 		{
 			DP(DBG_INFO, "  notifyRef        = %d", param->notifyRef);
 			DP(DBG_INFO, "  maxWaitForReply  = %d", param->maxWaitForReply);
 			DP(DBG_INFO, "  changeFlags      = %d", param->changeFlags);
 			DP(DBG_INFO, "  fromCapabilities = %d", param->fromCapabilities);
 			DP(DBG_INFO, "  toCapabilities   = %d", param->toCapabilities);
 
 			if (0 != mEngineArray)
 			{
 				for (UInt32 engineIndex = 0; engineIndex < mEngineArray->getCount(); engineIndex++)
 				{
 					PdjUsbAudioEngine	*engine = OSDynamicCast(
 															PdjUsbAudioEngine,
 															mEngineArray->getObject(engineIndex)
 												  );
 					if (engine)
 					{
 						switch (param->changeFlags)
 						{
 							case kIOPMSystemCapabilityWillChange:
 								engine->systemCapabilityWillChangeTo(param->fromCapabilities, param->toCapabilities);
 								break;
 
 							case kIOPMSystemCapabilityDidChange:
 								engine->systemCapabilityDidChangeTo(param->fromCapabilities, param->toCapabilities);
 								break;
 
 							default:
 								break;
 						}
 					}
 				}
 			}
 		}
 	}
 
 	DP(DBG_TRACE, "- " LOGSTRAUDIODEVICE "[%p]::sleepWakeHandlerAction(0x%08X, %p)",
 		this, messageType, messageArgument);
 }
#endif
