/*
 * Copyright (c) 1998-2009 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioDevice.h
 *
 * @brief	IOAudioDevice クラスの派生クラス．
 *			ドライバ全体のルートクラス．
 *			１個のデバイスの全体を表現する．
 *			AppleUSBAudioDevice クラスを２つに分離して
 *			PdjUsbAudioDevice クラスおよび PdjUsbAudioTopology クラスとした．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/
#ifndef PDJ_USB_AUDIO_DEVICE_H_FILE_
#define PDJ_USB_AUDIO_DEVICE_H_FILE_

/*------------------------------------------*/
/*				定数						*/
/*------------------------------------------*/

#pragma mark - Constants -

#define kStringBufferSize				255

// The following value is defined in USB 1.0 Class Spec section 5.2.2.4.3.2
#define kNegativeInfinity				0x8000

#define kWallTimeExtraPrecision			  10000ull
#define kMaxWallTimePerUSBCycle			1001000ull
#define kMinWallTimePerUSBCycle			 999000ull

#define kMaxTimeStampJitter				  10000ull						// <rdar://7378275>

#define kEngine							"engine"
#define kInterface						"interface"
#define kAltSetting						"altsetting"
#define kInputGainControls				"inputgaincontrols"
#define kInputMuteControls				"inputmutecontrols"
#define kOutputMuteControls				"outputmutecontrols"			// <rdar://6413207>
#define kPassThruVolControls			"passthruvolcontrols"			// <rdar://5366067>
#define kPassThruToggleControls			"passthrutogglecontrols"		// <rdar://5366067>
#define kOutputVolControls				"outputvolcontrols"
#define kPassThruPathsArray				"passthrupathsarray"
#define kPassThruSelectorControl		"passthruselectorcontrol"		// <rdar://5366067>
#define kDisplayRoutingPropertyKey		"DisplayRouting"				// <rdar://problem/7349398>


enum {
	kVolumeControl						= 1,
	kMuteControl						= 2
};

enum {
	kAudioStatusWordFormat				= 1,
	kInterruptDataMessageFormat			= 2
};

// The following return codes are used by PdjUsbAudioDevice to detail the status of a format change.
enum
{
	kPdjUsbFormatChangeNormal			= kIOReturnSuccess,
	kPdjUsbFormatChangeForced,
	kPdjUsbFormatChangeForceFailure,
	kPdjUsbFormatChangeError			= kIOReturnError
};

// <rdar://7446555>
#define MAX_ANCHOR_ENTRIES			(4096)						// <rdar://problem/7666699>
#define MIN_ENTRIES_APPLY_OFFSET	(MAX_ANCHOR_ENTRIES / 4)	// <rdar://problem/7666699>
#define MIN_FRAMES_APPLY_OFFSET		(512)						// <rdar://problem/7666699>

/*------------------------------------------*/
/*				型定義						*/
/*------------------------------------------*/

typedef struct
{
	U64		X[MAX_ANCHOR_ENTRIES];
	U64		Y[MAX_ANCHOR_ENTRIES];
	U128	XX[MAX_ANCHOR_ENTRIES];
	U128	XY[MAX_ANCHOR_ENTRIES];
	UInt32	index;
	UInt32	n;
	U64		sumX;
	U64		sumY;
	U128	sumXX;
	U128	sumXY;
#if (MAX_ANCHOR_ENTRIES <= 1024)
	U128	P;
	U128	Q;
#else
	U256	P;
	U256	Q;
#endif
	U256	QSumY;
	U256	Qn;
	U128	mExtraPrecision;

	UInt32	calculateOffset;	// <rdar://problem/7666699>
	bool	deviceStart;		// <rdar://problem/7666699>

} ANCHORTIME;

/*------------------------------------------*/
/*			PdjUsbAudioDevice クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbAudioDevice -

class PdjUsbAudioDevice : public IOAudioDevice
{
	OSDeclareDefaultStructors(PdjUsbAudioDevice)

public:

	typedef IOAudioDevice				super;

#pragma mark - Public Methods -

#ifdef DEBUGLOGGING
	void
	retain(
		void
	) const;
#endif

#ifdef DEBUGLOGGING
	void
	release(
		void
	) const;
#endif

	bool
	init(
		OSDictionary				*properties
	);

	void
	free(
		void
	);

	bool
	matchPropertyTable(
		OSDictionary				*table,
		SInt32						*score
	);

	bool
	start(
		IOService					*provider
	);

	void
	stop(
		IOService					*provider
	);

	IOReturn
	message(
		UInt32						type,
		IOService					*provider,
		void						*arg = 0
	);

	bool
	willTerminate(
		IOService					*provider,
		IOOptionBits				options
	);

	IOReturn
	performPowerStateChange(
		IOAudioDevicePowerState		oldPowerState,
		IOAudioDevicePowerState		newPowerState,
		UInt32						*microSecsUntilComplete
	);

	void
	setConfigurationApp(
		const char					*bundleID
	);

	IOUSBHostInterface *
	getUSBInterface(
		UInt8						interfaceNumber
	);

	IOReturn
	getStringDescriptor(
		UInt8						index,
		char						*buf,
		int							maxLen,
		UInt16						lang = 0x409
	);

	UInt8
	getManufacturerStringIndex(
		void
	);

	UInt8
	getProductStringIndex(
		void
	);

	UInt8
	getSerialNumberStringIndex(
		void
	);

	UInt16
	getVendorID(
		void
	);

	UInt16
	getProductID(
		void
	);

	OSNumber *
	getLocationID(
		void
	);

	UInt64
	getUSBFrameNumber(
		void
	);

	UInt8											// added for rdar://4168019
	getDeviceSpeed(
		void
	);

	bool											// added for rdar://3959606
	detectSplitTransactions(
		void
	);

#ifdef TARGET_PDJ
 	bool
 	checkForXHCI(
 		void
 	);
#else
	bool											// added for rdar://3727110
	checkForUHCI(
		void
	);
#endif

	IOReturn
	deviceRequest(
		StandardUSB::DeviceRequest	*request,
		void						*dataBuffer,
		UInt32						*bytesTransferred
	);

	IOReturn
	deviceRequest(
		StandardUSB::DeviceRequest	*request,
		IOMemoryDescriptor			*dataBuffer,
		IOUSBHostCompletion			*completion = NULL
	);

	char *
	terminalTypeString(
		UInt16						terminalType
	);

	char *											// <rdar://5811247>
	clockTypeString(
		UInt8						clockType
	);

	void											// <rdar://problem/7666699>
	calculateOffset(
		void
	);

	UInt64											// <rdar://7378275>
	getTimeForFrameNumber(
		UInt64						frameNumber
	);

	void											// <rdar://7378275>
	updateUSBCycleTime(
		void
	);

	IOReturn
	createControlsForInterface(
		IOAudioEngine				*audioEngine,
		UInt8						interfaceNum,
		UInt8						altSettingNum
	);

	UInt32
	formatChangeController(
		PdjUsbAudioEngine			*audioEngine,
		IOAudioStream				*audioStream,
		const IOAudioStreamFormat	*newFormat,
		const IOAudioSampleRate		*newSampleRate
	);

#ifdef PROVIDER_USB_DEVICE
	IOUSBHostInterface *
	findInterface(
		IOUSBFindInterfaceRequest	*request
	);
#endif

	static IOUSBHostPipe *
	findPipe(
		IOUSBHostInterface			*interface,
		IOUSBFindEndpointRequest	*request
	);

	PdjUsbConfigurationDictionary *
	getConfigDictionary(
		void
	)
	{
		return mConfigDictionary;
	}

	bool
	getSingleSampleRateDevice(
		void
	)
	{
		return mSingleSampleRateDevice;
	}

	void
	setSingleSampleRateDevice(
		bool						isSingleSampleRateDevice
	)
	{
		mSingleSampleRateDevice = isSingleSampleRateDevice;
	}

	void
	setShouldSyncSampleRates(
		PdjUsbAudioEngine			*problemEngine
	)
	{
		mFailingAudioEngine = problemEngine;
	}

	void
	setShouldResetEngine(
		PdjUsbAudioEngine			*problemEngine
	)
	{
		mEngineToRestart = problemEngine;
	}

	void
	requestDeviceRecovery(
		void
	)
	{
		mShouldAttemptDeviceRecovery = true;
	}

	bool
	recoveryRequested(
		void
	)
	{
		return mShouldAttemptDeviceRecovery;
	}

	bool
	initHardware(
		IOService					*provider
	);

protected:

#pragma mark - Protected Methods -

	static void										// <rdar://problem/6686515>
	initHardwareThread(
		PdjUsbAudioDevice			*audioDevice,
		IOService					*provider
	);

	static IOReturn									// <rdar://problem/6686515>
	initHardwareThreadAction(
		OSObject					*owner,
		void						*provider,
		void						*arg2,
		void						*arg3,
		void						*arg4
	);

	IOReturn
	protectedInitHardware(
		IOService					*provider
	);

	const StandardUSB::ConfigurationDescriptor *	// added for rdar://5495653
	getConfigurationDescriptor(
		void
	);

	bool
	createAudioEngines(
		void
	);

	bool
	createAudioEngine(
		OSArray						*streamInterfaceNumberArray
	);

	bool
	activateAudioEngines(
		void
	);

	IOReturn										// <rdar://problem/6892754>
	interfaceContainsSupportedFormat(
		UInt8						streamInterface,
		bool						*formatSupported
	);

	static void
	timerAction(
		OSObject					*owner,
		IOTimerEventSource			*sender
	);

	void
	doTimerAction(
		IOTimerEventSource			*timer
	);

	IOReturn
	getAnchorFrameAndTimeStamp(
		UInt64						*frame,
		AbsoluteTime				*time
	);

	IOReturn
	tryGetAnchorFrameAndTimeStamp(
		UInt64						*frame,
		AbsoluteTime				*time
	);

	static void
	updateAnchorTime(
		ANCHORTIME					*anchorTime,
		UInt64						X,
		UInt64						Y
	);

	void											// <rdar://problem/7666699>
	applyOffsetAmountToFilter(
		void
	);

	void											// [rdar://5165798]
	resetRateTimer(
		void
	);

	UInt64											// <rdar://problem/7666699>
	lastAnchorFrame(
		void
	);

#if DEBUGANCHORS
	void
	accumulateAnchor(
		UInt64						anchorFrame,
		AbsoluteTime				timeStamp
	);
#endif

	bool											// <rdar://problem/7666699>
	allEnginesStopped(
		void
	);

	void
	attemptDeviceRecovery(
		void
	);

	void
	checkForStatusInterruptEndpoint(
		void
	);

	static void
	statusInterruptHandler(
		void						*target,
		void						*parameter,
		IOReturn					status,
		UInt32						bufferSizeRemaining
	);

	static void
	processStatusInterrupt(
		void						*arg
	);

	static IOReturn
	runStatusInterruptTask(
		OSObject					*target,
		void						*arg0,
		void						*arg1,
		void						*arg2,
		void						*arg3
	);

	void
	handleStatusInterrupt(
		void
	);

	static UInt64
	getUSBCycleTime(
		ANCHORTIME					*anchorTime
	)
	{
		// The slope is the USB cycle time. This has the extra precision factor in it.
		return anchorTime->mExtraPrecision.lo;
	}

#ifdef TARGET_PDJ
 	static IOReturn
 	sleepWakeHandler(
 		void						*target,
 		void						*refCon,
 		UInt32						messageType,
 		IOService					*provider,
 		void						*messageArgument,
 		vm_size_t					argSize
 	);
 
 	void
 	sleepWakeHandlerAction(
 		UInt32						messageType,
 		void						*messageArgument
 	);
#endif

public:

	IOUSBHostInterface					*mControlInterface;
	PdjUsbAudioTopology					*mTopology;

	OSArray								*mRegisteredEngines;
	IORecursiveLock						*mRegisteredEnginesMutex;				// <rdar://problem/6369110>

	OSArray								*mRegisteredStreams;					// <rdar://6420832>
	IORecursiveLock						*mRegisteredStreamsMutex;				// <rdar://problem/6420832>

	UInt64								mWallTimePerUSBCycle;

	ANCHORTIME							mAnchorTime;							// <rdar://7378275>
	IOLock								*mTimeLock;								// <rdar://7378275>

protected:

#ifdef PROVIDER_USB_DEVICE
	IOUSBHostDevice						*mUSBDevice;
#endif

	IORecursiveLock						*mInterfaceLock;
	PdjUsbConfigurationDictionary		*mConfigDictionary;
	OSArray								*mEngineArray;
	bool								mTerminatingDriver;
	thread_call_t						mInitHardwareThread;					// <rdar://problem/6686515>

#ifdef TARGET_PDJ
 	IONotifier							*mSleepWakeNotifier;
#endif

	// These member variables are for anchored time stamp algorithm
	UInt64								mLastUSBFrame;
	UInt64								mLastWallTime_nanos;
#if DEBUGANCHORS
	UInt64								mAnchorFrames[kAnchorsToAccumulate];
	AbsoluteTime						mAnchorTimes[kAnchorsToAccumulate];
#endif

	// These member variables are for anchored time stamp algorithm
	IOTimerEventSource					*mUpdateTimer;
	bool								mSingleSampleRateDevice;
	PdjUsbAudioEngine					*mFailingAudioEngine;
	PdjUsbAudioEngine					*mEngineToRestart;						// [rdar://5355808]
	UInt32								mTimerCallCount;						// <rdar://problem/7378275>

	// This is for emergency device recovery
	bool								mShouldAttemptDeviceRecovery;

	// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
	IOUSBHostPipe						*mInterruptPipe;
	IOBufferMemoryDescriptor			*mInterruptEndpointMemoryDescriptor;	// <rdar://7000283>
	bool								mInterruptPipeStalled;					// <rdar://7865729>
	void								*mStatusInterruptBuffer;
	UInt32								mStatusInterruptBufferType;				// <rdar://5811247>
	IOUSBHostCompletion					mStatusInterruptCompletion;
	thread_call_t						mProcessStatusInterruptThread;			// <rdar://problem/6021475>

	UInt64								mRampUpdateCounter;						// <rdar://problem/7666699>
	UInt64								Xcopy[MAX_ANCHOR_ENTRIES];				// <rdar://problem/7666699>
	UInt64								Ycopy[MAX_ANCHOR_ENTRIES];				// <rdar://problem/7666699>
};

#endif /* PDJ_USB_AUDIO_DEVICE_H_FILE_ */
