/*
 * Copyright (c) 1998-2009 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioEngine.cpp
 *
 * @brief	IOAudioEngine クラスの派生クラス．
 *			PC ←→デバイス間のデータ転送を司る I/O エンジンを表現する．
 *			PdjUsbAudioDevice::createAudioEngine() においてインスタンス化される．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/

/*------------------------------------------*/
/*				インクルード				*/
/*------------------------------------------*/

#include "PdjUsbAudioCommon.h"

/*------------------------------------------*/
/*			PdjUsbAudioEngine クラス		*/
/*------------------------------------------*/

OSDefineMetaClassAndStructors(PdjUsbAudioEngine, super)

#pragma mark - Public Methods -

/****************************************************************************/
/**
 * @brief	インスタンス初期化．
 *			IOAudioEngine::init() とはシグネチャが異なる．
 *			PdjUsbAudioDevice::createAudioEngine() から呼び出され、
 *			オーディオストリームの interface 番号の一覧を渡される．
 *
 * @param[in]		*streamInterfaceNumberArray		オーディオストリームの interface 番号一覧．
 *
 * @retval			true			初期化成功．
 * @retval			false			初期化失敗．
 */
bool
PdjUsbAudioEngine::init(
	OSArray			*streamInterfaceNumberArray
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::init()", this);

	bool			result = false;

	FailIf(NULL == streamInterfaceNumberArray, Exit);
	FailIf(0 == streamInterfaceNumberArray->getCount(), Exit);

	FailIf(!super::init(NULL), Exit);

	mStreamInterfaceNumberArray = streamInterfaceNumberArray;
	mStreamInterfaceNumberArray->retain();

	mIOAudioStreamArray = OSArray::withCapacity(1);
	FailIf(NULL == mIOAudioStreamArray, Exit);

	// Change this to use defines from the IOAudioFamily when they are available
	IORegistryEntry::setProperty("IOAudioStreamSampleFormatByteOrder", "Little Endian");

	result = true;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::init()", this);
	return result;
}

/****************************************************************************/
/**
 * @brief	インスタンス解放．
 *			IOAudioEngine::free() をオーバーライド．
 *			参照カウントがゼロになった時呼び出される．
 */
void
PdjUsbAudioEngine::free(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::free()", this);

	// <rdar://5811247>
	if (NULL != mClockSelectorControl)
	{
		mClockSelectorControl->release();
		mClockSelectorControl = NULL;
	}

	if (NULL != mStreamInterfaceNumberArray)
	{
		mStreamInterfaceNumberArray->release();
		mStreamInterfaceNumberArray = NULL;
	}

	if (NULL != mIOAudioStreamArray)
	{
		mIOAudioStreamArray->release();
		mIOAudioStreamArray = NULL;
	}

	mMainOutputStream	= NULL;
	mMainInputStream	= NULL;

	super::free();

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::free()", this);
}

/****************************************************************************/
/**
 * @brief	サービスを IORegistry に登録．
 *			IOAudioEngine::registerService() をオーバーライド．
 *			このメソッドを呼び出すことによりオーディオデバイスが認識される．
 *
 * @param[in]		options			オプション指定．
 */
// <rdar://7295322> Asynchronous to prevent deadlock if the device or interface is terminated while
// registerService() is performing matching.
void
PdjUsbAudioEngine::registerService(
	IOOptionBits	options
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::registerService(0x%x)", this, options);

	if (0 == (kIOServiceSynchronous & options))
	{
		options |= kIOServiceAsynchronous;
	}

	super::registerService(options);

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::registerService(0x%x)", this, options);
}

/****************************************************************************/
/**
 * @brief	サービスマッチング．
 *			IOService::matchPropertyTable() をオーバーライド．
 *			ドライバ固有のマッチング条件があるときはここで判定を行う．
 *
 * @param[in]		*table			マッチングの対象となるプロパティ．
 * @param[in, out]	*score			プローブスコア．必要に応じて書き換えることができる．
 *
 * @retval			true			マッチング成功．
 * @retval			false			マッチング失敗．
 */
bool
PdjUsbAudioEngine::matchPropertyTable(
	OSDictionary	*table,
	SInt32			*score
)
{
	DP(DBG_VERBOSE, "+ " LOGSTRAUDIOENGINE "[%p]::matchPropertyTable(%p, %p)", this, table, score);

	bool			returnValue = false;

	if (super::matchPropertyTable(table, score))
	{
		if (
				IOService::compareProperty(table, kIDVendorString)
		   &&
				IOService::compareProperty(table, kIDProductString)
		   )
		{
			returnValue = true;
		}
	}

	DP(DBG_VERBOSE, "- " LOGSTRAUDIOENGINE "[%p]::matchPropertyTable(%p, %p) = %d", this, table, score, returnValue);

	return returnValue;
}

/****************************************************************************/
/**
 * @brief	サービス開始．
 *			IOAudioEngine::start() をオーバーライド．
 *			super::start() 内でハードウェアの初期化が行われる．
 *
 * @param[in]		*provider		サービスプロバイダ．PdjUsbAudioDevice クラスのインスタンス．
 * @param[in]		*device			IOAudioDevice．通常 provider と同一のオブジェクトである．
 *
 * @retval			true			開始成功．
 * @retval			false			開始失敗．
 */
bool
PdjUsbAudioEngine::start(
	IOService		*provider,
	IOAudioDevice	*device
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::start(%p)", this, provider);

	bool			resultCode = false;

	FailIf(NULL == device, Exit);

	mUSBAudioDevice = OSDynamicCast(PdjUsbAudioDevice, device);

	// <rdar://62775111> Retain a reference to the PdjUsbAudioDevice object here so that it doesn't goes away
	// while it is initializing.
	mUSBAudioDevice->retain();

	resultCode = super::start(provider, device);

Exit:
	// <rdar://62775111> In case it failed, clean up to prevent leakage as stop() will not be called.
	if (!resultCode)
	{
		if (NULL != mUSBAudioDevice)
		{
			mUSBAudioDevice->release();
			mUSBAudioDevice = NULL;
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::start(%p) = %d", this, provider, resultCode);
	return resultCode;
}

/****************************************************************************/
/**
 * @brief	サービス停止．
 *			IOAudioEngine::stop() をオーバーライド．
 *
 * @param[in]		*provider		サービスプロバイダ．PdjUsbAudioDevice クラスのインスタンス．
 */
void
PdjUsbAudioEngine::stop(
	IOService		*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::stop(%p)", this, provider);

	if (NULL != mPluginInitThread)
	{
		::thread_call_cancel(mPluginInitThread);
		::thread_call_free(mPluginInitThread);
		mPluginInitThread = NULL;
	}

	if (mPlugin)
	{
		mPlugin->close(this);
		mPlugin = NULL;
	}

	if (mUSBAudioDevice)
	{
		mUSBAudioDevice->release();
		mUSBAudioDevice = NULL;
	}

	super::stop(provider);

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::stop(%p) - rc = %d", this, provider, this->getRetainCount());
}

/****************************************************************************/
/**
 * @brief	サービス終了要求．
 *			IOService::requestTerminate() をオーバーライド．
 *			プロバイダがサービスを終了した際、このサービスも終了させてよいかどうか
 *			問い合わせるために呼び出される．
 *
 * @param[in]		*provider		サービスを終了したプロバイダ．
 *									PdjUsbAudioDevice クラスのインスタンス．
 * @param[in]		options			オプション指定．
 *
 * @retval			true			サービス終了要求を受け入れる．
 * @retval			false			サービス終了要求を拒否する．
 */
bool
PdjUsbAudioEngine::requestTerminate(
	IOService		*provider,
	IOOptionBits	options
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::requestTerminate(%p, %x)", this, provider, options);

	bool			result;

	// if interface or audio device
	if (mUSBAudioDevice == provider)
	{
		result = true;		// it is OK to terminate us
	}
	else
	{
		result = false;		// don't terminate us
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::requestTerminate(%p, %x) = %d", this, provider, options, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	サービスを終了させる．
 *			IOService::terminate() をオーバーライド．
 *			このサービスを終了させるために呼び出される．
 *
 * @param[in]		options			オプション指定．
 *
 * @retval			true			成功．このサービスは終了される．
 * @retval			false			失敗．このサービスは終了されない．
 */
bool
PdjUsbAudioEngine::terminate(
	IOOptionBits	options
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::terminate()", this);

	bool			shouldTerminate	= true;
	bool			result			= true;

	if (shouldTerminate)
	{
		result = super::terminate(options);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::terminate()", this);
	return result;
}

#ifdef TARGET_PDJ
 /****************************************************************************/
 /**
  * @brief	システムの電源状態遷移通知（予告）．
  *
  * @param[in]		fromCapabilities	遷移前のシステム電源状態．
  * @param[in]		toCapabilities		遷移後のシステム電源状態．
  */
 void
 PdjUsbAudioEngine::systemCapabilityWillChangeTo(
 	UInt32		fromCapabilities,
 	UInt32		toCapabilities
 )
 {
 	DP(DBG_WARNING, "+ " LOGSTRAUDIOENGINE "[%p]::systemCapabilityWillChangeTo(0x%08X, 0x%08X)",
 		this, fromCapabilities, toCapabilities);
 
 	if		(	(0 != (fromCapabilities & kIOPMSystemCapabilityAudio))		/* 就寝開始 */
 			 &&	(0 == (toCapabilities   & kIOPMSystemCapabilityAudio)))
 	{
 		mSystemSleep = 1;
 	}
 	else if (	(0 == (fromCapabilities & kIOPMSystemCapabilityAudio))		/* 起床開始 */
 			 &&	(0 != (toCapabilities   & kIOPMSystemCapabilityAudio)))
 	{

 // @note ここでストリーミングを再開すると、レイテンシずれが発生する場合がある
 //
 //		super::pauseAudioEngine();
 //			super::beginConfigurationChange();
 //				super::hardwareSampleRateChanged(&mCurSampleRate);
 //			super::completeConfigurationChange();
 //		super::resumeAudioEngine();

 		mSystemSleep = 3;
 	}
 
 	DP(DBG_WARNING, "- " LOGSTRAUDIOENGINE "[%p]::systemCapabilityWillChangeTo(0x%08X, 0x%08X)",
 		this, fromCapabilities, toCapabilities);
 }
 
 /****************************************************************************/
 /**
  * @brief	システムの電源状態遷移通知（完了）．
  *
  * @param[in]		fromCapabilities	遷移前のシステム電源状態．
  * @param[in]		toCapabilities		遷移後のシステム電源状態．
  */
 void
 PdjUsbAudioEngine::systemCapabilityDidChangeTo(
 	UInt32		fromCapabilities,
 	UInt32		toCapabilities
 )
 {
 	DP(DBG_WARNING, "+ " LOGSTRAUDIOENGINE "[%p]::systemCapabilityDidChangeTo(0x%08X, 0x%08X)",
 		this, fromCapabilities, toCapabilities);
 
 	if		(	(0 != (fromCapabilities & kIOPMSystemCapabilityAudio))		/* 就寝完了 */
 			 &&	(0 == (toCapabilities   & kIOPMSystemCapabilityAudio)))
 	{
 		mSystemSleep = 2;
 	}
 	else if (	(0 == (fromCapabilities & kIOPMSystemCapabilityAudio))		/* 起床完了 */
 			 &&	(0 != (toCapabilities   & kIOPMSystemCapabilityAudio)))
 	{
 //		if (mSystemSleep != 0)		/* Redmine #9880 */
 		{
 			if (mUSBAudioDevice)
 			{
 				IOCommandGate	*cg = mUSBAudioDevice->getCommandGate();
 				if (cg)
 				{
 					cg->runAction(this->restartEngineAction, (void *)this);
 				}
 			}
 		}
 
 		mSystemSleep = 0;
 	}
 
 	DP(DBG_WARNING, "- " LOGSTRAUDIOENGINE "[%p]::systemCapabilityDidChangeTo(0x%08X, 0x%08X)",
 		this, fromCapabilities, toCapabilities);
 }
 
 /****************************************************************************/
 /**
  * @brief	Engine をリスタート［コマンドゲート］．
  *
  * @param[in]		*device			PdjUsbAudioDevice クラスのインスタンス．
  * @param[in]		*engine			PdjUsbAudioEngine クラスのインスタンス．
  * @param[in]		*arg1			未使用．
  * @param[in]		*arg2			未使用．
  * @param[in]		*arg3			未使用．
  *
  * @return			IOReturn コード．
  */
 IOReturn
 PdjUsbAudioEngine::restartEngineAction(
 	OSObject		*device,
 	void			*engine,
 	void			*arg1,
 	void			*arg2,
 	void			*arg3
 )
 {
 	PdjUsbAudioEngine	*self;
 
 	FailIf(NULL == engine, Exit);
 
 	self = (PdjUsbAudioEngine *)engine;
 
 	self->restartEngine();
 
 Exit:
 	return kIOReturnSuccess;
 }
 
 /****************************************************************************/
 /**
  * @brief	Engine をリスタート［処理本体］．
  */
 void
 PdjUsbAudioEngine::restartEngine(
 	void
 )
 {
 	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::restartEngine()");
 
 	super::pauseAudioEngine();
 		super::beginConfigurationChange();
 			super::hardwareSampleRateChanged(&mCurSampleRate);
 		super::completeConfigurationChange();
 	super::resumeAudioEngine();
 
 	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::restartEngine()");
 }
#endif

/****************************************************************************/
/**
 * @brief	このオーディオエンジンを識別するための ID 文字列を生成．
 *			IOAudioEngine::getGlobalUniqueID() をオーバーライド．
 *			生成された ID はオーディオエンジンのプロパティの１つとして使用される．
 *			一般に言う GUID ではなく、USB のストリングディスクリプタ等を
 *			組み合わせた文字列となる．
 *
 * @return			一意の ID 文字列．
 */
OSString *
PdjUsbAudioEngine::getGlobalUniqueID(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::getGlobalUniqueID()");

	IOReturn		err = kIOReturnSuccess;
	OSObject		*nameObject	 = NULL;			// <rdar://problem/6398888>
	OSString		*nameString	 = NULL;
	UInt32			uniqueIDSize;
	char			*uniqueIDStr = NULL;
	OSString		*uniqueID	 = NULL;
	UInt32			locationID;
	char			locationIDString[kStringBufferSize];
	UInt8			stringIndex;
	UInt8			interfaceNumber;
	char			interfaceNumberString[kStringBufferSize];	// <rdar://7676010>
	char			productString[kStringBufferSize];
	char			manufacturerString[kStringBufferSize];
	char			serialNumberString[kStringBufferSize];
	char			tempString[kStringBufferSize];

	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == mUSBAudioDevice->mControlInterface, Exit);

	uniqueIDSize  = 5;		// This is the number of ":" in the final string, plus space for the trailing NULL that ends the string
	uniqueIDSize += ::strlen("com_pioneerdj_DJM-V10");

	manufacturerString[0] = 0;

	// <rdar://problem/6398888> Attempt to retrieve the manufacturer name from registry before reading it from the device
	nameObject = mUSBAudioDevice->mControlInterface->getDevice()->getProperty("USB Vendor Name");

	if (nameObject)
	{
		if (NULL != (nameString = OSDynamicCast(OSString, nameObject)))
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::getGlobalUniqueID() - Retrieved vendor name %s from registry",
				this, nameString->getCStringNoCopy());

			::strncpy(manufacturerString, nameString->getCStringNoCopy(), kStringBufferSize);

			err = kIOReturnSuccess;
		}
	}
	else
	{
		stringIndex = mUSBAudioDevice->getManufacturerStringIndex();

		if (0 != stringIndex)
		{
			err = mUSBAudioDevice->getStringDescriptor(stringIndex, manufacturerString, kStringBufferSize);
		}
	}

	if ((0 == manufacturerString[0]) || (kIOReturnSuccess != err))
	{
		::strncpy(manufacturerString, "Unknown Manufacturer", kStringBufferSize);
	}

	uniqueIDSize += ::strlen(manufacturerString);

	err = kIOReturnSuccess;

	productString[0] = 0;

	// <rdar://problem/6398888> Attempt to retrieve product name from registry before reading it from the device
	nameObject = mUSBAudioDevice->mControlInterface->getDevice()->getProperty("USB Product Name");

	if (nameObject)
	{
		if (NULL != (nameString = OSDynamicCast(OSString, nameObject)))
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::getGlobalUniqueID() - Retrieved product name %s from registry",
				this, nameString->getCStringNoCopy());

			::strncpy(productString, nameString->getCStringNoCopy(), kStringBufferSize);

			err = kIOReturnSuccess;
		}
	}
	else
	{
		stringIndex = mUSBAudioDevice->getProductStringIndex();

		if (0 != stringIndex)
		{
			err = mUSBAudioDevice->getStringDescriptor(stringIndex, productString, kStringBufferSize);
		}
	}

	if ((0 == productString[0]) || (kIOReturnSuccess != err))
	{
		::strncpy(productString, "Unknown USB Audio Device", kStringBufferSize);
	}

	uniqueIDSize += ::strlen(productString);

	// <rdar://problem/6470583>

	err = kIOReturnSuccess;

	serialNumberString[0] = 0;

	// <rdar://problem/6398888> Attempt to retrieve serial number from registry before reading it from the device
	nameObject = mUSBAudioDevice->mControlInterface->getDevice()->getProperty("USB Serial Number");

	if (nameObject)
	{
		if (NULL != (nameString = OSDynamicCast(OSString, nameObject)))
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::getGlobalUniqueID() - Retrieved serial number %s from registry",
				this, nameString->getCStringNoCopy());

			::strncpy(serialNumberString, nameString->getCStringNoCopy(), kStringBufferSize);

			err = kIOReturnSuccess;
		}
	}
	else
	{
		stringIndex = mUSBAudioDevice->getSerialNumberStringIndex();

		if (0 != stringIndex)
		{
			err = mUSBAudioDevice->getStringDescriptor(stringIndex, serialNumberString, kStringBufferSize);
		}
	}

	if ((0 == serialNumberString[0]) || (kIOReturnSuccess != err))
	{
		// device doesn't have a serial number, get its location ID
		OSNumber	*usbLocation = mUSBAudioDevice->getLocationID();

		if (NULL != usbLocation)
		{
			locationID = usbLocation->unsigned32BitValue();

			::snprintf(locationIDString, kStringBufferSize, "%x", locationID);
		}
		else
		{
			::strncpy(locationIDString, "Unknown location", kStringBufferSize);
		}

		uniqueIDSize += ::strlen(locationIDString);
	}
	else
	{
		// device has a serial number that we can use to track it
		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::getGlobalUniqueID() - Device has a serial number = %s",
			this, serialNumberString);

		uniqueIDSize += ::strlen(serialNumberString);
	}

	// Depending on whether this is a single or multi-streams engine, we may add the interface number.

	// <rdar://7676010> Concatenate the interface number(s).
	interfaceNumberString[0] = 0;
	for (UInt32 index = 0; index < mStreamInterfaceNumberArray->getCount(); index++)
	{
		OSNumber	*streamInterfaceNumber = OSDynamicCast(OSNumber, mStreamInterfaceNumberArray->getObject(index));

		if (NULL != streamInterfaceNumber)
		{
			interfaceNumber = streamInterfaceNumber->unsigned8BitValue();

			::strncpy(tempString, interfaceNumberString, kStringBufferSize);
			::snprintf(interfaceNumberString, 4, "%s%s%d", tempString, (index == 0) ? "" : ",", interfaceNumber);

			uniqueIDSize += ::strlen(interfaceNumberString) + ((index == 0) ? 0 : 1);
		}
	}

	uniqueIDStr = (char *)IOMalloc(uniqueIDSize);

	if (NULL != uniqueIDStr)
	{
		uniqueIDStr[0] = 0;

		if (0 == serialNumberString[0])
		{
			::snprintf(
				uniqueIDStr,
				uniqueIDSize,
				"com_pioneerdj_DJM-V10:%s:%s:%s",
				manufacturerString,
				productString,
				locationIDString
			);
		}
		else
		{
			::snprintf(
				uniqueIDStr,
				uniqueIDSize,
				"com_pioneerdj_DJM-V10:%s:%s:%s",
				manufacturerString,
				productString,
				serialNumberString
			);
		}

		// <rdar://7676010> Concatenate the interface number(s).
		::strlcat(uniqueIDStr, ":", uniqueIDSize);
		::strlcat(uniqueIDStr, interfaceNumberString, uniqueIDSize);

		uniqueID = OSString::withCString(uniqueIDStr);

		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::getGlobalUniqueID() - getGlobalUniqueID = %s",
			this, uniqueIDStr);

		IOFree(uniqueIDStr, uniqueIDSize);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::getGlobalUniqueID()");
	return uniqueID;
}

/****************************************************************************/
/**
 * @brief	ハードウェア初期化．
 *			IOAudioEngine::initHardware() をオーバーライド．
 *			super::start() から呼び出される．
 *			この関数から return 時には、オーディオ入出力を行う準備ができていなければならない．
 *
 * @param[in]		*provider		サービスプロバイダ．PdjUsbAudioDevice クラスのインスタンス．
 *
 * @retval			true			初期化成功．
 * @retval			false			初期化失敗．
 */
bool
PdjUsbAudioEngine::initHardware(
	IOService		*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::initHardware(%p)", this, provider);

	bool								resultBool					= false;
	PdjUsbConfigurationDictionary		*configDictionary;
	OSDictionary						*channelNamesDictionary		= NULL;		// <rdar://6430836>
	PdjUsbAudioStream					*masterStream				= NULL;
	bool								syncOutputCompensation		= false;	// <rdar://6343818>
	bool								syncInputCompensation		= false;	// <rdar://6343818>
	char								*streamDescription			= NULL;
	OSObject							*nameObject					= NULL;
	IOReturn							resultCode					= kIOReturnError;
	OSString							*coreAudioPluginPathString	= NULL;
	OSNumber							*idVendor					= NULL;
	OSNumber							*idProduct					= NULL;

	FailIf(!super::initHardware(provider), Exit);

	FailIf(NULL == mUSBAudioDevice, Exit);							// <rdar://7085810>
	FailIf(NULL == mUSBAudioDevice->mControlInterface, Exit);		// <rdar://7085810>
	FailIf(NULL == mStreamInterfaceNumberArray, Exit);
	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	// Choose default sampling rate (rdar://3866513)
	this->getDefaultSampleRate(&sampleRate);

	mCurSampleRate = sampleRate;

#ifdef TARGET_PDJ
#else
	// See if UHCI support is necessary
	mUHCISupport = mUSBAudioDevice->checkForUHCI();
#endif

	super::setSampleRate(&sampleRate);

	// Store default sample rate
	mDefaultAudioSampleRate = sampleRate;

	// <rdar://6430836> Dictionary for channel names
	channelNamesDictionary = OSDictionary::withCapacity(4);
	FailIf(NULL == channelNamesDictionary, Exit);

	IORegistryEntry::setProperty(kIOAudioEngineFullChannelNamesKey, channelNamesDictionary);
	channelNamesDictionary->release();

	mStartInputChannelID	= 1;
	mStartOutputChannelID	= 1;

	for (UInt32 streamIndex = 0; streamIndex < mStreamInterfaceNumberArray->getCount(); streamIndex++)
	{
		OSNumber	*streamInterfaceNumber = OSDynamicCast(OSNumber, mStreamInterfaceNumberArray->getObject(streamIndex));
		FailIf(NULL == streamInterfaceNumber, Exit);

#ifdef TARGET_PDJ
 		IOUSBHostInterface	*streamInterface = NULL;
 
 		for (
 				UInt32 numTriesForStreamInterface = 0;
 				numTriesForStreamInterface < kMaxTriesForStreamInterface;
 				numTriesForStreamInterface++
 			)
 		{
 			streamInterface = mUSBAudioDevice->getUSBInterface(mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
 			if (streamInterface)
 			{
 				break;
 			}
 
 			DP(DBG_WARNING, "! " LOGSTRAUDIOENGINE "[%p]::initHardware() - Can't retrieve USB interface #%d, sleeping 10 ms...",
 				this, streamInterfaceNumber->unsigned8BitValue());
 
 			IOSleep(10);
 
 			FailIf(mUSBAudioDevice->isInactive(), Exit);
 		}
 
 		FailIf(NULL == streamInterface, Exit);
#else
		IOUSBHostInterface	*streamInterface = mUSBAudioDevice->getUSBInterface(streamInterfaceNumber->unsigned8BitValue());
		FailIf(NULL == streamInterface, Exit);

		// <rdar://problem/6686515> Wait for stream interface nubs to appear to give opportunity for override kext properties to merge
		for (
				UInt32 numTriesForStreamPropertiesReady = 0;
				numTriesForStreamPropertiesReady < kMaxTriesForStreamPropertiesReady;
				numTriesForStreamPropertiesReady++
			)
		{
			OSObject	*propertiesReadyObject = streamInterface->getProperty("PDJXXXPropertiesReady");

			if (propertiesReadyObject)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::initHardware() - PDJXXXPropertiesReady found for stream #%d",
					this, streamInterfaceNumber->unsigned8BitValue());
				break;
			}

			DP(DBG_WARNING, "! " LOGSTRAUDIOENGINE "[%p]::initHardware() - PDJXXXPropertiesReady NOT found for stream #%d, sleeping 10 ms...",
				this, streamInterfaceNumber->unsigned8BitValue());

			IOSleep(10);

			FailIf(mUSBAudioDevice->isInactive(), Exit);
		}
#endif

		// Instantiate one PdjUsbAudioStream per IOUSBHostInterface.
		PdjUsbAudioStream	*audioStream = OSTypeAlloc(PdjUsbAudioStream);
		FailIf(NULL == audioStream, Exit);

		// If this engine supports more than 1 audio stream, show different description for the streams.
		if (mStreamInterfaceNumberArray->getCount() > 1)
		{
			nameObject = streamInterface->getProperty("USB Interface Name");

			if (nameObject)
			{
				OSString	*nameString;

				if (NULL != (nameString = OSDynamicCast(OSString, nameObject)))
				{
					DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::initHardware() - Retrieved product name %s",
						this, nameString->getCStringNoCopy());

					streamDescription = (char *)nameString->getCStringNoCopy();
				}
			}
		}

#ifdef TARGET_PDJ
 		FailWithAction(
 							!audioStream->initWithAudioEngine(
 												mUSBAudioDevice,
 												this,
 												streamInterfaceNumber->unsigned8BitValue(),
 												sampleRate,
 												streamDescription
 										  ),
 							audioStream->release(),
 							Exit
 					  );
#else
		FailWithAction(
							!audioStream->initWithAudioEngine(
												mUSBAudioDevice,
												this,
												streamInterface,
												sampleRate,
												streamDescription
										  ),
							audioStream->release(),
							Exit
					  );
#endif

		mIOAudioStreamArray->setObject(audioStream);

		if (kIOAudioStreamDirectionOutput == audioStream->getDirection())
		{
			if (NULL == mMainOutputStream)
			{
				mMainOutputStream = audioStream;
			}
			else
			{
				// <rdar://6343818> If the output is adaptive, use it as the main output stream.
				if (kAdaptiveSyncType == audioStream->getSyncType())
				{
					mMainOutputStream = audioStream;
				}
			}
		}

		if ((NULL == mMainInputStream) && (kIOAudioStreamDirectionInput == audioStream->getDirection()))
		{
			mMainInputStream = audioStream;
		}

		FailWithAction(!audioStream->configureAudioStream(sampleRate), audioStream->release(), Exit);

		FailWithAction(kIOReturnSuccess != (resultCode = super::addAudioStream(audioStream)), audioStream->release(), Exit);

		if (kIOAudioStreamDirectionOutput == audioStream->getDirection())
		{
			mStartOutputChannelID += audioStream->getFormat()->fNumChannels;
		}
		else
		{
			mStartInputChannelID += audioStream->getFormat()->fNumChannels;
		}

		audioStream->release();
	}

	// <rdar://6343818> Determine which of the stream is to be the master stream.
	if ((NULL != mMainOutputStream) && (kAdaptiveSyncType == mMainOutputStream->getSyncType()))
	{
		// The main output stream is present & adaptive, so designate it as the master stream to generate time stamps.
		masterStream = mMainOutputStream;

		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::initHardware(%p) - Main output stream (%p) is adaptive and designated as master",
			this, provider, mMainOutputStream);
	}
	else
	{
		// Otherwise, use the input stream as master stream if it is present.
		if (NULL != mMainInputStream)
		{
			masterStream = mMainInputStream;
		}
		else
		{
			masterStream = mMainOutputStream;
		}

		// If there is more than 1 audio streams present, then we need to compensate for synchronization between the input & output
		// streams running on the same engine.
		if (1 < mIOAudioStreamArray->getCount())
		{
			syncOutputCompensation = true;
		}
	}

	FailIf(NULL == masterStream, Exit);

	masterStream->setMasterStreamMode(true);

	DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::initHardware(%p) - Compensate for output synchronization: %d",
		this, provider, syncOutputCompensation);

	if (syncOutputCompensation)
	{
		// <rdar://6343818> Compensate for the synchronization on the output since the time stamps are generated on the input stream.
		for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
		{
			PdjUsbAudioStream	*audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));

			if ((NULL != audioStream) && (kIOAudioStreamDirectionOutput == audioStream->getDirection()))
			{
				audioStream->compensateForSynchronization(true);
				audioStream->updateSampleOffsetAndLatency();
			}
		}
	}

	DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::initHardware(%p) - Compensate for input synchronization: %d",
		this, provider, syncInputCompensation);

	if (syncInputCompensation)
	{
		// <rdar://6343818> Compensate for the synchronization on the input since the time stamps are generated on the output stream.
		for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
		{
			PdjUsbAudioStream	*audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));

			if ((NULL != audioStream) && (kIOAudioStreamDirectionInput == audioStream->getDirection()))
			{
				audioStream->compensateForSynchronization(true);
				audioStream->updateSampleOffsetAndLatency();
			}
		}
	}

	// <rdar://7558825> Clock domain for different synchronization type.
	switch (masterStream->getSyncType())
	{
		case kSynchronousSyncType:
		case kAdaptiveSyncType:
			this->setClockDomain(this->getSystemClockDomain());
			break;
		default:
			this->setClockDomain();
			break;
	}

	// If there is more than one audio streams per engine, then use the name of the control interface. Otherwise,
	// use the name of the stream interface if available.
	if (mIOAudioStreamArray->getCount() > 1)
	{
		nameObject = mUSBAudioDevice->mControlInterface->getProperty("USB Interface Name");
	}
	else
	{
		// [rdar://problem/5797877] Show different names for input and output interfaces of a USB Audio device
		if ((NULL != mMainOutputStream) && (NULL != mMainOutputStream->mStreamInterface))
		{
			nameObject = mMainOutputStream->mStreamInterface->getProperty("USB Interface Name");
		}
		else if ((NULL != mMainInputStream) && (NULL != mMainInputStream->mStreamInterface))
		{
			nameObject = mMainInputStream->mStreamInterface->getProperty("USB Interface Name");
		}
	}

	if (nameObject)
	{
		OSString	*nameString;

		if (NULL != (nameString = OSDynamicCast(OSString, nameObject)))
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::initHardware() - Retrieved product name %s",
				this, nameString->getCStringNoCopy());

			super::setDescription(nameString->getCStringNoCopy());
		}
	}

	// If there is more than one audio streams per engine, then look for the plugin path specified in the control interface
	// first. If it is not there, then look for the plugin path in the stream interface.
	// If there is only one audio stream per engine, then look for the the plugin path in the stream interface first. If it
	// is not there, look for the plugin path in the specified in the control interface.
	if (mIOAudioStreamArray->getCount() > 1)
	{
		coreAudioPluginPathString = OSDynamicCast(
										OSString,
										mUSBAudioDevice->mControlInterface->getProperty(kIOAudioEngineCoreAudioPlugInKey)
									);
	}

	if (NULL == coreAudioPluginPathString)
	{
		if ((NULL != mMainOutputStream) && (NULL != mMainOutputStream->mStreamInterface))
		{
			coreAudioPluginPathString = OSDynamicCast(
											OSString,
											mMainOutputStream->mStreamInterface->getProperty(kIOAudioEngineCoreAudioPlugInKey)
										);
		}
		else if ((NULL != mMainInputStream) && (NULL != mMainInputStream->mStreamInterface))
		{
			coreAudioPluginPathString = OSDynamicCast(
											OSString,
											mMainInputStream->mStreamInterface->getProperty(kIOAudioEngineCoreAudioPlugInKey)
										);
		}
	}

	if (NULL == coreAudioPluginPathString)
	{
		coreAudioPluginPathString = OSDynamicCast(
										OSString,
										mUSBAudioDevice->mControlInterface->getProperty(kIOAudioEngineCoreAudioPlugInKey)
									);
	}

	if (NULL != coreAudioPluginPathString)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::initHardware() - Retrieved CoreAudio plugin path %s",
			this, coreAudioPluginPathString->getCStringNoCopy());

		IORegistryEntry::setProperty(kIOAudioEngineCoreAudioPlugInKey, coreAudioPluginPathString);
	}

	resultBool = true;

	// Ask for plugin to load (if it exists)
	idVendor = OSNumber::withNumber(mUSBAudioDevice->getVendorID(), 16);
	if (NULL != idVendor)
	{
		IORegistryEntry::setProperty(kIDVendorString, idVendor);
		idVendor->release();
	}

	idProduct = OSNumber::withNumber(mUSBAudioDevice->getProductID(), 16);
	if (NULL != idProduct)
	{
		IORegistryEntry::setProperty(kIDProductString, idProduct);
		idProduct->release();
	}

Exit:
	// <rdar://problem/6576824> PdjUsbAudio fails to unload properly
	if (!resultBool)
	{
		// <rdar://7295322> Clean up the default audio controls in the case when the engine becomes inactive while the starting.
		// The controls doesn't get detach from the engine by IOAudioEngine::removeAllDefaultAudioControls() as the engine
		// is already inactive.
		if (defaultAudioControls)
		{
			if (IOService::isInactive())
			{
				OSCollectionIterator	*controlIterator = OSCollectionIterator::withCollection(defaultAudioControls);

				if (controlIterator)
				{
					IOAudioControl	*control;

					while ((control = (IOAudioControl *)(controlIterator->getNextObject())))
					{
						control->detach(this);
					}

					controlIterator->release();
				}
			}
		}

		this->stop(provider);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::initHardware(%p), resultCode = %x, resultBool = %d",
		this, provider, resultCode, resultBool);
	return resultBool;
}

/****************************************************************************/
/**
 * @brief	オーディオストリームに対応する USB interface をオープン．
 */
void
PdjUsbAudioEngine::openStreamInterfaces(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::openStreamInterfaces()");

	FailIf(NULL == mIOAudioStreamArray, Exit);

	for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
	{
		PdjUsbAudioStream	*audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));
		FailIf(NULL == audioStream, Exit);

		audioStream->openStreamInterface();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::openStreamInterfaces()");
	return;
}

/****************************************************************************/
/**
 * @brief	オーディオストリームに対応する USB interface をクローズ．
 */
void
PdjUsbAudioEngine::closeStreamInterfaces(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::closeStreamInterfaces()");

	FailIf(NULL == mIOAudioStreamArray, Exit);

	for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
	{
		PdjUsbAudioStream	*audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));
		FailIf(NULL == audioStream, Exit);

		audioStream->closeStreamInterface();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::closeStreamInterfaces()");
	return;
}

/****************************************************************************/
/**
 * @brief	オーディオ I/O エンジン開始．
 *			IOAudioEngine::performAudioEngineStart() をオーバーライド．
 *			ストリーミングを開始する．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioEngine::performAudioEngineStart(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::performAudioEngineStart()", this);

	IOReturn		resultCode = kIOReturnError;
	UInt32			usbFramesToDelay = 0;

	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == mIOAudioStreamArray, Exit);

	if (mUSBAudioDevice->mAnchorTime.n == 0)
	{
		// We have to have an anchor frame and time before we can take a time stamp. Generate one now.
		DP(DBG_WARNING, "! " LOGSTRAUDIOENGINE "[%p]::performAudioEngineStart() - Getting an anchor for the first timestamp.", this);
		mUSBAudioDevice->updateUSBCycleTime();
		FailIf(mUSBAudioDevice->mAnchorTime.n == 0, Exit);
	}

	mUSBAudioDevice->calculateOffset();		// <rdar://problem/7666699>

#ifdef TARGET_PDJ
 	super::numErasesPerBuffer = kMinimumNumErasesPerBuffer;
 
 	IOService::requireMaxBusStall(10 * 1000);
 	IOService::requireMaxInterruptDelay(10 * 1000);
#endif

	resultCode = kIOReturnSuccess;

	if (mPlugin)
	{
		mPlugin->pluginStart();
	}

	for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
	{
		PdjUsbAudioStream	*audioStream;

		audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));

		if (audioStream && audioStream->mPlugin)
		{
			audioStream->mPlugin->pluginStart();
		}
	}

	if (!mUSBStreamRunning)
	{
		UInt64		currentUSBFrame;

#ifdef TARGET_PDJ
 		mShouldRestartStreaming = false;
 
 		mUSBInputIndex	= 0;
 		mUSBOutputIndex	= 0;
 
 		::bzero(mUSBInputSizeList, sizeof(mUSBInputSizeList));
#endif

#if DEBUGLATENCY
		mHaveClipped = false;
#endif

		for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
		{
			PdjUsbAudioStream	*audioStream;

			resultCode = kIOReturnError;

			audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));
			FailIf(NULL == audioStream, Exit);

			resultCode = audioStream->prepareUSBStream();
			FailIf(kIOReturnSuccess != resultCode, Exit);

			UInt32		lockDelayFrames = audioStream->getLockDelayFrames();

			if (usbFramesToDelay < lockDelayFrames)
			{
				usbFramesToDelay = lockDelayFrames;
			}
		}

		currentUSBFrame = mUSBAudioDevice->getUSBFrameNumber();
		// Added an offset to compensate for the time needed for some of the operation in PdjUsbAudioStream::startUSBStream();
		currentUSBFrame += kStartDelayOffset;

#ifdef DEBUGLOGGING
		mNextSampleFrame = 0xFFFFFFFF;
#endif

		for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
		{
			PdjUsbAudioStream	*audioStream;

			resultCode = kIOReturnError;

			audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));
			FailIf(NULL == audioStream, Exit);

			resultCode = audioStream->startUSBStream(currentUSBFrame, usbFramesToDelay);
			FailIf(kIOReturnSuccess != resultCode, Exit);
		}

		if (kIOReturnSuccess == resultCode)
		{
			if (0 != usbFramesToDelay)
			{
				// [rdar://5083342] Sleep for the amount of frames delayed.
				IOSleep(usbFramesToDelay);
			}

			mUSBStreamRunning = true;
		}
	}
	else
	{
		DP(DBG_WARNING, "! " LOGSTRAUDIOENGINE "::performAudioEngineStart() - Streaming Already Running.");
	}

Exit:
	if (resultCode != kIOReturnSuccess)
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIOENGINE "[%p]::performAudioEngineStart() - NOT started, error = 0x%x", this, resultCode);

		// It is safe to call performAudioEngineStop from here because we are failing the performAudioEngineStart method.
		this->performAudioEngineStop();

		if (
				(
					(kIOReturnNotResponding == resultCode)
				||
					(kIOReturnExclusiveAccess == resultCode)
				)
		   &&
				(! mUSBAudioDevice->recoveryRequested())
		   )
		{
			// The device is in an odd state. We should ask for a device recovery attempt.
			if (mUSBAudioDevice)
			{
				DP(DBG_ERROR, "! " LOGSTRAUDIOENGINE "[%p]::performAudioEngineStart() - Device not responding! Requesting a recovery attempt.", this);
				mUSBAudioDevice->requestDeviceRecovery();
			}
		}
	}
	else
	{
		DP(DBG_INFO, "\n");
		DP(DBG_INFO, "      sampleRate->whole = %u", super::getSampleRate()->whole);
		DP(DBG_INFO, "\n");
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::performAudioEngineStart() = 0x%x", this, resultCode);
	return resultCode;
}

/****************************************************************************/
/**
 * @brief	オーディオ I/O エンジン停止．
 *			IOAudioEngine::performAudioEngineStop() をオーバーライド．
 *			ストリーミングを停止する．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioEngine::performAudioEngineStop(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::performAudioEngineStop()", this);

	PdjUsbAudioStream	*audioStream;

	for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
	{
		audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));

		if (audioStream && audioStream->mPlugin)
		{
			audioStream->mPlugin->pluginStop();
		}
	}

	if (mPlugin)
	{
		mPlugin->pluginStop();
	}

	if (mUSBStreamRunning)
	{
		for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
		{
			audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));

			if (NULL != audioStream)
			{
				audioStream->stopUSBStream();
			}
		}
	}
	else
	{
		DP(DBG_WARNING, "! " LOGSTRAUDIOENGINE "::performAudioEngineStop() - Streaming Already Stopped.");
	}

	mUSBStreamRunning = false;

#ifdef TARGET_PDJ
 	IOService::requireMaxInterruptDelay(0);
 	IOService::requireMaxBusStall(0);
#endif

	if (NULL != mUSBAudioDevice)							// <rdar://problem/7779397>
	{
		mUSBAudioDevice->mAnchorTime.deviceStart = false;	// <rdar://problem/7666699>
	}

	DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::performAudioEngineStop() - stopped", this);

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::performAudioEngineStop()", this);
	return kIOReturnSuccess;
}

/****************************************************************************/
/**
 * @brief	オーディオフォーマット変更．
 *			IOAudioEngine::performFormatChange() をオーバーライド．
 *
 * @param[in]		*audioStream	フォーマット変更の対象となるストリーム．
 * @param[in]		*newFormat		変更後のフォーマット．
 * @param[in]		*newSampleRate	変更後のサンプリング周波数．
 *
 * @return			IOReturn コード．
 */
// This gets called when the HAL wants to select one of the different formats that we made available via mMainStream->addAvailableFormat
IOReturn
PdjUsbAudioEngine::performFormatChange(
	IOAudioStream				*audioStream,
	const IOAudioStreamFormat	*newFormat,
	const IOAudioSampleRate		*newSampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::performFormatChange(%p, %p, %p)",
		this, audioStream, newFormat, newSampleRate);

	IOReturn					result = kIOReturnError;
	UInt32						controllerResult;
	bool						streamIsRunning = false;

	FailIf(NULL == mUSBAudioDevice, Exit);

	streamIsRunning = mUSBStreamRunning;

	if (streamIsRunning)
	{
		super::pauseAudioEngine();
	}

	controllerResult = mUSBAudioDevice->formatChangeController(this, audioStream, newFormat, newSampleRate);

	switch (controllerResult)
	{
		case kPdjUsbFormatChangeNormal:
			result = kIOReturnSuccess;
			break;

		case kPdjUsbFormatChangeForced:
			DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::performFormatChange() - This request was forced.", this);
			result = kIOReturnSuccess;
			break;

		case kPdjUsbFormatChangeForceFailure:
			DP(DBG_ERROR, "! " LOGSTRAUDIOENGINE "[%p]::performFormatChange() - Force of this request was attempted but failed.", this);
			result = kIOReturnSuccess;
			break;

		case kPdjUsbFormatChangeError:
		default:
			DP(DBG_ERROR, "! " LOGSTRAUDIOENGINE "[%p]::performFormatChange() - Error encountered.", this);
			result = kIOReturnError;
	}

Exit:
	if (streamIsRunning)
	{
		super::resumeAudioEngine();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::performFormatChange() = 0x%x", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオフォーマット変更．
 *			PdjUsbAudioDevice::formatChangeController() から呼び出される．
 *
 * @param[in]		*audioStream		PdjUsbAudioStream クラスのインスタンス．
 * @param[in]		*newFormat			変更後のオーディオフォーマット．
 * @param[in]		*newSampleRate		変更後のサンプリング周波数．
 *
 * @return			IOReturn コード．
 */
// [rdar://3918719] The following method now does the work of performFormatChange
// after being regulated by PdjUsbAudioDevice::formatChangeController().
IOReturn
PdjUsbAudioEngine::controlledFormatChange(
	IOAudioStream				*audioStream,
	const IOAudioStreamFormat	*newFormat,
	const IOAudioSampleRate		*newSampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::controlledFormatChange(%p, %p, %p)", this, audioStream, newFormat, newSampleRate);

	IOReturn							result = kIOReturnError;
	PdjUsbAudioStream					*pdjUsbAudioStream;
	PdjUsbConfigurationDictionary		*configDictionary = NULL;
	UInt32								numSamplesInBuffer;
	bool								needToChangeChannels;
	OSDictionary						*channelNamesDictionary = NULL;		// <rdar://6430836>
	UInt32								averageSampleRate;

	FailIf(NULL == mStreamInterfaceNumberArray, Exit);
	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == mUSBAudioDevice->mControlInterface, Exit);		// <rdar://7085810>
	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	if ((newFormat == NULL) && (audioStream != NULL))
	{
		result = kIOReturnSuccess;
		goto Exit;
	}

	if (NULL == audioStream)
	{
		// This is an emergency format change request initiated to keep the input and output at the same sample rate.
		// We need to supply the correct stream for this engine.
		audioStream = (NULL != mMainOutputStream) ? mMainOutputStream : mMainInputStream;

		// We also need to get the format if it wasn't supplied.
		if (NULL == newFormat)
		{
			newFormat = audioStream->getFormat();
		}
	}

	FailIf(NULL == audioStream,	Exit);
	FailIf(NULL == newFormat,	Exit);

	pdjUsbAudioStream = OSDynamicCast(PdjUsbAudioStream, audioStream);
	FailIf(NULL == pdjUsbAudioStream, Exit);

	result = pdjUsbAudioStream->controlledFormatChange(newFormat, newSampleRate);
	FailIf(kIOReturnSuccess != result, Exit);

	for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
	{
		PdjUsbAudioStream	*stream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));
		FailIf(NULL == stream, Exit);

		if (stream != pdjUsbAudioStream)
		{
			stream->controlledFormatChange(stream->getFormat(), newSampleRate);
		}
	}

	if (newFormat->fNumChannels != audioStream->getFormat()->fNumChannels)
	{
		needToChangeChannels = true;

		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::controlledFormatChange() - Need to adjust channel controls (cur = %u, new = %u)",
			this, audioStream->getFormat()->fNumChannels, newFormat->fNumChannels);

		if (kIOAudioStreamDirectionOutput == pdjUsbAudioStream->getDirection())
		{
			// check for mono mode
			if (1 == newFormat->fNumChannels)
			{
				mUSBAudioDevice->mTopology->setMonoState(true);
			}
			else
			{
				mUSBAudioDevice->mTopology->setMonoState(false);
			}
		}
	}
	else
	{
		needToChangeChannels = false;
		// DP(DBG_INFO, "No need to adjust channel controls.");
	}

	if (newSampleRate)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::controlledFormatChange() - Changing sampling rate to %u", this, newSampleRate->whole);
		mCurSampleRate = *newSampleRate;
	}
	else
	{
		DP(DBG_INFO, "Keeping existing sampling rate of %u", mCurSampleRate.whole);
	}

	// If both input & output streams are present, then use the output stream format.
	if ((pdjUsbAudioStream == mMainOutputStream) || ((pdjUsbAudioStream == mMainInputStream) && (NULL == mMainOutputStream)))
	{
		if (mPlugin)
		{
			mPlugin->pluginSetFormat(newFormat, &mCurSampleRate);
		}
	}

	super::setNumSampleFramesPerBuffer(0);

	averageSampleRate = mCurSampleRate.whole;	// Set this as the default until we are told otherwise

	DP(DBG_INFO, "averageSampleRate = %u", averageSampleRate);

	// Need a minimum of two pages in OHCI/UHCI
	numSamplesInBuffer  = averageSampleRate / 4;
	numSamplesInBuffer += (PAGE_SIZE * 2 - 1);
	numSamplesInBuffer /=  PAGE_SIZE * 2;
	numSamplesInBuffer *=  PAGE_SIZE * 2;

	DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::controlledFormatChange() - New numSamplesInBuffer = %u", this, numSamplesInBuffer);

	super::setNumSampleFramesPerBuffer(numSamplesInBuffer);

	if (needToChangeChannels)
	{
		super::beginConfigurationChange();

		// <rdar://6430836> Dictionary for channel names
		channelNamesDictionary = OSDictionary::withCapacity(4);
		FailIf(NULL == channelNamesDictionary, Exit);

		IORegistryEntry::setProperty(kIOAudioEngineFullChannelNamesKey, channelNamesDictionary);
		channelNamesDictionary->release();

		super::removeAllDefaultAudioControls();

		for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
		{
			PdjUsbAudioStream	*stream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));

			if (NULL != stream)
			{
				mUSBAudioDevice->createControlsForInterface(this, stream->mInterfaceNumber, stream->mAlternateSettingID);
			}
		}

		// <rdar://5811247>
		if (IP_VERSION_02_00 == mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
		{
			// Clock selector control for USB Audio 2.0 devices
			this->doClockSelectorSetup(pdjUsbAudioStream->mInterfaceNumber, pdjUsbAudioStream->mAlternateSettingID, mCurSampleRate.whole);
		}

		super::completeConfigurationChange();
	}

	DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::controlledFormatChange() - newFormat->fNumChannels = %u, newFormat->fBitWidth %u",
		this, newFormat->fNumChannels, newFormat->fBitWidth);

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::controlledFormatChange() = 0x%x", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	使用可能なオーディオフォーマットを更新．
 *
 * @param[in]		sampleRate			サンプリング周波数．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbAudioEngine::republishAvailableFormats(
	IOAudioSampleRate	sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::republishAvailableFormats()");

	PdjUsbConfigurationDictionary		*configDictionary = NULL;

	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
	{
		PdjUsbAudioStream	*audioStream = OSDynamicCast(
												PdjUsbAudioStream,
												mIOAudioStreamArray->getObject(streamIndex)
										   );
		if (NULL != audioStream)
		{
			audioStream->clearAvailableFormats();
			audioStream->addAvailableFormats(configDictionary, sampleRate);
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::republishAvailableFormats()");
	return kIOReturnSuccess;
}

/****************************************************************************/
/**
 * @brief	入力データを整数型から浮動小数点型に変換．
 *			IOAudioEngine::convertInputSamples() をオーバーライド．
 *			オーディオ入力ストリーミング時、上位から定期的に呼び出される．
 *
 * @param[in]		*sampleBuf			変換前の整数型オーディオバッファ．
 * @param[out]		*destBuf			変換後の浮動小数点型オーディオバッファ．
 * @param[in]		firstSampleFrame	変換開始位置（サンプルフレーム単位）．
 * @param[in]		numSampleFrames		変換するサンプルフレーム数．
 * @param[in]		*streamFormat		オーディオフォーマット．
 * @param[in]		*audioStream		PdjUsbAudioStream クラスのインスタンス．
 *
 * @return			IOReturn コード．
 */
// When convertInputSamples is called, we have a window of samples that might possibly still be in transit on the USB bus.
// The number of samples that might be in transit depends on how long our USB read completion routines have been held off.
// Best case is that we have already coalesced all the samples that have been recorded because we weren't held off.
// Worst case is that we've been held off for longer than (framesPerList * mNumUSBFrameListsToQueue) milliseconds and we haven't
// coalesced anything that's been recorded and we don't have any more reads queued up and we're being asked for something
// that hasn't been recorded.
// The normal case should be not much worse than the best case and that's what we're trying to make better.
// The case we are really trying to fix is when we have been held off for (framesPerList * 1 or 2) ms and we have the data
// that's being asked for, it just hasn't been coalesced yet.
// What we have is a window of samples that are outstanding, either they have been recorded and we haven't gotten the
// completion routine called for them, or they are still in the future and haven't been recorded yet. We must figure out
// where that window is in our buffer and if the request is for sample inside of that window, coalesce and return those
// samples. If the request is for samples outside of that window, just return those samples because there is no possibility
// that they are in the recording buffer (it's either the worst case, or someone is asking for old data).
// The window goes from (mCurrentFrameList + 1) to (mCurrentFrameList + numFramesListsToQueue) which is at most
// mReadUSBFrameListSize * mNumUSBFrameListsToQueue in size. It's actually probably smaller than that, but we'll assume that it's
// that big and if can't get coalesce enough bytes from the read buffer, then we'll return old data since there is nothing
// else that we can do anyway (perhaps we could return an error to the HAL too).
IOReturn
PdjUsbAudioEngine::convertInputSamples(
	const void					*sampleBuf,
	void						*destBuf,
	UInt32						firstSampleFrame,
	UInt32						numSampleFrames,
	const IOAudioStreamFormat	*streamFormat,
	IOAudioStream				*audioStream
)
{
	DP(DBG_READ, "+ " LOGSTRAUDIOENGINE "::convertInputSamples(%p, %p, %u, %u, %p, %p)",
		sampleBuf, destBuf, firstSampleFrame, numSampleFrames, streamFormat, audioStream);

	IOReturn					result = kIOReturnSuccess;
	IOReturn					coalescenceErrorCode = kIOReturnSuccess;
	PdjUsbAudioStream			*pdjUsbAudioStream;
	UInt32						lastSampleByte;
	UInt32						windowStartByte;

	// <rdar://7298196> Verify that the sample & destination buffers are valid.
	// Also validate that the firstSampleFrame is not equal or larger than the number of sample frames in the buffer.
	FailWithAction(NULL == sampleBuf,	result = kIOReturnError, Exit);
	FailWithAction(NULL == destBuf,		result = kIOReturnError, Exit);
	FailWithAction(super::getNumSampleFramesPerBuffer() <= firstSampleFrame, result = kIOReturnError, Exit);

	pdjUsbAudioStream = OSDynamicCast(PdjUsbAudioStream, audioStream);
	FailWithAction(NULL == pdjUsbAudioStream, result = kIOReturnError, Exit);

#ifdef DEBUGLOGGING
	if (mNextSampleFrame != 0xFFFFFFFF)
	{
		if (mNextSampleFrame < firstSampleFrame)
		{
			DP(DBG_WARNING, "Skip %u frames", firstSampleFrame - mNextSampleFrame);
		}
		else if (mNextSampleFrame > firstSampleFrame)
		{
			DP(DBG_WARNING, "Overlap %u frames", mNextSampleFrame - firstSampleFrame);
		}
	}

	mNextSampleFrame = firstSampleFrame + numSampleFrames;

	if (mNextSampleFrame >= super::getNumSampleFramesPerBuffer())
	{
		mNextSampleFrame -= super::getNumSampleFramesPerBuffer();
	}
#endif

//	if (!pdjUsbAudioStream->mHaveTakenFirstTimeStamp)
//	{
//		DP(DBG_WARNING, "! " LOGSTRAUDIOENGINE "::convertInputSamples() - called before first time stamp!");
//	}

	// <rdar://7298196> Only perform the conversion if the audio engine is running.
	if (mUSBStreamRunning)
	{
		pdjUsbAudioStream->queueInputFrames();

#if 1 // enabled for [3091812]
		// <rdar://problem/7378275>
		if (pdjUsbAudioStream->mCoalescenceMutex)
		{
			IORecursiveLockLock(pdjUsbAudioStream->mCoalescenceMutex);
		}

		lastSampleByte = (firstSampleFrame + numSampleFrames) * streamFormat->fNumChannels * (streamFormat->fBitWidth / 8);

		// Is the request inside our window of possibly recorded samples?
		if ((pdjUsbAudioStream->mBufferOffset + 1) > pdjUsbAudioStream->getSampleBufferSize())
		{
			windowStartByte = 0;
		}
		else
		{
			windowStartByte = pdjUsbAudioStream->mBufferOffset + 1;
		}

		UInt32		windowEndByte;

		windowEndByte = windowStartByte + (pdjUsbAudioStream->mNumUSBFrameListsToQueue * pdjUsbAudioStream->mReadUSBFrameListSize);

		if (windowEndByte > pdjUsbAudioStream->getSampleBufferSize())
		{
			windowEndByte -= pdjUsbAudioStream->getSampleBufferSize();
		}

		if (
				((windowStartByte < lastSampleByte ) && (lastSampleByte  < windowEndByte  ))
		   ||
				((lastSampleByte  < windowEndByte  ) && (windowEndByte   < windowStartByte))
		   ||
				((windowEndByte   < windowStartByte) && (windowStartByte < lastSampleByte ))
		   )
		{
			// DP(DBG_INFO, "%u, %u, %u, %u, %u",
			//		firstSampleFrame * 4,
			//		numSampleFrames,
			//		lastSampleByte,
			//		windowStartByte,
			//		windowEndByte);

#ifdef BUG_FIX
			UInt32		numBytesToCoalesce;

			if (pdjUsbAudioStream->mBufferOffset < lastSampleByte)
			{
				numBytesToCoalesce = lastSampleByte - pdjUsbAudioStream->mBufferOffset;
			}
			else
			{
				// Have to wrap around the buffer.
				numBytesToCoalesce = pdjUsbAudioStream->getSampleBufferSize() - pdjUsbAudioStream->mBufferOffset + lastSampleByte;
			}

			if (numBytesToCoalesce < (pdjUsbAudioStream->getSampleBufferSize() / 2))
			{
				// [rdar://5355808] Keep track of sample data underruns.
				coalescenceErrorCode = pdjUsbAudioStream->coalesceInputSamples(numBytesToCoalesce, NULL);

				DP(DBG_READ, "! " LOGSTRAUDIOENGINE "::convertInputSamples() - Coalesce from convert %u bytes",
					numBytesToCoalesce);
			}
#else
			if (pdjUsbAudioStream->mBufferOffset < lastSampleByte)
			{
				// [rdar://5355808] Keep track of sample data underruns.
				coalescenceErrorCode = pdjUsbAudioStream->coalesceInputSamples(lastSampleByte - pdjUsbAudioStream->mBufferOffset, NULL);

				DP(DBG_WARNING, "! " LOGSTRAUDIOENGINE "::convertInputSamples() - Coalesce from convert %u bytes",
					lastSampleByte - pdjUsbAudioStream->mBufferOffset);
			}
			else
			{
				// Have to wrap around the buffer.
				UInt32		numBytesToCoalesce = pdjUsbAudioStream->getSampleBufferSize() - pdjUsbAudioStream->mBufferOffset + lastSampleByte;

				// [rdar://5355808] Keep track of sample data underruns.
				coalescenceErrorCode = pdjUsbAudioStream->coalesceInputSamples(numBytesToCoalesce, NULL);

				DP(DBG_WARNING, "! " LOGSTRAUDIOENGINE "::convertInputSamples() - Coalesce from convert %u bytes (wrapping)",
					numBytesToCoalesce);
			}
#endif
		}

		// <rdar://problem/7378275>
		if (pdjUsbAudioStream->mCoalescenceMutex)
		{
			IORecursiveLockUnlock(pdjUsbAudioStream->mCoalescenceMutex);
		}
#endif

		result = ::convertFromPdjUsbAudioInputStream_NoWrap(sampleBuf, destBuf, firstSampleFrame, numSampleFrames, streamFormat);

		if (pdjUsbAudioStream->mPlugin)
		{
			pdjUsbAudioStream->mPlugin->pluginProcessInput((float *)destBuf, numSampleFrames, streamFormat->fNumChannels);
		}

		if (mPlugin)
		{
			if (pdjUsbAudioStream == mMainInputStream)
			{
				mPlugin->pluginProcessInput((float *)destBuf, numSampleFrames, streamFormat->fNumChannels);
			}
		}
#if 0
		if (0 == firstSampleFrame) // if this convert began with sample 0
		{
			DP(DBG_INFO, "sample 0 converted on USB frame %llu", mUSBAudioDevice->getUSBFrameNumber());
		}
#endif
	}

Exit:
	// [rdar://5355808] Keep track of sample data underruns.
	if (kIOReturnSuccess != coalescenceErrorCode)
	{
		result = coalescenceErrorCode;
	}

	if (kIOReturnSuccess != result)
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIOENGINE "::convertInputSamples() = 0x%x", result);
	}

	DP(DBG_READ, "- " LOGSTRAUDIOENGINE "::convertInputSamples() = 0x%x", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	出力データを浮動小数点型から整数型に変換．
 *			IOAudioEngine::clipOutputSamples() をオーバーライド．
 *			オーディオ出力ストリーミング時、上位から定期的に呼び出される．
 *
 * @param[in]		*mixBuf				変換前の浮動小数点型オーディオバッファ．
 * @param[out]		*sampleBuf			変換後の整数型オーディオバッファ．
 * @param[in]		firstSampleFrame	変換開始位置（サンプルフレーム単位）．
 * @param[in]		numSampleFrames		変換するサンプルフレーム数．
 * @param[in]		*streamFormat		オーディオフォーマット．
 * @param[in]		*audioStream		PdjUsbAudioStream クラスのインスタンス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioEngine::clipOutputSamples(
	const void					*mixBuf,
	void						*sampleBuf,
	UInt32						firstSampleFrame,
	UInt32						numSampleFrames,
	const IOAudioStreamFormat	*streamFormat,
	IOAudioStream				*audioStream
)
{
	DP(DBG_WRITE, "+ " LOGSTRAUDIOENGINE "::clipOutputSamples()");

	IOReturn					result = kIOReturnError;
	PdjUsbAudioStream			*pdjUsbAudioStream;

	pdjUsbAudioStream = OSDynamicCast(PdjUsbAudioStream, audioStream);
	FailIf(NULL == pdjUsbAudioStream, Exit);

	pdjUsbAudioStream->queueOutputFrames();

	if (streamFormat->fIsMixable)
	{
		if (pdjUsbAudioStream->mPlugin)
		{
			pdjUsbAudioStream->mPlugin->pluginProcessOutput(
											(Float32 *)mixBuf + (firstSampleFrame * streamFormat->fNumChannels),
											numSampleFrames,
											streamFormat->fNumChannels
										);
		}

		if (mPlugin)
		{
			if (pdjUsbAudioStream == mMainOutputStream)
			{
				mPlugin->pluginProcessOutput(
								(Float32 *)mixBuf + (firstSampleFrame * streamFormat->fNumChannels),
								numSampleFrames,
								streamFormat->fNumChannels
						 );
			}
		}

		result = ::clipPdjUsbAudioToOutputStream(mixBuf, sampleBuf, firstSampleFrame, numSampleFrames, streamFormat);

#if DEBUGLATENCY
		if (!mHaveClipped)
		{
			mHaveClipped = true;

			DP(DBG_INFO, "! FIRST clip at sample frame %u when current USB Frame is %llu",
				firstSampleFrame, mUSBAudioDevice->getUSBFrameNumber());
			DP(DBG_INFO, "! " LOGSTRAUDIOENGINE "::clipOutputSamples() - Sample frame %u queued for USB frame %llu (current frame = %llu)",
				firstSampleFrame, pdjUsbAudioStream->getQueuedFrameForSample(firstSampleFrame), mUSBAudioDevice->getUSBFrameNumber());
		}
		else if ((0 == firstSampleFrame) || true)
		{
			DP(DBG_INFO, "! " LOGSTRAUDIOENGINE "::clipOutputSamples() - Index %u of sample buffer written when current USB frame is %llu",
				firstSampleFrame, mUSBAudioDevice->getUSBFrameNumber());
			DP(DBG_INFO, "! " LOGSTRAUDIOENGINE "::clipOutputSamples() - Sample frame %u queued for USB frame %llu (current frame = %llu)",
				firstSampleFrame, pdjUsbAudioStream->getQueuedFrameForSample(firstSampleFrame), mUSBAudioDevice->getUSBFrameNumber());
		}
#endif

#ifdef TARGET_PDJ
#else
		if (mUHCISupport)
		{
			// If we've wrapped, copy to sampleBuffer extension create/keep mLastClippedFrame for nonmixable cases.
			// If we return and see frame count lower than mLastClippedFrame, we've wrapped.
			UInt16		alternateFrameSize = pdjUsbAudioStream->getAlternateFrameSize();
			UInt32		start = firstSampleFrame * streamFormat->fNumChannels * (streamFormat->fBitWidth / 8);
						// first byte index of first sample that is being clipped

			if (start < alternateFrameSize)
			{
				// This is correct because mAverageFrameSize is in bytes. size is the amount of data we want to copy into the scribble-ahead area.
				UInt32		size = alternateFrameSize - start;

				::memcpy(
					&(((char *)sampleBuf)[pdjUsbAudioStream->mSampleBufferSize + start]),
					&(((char *)sampleBuf)[                                       start]),
					size
				);

				DP(DBG_UHCI, "? " LOGSTRAUDIOENGINE "::clipOutputSamples() - firstSampleFrame = %u. Just copied %u bytes from %u to %u \n",
					firstSampleFrame, size, start, start + pdjUsbAudioStream->mSampleBufferSize);
			}
		}
#endif
	}
	else
	{
		// Non-mixable case
		UInt32		offset = firstSampleFrame * streamFormat->fNumChannels * (streamFormat->fBitWidth / 8);

		::memcpy(
			(UInt8 *)sampleBuf + offset,
			(UInt8 *)mixBuf,
			numSampleFrames * streamFormat->fNumChannels * (streamFormat->fBitWidth / 8)
		);

		result = kIOReturnSuccess;
	}

Exit:
	DP(DBG_WRITE, "- " LOGSTRAUDIOENGINE "::clipOutputSamples() = 0x%x", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオ出力用リングバッファをリセット．
 *			IOAudioEngine::resetClipPosition() をオーバーライド．
 *
 * @param[in]		*audioStream		リセットされた PdjUsbAudioStream のインスタンス．
 * @param[in]		clipSampleFrame		リセットが発生したときのバッファ位置．
 */
void
PdjUsbAudioEngine::resetClipPosition(
	IOAudioStream	*audioStream,
	UInt32			clipSampleFrame
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::resetClipPosition()");

	PdjUsbAudioStream	*pdjUsbAudioStream;

	pdjUsbAudioStream = OSDynamicCast(PdjUsbAudioStream, audioStream);
	FailIf(NULL == pdjUsbAudioStream, Exit);

	if (pdjUsbAudioStream->mPlugin)
	{
		pdjUsbAudioStream->mPlugin->pluginReset();
	}

	if (mPlugin)
	{
		mPlugin->pluginReset();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::resetClipPosition()");
	return;
}

#ifdef TARGET_PDJ
#else
/****************************************************************************/
/**
 * @brief	出力データを消去．
 *			IOAudioEngine::eraseOutputSamples() をオーバーライド．
 *
 * @param[out]		*mixBuf				消去対象の浮動小数点型オーディオ出力バッファ．
 * @param[out]		*sampleBuf			消去対象の整数型オーディオ出力バッファ．
 * @param[in]		firstSampleFrame	消去開始位置（サンプルフレーム単位）．
 * @param[in]		numSampleFrames		消去するサンプルフレーム数．
 * @param[in]		*streamFormat		オーディオフォーマット．
 * @param[in]		*audioStream		PdjUsbAudioStream クラスのインスタンス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioEngine::eraseOutputSamples(
	const void					*mixBuf,
	void						*sampleBuf,
	UInt32						firstSampleFrame,
	UInt32						numSampleFrames,
	const IOAudioStreamFormat	*streamFormat,
	IOAudioStream				*audioStream
)
{
	DP(DBG_WRITE, "+ " LOGSTRAUDIOENGINE "::eraseOutputSamples()");

	super::eraseOutputSamples(mixBuf, sampleBuf, firstSampleFrame, numSampleFrames, streamFormat, audioStream);

	// if on a UHCI connection and using output, erase extended buffer area; necessary to avoid an audio artifact after stopping the stream
	if (mUHCISupport)
	{
		PdjUsbAudioStream	*pdjUsbAudioStream = OSDynamicCast(PdjUsbAudioStream, audioStream);
		FailIf(NULL == pdjUsbAudioStream, Exit);

		if (pdjUsbAudioStream->getDirection() == kIOAudioStreamDirectionOutput)
		{
			UInt16		alternateFrameSize = pdjUsbAudioStream->getAlternateFrameSize();
			UInt32		start = firstSampleFrame * streamFormat->fNumChannels * (streamFormat->fBitWidth / 8);
						// first byte index of first sample that is being clipped

			if (start < alternateFrameSize)
			{
				UInt32		size = alternateFrameSize - start;
				::bzero(&(((char *)sampleBuf)[pdjUsbAudioStream->mSampleBufferSize + start]), size);
			}
		}
	}

Exit:
	DP(DBG_WRITE, "- " LOGSTRAUDIOENGINE "::eraseOutputSamples()");
	return kIOReturnSuccess;
}
#endif

/****************************************************************************/
/**
 * @brief	オーディオリングバッファの現在位置を取得．
 *			IOAudioEngine::getCurrentSampleFrame() をオーバーライド．
 *
 * @return			リングバッファの現在位置（サンプルフレーム単位）．
 */
UInt32
PdjUsbAudioEngine::getCurrentSampleFrame(
	void
)
{
	DP(DBG_VERBOSE, "+ " LOGSTRAUDIOENGINE "::getCurrentSampleFrame()");

	UInt32		currentSampleFrame = 0;

	FailIf((NULL == mMainOutputStream) && (NULL == mMainInputStream), Exit);

	// <rdar://6343818> getCurrentSampleFrame() is used by the audio engine to perform erase on the
	// output stream, so return the output stream value if it is present.
	if (mMainOutputStream)
	{
		currentSampleFrame = mMainOutputStream->getCurrentSampleFrame();
	}
	else if (mMainInputStream)
	{
		currentSampleFrame = mMainInputStream->getCurrentSampleFrame();
	}

Exit:
	DP(DBG_VERBOSE, "- " LOGSTRAUDIOENGINE "::getCurrentSampleFrame()");
	return currentSampleFrame;
}

/****************************************************************************/
/**
 * @brief	タイムスタンプを上位に報告．
 *			オーディオリングバッファが末尾に達し、先頭に折り返したときは
 *			折り返し発生時刻を上位に報告しなければならない．
 *
 * @param[in]		incrementLoopCount	true ならリングバッファのループ回数をインクリメントする．
 *										初回のタイムスタンプ報告では false とする．
 * @param[in]		*timestamp			折り返し発生時のタイムスタンプ．
 */
void
PdjUsbAudioEngine::takeTimeStamp(
	bool			incrementLoopCount,
	AbsoluteTime	*timestamp
)
{
#if LOGTIMESTAMPS
	UInt64			time_nanos;

	::absolutetime_to_nanoseconds(*timestamp, &time_nanos);

//	if (getDirection() == kIOAudioStreamDirectionInput)
//	if (getDirection() == kIOAudioStreamDirectionOutput)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::takeTimeStamp(%d, %p) = %llu ns",
			this, incrementLoopCount, timestamp, time_nanos);
	}
#endif

#ifdef TARGET_PDJ
 	UInt64		time_nanos;
 
 	::absolutetime_to_nanoseconds(*timestamp, &time_nanos);
 
 	UInt64		sampling_rate		= time_nanos - mLastTimeStamp_nanos;
 	UInt64		numSamplesInBuffer	= 0;
 
 	if (mMainOutputStream)
 	{
 		numSamplesInBuffer = mMainOutputStream->mNumSamplesInBuffer;
 	}
 
 	if (sampling_rate > 0)
 	{
 		sampling_rate = numSamplesInBuffer / sampling_rate;
 	}
 
 	DP(DBG_INFO, "OUT ----------------------------> %llu, %llu", time_nanos, sampling_rate);
 
 	mLastTimeStamp_nanos = time_nanos;
#endif

	super::takeTimeStamp(incrementLoopCount, timestamp);
}

/****************************************************************************/
/**
 * @brief	登録済みの既定オーディオコントロールの集合をコピー．
 *
 * @return			コピーされた既定オーディオコントロールの集合．
 */
// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
OSSet *
PdjUsbAudioEngine::copyDefaultAudioControls(
	void
)
{
	// <rdar://7695055> If there are controls present, copy it.
	if ((defaultAudioControls) && (defaultAudioControls->getCount()))
	{
		return OSSet::withSet(defaultAudioControls, defaultAudioControls->getCount());
	}
	else
	{
		return NULL;
	}
}

/****************************************************************************/
/**
 * @brief	タイマハンドラ．
 *			PdjUsbAudioDevice のタイマハンドラから呼び出される．
 */
// <rdar://5811247>
void
PdjUsbAudioEngine::runPolledTask(
	void
)
{
	DP(DBG_TIMER, "+ " LOGSTRAUDIOENGINE "::runPolledTask()");

	FailIf(NULL == mUSBAudioDevice, Exit);

	// <rdar://5811247> Restore the clock selector value when the change failed.
	if (mRestoreClockSelection)
	{
		OSNumber	*number = OSNumber::withNumber(mRestoreClockSelectionValue, 32);

		if (NULL != number)
		{
			if (NULL != mClockSelectorControl)
			{
				mClockSelectorControl->hardwareValueChanged(number);
			}

			number->release();
		}

		mRestoreClockSelection = false;
	}
#if POLLCLOCKSTATUS
	// <rdar://5811247>
	else if (mShouldPollClockStatus)
	{
		if (0 == mPollClockStatusCounter)
		{
			this->updateClockStatus(mCurrentClockSourceID);
		}

		mPollClockStatusCounter++;

		if ((1024 / kRefreshInterval) == mPollClockStatusCounter)	// Poll once every 1024 millisecond.
		{
			mPollClockStatusCounter = 0;
		}
	}
#endif
#ifdef TARGET_PDJ
 	else if (mShouldRestartStreaming)
 	{
 		if (mUSBStreamRunning)
 		{
 			super::pauseAudioEngine();
 				super::beginConfigurationChange();
 					super::hardwareSampleRateChanged(&mCurSampleRate);
 				super::completeConfigurationChange();
 			super::resumeAudioEngine();
 		}
 
 		mShouldRestartStreaming = false;
 	}
#endif

Exit:
	DP(DBG_TIMER, "- " LOGSTRAUDIOENGINE "::runPolledTask()");
	return;
}

/****************************************************************************/
/**
 * @brief	チャンネル名を更新．
 *
 * @param[in]		*thePath		Audio Function Topology 上のパス．
 * @param[in]		interfaceNum	USB interface 番号．
 * @param[in]		altSettingNum	alternate setting 番号．
 */
// <rdar://6430836> <rdar://problem/6706026> Update the channel names in the IOAudioEngineFullChannelNames dictionary for the specified stream.
void
PdjUsbAudioEngine::updateChannelNames(
	OSArray			*thePath,
	UInt8			interfaceNum,
	UInt8			altSettingNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::updateChannelNames(%d, %d)", this, interfaceNum, altSettingNum);

	OSDictionary						*oldChannelNamesDictionary = NULL;	// <rdar://6500720>
	OSDictionary						*newChannelNamesDictionary = NULL;	// <rdar://6500720>
	PdjUsbConfigurationDictionary		*configDictionary = NULL;
	UInt8								controlInterfaceNum;
	char								keyString[32];
	OSString							*nameString = NULL;
	AudioClusterDescriptor				clusterDescriptor = { 0, 0, 0 };	// <rdar://problem/6706026>
	UInt8								subType;
	OSNumber							*number;
	UInt8								unitID = 0;

	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == mUSBAudioDevice->mControlInterface, Exit);		// <rdar://7085810>

	controlInterfaceNum = mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	for (UInt32 streamIndex = 0; streamIndex < mIOAudioStreamArray->getCount(); streamIndex++)
	{
		PdjUsbAudioStream	*audioStream = OSDynamicCast(PdjUsbAudioStream, mIOAudioStreamArray->getObject(streamIndex));

		if (
				(NULL != audioStream)
		   &&
				(interfaceNum == audioStream->mInterfaceNumber)
		   &&
				(altSettingNum == audioStream->mAlternateSettingID)
		   )
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateChannelNames(%d, %d) - Found audio stream = %p",
				this, interfaceNum, altSettingNum, audioStream);

			// <rdar://6500270> Panic trying to change a collection in the registry on K64.
			// IORegistryEntry object marks collections immutable when set as properties of a registry entry that's attached to a plane.
			oldChannelNamesDictionary = OSDynamicCast(OSDictionary, IORegistryEntry::getProperty(kIOAudioEngineFullChannelNamesKey));
			FailIf(NULL == oldChannelNamesDictionary, Exit);

			newChannelNamesDictionary = OSDictionary::withDictionary(oldChannelNamesDictionary);
			FailIf(NULL == newChannelNamesDictionary, Exit);

			// <rdar://problem/6706026>
			// Find the channel names from either the mixer, processing, extension unit, or input terminal closest to the output terminal.
			// The driver should use the first unit that has an audio cluster descriptor.
			FailIf(NULL == thePath, Exit);

			// Search from the unit connected to the output terminal (index 0 + 1) to the input terminal
			for (UInt32 unitIndex = 1; unitIndex < thePath->getCount(); unitIndex++)
			{
				number = OSDynamicCast(OSNumber, thePath->getObject(unitIndex));

				if (
						(NULL != number)
				   &&
						(kIOReturnSuccess == configDictionary->getSubType(&subType, controlInterfaceNum, 0, number->unsigned8BitValue()))
				   )
				{
					if (
							(MIXER_UNIT		 == subType)
					   ||
							(PROCESSING_UNIT == subType)
					   ||
							(EXTENSION_UNIT	 == subType)
					   ||
							(INPUT_TERMINAL	 == subType)
					   )
					{
						unitID = number->unsigned8BitValue();

						FailIf(
									kIOReturnSuccess != configDictionary->getAudioClusterDescriptor(
																				&clusterDescriptor,
																				controlInterfaceNum,
																				0,
																				unitID
																		  ),
									Exit
							  );
						if (clusterDescriptor.bNrChannels > 0)	// We found an audio cluster descriptor
						{
							break;
						}
					}
				}
			}

			DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateChannelNames(%u, %u, %u) - Audio cluster descriptor { %u, %u, %u }",
				this, unitID, interfaceNum, altSettingNum,
				clusterDescriptor.bNrChannels,
				clusterDescriptor.bmChannelConfig,
				clusterDescriptor.iChannelNames);

#ifdef TARGET_PDJ
 			OSArray		*channelNameArray = OSDynamicCast(OSArray, audioStream->mStreamInterface->getProperty("Channel Name"));
#endif
			for (UInt8 channelIndex = 0; channelIndex < clusterDescriptor.bNrChannels; channelIndex++)
			{
				nameString = this->getChannelNameString(unitID, channelIndex + 1);	// <rdar://problem/6706026>
#ifdef TARGET_PDJ
 				if ((NULL == nameString) && (NULL != channelNameArray))
 				{
 					nameString = OSDynamicCast(OSString, channelNameArray->getObject(channelIndex));
 
 					if (nameString)
 					{
 						nameString->retain();
 					}
 				}
#endif

				if (NULL != nameString)
				{
					keyString[0] = 0;

					if (kIOAudioStreamDirectionOutput == audioStream->getDirection())
					{
						::snprintf(
							keyString,
							sizeof(keyString),
							kIOAudioEngineFullChannelNameKeyOutputFormat,
							audioStream->getStartingChannelID() + channelIndex
						);
					}
					else
					{
						::snprintf(
							keyString,
							sizeof(keyString),
							kIOAudioEngineFullChannelNameKeyInputFormat,
							audioStream->getStartingChannelID() + channelIndex
						);
					}

					DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateChannelNames(%d, %d, %d) - Setting %s to %s",
						this, unitID, interfaceNum, altSettingNum, keyString, nameString->getCStringNoCopy());

					newChannelNamesDictionary->setObject(keyString, nameString);	// <rdar://6500270>

					nameString->release();
				}
			}

			// <rdar://6500270> Update the dictionary.
			IORegistryEntry::setProperty(kIOAudioEngineFullChannelNamesKey, newChannelNamesDictionary);
			newChannelNamesDictionary->release();
			break;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::updateChannelNames(%d, %d, %d)", this, unitID, interfaceNum, altSettingNum);
	return;
}

/****************************************************************************/
/**
 * @brief	デバイスのクロック状態を更新．
 *
 * @param[in]		clockID			クロック ID．
 */
// <rdar://5811247>
void
PdjUsbAudioEngine::updateClockStatus(
	UInt8			clockID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::updateClockStatus()");

	UInt32			clockRate;
	bool			clockValidity = false;

	if ((0 != clockID) && (clockID == mCurrentClockSourceID))
	{
		if (kIOReturnSuccess == mUSBAudioDevice->mTopology->getCurClockSourceSamplingFrequency(clockID, &clockRate, &clockValidity))
		{
			if (NULL != mClockSelectorControl)
			{
				if (!clockValidity)
				{
					SInt32		currentValue = mClockSelectorControl->getIntValue();

					// The current selected clock is no longer valid. Switch over to a valid clock source.
					OSArray		*availableSelections = OSDynamicCast(
															OSArray,
															mClockSelectorControl->getProperty(kIOAudioSelectorControlAvailableSelectionsKey)
													   );
					if (NULL != availableSelections)
					{
						for (UInt32 index = 0; index < availableSelections->getCount(); index++)
						{
							OSDictionary	*selectionDictionary = OSDynamicCast(OSDictionary, availableSelections->getObject(index));

							if (NULL != selectionDictionary)
							{
								OSNumber	*selectionNumber = OSDynamicCast(
																	OSNumber,
																	selectionDictionary->getObject(kIOAudioSelectorControlSelectionValueKey)
															   );
								if (NULL != selectionNumber)
								{
									SInt32		selection = (SInt32)selectionNumber->unsigned32BitValue();

									if (selection != currentValue)
									{
										clockID = (selection >> 24) & 0xFF;

										if (
												(
													kIOReturnSuccess == mUSBAudioDevice->mTopology->getCurClockSourceSamplingFrequency(
																										clockID,
																										&clockRate,
																										&clockValidity
																									)
												)
										   &&
												(clockValidity)
										   )
										{
											// Found a clock source which is valid. Switch over to it.
											DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - Switch over to selection = 0x%x",
												this, selection);
											mClockSelectorControl->setValue(selection);
											break;
										}
									}
								}
							}
						}
					}
				}
				else
				{
					// <rdar://6945472> If the sample rate has changed, then it needs to be republished and CoreAudio be informed of the change.
					IOAudioSampleRate	sampleRate = this->getCurrentClockPathSampleRate();

					DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - Current sample rate = %u, new sample rate = %u",
						this, mCurSampleRate.whole, sampleRate.whole);

					if (mCurSampleRate.whole != sampleRate.whole)
					{
						super::pauseAudioEngine();

						super::beginConfigurationChange();

						DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - Re-publishing audio formats", this);
						// Re-publish all the streams supported sample rates & formats.
						this->republishAvailableFormats(sampleRate);

						DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - Performing format - sample rate changes", this);
						// Do the sample rate change.
						FailMessage(kIOReturnSuccess != this->performFormatChange(NULL, NULL, &sampleRate));

						DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - Informing CoreAudio of the sample rate change", this);
						// Inform CoreAudio that the sample rate has changed.
						super::hardwareSampleRateChanged(&sampleRate);

						super::completeConfigurationChange();

						super::resumeAudioEngine();
					}
				}
			}
			else
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - No clock selector present - update audio formats if necessary", this);

				// Don't have a selector so we need to republish the available format if the current clock validity state or sample rate is not
				// the same as what we thought was.
				// <rdar://6945472>
				if (
						mShouldRepublishFormat
				   ||
						(clockValidity != mClockSourceValidity)
				   ||
						(mCurSampleRate.whole != sampleRate.whole)
				   )
				{
					super::pauseAudioEngine();

					super::beginConfigurationChange();

					mClockSourceValidity = clockValidity;

					IOAudioSampleRate	sampleRate = this->getCurrentClockPathSampleRate();

					DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - current sample rate = %u, new sample rate = %u",
						this, mCurSampleRate.whole, sampleRate.whole);

					DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - Re-publishing audio formats", this);
					// Re-publish all the streams supported sample rates & formats.
					this->republishAvailableFormats(sampleRate);

					DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - Performing format - sample rate changes", this);
					// Do the sample rate change.
					FailMessage(kIOReturnSuccess != this->performFormatChange(NULL, NULL, &sampleRate));

					DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::updateClockStatus() - Informing CoreAudio of the sample rate change", this);
					// <rdar://6945472> Inform CoreAudio that the sample rate has changed.
					super::hardwareSampleRateChanged(&sampleRate);

					mShouldRepublishFormat = false;

					super::completeConfigurationChange();

					super::resumeAudioEngine();
				}
			}
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::updateClockStatus()");
}

/****************************************************************************/
/**
 * @brief	プラグイン登録．
 *			プラグインがアタッチされた場合に、プラグインから呼び出される．
 *
 * @param[in]		*thePlugin		アタッチされたプラグイン．
 */
void
PdjUsbAudioEngine::registerPlugin(
	PdjUsbAudioPlugin	*thePlugin
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::registerPlugin()");

	mPlugin = thePlugin;

	mPluginInitThread = ::thread_call_allocate((thread_call_func_t)pluginLoaded, (thread_call_param_t)this);

	if (NULL != mPluginInitThread)
	{
		::thread_call_enter(mPluginInitThread);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::registerPlugin()");
}

/****************************************************************************/
/**
 * @brief	デバイス設定用の外部アプリを登録．
 *			プラグインからの要求により、外部アプリの関連付けを行う．
 *
 * @param[in]		*bundleID		外部アプリのバンドル ID 文字列．
 */
void
PdjUsbAudioEngine::pluginSetConfigurationApp(
	const char		*bundleID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::pluginSetConfigurationApp()");

	if (bundleID)
	{
		mUSBAudioDevice->setConfigurationApp(bundleID);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::pluginSetConfigurationApp()");
}

/****************************************************************************/
/**
 * @brief	デバイスリクエスト．
 *			プラグインからの要求により、USB デバイスリクエストを発行する．
 *
 * @param[in]		*request		デバイスリクエスト．
 * @param[in, out]	*dataBuffer		デバイスリクエストのデータ．
 * @param[in]		*completion		完了ルーチン．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioEngine::pluginDeviceRequest(
	StandardUSB::DeviceRequest	*request,
	IOMemoryDescriptor			*dataBuffer,
	IOUSBHostCompletion			*completion
)
{
	IOReturn			result = kIOReturnBadArgument;

	if (request)
	{
		result = mUSBAudioDevice->deviceRequest(request, dataBuffer, completion);
	}

	return result;
}

#pragma mark - Protected Methods -

/****************************************************************************/
/**
 * @brief	プラグインがロードされたときの処理．
 *
 * @param[in]		*usbAudioEngineObject	プラグインがアタッチされた
 *											PdjUsbAudioEngine のインスタンス．
 *
 * @return			IOReturn コード．
 */
void
PdjUsbAudioEngine::pluginLoaded(
	PdjUsbAudioEngine	*usbAudioEngineObject
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::pluginLoaded()");

	// It can only be output or input for plugin at the engine level. For engine with multiple streams,
	// plugin should be instantiated at stream level. If a plugin appears at the engine level, treat it as
	// an output plugin.
	if (
			(usbAudioEngineObject->mPlugin)
	   &&
			(
				(usbAudioEngineObject->mMainOutputStream)
			||
				(usbAudioEngineObject->mMainInputStream)
			)
	   )
	{
		usbAudioEngineObject->mPlugin->open(usbAudioEngineObject);

		IOReturn	result = usbAudioEngineObject->mPlugin->pluginInit(
																usbAudioEngineObject,
																usbAudioEngineObject->mUSBAudioDevice->getVendorID(),
																usbAudioEngineObject->mUSBAudioDevice->getProductID()
															);
		if (result == kIOReturnSuccess)
		{
			DP(DBG_INFO, "success initing the plugin");

			if (usbAudioEngineObject->mMainOutputStream)
			{
				usbAudioEngineObject->mPlugin->pluginSetDirection(
													(IOAudioStreamDirection)kIOAudioStreamDirectionOutput
											   );

				usbAudioEngineObject->mPlugin->pluginSetFormat(
													usbAudioEngineObject->mMainOutputStream->getFormat(),
													&usbAudioEngineObject->sampleRate
											   );
			}
			else if (usbAudioEngineObject->mMainInputStream)
			{
				usbAudioEngineObject->mPlugin->pluginSetDirection(
													(IOAudioStreamDirection)kIOAudioStreamDirectionInput
											   );

				usbAudioEngineObject->mPlugin->pluginSetFormat(
													usbAudioEngineObject->mMainInputStream->getFormat(),
													&usbAudioEngineObject->sampleRate
											   );
			}
		}
		else
		{
			DP(DBG_INFO, "Error initing the plugin");

			usbAudioEngineObject->mPlugin->close(usbAudioEngineObject);
			usbAudioEngineObject->mPlugin = NULL;
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::pluginLoaded()");
}

/****************************************************************************/
/**
 * @brief	各種オーディオコントロールの値が変更されるときに呼び出されるコールバック．
 *
 * @param[in]		*target			このコールバックを登録した PdjUsbAudioEngine のインスタンス．
 * @param[in]		*audioControl	値が変化した IOAudioControl のインスタンス．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbAudioEngine::controlChangedHandler(
	OSObject				*target,
	IOAudioControl			*audioControl,
	SInt32					oldValue,
	SInt32					newValue
)
{
	IOReturn				result = kIOReturnError;
	PdjUsbAudioEngine		*self;

	self = OSDynamicCast(PdjUsbAudioEngine, target);
	FailIf(NULL == self, Exit);

	result = self->protectedControlChangedHandler(audioControl, oldValue, newValue);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	各種オーディオコントロールの値が変更されるときの処理．
 *
 * @param[in]		*audioControl	値が変化した IOAudioControl のインスタンス．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbAudioEngine::protectedControlChangedHandler(
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	IOReturn		result = kIOReturnError;

	switch (audioControl->getType())
	{
		case kIOAudioControlTypeSelector:
			if (kIOAudioSelectorControlSubTypeClockSource == audioControl->getSubType())
			{
				result = this->doClockSelectorChange(audioControl, oldValue, newValue);
			}
			break;
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	クロックセレクタ設定．
 *
 * @param[in]		interfaceNum	USB interface 番号．
 * @param[in]		altSettingNum	alternate settings 番号．
 * @param[in]		sampleRate		サンプリング周波数．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbAudioEngine::doClockSelectorSetup(
	UInt8			interfaceNum,
	UInt8			altSettingNum,
	UInt32			sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::doClockSelectorSetup(0x%x, 0x%x, %u)",
		this, interfaceNum, altSettingNum, sampleRate);

	IOReturn						result = kIOReturnError;
	PdjUsbConfigurationDictionary	*configDictionary = NULL;
	UInt8							controlInterfaceNum;
	OSNumber						*clockIDNumber;
	UInt8							clockID;
	UInt8							clockSelectorID = 0;
	OSArray							*clockPathGroup = NULL;
	UInt8							clockPathGroupIndex = 0;
	OSArray							*clockPath;
	UInt8							clockPathIndex = 0;
	UInt32							clockSelection = 0;
#if POLLCLOCKSTATUS
	bool							hasNonProgrammableClockSource = false;
#endif

	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == mUSBAudioDevice->mControlInterface, Exit);
	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	controlInterfaceNum = mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	if (NULL == mClockSelectorControl)
	{
		clockPath = mUSBAudioDevice->mTopology->getOptimalClockPath(this, interfaceNum, altSettingNum, sampleRate, NULL, &clockPathGroupIndex);
		FailIf(NULL == clockPath, Exit);

		FailIf(NULL == (clockIDNumber = OSDynamicCast(OSNumber, clockPath->getLastObject())), Exit);

		clockID = clockIDNumber->unsigned8BitValue();

		result = mUSBAudioDevice->mTopology->getClockSelectorIDAndPathIndex(&clockSelectorID, &clockPathIndex, clockPath);

		FailIf(kIOReturnSuccess != result, Exit);

		result = mUSBAudioDevice->mTopology->getClockPathCurSampleRate(NULL, &mClockSourceValidity, NULL, clockPath);		// <rdar://6945472>
		FailIf(kIOReturnSuccess != result, Exit);

		if (!mClockSourceValidityInitialized)
		{
			mClockSourceValidityInitialized = true;
			mShouldRepublishFormat = true;
		}
	}
	else
	{
		clockID				= (mClockSelectorControl->getIntValue() >> 24) & 0xFF;
		clockSelectorID		= (mClockSelectorControl->getIntValue() >> 16) & 0xFF;
		clockPathGroupIndex	= (mClockSelectorControl->getIntValue() >>  8) & 0xFF;
		clockPathIndex		= (mClockSelectorControl->getIntValue() >>  0) & 0xFF;

		result = kIOReturnSuccess;

		mClockSelectorControl->release();
		mClockSelectorControl = NULL;
	}

	mCurrentClockSourceID		= clockID;
	mCurrentClockPathGroupIndex	= clockPathGroupIndex;
	mCurrentClockPathIndex		= clockPathIndex;

	// If there is a clock selector in the path, then create a clock selector control.
	if ((kIOReturnSuccess == result) && (0 != clockSelectorID) && (0 != clockPathIndex))
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::doClockSelectorSetup(0x%x, 0x%x, %u) - clockSelectorID = %u, clockPathIndex = %u",
			this, interfaceNum, altSettingNum, sampleRate, clockSelectorID, clockPathIndex);

		FailIf(kIOReturnSuccess != mUSBAudioDevice->mTopology->setCurClockSelector(clockSelectorID, clockPathIndex), Exit);

		clockSelection = (
							((UInt32)clockID		 << 24)
						 |
							((UInt32)clockSelectorID << 16)
						 |
							(clockPathGroupIndex	 <<  8)
						 |
							(clockPathIndex			 <<  0)
						 );

		mClockSelectorControl = IOAudioSelectorControl::create(
															clockSelection,
															kIOAudioControlChannelIDAll,
															kIOAudioControlChannelNameAll,
															0,
															kIOAudioSelectorControlSubTypeClockSource,
															kIOAudioControlUsageInput
														);
		FailIf(NULL == mClockSelectorControl, Exit);

#ifdef TARGET_PDJ
#else
		mClockSelectorControl->setValueChangeHandler(controlChangedHandler, this);
#endif

		clockPathGroup = mUSBAudioDevice->mTopology->getClockPathGroup(interfaceNum, altSettingNum);
		FailIf(NULL == clockPathGroup, Exit);

		for (UInt32 pathIndex = 0; pathIndex < clockPathGroup->getCount(); pathIndex++)
		{
			UInt8		subType;

			FailIf(NULL == (clockPath = OSDynamicCast(OSArray, clockPathGroup->getObject(pathIndex))), Exit);
			FailIf(NULL == (clockIDNumber = OSDynamicCast(OSNumber, clockPath->getLastObject())), Exit);

			clockID = clockIDNumber->unsigned8BitValue();

			FailIf(kIOReturnSuccess != (result = configDictionary->getSubType(&subType, controlInterfaceNum, 0, clockID)), Exit);

			DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::doClockSelectorSetup(0x%x, 0x%x, %u) - %u: clockID = %u, subType = %u",
				this, interfaceNum, altSettingNum, sampleRate, pathIndex, clockID, subType);

			if (USBAUDIO_0200::CLOCK_SOURCE == subType)
			{
				UInt8		clockType;
				UInt8		assocTerminal	= 0;
				UInt16		terminalType	= 0;
				UInt8		stringIndex;
				char		*clockSelectionString = NULL;

				clockSelection = (
									((UInt32)clockID		 << 24)
								 |
									((UInt32)clockSelectorID << 16)
								 |
									(clockPathGroupIndex	 <<  8)
								 |
									((pathIndex + 1)		 <<  0)
								 );

				if (!mClockSelectorControl->valueExists(clockSelection))
				{
					char	tempString[kStringBufferSize] = { 0 };

					if (
							(kIOReturnSuccess == configDictionary->getStringIndex(&stringIndex, controlInterfaceNum, 0, clockID))
					   &&
							(0 != stringIndex)
					   &&
							(kIOReturnSuccess == mUSBAudioDevice->getStringDescriptor(stringIndex, tempString, kStringBufferSize))
					   )
					{
						// There is a string specified for this clock source. Use this for the name.
						clockSelectionString = tempString;
					}
					else
					{
						// No string specified. Generate one based on the following:
						// 1. If it is internal clock type, it should be "Device"
						// 2. If it is external clock type, check the associated terminal type. If it is present, then check the
						// following:
						// (a) If it is USB streaming, then it should be "Mac Sync".
						// (b) Otherwise, use the name of the terminal, if available.
						// (c) If no terminal name is available, then fallback to the terminal type.
						// 3. If there is no assoc terminal specified, then use the clock type to generate the string.
						FailIf(
									kIOReturnSuccess != (
															result = configDictionary->getClockSourceClockType(
																							&clockType,
																							controlInterfaceNum,
																							0,
																							clockID
																					   )
														),
									Exit
							  );

						switch (clockType)
						{
							case USBAUDIO_0200::CLOCK_TYPE_EXTERNAL:
								FailIf(
											kIOReturnSuccess != (
																	result = configDictionary->getClockSourceAssocTerminal(
																									&assocTerminal,
																									controlInterfaceNum,
																									0,
																									clockID
																							   )
																),
											Exit
									  );

								if (0 != assocTerminal)
								{
									FailIf(
											kIOReturnSuccess != (
																	result = configDictionary->getInputTerminalType(
																									&terminalType,
																									controlInterfaceNum,
																									0,
																									assocTerminal
																							   )
																),
											Exit
										  );

									if (USB_STREAMING == terminalType)
									{
#if LOCALIZABLE
										clockSelectionString = (char *)"StringMacSync";
#else
										clockSelectionString = (char *)"Mac Sync";
#endif
									}
									else
									{
										if (
												(
													kIOReturnSuccess == configDictionary->getStringIndex(
																								&stringIndex,
																								controlInterfaceNum,
																								0,
																								assocTerminal
																						  )
												)
										   &&
												(0 != stringIndex)
										   &&
												(
													kIOReturnSuccess == mUSBAudioDevice->getStringDescriptor(
																								stringIndex,
																								tempString,
																								kStringBufferSize
																						 )
												)
										   )
										{
											// There is a string specified for the assoc terminal. Use this for the name.
											clockSelectionString = tempString;
										}
										else
										{
											clockSelectionString = mUSBAudioDevice->terminalTypeString(terminalType);
										}
									}
								}
								else
								{
									clockSelectionString = mUSBAudioDevice->clockTypeString(clockType);
								}

								break;

							case USBAUDIO_0200::CLOCK_TYPE_INTERNAL_FIXED:
							case USBAUDIO_0200::CLOCK_TYPE_INTERNAL_VARIABLE:
							case USBAUDIO_0200::CLOCK_TYPE_INTERNAL_PROGRAMMABLE:
#if LOCALIZABLE
								clockSelectionString = (char *)"StringDevice";
#else
								clockSelectionString = (char *)"Device";
#endif
								break;
						}
					}

					if (NULL != clockSelectionString)
					{
						mClockSelectorControl->addAvailableSelection(clockSelection, clockSelectionString);
					}
				}

#if POLLCLOCKSTATUS
				if (!configDictionary->clockSourceHasFrequencyControl(controlInterfaceNum, 0, clockID, true))
				{
					hasNonProgrammableClockSource = true;
				}
#endif
			}
		}

#ifdef TARGET_PDJ
#else
		super::addDefaultAudioControl(mClockSelectorControl);
#endif
	}
#if POLLCLOCKSTATUS
	else
	{
		hasNonProgrammableClockSource = !configDictionary->clockSourceHasFrequencyControl(
																controlInterfaceNum,
																0,
																mCurrentClockSourceID,
																true
														   );
	}
#endif

	this->determineMacSyncMode(mCurrentClockSourceID);

#if POLLCLOCKSTATUS
	// Only poll if there isn't an interrupt endpoint, and at least one of the clock source is non-programmable.
	mShouldPollClockStatus = hasNonProgrammableClockSource && (!configDictionary->hasInterruptEndpoint(controlInterfaceNum, 0));

	mPollClockStatusCounter = 0;

	DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::doClockSelectorSetup(0x%x, 0x%x, %u) - Should poll = %d",
		this, interfaceNum, altSettingNum, sampleRate, mShouldPollClockStatus);
#endif

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::doClockSelectorSetup(0x%x, 0x%x, %u) = 0x%x",
		this, interfaceNum, altSettingNum, sampleRate, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックセレクタ変更．
 *
 * @param[in]		*audioControl	変更されたクロックソースコントロール．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbAudioEngine::doClockSelectorChange(
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::doClockSelectorChange(%p, 0x%x, 0x%x)",
		this, audioControl, oldValue, newValue);

	IOReturn			result = kIOReturnError;
	UInt8				clockSourceID;
	UInt8				clockSelectorID;
	UInt8				newClockPathGroupIndex;
	UInt8				newClockPathIndex;
	OSArray				*newClockPathGroup;
	OSArray				*newClockPath;
	UInt8				oldClockPathGroupIndex;
	UInt8				oldClockPathIndex;
	OSArray				*oldClockPathGroup;
	OSArray				*oldClockPath;
	bool				clockValidity = false;
	IOAudioSampleRate	newSampleRate;
	IOAudioSampleRate	oldSampleRate;

	clockSourceID			= (newValue >> 24) & 0xFF;
	clockSelectorID			= (newValue >> 16) & 0xFF;
	newClockPathGroupIndex	= (newValue >>  8) & 0xFF;
	newClockPathIndex		= (newValue >>  0) & 0xFF;

	oldClockPathGroupIndex	= (oldValue >>  8) & 0xFF;
	oldClockPathIndex		= (oldValue >>  0) & 0xFF;

	// When switching the clocks, determine if the device is locked to the selected clock. This is done by checking the clock validity
	// bit on the clock source. If it is valid, then perform the switch to the clock. Otherwise, just return success and schedule a
	// restore of the UI so that it switches back to the old value.
	if (oldValue != newValue)
	{
		super::pauseAudioEngine();

		super::beginConfigurationChange();

		// Determine if the selected clock source is valid.
		newClockPathGroup = mUSBAudioDevice->mTopology->getClockPathGroup(newClockPathGroupIndex);
		FailIf(NULL == newClockPathGroup, Exit);

		newClockPath = OSDynamicCast(OSArray, newClockPathGroup->getObject(newClockPathIndex - 1));
		FailIf(NULL == newClockPath, Exit)

		newSampleRate.whole		= 0;
		newSampleRate.fraction	= 0;

		oldSampleRate.whole		= 0;
		oldSampleRate.fraction	= 0;

		result = mUSBAudioDevice->mTopology->getClockPathCurSampleRate(
												&newSampleRate.whole,
												&clockValidity,
												NULL,
												newClockPath
											 );	// <rdar://6945472>

		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::doClockSelectorChange(%p, 0x%x, 0x%x) - result = 0x%x, sample rate = %u, clockValidity = %d",
			this, audioControl, oldValue, newValue, result, newSampleRate.whole, clockValidity);

		if (kIOReturnSuccess == result)
		{
			if (clockValidity)
			{
				// Set the old clock path to the same sample rate as the new one. Some devices lock their external clock using
				// the internal one as the reference, so it doesn't lock if the frequency is not the same (or close to it).
				oldClockPathGroup = mUSBAudioDevice->mTopology->getClockPathGroup(oldClockPathGroupIndex);
				FailIf(NULL == oldClockPathGroup, Exit);

				oldClockPath = OSDynamicCast(OSArray, oldClockPathGroup->getObject(oldClockPathIndex - 1));
				FailIf(NULL == oldClockPath, Exit)

				result = mUSBAudioDevice->mTopology->getClockPathCurSampleRate(
														&oldSampleRate.whole,
														NULL,
														NULL,
														oldClockPath
													 );	// <rdar://6945472>
				FailIf(kIOReturnSuccess != result, Exit);

				result = mUSBAudioDevice->mTopology->setClockPathCurSampleRate(newSampleRate.whole, oldClockPath);
				FailIf(kIOReturnSuccess != result, Exit);

				mCurrentClockSourceID  = clockSourceID;
				mCurrentClockPathIndex = newClockPathIndex;

				// Re-publish all the streams supported sample rates & formats.
				this->republishAvailableFormats(newSampleRate);

				// Do the sample rate change.
				result = this->performFormatChange(NULL, NULL, &newSampleRate);
				FailMessage(kIOReturnSuccess != result);

				if (kIOReturnSuccess == result)
				{
					this->determineMacSyncMode(mCurrentClockSourceID);

					// Inform CoreAudio that the sample rate has changed.
					super::hardwareSampleRateChanged(&newSampleRate);
				}
				else
				{
					// Failed to switch over to the new sample rate. Try to recover.
					mCurrentClockSourceID = (oldValue >> 24) & 0xFF;
					mCurrentClockPathIndex = oldClockPathIndex;

					// Re-publish all the streams supported sample rates & formats.
					this->republishAvailableFormats(oldSampleRate);

					// Do the sample rate change.
					FailMessage(kIOReturnSuccess != this->performFormatChange(NULL, NULL, &oldSampleRate));

					// Indicate that we need to restore the clock selection in the polled task.
					mRestoreClockSelection = true;
					mRestoreClockSelectionValue = oldValue;

					result = kIOReturnSuccess;
				}
			}
			else
			{
				// No valid clock on the selected clock source. Schedule a restore of the old selector value.
				mRestoreClockSelection = true;
				mRestoreClockSelectionValue = oldValue;
				result = kIOReturnSuccess;
			}
		}

		super::completeConfigurationChange();

		super::resumeAudioEngine();
	}
	else
	{
		result = mUSBAudioDevice->mTopology->setCurClockSelector(clockSelectorID, newClockPathIndex);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::doClockSelectorChange(%p, 0x%x, 0x%x) = 0x%x",
		this, audioControl, oldValue, newValue, result);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースが Mac 方式で同期されているかどうか判定．
 *			判定結果に応じてマスタストリームの設定を行う．
 *
 * @param[in]		clockID			クロック ID．
 *
 * @retval			true			クロックソースは Mac 方式で同期されている．
 * @retval			false			Mac 方式ではない．
 */
// <rdar://5811247> Determine if the engine is running in Mac sync mode, and setup the appropriate stream to be the master stream.
bool
PdjUsbAudioEngine::determineMacSyncMode(
	UInt8			clockID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::determineMacSyncMode(%d)", this, clockID);

	bool								macSyncMode			= false;
	PdjUsbConfigurationDictionary		*configDictionary	= NULL;
	PdjUsbAudioStream					*masterStream		= NULL;		// <rdar://7558825>
	UInt8								clockType			= 0xFF;
	UInt8								assocTerminal		= 0;
	UInt16								terminalType		= 0;
	UInt8								controlInterfaceNum;

	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == mUSBAudioDevice->mControlInterface, Exit);
	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	controlInterfaceNum = mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	FailIf(kIOReturnSuccess != configDictionary->getClockSourceClockType(&clockType, controlInterfaceNum, 0, clockID), Exit);
	FailIf(kIOReturnSuccess != configDictionary->getClockSourceAssocTerminal(&assocTerminal, controlInterfaceNum, 0, clockID), Exit);

	if ((USBAUDIO_0200::CLOCK_TYPE_EXTERNAL == clockType) && (0 != assocTerminal))
	{
		FailIf(kIOReturnSuccess != configDictionary->getInputTerminalType(&terminalType, controlInterfaceNum, 0, assocTerminal), Exit);

		macSyncMode = (USB_STREAMING == terminalType);
	}

	if ((NULL != mMainOutputStream) && (NULL != mMainInputStream))
	{
		if ((macSyncMode) || (kAdaptiveSyncType == mMainOutputStream->getSyncType()))
		{
			mMainInputStream->setMasterStreamMode(false);
			mMainOutputStream->setMasterStreamMode(true);

			masterStream = mMainOutputStream;	// <rdar://7558825>
		}
		else
		{
			mMainInputStream->setMasterStreamMode(true);
			mMainOutputStream->setMasterStreamMode(false);

			masterStream = mMainInputStream;	// <rdar://7558825>
		}
	}
	else if (NULL != mMainOutputStream)		// <rdar://7558825>
	{
		masterStream = mMainOutputStream;
	}
	else if (NULL != mMainInputStream)		// <rdar://7558825>
	{
		masterStream = mMainInputStream;
	}

	// <rdar://7558825> Clock domain for different synchronization type.
	if (masterStream)
	{
		if (
				(macSyncMode)
		   ||
				(kSynchronousSyncType == masterStream->getSyncType())
		   ||
				(kAdaptiveSyncType == masterStream->getSyncType())
		   )
		{
			this->setClockDomain(this->getSystemClockDomain());
		}
		else
		{
			this->setClockDomain();
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::determineMacSyncMode(%d) = %d", this, clockID, macSyncMode);
	return macSyncMode;
}

/****************************************************************************/
/**
 * @brief	システムのクロックドメインを取得．
 *
 * @return			クロックサービスへのポインタ．
 */
// <rdar://7558825> Get the system clock domain
UInt32
PdjUsbAudioEngine::getSystemClockDomain(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::getSystemClockDomain()");

	UInt32			result				 = kIOAudioNewClockDomain;
	OSDictionary	*serviceMatchingDict = NULL;
	OSIterator		*iterator			 = NULL;
	OSObject		*object;

	serviceMatchingDict = IOService::serviceMatching("AppleHDAController");
	FailIf(NULL == serviceMatchingDict, Exit);

	iterator = IOService::getMatchingServices(serviceMatchingDict);
	FailIf(NULL == iterator, Exit);

	object = iterator->getNextObject();

	while (object)
	{
		IOService	*service = OSDynamicCast(IOService, object);

		if (service)
		{
			IOService	*provider = service->getProvider();

			if (provider)
			{
				if (0 == ::strncmp(provider->getName(), "HDEF", sizeof("HDEF")))
				{
					result = (UInt32)(uintptr_t)(service);
					break;
				}
			}
		}

		object = iterator->getNextObject();
	}

Exit:
	if (serviceMatchingDict)
	{
		serviceMatchingDict->release();
	}

	if (iterator)
	{
		iterator->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::getSystemClockDomain() = %u", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	サンプリング周波数の初期値を決定．
 *
 * @param[out]		*sampleRate		サンプリング周波数の初期値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioEngine::getDefaultSampleRate(
	IOAudioSampleRate	*sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "[%p]::getDefaultSampleRate()", this);

	IOReturn						result = kIOReturnError;
	bool							foundSampleRate = false;
	IOAudioSampleRate				newSampleRate;
	PdjUsbConfigurationDictionary	*configDictionary = NULL;
	OSNumber						*streamInterfaceNumber;
	UInt8							interfaceNumber;
	UInt8							newAltSettingID = 0;
	UInt16							format;

	FailIf(NULL == mStreamInterfaceNumberArray, Exit);

	newSampleRate.whole		= kDefaultSamplingRate;
	newSampleRate.fraction	= 0;

	configDictionary = mUSBAudioDevice->getConfigDictionary();

	for (UInt32 streamIndex = 0; streamIndex < mStreamInterfaceNumberArray->getCount(); streamIndex++)
	{
		result = kIOReturnError;

		streamInterfaceNumber = OSDynamicCast(OSNumber, mStreamInterfaceNumberArray->getObject(streamIndex));

		FailIf(NULL == streamInterfaceNumber, Exit);

		interfaceNumber = streamInterfaceNumber->unsigned8BitValue();

		// If possible, never pick anything other than PCM for default.
		// We, as the driver, have to pick a default sample rate, size, and number of channels, so try 16-bit stereo 44.1kHz
		result = configDictionary->getAltSettingWithSettings(
										&newAltSettingID,
										interfaceNumber,
										kChannelDepth_STEREO,
										kBitDepth_16bits,
										newSampleRate.whole
								   );
		if (
				(kIOReturnSuccess == result)
		   &&
				(kIOReturnSuccess == configDictionary->getFormat(&format, interfaceNumber, newAltSettingID))
		   &&
				(PCM == format)
		   )
		{
			foundSampleRate = true;
			break;
		}
	}

	if (!foundSampleRate)
	{
		for (UInt32 streamIndex = 0; streamIndex < mStreamInterfaceNumberArray->getCount(); streamIndex++)
		{
			result = kIOReturnError;

			streamInterfaceNumber = OSDynamicCast(OSNumber, mStreamInterfaceNumberArray->getObject(streamIndex));
			FailIf(NULL == streamInterfaceNumber, Exit);

			interfaceNumber = streamInterfaceNumber->unsigned8BitValue();

			// Didn't have stereo, so try mono
			result = configDictionary->getAltSettingWithSettings(
											&newAltSettingID,
											interfaceNumber,
											kChannelDepth_MONO,
											kBitDepth_16bits,
											newSampleRate.whole
									   );
			if (
					(kIOReturnSuccess == result)
			   &&
					(kIOReturnSuccess == configDictionary->getFormat(&format, interfaceNumber, newAltSettingID))
			   &&
					(PCM == format)
			   )
			{
				foundSampleRate = true;
				break;
			}
		}
	}

	if (!foundSampleRate)
	{
		for (UInt32 streamIndex = 0; streamIndex < mStreamInterfaceNumberArray->getCount(); streamIndex++)
		{
			result = kIOReturnError;

			streamInterfaceNumber = OSDynamicCast(OSNumber, mStreamInterfaceNumberArray->getObject(streamIndex));
			FailIf(NULL == streamInterfaceNumber, Exit);

			interfaceNumber = streamInterfaceNumber->unsigned8BitValue();

			// Don't have a mono or stereo 16-bit 44.1kHz interface, so try for a stereo 16-bit interface with any sample rate
			if (
					(
						kIOReturnSuccess == (
												result = configDictionary->getAltSettingWithSettings(
																				&newAltSettingID,
																				interfaceNumber,
																				kChannelDepth_STEREO,
																				kBitDepth_16bits
																		   )
											)
					)
			   &&
					(
						kIOReturnSuccess == configDictionary->getFormat(&format, interfaceNumber, newAltSettingID)
					&&
						(PCM == format)
					)
			   )
			{
				// we'll run at the highest sample rate that the device has at stereo 16-bit
				configDictionary->getHighestSampleRate(&(newSampleRate.whole), interfaceNumber, newAltSettingID);

				foundSampleRate = true;

				break;
			}
		}
	}

	if (!foundSampleRate)
	{
		for (UInt32 streamIndex = 0; streamIndex < mStreamInterfaceNumberArray->getCount(); streamIndex++)
		{
			result = kIOReturnError;

			streamInterfaceNumber = OSDynamicCast(OSNumber, mStreamInterfaceNumberArray->getObject(streamIndex));
			FailIf(NULL == streamInterfaceNumber, Exit);

			interfaceNumber = streamInterfaceNumber->unsigned8BitValue();

			// Don't have a stereo 16-bit interface, so try for a mono 16-bit interface with any sample rate
			if (
					(
						kIOReturnSuccess == (
												result = configDictionary->getAltSettingWithSettings(
																				&newAltSettingID,
																				interfaceNumber,
																				kChannelDepth_MONO,
																				kBitDepth_16bits
																		   )
											)
					)
			   &&
					(
						kIOReturnSuccess == configDictionary->getFormat(&format, interfaceNumber, newAltSettingID)
					&&
						(PCM == format)
					)
			   )
			{
				// we'll run at the highest sample rate that the device has at mono 16-bit
				configDictionary->getHighestSampleRate(&(newSampleRate.whole), interfaceNumber, newAltSettingID);

				foundSampleRate = true;

				break;
			}
		}
	}

	if (!foundSampleRate)
	{
		result = kIOReturnError;

		streamInterfaceNumber = OSDynamicCast(OSNumber, mStreamInterfaceNumberArray->getObject(0));
		FailIf(NULL == streamInterfaceNumber, Exit);

		interfaceNumber = streamInterfaceNumber->unsigned8BitValue();

		// Just take the first interface.
		newAltSettingID = configDictionary->alternateSettingZeroCanStream(interfaceNumber) ? 0 : 1;

		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::getDefaultSampleRate() - Taking first available alternate setting (%d)",
			this, newAltSettingID);

		FailIf(
					kIOReturnSuccess != (
											result = configDictionary->getHighestSampleRate(
																			&(newSampleRate.whole),
																			interfaceNumber,
																			newAltSettingID
																	   )
										),
					Exit
			  );
	}

	DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::getDefaultSampleRate() - Default sample rate is %u", this, newSampleRate.whole);
	DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::getDefaultSampleRate() - Default alternate setting ID is %u", this, newAltSettingID);

	FailIf(0 == newSampleRate.whole, Exit);

	*sampleRate = newSampleRate;

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "[%p]::getDefaultSampleRate(%u) = 0x%x", this, sampleRate->whole, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	現在のクロックパスにおけるサンプリング周波数を取得．
 *
 * @return			取得したサンプリング周波数．
 */
// <rdar://6945472>
IOAudioSampleRate
PdjUsbAudioEngine::getCurrentClockPathSampleRate(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::getCurrentClockPathSampleRate()");

	OSArray				*clockPathGroup;
	OSArray				*clockPath;
	IOAudioSampleRate	sampleRate;

	sampleRate.whole	= 0;
	sampleRate.fraction	= 0;

	// <rdar://6945472> Get the current clock path sample rate.
	clockPathGroup = mUSBAudioDevice->mTopology->getClockPathGroup(mCurrentClockPathGroupIndex);
	FailIf(NULL == clockPathGroup, Exit);

	clockPath = OSDynamicCast(OSArray, clockPathGroup->getObject(mCurrentClockPathIndex - 1));
	FailIf(NULL == clockPath, Exit)

	FailIf(
				kIOReturnSuccess != mUSBAudioDevice->mTopology->getClockPathCurSampleRate(
																	&sampleRate.whole,
																	NULL,
																	NULL,
																	clockPath
																),
				Exit
		  );

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::getCurrentClockPathSampleRate()");
	return sampleRate;
}

/****************************************************************************/
/**
 * @brief	チャンネル名取得．
 *
 * @param[in]		unitID			オーディオユニット ID．
 * @param[in]		channelNum		チャンネル番号．
 *
 * @return			チャンネル名文字列．
 */
// <rdar://6430836> <rdar://problem/6706026> Retrieve the channel name for the specified channel.
OSString *
PdjUsbAudioEngine::getChannelNameString(
	UInt8			unitID,
	UInt8			channelNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOENGINE "::getChannelNameString()");

	PdjUsbConfigurationDictionary		*configDictionary = NULL;
	UInt8								controlInterfaceNum;
	AudioClusterDescriptor				clusterDescriptor;
	UInt8								numPreDefinedChannels = 0;
	char								stringBuffer[kStringBufferSize];
	IOReturn							result = kIOReturnNotFound;
	OSString							*theString = NULL;

	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == mUSBAudioDevice->mControlInterface, Exit);		// <rdar://7085810>

	controlInterfaceNum = mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	FailIf(kIOReturnSuccess != configDictionary->getAudioClusterDescriptor(&clusterDescriptor, controlInterfaceNum, 0, unitID), Exit);

	DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "[%p]::getChannelNameString(%u, %u) - Audio cluster descriptor { %u, %u, %u }",
		this, unitID, channelNum, clusterDescriptor.bNrChannels, clusterDescriptor.bmChannelConfig, clusterDescriptor.iChannelNames);

	// See if the channel has a predefined spatial location.
	for (UInt8 channelIndex = 0; channelIndex < sizeof(UInt32) * 8; channelIndex++)		// <rdar://problem/6679955>
	{
		if (clusterDescriptor.bmChannelConfig & (1 << channelIndex))
		{
			numPreDefinedChannels++;
		}

		if (numPreDefinedChannels == channelNum)
		{
			// Found the channel that has pre-defined spatial location. Use the spatial location name.
			::strncpy(stringBuffer, this->channelConfigString(channelIndex), kStringBufferSize);
			result = kIOReturnSuccess;
			break;
		}
	}

	if ((numPreDefinedChannels != channelNum) && (channelNum <= clusterDescriptor.bNrChannels))
	{
		// Channel doesn't have a predefined spatial location.
		// See if iChannelNames is present, and use it to get the name.
		if (0 != clusterDescriptor.iChannelNames)
		{
			result = mUSBAudioDevice->getStringDescriptor(
											clusterDescriptor.iChannelNames + channelNum - 1 - numPreDefinedChannels,
											stringBuffer,
											kStringBufferSize
									  );
			DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "::getChannelNameString(%d, %d) - stringIndex = %d, stringBuffer = %s, result = 0%x",
				unitID, channelNum, clusterDescriptor.iChannelNames + channelNum - 1 - numPreDefinedChannels, stringBuffer, result);
		}
	}

	if (kIOReturnSuccess == result)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOENGINE "::getChannelNameString() - terminalID = %d, channelNum = %d, stringBuffer = %s",
			unitID, channelNum, stringBuffer);

		theString = OSString::withCString(stringBuffer);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOENGINE "::getChannelNameString()");
	return theString;
}

/****************************************************************************/
/**
 * @brief	標準のチャンネル名．
 *
 * @param[in]		channel			UAC 2.0 ディスクリプタの bmChannelConfig フィールド
 *									で指定されたチャンネルの空間配置．
 *
 * @return			チャンネル名文字列．
 */
char *
PdjUsbAudioEngine::channelConfigString(
	UInt8			channel
)
{
	char			*channelNameString;

	switch (channel) {
#if LOCALIZABLE
		case 0:		channelNameString = (char *)"StringFrontLeft";					break;
		case 1:		channelNameString = (char *)"StringFrontRight";					break;
		case 2:		channelNameString = (char *)"StringFrontCenter";				break;
		case 3:		channelNameString = (char *)"StringLowFrequencyEffects";		break;
		case 4:		channelNameString = (char *)"StringBackLeft";					break;
		case 5:		channelNameString = (char *)"StringBackRight";					break;
		case 6:		channelNameString = (char *)"StringFrontLeftofCenter";			break;
		case 7:		channelNameString = (char *)"StringFrontRightofCenter";			break;
		case 8:		channelNameString = (char *)"StringBackCenter";					break;
		case 9:		channelNameString = (char *)"StringSideLeft";					break;
		case 10:	channelNameString = (char *)"StringSideRight";					break;
		case 11:	channelNameString = (char *)"StringTopCenter";					break;
		case 12:	channelNameString = (char *)"StringTopFrontLeft";				break;
		case 13:	channelNameString = (char *)"StringTopFrontCenter";				break;
		case 14:	channelNameString = (char *)"StringTopFrontRight";				break;
		case 15:	channelNameString = (char *)"StringTopBackLeft";				break;
		case 16:	channelNameString = (char *)"StringTopBackCenter";				break;
		case 17:	channelNameString = (char *)"StringTopBackRight";				break;
		case 18:	channelNameString = (char *)"StringTopFrontLeftofCenter";		break;
		case 19:	channelNameString = (char *)"StringTopFrontRightofCenter";		break;
		case 20:	channelNameString = (char *)"StringLeftLowFrequencyEffects";	break;
		case 21:	channelNameString = (char *)"StringRightLowFrequencyEffects";	break;
		case 22:	channelNameString = (char *)"StringTopSideLeft";				break;
		case 23:	channelNameString = (char *)"StringTopSideRight";				break;
		case 24:	channelNameString = (char *)"StringBottomCenter";				break;
		case 25:	channelNameString = (char *)"StringBackLeftofCenter";			break;
		case 26:	channelNameString = (char *)"StringBackRightofCenter";			break;
		case 27:
		case 28:
		case 29:
		case 30:	channelNameString = (char *)"StringReserved";					break;
		case 31:	channelNameString = (char *)"StringRawData";					break;
		default:	channelNameString = (char *)"StringUnknown";					break;
#else
		case 0:		channelNameString = (char *)"Front Left";						break;
		case 1:		channelNameString = (char *)"Front Right";						break;
		case 2:		channelNameString = (char *)"Front Center";						break;
		case 3:		channelNameString = (char *)"Low Frequency Effects";			break;
		case 4:		channelNameString = (char *)"Back Left";						break;
		case 5:		channelNameString = (char *)"Back Right";						break;
		case 6:		channelNameString = (char *)"Front Left of Center";				break;
		case 7:		channelNameString = (char *)"Front Right of Center";			break;
		case 8:		channelNameString = (char *)"Back Center";						break;
		case 9:		channelNameString = (char *)"Side Left";						break;
		case 10:	channelNameString = (char *)"Side Right";						break;
		case 11:	channelNameString = (char *)"Top Center";						break;
		case 12:	channelNameString = (char *)"Top Front Left";					break;
		case 13:	channelNameString = (char *)"Top Front Center";					break;
		case 14:	channelNameString = (char *)"Top Front Right";					break;
		case 15:	channelNameString = (char *)"Top Back Left";					break;
		case 16:	channelNameString = (char *)"Top Back Center";					break;
		case 17:	channelNameString = (char *)"Top Back Right";					break;
		case 18:	channelNameString = (char *)"Top Front Left of Center";			break;
		case 19:	channelNameString = (char *)"Top Front Right of Center";		break;
		case 20:	channelNameString = (char *)"Left Low Frequency Effects";		break;
		case 21:	channelNameString = (char *)"Right Low Frequency Effects";		break;
		case 22:	channelNameString = (char *)"Top Side Left";					break;
		case 23:	channelNameString = (char *)"Top Side Right";					break;
		case 24:	channelNameString = (char *)"Bottom Center";					break;
		case 25:	channelNameString = (char *)"Back Left of Center";				break;
		case 26:	channelNameString = (char *)"Back Right of Center";				break;
		case 27:
		case 28:
		case 29:
		case 30:	channelNameString = (char *)"Reserved";							break;
		case 31:	channelNameString = (char *)"Raw Data";							break;
		default:	channelNameString = (char *)"Unknown";							break;
#endif
	}

	return channelNameString;
}

#ifdef TARGET_PDJ
 /****************************************************************************/
 /**
  * @brief	Implicit Feedback のための受信パケットサイズのリストを取得．
  *
  * @param[in]		input					true: IN 側，false: OUT 側．
  * @param[in]		numTransactionsPerList	mNumTransactionsPerList．
  *
  * @return			受信パケットサイズのリスト．
  */
 UInt16 *
 PdjUsbAudioEngine::getInputSizeList(
 	bool			input,
 	UInt32			numTransactionsPerList
 )
 {
 	if (input)
 	{
 		UInt16	*p = &mUSBInputSizeList[mUSBInputIndex * numTransactionsPerList];
 
 		mUSBInputIndex = (mUSBInputIndex + 1) % (4 * PLAY_NUM_USB_FRAMES_PER_LIST / RECORD_NUM_USB_FRAMES_PER_LIST);
 
 		return p;
 	}
 	else
 	{
 		UInt16	*p = &mUSBInputSizeList[mUSBOutputIndex * numTransactionsPerList];
 
 		mUSBOutputIndex = (mUSBOutputIndex + 1) % 4;
 
 		return p;
 	}
 }
 
 void
 IOAudioEngine::lockStreamForIO(
 	IOAudioStream	*stream
 )
 {
 	stream->lockStreamForIO();
 }
 
 void
 IOAudioEngine::unlockStreamForIO(
 	IOAudioStream	*stream
 )
 {
 	stream->unlockStreamForIO();
 }
#endif
