/*
 * Copyright (c) 1998-2009 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioEngine.h
 *
 * @brief	IOAudioEngine クラスの派生クラス．
 *			PC ←→デバイス間のデータ転送を司る I/O エンジンを表現する．
 *			PdjUsbAudioDevice::createAudioEngine() においてインスタンス化される．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/
#ifndef PDJ_USB_AUDIO_ENGINE_H_FILE_
#define PDJ_USB_AUDIO_ENGINE_H_FILE_

/*------------------------------------------*/
/*				定数						*/
/*------------------------------------------*/

#pragma mark - Constants -

#define kSampleRate_44100						44100
#define kDefaultSamplingRate					kSampleRate_44100

#define kBitDepth_16bits						16

#define kChannelDepth_MONO						1
#define kChannelDepth_STEREO					2

#define kRefreshInterval						128

#define kAnchorSamplingFreqSec					(1024)							// <rdar://problem/7378275>
#define kAnchorSamplingFreq1					(kAnchorSamplingFreqSec / 64)	// <rdar://problem/7378275>
#define kAnchorSamplingFreq2					(kAnchorSamplingFreqSec / 32)	// <rdar://problem/7378275>
#define kAnchorSamplingFreq3					(kAnchorSamplingFreqSec / 16)	// <rdar://problem/7378275>
#define kAnchorSamplingFreq4					(kAnchorSamplingFreqSec /  8)	// <rdar://problem/7378275>

#ifdef TARGET_PDJ
 #define kMinimumNumErasesPerBuffer				12
#endif

#define kFormatChangeDelayInMs					667
#define kStartDelayOffset						5

#ifdef TARGET_PDJ
 #define kMaxTriesForStreamInterface			300		// 300 x 10ms = 3 second timeout
#else
 #define kMaxTriesForStreamPropertiesReady		500		// <rdar://problem/6686515> 500 x 10ms = 5 second timeout
#endif

// <rdar://5131786> Publish property for USB audio plugin.
#define kIDVendorString							"idVendor"
#define kIDProductString						"idProduct"

/*------------------------------------------*/
/*			PdjUsbAudioEngine クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbAudioEngine -

class PdjUsbAudioEngine : public IOAudioEngine
{
	OSDeclareDefaultStructors(PdjUsbAudioEngine)

public:

	typedef IOAudioEngine				super;

#pragma mark - Public Methods -

	bool
	init(
		OSArray						*streamInterfaceNumberArray
	);

	void
	free(
		void
	);

	void									// <rdar://7295322>
	registerService(
		IOOptionBits				options = 0
	);

	bool
	matchPropertyTable(
		OSDictionary				*table,
		SInt32						*score
	);

	bool
	start(
		IOService					*provider,
		IOAudioDevice				*device
	);

	void
	stop(
		IOService					*provider
	);

	bool
	requestTerminate(
		IOService					*provider,
		IOOptionBits				options
	);

	bool
	terminate(
		IOOptionBits				options = 0
	);

#ifdef TARGET_PDJ
 	void
 	systemCapabilityWillChangeTo(
 		UInt32						fromCapabilities,
 		UInt32						toCapabilities
 	);
 
 	void
 	systemCapabilityDidChangeTo(
 		UInt32						fromCapabilities,
 		UInt32						toCapabilities
 	);
#endif

	OSString *
	getGlobalUniqueID(
		void
	);

	bool
	initHardware(
		IOService					*provider
	);

	void
	openStreamInterfaces(
		void
	);

	void
	closeStreamInterfaces(
		void
	);

	IOReturn
	performAudioEngineStart(
		void
	);

	IOReturn
	performAudioEngineStop(
		void
	);

	IOReturn
	performFormatChange(
		IOAudioStream				*audioStream,
		const IOAudioStreamFormat	*newFormat,
		const IOAudioSampleRate		*newSampleRate
	);

	IOReturn
	controlledFormatChange(
		IOAudioStream				*audioStream,
		const IOAudioStreamFormat	*newFormat,
		const IOAudioSampleRate		*newSampleRate
	);

	IOReturn								// <rdar://5811247>
	republishAvailableFormats(
		IOAudioSampleRate			sampleRate
	);

	IOReturn
	convertInputSamples(
		const void					*sampleBuf,
		void						*destBuf,
		UInt32						firstSampleFrame,
		UInt32						numSampleFrames,
		const IOAudioStreamFormat	*streamFormat,
		IOAudioStream				*audioStream
	);

	IOReturn
	clipOutputSamples(
		const void					*mixBuf,
		void						*sampleBuf,
		UInt32						firstSampleFrame,
		UInt32						numSampleFrames,
		const IOAudioStreamFormat	*streamFormat,
		IOAudioStream				*audioStream
	);

	void
	resetClipPosition(
		IOAudioStream				*audioStream,
		UInt32						clipSampleFrame
	);

#ifdef TARGET_PDJ
#else
	// for UHCI
	IOReturn
	eraseOutputSamples(
		const void					*mixBuf,
		void						*sampleBuf,
		UInt32						firstSampleFrame,
		UInt32						numSampleFrames,
		const IOAudioStreamFormat	*streamFormat,
		IOAudioStream				*audioStream
	);
#endif

	UInt32
	getCurrentSampleFrame(
		void
	);

	void
	takeTimeStamp(
		bool						incrementLoopCount = true,
		AbsoluteTime				*timestamp = NULL
	);

	// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
	OSSet *
	copyDefaultAudioControls(
		void
	);

	void									// <rdar://5811247>
	runPolledTask(
		void
	);

	void									// <rdar://6430836> <rdar://problem/6706026>
	updateChannelNames(
		OSArray						*thePath,
		UInt8						interfaceNum,
		UInt8						altSettingNum
	);

	void									// <rdar://5811247>
	updateClockStatus(
		UInt8						clockID
	);

	void
	setOutputSampleLatency(
		UInt32						numSamples
	)
	{
		super::setOutputSampleLatency(numSamples);
	}

	void
	setInputSampleLatency(
		UInt32						numSamples
	)
	{
		super::setInputSampleLatency(numSamples);
	}

	void
	registerPlugin(
		PdjUsbAudioPlugin			*thePlugin
	);

	void
	pluginSetConfigurationApp(
		const char					*bundleID
	);

	IOReturn
	pluginDeviceRequest(
		StandardUSB::DeviceRequest	*request,
		IOMemoryDescriptor			*dataBuffer,
		IOUSBHostCompletion			*completion
	);

#ifdef TARGET_PDJ
 	UInt16 *
 	getInputSizeList(
 		bool						input,
 		UInt32						numTransactionsPerList
 	);
 
 	void
 	requestStreamingRestart(
 		void
 	)
 	{
 		mShouldRestartStreaming = true;
 	}
#endif

	UInt32
	getNumSampleFramesPerBuffer(
		void
	)
	{
		return super::getNumSampleFramesPerBuffer();
	}

protected:

#pragma mark - Protected Methods -

#ifdef TARGET_PDJ
 	static IOReturn
 	restartEngineAction(
 		OSObject		*device,
 		void			*engine,
 		void			*arg1,
 		void			*arg2,
 		void			*arg3
 	);
 
 	void
 	restartEngine(
 		void
 	);
#endif

	static void
	pluginLoaded(
		PdjUsbAudioEngine			*usbAudioEngineObject
	);

	static IOReturn							// <rdar://5811247>
	controlChangedHandler(
		OSObject					*target,
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	IOReturn								// <rdar://5811247>
	protectedControlChangedHandler(
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	IOReturn								// <rdar://5811247>
	doClockSelectorSetup(
		UInt8						interfaceNum,
		UInt8						altSettingNum,
		UInt32						sampleRate
	);

	IOReturn								// <rdar://5811247>
	doClockSelectorChange(
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	bool									// <rdar://5811247>
	determineMacSyncMode(
		UInt8						clockID
	);

	UInt32									// <rdar://7558825>
	getSystemClockDomain(
		void
	);

	IOReturn								// added for rdar://3866513
	getDefaultSampleRate(
		IOAudioSampleRate			*sampleRate
	);

	IOAudioSampleRate						// <rdar://6945472>
	getCurrentClockPathSampleRate(
		void
	);

	OSString *								// <rdar://6430836> <rdar://problem/6706026>
	getChannelNameString(
		UInt8						unitID,
		UInt8						channelNum
	);

	char *									// <rdar://6430836>
	channelConfigString(
		UInt8						channel
	);

	bool
	driverDesiresHiResSampleIntervals(
		void
	)
	{
		return false;
	}

public:

	PdjUsbAudioStream					*mMainOutputStream;
	PdjUsbAudioStream					*mMainInputStream;
	OSArray								*mIOAudioStreamArray;
	bool								mUSBStreamRunning;
	IOAudioSampleRate					mDefaultAudioSampleRate;	// Default sample rate are stored.
	UInt32								mStartOutputChannelID;
	UInt32								mStartInputChannelID;
	IOAudioSelectorControl				*mClockSelectorControl;				// <rdar://5811247>
	UInt8								mCurrentClockSourceID;				// <rdar://5811247>
	UInt8								mCurrentClockPathIndex;				// <rdar://5811247>

protected:

	thread_call_t						mPluginInitThread;
	PdjUsbAudioPlugin					*mPlugin;
	PdjUsbAudioDevice					*mUSBAudioDevice;
	OSArray								*mStreamInterfaceNumberArray;
	IOAudioSampleRate					mCurSampleRate;

#ifdef TARGET_PDJ
 	UInt32								mSystemSleep;
 	bool								mShouldRestartStreaming;
 	UInt64								mLastTimeStamp_nanos;
#else
	bool								mUHCISupport;
#endif

#ifdef DEBUGLOGGING
	UInt32								mNextSampleFrame;
#endif

#if DEBUGLATENCY
	bool								mHaveClipped;
#endif

	// Clock control
	UInt8								mCurrentClockPathGroupIndex;		// <rdar://5811247>
	bool								mRestoreClockSelection;				// <rdar://5811247>
	UInt32								mRestoreClockSelectionValue;		// <rdar://5811247>
	bool								mShouldPollClockStatus;				// <rdar://5811247>
	UInt32								mPollClockStatusCounter;			// <rdar://5811247>
	bool								mClockSourceValidity;				// <rdar://5811247>
	bool								mClockSourceValidityInitialized;	// <rdar://5811247>
	bool								mShouldRepublishFormat;				// <rdar://5811247>

#ifdef TARGET_PDJ
 	UInt16								mUSBInputSizeList[2048];
 	UInt16								mUSBInputIndex;
 	UInt16								mUSBOutputIndex;
#endif
};

#endif /* PDJ_USB_AUDIO_ENGINE_H_FILE_ */
