/*
 * Copyright (c) 1998-2008 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioPlugin.cpp
 *
 * @brief	IOService クラスの派生クラス．
 *			オーディオ入出力データに対し追加の処理を行うプラグイン．
 *			PdjUsbAudioStream または PdjUsbAudioEngine に
 *			アタッチすると有効になる．
 *			アタッチするには Info.plist を書き換える必要がある．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/

/*------------------------------------------*/
/*				インクルード				*/
/*------------------------------------------*/

#include "PdjUsbAudioCommon.h"

/*------------------------------------------*/
/*			PdjUsbAudioPlugin クラス		*/
/*------------------------------------------*/

OSDefineMetaClassAndStructors(PdjUsbAudioPlugin, super)

OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  0)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  1)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  2)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  3)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  4)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  5)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  6)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  7)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  8)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin,  9)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin, 10)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin, 11)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin, 12)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin, 13)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin, 14)
OSMetaClassDefineReservedUnused(PdjUsbAudioPlugin, 15)

#pragma mark - Public Methods -

/****************************************************************************/
/**
 * @brief	サービス開始．
 *			IOService::start() をオーバーライド．
 *			サービスプロバイダ（PdjUsbAudioStream または
 *			PdjUsbAudioEngine）に自身を登録する．
 *
 * @param[in]		*provider		サービスプロバイダ．
 *									PdjUsbAudioStream クラスまたは
 *									PdjUsbAudioEngine クラスのインスタンス．
 *
 * @retval			true			開始成功．
 * @retval			false			開始失敗．
 */
bool
PdjUsbAudioPlugin::start(
	IOService		*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::start()");

	bool			result = false;

	FailIf(!super::start(provider), Exit);

	reserved = (ExpansionData *)IOMalloc(sizeof(struct ExpansionData));
	FailIf(NULL == reserved, Exit);

	reserved->streamProvider = OSDynamicCast(PdjUsbAudioStream, provider);

	if (reserved->streamProvider)
	{
		// Tell PdjUsbAudioStream that we're loaded
		reserved->streamProvider->registerPlugin(this);
	}
	else
	{
		mOurProvider = OSDynamicCast(PdjUsbAudioEngine, provider);

		if (mOurProvider)
		{
			// Tell PdjUsbAudioEngine that we're loaded
			mOurProvider->registerPlugin(this);
		}
	}

	result = true;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::start() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	サービス停止．
 *			IOService::stop() をオーバーライド．
 *
 * @param[in]		*provider		サービスプロバイダ．
 *									PdjUsbAudioStream クラスまたは
 *									PdjUsbAudioEngine クラスのインスタンス．
 */
void
PdjUsbAudioPlugin::stop(
	IOService		*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::stop()");

	// Tell the system that we're not an available resource anymore
	super::publishResource("PdjUsbAudioPlugin", NULL);

	if (reserved)
	{
		IOFree(reserved, sizeof(struct ExpansionData));
	}

	super::stop(provider);

	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::stop()");
}

/****************************************************************************/
/**
 * @brief	デバイス設定用の外部アプリを登録．
 *			アタッチ先のドライバに外部アプリを関連付ける．
 *
 * @param[in]		*bundleID		外部アプリのバンドル ID 文字列．
 */
void
PdjUsbAudioPlugin::pluginSetConfigurationApp(
	const char		*bundleID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::pluginSetConfigurationApp()");

	if (reserved && reserved->streamProvider)
	{
		reserved->streamProvider->pluginSetConfigurationApp(bundleID);
	}
	else if (mOurProvider)
	{
		mOurProvider->pluginSetConfigurationApp(bundleID);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::pluginSetConfigurationApp()");
}

/****************************************************************************/
/**
 * @brief	デバイスリクエスト．
 *			アタッチ先のドライバを通じて、USB デバイスリクエストを発行する．
 *
 * @param[in]		*request		デバイスリクエスト．
 * @param[in, out]	*dataBuffer		デバイスリクエストのデータ．
 * @param[in]		*completion		完了ルーチン．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioPlugin::pluginDeviceRequest(
	StandardUSB::DeviceRequest	*request,
	IOMemoryDescriptor			*dataBuffer,
	IOUSBHostCompletion			*completion
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::pluginDeviceRequest()");

	IOReturn			result = kIOReturnError;

	if (reserved && reserved->streamProvider)
	{
		result = reserved->streamProvider->pluginDeviceRequest(request, dataBuffer, completion);
	}
	else if (mOurProvider)
	{
		result = mOurProvider->pluginDeviceRequest(request, dataBuffer, completion);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::pluginDeviceRequest() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	プラグイン初期化．
 *			プラグインがサービスプロバイダに登録されたときに呼び出される．
 *
 * @param[in]		*provider		サービスプロバイダ．
 *									PdjUsbAudioStream クラスまたは
 *									PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		vendorID		デバイスの Vendor ID．
 * @param[in]		productID		デバイスの Product ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioPlugin::pluginInit(
	IOService		*provider,
	UInt16			vendorID,
	UInt16			productID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::pluginInit()");

	IOReturn		result = kIOReturnSuccess;

	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::pluginInit() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	入出力方向通知．
 *			ストリーミングの方向が入力か出力か、サービスプロバイダから通知される．
 *
 * @param[in]		direction		kIOAudioStreamDirectionOutput または
 *									kIOAudioStreamDirectionInput．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioPlugin::pluginSetDirection(
	IOAudioStreamDirection	direction
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::pluginSetDirection()");

	IOReturn		result = kIOReturnSuccess;

	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::pluginSetDirection() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオフォーマット通知．
 *			オーディオフォーマットとサンプリング周波数が、サービスプロバイダから通知される．
 *
 * @param[in]		*newFormat		オーディオフォーマット．
 * @param[in]		*newSampleRate	サンプリング周波数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioPlugin::pluginSetFormat(
	const IOAudioStreamFormat	*const newFormat,
	const IOAudioSampleRate		*const newSampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::pluginSetFormat()");

	IOReturn		result = kIOReturnSuccess;

	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::pluginSetFormat() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオエンジン開始通知．
 *			オーディオエンジン開始時に、サービスプロバイダから呼び出される．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioPlugin::pluginStart(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::pluginStart()");

	IOReturn		result = kIOReturnSuccess;

	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::pluginStart() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオエンジン停止通知．
 *			オーディオエンジン停止時に、サービスプロバイダから呼び出される．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioPlugin::pluginStop(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::pluginStop()");

	IOReturn		result = kIOReturnSuccess;

	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::pluginStop() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオ出力用リングバッファのリセット通知．
 *			オーディオ出力用リングバッファがリセットされたときに、
 *			サービスプロバイダから呼び出される．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioPlugin::pluginReset(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOPLUGIN "::pluginReset()");

	IOReturn		result = kIOReturnSuccess;

	DP(DBG_TRACE, "- " LOGSTRAUDIOPLUGIN "::pluginReset() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオ出力データの処理．
 *			出力直前のオーディオデータ（浮動小数点型）を取得または加工することができる．
 *
 * @param[in, out]	*mixBuf				ミックスバッファ中のオーディオデータ．
 * @param[in]		numSampleFrames		取得または加工できるサンプルフレーム数．
 * @param[in]		numChannels			ストリームのチャンネル数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioPlugin::pluginProcessOutput(
	float			*mixBuf,
	UInt32			numSampleFrames,
	UInt32			numChannels
)
{
	DP(DBG_WRITE, "+ " LOGSTRAUDIOPLUGIN "::pluginProcessOutput()");

	IOReturn		result = kIOReturnSuccess;

	DP(DBG_WRITE, "- " LOGSTRAUDIOPLUGIN "::pluginProcessOutput() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオ入力データの処理．
 *			入力直後のオーディオデータ（浮動小数点型）を取得または加工することができる．
 *
 * @param[in]		*destBuf			入力バッファ中のオーディオデータ．
 * @param[in]		numSampleFrames		取得または加工できるサンプルフレーム数．
 * @param[in]		numChannels			ストリームのチャンネル数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioPlugin::pluginProcessInput(
	float			*destBuf,
	UInt32			numSampleFrames,
	UInt32			numChannels
)
{
	DP(DBG_READ, "+ " LOGSTRAUDIOPLUGIN "::pluginProcessInput()");

	IOReturn		result = kIOReturnSuccess;

	DP(DBG_READ, "- " LOGSTRAUDIOPLUGIN "::pluginProcessInput() = %d", result);
	return result;
}
