/*
 * Copyright (c) 1998-2009 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioTopology.h
 *
 * @brief	OSObject クラスの派生クラス．
 *			USB Audio Class の Audio Function Topology に関わる処理．
 *			AppleUSBAudioDevice クラスを２つに分離して
 *			PdjUsbAudioDevice クラスおよび PdjUsbAudioTopology クラスとした．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/
#ifndef PDJ_USB_AUDIO_TOPOLOGY_H_FILE_
#define PDJ_USB_AUDIO_TOPOLOGY_H_FILE_

/*------------------------------------------*/
/*			PdjUsbAudioTopology クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbAudioTopology -

class PdjUsbAudioTopology : public OSObject
{
	OSDeclareDefaultStructors(PdjUsbAudioTopology)

public:

	typedef OSObject					super;

#pragma mark - Public Methods -

	static PdjUsbAudioTopology *
	create(
		PdjUsbAudioDevice			*audioDevice
	);

	bool
	init(
		PdjUsbAudioDevice			*audioDevice
	);

	void
	free(
		void
	);

	OSArray *
	buildConnectionGraph(
		UInt8						controlInterfaceNum
	);

	OSArray *										// [rdar://4801032]
	buildClockGraph(
		UInt8						controlInterfaceNum
	);

	OSArray *										// [rdar://4867843], <rdar://5811247>
	getOptimalClockPath(
		PdjUsbAudioEngine			*thisEngine,
		UInt8						streamInterface,
		UInt8						altSetting,
		UInt32						sampleRate,
		bool						*otherEngineNeedSampleRateChange,
		UInt8						*clockPathGroupIndex = NULL
	);

	OSArray *										// <rdar://5811247>
	getClockPathGroup(
		UInt8						streamInterface,
		UInt8						altSetting,
		UInt8						*clockPathGroupIndex = NULL
	);

	OSArray *										// <rdar://5811247>
	getClockPathGroup(
		UInt8						pathGroupIndex
	);

	OSArray *
	findStreamsWithCommonClocks(
		OSArray						*availableStreamList
	);

	OSArray *
	findStreamsWithCommonSampleRates(
		OSArray						*availableStreamList
	);

	OSArray *
	findStreamsWithCompatibleEndpoints(
		OSArray						*availableStreamList
	);

	IOReturn										// [rdar://4867779]
	addSampleRatesFromClockSpace(
		void
	);

	IOReturn										// <rdar://6945472>
	setClockPathCurSampleRate(
		UInt32						sampleRate,
		OSArray						*clockPath,
		bool						failIfReadOnly = false
	);

	IOReturn										// <rdar://6945472>
	getClockPathCurSampleRate(
		UInt32						*sampleRate,
		bool						*validity,
		bool						*isReadOnly,
		OSArray						*clockPath
	);

	IOReturn
	setCurClockSourceSamplingFrequency(
		UInt8						unitID,
		UInt32						samplingFrequency
	);

	IOReturn
	getCurClockSourceSamplingFrequency(
		UInt8						unitID,
		UInt32						*samplingFrequency,
		bool						*validity
	);

	IOReturn
	setCurClockSelector(
		UInt8						unitID,
		UInt8						selector
	);

	IOReturn										// <rdar://5811247>
	getClockSelectorIDAndPathIndex(
		UInt8						*selectorID,
		UInt8						*pathIndex,
		OSArray						*clockPath
	);

	IOReturn
	doControlStuff(
		IOAudioEngine				*audioEngine,
		UInt8						interfaceNum,
		UInt8						altSettingNum
	);

	void
	controlHasChangedOnDevice(
		UInt8						controlID,
		OSSet						*defaultAudioControls
	);

	IOReturn
	getBothEngines(
		PdjUsbAudioEngine			**firstEngine,
		PdjUsbAudioEngine			**secondEngine
	);

	PdjUsbAudioEngine *
	otherEngine(
		PdjUsbAudioEngine			*thisEngine
	);

	void
	setMonoState(
		bool						state
	)
	{
		mDeviceIsInMonoMode = state;
	}

protected:

#pragma mark - Protected Methods -

	OSArray *
	buildPath(
		UInt8						controlInterfaceNum,
		UInt8						startingUnitID,
		OSArray						*allPaths,
		OSArray						*thisPath
	);

	void											// [rdar://4801032]
	buildClockPath(
		UInt8						controlInterfaceNum,
		UInt8						startingUnitID,
		OSArray						*allPaths,
		OSArray						*thisPath
	);

	UInt32											// [rdar://4867843]
	determineClockPathUnitUsage(
		PdjUsbAudioEngine			*thisEngine,
		OSArray						*thisClockPath
	);

	bool											// [rdar://4867843]
	clockPathCrossed(
		OSArray						*clockPathA,
		OSArray						*clockPathB
	);

	bool
	isClockCommonWithAtLeastOneStreamsInList(
		OSNumber					*refStreamInterfaceNumber,
		OSArray						*streamInterfaceNumberList
	);

	bool
	isClockCommonWithAllStreamsInList(
		OSNumber					*refStreamInterfaceNumber,
		OSArray						*streamInterfaceNumberList
	);

	bool
	streamsHaveCommonClocks(
		OSNumber					*streamInterfaceNumberA,
		OSNumber					*streamInterfaceNumberB
	);

	bool
	isClockPathCrossed(
		OSNumber					*streamInterfaceNumber,
		OSArray						*otherClockPath,
		UInt32						sampleRate
	);

	bool											// [rdar://4867843]
	supportSampleRateInClockPath(
		OSArray						*pathArray,
		UInt32						sampleRate
	);

	IOReturn
	getNumSampleRatesForClockPath(
		UInt8						*numSampleRates,
		OSArray						*clockPath
	);

	IOReturn
	getIndexedSampleRatesForClockPath(
		SubRange32					*sampleRates,
		OSArray						*clockPath,
		UInt32						rangeIndex
	);

	IOReturn										// [rdar://4867779]
	addSampleRatesFromClockPath(
		OSArray						*path,
		UInt8						streamInterface,
		UInt8						altSetting
	);

	IOReturn										// [rdar://4867779]
	getClockSourceSampleRates(
		OSArray						**sampleRates,
		UInt8						clockSource
	);

	IOReturn
	getNumClockSourceSamplingFrequencySubRanges(
		UInt8						unitID,
		UInt16						*numSubRanges
	);

	IOReturn
	getIndexedClockSourceSamplingFrequencySubRange(
		UInt8						unitID,
		SubRange32					*subRange,
		UInt16						subRangeIndex
	);

	IOReturn
	getCurClockMultiplier(
		UInt8						unitID,
		UInt16						*numerator,
		UInt16						*denominator
	);

	IOReturn
	setClockSetting(
		UInt8						controlSelector,
		UInt8						unitID,
		UInt8						requestType,
		void						*target,
		UInt16						length
	);

	IOReturn
	getClockSetting(
		UInt8						controlSelector,
		UInt8						unitID,
		UInt8						requestType,
		void						*target,
		UInt16						length
	);

	IOReturn
	setCurMute(
		UInt8						unitID,
		UInt8						channelNumber,
		SInt16						mute
	);

	IOReturn
	getCurMute(
		UInt8						unitID,
		UInt8						channelNumber,
		SInt16						*target
	);

	IOReturn
	setCurVolume(
		UInt8						unitID,
		UInt8						channelNumber,
		SInt16						volume
	);

	IOReturn
	getCurVolume(
		UInt8						unitID,
		UInt8						channelNumber,
		SInt16						*target
	);

	IOReturn
	getMinVolume(
		UInt8						unitID,
		UInt8						channelNumber,
		SInt16						*target
	);

	IOReturn
	getMaxVolume(
		UInt8						unitID,
		UInt8						channelNumber,
		SInt16						*target
	);

	IOReturn
	getVolumeResolution(
		UInt8						unitID,
		UInt8						channelNumber,
		UInt16						*target
	);

	IOReturn
	setFeatureUnitSetting(
		UInt8						controlSelector,
		UInt8						unitID,
		UInt8						channelNumber,
		UInt8						requestType,
		UInt16						newValue,
		UInt16						newValueLen
	);

	IOReturn
	getFeatureUnitSetting(
		UInt8						controlSelector,
		UInt8						unitID,
		UInt8						channelNumber,
		UInt8						requestType,
		SInt16						*target
	);

	IOReturn
	getFeatureUnitRange(
		UInt8						controlSelector,
		UInt8						unitID,
		UInt8						channelNumber,
		UInt8						requestType,
		SubRange16					*target
	);

	IOReturn
	setSelectorSetting(
		UInt8						selectorID,
		UInt8						setting
	);

	UInt8
	getSelectorSetting(
		UInt8						selectorID
	);

	IOFixed
	convertUSBVolumeTodB(
		SInt16						volume
	);

	bool
	isSampleRateCommonWithAtLeastOneStreamsInList(
		OSNumber					*refStreamInterfaceNumber,
		OSArray						*streamInterfaceNumberList
	);

	bool
	isSampleRateCommonWithAllStreamsInList(
		OSNumber					*refStreamInterfaceNumber,
		OSArray						*streamInterfaceNumberList
	);

	bool
	streamsHaveCommonSampleRates(
		OSNumber					*streamInterfaceNumberA,
		OSNumber					*streamInterfaceNumberB
	);

	OSArray *
	getSampleRatesFromStreamInterface(
		OSNumber					*streamInterfaceNumber
	);

	bool
	compareSampleRates(
		OSArray						*sampleRatesA,
		OSArray						*sampleRatesB
	);

	void
	mergeSampleRates(
		OSArray						*thisArray,
		OSArray						*otherArray
	);

	OSArray *
	findStreamsWithDirectionAndSyncType(
		OSArray						*availableStreamList,
		UInt8						direction,
		UInt8						syncType
	);

	bool
	streamEndpointsHaveSpecifiedDirectionAndSyncType(
		OSNumber					*streamInterfaceNumber,
		UInt8						endpointDirection,
		UInt8						endpointSyncType
	);

	IOReturn										// <rdar://6420832>
	registerEngineInfo(
		PdjUsbAudioEngine			*usbAudioEngine
	);

	SInt32
	getEngineInfoIndex(
		PdjUsbAudioEngine			*inAudioEngine
	);

	SInt32											// <rdar://6420832>
	getStreamInfoIndex(
		UInt8						interfaceNum
	);

	IOReturn										// <rdar://6420832>
	registerStreamInfo(
		UInt8						interfaceNum,
		UInt8						altSettingNum
	);

	UInt8											// <rdar://6413207>
	getDefaultOutputTerminalID(
		UInt8						inputTerminalID
	);

	UInt32											// <rdar://6413207>
	getNumConnectedOutputTerminals(
		UInt8						inputTerminalID
	);

	UInt8
	getBestFeatureUnitInPath(
		OSArray						*thePath,
		UInt32						direction,
		UInt8						interfaceNum,
		UInt8						altSettingNum,
		UInt32						controlTypeWanted
	);

	void											// <rdar://6413207>
	addVolumeControls(
		PdjUsbAudioEngine			*usbAudioEngine,
		UInt8						featureUnitID,
		UInt8						terminalID,
		UInt8						interfaceNum,
		UInt8						altSettingNum,
		UInt32						usage
	);

	void											// <rdar://6413207>
	addMuteControl(
		PdjUsbAudioEngine			*usbAudioEngine,
		UInt8						featureUnitID,
		UInt8						terminalID,
		UInt8						interfaceNum,
		UInt8						altSettingNum,
		UInt32						usage
	);

	UInt32											// <rdar://6523676>, <rdar://problem/6656791>
	getPathIndexForSelectorSetting(
		OSArray						*arrayOfPathsFromOutputTerminal,
		UInt32						pathsToOutputTerminalN,
		UInt32						graphPathIndex,
		UInt8						selectorUnitIndex,
		UInt8						selectorSetting
	);

	IOReturn										// <rdar://problem/6656791>
	addSelectorSourcesToSelectorControl(
		IOAudioSelectorControl		*theSelectorControl,
		OSArray						*arrayOfPathsFromOutputTerminal,
		UInt32						pathsToOutputTerminalN,
		UInt32						graphPathIndex,
		UInt8						selectorIndex
	);

	OSArray *										// <rdar://5366067>
	getPlaythroughPaths(
		UInt8						inputTerminalID
	);

	IOReturn										// <rdar://5366067>
	doPlaythroughSetup(
		PdjUsbAudioEngine			*usbAudioEngine,
		OSArray						*playThroughPaths,
		UInt8						interfaceNum,
		UInt8						altSettingNum,
		UInt8						inputTerminalID
	);

	UInt8
	pathsContaining(
		UInt8						unitID
	);

	UInt8											// <rdar://5366067>
	pathsContainingFeatureUnitButNotMixerUnit(
		UInt8						featureUnitID,
		UInt8						mixerUnitID
	);

	OSString *
	getNameForPath(
		OSArray						*arrayOfPathsFromOutputTerminal,
		UInt32						*pathIndex,
		UInt8						startingPoint
	);

	OSString *
	getNameForMixerPath(
		OSArray						*arrayOfPathsFromOutputTerminal,
		UInt32						*pathIndex,
		UInt8						startingPoint
	);

	OSString *										// <rdar://6430836>
	getNameForTerminal(
		UInt8						terminalID,
		UInt8						direction
	);

	static IOReturn
	controlChangedHandler(
		OSObject					*target,
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	IOReturn
	protectedControlChangedHandler(
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	IOReturn
	doVolumeControlChange(
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	IOReturn
	doToggleControlChange(
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	IOReturn
	doSelectorControlChange(
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	IOReturn
	doInputSelectorChange(
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	IOReturn										// <rdar://6413207>
	doOutputSelectorChange(
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

	IOReturn
	doPassThruSelectorChange(
		IOAudioControl				*audioControl,
		SInt32						oldValue,
		SInt32						newValue
	);

protected:

	PdjUsbAudioDevice					*mAudioDevice;
	PdjUsbConfigurationDictionary		*mConfigDictionary;
	IOUSBHostInterface					*mControlInterface;
	OSArray								*mControlGraph;
	OSArray								*mClockGraph;
	bool								mDeviceIsInMonoMode;	// this flag is set by PdjUsbAudioEngine::controlledFormatChange
	OSArray								*mMonoControlsArray;
};

#endif /* PDJ_USB_AUDIO_TOPOLOGY_H_FILE_ */
