/*
 * Copyright (c) 1998-2010 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioStream.cpp
 *
 * @brief	IOAudioStream クラスの派生クラス．
 *			オーディオストリーミング用のリングバッファを表現する．
 *			PdjUsbAudioEngine::initHardware() においてインスタンス化される．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/

/*------------------------------------------*/
/*				インクルード				*/
/*------------------------------------------*/

#include "PdjUsbAudioCommon.h"

/*------------------------------------------*/
/*				マクロ						*/
/*------------------------------------------*/

#if DEBUGTIMESTAMPS
 #define MAGNITUDEOF(x)		(((x) > 0) ? (x) : (-(x)))
#endif

/*------------------------------------------*/
/*			PdjUsbAudioStream クラス		*/
/*------------------------------------------*/

OSDefineMetaClassAndStructors(PdjUsbAudioStream, super)

#pragma mark - Public Methods -

/****************************************************************************/
/**
 * @brief	インスタンス初期化．
 *			IOAudioStream::initWithAudioEngine() とはシグネチャが異なる．
 *			PdjUsbAudioEngine::initHardware() から呼び出される．
 *
 * @param[in]		*device				PdjUsbAudioDevice のインスタンス．
 * @param[in]		*engine				PdjUsbAudioEngine のインスタンス．
 * @param[in]		*streamInterface	このストリームに対応する USB interface．
 * @param[in]		sampleRate			サンプリング周波数の初期値．
 * @param[in]		*streamDescription	ストリームの名称．
 * @param[in]		*properties			デバイスプロパティ．
 *
 * @retval			true			初期化成功．
 * @retval			false			初期化失敗．
 */
bool
PdjUsbAudioStream::initWithAudioEngine(
	PdjUsbAudioDevice					*device,
	PdjUsbAudioEngine					*engine,
#ifdef TARGET_PDJ
 	UInt16								interfaceNumber,
#else
	IOUSBHostInterface					*streamInterface,
#endif
	IOAudioSampleRate					sampleRate,
	const char							*streamDescription,
	OSDictionary						*properties
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::initWithAudioEngine()", this);

	PdjUsbConfigurationDictionary		*configDictionary;
	bool								result = false;
	UInt32								startChannelID;

	FailIf(NULL == device, Exit);
	FailIf(NULL == engine, Exit);
#ifdef TARGET_PDJ
#else
	FailIf(NULL == streamInterface, Exit);
#endif

	mUSBAudioDevice = device;
	mUSBAudioDevice->retain();

	mUSBAudioEngine = engine;
	mUSBAudioEngine->retain();

#ifdef TARGET_PDJ
 	mStreamInterface = mUSBAudioDevice->mControlInterface;
#else
	mStreamInterface = streamInterface;
#endif
	mStreamInterface->retain();

#ifdef TARGET_PDJ
 	mInterfaceNumber = interfaceNumber;
#else
	mInterfaceNumber = mStreamInterface->getInterfaceDescriptor()->bInterfaceNumber;
#endif
	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::initWithAudioEngine() - mInterfaceNumber = %d", this, mInterfaceNumber);

	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	// Choose default alternate setting ID and sampling rate ( rdar://3866513 )

	FailIf(kIOReturnSuccess != this->getDefaultSettings(&mAlternateSettingID, &sampleRate), Exit);

	FailIf(
			kIOReturnSuccess != configDictionary->getIsocEndpointDirection(
														&mDirection,
														mInterfaceNumber,
														mAlternateSettingID
												  ),
			Exit
		  );

	startChannelID = (kIOAudioStreamDirectionOutput == mDirection)
						? mUSBAudioEngine->mStartOutputChannelID
						: mUSBAudioEngine->mStartInputChannelID;

	FailIf(
			!super::initWithAudioEngine(
						engine,
						(IOAudioStreamDirection)mDirection,
						startChannelID,
						streamDescription,
						properties
					),
			Exit
		  );

	// Safeguard against USB-Audio 2.0 descriptors for rdar://4798933.
	// Check both INTERFACE_PROTOCOL_UNDEFINED and IP_PROTOCOL_VERSION to protect against unsupported
	// protocols.
	FailIf(
			(
				(INTERFACE_PROTOCOL_UNDEFINED != mStreamInterface->getInterfaceDescriptor()->bInterfaceProtocol)
			&&
				(IP_VERSION_02_00 != mStreamInterface->getInterfaceDescriptor()->bInterfaceProtocol)
			),
			Exit
		  );

	// Change this to use defines from the IOAudioFamily when they are available
	IORegistryEntry::setProperty("IOAudioStreamSampleFormatByteOrder", "Little Endian");

	// Attach to the stream interface so we get willTerminate message.
	IOService::attach(mStreamInterface);

	mInitStampDifference = true;	// <rdar://problem/7378275>

	result = true;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::initWithAudioEngine()", this);
	return result;
}

/****************************************************************************/
/**
 * @brief	インスタンス解放．
 *			IOAudioStream::free() をオーバーライド．
 *			参照カウントがゼロになった時呼び出される．
 */
void
PdjUsbAudioStream::free(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::free()", this);

	if (NULL != mCoalescenceMutex)
	{
		IORecursiveLockFree(mCoalescenceMutex);		// <rdar://problem/7378275>
		mCoalescenceMutex = NULL;
	}

	if (NULL != mFrameQueuedForList)
	{
		delete[] mFrameQueuedForList;
		mFrameQueuedForList = NULL;
	}

	if (mAverageSampleRateBuffer)
	{
		// <rdar://7000283> Pointer was obtained from mAssociatedEndpointMemoryDescriptor,
		// so no need to free it explicitly here.
		mAverageSampleRateBuffer = NULL;
	}

	if (mAssociatedEndpointMemoryDescriptor)
	{
		mAssociatedEndpointMemoryDescriptor->release();
		mAssociatedEndpointMemoryDescriptor = NULL;
	}

	if (mUSBBufferDescriptor)
	{
		mUSBBufferDescriptor->release();
		mUSBBufferDescriptor = NULL;
	}

#ifdef BUG_FIX
	if (mWrapRangeDescriptor)
	{
		mWrapRangeDescriptor->release();
		mWrapRangeDescriptor = NULL;
	}

	if (mWrapDescriptors[0])
	{
		mWrapDescriptors[0]->release();
		mWrapDescriptors[0] = NULL;
	}

	if (mWrapDescriptors[1])
	{
		mWrapDescriptors[1]->release();
		mWrapDescriptors[1] = NULL;
	}
#else
	if (mWrapRangeDescriptor)
	{
		mWrapRangeDescriptor->release();
		mWrapDescriptors[0]->release();
		mWrapDescriptors[1]->release();
		mWrapRangeDescriptor = NULL;
	}
#endif

	if (mSampleBufferMemoryDescriptor)
	{
		mSampleBufferMemoryDescriptor->release();
		mSampleBufferMemoryDescriptor = NULL;
	}

	if (NULL != mSampleBufferDescriptors)
	{
		for (UInt32 i = 0; i < mNumUSBFrameLists; i++)
		{
			if (NULL != mSampleBufferDescriptors[i])
			{
				mSampleBufferDescriptors[i]->release();
				mSampleBufferDescriptors[i] = NULL;
			}
		}

		IOFree(mSampleBufferDescriptors, mNumUSBFrameLists * sizeof(IOSubMemoryDescriptor *));
		mSampleBufferDescriptors = NULL;
	}

	if (NULL != mUSBIsocFrames)
	{
		IOFree(mUSBIsocFrames, mNumUSBFrameLists * mNumTransactionsPerList * sizeof(IOUSBHostIsochronousFrame));
		mUSBIsocFrames = NULL;
	}

	if (NULL != mUSBCompletion)
	{
		IOFree(mUSBCompletion, mNumUSBFrameLists * sizeof(IOUSBHostIsochronousCompletion));
		mUSBCompletion = NULL;
	}

	if (mUSBAudioDevice)
	{
		mUSBAudioDevice->release();
		mUSBAudioDevice = NULL;
	}

	if (mUSBAudioEngine)
	{
		mUSBAudioEngine->release();
		mUSBAudioEngine = NULL;
	}

	if (mStreamInterface)
	{
		mStreamInterface->release();
		mStreamInterface = NULL;
	}

	super::free();

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::free()", this);
}

/****************************************************************************/
/**
 * @brief	サービスを IORegistry に登録．
 *			IOService::registerService() をオーバーライド．
 *
 * @param[in]		options			オプション指定．
 */
// <rdar://7295322> Asynchronous to prevent deadlock if the device or interface is terminated while
// registerService() is performing matching.
void
PdjUsbAudioStream::registerService(
	IOOptionBits	options
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::registerService(0x%x)", this, options);

	if (0 == (kIOServiceSynchronous & options))
	{
		options |= kIOServiceAsynchronous;
	}

	super::registerService(options);

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::registerService(0x%x)", this, options);
}

/****************************************************************************/
/**
 * @brief	サービスマッチング．
 *			IOService::matchPropertyTable() をオーバーライド．
 *			ドライバ固有のマッチング条件があるときはここで判定を行う．
 *
 * @param[in]		*table			マッチングの対象となるプロパティ．
 * @param[in, out]	*score			プローブスコア．必要に応じて書き換えることができる．
 *
 * @retval			true			マッチング成功．
 * @retval			false			マッチング失敗．
 */
bool
PdjUsbAudioStream::matchPropertyTable(
	OSDictionary	*table,
	SInt32			*score
)
{
	DP(DBG_VERBOSE, "+ " LOGSTRAUDIOSTREAM "[%p]::matchPropertyTable(%p, %p)", this, table, score);

	bool			returnValue = false;

	if (IOService::matchPropertyTable(table, score))
	{
		if (
				IOService::compareProperty(table, kIDVendorString)
		   &&
				IOService::compareProperty(table, kIDProductString)
		   &&
				IOService::compareProperty(table, kIOAudioStreamDirectionKey)
		   )
		{
			returnValue = true;
		}
	}

	DP(DBG_VERBOSE, "- " LOGSTRAUDIOSTREAM "[%p]::matchPropertyTable(%p, %p) = %d", this, table, score, returnValue);
	return returnValue;
}

/****************************************************************************/
/**
 * @brief	サービス停止．
 *			IOAudioStream::stop() をオーバーライド．
 *
 * @param[in]		*provider		サービスプロバイダ．PdjUsbAudioEngine クラスのインスタンス．
 */
void
PdjUsbAudioStream::stop(
	IOService		*provider
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::stop(%p)", this, provider);

	if (NULL != mPluginInitThread)
	{
		::thread_call_cancel(mPluginInitThread);
		::thread_call_free(mPluginInitThread);
		mPluginInitThread = NULL;
	}

	if (mPlugin)
	{
		mPlugin->close(this);
		mPlugin = NULL;
	}

	if (mUSBAudioDevice)
	{
		mUSBAudioDevice->release();
		mUSBAudioDevice = NULL;
	}

	if (mUSBAudioEngine)
	{
		mUSBAudioEngine->release();
		mUSBAudioEngine = NULL;
	}

	if (mPipe)
	{
		mPipe->release();
		mPipe = NULL;
	}

	if (mAssociatedPipe)
	{
		mAssociatedPipe->release();
		mAssociatedPipe = NULL;
	}

	// [rdar://4287899] We don't expect the stream interface to need closing unless the following conditions are true.
	if (
			(mStreamInterface)
	   &&
			(
				(provider == mUSBAudioEngine)
			||
				(provider == mStreamInterface)
			)
	   &&
			(mStreamInterface->isOpen())
	   )
	{
		DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "[%p]::stop() - mStreamInterface was still open when stop() was called. Closing ...", this);

		mStreamInterface->close(this);
		mStreamInterface->release();
		mStreamInterface = NULL;
	}

	super::stop(provider);

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::stop(%p) - rc = %d", this, provider, this->getRetainCount());
}

/****************************************************************************/
/**
 * @brief	サービス終了要求．
 *			IOService::requestTerminate() をオーバーライド．
 *			プロバイダがサービスを終了した際、このサービスも終了させてよいかどうか
 *			問い合わせるために呼び出される．
 *
 * @param[in]		*provider		サービスを終了したプロバイダ．
 *									PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		options			オプション指定．
 *
 * @retval			true			サービス終了要求を受け入れる．
 * @retval			false			サービス終了要求を拒否する．
 */
bool
PdjUsbAudioStream::requestTerminate(
	IOService		*provider,
	IOOptionBits	options
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::requestTerminate(%p, %x)", this, provider, options);

	bool			result;

	// if interface or audio device
	if ((mUSBAudioEngine == provider) || (mStreamInterface == provider))
	{
		result = true;		// it is OK to terminate us
	}
	else
	{
		result = false;		// don't terminate us
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::requestTerminate(%p, %x) = %d", this, provider, options, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	プロバイダ終了通知．
 *			IOService::willTerminate() をオーバーライド．
 *			サービスプロバイダがサービスを終了した時に呼び出される．
 *
 * @param[in]		*provider		サービスを終了したプロバイダ．
 *									PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		options			オプション指定．
 *
 * @retval			true			プロバイダの終了を受け入れる．
 * @retval			false			プロバイダの終了に対処できない．
 */
bool
PdjUsbAudioStream::willTerminate(
	IOService		*provider,
	IOOptionBits	options
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::willTerminate(%p)", this, provider);

	if ((mUSBAudioEngine == provider) || (mStreamInterface == provider))
	{
		mTerminatingDriver = true;

		// [rdar://5535738] Always close the stream interface here. Don't do it in the isoc completion routines.

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::willTerminate() - Closing stream interface", this);

		if ((mUSBStreamRunning) && (NULL != mPipe))
		{
			mPipe->abort();

			if (0 == mShouldStop)
			{
				mShouldStop++;
			}
		}

		if (NULL != mStreamInterface)
		{
			mStreamInterface->close(this);
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::willTerminate()", this);
	return super::willTerminate(provider, options);
}

/****************************************************************************/
/**
 * @brief	サービスを終了させる．
 *			IOService::terminate() をオーバーライド．
 *			このサービスを終了させるために呼び出される．
 *
 * @param[in]		options			オプション指定．
 *
 * @retval			true			成功．このサービスは終了される．
 * @retval			false			失敗．このサービスは終了されない．
 */
bool
PdjUsbAudioStream::terminate(
	IOOptionBits	options
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::terminate()", this);

	bool			shouldTerminate = true;
	bool			result = true;

	if (shouldTerminate)
	{
		result = super::terminate(options);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::terminate()", this);
	return result;
}

/****************************************************************************/
/**
 * @brief	対応する USB interface をオープン．
 *
 * @retval			true			成功．
 * @retval			false			失敗．
 */
bool
PdjUsbAudioStream::openStreamInterface(
	void
)
{
#ifdef TARGET_PDJ
 	return true;
#else
	bool			result = false;

	// <rdar://6277511> Make sure that mStreamInterface is not null.
	if (NULL != mStreamInterface)
	{
		result = mStreamInterface->open(this);
	}

	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	対応する USB interface をクローズ．
 */
void
PdjUsbAudioStream::closeStreamInterface(
	void
)
{
#ifdef TARGET_PDJ
#else
	// <rdar://6277511> Make sure that mStreamInterface is not null.
	if (NULL != mStreamInterface)
	{
		mStreamInterface->close(this);
	}
#endif
}

/****************************************************************************/
/**
 * @brief	ストリームの初期設定．
 *			PdjUsbAudioEngine::initHardware() から呼び出される．
 *
 * @param[in]		sampleRate		サンプリング周波数．
 *
 * @retval			true			成功．
 * @retval			false			失敗．
 */
bool
PdjUsbAudioStream::configureAudioStream(
	IOAudioSampleRate	sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::configureAudioStream()", this);

	bool								resultBool = false;
	OSNumber							*idVendor = NULL;
	OSNumber							*idProduct = NULL;
	PdjUsbConfigurationDictionary		*configDictionary;
	IOAudioStreamFormat					streamFormat;
	IOReturn							resultCode = kIOReturnError;
	UInt32								index;
	UInt16								terminalType;
	UInt16								format;

#if STAGGERINTERFACES
	if (mInterfaceNumber % 2 != 1)
	{
		IOSleep(1000);
	}
#endif

	mTerminatingDriver	= false;
	mCoalescenceMutex	= NULL;

	FailIf(NULL == mUSBAudioDevice, Exit);							// <rdar://7085810>
	FailIf(NULL == mUSBAudioDevice->mControlInterface, Exit);		// <rdar://7085810>
	FailIf(NULL == mStreamInterface, Exit);

	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	if (kIOAudioStreamDirectionInput == mDirection)
	{
		// look for a streaming output terminal that's connected to a non-streaming input terminal
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::configureAudioStream() - This is an input type endpoint (mic, etc.)", this);

		index = 0;

		do
		{
			resultCode = configDictionary->getIndexedInputTerminalType(
												&terminalType,
												mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
												0,
												index++
										   );
		} while ((terminalType == INPUT_UNDEFINED) && (index < 256) && (kIOReturnSuccess == resultCode));

		mNumUSBFrameLists		 = RECORD_NUM_USB_FRAME_LISTS;
		mNumUSBFramesPerList	 = RECORD_NUM_USB_FRAMES_PER_LIST;
		mNumUSBFrameListsToQueue = RECORD_NUM_USB_FRAME_LISTS_TO_QUEUE;

		// We need a mutex for coalesceInputSamples() in case something goes wrong at the start of the stream.
		mCoalescenceMutex = IORecursiveLockAlloc();
		FailIf(NULL == mCoalescenceMutex, Exit);
	}
	else if (kIOAudioStreamDirectionOutput == mDirection)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::configureAudioStream() - This is an output type endpoint (speaker, etc.)", this);

		index = 0;

		do
		{
			resultCode = configDictionary->getIndexedOutputTerminalType(
												&terminalType,
												mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
												0,
												index++
										   );
		} while ((terminalType == OUTPUT_UNDEFINED) && (index < 256) && (kIOReturnSuccess == resultCode));

		mNumUSBFrameLists		 = PLAY_NUM_USB_FRAME_LISTS;
		mNumUSBFramesPerList	 = PLAY_NUM_USB_FRAMES_PER_LIST;
		mNumUSBFrameListsToQueue = PLAY_NUM_USB_FRAME_LISTS_TO_QUEUE;
	}
	else
	{
		FailIf("Couldn't get the endpoint direction!", Exit);
	}

#ifdef TARGET_PDJ
#else
	// See if UHCI support is necessary
	mUHCISupport = mUSBAudioDevice->checkForUHCI();
	// mUHCISupport = true;
#endif

	mSplitTransactions = mUSBAudioDevice->detectSplitTransactions();

	mFrameQueuedForList = NULL;

	// Allocate frame list time stamp array
	mFrameQueuedForList = new UInt64[mNumUSBFrameLists];
	FailIf(NULL == mFrameQueuedForList, Exit);

	super::setTerminalType(terminalType);

	// [rdar://4801012] We can't allocate the isoc frames yet because we don't know how many transactions there will be per frame list until
	// the alternate setting is chosen. This must be done in controlledFormatChange().
	// mUSBIsocFrames = (IOUSBHostIsochronousFrame *)IOMalloc(mNumUSBFrameLists * mNumUSBFramesPerList * sizeof(IOUSBHostIsochronousFrame));

	mUSBCompletion = (IOUSBHostIsochronousCompletion *)IOMalloc(mNumUSBFrameLists * sizeof(IOUSBHostIsochronousCompletion));

	mSampleBufferDescriptors = (IOSubMemoryDescriptor **)IOMalloc(mNumUSBFrameLists * sizeof(IOSubMemoryDescriptor *));

	::bzero(mSampleBufferDescriptors, mNumUSBFrameLists * sizeof(IOSubMemoryDescriptor *));

	mWrapDescriptors[0] = OSTypeAlloc(IOSubMemoryDescriptor);
	mWrapDescriptors[1] = OSTypeAlloc(IOSubMemoryDescriptor);

	FailIf(NULL == mWrapDescriptors[0], Exit);
	FailIf(NULL == mWrapDescriptors[1], Exit);
	FailIf(NULL == mUSBCompletion, Exit);
	FailIf(NULL == mSampleBufferDescriptors, Exit);

	FailIf(kIOReturnSuccess != this->addAvailableFormats(configDictionary, sampleRate), Exit);

	mCurSampleRate = sampleRate;

	// Tell the IOAudioFamily what format we are going to be running in.
	// <rdar://problem/6892754> 10.5.7 Regression: Devices with unsupported formats stopped working
	FailIf(kIOReturnSuccess != configDictionary->getFormat(&format, mInterfaceNumber, mAlternateSettingID), Exit);

	streamFormat.fBitWidth = 0;

	if ((PCM == format) || (IEC1937_AC3 == format))
	{
		UInt8	numChannels;

		FailIf(kIOReturnSuccess != configDictionary->getNumChannels(&numChannels, mInterfaceNumber, mAlternateSettingID), Exit);

		streamFormat.fNumChannels = numChannels;

		FailIf(kIOReturnSuccess != configDictionary->getBitResolution(&(streamFormat.fBitDepth), mInterfaceNumber, mAlternateSettingID), Exit);
		FailIf(kIOReturnSuccess != configDictionary->getSubframeSize( &(streamFormat.fBitWidth), mInterfaceNumber, mAlternateSettingID), Exit);
	}

	streamFormat.fBitWidth	*= 8;
	streamFormat.fAlignment	 = kIOAudioStreamAlignmentLowByte;
	streamFormat.fByteOrder	 = kIOAudioStreamByteOrderLittleEndian;
	streamFormat.fDriverTag	 = (mInterfaceNumber << 16) | mAlternateSettingID;

	switch (format)
	{
		case PCM:
			streamFormat.fSampleFormat			= kIOAudioStreamSampleFormatLinearPCM;
			streamFormat.fNumericRepresentation	= kIOAudioStreamNumericRepresentationSignedInt;
			streamFormat.fIsMixable				= true;
			break;

		case AC3:	// just starting to stub something in for AC-3 support
			streamFormat.fSampleFormat			= kIOAudioStreamSampleFormatAC3;
			streamFormat.fNumericRepresentation	= kIOAudioStreamNumericRepresentationSignedInt;
			streamFormat.fIsMixable				= false;
			streamFormat.fNumChannels			= 6;
			streamFormat.fBitDepth				= 16;
			streamFormat.fBitWidth				= 16;
			streamFormat.fByteOrder				= kIOAudioStreamByteOrderBigEndian;
			break;

		case IEC1937_AC3:
			streamFormat.fSampleFormat			= kIOAudioStreamSampleFormat1937AC3;
			streamFormat.fNumericRepresentation	= kIOAudioStreamNumericRepresentationSignedInt;
			streamFormat.fIsMixable				= false;
			break;

		default:
			FailIf("Interface doesn't have any supported formats!", Exit);
	}

	// Store default stream format
	mDefaultAudioStreamFormat = streamFormat;

#ifdef TARGET_PDJ
 	FailIf(kIOReturnSuccess != (resultCode = mStreamInterface->selectAlternateSetting(kRootAlternateSetting)), Exit);
#else
	FailIf(!mStreamInterface->open(this), Exit);
	// Have to open the interface because calling setFormat will call performFormatChange, which expects the interface to be open.

	FailIf(kIOReturnSuccess != (resultCode = mStreamInterface->selectAlternateSetting(kRootAlternateSetting)), Exit);
#endif
	// Select the first alternate interface to init the hardware

	FailIf(kIOReturnSuccess != (resultCode = this->setFormat(&streamFormat)), Exit);

	// Verify that this 'start' request is targeting a USB Audio Stream interface
	// (i.e. it must be an audio class and a stream subclass).

#ifdef TARGET_PDJ
#else
	UInt8		deviceClass;
	UInt8		deviceSubclass;

	FailIf(kIOReturnSuccess != configDictionary->getInterfaceClass(&deviceClass, mInterfaceNumber, mAlternateSettingID), Exit);
	FailIf(AUDIO != deviceClass, Exit);

	FailIf(kIOReturnSuccess != configDictionary->getInterfaceSubClass(&deviceSubclass, mInterfaceNumber, mAlternateSettingID), Exit);
	FailIf(AUDIOSTREAMING != deviceSubclass, Exit);
#endif

	resultBool = true;

	// Ask for plugin to load (if it exists)
	idVendor = OSNumber::withNumber(mUSBAudioDevice->getVendorID(), 16);
	if (NULL != idVendor)
	{
		IORegistryEntry::setProperty(kIDVendorString, idVendor);
		idVendor->release();
	}

	idProduct = OSNumber::withNumber(mUSBAudioDevice->getProductID(), 16);
	if (NULL != idProduct)
	{
		IORegistryEntry::setProperty(kIDProductString, idProduct);
		idProduct->release();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::configureAudioStream(), resultCode = %x, resultBool = %d", this, resultCode, resultBool);
	return resultBool;
}

/****************************************************************************/
/**
 * @brief	オーディオストリーミングのオフセットとレイテンシを更新．
 */
void
PdjUsbAudioStream::updateSampleOffsetAndLatency(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::updateSampleOffsetAndLatency()");

	UInt32			minimumSafeSampleOffset;
	UInt32			cautiousSafeSampleOffset;
	UInt16			averageFrameSamples;
	UInt16			additionalSampleFrameFreq;
	bool			highSpeedCompensation;
	UInt32			newSampleOffset;
	UInt32			newSampleLatency;
	OSDictionary	*sampleOffsetDictionary	 = NULL;
	OSDictionary	*sampleLatencyDictionary = NULL;

	// <rdar://problem/7378275>
	averageFrameSamples			= mCurSampleRate.whole / 1000;	// per ms
	additionalSampleFrameFreq	= mCurSampleRate.whole - (averageFrameSamples * 1000);

	if (kIOAudioStreamDirectionInput == mDirection)
	{
		// Check to see if latency should be higher for EHCI (rdar://3959606 )
		if (mSplitTransactions)
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - Compensating for high speed timing difference in sample offset", this);

			highSpeedCompensation = true;
		}
		else
		{
			highSpeedCompensation = false;
		}

		minimumSafeSampleOffset = averageFrameSamples + 1;

		cautiousSafeSampleOffset = minimumSafeSampleOffset + (minimumSafeSampleOffset / kUSBInputRecoveryTimeFraction);

		newSampleOffset = cautiousSafeSampleOffset;

		// <rdar://6343818> Adjust the safety offset by 1.875ms to compensate for the time stamps generation
		// in the output stream when both the input & output streams are on the same engine.
		if (mSyncCompensation)
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - Compensating for time stamp generation in sample offset", this);

			newSampleOffset += cautiousSafeSampleOffset * 3 / 2;
		}

		// Check to see if there is an override in the vendor specific kext.
		sampleOffsetDictionary = OSDynamicCast(OSDictionary, mStreamInterface->getProperty(kIOAudioEngineInputSampleOffsetKey));

		if (NULL == sampleOffsetDictionary)
		{
			sampleOffsetDictionary = OSDynamicCast(OSDictionary, mStreamInterface->getProperty(kIOAudioEngineSampleOffsetKey));
		}

		if (NULL != sampleOffsetDictionary)
		{
			char	sampleOffsetKeyString[16];

			::snprintf(sampleOffsetKeyString, 16, "%u", mCurSampleRate.whole);

			OSNumber	*sampleOffset = OSDynamicCast(OSNumber, sampleOffsetDictionary->getObject(sampleOffsetKeyString));

			if (NULL != sampleOffset)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - override input sample offset (%u) to %u sample frames",
					this, newSampleOffset, sampleOffset->unsigned32BitValue());

				newSampleOffset = sampleOffset->unsigned32BitValue();
			}
		}

		// Add an extra frame and a half of samples to the offset if going through a USB 2.0 hub.
		newSampleOffset += highSpeedCompensation ? (5 * minimumSafeSampleOffset / 3) : 0;

        // @note workaround for longrun noise on macOS 10.12.6.
        newSampleOffset = (UInt32)((float)newSampleOffset * 1.72);
        
		// Set the offset for input devices (microphones, etc.)
		mUSBAudioEngine->setInputSampleOffset(newSampleOffset);

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - setting input sample offset to %u sample frames",
			this, newSampleOffset);

		newSampleLatency = averageFrameSamples * (/* kMinimumFrameOffset + */ 1);

		// Check to see if there is an override in the vendor specific kext.
		sampleLatencyDictionary = OSDynamicCast(OSDictionary, mStreamInterface->getProperty(kIOAudioEngineInputSampleLatencyKey));

		if (NULL != sampleLatencyDictionary)
		{
			char	sampleLatencyKeyString[16];

			::snprintf(sampleLatencyKeyString, 16, "%u", mCurSampleRate.whole);

			OSNumber	*sampleLatency = OSDynamicCast(OSNumber, sampleLatencyDictionary->getObject(sampleLatencyKeyString));

			if (NULL != sampleLatency)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - override input sample latency (%u) to %u sample frames",
					this, newSampleLatency, sampleLatency->unsigned32BitValue());

				newSampleLatency = sampleLatency->unsigned32BitValue();
			}
		}

		// setSampleLatency chosen via heuristics
		mUSBAudioEngine->setInputSampleLatency(newSampleLatency);

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - setting input sample latency to %u sample frames",
			this, newSampleLatency);
	}
	else	// Output case.
	{
		cautiousSafeSampleOffset = averageFrameSamples + 1;

		// <rdar://6343818> Adjust the safety offset by + 0.5ms to compensate for the time stamps generation
		// in the input stream when both the input & output streams are on the same engine.
		if (mSyncCompensation)
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - Compensating for time stamp generation in sample offset", this);

			minimumSafeSampleOffset = cautiousSafeSampleOffset;
		}
		else
		{
			minimumSafeSampleOffset = cautiousSafeSampleOffset / 2;
		}

		newSampleOffset = minimumSafeSampleOffset;

		// Check to see if there is an override in the vendor specific kext.
		sampleOffsetDictionary = OSDynamicCast(OSDictionary, mStreamInterface->getProperty(kIOAudioEngineSampleOffsetKey));

		if (NULL != sampleOffsetDictionary)
		{
			char	sampleOffsetKeyString[16];

			::snprintf(sampleOffsetKeyString, 16, "%u", mCurSampleRate.whole);

			OSNumber	*sampleOffset = OSDynamicCast(OSNumber, sampleOffsetDictionary->getObject(sampleOffsetKeyString));

			if (NULL != sampleOffset)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - override output sample offset (%u) to %u sample frames",
					this, newSampleOffset, sampleOffset->unsigned32BitValue());

				newSampleOffset = sampleOffset->unsigned32BitValue();
			}
		}

		// Set the offset for output devices (speakers, etc.) to 1 USB frame (+1 ms to latency).
		// This is necessary to ensure that samples are not clipped to a portion of the buffer whose DMA is in process.
		mUSBAudioEngine->setOutputSampleOffset(newSampleOffset);

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - setting output sample offset to %u sample frames",
			this, newSampleOffset);

		newSampleLatency = additionalSampleFrameFreq ? (averageFrameSamples + 1) : averageFrameSamples;

		// Check to see if there is an override in the vendor specific kext.
		sampleLatencyDictionary = OSDynamicCast(OSDictionary, mStreamInterface->getProperty(kIOAudioEngineOutputSampleLatencyKey));

		if (NULL != sampleLatencyDictionary)
		{
			char	sampleLatencyKeyString[16];

			::snprintf(sampleLatencyKeyString, 16, "%u", mCurSampleRate.whole);

			OSNumber	*sampleLatency = OSDynamicCast(OSNumber, sampleLatencyDictionary->getObject(sampleLatencyKeyString));

			if (NULL != sampleLatency)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - override output sample latency (%u) to %u sample frames",
					this, newSampleLatency, sampleLatency->unsigned32BitValue());

				newSampleLatency = sampleLatency->unsigned32BitValue();
			}
		}

		// setSampleLatency chosen via heuristics
		mUSBAudioEngine->setOutputSampleLatency(newSampleLatency);

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::updateSampleOffsetAndLatency() - setting output sample latency to %u sample frames",
			this, newSampleLatency);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::updateSampleOffsetAndLatency()");
}

/****************************************************************************/
/**
 * @brief	使用可能なオーディオフォーマットを調べる．
 *
 * @param[in]		*configDictionary	USB Audio Class のディスクリプタ．
 * @param[in]		sampleRate			サンプリング周波数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::addAvailableFormats(
	PdjUsbConfigurationDictionary		*configDictionary,
	IOAudioSampleRate					sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::addAvailableFormats(%p)", this, configDictionary);

	IOReturn						result = kIOReturnError;
	IOAudioStreamFormat				streamFormat;
	IOAudioStreamFormatExtension	streamFormatExtension;
	IOAudioSampleRate				lowSampleRate;
	IOAudioSampleRate				highSampleRate;
	OSArray							*sampleRates = NULL;
	OSObject						*arrayObject = NULL;
	OSNumber						*arrayNumber = NULL;
	UInt32							thisSampleRate;
	UInt32							otherSampleRate;
	UInt16							format;
	UInt8							numAltInterfaces;
	UInt8							numSampleRates;
	UInt8							altSettingIndex;
	UInt8							numChannels;
	UInt8							candidateAC3AltSetting = 0;
	bool							hasNativeAC3Format = false;
	bool							hasDigitalOutput = false;
	bool							isClockSourceProgrammable = true;	// <rdar://5811247>

	FailIf(NULL == mUSBAudioDevice, Exit);							// <rdar://7085810>
	FailIf(NULL == mUSBAudioEngine, Exit);							// <rdar://7085810>
	FailIf(NULL == mUSBAudioDevice->mControlInterface, Exit);		// <rdar://7085810>
	FailIf(NULL == configDictionary, Exit);

	FailIf(kIOReturnSuccess != configDictionary->getNumAltSettings(&numAltInterfaces, mInterfaceNumber), Exit);

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::addAvailableFormats() - There are %d alternate interfaces @ interface %d",
		this, numAltInterfaces, mInterfaceNumber);

	// <rdar://5811247>
	if (
			(IP_VERSION_02_00 == mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	   &&
			(0 != mUSBAudioEngine->mCurrentClockSourceID)
	   )
	{
		isClockSourceProgrammable = configDictionary->clockSourceHasFrequencyControl(
															mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
															0,
															mUSBAudioEngine->mCurrentClockSourceID,
															true
													  );
	}

	// Find all of the available formats on the device.
	altSettingIndex = configDictionary->alternateSettingZeroCanStream(mInterfaceNumber) ? 0 : 1;

	for ( ; altSettingIndex < numAltInterfaces; altSettingIndex++)
	{
		// <rdar://5811247>If the clock source is present, use that to find out if the clock source is programmable.
		// If it is programmable, then do what it is done now. If it is not programmable, then only 1 sample rate is supported, which
		// is the sample rate of the clock source.
		if (isClockSourceProgrammable)
		{
			// [rdar://5067229]
			if (kIOReturnSuccess != configDictionary->getNumSampleRates(&numSampleRates, mInterfaceNumber, altSettingIndex))
			{
				continue;
			}

			sampleRates = configDictionary->getSampleRates(mInterfaceNumber, altSettingIndex);
		}
		else
		{
			numSampleRates = 0;

			sampleRates = NULL;
		}

		// [rdar://5284099] Check the format before deciding whether to retrieve the following values.
		FailIf(kIOReturnSuccess != configDictionary->getFormat(&format, mInterfaceNumber, altSettingIndex), Exit);

		streamFormat.fBitWidth = 0;

		if ((PCM == format) || (IEC1937_AC3 == format))
		{
			FailIf(
						kIOReturnSuccess != configDictionary->getNumChannels(
																	&numChannels,
																	mInterfaceNumber,
																	altSettingIndex
															  ),
						Exit
				  );

			FailIf(
						kIOReturnSuccess != configDictionary->getBitResolution(
																	&(streamFormat.fBitDepth),
																	mInterfaceNumber,
																	altSettingIndex
															  ),
						Exit
				  );

			FailIf(
						kIOReturnSuccess != configDictionary->getSubframeSize(
																	&(streamFormat.fBitWidth),
																	mInterfaceNumber,
																	altSettingIndex
															  ),
						Exit
				  );
		}
		else
		{
			numChannels = 0;
		}

		streamFormat.fNumChannels	 = numChannels;
		streamFormat.fBitWidth		*= 8;
		streamFormat.fAlignment		 = kIOAudioStreamAlignmentLowByte;
		streamFormat.fByteOrder		 = kIOAudioStreamByteOrderLittleEndian;
		streamFormat.fDriverTag		 = (mInterfaceNumber << 16) | altSettingIndex;

		streamFormatExtension.fVersion			= kFormatExtensionCurrentVersion;
		streamFormatExtension.fFlags			= 0;
		streamFormatExtension.fFramesPerPacket	= 1;
		streamFormatExtension.fBytesPerPacket	= numChannels * (streamFormat.fBitWidth / 8);

		switch (format)
		{
			case PCM:
				streamFormat.fSampleFormat			= kIOAudioStreamSampleFormatLinearPCM;
				streamFormat.fNumericRepresentation	= kIOAudioStreamNumericRepresentationSignedInt;
				streamFormat.fIsMixable				= true;

				if (
						( 2 == streamFormat.fNumChannels)
				   &&
						(16 == streamFormat.fBitDepth)
				   &&
						(16 == streamFormat.fBitWidth)
				   )
				{
					candidateAC3AltSetting = altSettingIndex;
				}

				break;

			case AC3:	// just starting to stub something in for AC-3 support
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::addAvailableFormats() - Variable bit rate AC-3 audio format type", this);
				continue;	// We're not supporting this at the moment, so just skip it.
#if 0
				streamFormat.fSampleFormat			= kIOAudioStreamSampleFormatAC3;
				streamFormat.fIsMixable				= false;
				streamFormat.fNumChannels			= 6;
				streamFormat.fNumericRepresentation	= kIOAudioStreamNumericRepresentationSignedInt;
				streamFormat.fBitDepth				= 16;
				streamFormat.fBitWidth				= 16;
				streamFormat.fByteOrder				= kIOAudioStreamByteOrderBigEndian;

				FailIf(
							kIOReturnSuccess != configDictionary->getAC3BSID(
																		&(streamFormatExtension.fFlags),
																		mInterfaceNumber,
																		altSettingIndex
																  ),
							Exit
					  );

//				streamFormatExtension.fFlags			= USBToHostLong(configDictionary->GetAC3BSID(mInterfaceNumber, altSettingIndex));
//				streamFormatExtension.fFramesPerPacket	= configDictionary->GetSamplesPerFrame(mInterfaceNumber, altSettingIndex);
				streamFormatExtension.fFramesPerPacket	= 1536;
//				streamFormatExtension.fBytesPerPacket	= ((configDictionary->GetMaxBitRate(mInterfaceNumber, altSettingIndex) * 1024 / 8) + 500) / 1000;
				streamFormatExtension.fBytesPerPacket	= streamFormatExtension.fFramesPerPacket * streamFormat.fNumChannels * (streamFormat.fBitWidth / 8);

				break;
#endif

			case IEC1937_AC3:
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::addAvailableFormats() - IEC1937 AC-3 audio format type", this);

				hasNativeAC3Format = true;

				streamFormat.fSampleFormat			= kIOAudioStreamSampleFormat1937AC3;
				streamFormat.fNumericRepresentation	= kIOAudioStreamNumericRepresentationSignedInt;
				streamFormat.fIsMixable				= false;

				streamFormatExtension.fFramesPerPacket	= 1536;
				streamFormatExtension.fBytesPerPacket	= streamFormatExtension.fFramesPerPacket * streamFormat.fNumChannels * (streamFormat.fBitWidth / 8);

				break;

			default:
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::addAvailableFormats() - Interface format = 0x%x not published.", this, format);
				continue;	// skip this alternate interface
		}

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::addAvailableFormats() - Interface %d, Alt %d has a", this, mInterfaceNumber, altSettingIndex);
		DP(DBG_INFO, "     %d bit interface, ", streamFormat.fBitDepth);
		DP(DBG_INFO, "     %u channel(s), and ", streamFormat.fNumChannels);
		DP(DBG_INFO, "     %d sample rate(s), which is/are:", numSampleRates);

		if (numSampleRates && sampleRates)
		{
			for (UInt8 rateIndex = 0; rateIndex < numSampleRates; rateIndex++)
			{
				FailIf(NULL == (arrayObject = sampleRates->getObject(rateIndex)), Exit);
				FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

				thisSampleRate = arrayNumber->unsigned32BitValue();

				DP(DBG_INFO, "          %u", thisSampleRate);

				lowSampleRate.whole		= thisSampleRate;
				lowSampleRate.fraction	= 0;

				super::addAvailableFormat(&streamFormat, &streamFormatExtension, &lowSampleRate, &lowSampleRate);

				if (kIOAudioStreamSampleFormatLinearPCM == streamFormat.fSampleFormat)
				{
					streamFormat.fIsMixable = false;

					super::addAvailableFormat(&streamFormat, &streamFormatExtension, &lowSampleRate, &lowSampleRate);

					streamFormat.fIsMixable = true;		// set it back to true for next time through the loop
				}
			}

			DP(DBG_INFO, "");
		}
		else if (sampleRates)
		{
			FailIf(NULL == (arrayObject = sampleRates->getObject(0)), Exit);
			FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

			thisSampleRate = arrayNumber->unsigned32BitValue();

			FailIf(NULL == (arrayObject = sampleRates->getObject(1)), Exit);
			FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

			otherSampleRate = arrayNumber->unsigned32BitValue();

			DP(DBG_INFO, "          %u to %u", thisSampleRate, otherSampleRate);

			lowSampleRate.whole		= thisSampleRate;
			lowSampleRate.fraction	= 0;

			highSampleRate.whole	= otherSampleRate;
			highSampleRate.fraction	= 0;

			super::addAvailableFormat(&streamFormat, &streamFormatExtension, &lowSampleRate, &highSampleRate);

			if (kIOAudioStreamSampleFormatLinearPCM == streamFormat.fSampleFormat)
			{
				streamFormat.fIsMixable = false;

				super::addAvailableFormat(&streamFormat, &streamFormatExtension, &lowSampleRate, &highSampleRate);
			}
		}
		else
		{
			// <rdar://5811247>
			if (!isClockSourceProgrammable)
			{
				OSArray		*clockPath = NULL;
				OSArray		*clockPathGroup = mUSBAudioDevice->mTopology->getClockPathGroup(mInterfaceNumber, altSettingIndex);

				FailIf(NULL == clockPathGroup, Exit);

				FailIf(
							NULL == (
										clockPath = OSDynamicCast(
														OSArray,
														clockPathGroup->getObject(
																			mUSBAudioEngine->mCurrentClockPathIndex - 1
																		)
													)
									),
							Exit
					  );

				// <rdar://6945472>
				if (
						kIOReturnSuccess == mUSBAudioDevice->mTopology->getClockPathCurSampleRate(
																			&thisSampleRate,
																			NULL,
																			NULL,
																			clockPath
																		)
				   )
				{
					DP(DBG_INFO, "          %u", thisSampleRate);

					lowSampleRate.whole		= thisSampleRate;
					lowSampleRate.fraction	= 0;

					super::addAvailableFormat(&streamFormat, &streamFormatExtension, &lowSampleRate, &lowSampleRate);
				}
			}
		}
	} // for altSettingIndex

	/*
	configDictionary->getOutputTerminalType(
							&terminalType,
							mUSBAudioDevice->mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
							0,
							altSettingIndex
					  );

	switch (terminalType)
	{
		case EXTERNAL_DIGITAL_AUDIO_INTERFACE:
		case EXTERNAL_SPDIF_INTERFACE:
		case EMBEDDED_DVD_AUDIO:
			hasDigitalOutput = true;
			break;

		default:
			hasDigitalOutput = false;
			break;
	}
	*/

	if (
			(hasDigitalOutput)
	   &&
			(!hasNativeAC3Format)
	   &&
			(0 != candidateAC3AltSetting)
	   &&
			(kIOAudioStreamDirectionOutput == super::getDirection())
	   )
	{
		FailIf(
					kIOReturnSuccess != configDictionary->getNumSampleRates(
																&numSampleRates,
																mInterfaceNumber,
																candidateAC3AltSetting
														  ),
					Exit
			  );

		sampleRates = configDictionary->getSampleRates(mInterfaceNumber, candidateAC3AltSetting);

		FailIf(
					kIOReturnSuccess != configDictionary->getNumChannels(
																&numChannels,
																mInterfaceNumber,
																candidateAC3AltSetting
														  ),
					Exit
			  );

		streamFormat.fNumChannels = numChannels;

		FailIf(
					kIOReturnSuccess != configDictionary->getBitResolution(
																&(streamFormat.fBitDepth),
																mInterfaceNumber,
																candidateAC3AltSetting
														  ),
					Exit
			  );

		FailIf(
					kIOReturnSuccess != configDictionary->getSubframeSize(
																&(streamFormat.fBitWidth),
																mInterfaceNumber,
																candidateAC3AltSetting
														  ),
					Exit
			  );

		streamFormat.fBitWidth				*= 8;
		streamFormat.fAlignment				 = kIOAudioStreamAlignmentLowByte;
		streamFormat.fByteOrder				 = kIOAudioStreamByteOrderLittleEndian;
		streamFormat.fDriverTag				 = (mInterfaceNumber << 16) | candidateAC3AltSetting;
		streamFormat.fSampleFormat			 = kIOAudioStreamSampleFormat1937AC3;
		streamFormat.fNumericRepresentation	 = kIOAudioStreamNumericRepresentationSignedInt;
		streamFormat.fIsMixable				 = false;

		streamFormatExtension.fVersion			= kFormatExtensionCurrentVersion;
		streamFormatExtension.fFlags			= 0;
		streamFormatExtension.fFramesPerPacket	= 1536;

		streamFormatExtension.fBytesPerPacket	= (
														streamFormatExtension.fFramesPerPacket
												  *
														streamFormat.fNumChannels
												  *
														(streamFormat.fBitWidth / 8)
												  );

		if (numSampleRates && sampleRates)
		{
			for (UInt8 rateIndex = 0; rateIndex < numSampleRates; rateIndex++)
			{
				FailIf(NULL == (arrayObject = sampleRates->getObject(rateIndex)), Exit);
				FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

				thisSampleRate = arrayNumber->unsigned32BitValue();

				lowSampleRate.whole		= thisSampleRate;
				lowSampleRate.fraction	= 0;

				super::addAvailableFormat(&streamFormat, &streamFormatExtension, &lowSampleRate, &lowSampleRate);
			}
		}
		else if (sampleRates)
		{
			FailIf(NULL == (arrayObject = sampleRates->getObject(0)), Exit);
			FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

			thisSampleRate = arrayNumber->unsigned32BitValue();

			FailIf(NULL == (arrayObject = sampleRates->getObject(1)), Exit);
			FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

			otherSampleRate = arrayNumber->unsigned32BitValue();

			lowSampleRate.whole		= thisSampleRate;
			lowSampleRate.fraction	= 0;

			highSampleRate.whole	= otherSampleRate;
			highSampleRate.fraction	= 0;

			super::addAvailableFormat(&streamFormat, &streamFormatExtension, &lowSampleRate, &highSampleRate);
		}
	} // if hasDigitalOutput ...

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::addAvailableFormats(%p) = 0x%x", this, configDictionary, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオフォーマットを設定．
 *			IOAudioStream::setFormat() をオーバーライド．
 *
 * @param[in]		*streamFormat		設定するフォーマット．
 * @param[in]		callDriver			true なら PdjUsbAudioEngine::performFormatChange() が呼び出される．
 *
 * @return			IOReturn コード．
 */
// <rdar://7259238>
IOReturn
PdjUsbAudioStream::setFormat(
	const IOAudioStreamFormat	*streamFormat,
	bool						callDriver
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::setFormat(%p, %d)", this, streamFormat, callDriver);

	IOReturn	result = super::setFormat(streamFormat, callDriver);

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::setFormat(%p, %d) = 0x%x", this, streamFormat, callDriver, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオフォーマットを設定．
 *			IOAudioStream::setFormat() をオーバーライド．
 *
 * @param[in]		*streamFormat		設定するフォーマット．
 * @param[in]		*formatExtension	追加のフォーマット情報．
 * @param[in]		*formatDict			フォーマット情報を OSDictionary 形式に変換したもの．
 * @param[in]		callDriver			true なら PdjUsbAudioEngine::performFormatChange() が呼び出される．
 *
 * @return			IOReturn コード．
 */
// <rdar://7259238>
IOReturn
PdjUsbAudioStream::setFormat(
	const IOAudioStreamFormat			*streamFormat,
	const IOAudioStreamFormatExtension	*formatExtension,
	OSDictionary						*formatDict,
	bool								callDriver
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::setFormat(%p, %p, %p, %d)",
		this, streamFormat, formatExtension, formatDict, callDriver);

	IOReturn							result = kIOReturnError;
	bool								streamIsRunning = false;

	FailIf(0 == mUSBAudioEngine, Exit);

	streamIsRunning = mUSBStreamRunning;

	if (streamIsRunning)
	{
		mUSBAudioEngine->pauseAudioEngine();
	}

	mUSBAudioEngine->beginConfigurationChange();

	result = super::setFormat(streamFormat, formatExtension, formatDict, callDriver);

	if (kIOReturnSuccess == result)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::setFormat(%p, %p, %p, %d) - Delaying %u ms...",
			this, streamFormat, formatExtension, formatDict, callDriver, kFormatChangeDelayInMs);

		// Wait a bit after format change so that the USB audio device has a chance to catch up.
		IOSleep(kFormatChangeDelayInMs);
	}

	// Send an engine change notification so that the HAL refreshes its settings.
	mUSBAudioEngine->completeConfigurationChange();

	if (streamIsRunning)
	{
		mUSBAudioEngine->resumeAudioEngine();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::setFormat(%p, %p, %p, %d) = 0x%x",
		this, streamFormat, formatExtension, formatDict, callDriver, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオフォーマット変更．
 *			PdjUsbAudioEngine::controlledFormatChange() から呼び出される．
 *
 * @param[in]		*newFormat		変更後のフォーマット．
 * @param[in]		*newSampleRate	変更後のサンプリング周波数．
 *
 * @return			IOReturn コード．
 */
// [rdar://3918719] The following method now does the work of performFormatChange
// after being regulated by PdjUsbAudioDevice::formatChangeController().
IOReturn
PdjUsbAudioStream::controlledFormatChange(
	const IOAudioStreamFormat	*newFormat,
	const IOAudioSampleRate		*newSampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange(%p, %p)", this, newFormat, newSampleRate);

	IOReturn							result = kIOReturnError;
	PdjUsbConfigurationDictionary		*configDictionary = NULL;
	void								*sampleBuffer;
	UInt32								numSamplesInBuffer;
	UInt16								averageFrameSamples;
	UInt16								additionalSampleFrameFreq;
	UInt16								maxPacketSize;
	UInt8								endpointAddress;
	UInt8								newAlternateSettingID;
	UInt8								newDirection;
	UInt8								oldTransactionsPerFrame;
	UInt32								remainder;								// <rdar://problem/6954295>
	IOAudioSampleRate					sampleRate;								// <rdar://6945472>
	bool								needToUpdateStampDifference = false;	// <rdar://problem/7378275>

	FailIf(NULL == mStreamInterface, Exit);
	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);
	FailIf(NULL == newFormat, Exit);

	// Can't rely on the driver tag to be correct because IOAudioFamily only looks for formats without respect to sample rate,
	// but it's an optimization in the general case.
	mInterfaceNumber = (UInt8)(newFormat->fDriverTag >> 16);

	newAlternateSettingID = (UInt8)(newFormat->fDriverTag);

	if (newFormat->fNumChannels != super::format.fNumChannels)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - Need to adjust channel controls (cur = %u, new = %u)",
			this, super::format.fNumChannels, newFormat->fNumChannels);

		if (kIOAudioStreamDirectionOutput == mDirection)
		{
			// check for mono mode
			if (1 == newFormat->fNumChannels)
			{
				mUSBAudioDevice->mTopology->setMonoState(true);
			}
			else
			{
				mUSBAudioDevice->mTopology->setMonoState(false);
			}
		}
	}
	else
	{
		// DP(DBG_INFO, "No need to adjust channel controls.");
	}

	if (newSampleRate)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - Changing sampling rate to %u",
			this, newSampleRate->whole);

		sampleRate = *newSampleRate;	// <rdar://6945472>

		needToUpdateStampDifference = true;		// <rdar://problem/7378275>
	}
	else
	{
		// DP(DBG_INFO, "Keeping existing sampling rate of %d", mCurSampleRate.whole);
		sampleRate = mCurSampleRate;	// <rdar://6945472>
	}

	if (!configDictionary->verifySampleRateIsSupported(mInterfaceNumber, newAlternateSettingID, sampleRate.whole))		// <rdar://6945472>
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - %u channel %d bit @ %u Hz is not supported by alternate setting %u.",
			this, newFormat->fNumChannels, newFormat->fBitDepth, sampleRate.whole, newAlternateSettingID);

		FailIf(
					kIOReturnSuccess != configDictionary->getAltSettingWithSettings(
																&newAlternateSettingID,
																mInterfaceNumber,
																newFormat->fNumChannels,
																newFormat->fBitDepth,
																sampleRate.whole
														  ),
					Exit
			  );
	}

	FailIf(kIOReturnSuccess != configDictionary->getIsocEndpointDirection(&newDirection, mInterfaceNumber, newAlternateSettingID), Exit);
	FailIf(newDirection != mDirection, Exit);

	// Set the sampling rate on the device [rdar://4867843], <rdar://6945472>
	if (IP_VERSION_02_00 == mStreamInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	{
		// <rdar://5811247>
		OSArray		*pathArray = NULL;

		if ((NULL != mUSBAudioEngine->mClockSelectorControl) && (0 != mUSBAudioEngine->mCurrentClockPathIndex))
		{
			// The clock source to use is dependent on what the clock selector is set to. Ask the engine what
			// is the current clock selector is pointed to, and use that to set the sample rate.
			UInt32		pathIndex = mUSBAudioEngine->mCurrentClockPathIndex;
			FailIf(0 == pathIndex, Exit);

			OSArray		*clockPathGroup = mUSBAudioDevice->mTopology->getClockPathGroup(mInterfaceNumber, newAlternateSettingID);
			FailIf(NULL == clockPathGroup, Exit);

			pathArray = OSDynamicCast(OSArray, clockPathGroup->getObject(pathIndex - 1));
		}
		else
		{
			pathArray = mUSBAudioDevice->mTopology->getOptimalClockPath(
														mUSBAudioEngine,
														mInterfaceNumber,
														newAlternateSettingID,
														sampleRate.whole,
														NULL
													);
		}

		FailIf(NULL == pathArray, Exit);

		FailIf(kIOReturnSuccess != mUSBAudioDevice->mTopology->setClockPathCurSampleRate(sampleRate.whole, pathArray, true), Exit);

		mActiveClockPath = pathArray;
	}

	mCurSampleRate = sampleRate;

	if (mPlugin)
	{
		mPlugin->pluginSetFormat(newFormat, &mCurSampleRate);
	}

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - about to set: mInterfaceNumber = %d & newAlternateSettingID = %d",
		this, mInterfaceNumber, newAlternateSettingID);

	mAlternateSettingID = newAlternateSettingID;

	oldTransactionsPerFrame = mTransactionsPerUSBFrame;

	// [rdar://4801012] Must determine the number of transfer opportunities per millisecond.
	if (
			(IP_VERSION_02_00 == mStreamInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	   &&
			(kUSBDeviceSpeedHigh == mUSBAudioDevice->getDeviceSpeed())
	   )
	{
		UInt8		interval;

		FailIf(
					kIOReturnSuccess != configDictionary->getIsocEndpointInterval(
																&interval,
																mInterfaceNumber,
																mAlternateSettingID,
																mDirection
														  ),
					Exit
			  );

		if (0 == interval)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - ERROR! Isoc endpoint has a refresh interval of 0! Treating as 4 ...", this);

			mTransactionsPerUSBFrame = 1;
		}
		else
		{
			FailIf(interval > 4, Exit);

			mTransactionsPerUSBFrame = 8 >> (interval - 1);
		}
	}
	else
	{
#ifdef TARGET_PDJ
 		mTransactionsPerUSBFrame = 8;
#else
		mTransactionsPerUSBFrame = 1;
#endif
	}

	// [rdar://4801012] Now determine the number of transactions per list.
	mNumTransactionsPerList = mNumUSBFramesPerList * mTransactionsPerUSBFrame;

	// [rdar://4801012] Allocate the isoc frames if necessary.
	if (
			(NULL != mUSBIsocFrames)
	   &&
			(oldTransactionsPerFrame != mTransactionsPerUSBFrame)
	   )
	{
		// Dispose of the current isoc frames.
		IOFree(mUSBIsocFrames, mNumUSBFrameLists * mNumUSBFramesPerList * oldTransactionsPerFrame * sizeof(IOUSBHostIsochronousFrame));
		mUSBIsocFrames = NULL;
	}

	if (NULL == mUSBIsocFrames)
	{
		mUSBIsocFrames = (IOUSBHostIsochronousFrame *)IOMalloc(mNumUSBFrameLists * mNumTransactionsPerList * sizeof(IOUSBHostIsochronousFrame));
		FailIf(NULL == mUSBIsocFrames, Exit);	// <rdar://7568547>

		// <rdar://7568547> Initialize the USB Isoc frames so that coalesceInputSamples() will not panic
		// due to uninitialized values in frStatus & frActCount.
		this->initializeUSBFrameList(mUSBIsocFrames, mNumUSBFrameLists * mNumTransactionsPerList);
	}

	// Set the sampling rate on the endpoint
	if (configDictionary->asEndpointHasSampleFreqControl(mInterfaceNumber, mAlternateSettingID))
	{
		FailIf(
					kIOReturnSuccess != configDictionary->getIsocEndpointAddress(
																&endpointAddress,
																mInterfaceNumber,
																mAlternateSettingID,
																mDirection
														  ),
					Exit
			  );

		(void)this->setSampleRateControl(endpointAddress, mCurSampleRate.whole);
		// no need to check the error, it's not a real problem if it doesn't work
	}

	// Set this as the default until we are told otherwise <rdar://problem/6954295>
	// Take the current sample rate (in Hz) and and transform it into samples per packet represented in 16.16 fixed point.
	// When calculating the fractional part, store the fraction x 1000 to maintain precision.
	mSamplesPerPacket.whole = mCurSampleRate.whole / (mTransactionsPerUSBFrame * 1000);

	remainder = mCurSampleRate.whole - (mSamplesPerPacket.whole * mTransactionsPerUSBFrame * 1000);
	// same as (mCurSampleRate.whole % 1000) * mTransactionsPerUSBFrame

	mSamplesPerPacket.fraction = (remainder * 65536) / mTransactionsPerUSBFrame;

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - mSamplesPerPacket: %u(whole) %u(fraction)",
		this, mSamplesPerPacket.whole, mSamplesPerPacket.fraction);

	this->calculateSamplesPerPacket(mCurSampleRate.whole, &averageFrameSamples, &additionalSampleFrameFreq);

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - averageFrameSamples = %d", this, averageFrameSamples);

	mSampleSize			= newFormat->fNumChannels * (newFormat->fBitWidth / 8);
	mAverageFrameSize	= averageFrameSamples * mSampleSize;
	mAlternateFrameSize	= (averageFrameSamples + 1) * mSampleSize;

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - mAverageFrameSize = %d, mAlternateFrameSize = %d",
		this, mAverageFrameSize, mAlternateFrameSize);

	mOverrunsThreshold = kOverrunsThreshold;	// <rdar://6411577> Threshold for overruns.

	// You have to make the read buffer the size of the alternate frame size because we have to ask for mAlternateFrameSize bytes
	// with each read. If you don't make the buffer big enough, you don't get all the data from the last frame...
	// USB says that if the device is running at an even multiple of the bus clock (i.e. 48KHz) that it can send frames that
	// have +/- one sample (i.e. 47, 48, or 49 samples per frame) from the average. This differs from when it's not a even
	// multiple and it can send only +1.5 sample from the average.
	if (kIOAudioStreamDirectionInput == mDirection)
	{
		// mReadUSBFrameListSize = mAlternateFrameSize * mNumTransactionsPerList;
		// [rdar://5355808] [rdar://5889101] Be a little more lenient than the spec dictates to accommodate ill-behaved devices if possible.
		FailIf(
					kIOReturnSuccess != configDictionary->getIsocEndpointMaxPacketSize(
																&maxPacketSize,
																mInterfaceNumber,
																mAlternateSettingID,
																mDirection
														  ),
					Exit
			  );

		mReadUSBFrameListSize = (
									((mAlternateFrameSize + (2 * mSampleSize)) < maxPacketSize)
										? (mAlternateFrameSize + (2 * mSampleSize))
										: maxPacketSize
								);

		mReadUSBFrameListSize *= mNumTransactionsPerList;
	}

	// Need a minimum of two pages in OHCI/UHCI
	numSamplesInBuffer  = mCurSampleRate.whole / 4;		// <rdar://problem/6954295>
	numSamplesInBuffer += (PAGE_SIZE * 2) - 1;
	numSamplesInBuffer /=  PAGE_SIZE * 2;
	numSamplesInBuffer *=  PAGE_SIZE * 2;

	mSampleBufferSize = numSamplesInBuffer * mSampleSize;

#ifdef TARGET_PDJ
 	mNumSamplesInBuffer = (UInt64)numSamplesInBuffer * 1000000000000ull;
#else
	if ((mUHCISupport) && (kIOAudioStreamDirectionInput != mDirection))		// <rdar://6564854> For output sample buffer only.
	{
		mSampleBufferSizeExtended = mSampleBufferSize + PAGE_SIZE;
	}
	else
	{
		mSampleBufferSizeExtended = mSampleBufferSize;
	}
#endif

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - New mSampleBufferSize = %u numSamplesInBuffer = %u",
		this, mSampleBufferSize, numSamplesInBuffer);

	// <rdar://problem/7378275>
	if (needToUpdateStampDifference || mInitStampDifference)
	{
		// Prime the filter with the nominal sample rate.
		mFilterWritePointer			 = 0;
		mLastFilteredStampDifference = this->jitterFilter(((1000000000ull * numSamplesInBuffer) / mCurSampleRate.whole), 0);
		mNumTimeStamp				 = 1;
		mInitStampDifference		 = false;
	}

	FailIf((NULL == mSampleBufferDescriptors) && (mNumUSBFrameLists > 0), Exit);

	for (UInt32 i = 0; i < mNumUSBFrameLists; i++)
	{
		if (NULL != mSampleBufferDescriptors[i])
		{
			mSampleBufferDescriptors[i]->release();
			mSampleBufferDescriptors[i] = NULL;
		}
	}

	if (kIOAudioStreamDirectionInput == mDirection)
	{
		if (NULL != mUSBBufferDescriptor)
		{
			mUSBBufferDescriptor->release();
		}

		mUSBBufferDescriptor = this->allocateBufferDescriptor(kIODirectionIn, mNumUSBFrameLists * mReadUSBFrameListSize, PAGE_SIZE);
		FailIf(NULL == mUSBBufferDescriptor, Exit);

		mReadBuffer = mUSBBufferDescriptor->getBytesNoCopy();
		FailIf(NULL == mReadBuffer, Exit);

		for (UInt32 i = 0; i < mNumUSBFrameLists; i++)
		{
			mSampleBufferDescriptors[i] = OSTypeAlloc(IOSubMemoryDescriptor);
			FailIf(NULL == mSampleBufferDescriptors[i], Exit);

			mSampleBufferDescriptors[i]->initSubRange(
												mUSBBufferDescriptor,
												i * mReadUSBFrameListSize,
												mReadUSBFrameListSize,
												kIODirectionIn
										 );
		}

		if (NULL != mSampleBufferMemoryDescriptor)
		{
			super::setSampleBuffer(NULL, 0);

			mSampleBufferMemoryDescriptor->release();
		}

		mSampleBufferMemoryDescriptor = IOBufferMemoryDescriptor::withOptions(kIODirectionInOut, mSampleBufferSize, PAGE_SIZE);
		FailIf(NULL == mSampleBufferMemoryDescriptor, Exit);

		sampleBuffer = mSampleBufferMemoryDescriptor->getBytesNoCopy();
	}
	else
	{
		// This is the output case.
		if (NULL != mUSBBufferDescriptor)
		{
			super::setSampleBuffer(NULL, 0);

			mUSBBufferDescriptor->release();
		}

#ifdef TARGET_PDJ
#else
		if (mUHCISupport)
		{
			// Allocate an additional alternate frame size (mSampleBufferSizeExtended total)
			// as our scribble ahead for frames that would be wrapped in clipOutputSamples
			mUSBBufferDescriptor = this->allocateBufferDescriptor(kIODirectionOut, mSampleBufferSizeExtended, PAGE_SIZE);
		}
		else
#endif
		{
			mUSBBufferDescriptor = this->allocateBufferDescriptor(kIODirectionOut, mSampleBufferSize, PAGE_SIZE);
		}

		FailIf(NULL == mUSBBufferDescriptor, Exit);

		for (UInt32 i = 0; i < mNumUSBFrameLists; i++)
		{
			if (NULL != mSampleBufferDescriptors[i])
			{
				mSampleBufferDescriptors[i]->release();
			}

			mSampleBufferDescriptors[i] = OSTypeAlloc(IOSubMemoryDescriptor);
			FailIf(NULL == mSampleBufferDescriptors[i], Exit);

			mSampleBufferDescriptors[i]->initSubRange(mUSBBufferDescriptor, 0, mSampleBufferSize, kIODirectionOut);

#ifdef BUG_FIX
#else
			result = mSampleBufferDescriptors[i]->prepare();
			FailIf(kIOReturnSuccess != result, Exit);
#endif
		}

		sampleBuffer = mUSBBufferDescriptor->getBytesNoCopy();
		FailIf(NULL == sampleBuffer, Exit);
	}

	super::setSampleBuffer(sampleBuffer, mSampleBufferSize);

	this->updateSampleOffsetAndLatency();

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() - newFormat->fNumChannels = %u, newFormat->fBitWidth %d",
		this, newFormat->fNumChannels, newFormat->fBitWidth);

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::controlledFormatChange() = 0x%x", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	USB ストリーミングの準備．
 *			PdjUsbAudioEngine::performAudioEngineStart() から呼び出される．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::prepareUSBStream(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::prepareUSBStream()", this);

	IOReturn							resultCode = kIOReturnError;	// Assume the entire method will fail.
	UInt32								remainder;						// <rdar://problem/6954295>
	PdjUsbConfigurationDictionary		*configDictionary;
	UInt8								endpointAddress;
	IOUSBFindEndpointRequest			audioIsochEndpoint;
	UInt16								averageFrameSamples;
	UInt16								additionalSampleFrameFreq;
	UInt16								maxPacketSize = 0;

	// Start the IO audio engine
	// Enable interrupts for this stream
	// The interrupt should be triggered at the start of the sample buffer
	// The interrupt handler should increment the fCurrentLoopCount and fLastLoopTime fields

	mCurrentFrameList			= 0;
	mSafeErasePoint				= 0;
	mNextSafeErasePoint			= 0;
	mBufferOffset				= 0;
	mLastPreparedBufferOffset	= 0;	// Start playing from the start of the buffer
	mFractionalSamplesLeft		= 0;	// Reset our parital frame list info
	mOverrunsCount				= 0;
	mShouldStop					= 0;

	// Set this as the default until we are told otherwise <rdar://problem/6954295>
	// Take the current sample rate (in Hz) and and transform it into samples per packet represented in 16.16 fixed point.
	// When calculating the fractional part, store the fraction x 1000 to maintain precision.
	mSamplesPerPacket.whole = mCurSampleRate.whole / (mTransactionsPerUSBFrame * 1000);

	remainder = mCurSampleRate.whole - (mSamplesPerPacket.whole * mTransactionsPerUSBFrame * 1000);
	// same as (mCurSampleRate.whole % 1000) * mTransactionsPerUSBFrame

	mSamplesPerPacket.fraction = (remainder * 65536) / mTransactionsPerUSBFrame;

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::prepareUSBStream() - mSamplesPerPacket: %u(whole) %u(fraction)",
		this, mSamplesPerPacket.whole, mSamplesPerPacket.fraction);

	FailIf((mNumUSBFrameLists < mNumUSBFrameListsToQueue), Exit);
	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);
	FailIf(NULL == mStreamInterface, Exit);

#ifdef TARGET_PDJ
 	if (mInterfaceNumber == 1)
 	{
 		resultCode = mStreamInterface->selectAlternateSetting(mAlternateSettingID);
 
 		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::prepareUSBStream() - mStreamInterface->selectAlternateSetting(%d) = 0x%x",
 			this, mAlternateSettingID, resultCode);
 
 		FailIf(kIOReturnSuccess != resultCode, Exit);
 	}
#else
	resultCode = mStreamInterface->selectAlternateSetting(mAlternateSettingID);

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::prepareUSBStream() - mStreamInterface->selectAlternateSetting(%d) = 0x%x",
		this, mAlternateSettingID, resultCode);

	FailIf(kIOReturnSuccess != resultCode, Exit);
#endif

	if (configDictionary->asEndpointHasSampleFreqControl(mInterfaceNumber, mAlternateSettingID))
	{
		FailIf(
					kIOReturnSuccess != configDictionary->getIsocEndpointAddress(
																&endpointAddress,
																mInterfaceNumber,
																mAlternateSettingID,
																mDirection
														  ),
					Exit
			  );

		(void)this->setSampleRateControl(endpointAddress, mCurSampleRate.whole);
		// no need to check the error, it's not a real problem if it doesn't work
	}

	// Acquire a PIPE for the isochronous stream.
	audioIsochEndpoint.type		 = kUSBIsoc;
	audioIsochEndpoint.direction = mDirection;

	mPipe = PdjUsbAudioDevice::findPipe(mStreamInterface, &audioIsochEndpoint);
	FailIf(NULL == mPipe, Exit);

	mPipe->retain();

#ifdef TARGET_PDJ
#else
	if (super::getDirection() == kIOAudioStreamDirectionOutput)
	{
		// Not concerned with errors in this function at this time.
		(void)this->checkForFeedbackEndpoint(configDictionary);
	}
#endif

	this->calculateSamplesPerPacket(mCurSampleRate.whole, &averageFrameSamples, &additionalSampleFrameFreq);

	// [rdar://4664738] Check the maximum packet size of the isoc data endpoint.
	FailIf(
				kIOReturnSuccess != configDictionary->getIsocEndpointMaxPacketSize(
															&maxPacketSize,
															mInterfaceNumber,
															mAlternateSettingID,
															mDirection
													  ),
				Exit
		  );

	// [rdar://4664738] Disallow values of zero for the maximum packet size.
	FailIf(0 == maxPacketSize, Exit);

	// [rdar://2750290] Make sure we have enough bandwidth (and give back any that we won't be using).
	const IOAudioStreamFormat	*theFormat;
	UInt16						maxFrameSize;

	theFormat = &(super::format);

	if (kIOAudioStreamDirectionInput == mDirection)
	{
		// [rdar://5355808] [rdar://5889101]
		// Be a little more lenient with this calculation if possible to safeguard against ill-behaved devices.
		maxFrameSize = (averageFrameSamples + 3) * (theFormat->fNumChannels * (theFormat->fBitWidth / 8));

		// [rdar://4664738] Never call setPipePolicy in excess of the maximum packet size for the endpoint.
		maxFrameSize = (maxFrameSize > maxPacketSize) ? maxPacketSize : maxFrameSize;

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::prepareUSBStream() - maxPacketSize = %d, maxFrameSize = %d",
			this, maxPacketSize, maxFrameSize);

		mReadUSBFrameSize = maxFrameSize;
	}
	else
	{
		if ((0 == additionalSampleFrameFreq) && (NULL == mAssociatedPipe))		// [rdar://5032866]
		{
#ifdef TARGET_PDJ
 			maxFrameSize = (averageFrameSamples + 3) * (theFormat->fNumChannels * (theFormat->fBitWidth / 8));
#else
			maxFrameSize = (averageFrameSamples    ) * (theFormat->fNumChannels * (theFormat->fBitWidth / 8));
#endif
		}
		else
		{
			maxFrameSize = (averageFrameSamples + 1) * (theFormat->fNumChannels * (theFormat->fBitWidth / 8));
		}
	}

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::prepareUSBStream() - calling SetPipePolicy(%d)", this, maxFrameSize);

#ifdef TARGET_PDJ
 	resultCode = kIOReturnSuccess;
#else
	resultCode = mPipe->SetPipePolicy(maxFrameSize, 0);
	FailIf(kIOReturnSuccess != resultCode, Exit);
#endif

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::prepareUSBStream() = %x", this, resultCode);
	return resultCode;
}

/****************************************************************************/
/**
 * @brief	USB ストリーミングの開始．
 *			PdjUsbAudioEngine::performAudioEngineStart() から呼び出される．
 *
 * @param[in]		currentUSBFrame		現在の USB フレーム番号．
 * @param[in]		usbFramesToDelay	ストリーミング開始時に遅延させる USB フレーム数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::startUSBStream(
	UInt64			currentUSBFrame,
	UInt32			usbFramesToDelay
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::startUSBStream(%llu, %u)", this, currentUSBFrame, usbFramesToDelay);

	IOReturn							resultCode = kIOReturnError;
	IOReturn							interimResult;
	PdjUsbConfigurationDictionary		*configDictionary;
	UInt16								averageFrameSamples;
	UInt16								additionalSampleFrameFreq;
	UInt16								maxPacketSize = 0;
	bool								encounteredQueuingError = false;
	const IOAudioStreamFormat			*theFormat;
	UInt16								maxFrameSize;

	FailIf((mNumUSBFrameLists < mNumUSBFrameListsToQueue), Exit);
	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);
	FailIf(NULL == mStreamInterface, Exit);

#if DEBUGTIMESTAMPS
	mStampDrift					 = 0ll;		// <rdar://problem/7378275>
#endif
	mLastRawTimeStamp_nanos		 = 0ull;	// <rdar://problem/7378275>
	mLastFilteredTimeStamp_nanos = 0ull;	// <rdar://problem/7378275>
	mLastWrapFrame				 = 0ull;

	this->calculateSamplesPerPacket(mCurSampleRate.whole, &averageFrameSamples, &additionalSampleFrameFreq);

	theFormat = &(super::format);

	// [rdar://4664738] Check the maximum packet size of the isoc data endpoint.
	FailIf(
				kIOReturnSuccess != configDictionary->getIsocEndpointMaxPacketSize(
															&maxPacketSize,
															mInterfaceNumber,
															mAlternateSettingID,
															mDirection
													  ),
				Exit
		  );

	// [rdar://4664738] Disallow values of zero for the maximum packet size.
	FailIf(0 == maxPacketSize, Exit);

	// [rdar://2750290] Make sure we have enough bandwidth (and give back any that we won't be using).
	if (kIOAudioStreamDirectionInput == mDirection)
	{
		// [rdar://5355808] [rdar://5889101]
		// Be a little more lenient with this calculation if possible to safeguard against ill-behaved devices.
		maxFrameSize = (averageFrameSamples + 3) * (theFormat->fNumChannels * (theFormat->fBitWidth / 8));

		// [rdar://4664738] Never call setPipePolicy in excess of the maximum packet size for the endpoint.
		maxFrameSize = (maxFrameSize > maxPacketSize) ? maxPacketSize : maxFrameSize;

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() - maxPacketSize = %d, maxFrameSize = %d",
			this, maxPacketSize, maxFrameSize);

		mReadUSBFrameSize = maxFrameSize;
	}
	else
	{
#if PRIMEISOCINPUT
		if ((0 == additionalSampleFrameFreq) && (NULL == mAssociatedPipe))		// [rdar://5032866]
		{
#ifdef TARGET_PDJ
 			maxFrameSize = (averageFrameSamples + 3) * (theFormat->fNumChannels * (theFormat->fBitWidth / 8));
#else
			maxFrameSize = (averageFrameSamples    ) * (theFormat->fNumChannels * (theFormat->fBitWidth / 8));
#endif
		}
		else
		{
			maxFrameSize = (averageFrameSamples + 1) * (theFormat->fNumChannels * (theFormat->fBitWidth / 8));
		}
#endif
	}

	// The current frame is already in processing, and it may be nearly done.
	// Must queue a minimum of kMinimumFrameOffset USB frames in the future to ensure that our DMA occurs when we request it.
	mUSBFrameToQueue = currentUSBFrame + kMinimumFrameOffset;

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() - mUSBFrameToQueue = %llu", this, mUSBFrameToQueue);

	if (NULL != mAssociatedPipe)
	{
		mNextSyncReadFrame = mUSBFrameToQueue;
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() - Starting feedback endpoint stream at frame %llu",
			this, mNextSyncReadFrame);

		(void)mAssociatedPipe->io(
									mAssociatedEndpointMemoryDescriptor,
									&mSampleRateFrame,
									1,
									mNextSyncReadFrame,
									&mSampleRateCompletion
							   );

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() - Feedback endpoint stream started.", this);
	}

	// Note that we haven't taken our first time stamp yet. This will help us determine when we should take it.
	mHaveTakenFirstTimeStamp = false;

	if (super::getDirection() == kIOAudioStreamDirectionInput)
	{
#if PRIMEISOCINPUT
		UInt16		maxConsecutiveFramesToPrime;
		UInt16		transactionsToQueue;
		UInt16		remainingFrames;

		maxConsecutiveFramesToPrime = mSampleBufferSize / maxFrameSize;

		FailIf(0 == maxConsecutiveFramesToPrime, Exit);

		// Add one transaction for every buffer size in bytes and one for any partial remainder
		transactionsToQueue = (
									( usbFramesToDelay / maxConsecutiveFramesToPrime )
							  +
									((usbFramesToDelay % maxConsecutiveFramesToPrime) ? 1 : 0)
							  );

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() - Priming input stream at frame %llu (%u USB frames) in %u transaction(s)",
			this, mUSBFrameToQueue, usbFramesToDelay, transactionsToQueue);

		remainingFrames = usbFramesToDelay;

		// No longer does the following as it it compensated by kStartDelayOffset in PdjUsbAudioEngine::performAudioEngineStart().
		// [rdar://5148788] Update the frame to queue. Some of the above operations could cost as as much as a millisecond if we get time-sliced out.
		// currentUSBFrame = mStreamInterface->getDevice()->getFrameNumber();
		// mUSBFrameToQueue = currentUSBFrame + kMinimumFrameOffset;

		for (UInt16 currentTransaction = 0; currentTransaction < transactionsToQueue; currentTransaction++)
		{
			if (remainingFrames < maxConsecutiveFramesToPrime)
			{
				// This is the last transaction
				this->primeInputPipe(mPipe, maxFrameSize, remainingFrames);

				remainingFrames = 0;
			}
			else
			{
				// We will have more to queue after this
				this->primeInputPipe(mPipe, maxFrameSize, maxConsecutiveFramesToPrime);

				remainingFrames -= maxConsecutiveFramesToPrime;
			}
		}
#endif

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() - Starting input stream at frame %llu",
			this, mUSBFrameToQueue);

		for (UInt32 frameListNum = mCurrentFrameList; frameListNum < mNumUSBFrameListsToQueue; frameListNum++)
		{
			interimResult = this->readFrameList(frameListNum);

			if (kIOReturnSuccess != interimResult)
			{
				DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() - readFrameList (%u) failed with error 0x%x!",
					this, frameListNum, interimResult);

				encounteredQueuingError = true;
			}
		}
	}
	else
	{
#if PRIMEISOCINPUT
		mUSBFrameToQueue += usbFramesToDelay;
#endif

		// This cast can be removed when AbsoluteTime is UInt64.
		AbsoluteTime_to_scalar(&(mUSBIsocFrames[0].timeStamp)) = 0xFFFFFFFFFFFFFFFFull;

		mUSBIsocFrames[0].status = kUSBLowLatencyIsochTransferKey;

		for (UInt32 frameListNum = mCurrentFrameList; frameListNum < mNumUSBFrameListsToQueue; frameListNum++)
		{
			interimResult = this->writeFrameList(frameListNum);

			if (kIOReturnSuccess != interimResult)
			{
				DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() - writeFrameList (%u) failed with error 0x%x!",
					this, frameListNum, interimResult);

				encounteredQueuingError = true;
			}
		}
	}

	// Here we need to determine if the stream is started to our satisfaction before returning.
	// Currently, we expect to queue exacty mNumUSBFrameListsToQueue.

	if (encounteredQueuingError)
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() - Stream will *NOT* start because of queuing errors", this);

		resultCode = kIOReturnError;
	}
	else
	{
		mUSBStreamRunning = true;

		resultCode = kIOReturnSuccess;
	}

	if (kIOReturnSuccess == resultCode)
	{
#ifdef TARGET_PDJ
 		DP(DBG_WARNING, "  ---------- Starting Stream (interface %d, alternate setting %d) ----------",
 			mInterfaceNumber, mAlternateSettingID);
 		DP(DBG_WARNING, "      sampling rate = %u", mCurSampleRate.whole);
#else
		DP(DBG_INFO, "\n");
		DP(DBG_INFO, "  -------------------- Starting Stream (interface %d, alternate setting %d) --------------------",
			mInterfaceNumber, mAlternateSettingID);
		DP(DBG_INFO, "      format = %p",				&(super::format));
		DP(DBG_INFO, "          fNumChannels = %u",		super::format.fNumChannels);
		DP(DBG_INFO, "          fBitDepth    = %d",		super::format.fBitDepth);
		DP(DBG_INFO, "          fDriverTag   = 0x%x",	super::format.fDriverTag);
		DP(DBG_INFO, "\n");
#endif
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::startUSBStream() = %x", this, resultCode);
	return resultCode;
}

/****************************************************************************/
/**
 * @brief	USB ストリーミングの停止．
 *			PdjUsbAudioEngine::performAudioEngineStop() から呼び出される．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::stopUSBStream(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::stopUSBStream()", this);

	if (0 == mShouldStop)
	{
		mShouldStop = 1;
	}

	if (NULL != mPipe)
	{
		if (!mTerminatingDriver)
		{
			// <rdar://7251353> abort() needs to be called first to cancel all existing transfers before returning the
			// bandwidth via SetPipePolicy().
			// <rdar://6277511> Abort the pipe to cancel all pending transactions.
			mPipe->abort();

			// Don't call USB if we are being terminated because we could deadlock their workloop.
//			mPipe->SetPipePolicy(0, 0);
		}

		// Have to close the current pipe so we can open a new one because changing the alternate interface will tear down the current pipe
		mPipe->release();
		mPipe = NULL;
	}

	if (NULL != mAssociatedPipe)
	{
		if (!mTerminatingDriver)
		{
			// <rdar://6277511> Abort the pipe to cancel all pending transactions.
			mAssociatedPipe->abort();
		}

		mAssociatedPipe->release();
		mAssociatedPipe = NULL;
	}

	if (!mTerminatingDriver)
	{
		// Don't call USB if we are being terminated because we could deadlock their workloop.
		if (NULL != mStreamInterface)
		{
			// if we don't have an interface, message() got called and we are being terminated
 			(void)mStreamInterface->selectAlternateSetting(kRootAlternateSetting);
		}
	}

	mUSBStreamRunning = false;

#ifdef TARGET_PDJ
 	DP(DBG_WARNING, "  ---------- Streaming Stopped (interface %d, alternate setting %d) ----------",
 		mInterfaceNumber, mAlternateSettingID);
#else
	DP(DBG_INFO, "  -------------------- Streaming Stopped (interface %d, alternate setting %d) --------------------",
		mInterfaceNumber, mAlternateSettingID);
#endif

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::stopUSBStream()", this);
	return kIOReturnSuccess;
}

/****************************************************************************/
/**
 * @brief	Isochronous OUT 転送の予約数が不足の場合は補充する．
 */
void
PdjUsbAudioStream::queueOutputFrames(
	void
)
{
	DP(DBG_WRITE, "+ " LOGSTRAUDIOSTREAM "::queueOutputFrames()");

	UInt64			curUSBFrameNumber;
	UInt64			framesLeftInQueue;

	if ((0 == mShouldStop) && (!mInCompletion) && (NULL != mStreamInterface))
	{
		curUSBFrameNumber = mStreamInterface->getDevice()->getFrameNumber();

		framesLeftInQueue = mUSBFrameToQueue - curUSBFrameNumber;

		if (framesLeftInQueue < (mNumUSBFramesPerList * (mNumUSBFrameListsToQueue / 2)) / 2)	// <rdar://problem/6327095>
		{
			DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "::queueOutputFrames() - Queue a write from clipOutputSamples: framesLeftInQueue = %llu",
				framesLeftInQueue);

			writeHandler(
				this,
				mUSBCompletion[mCurrentFrameList].parameter,
				kIOReturnSuccess,
				&mUSBIsocFrames[mCurrentFrameList * mNumUSBFramesPerList]
			);
		}
	}

	DP(DBG_WRITE, "- " LOGSTRAUDIOSTREAM "::queueOutputFrames()");
}

/****************************************************************************/
/**
 * @brief	Isochronous IN 転送の予約数が不足の場合は補充する．
 */
void
PdjUsbAudioStream::queueInputFrames(
	void
)
{
	DP(DBG_READ, "+ " LOGSTRAUDIOSTREAM "::queueInputFrames()");

#if 1	// enabled for [3091812]
	UInt64			curUSBFrameNumber;
	UInt64			framesLeftInQueue;

	if ((0 == mShouldStop) && (!mInCompletion) && (NULL != mStreamInterface))
	{
		curUSBFrameNumber = mStreamInterface->getDevice()->getFrameNumber();

		framesLeftInQueue = mUSBFrameToQueue - curUSBFrameNumber;

		if (framesLeftInQueue < (mNumUSBFramesPerList * (mNumUSBFrameListsToQueue / 2)) / 2)	// <rdar://problem/6327095>
		{
			while ((framesLeftInQueue < mNumUSBFramesPerList * (mNumUSBFrameListsToQueue - 1)) && (0 == mShouldStop))
			{
				DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "::queueInputFrames() - Queue a read from convertInputSamples: framesLeftInQueue = %llu",
					framesLeftInQueue);

				readHandler(
					this,
					mUSBCompletion[mCurrentFrameList].parameter,
					kIOReturnSuccess,
					&mUSBIsocFrames[mCurrentFrameList * mNumUSBFramesPerList]
				);

				curUSBFrameNumber = mStreamInterface->getDevice()->getFrameNumber();

				framesLeftInQueue = mUSBFrameToQueue - curUSBFrameNumber;
			}
		}
	}
#endif

	DP(DBG_READ, "- " LOGSTRAUDIOSTREAM "::queueInputFrames()");
}

/****************************************************************************/
/**
 * @brief	デバイスからの受信データをオーディオ入力用のリングバッファに書き込む．
 *
 * @param[in]		numBytesToCoalesce	USB 受信完了ルーチンから呼び出された場合は、ゼロ．
 *										PdjUsbAudioEngine::convertInputSamples() から
 *										呼び出された場合は、リングバッファに書き込むべきバイト数．
 * @param[in]		*pFrames			Isochronous IN 転送による受信データ．
 *
 * @return			IOReturn コード．
 */
// This function is called from both the IOProc's call to convertInputSamples and by the readHandler.
// To figure out where to start coalescing from, it looks at the mCurrentFrameList, which is updated by the readHandler.
// It will copy from currentFameList+1 the number of bytes requested or one USB frame list.
// When numBytesToCoalesce == 0 it will coalesce the current USB frame list (however big it is).
// If numBytesToCoalesce != 0, it will coalesce that many bytes starting from the current frame list and going to the next one if needed.
// When called from the readHandler it will just coalesce one USB frame starting from mCurrentFrameList.
// When called from convertInputSamples, it will convert the number of bytes that corresponds to the number of samples that are being asked to be converted,
// starting from mCurrentFrameList.
IOReturn
PdjUsbAudioStream::coalesceInputSamples(
	UInt32						numBytesToCoalesce,
	IOUSBHostIsochronousFrame	*pFrames
)
{
	DP(DBG_READ, "+ " LOGSTRAUDIOSTREAM "[%p]::coalesceInputSamples(%u, %p)", this, numBytesToCoalesce, pFrames);

	IOReturn					result = kIOReturnSuccess;
	AbsoluteTime				time;
	UInt32						usbFrameIndex;
	UInt32						firstUSBFrameIndex;		// <rdar://6094454>
	UInt32						totalNumUSBFrames;		// <rdar://6094454>
	UInt32						numFramesChecked;
	UInt32						numBytesToCopy;
	UInt32						numBytesToEnd;
	UInt32						originalBufferOffset = 0;
	SInt32						numBytesLeft;
	UInt32						preWrapBytes = 0;
	UInt32						byteCount = 0;
	UInt8						*source;
	UInt8						*dest;
	bool						done;
#if DEBUGINPUT
	// <rdar://problem/7378275>
	UInt32						numBytesOnLastCopy			 = 0;
	UInt32						numBytesMargin				 = -1;
	UInt32						usbFrameIndexOnLastCopy		 = 0;
	UInt32						firstUSBFrameIndexOnLastCopy = 0;
	IOUSBHostIsochronousFrame	*pFramesOnLastCopy			 = NULL;
#endif

	if (mCoalescenceMutex)
	{
		IORecursiveLockLock(mCoalescenceMutex);
	}

	if (0 != numBytesToCoalesce)
	{
		// This is being called from the CoreAudio thread

		originalBufferOffset = mBufferOffset;	// So that when we later get called from the readHandler that we'll put everything in the right spot

		DP(DBG_INFO, "! " LOGSTRAUDIOSTREAM "[%p]::coalesceInputSamples() - Coalesce from %u %u bytes (framelist %u) on CoreAudio thread",
			this, originalBufferOffset, numBytesToCoalesce, mCurrentFrameList);

		if (mMasterMode && !mHaveTakenFirstTimeStamp)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::coalesceInputSamples() - CoreAudio thread is asking for samples without having been sent a timestamp!", this);
		}
	}
	else
	{
		// This is not being called from the CoreAudio thread
	}

	if (NULL == pFrames)
	{
		pFrames = &mUSBIsocFrames[mCurrentFrameList * mNumTransactionsPerList];
	}

	dest	= (UInt8 *)super::getSampleBuffer() + mBufferOffset;
	source	= (UInt8 *)mReadBuffer + (mCurrentFrameList * mReadUSBFrameListSize);

	// <rdar://6094454> Pre-compute these values here instead in the while loop. There is a race condition where
	// mCurrentFrameList is updated in the readHandler(), and if it changes, then it could cause usbFrameIndex to get
	// out of range when accessing pFrames. firstUSBFrameIndex should be tied to pFrames, so it should only change
	// when pFrames changes in the wrap situation. totalNumUSBFrames shouldn't change at all.
	firstUSBFrameIndex	= mCurrentFrameList * mNumTransactionsPerList;
	totalNumUSBFrames	= mNumUSBFrameLists * mNumTransactionsPerList;

	usbFrameIndex	 = 0;
	numFramesChecked = 0;
	numBytesLeft	 = numBytesToCoalesce;
	done			 = false;

	while (
				(!done)
		  &&
				('llit' != pFrames[usbFrameIndex].status)			// IOUSBFamily is processing this now
		  &&
				(-1 != pFrames[usbFrameIndex].status)				// IOUSBFamily hasn't gotten here yet
		  )
	{
		// Log unusual status here
		// [rdar://5889101]
		if (
				(!mShouldStop)
		   &&
				(
					(kIOReturnSuccess != pFrames[usbFrameIndex].status)
				&&
					(
						(kIOReturnUnderrun != pFrames[usbFrameIndex].status)
					||
						(pFrames[usbFrameIndex].completeCount < (mAverageFrameSize - 2 * mSampleSize))
					)
				)
		   )
		{
			DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "[%p]::coalesceInputSamples() - encountered unusual frame with status 0x%x in frame list %u",
				this, pFrames[usbFrameIndex].status, usbFrameIndex);
			DP(DBG_WARNING, "     pFrames[%u].status = 0x%x", usbFrameIndex, pFrames[usbFrameIndex].status);
			DP(DBG_WARNING, "     pFrames[%u].requestCount = %u", usbFrameIndex, pFrames[usbFrameIndex].requestCount);
			DP(DBG_WARNING, "     pFrames[%u].completeCount = %u", usbFrameIndex, pFrames[usbFrameIndex].completeCount);
			DP(DBG_WARNING, "     pFrames[%u].timeStamp = 0x%llx", usbFrameIndex, __OSAbsoluteTime(pFrames[usbFrameIndex].timeStamp));

			// break;

			// <rdar://6902105>, <rdar://6411577> Workaround for issue where the device sends more data than it should.
			// This causes overruns and the USB host controller may indicate that the frActCount is zero (different
			// host controller behaves differently).
			if ((kIOReturnOverrun == pFrames[usbFrameIndex].status) && (0 == pFrames[usbFrameIndex].completeCount))
			{
				// Set the frActCount to frReqCount so that at least the timing is somewhat preserved and not dropping the
				// whole packet.
				pFrames[usbFrameIndex].completeCount = pFrames[usbFrameIndex].requestCount;

				mOverrunsCount++;

				// If there is too many overruns, the audio stream is possibly corrupt constantly, so restart the
				// audio engine if the engine has multiple streams and this input stream is the master stream.
				// This is to prevent the continous corruptions.
				if (mMasterMode && (mOverrunsCount >= mOverrunsThreshold))
				{
					if (mUSBAudioDevice && mUSBAudioEngine && mUSBAudioEngine->mIOAudioStreamArray)
					{
						if (1 < mUSBAudioEngine->mIOAudioStreamArray->getCount())
						{
							// Reset the engine to prevent constant corruption.
							mUSBAudioDevice->setShouldResetEngine(mUSBAudioEngine);
						}
					}
				}
			}
		}

		numBytesToEnd = super::getSampleBufferSize() - mBufferOffset;

		// We should take the first time stamp now if we are receiving our first byte when we expect; otherwise wait until the first buffer loop.
		if (
				(!mHaveTakenFirstTimeStamp)
		   &&
				(0 == mBufferOffset)
		   &&
				(pFrames[usbFrameIndex].completeCount > 0)
		   )
		{
			if (mMasterMode && !mShouldStop)	// <rdar://problem/7378275>
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::coalesceInputSamples() - Taking first time stamp.");

				time = this->generateTimeStamp(((SInt32)usbFrameIndex - 1), 0, 0);	// <rdar://problem/7378275>

				this->takeTimeStamp(false, &time);
			}
#ifdef TARGET_PDJ
 			else if (!mMasterMode && !mShouldStop)
 			{
 				time = this->generateTimeStamp(((SInt32)usbFrameIndex - 1), 0, 0);
 
 				this->logTimeStamp(false, &time);
 			}
#endif
		}

		if ((UInt32)(pFrames[usbFrameIndex].completeCount) >= numBytesToEnd)	// <rdar://problem/7378275>
		{
			// This copy will wrap
			numBytesToCopy = numBytesToEnd;

			// Store numbers for time stamping
			preWrapBytes = numBytesToEnd;

			byteCount = pFrames[usbFrameIndex].completeCount;
		}
		else
		{
			numBytesToCopy = pFrames[usbFrameIndex].completeCount;

			if (0 == numBytesToCoalesce)
			{
				pFrames[usbFrameIndex].completeCount = 0;

#ifdef DEBUGLOGGING
				// We don't want to see these frames logged as errors later, so cook the error code if necessary.
				if (kIOReturnUnderrun == pFrames[usbFrameIndex].status)
				{
					pFrames[usbFrameIndex].status = kIOReturnSuccess;
				}
#endif
			}
		}

#if DEBUGINPUT
		// <rdar://problem/7378275>
		if (pFrames[usbFrameIndex].completeCount >= numBytesLeft)
		{
			numBytesOnLastCopy				= numBytesLeft;
			usbFrameIndexOnLastCopy			= usbFrameIndex;
			firstUSBFrameIndexOnLastCopy	= firstUSBFrameIndex;
			pFramesOnLastCopy				= pFrames;
		}
#endif

		if (0 != numBytesToCopy)
		{
			::memcpy(dest, source, numBytesToCopy);

			mBufferOffset	+= numBytesToCopy;
			numBytesLeft	-= numBytesToCopy;
		}

		UInt32		numBytesCopied = numBytesToCopy;

		if (pFrames[usbFrameIndex].completeCount >= numBytesToEnd)		// <rdar://problem/7378275>
		{
			numBytesToCopy = pFrames[usbFrameIndex].completeCount - numBytesToEnd;

			dest = (UInt8 *)super::getSampleBuffer();

			::memcpy(dest, source + numBytesCopied, numBytesToCopy);

			mBufferOffset	 = numBytesToCopy;
			numBytesLeft	-= numBytesToCopy;

			if (0 == numBytesToCoalesce)
			{
				if (mMasterMode && !mShouldStop)	// <rdar://problem/7378275>
				{
					// we have wrapped and we were called by the completion routine -- take timestamp
					// Calculate a time stamp based on our filtered rates
					time = this->generateTimeStamp(usbFrameIndex, preWrapBytes, byteCount);

					this->takeTimeStamp(true, &time);
				}
#ifdef TARGET_PDJ
 				else if (!mMasterMode && !mShouldStop)
 				{
 					time = this->generateTimeStamp(usbFrameIndex, preWrapBytes, byteCount);
 
 					this->logTimeStamp(true, &time);
 				}
#endif
			}
		}

		dest	+= numBytesToCopy;
		source	+= pFrames[usbFrameIndex].requestCount;

		usbFrameIndex++;
		numFramesChecked++;

		// <rdar://6094454> Use the pre-computed value of firstUSBFrameIndex and totalNumUSBFrames.
		// firstUSBFrameIndex should be tied to pFrames, so it should only change when pFrames changes
		// in the wrap situation. totalNumUSBFrames shouldn't change at all.
		if (
				(0 != numBytesToCoalesce)
		   &&
				((usbFrameIndex + firstUSBFrameIndex) == totalNumUSBFrames)
		   )
		{
			pFrames = &mUSBIsocFrames[0];	// wrap around the frame list and keep trying to coalesce

			usbFrameIndex = 0;

			firstUSBFrameIndex = 0;			// <rdar://6094454> Start at frame# 0.

			source = (UInt8 *)mReadBuffer;

			// DP(DBG_READ, "wrapping coalesce numBytesToCoalesce = %d", numBytesToCoalesce);
		}

		if (
				((0 == numBytesToCoalesce) && (mNumTransactionsPerList == usbFrameIndex))
				// We've coalesced the current frame list
		   ||
				((0 != numBytesToCoalesce) && (0 >= numBytesLeft))
				// We've coalesced the requested number of bytes
		   ||
				((0 != numBytesToCoalesce) && (numFramesChecked >= (mNumTransactionsPerList * mNumUSBFrameLists)))
				// We've gone through all the frame lists and there's nothing left to coalesce (starvation case)
		   )
		{
			done = true;
#if DEBUGINPUT
			// <rdar://problem/7378275>
			if ((0 != numBytesToCoalesce) && (0 >= numBytesLeft))
			{
				UInt32	actualCount = 0;

				while (
						('llit' != pFramesOnLastCopy[usbFrameIndexOnLastCopy].status)		// IOUSBFamily is processing this now
					  &&
						(-1 != pFramesOnLastCopy[usbFrameIndexOnLastCopy].status)
					  )
				{
					actualCount += pFramesOnLastCopy[usbFrameIndexOnLastCopy].completeCount;

					usbFrameIndexOnLastCopy++;

					if ((usbFrameIndexOnLastCopy + firstUSBFrameIndexOnLastCopy) == totalNumUSBFrames)
					{
						pFramesOnLastCopy = &mUSBIsocFrames[0];		// wrap around the frame list

						usbFrameIndexOnLastCopy		 = 0;
						firstUSBFrameIndexOnLastCopy = 0;			// Start at frame# 0.
					}
				}

				if (actualCount > 0)
				{
					numBytesMargin = actualCount - numBytesOnLastCopy;

					DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "[%p]::coalesceInputSamples() - numBytesMargin: %u frames: %u\n",
						this, numBytesMargin, (numBytesMargin / mSampleSize));
				}
			}
#endif
		}
	}

	if (0 != numBytesToCoalesce)
	{
		mBufferOffset = originalBufferOffset;
	}

	// Log here if we are requesting more bytes than is possible to coalesce in mNumTransactionsPerList.
	if (
			(0 != numBytesToCoalesce)
	   &&
			(numBytesLeft > 0)
	   &&
			(NULL != mStreamInterface)
	   )
	{
		DP(DBG_INFO, "! " LOGSTRAUDIOSTREAM "[%p]::coalesceInputSamples() - Requested: %u, Remaining: %d on frame list %u",
			this, numBytesToCoalesce, numBytesLeft, mCurrentFrameList);
	}

	if (mCoalescenceMutex)
	{
		IORecursiveLockUnlock(mCoalescenceMutex);	// <rdar://problem/7378275>
	}

	if (kIOReturnSuccess != result)
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::coalesceInputSamples(%u, %p) = 0x%x",
			this, numBytesToCoalesce, pFrames, result);
	}

	DP(DBG_READ, "- " LOGSTRAUDIOSTREAM "[%p]::coalesceInputSamples(%u, %p)", this, numBytesToCoalesce, pFrames);
	return result;
}

/****************************************************************************/
/**
 * @brief	タイムスタンプを上位に報告．
 *			オーディオリングバッファが末尾に達し、先頭に折り返したときは
 *			折り返し発生時刻を上位に報告しなければならない．
 *
 * @param[in]		incrementLoopCount	true ならリングバッファのループ回数をインクリメントする．
 *										初回のタイムスタンプ報告では false とする．
 * @param[in]		*timestamp			折り返し発生時のタイムスタンプ．
 */
void
PdjUsbAudioStream::takeTimeStamp(
	bool			incrementLoopCount,
	AbsoluteTime	*timestamp
)
{
	DP(DBG_TIMESTAMP, "+ " LOGSTRAUDIOSTREAM "::takeTimeStamp()");

	if (NULL != mUSBAudioEngine)
	{
		if (!mHaveTakenFirstTimeStamp)
		{
			mUSBAudioEngine->takeTimeStamp(false, timestamp);

			DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::takeTimeStamp(0, %p) - First timestamp taken", this, timestamp);

			mHaveTakenFirstTimeStamp = true;
		}
		else
		{
			mUSBAudioEngine->takeTimeStamp(incrementLoopCount, timestamp);
		}
	}

	DP(DBG_TIMESTAMP, "- " LOGSTRAUDIOSTREAM "::takeTimeStamp()");
}

#ifdef TARGET_PDJ
 /****************************************************************************/
 /**
  * @brief	タイムスタンプをデバッグログに出力．
  *			オーディオリングバッファが末尾に達し、先頭に折り返したときの
  *			折り返し発生時刻をログ出力する．
  *
  * @param[in]		incrementLoopCount	true ならリングバッファのループ回数をインクリメントする．
  *										初回のタイムスタンプ報告では false とする．
  * @param[in]		*timestamp			折り返し発生時のタイムスタンプ．
  */
 void
 PdjUsbAudioStream::logTimeStamp(
 	bool			incrementLoopCount,
 	AbsoluteTime	*timestamp
 )
 {
 	DP(DBG_TIMESTAMP, "+ " LOGSTRAUDIOSTREAM "::logTimeStamp()");
 
 	if (!mHaveTakenFirstTimeStamp)
 	{
 		mHaveTakenFirstTimeStamp = true;
 	}
 
 	UInt64		time_nanos;
 
 	::absolutetime_to_nanoseconds(*timestamp, &time_nanos);
 
 	UInt64		sampling_rate = time_nanos - mLastTimeStamp_nanos;
 
 	if (sampling_rate > 0)
 	{
 		sampling_rate = mNumSamplesInBuffer / sampling_rate;
 	}
 
 	DP(DBG_INFO, "IN ---> %llu, %llu", time_nanos, sampling_rate);
 
 	mLastTimeStamp_nanos = time_nanos;
 
 	DP(DBG_TIMESTAMP, "- " LOGSTRAUDIOSTREAM "::logTimeStamp()");
 }
#endif

/****************************************************************************/
/**
 * @brief	オーディオリングバッファの現在位置を取得．
 *
 * @return			リングバッファの現在位置（サンプルフレーム単位）．
 */
UInt32
PdjUsbAudioStream::getCurrentSampleFrame(
	void
)
{
	DP(DBG_VERBOSE, "+ " LOGSTRAUDIOSTREAM "::getCurrentSampleFrame()");

	const IOAudioStreamFormat	*theFormat;
	UInt32						currentSampleFrame;

	theFormat = &(super::format);

	if (super::getDirection() == kIOAudioStreamDirectionOutput)
	{
		currentSampleFrame = mSafeErasePoint;
	}
	else
	{
		currentSampleFrame = (mBufferOffset == mSampleBufferSize) ? 0 : mBufferOffset;
	}

	currentSampleFrame /= (theFormat->fNumChannels * (theFormat->fBitWidth / 8));

	DP(DBG_VERBOSE, "- " LOGSTRAUDIOSTREAM "::getCurrentSampleFrame() = %u", currentSampleFrame);
	return currentSampleFrame;
}

#if DEBUGLATENCY
/****************************************************************************/
/**
 * @brief	サンプルフレームに対応する USB フレームを検索．
 *
 * @param[in]		sampleFrame		サンプルフレーム番号．
 *
 * @return			USB フレーム番号．
 */
UInt64
PdjUsbAudioStream::getQueuedFrameForSample(
	UInt32			sampleFrame
)
{
	// DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::getQueuedFrameForSample(%u)", sampleFrame);

	UInt64			usbFrame = 0ull;
	UInt32			sampleByteOffset;
	UInt32			bufferByteOffset;
	UInt32			bytesToQueuePoint;
	UInt32			frameListNumber;

	FailIf(NULL == mFrameQueuedForList, Exit);
	FailIf(0 == mSampleSize, Exit);

	sampleByteOffset = sampleFrame * mSampleSize;

	bytesToQueuePoint = (
							(mLastPreparedBufferOffset > sampleByteOffset ? 0 : super::getSampleBufferSize())
						+
							mLastPreparedBufferOffset
						-
							sampleByteOffset
						);

	// DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::getQueuedFrameForSample(%u) - mLastPreparedBufferOffset = %u, sampleByteOffset = %u, bytesToQueuePoint = %u, sampleBufferSize = %u",
	//		sampleFrame, mLastPreparedBufferOffset, sampleByteOffset, bytesToQueuePoint, super::getSampleBufferSize());

	if (bytesToQueuePoint > mThisFrameListSize + mLastFrameListSize)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::getQueuedFrameForSample(%u) - sample frame is not queued in the previous two frame lists", sampleFrame);
	}
	else
	{
		// Find the USB frame on which this sample frame is queued to be transmitted on the bus
		if (bytesToQueuePoint <= mThisFrameListSize)
		{
			// This sample frame is queued to go out in the most recently queued frame list (bad if this has been clipped!)
			// Store the initial offset.
			bufferByteOffset = mThisFrameListSize;

			// Find the frame list number.
			// frameListNumber = (mNumUSBFrameLists + mCurrentFrameList + mNumUSBFrameListsToQueue - 2) % mNumUSBFrameLists;
			frameListNumber = (mCurrentFrameList + 1) % mNumUSBFrameLists;

			// DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::getQueuedFrameForSample(%u) - sample frame is queued in *PREVIOUS* frame list (%d) (mThisFrameListSize = %u, mLastFrameListSize = %u)",
			//		sampleFrame, frameListNumber, mThisFrameListSize, mLastFrameListSize);
		}
		else
		{
			// This sample frame is queued to go out in the least recently queued frame list (expected if this has been clipped)
			// Store the intial offset.
			bufferByteOffset = mThisFrameListSize + mLastFrameListSize;

			// Find the frame list number.
			// frameListNumber = (mNumUSBFrameLists + mCurrentFrameList + mNumUSBFrameListsToQueue - 3) % mNumUSBFrameLists;
			frameListNumber = mCurrentFrameList;

			// DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::getQueuedFrameForSample(%u) - sample frame was queued in frame list %d (mThisFrameListSize = %u, mLastFrameListSize = %u)",
			//		sampleFrame, frameListNumber, mThisFrameListSize, mLastFrameListSize);
		}

		// Get the first byte of the frame list
		bufferByteOffset = (super::getSampleBufferSize() + mLastPreparedBufferOffset - bufferByteOffset) % super::getSampleBufferSize();
		// DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::getQueuedFrameForSample(%u) - frame list %d starting queueing at byte %u",
		//		sampleFrame, frameListNumber, bufferByteOffset);

		// We've already determined in which frame list the sample frame lies and the buffer offset at which we should start. We no longer care about the
		// actual buffer offset, but we must preserve the order of the buffer offset and the sample byte offset. We can do this by "unrolling" the ring
		// buffer and not marking the wrap any more.

		if (bufferByteOffset > sampleByteOffset)
		{
			// Add a buffer size so we don't have to worry around looping back around to the zero byte again
			sampleByteOffset += super::getSampleBufferSize();
		}

		// Find the sample byte
		for (UInt32 i = 0; i < mNumTransactionsPerList; i++)
		{
			bufferByteOffset += mUSBIsocFrames[frameListNumber * mNumTransactionsPerList].requestCount;

			// DP(DBG_INFO, "i: %d, bufferByteOffset = %u", i, bufferByteOffset);

			if (sampleByteOffset < bufferByteOffset)
			{
				// The sample frame is queued to go out in this frame.
				usbFrame = mFrameQueuedForList[frameListNumber] + i;

				break;
			}
		}

		FailIf(0 == usbFrame, Exit);
	}

Exit:
	// DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::getQueuedFrameForSample(%u) = %llu", sampleFrame, usbFrame);
	return usbFrame;

}
#endif

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの同期タイプを取得．
 *
 * @return			エンドポイントディスクリプタの bmAttributes フィールドで指定されている同期タイプ．
 */
UInt8
PdjUsbAudioStream::getSyncType(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::getSyncType()");

#ifdef TARGET_PDJ
    UInt8								syncType = kNoneSyncType;
    //UInt8								syncType = kAdaptiveSyncType;
    //UInt8								syncType = kAsynchSyncType;
#else
	PdjUsbConfigurationDictionary		*configDictionary;
	UInt8								direction = 0;
	UInt8								address = 0;
	UInt8								syncType = kNoneSyncType;

	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	FailIf(kIOReturnSuccess != configDictionary->getIsocEndpointDirection(&direction, mInterfaceNumber, mAlternateSettingID), Exit);
	FailIf(kIOReturnSuccess != configDictionary->getIsocEndpointAddress(&address, mInterfaceNumber, mAlternateSettingID, direction), Exit);
	FailIf(kIOReturnSuccess != configDictionary->getIsocEndpointSyncType(&syncType, mInterfaceNumber, mAlternateSettingID, address), Exit);

Exit:
#endif

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::getSyncType() = %d", syncType);
	return syncType;
}

/****************************************************************************/
/**
 * @brief	lock delay 値を取得．
 *
 * @return			USB Audio Class のエンドポイントディスクリプタで指定された
 *					LockDelay 値 [mSec]
 */
UInt32
PdjUsbAudioStream::getLockDelayFrames(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::getLockDelayFrames()");

	PdjUsbConfigurationDictionary		*configDictionary;
	UInt32								usbFramesToDelay			= 0;
	UInt16								averageFrameSamples;
	UInt16								additionalSampleFrameFreq;
	UInt8								lockDelay					= 0;
	UInt8								lockDelayUnits				= 0;

	FailIf(NULL == (configDictionary = mUSBAudioDevice->getConfigDictionary()), Exit);

	this->calculateSamplesPerPacket(mCurSampleRate.whole, &averageFrameSamples, &additionalSampleFrameFreq);

	// If successful, this operation will advance the first frame to queue, so this must be done prior to reading the frame lists.
	// [rdar://5083342] Use the lockDelay if available to determine how many USB frames to prime.

	FailIf(
				kIOReturnSuccess != configDictionary->asEndpointGetLockDelay(
															&lockDelay,
															mInterfaceNumber,
															mAlternateSettingID
													  ),
				Exit
		  );

	FailIf(
				kIOReturnSuccess != configDictionary->asEndpointGetLockDelayUnits(
															&lockDelayUnits,
															mInterfaceNumber,
															mAlternateSettingID
													  ),
				Exit
		  );

	if ((lockDelay) && (lockDelayUnits))
	{
		switch (lockDelayUnits)
		{
			case kLockDelayUnitMilliseconds:
				usbFramesToDelay = lockDelay;
				break;

			case kLockDelayUnitsDecodedPCMSamples:
				// Add one USB frame for every maximum packet size in samples and one for any partial remainder
				usbFramesToDelay = (
										( lockDelay / (averageFrameSamples + 1) )
								   +
										((lockDelay % (averageFrameSamples + 1)) ? 1 : 0)
								   );
				break;

			default:
				usbFramesToDelay = kNumUSBFramesToPrime;
				break;
		}
	}
	else
	{
		usbFramesToDelay = kNumUSBFramesToPrime;
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::getLockDelayFrames() = %u", usbFramesToDelay);
	return usbFramesToDelay;
}

/****************************************************************************/
/**
 * @brief	プラグイン登録．
 *			プラグインがアタッチされた場合に、プラグインから呼び出される．
 *
 * @param[in]		*thePlugin		アタッチされたプラグイン．
 */
void
PdjUsbAudioStream::registerPlugin(
	PdjUsbAudioPlugin	*thePlugin
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::registerPlugin(%p)", thePlugin);

	mPlugin = thePlugin;

	mPluginInitThread = ::thread_call_allocate((thread_call_func_t)pluginLoaded, (thread_call_param_t)this);

	if (NULL != mPluginInitThread)
	{
		::thread_call_enter(mPluginInitThread);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::registerPlugin()");
}

/****************************************************************************/
/**
 * @brief	デバイス設定用の外部アプリを登録．
 *			プラグインからの要求により、外部アプリの関連付けを行う．
 *
 * @param[in]		*bundleID		外部アプリのバンドル ID 文字列．
 */
void
PdjUsbAudioStream::pluginSetConfigurationApp(
	const char		*bundleID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::pluginSetConfigurationApp(%s)", bundleID);

	if (bundleID)
	{
		mUSBAudioDevice->setConfigurationApp(bundleID);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::pluginSetConfigurationApp()");
}

/****************************************************************************/
/**
 * @brief	デバイスリクエスト．
 *			プラグインからの要求により、USB デバイスリクエストを発行する．
 *
 * @param[in]		*request		デバイスリクエスト．
 * @param[in, out]	*dataBuffer		デバイスリクエストのデータ．
 * @param[in]		*completion		完了ルーチン．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::pluginDeviceRequest(
	StandardUSB::DeviceRequest	*request,
	IOMemoryDescriptor			*dataBuffer,
	IOUSBHostCompletion			*completion
)
{
	IOReturn			result = kIOReturnBadArgument;

	if (request)
	{
		result = mUSBAudioDevice->deviceRequest(request, dataBuffer, completion);
	}

	return result;
}

#pragma mark - Protected Methods -

/****************************************************************************/
/**
 * @brief	プラグインがロードされたときの処理．
 *
 * @param[in]		*usbAudioStreamObject	プラグインがアタッチされた
 *											PdjUsbAudioStream のインスタンス．
 */
void
PdjUsbAudioStream::pluginLoaded(
	PdjUsbAudioStream	*usbAudioStreamObject
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::pluginLoaded()");

	IOReturn				result;

	if (usbAudioStreamObject->mPlugin && usbAudioStreamObject->mStreamInterface)
	{
		usbAudioStreamObject->mPlugin->open(usbAudioStreamObject);

		result = usbAudioStreamObject->mPlugin->pluginInit(
													usbAudioStreamObject,
													USBToHost16(usbAudioStreamObject->mStreamInterface->getDevice()->getDeviceDescriptor()->idVendor),
													USBToHost16(usbAudioStreamObject->mStreamInterface->getDevice()->getDeviceDescriptor()->idProduct)
												);
		if (result == kIOReturnSuccess)
		{
			DP(DBG_INFO, "success initing the plugin");

			usbAudioStreamObject->mPlugin->pluginSetDirection(
												(IOAudioStreamDirection)usbAudioStreamObject->mDirection
										   );

			usbAudioStreamObject->mPlugin->pluginSetFormat(
												usbAudioStreamObject->getFormat(),
												&usbAudioStreamObject->mCurSampleRate
										   );
		}
		else
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::pluginLoaded() - Error initing the plugin");

			usbAudioStreamObject->mPlugin->close(usbAudioStreamObject);
			usbAudioStreamObject->mPlugin = NULL;
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::pluginLoaded()");
}

/****************************************************************************/
/**
 * @brief	Alternate Setting とサンプリング周波数の初期値を取得．
 *
 * @param[out]		*altSettingID	ストリーミング interface の alternate setting 番号．
 * @param[in, out]	*sampleRate		サンプリング周波数の初期値．
 *
 * @return			IOReturn コード．
 */
// getDefaultSettings added for rdar://3866513
IOReturn
PdjUsbAudioStream::getDefaultSettings(
	UInt8				*altSettingID,
	IOAudioSampleRate	*sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::getDefaultSettings()", this);

	IOReturn							result = kIOReturnError;
	UInt16								format;
	UInt8								newAltSettingID;
	IOAudioSampleRate					newSampleRate;
	PdjUsbConfigurationDictionary		*configDictionary = NULL;

	newSampleRate.whole		= sampleRate->whole;
	newSampleRate.fraction	= 0;

	configDictionary = mUSBAudioDevice->getConfigDictionary();

	// The sample rate is passsed in, so all we should do here is to pick the sample format (size, and number of channels),
	// so try 16-bit stereo. If possible, never pick anything other than PCM for default.
	result = configDictionary->getAltSettingWithSettings(
									&newAltSettingID,
									mInterfaceNumber,
									kChannelDepth_STEREO,
									kBitDepth_16bits,
									newSampleRate.whole
							   );
	if (
			(kIOReturnSuccess != result)
	   ||
			(
				(kIOReturnSuccess == configDictionary->getFormat(&format, mInterfaceNumber, newAltSettingID))
			&&
				(PCM != format)
			)
	   )
	{
		// Didn't have stereo, so try mono
		result = configDictionary->getAltSettingWithSettings(
										&newAltSettingID,
										mInterfaceNumber,
										kChannelDepth_MONO,
										kBitDepth_16bits,
										newSampleRate.whole
								   );
	}

	if (
			(kIOReturnSuccess != result)
	   ||
			(
				(kIOReturnSuccess == configDictionary->getFormat(&format, mInterfaceNumber, newAltSettingID))
			&&
				(PCM != format)
			)
	   )
	{
		UInt8		numAltSettings = 0;

		if (kIOReturnSuccess == configDictionary->getNumAltSettings(&numAltSettings, mInterfaceNumber))
		{
			bool	startAtZero = configDictionary->alternateSettingZeroCanStream(mInterfaceNumber);

			for (UInt8 altSetting = (startAtZero ? 0 : 1); altSetting < numAltSettings; altSetting++)
			{
				// Don't have a mono or stereo 16-bit interface, so try for any format at the given sample rate.
				if (
						(configDictionary->verifySampleRateIsSupported(mInterfaceNumber, altSetting, newSampleRate.whole))
				   &&
						(kIOReturnSuccess == configDictionary->getFormat(&format, mInterfaceNumber, altSetting))
				   &&
						(PCM == format)
				   )
				{
					newAltSettingID = altSetting;

					result = kIOReturnSuccess;

					break;
				}
			}
		}
	}

	if (
			(kIOReturnSuccess != result)
	   ||
			(
				(kIOReturnSuccess == configDictionary->getFormat(&format, mInterfaceNumber, newAltSettingID))
			&&
				(PCM != format)
			)
	   )
	{
		// Just take the first interface.
		newAltSettingID = configDictionary->alternateSettingZeroCanStream(mInterfaceNumber) ? 0 : 1;

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::getDefaultSettings() - Taking first available alternate setting (%d)",
			this, newAltSettingID);

		FailIf(
					kIOReturnSuccess != (
											result = configDictionary->getHighestSampleRate(
																			&(newSampleRate.whole),
																			mInterfaceNumber,
																			newAltSettingID
																	   )
										),
					Exit
			  );
	}

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::getDefaultSettings() - Default sample rate is %u", this, newSampleRate.whole);
	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::getDefaultSettings() - Default alternate setting ID is %u", this, newAltSettingID);

	FailIf(0 == newSampleRate.whole, Exit);

	*sampleRate		= newSampleRate;
	*altSettingID	= newAltSettingID;

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::getDefaultSettings(%d, %u) = 0x%x", this, *altSettingID, sampleRate->whole, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	フィードバックエンドポイントを検索．
 *			Asynchronous OUT エンドポイントに対応するフィードバックエンドポイントを探す．
 *			出力ストリームの場合のみ呼び出される．
 *
 * @param[in]		*configDictionary	USB Audio Class のディスクリプタ．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::checkForFeedbackEndpoint(
	PdjUsbConfigurationDictionary		*configDictionary
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::checkForFeedbackEndpoint()");

	IOReturn					result = kIOReturnSuccess;
	IOUSBFindEndpointRequest	associatedEndpoint;
	UInt16						maxPacketSize;
	UInt8						assocEndpoint;
	UInt8						address;
	UInt8						syncType;
	UInt16						framesUntilRefresh;

	mAssociatedPipe = NULL;

	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == mStreamInterface, Exit);

	FailIf(configDictionary->getIsocEndpointAddress(&address, mInterfaceNumber, mAlternateSettingID, mDirection), Exit);
	FailIf(configDictionary->getIsocEndpointSyncType(&syncType, mInterfaceNumber, mAlternateSettingID, address), Exit);

	if (kAsynchSyncType == syncType)
	{
		// DP(DBG_INFO, "checking endpoint %d for an associated endpoint", address);

		FailIf(
					kIOReturnSuccess != configDictionary->getIsocAssociatedEndpointAddress(
																&assocEndpoint,
																mInterfaceNumber,
																mAlternateSettingID,
																address
														  ),
					Exit
			  );

		if (assocEndpoint != 0)
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::checkForFeedbackEndpoint() - assocEndpoint = 0x%x", this, assocEndpoint);

			FailIf(
						kIOReturnSuccess != configDictionary->getIsocAssociatedEndpointRefreshInt(
																	&mRefreshInterval,
																	mInterfaceNumber,
																	mAlternateSettingID,
																	assocEndpoint
															  ),
						Exit
				  );

			FailIf(
						kIOReturnSuccess != configDictionary->getIsocAssociatedEndpointMaxPacketSize(
																	&maxPacketSize,
																	mInterfaceNumber,
																	mAlternateSettingID,
																	assocEndpoint
															  ),
						Exit
				  );

			if (kUSBDeviceSpeedHigh == mUSBAudioDevice->getDeviceSpeed())
			{
				// Request 4 bytes for the 16.16 value if the endpoint allows it
				mFeedbackPacketSize = (maxPacketSize < kFixedPoint16_16ByteSize) ? maxPacketSize : kFixedPoint16_16ByteSize;
			}
			else
			{
				mFeedbackPacketSize = kFixedPoint10_14ByteSize;
			}

			DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::checkForFeedbackEndpoint() - Synch endpoint has refresh interval %d, feedback packet size %d",
				this, mRefreshInterval, mFeedbackPacketSize);

			mRefreshInterval = mRefreshInterval ? mRefreshInterval : kMinimumSyncRefreshInterval;

			framesUntilRefresh = 1 << mRefreshInterval;		// the same as (2 ^ mRefreshInterval)

			// If the hardware needs to be updated more often than PLAY_NUM_USB_FRAMES_PER_LIST ms,
			// change list size to PLAY_NUM_USB_FRAMES_PER_LIST_SYNC frames.
			if (framesUntilRefresh < mNumUSBFramesPerList)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::checkForFeedbackEndpoint() - Need to adjust mNumUSBFramesPerList: %u < %u",
					this, framesUntilRefresh, mNumUSBFramesPerList);

				if (NULL != mUSBIsocFrames)
				{
					DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::checkForFeedbackEndpoint() - Disposing of current mUSBIsocFrames[%p]", this, mUSBIsocFrames);

					IOFree(mUSBIsocFrames, mNumUSBFrameLists * mNumTransactionsPerList * sizeof(IOUSBHostIsochronousFrame));
					mUSBIsocFrames = NULL;
				}

				mNumUSBFramesPerList	= PLAY_NUM_USB_FRAMES_PER_LIST_SYNC;
				mNumTransactionsPerList	= mNumUSBFramesPerList * mTransactionsPerUSBFrame;
				mNumUSBFrameLists		= mNumUSBFrameListsToQueue * 2;

				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::checkForFeedbackEndpoint() - mNumUSBFramesPerList = %u, mNumUSBFrameListsToQueue = %u, mNumUSBFrameLists = %u",
					this, mNumUSBFramesPerList, mNumUSBFrameListsToQueue, mNumUSBFrameLists);

				mUSBIsocFrames = (IOUSBHostIsochronousFrame *)IOMalloc(mNumUSBFrameLists * mNumTransactionsPerList * sizeof(IOUSBHostIsochronousFrame));
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::checkForFeedbackEndpoint() - mUSBIsocFrames is now %p", this, mUSBIsocFrames);
				FailIf(NULL == mUSBIsocFrames, Exit);

				// <rdar://7568547> Initialize the USB Isoc frames so that coalesceInputSamples() will not panic
				// due to uninitialized values in frStatus & frActCount.
				this->initializeUSBFrameList(mUSBIsocFrames, mNumUSBFrameLists * mNumTransactionsPerList);
			}

			associatedEndpoint.type			 = kUSBIsoc;
			associatedEndpoint.direction	 = kUSBIn;				// The associated endpoint always goes "in"
			// The sample rate should be either a 3-byte 10.14 or a 4-byte 16.16
			associatedEndpoint.maxPacketSize = mFeedbackPacketSize;
			associatedEndpoint.interval		 = 0xFF;

			mAssociatedPipe = PdjUsbAudioDevice::findPipe(mStreamInterface, &associatedEndpoint);
			FailWithAction(NULL == mAssociatedPipe, result = kIOReturnError, Exit);

			if (NULL == mAssociatedEndpointMemoryDescriptor)
			{
				// <rdar://7000283> Remove the use of deprecated IOMallocContiguous(). Use IOBufferMemoryDescriptor to allocate memory instead.
				mAssociatedEndpointMemoryDescriptor = IOBufferMemoryDescriptor::withOptions(kIODirectionInOut, sizeof(UInt32), sizeof(UInt32));
				FailIf(NULL == mAssociatedEndpointMemoryDescriptor, Exit);

				mAverageSampleRateBuffer = (UInt32 *)mAssociatedEndpointMemoryDescriptor->getBytesNoCopy();
				FailIf(NULL == mAverageSampleRateBuffer, Exit);

				::bzero(mAverageSampleRateBuffer, sizeof(UInt32));
			}

			mSampleRateFrame.status			= -1;
			mSampleRateFrame.requestCount	= mFeedbackPacketSize;
			mSampleRateFrame.completeCount	= 0;

			mSampleRateCompletion.owner		= (void *)this;
			mSampleRateCompletion.action	= sampleRateHandler;
			mSampleRateCompletion.parameter	= 0;

			mAssociatedPipe->retain();
		} // if (assocEndpoint != 0)
		else
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::checkForFeedbackEndpoint() - No associated synch endpoint found.", this);
		}
	} // if (kAsynchSyncType == syncType)
	else
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::checkForFeedbackEndpoint() - No associated synch endpoint.", this);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::checkForFeedbackEndpoint() = 0x%x", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	サンプリング周波数設定のデバイスリクエストを発行．
 *
 * @param[in]		address			エンドポイントアドレス．
 * @param[in]		sampleRate		設定するサンプリング周波数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::setSampleRateControl(
	UInt8			address,
	UInt32			sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::setSampleRateControl()");

	IOReturn					result = kIOReturnError;
	StandardUSB::DeviceRequest	devReq;
	UInt32						theSampleRate;
	UInt32						bytesTransferred = 0;

	FailIf(NULL == mUSBAudioDevice, Exit);

	// if (usbAudio->IsocEndpointHasSampleFreqControl(mInterfaceNumber, mAlternateSettingID))

	devReq.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBClass, kUSBEndpoint);
	devReq.bRequest		 = SET_CUR;
	devReq.wValue		 = (SAMPLING_FREQ_CONTROL << 8) | 0;
	devReq.wIndex		 = address;
	devReq.wLength		 = 3;
	theSampleRate		 = OSSwapHostToLittleInt32(sampleRate);

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::setSampleRateControl() - Control interface %u, alt setting %u, endpoint address 0x%x, sample rate (little endian) 0x%x",
		this, mInterfaceNumber, mAlternateSettingID, devReq.wIndex, theSampleRate);

	result = mUSBAudioDevice->deviceRequest(&devReq, &theSampleRate, &bytesTransferred);

	FailIf(kIOReturnSuccess != result, Exit);

Exit:

	if (kIOReturnSuccess != result)
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::setSampleRateControl() = 0x%x", this, result);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::setSampleRateControl() = 0x%x", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	メモリディスクリプタを確保．
 *
 * @param[in]		options			メモリディスクリプタのオプション指定．
 * @param[in]		capacity		確保するバイト数．
 * @param[in]		alignment		確保するメモリのアラインメント．
 *
 * @return			確保したメモリディスクリプタ．
 */
// [rdar://4487489] - Use this method to allocate all USB buffers.
IOBufferMemoryDescriptor *
PdjUsbAudioStream::allocateBufferDescriptor(
	IOOptionBits	options,
	vm_size_t		capacity,
	vm_offset_t		alignment
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "[%p]::allocateBufferDescriptor()", this);

	const mach_vm_address_t		physicalMask = 0xFFFFFFFFFFFFF000ull;

	IOBufferMemoryDescriptor	*bufferDescriptorPtr = IOBufferMemoryDescriptor::inTaskWithPhysicalMask(
															kernel_task,
															options,
															capacity,
															physicalMask
													   );

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "[%p]::allocateBufferDescriptor() = %p", this, bufferDescriptorPtr);
	return bufferDescriptorPtr;
}

/****************************************************************************/
/**
 * @brief	USB フレームリスト初期化．
 *
 * @param[in]		*usbIsocFrames	初期化する USB フレームリスト．
 * @param[in]		numFrames		リスト長．
 */
// <rdar://7568547> Initialize the USB frame list to proper values.
void
PdjUsbAudioStream::initializeUSBFrameList(
	IOUSBHostIsochronousFrame	*usbIsocFrames,
	UInt32						numFrames
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::initializeUSBFrameList()");

	for (UInt32 frameIndex = 0; frameIndex < numFrames; frameIndex++)
	{
		usbIsocFrames[frameIndex].status		= -1;
		usbIsocFrames[frameIndex].requestCount	= 0;
		usbIsocFrames[frameIndex].completeCount	= 0;

		AbsoluteTime_to_scalar(&(usbIsocFrames[frameIndex].timeStamp)) = 0ull;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::initializeUSBFrameList()");
}

/****************************************************************************/
/**
 * @brief	Isochronous OUT 転送用 USB フレームリストの準備．
 *
 * @param[in]		frameListNum	USB フレームリスト番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::prepareWriteFrameList(
	UInt32			frameListNum
)
{
	DP(DBG_WRITE, "+ " LOGSTRAUDIOSTREAM "::prepareWriteFrameList()");

	IOReturn		result = kIOReturnError;	// assume failure
	UInt32			thisFrameListSize = 0;
#if DEBUGLATENCY
	UInt32			frameListByteCount = 0;		// this is always updated, regardless of wrapping
#endif
	UInt32			thisFrameSize;
	UInt32			firstFrame;
	UInt32			numBytesToBufferEnd;
	UInt32			lastPreparedByte;
	UInt32			remainderedSamples;
	UInt16			integerSamplesInFrame;
	UInt16			averageSamplesInFrame;
	UInt16			bytesAfterWrap = 0;			// for UHCI support
	bool			haveWrapped = false;
#ifdef TARGET_PDJ
 	UInt16			*list = mUSBAudioEngine->getInputSizeList(false, mNumTransactionsPerList);
#endif

	FailIf(0 == mTransactionsPerUSBFrame, Exit);

	firstFrame = frameListNum * mNumTransactionsPerList;

	mUSBCompletion[frameListNum].owner		= (void *)this;
	mUSBCompletion[frameListNum].action		= writeHandler;
	mUSBCompletion[frameListNum].parameter	= 0;				// Set to number of bytes from the 0 wrap, 0 if this buffer didn't wrap

	numBytesToBufferEnd = super::getSampleBufferSize() - mLastPreparedBufferOffset;

	lastPreparedByte = mLastPreparedBufferOffset;

	// <rdar://problems/5600254> Calculate the sample rate in terms of transactions instead of milliseconds. For full speed devices, this changes nothing.
	// <rdar://problems/6954295> Store Async feedback in samples per frame/microframe as a 16.16 fixed point number
	averageSamplesInFrame	= mSamplesPerPacket.whole;
	remainderedSamples		= mSamplesPerPacket.fraction;

	for (UInt32 numTransactionsPrepared = 0; numTransactionsPrepared < mNumTransactionsPerList; numTransactionsPrepared++)
	{
		// [rdar://5600254] Remaindered samples are to be determined on a transaction basis, not a USB frame basis.
		integerSamplesInFrame = averageSamplesInFrame;

		mFractionalSamplesLeft += remainderedSamples;

		if (mFractionalSamplesLeft >= kSampleFractionAccumulatorRollover)	// <rdar://problem/6954295>
		{
			integerSamplesInFrame++;

			mFractionalSamplesLeft -= kSampleFractionAccumulatorRollover;	// <rdar://problem/6954295>
		}

#ifdef TARGET_PDJ
 		if (   (list[numTransactionsPrepared] >= (integerSamplesInFrame - 3))
 			&& (list[numTransactionsPrepared] <= (integerSamplesInFrame + 3)))
 		{
 			integerSamplesInFrame = list[numTransactionsPrepared];
 		}
#endif

		thisFrameSize = integerSamplesInFrame * mSampleSize;

#if DEBUGLATENCY
		frameListByteCount += thisFrameSize;
#endif

		if (thisFrameSize >= numBytesToBufferEnd)
		{
			bytesAfterWrap = thisFrameSize - numBytesToBufferEnd;

			mNumFramesInFirstList = numTransactionsPrepared + 1;

			mUSBCompletion[frameListNum].parameter = (void *)(uintptr_t)((mNumFramesInFirstList << 16) | bytesAfterWrap);	// Number of bytes after wrap

#ifdef TARGET_PDJ
#else
			if (mUHCISupport)
			{
				DP(DBG_UHCI, "? " LOGSTRAUDIOSTREAM "::prepareWriteFrameList(): Wrapping because (thisFrameSize = ) %u >= (numBytesToBufferEnd = ) %u",
					thisFrameSize, numBytesToBufferEnd);
				DP(DBG_UHCI, "? " LOGSTRAUDIOSTREAM "::prepareWriteFrameList(): bytesAfterWrap = %u, numTransactionsPrepared = %u, numTransactionsPerList %u",
					bytesAfterWrap, numTransactionsPrepared, mNumTransactionsPerList);

				mWrapDescriptors[0]->initSubRange(
											mUSBBufferDescriptor,
											mLastPreparedBufferOffset,
											super::getSampleBufferSize() + bytesAfterWrap - mLastPreparedBufferOffset,
											kIODirectionOut
									 );

				DP(DBG_UHCI, "? " LOGSTRAUDIOSTREAM "::prepareWriteFrameList(): initSubRange 0: %u to %u",
					mLastPreparedBufferOffset, super::getSampleBufferSize() + bytesAfterWrap);
				DP(DBG_UHCI, "? " LOGSTRAUDIOSTREAM "::prepareWriteFrameList(): %d frames in first list",
					mNumFramesInFirstList);
			}
			else
#endif
			{
				mWrapDescriptors[0]->initSubRange(
											mUSBBufferDescriptor,
											mLastPreparedBufferOffset,
											super::getSampleBufferSize() - mLastPreparedBufferOffset,
											kIODirectionOut
									 );
			}

			numBytesToBufferEnd = super::getSampleBufferSize() - bytesAfterWrap;

			lastPreparedByte = bytesAfterWrap;

			haveWrapped = true;
		}
		else
		{
			thisFrameListSize	+= thisFrameSize;
			lastPreparedByte	+= thisFrameSize;
			numBytesToBufferEnd	-= thisFrameSize;
		}

		mUSBIsocFrames[firstFrame + numTransactionsPrepared].status			= -1;
		mUSBIsocFrames[firstFrame + numTransactionsPrepared].completeCount	= 0;
		mUSBIsocFrames[firstFrame + numTransactionsPrepared].requestCount	= thisFrameSize;
	} // for numTransactionsPrepared

	if (haveWrapped)
	{
		mNeedTimeStamps = true;

#ifdef TARGET_PDJ
#else
		if (mUHCISupport)
		{
			// DP(DBG_INFO, "bytesAfterWrap = %d lastPreparedByte = %d numBytesToBufferEnd = %d mNumFramesInFirstList = %d",
			//		bytesAfterWrap, lastPreparedByte, numBytesToBufferEnd, mNumFramesInFirstList);

			mWrapDescriptors[1]->initSubRange(mUSBBufferDescriptor, bytesAfterWrap, lastPreparedByte - bytesAfterWrap, kIODirectionOut);

			DP(DBG_UHCI, "? " LOGSTRAUDIOSTREAM "::prepareWriteFrameList(): initSubRange 1: %u to %u", bytesAfterWrap, lastPreparedByte);

			if (lastPreparedByte != bytesAfterWrap)
			{
				// This is where we setup our extra completion for the second wrap write
				mExtraUSBCompletion.owner = (void *)this;
				mExtraUSBCompletion.action = writeHandlerForUHCI;
				// mExtraUSBCompletion.parameter = (void *)((mNumFramesInFirstList << 16) | bytesAfterWrap);
			}
		}
		else
#endif
		{
			mWrapDescriptors[1]->initSubRange(mUSBBufferDescriptor, 0, lastPreparedByte, kIODirectionOut);

			if (NULL != mWrapRangeDescriptor)
			{
				mWrapRangeDescriptor->release();
				mWrapRangeDescriptor = NULL;
			}

			mWrapRangeDescriptor = IOMultiMemoryDescriptor::withDescriptors(
																(IOMemoryDescriptor **)mWrapDescriptors,
																2,
																kIODirectionOut,
																true
															);
			FailIf(NULL == mWrapRangeDescriptor, Exit);
		}
	}
	else
	{
		mSampleBufferDescriptors[frameListNum]->initSubRange(
													mUSBBufferDescriptor,
													mLastPreparedBufferOffset,
													thisFrameListSize,
													kIODirectionOut
												);
	}

	mSafeErasePoint				= mNextSafeErasePoint;
	mNextSafeErasePoint			= mLastPreparedBufferOffset;
	mLastPreparedBufferOffset	= lastPreparedByte;
#if DEBUGLATENCY
	mLastFrameListSize			= mThisFrameListSize;
	mThisFrameListSize			= frameListByteCount;
#endif

	result = kIOReturnSuccess;

Exit:
	DP(DBG_WRITE, "- " LOGSTRAUDIOSTREAM "::prepareWriteFrameList() = 0x%x", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous OUT 転送起動．
 *
 * @param[in]		frameListNum	USB フレームリスト番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::writeFrameList(
	UInt32			frameListNum
)
{
	DP(DBG_WRITE, "+ " LOGSTRAUDIOSTREAM "::writeFrameList()");

	IOReturn		result = kIOReturnError;

	FailIf(NULL == mStreamInterface, Exit);

	result = this->prepareWriteFrameList(frameListNum);
	FailIf(kIOReturnSuccess != result, Exit);

	result = kIOReturnError;	// reset the error in case mPipe is null
	FailIf(NULL == mPipe, Exit);

	if (mNeedTimeStamps)
	{
#ifdef TARGET_PDJ
#else
		if (mUHCISupport)
		{
			// We might have to do two separate writes here. Do the first one, and then if necessary, do the second.
			DP(DBG_UHCI, "? " LOGSTRAUDIOSTREAM "::writeFrameList() - Writing mSampleBufferWrapDescriptors[0]");

			result = mPipe->io(
								mWrapDescriptors[0],
								&mUSBIsocFrames[frameListNum * mNumTransactionsPerList],
								mNumFramesInFirstList,
								mUSBFrameToQueue,
								&mUSBCompletion[frameListNum]
							);

			// mNumFramesInFirstList must be less than mNumUSBFramesPerList if we wrapped.
			if (mNumFramesInFirstList < mNumTransactionsPerList)
			{
				DP(DBG_UHCI, "? " LOGSTRAUDIOSTREAM "::writeFrameList() - writeFrameList: Writing mSampleBufferWrapDescriptors[1] to frame %llu",
					mUSBFrameToQueue + mNumFramesInFirstList);

				result = mPipe->io(
									mWrapDescriptors[1],
									&mUSBIsocFrames[frameListNum * mNumTransactionsPerList + mNumFramesInFirstList],
									mNumTransactionsPerList - mNumFramesInFirstList,
									mUSBFrameToQueue + mNumFramesInFirstList,
									&mExtraUSBCompletion
								);

				// result = thePipe->io(
				//						mSampleBufferWrapDescriptors[1],
				//						&theUSBIsocFrames[frameListNumber * numUSBFramesPerList + mNumFramesInFirstList],
				//						numUSBFramesPerList - mNumFramesInFirstList,
				//						usbFrameToQueueAt + mNumFramesInFirstList,
				//						&extraUsbCompletion
				//					 );
			}
		}
		else
#endif
		{
			result = mPipe->io(
								mWrapRangeDescriptor,
								&mUSBIsocFrames[frameListNum * mNumTransactionsPerList],
								mNumTransactionsPerList,
								mUSBFrameToQueue,
								&mUSBCompletion[frameListNum]
							);
		}

		mNeedTimeStamps = false;
	}
	else
	{
		result = mPipe->io(
							mSampleBufferDescriptors[frameListNum],
							&mUSBIsocFrames[frameListNum * mNumTransactionsPerList],
							mNumTransactionsPerList,
							mUSBFrameToQueue,
							&mUSBCompletion[frameListNum]
						);
	}

	FailIf(result != kIOReturnSuccess, Exit);

	// keep track of this frame number for time stamping
	if (NULL != mFrameQueuedForList)
	{
		mFrameQueuedForList[frameListNum] = mUSBFrameToQueue;
	}

	mUSBFrameToQueue += mNumUSBFramesPerList;

Exit:
	if (kIOReturnSuccess != result)
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::writeFrameList() - failed with error 0x%x", this, result);
	}

	DP(DBG_WRITE, "- " LOGSTRAUDIOSTREAM "::writeFrameList() = 0x%x", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous IN 転送起動．
 *
 * @param[in]		frameListNum	USB フレームリスト番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioStream::readFrameList(
	UInt32			frameListNum
)
{
	DP(DBG_READ, "+ " LOGSTRAUDIOSTREAM "::readFrameList()");

	IOReturn		result = kIOReturnError;	// assume failure
	UInt32			firstFrame;

	firstFrame = frameListNum * mNumTransactionsPerList;

	mUSBCompletion[frameListNum].owner		= (void *)this;
	mUSBCompletion[frameListNum].action		= readHandler;
	mUSBCompletion[frameListNum].parameter	= (void *)(uintptr_t)frameListNum;	// what frame list index this buffer is

	for (UInt32 numTransactionsPrepared = 0; numTransactionsPrepared < mNumTransactionsPerList; numTransactionsPrepared++)
	{
		mUSBIsocFrames[firstFrame + numTransactionsPrepared].status			= -1;
		mUSBIsocFrames[firstFrame + numTransactionsPrepared].completeCount	= 0;
		mUSBIsocFrames[firstFrame + numTransactionsPrepared].requestCount	= mReadUSBFrameSize;
		// [rdar://5355808] This should be the smaller of the calculated size and the maxPacketSize.

		// This casting can be removed when AbsoluteTime is changed to UInt64 in Tiger.
		AbsoluteTime_to_scalar(&(mUSBIsocFrames[firstFrame + numTransactionsPrepared].timeStamp)) = 0ull;
	}

	if (NULL != mPipe)
	{
		result = mPipe->io(
							mSampleBufferDescriptors[frameListNum],
							&mUSBIsocFrames[firstFrame],
							mNumTransactionsPerList,
							mUSBFrameToQueue,
							&mUSBCompletion[frameListNum]
						);
		if (kIOReturnSuccess != result)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::readFrameList() - Error 0x%x reading from pipe", this, result);
		}

		// keep track of this frame number for time stamping
		if (NULL != mFrameQueuedForList)
		{
			mFrameQueuedForList[frameListNum] = mUSBFrameToQueue;
		}

		mUSBFrameToQueue += mNumUSBFramesPerList;
	}
	else
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::readFrameList() - mPipe is NULL!", this);
	}

	DP(DBG_READ, "- " LOGSTRAUDIOSTREAM "::readFrameList() = 0x%x", result);
	return result;
}

#if PRIMEISOCINPUT
/****************************************************************************/
/**
 * @brief	Isochronous IN 転送起動（読み捨て用）．
 *			入力ストリーミング開始時に数フレーム読み捨てる場合に使用される．
 *
 * @param[in]		*pipeToPrime		USB パイプ．
 * @param[in]		bytesPerUSBFrame	USB フレームあたりのバイト数．
 * @param[in]		usbFramesToDelay	読み捨てる USB フレーム数．
 */
// This method starts an input isoc stream to a device and disregards kNumIsocFramesToPrime frames.
#ifdef BUG_FIX
void
PdjUsbAudioStream::primeInputPipe(
	IOUSBHostPipe	*pipeToPrime,
	UInt32			bytesPerUSBFrame,
	UInt32			usbFramesToDelay
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::primeInputPipe()");

	IOReturn		result;
	bool			dataWrittenToPipe	= false;
	UInt32			transactionsToDelay	= usbFramesToDelay * mTransactionsPerUSBFrame;

	FailIf(NULL == pipeToPrime, Exit);
	FailIf(0 == bytesPerUSBFrame, Exit);

	// Make sure the lock delay isn't too large.
	FailIf((bytesPerUSBFrame * usbFramesToDelay) > mSampleBufferSize, Exit);

	mPrimeInputIsocFrames = (IOUSBHostIsochronousFrame *)IOMalloc(transactionsToDelay * sizeof(IOUSBHostIsochronousFrame));
	FailIf(NULL == mPrimeInputIsocFrames, Exit);

	mPrimeInputCompletion.owner		= (void *)this;
	mPrimeInputCompletion.action	= primeInputPipeHandler;
	mPrimeInputCompletion.parameter	= (void *)(uintptr_t)transactionsToDelay;		// So we know how many frames to free in the completion

	// Initialize isoc frame list
	for (UInt32 i = 0; i < transactionsToDelay; i++)
	{
		mPrimeInputIsocFrames[i].status			= -1;
		mPrimeInputIsocFrames[i].requestCount	= bytesPerUSBFrame;
		mPrimeInputIsocFrames[i].completeCount	= 0;
	}

	// attempt to read from USB pipe
	result = pipeToPrime->io(
								mUSBBufferDescriptor,
								mPrimeInputIsocFrames,
								transactionsToDelay,
								mUSBFrameToQueue,
								&mPrimeInputCompletion
						  );
	if (result == kIOReturnSuccess)
	{
		// Our request was queued. We can let the completion handle memory deallocation.
		dataWrittenToPipe = true;

		// Advance the frame to queue by the number of frames primed
		mUSBFrameToQueue += usbFramesToDelay;

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::primeInputPipe(%p, %u) - %u frames primed. mUSBFrameToQueue = %llu",
			pipeToPrime, bytesPerUSBFrame, usbFramesToDelay, mUSBFrameToQueue);
	}

Exit:
	if (!dataWrittenToPipe)
	{
		if (mPrimeInputIsocFrames)
		{
			IOFree(mPrimeInputIsocFrames, transactionsToDelay * sizeof(IOUSBHostIsochronousFrame));
			mPrimeInputIsocFrames = NULL;
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::primeInputPipe()");
}
#else // #ifdef BUG_FIX
void
PdjUsbAudioStream::primeInputPipe(
	IOUSBHostPipe	*pipeToPrime,
	UInt32			bytesPerUSBFrame,
	UInt32			usbFramesToDelay
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::primeInputPipe()");

	// IOUSBHostIsochronousFrame	*isocFrames = NULL;
	IOReturn					result;
	bool						dataWrittenToPipe = false;

	FailIf(NULL == pipeToPrime, Exit);
	FailIf(0 == bytesPerUSBFrame, Exit);

	// Make sure the lock delay isn't too large.
	FailIf((bytesPerUSBFrame * usbFramesToDelay) > mSampleBufferSize, Exit);

	mPrimeInputIsocFrames = (IOUSBHostIsochronousFrame *)IOMalloc(usbFramesToDelay * sizeof(IOUSBHostIsochronousFrame));
	FailIf(NULL == mPrimeInputIsocFrames, Exit);

	mPrimeInputCompletion.owner		= (void *)this;
	mPrimeInputCompletion.action	= primeInputPipeHandler;
	mPrimeInputCompletion.parameter	= (void *)(uintptr_t)usbFramesToDelay;		// So we know how many frames to free in the completion

	// Initialize isoc frame list
	for (UInt32 i = 0; i < usbFramesToDelay; i++)
	{
		mPrimeInputIsocFrames[i].status			= -1;
		mPrimeInputIsocFrames[i].requestCount	= bytesPerUSBFrame;
		mPrimeInputIsocFrames[i].completeCount	= 0;
	}

	// attempt to read from USB pipe
	result = pipeToPrime->io(
								mUSBBufferDescriptor,
								mPrimeInputIsocFrames,
								usbFramesToDelay,
								mUSBFrameToQueue,
								&mPrimeInputCompletion
						  );
	if (result == kIOReturnSuccess)
	{
		// Our request was queued. We can let the completion handle memory deallocation.
		dataWrittenToPipe = true;

		// Advance the frame to queue by the number of frames primed
		mUSBFrameToQueue += usbFramesToDelay;

		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::primeInputPipe(%p, %u) - %u frames primed. mUSBFrameToQueue = %llu",
			pipeToPrime, bytesPerUSBFrame, usbFramesToDelay, mUSBFrameToQueue);
	}

Exit:
	if (!dataWrittenToPipe)
	{
		if (mPrimeInputIsocFrames)
		{
			IOFree(mPrimeInputIsocFrames, usbFramesToDelay * sizeof(IOUSBHostIsochronousFrame));
			mPrimeInputIsocFrames = NULL;
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::primeInputPipe()");
}
#endif // #ifdef BUG_FIX
#endif // #ifdef PRIMEISOCINPUT

/****************************************************************************/
/**
 * @brief	Isochronous OUT 転送完了ルーチン．
 *
 * @param[in]		*object		OUT 転送を起動した PdjUsbAudioStream のインスタンス．
 * @param[in]		*parameter	OUT 転送起動時に設定したパラメータ．出力用バッファの折り返しに関する情報．
 * @param[in]		result		OUT 転送の完了ステータス．
 * @param[in]		*pFrames	OUT 転送した USB フレーム．
 */
void
PdjUsbAudioStream::writeHandler(
	void						*object,
	void						*parameter,
	IOReturn					result,
	IOUSBHostIsochronousFrame	*pFrames
)
{
	DP(DBG_WRITE, "+ " LOGSTRAUDIOSTREAM "::writeHandler()");

	PdjUsbAudioStream			*self;
	AbsoluteTime				time;
	UInt64						curUSBFrameNumber;
	UInt32						byteOffset;
	UInt32						frameIndex;
	UInt32						byteCount;
	UInt32						preWrapBytes;
	SInt64						frameDifference;
	SInt32						expectedFrames;

	self = (PdjUsbAudioStream *)object;
	FailIf(self->mInCompletion, Exit);

	self->mInCompletion = true;

	FailIf(NULL == self->mStreamInterface, Exit);

	curUSBFrameNumber = self->mStreamInterface->getDevice()->getFrameNumber();

	DP(DBG_WRITE, "? " LOGSTRAUDIOSTREAM "::writeHandler() - writeHandler: curUSBFrameNumber = %llu parameter = %p mUSBFrameToQueue = %llu",
		curUSBFrameNumber, parameter, self->mUSBFrameToQueue);

	frameDifference = (SInt64)(self->mUSBFrameToQueue - curUSBFrameNumber);

	expectedFrames = (SInt32)(self->mNumUSBFramesPerList * (self->mNumUSBFrameListsToQueue / 2)) + 1;

	DP(DBG_WRITE, "? " LOGSTRAUDIOSTREAM "::writeHandler() - %llu ?> %d", frameDifference, expectedFrames);

	// This logical expression checks to see if IOUSBFamily fell behind. If so, we don't need to advance the frame list
#ifdef TARGET_PDJ
 	if (frameDifference > expectedFrames)
#else
	if (
			(frameDifference > expectedFrames)
	   &&
			(
				(!(self->mUHCISupport))		// This is not a UHCI connection
			||
				(0 == parameter)			// or this is a wrapping condition for a UHCI connection
			)
	   )
#endif
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::writeHandler() - Not advancing frame list");

		goto Exit;
	}

	if (kIOReturnAborted != result)
	{
		UInt32		numberOfFramesToCheck;

		if (kIOReturnSuccess != result)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::writeHandler() - Frame list %u write returned with error 0x%x",
				self->mCurrentFrameList, result);
		}

#ifdef TARGET_PDJ
 		numberOfFramesToCheck = self->mNumTransactionsPerList;
#else
		numberOfFramesToCheck = (self->mUHCISupport && (UInt32)(uintptr_t)parameter)
									? self->mNumFramesInFirstList
									: self->mNumTransactionsPerList;
#endif

		if ((self->mMasterMode) && (!(self->mHaveTakenFirstTimeStamp)) && (0 == self->mBufferOffset))
		{
			if (pFrames)
			{
				// Check to see if we should take our first time stamp in this frame list.
				for (UInt16 i = 0; i < numberOfFramesToCheck; i++)
				{
					if ((pFrames[i].completeCount) && (!self->mShouldStop))	// <rdar://problem/7378275>
					{
						// We should take our first time stamp here.
						// Here, i represents the first frame in the framelist with a nonzero frActCount, i.e., our first isoc data transfer.
						DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::writeHandler() - Taking first time stamp on frame list %u frame index %u",
							self->mCurrentFrameList, i);
						DP(DBG_INFO, "     pFrames[%d].status = %d",		i, pFrames[i].status);
						DP(DBG_INFO, "     pFrames[%d].requestCount = %u",	i, pFrames[i].requestCount);
						DP(DBG_INFO, "     pFrames[%d].completeCount = %u",	i, pFrames[i].completeCount);
						DP(DBG_INFO, "     pFrames[%d].timeStamp = 0x%llx",	i, __OSAbsoluteTime(pFrames[i].timeStamp));

						time = self->generateTimeStamp(((SInt32)i - 1), 0, 0);		// <rdar://problem/7378275>

						self->takeTimeStamp(false, &time);

						break;
					}
				}
			}
		}

#ifdef DEBUGLOGGING
		// Comb the isoc frame list for alarming statuses.
		if (pFrames)
		{
			for (UInt16 i = 0; i < numberOfFramesToCheck; i++)
			{
				if (
						(kIOReturnSuccess != pFrames[i].status)
				   ||
						(pFrames[i].completeCount != pFrames[i].requestCount)
				   )
				{
					DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "::writeHandler() - pFrames[%d].status = 0x%x", i, pFrames[i].status);
					DP(DBG_WARNING, "     pFrames[%d].requestCount = %u",	i, pFrames[i].requestCount);
					DP(DBG_WARNING, "     pFrames[%d].completeCount = %u",	i, pFrames[i].completeCount);
					DP(DBG_WARNING, "     pFrames[%d].timeStamp = 0x%llx",	i, __OSAbsoluteTime(pFrames[i].timeStamp));
					// IOLog("DJM-V10 USBAudio: ERROR on output! Short packet of size %u encountered when %u bytes were requested.\n",
					//		pFrames[i].completeCount, pFrames[i].requestCount);
				}
			}
		}
#endif

		// skip ahead and see if that helps
		if (self->mUSBFrameToQueue <= curUSBFrameNumber)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::writeHandler - Fell behind! mUSBFrameToQueue = %llu, curUSBFrameNumber = %llu",
				self->mUSBFrameToQueue, curUSBFrameNumber);

			self->mUSBFrameToQueue = curUSBFrameNumber + kMinimumFrameOffset;
		}
	}

	if (0 != parameter)
	{
		if ((self->mMasterMode) && (!self->mShouldStop))	// <rdar://problem/7378275>
		{
			if (pFrames)
			{
				// Take a timestamp
				byteOffset = (UInt32)(uintptr_t)parameter & 0xFFFF;
				// <rdar://6300220> Fixed truncation to 8-bit bug which causes audio output to click/pop as the
				// generated time stamp jitters more than 1ms (HALLab will show a red too early time stamp).

				frameIndex = ((UInt32)(uintptr_t)parameter >> 16) - 1;	// zero-indexed frame index in mCurrentFrameList

				byteCount = pFrames[frameIndex].completeCount;			// number of bytes written

				preWrapBytes = byteCount - byteOffset;					// number of bytes written - wrapped bytes

				time = self->generateTimeStamp(frameIndex, preWrapBytes, byteCount);

				self->takeTimeStamp(true, &time);
			}
		}

#ifdef TARGET_PDJ
#else
		// Now that we've taken the time stamp, if this is UHCI and the first of two writes, we need to exit now.
		// writeHandlerForUHCI will advance the frame list and queue the next write for us. If we do not let the
		// next write be queued by writeHandlerForUHCI, we will get intermittent artifacts after many minutes of
		// streaming.
		if ((self->mUHCISupport) && (frameDifference > expectedFrames - 1))
		{
			// Check to see if we should stop since we're about to skip the normal check
			if (self->mShouldStop > 0)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::writeHandler() - stopping: %u", self->mShouldStop);

				self->mShouldStop++;
			}

			// We're done for now. writeHandlerForUHCI will handle the rest.
			goto Exit;
		} // if this is the first of two UHCI write callbacks
#endif
	} // if we have wrapped

	if (self->mCurrentFrameList == self->mNumUSBFrameLists - 1)
	{
		self->mCurrentFrameList = 0;
	}
	else
	{
		self->mCurrentFrameList++;
	}

	if (self->mShouldStop > 0)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::writeHandler() - stopping: %u", self->mShouldStop);

		self->mShouldStop++;
	}
	else
	{
		// Queue another write
		UInt32	frameListToWrite = (self->mCurrentFrameList - 1) + self->mNumUSBFrameListsToQueue;

		if (frameListToWrite >= self->mNumUSBFrameLists)
		{
			frameListToWrite -= self->mNumUSBFrameLists;
		}

		self->writeFrameList(frameListToWrite);
	}

Exit:
	self->mInCompletion = false;

	DP(DBG_WRITE, "- " LOGSTRAUDIOSTREAM "::writeHandler()");
}

#ifdef TARGET_PDJ
#else
/****************************************************************************/
/**
 * @brief	Isochronous OUT 転送完了ルーチン（UHCI 用）．
 *
 * @param[in]		*object		OUT 転送を起動した PdjUsbAudioStream のインスタンス．
 * @param[in]		*parameter	OUT 転送起動時に設定したパラメータ．出力用バッファの折り返しに関する情報．
 * @param[in]		result		OUT 転送の完了ステータス．
 * @param[in]		*pFrames	OUT 転送した USB フレーム．
 */
void
PdjUsbAudioStream::writeHandlerForUHCI(
	void						*object,
	void						*parameter,
	IOReturn					result,
	IOUSBHostIsochronousFrame	*pFrames
)
{
	DP(DBG_WRITE, "+ " LOGSTRAUDIOSTREAM "::writeHandlerForUHCI()");

	PdjUsbAudioStream			*self;
	UInt64						curUSBFrameNumber;
	UInt32						frameListToWrite;

	self = (PdjUsbAudioStream *)object;
	FailIf(self->mInCompletion, Exit);

	self->mInCompletion = true;

	FailIf(NULL == self->mStreamInterface, Exit);

	curUSBFrameNumber = self->mStreamInterface->getDevice()->getFrameNumber();

	DP(DBG_UHCI, "? " LOGSTRAUDIOSTREAM "[%p]::writeHandlerForUHCI() - writeHandlerForUHCI: curUSBFrameNumber = %llu parameter = %p mUSBFrameToQueue = %llu",
		self, curUSBFrameNumber, parameter, self->mUSBFrameToQueue);

	DP(DBG_UHCI, "? " LOGSTRAUDIOSTREAM "[%p]::writeHandlerForUHCI() - %lld ?> %d",
		self,
		(SInt64)(self->mUSBFrameToQueue - curUSBFrameNumber),
		(SInt32)(self->mNumUSBFramesPerList * (self->mNumUSBFrameListsToQueue / 2)) + 1);

	// This logical expression checks to see if IOUSBFamily fell behind. If so, we don't need to advance the frame list

	if (kIOReturnAborted != result)
	{
		if (kIOReturnSuccess != result)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::writeHandlerForUHCI() - Frame list %u (split for UHCI) write returned with error 0x%x",
				self->mCurrentFrameList, result);
		}

#ifdef DEBUGLOGGING
		UInt32		numberOfFramesToCheck;

		// Comb the isoc frame list for alarming statuses.
		numberOfFramesToCheck = self->mUHCISupport
									? (self->mNumTransactionsPerList - self->mNumFramesInFirstList)
									: (self->mNumTransactionsPerList);

		if (pFrames)
		{
			for (UInt16 i = 0; i < numberOfFramesToCheck; i++)
			{
				if (
						(kIOReturnSuccess != pFrames[i].status)
				   ||
						(pFrames[i].completeCount != pFrames[i].requestCount)
				   )
				{
					DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "::writeHandlerForUHCI() - pFrames[%d].status = 0x%x ", i, pFrames[i].status);
					DP(DBG_WARNING, "     pFrames[%d].requestCount = %u",	i, pFrames[i].requestCount);
					DP(DBG_WARNING, "     pFrames[%d].completeCount = %u",	i, pFrames[i].completeCount);
					DP(DBG_WARNING, "     pFrames[%d].timeStamp = 0x%llx",	i, __OSAbsoluteTime(pFrames[i].timeStamp));
					// IOLog("DJM-V10 USBAudio: ERROR on output! Short packet of size %u encountered when %u bytes were requested.\n",
					//		pFrames[i].completeCount, pFrames[i].requestCount);
				}
			}
		}
#endif

		// skip ahead and see if that helps
		if (self->mUSBFrameToQueue <= curUSBFrameNumber)
		{
			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::writeHandlerForUHCI() - Fell behind! mUSBFrameToQueue = %llu, curUSBFrameNumber = %llu",
				self, self->mUSBFrameToQueue, curUSBFrameNumber);
			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::writeHandlerForUHCI() - Skipping ahead ...", self);

			self->mUSBFrameToQueue = curUSBFrameNumber + kMinimumFrameOffset;
		}
	}

	// Advance the frame list
	if (self->mCurrentFrameList == self->mNumUSBFrameLists - 1)
	{
		self->mCurrentFrameList = 0;
	}
	else
	{
		self->mCurrentFrameList++;
	}

	// Queue another write if we don't need to stop
	// self->mShouldStop is handled by writeHandler()
	if (0 == self->mShouldStop)
	{
		// Queue another write
		frameListToWrite = (self->mCurrentFrameList - 1) + self->mNumUSBFrameListsToQueue;

		if (frameListToWrite >= self->mNumUSBFrameLists)
		{
			frameListToWrite -= self->mNumUSBFrameLists;
		}

		self->writeFrameList(frameListToWrite);
	}
	else
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::writeHandlerForUHCI() - Halting.", self);
	}

Exit:
	self->mInCompletion = false;

	DP(DBG_WRITE, "- " LOGSTRAUDIOSTREAM "::writeHandlerForUHCI()");
}
#endif

/****************************************************************************/
/**
 * @brief	Isochronous IN 転送完了ルーチン．
 *
 * @param[in]		*object		IN 転送を起動した PdjUsbAudioStream のインスタンス．
 * @param[in]		*parameter	IN 転送起動時に設定したパラメータ．USB フレームリストのインデックス．
 * @param[in]		result		IN 転送の完了ステータス．
 * @param[in]		*pFrames	IN 転送した USB フレーム．
 */
void
PdjUsbAudioStream::readHandler(
	void						*object,
	void						*parameter,
	IOReturn					result,
	IOUSBHostIsochronousFrame	*pFrames
)
{
	DP(DBG_READ, "+ " LOGSTRAUDIOSTREAM "::readHandler()");

	PdjUsbAudioStream			*self;
	UInt64						currentUSBFrameNumber;
	UInt32						thisActCount = 0;
	UInt32						minimumUSBFrameSize = 0;
	IOReturn					thisStatus = 0;
	bool						flagOverrun;
	UInt32						frameListToRead;

	self = (PdjUsbAudioStream *)object;
	FailIf(self->mInCompletion, Exit);

	self->mInCompletion = true;

	if (
			(self->mUSBAudioDevice)
	   &&
			(!self->mUSBAudioDevice->getSingleSampleRateDevice())	// We didn't know this was a single sample rate device at this time
	   &&
			(kIOReturnOverrun == result)							// This is what IOUSBFamily should be reporting on an overrun
	   )
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::readHandler() - Encountered fatal error 0x%x on frame list %u (frReqCount = %u).",
			result, self->mCurrentFrameList, pFrames->requestCount);

		// [rdar://5417631] Drop a once-per-attach log message that the device is generating overruns.
		if (!self->mGeneratesOverruns)
		{
			IOLog("WARNING: DJM-V10 USBAudio has detected that a connected USB audio device is sending too much audio data.\n");
			IOLog("WARNING: This USB audio device may not function properly. Please notify the device manufacturer.\n");

			self->mGeneratesOverruns = true;
		}

		flagOverrun = true;

		// [rdar://4456484] If every frame in this frame list generated an overrun, we may need to take drastic measures.
		if (pFrames)
		{
			for (UInt16 frameIndex = 0; frameIndex < self->mNumTransactionsPerList; frameIndex++)
			{
				if (kIOReturnOverrun != pFrames[frameIndex].status)
				{
					flagOverrun = false;

					break;
				}
			}
		}

		if (flagOverrun)
		{
			// This is a fatal error.
			// Notify the PdjUsbAudioDevice to sync the sample rates when possible if this device has two streaming interfaces.
			self->mUSBAudioDevice->setShouldSyncSampleRates(self->mUSBAudioEngine);

			goto Exit;	// [rdar://5889101]
		}
	}

	FailIf(NULL == self->mStreamInterface, Exit);

	currentUSBFrameNumber = self->mStreamInterface->getDevice()->getFrameNumber();

	if (kIOReturnAborted != result)
	{
#if 1	// enabled for [3091812]
#ifdef BUG_FIX
		if (
				(0 == self->mShouldStop)
		   &&
				(
					(SInt32)(self->mUSBFrameToQueue - currentUSBFrameNumber)
				>
					(SInt32)(self->mNumUSBFramesPerList * self->mNumUSBFrameListsToQueue)
				)
		   )
#else
		if (
				(0 == self->mShouldStop)
		   &&
				(
					(SInt32)(self->mUSBFrameToQueue - currentUSBFrameNumber)
				>
					(SInt32)(self->mNumUSBFramesPerList * (self->mNumUSBFrameListsToQueue - 1))
				)
		   )
#endif
		{
			// The frame list that this would have queued has already been queued by convertInputSamples
			DP(DBG_WARNING, "Not queuing a frame list in readHandler (%d)", (SInt32)(self->mUSBFrameToQueue - currentUSBFrameNumber));

			goto Exit;
		}
#endif

		if (pFrames)
		{
#ifdef TARGET_PDJ
 			UInt16	*list = self->mUSBAudioEngine->getInputSizeList(true, self->mNumTransactionsPerList);
#endif
			// Comb the returned statuses for alarming statuses
			for (UInt8 frameIndex = 0; frameIndex < self->mNumTransactionsPerList; frameIndex++)
			{
				thisStatus	 = pFrames[frameIndex].status;
				thisActCount = pFrames[frameIndex].completeCount;

#ifdef TARGET_PDJ
 				list[frameIndex] = thisActCount / self->mSampleSize;
#endif

				// [rdar://5355808] [rdar://5889101]
				minimumUSBFrameSize = self->mAverageFrameSize - (2 * self->mSampleSize);
#ifdef DEBUGLOGGING
				if (
						(!(self->mShouldStop))
				   &&
						(thisStatus != kIOReturnSuccess)
				   &&
						(
							(thisStatus != kIOReturnUnderrun)
						||
							(
								(thisStatus == kIOReturnUnderrun)
							&&
								(thisActCount < minimumUSBFrameSize)
							)
						)
				   )
				{
					DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::readHandler() - Frame list %u frame index %u returned error 0x%x (frActCount = %u, result = 0x%x)",
						self->mCurrentFrameList, frameIndex, thisStatus, thisActCount, result);
				}
#endif

				if (thisActCount < minimumUSBFrameSize)
				{
					// IOLog("DJM-V10 USBAudio: ERROR on input! Short packet of size %u encountered when %u bytes were requested.\n",
					//		thisActCount, pFrames[frameIndex].requestCount);
				}

				if (kIOReturnNotResponding == thisStatus)
				{
					if ((self->mUSBAudioDevice) && (!self->mUSBAudioDevice->recoveryRequested()))
					{
						self->mUSBAudioDevice->requestDeviceRecovery();
					}
				}
			}
		}
	}

	if ((kIOReturnSuccess != result) && (kIOReturnAborted != result) && (kIOReturnUnderrun != result))
	{
		// skip ahead and see if that helps
		if (self->mUSBFrameToQueue <= currentUSBFrameNumber)
		{
			self->mUSBFrameToQueue = currentUSBFrameNumber + kMinimumFrameOffset;
		}
	}

	if (kIOReturnAborted != result)
	{
		self->coalesceInputSamples(0, pFrames);
	}

	if (self->mShouldStop > 0)
	{
		if ((self->mShouldStop == 1) || (self->mShouldStop == self->mNumUSBFrameListsToQueue))
		{
			// Only really care about the first and last stopped frame lists.
			DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::readHandler() - stopping: %u", self->mShouldStop);
		}

		self->mShouldStop++;
	}
	else if (kIOReturnAborted != result)
	{
		// <rdar://7568547> Acquire the lock when updating the mCurrentFrameList value so that it is
		// consistent when accessed in coalesceInputSamples().
		if (self->mCoalescenceMutex)
		{
			IORecursiveLockLock(self->mCoalescenceMutex);
		}

		if (self->mCurrentFrameList == self->mNumUSBFrameLists - 1)
		{
			self->mCurrentFrameList = 0;
		}
		else
		{
			self->mCurrentFrameList++;
		}

		// <rdar://7568547> Release the lock.
		if (self->mCoalescenceMutex)
		{
			IORecursiveLockUnlock(self->mCoalescenceMutex);
		}

		frameListToRead = (self->mCurrentFrameList - 1) + self->mNumUSBFrameListsToQueue;

		if (frameListToRead >= self->mNumUSBFrameLists)
		{
			frameListToRead -= self->mNumUSBFrameLists;
		}

		self->readFrameList(frameListToRead);
	}

Exit:
	self->mInCompletion = false;

	DP(DBG_READ, "- " LOGSTRAUDIOSTREAM "::readHandler()");
}

#if PRIMEISOCINPUT
/****************************************************************************/
/**
 * @brief	Isochronous IN 転送完了ルーチン（読み捨て用）．
 *			入力ストリーミング開始時に数フレーム読み捨てる場合に使用される．
 *
 * @param[in]		*object		IN 転送を起動した PdjUsbAudioStream のインスタンス．
 * @param[in]		*parameter	IN 転送起動時に設定したパラメータ．読み捨てる USB フレーム数．
 * @param[in]		result		IN 転送の完了ステータス．
 * @param[in]		*pFrames	IN 転送した USB フレーム．
 */
void
PdjUsbAudioStream::primeInputPipeHandler(
	void						*object,
	void						*parameter,
	IOReturn					result,
	IOUSBHostIsochronousFrame	*pFrames
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::primeInputPipeHandler(%p, %lu, 0x%x)",
		object, (uintptr_t)parameter, result);

	PdjUsbAudioStream			*self;
	UInt32						transactionsToDelay;

	self = (PdjUsbAudioStream *)object;
	FailIf(NULL == self, Exit);

	// If any analysis of primed input is required, insert it after here.

	// If any analysis of primed input is required, insert it before here.

Exit:
	if (self)
	{
		if (self->mPrimeInputIsocFrames)
		{
			transactionsToDelay = (UInt32)(uintptr_t)(parameter);

			IOFree(self->mPrimeInputIsocFrames, transactionsToDelay * sizeof(IOUSBHostIsochronousFrame));

			self->mPrimeInputIsocFrames = NULL;
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::primeInputPipeHandler()");
}
#endif

/****************************************************************************/
/**
 * @brief	フィードバックエンドポイントの IN 転送完了ルーチン．
 *
 * @param[in]		*target		IN 転送を起動した PdjUsbAudioStream のインスタンス．
 * @param[in]		*parameter	IN 転送起動時に設定したパラメータ．未使用．
 * @param[in]		result		IN 転送の完了ステータス．
 * @param[in]		*pFrames	IN 転送した USB フレーム．
 */
/*
	The purpose of this function is to deal with asynchronous synchronization of isochronous output streams.
	On devices that can lock their output clock to an external source, they can report that value to the driver
	so that the driver doesn't feed data too quickly or too slowly to the device (so that the device's FIFO's
	don't overrun or underrun).

	The device returns a 10.14 unsigned fixed point value in a 24 bit result or a 16.16 unsigned fixed point
	value in a 32 bit result. This value says how many samples per frame (or microframe) the device wants for the current
	sampling period. The device reports the current sampling period in its feedback/synch endpoint, which
	can be retrieved with the GetIsocAssociatedEndpointRefreshInt call (interpreted as 2^(10-x) frames where
	x is the value returned by GetIsocAssociatedEndpointRefreshInt).

	The endpoint should not be read from more often than once every 2^(10-x) frames as the number isn't updated
	by the device any more often than that. Because x can range from 1 to 9, the sample rate may need to be
	adjusted anywhere from once every 2 frames to once every 512 frames.

	If the number returned is larger than the last number returned, the device needs more data, if it is smaller
	than the previous value, the device needs less data.

	In typical usage, the value should not change by a large value (less than 1% of the clock value).
	A typical result would be a value of 0x0b0667 which in 10.14 is 44.10004. This is represented in the
	driver as 0x2c199c which is the 16.16 value for 44.10004.

	See this radar for more details: <rdar://problem/6954295> Store Async feedback in samples per frame/microframe as
	a 16.16 fixed point number
*/
void
PdjUsbAudioStream::sampleRateHandler(
	void						*target,
	void						*parameter,
	IOReturn					result,
	IOUSBHostIsochronousFrame	*pFrames
)
{
	DP(DBG_FEEDBACK, "+ " LOGSTRAUDIOSTREAM "::sampleRateHandler()");

	PdjUsbAudioStream		*self;
	UInt32					requestedSamplesPerFrame;		// sample frames requested per frame or microframe <rdar://problem/6954295>
	IOAudioSamplesPerFrame	oldSamplesPerFrame;				// sample frames per frame or microframe <rdar://problem/6954295>
	IOAudioSamplesPerFrame	newSamplesPerFrame;				// sample frames per frame or microframe <rdar://problem/6954295>
	UInt32					framesToAdvance = 0;
	IOReturn				readStatus = kIOReturnError;
	UInt32					sampleRateBuffer;				// <rdar://problem/6954295>

	self = (PdjUsbAudioStream *)target;

	if ((pFrames) && ((kIOReturnSuccess == result) || (kIOReturnUnderrun == result)))
	{
		// <rdar://problem/6954295>
		sampleRateBuffer = *(self->mAverageSampleRateBuffer);

		requestedSamplesPerFrame = USBToHostLong(sampleRateBuffer);

		oldSamplesPerFrame = self->mSamplesPerPacket;

		switch (pFrames->completeCount)
		{
			case kFixedPoint10_14ByteSize:
				// Assign 10.14 value to an IOAudioSamplesPerFrame struct <rdar://problem/6954295>
				requestedSamplesPerFrame	= requestedSamplesPerFrame << 2;

				newSamplesPerFrame.whole	= requestedSamplesPerFrame >> 16;
				newSamplesPerFrame.fraction	= (requestedSamplesPerFrame & 0x0000FFFF) * 1000;	// fraction is stored x 1000 to maintain precision
				break;

			case kFixedPoint16_16ByteSize:
				// Assign 16.16 value to an IOAudioSamplesPerFrame struct <rdar://problem/6954295>
				newSamplesPerFrame.whole	= requestedSamplesPerFrame >> 16;
				newSamplesPerFrame.fraction	= (requestedSamplesPerFrame & 0x0000FFFF) * 1000;	// fraction is stored x 1000 to maintain precision
				break;

			default:
				// We shouldn't get here. Set newSamplesPerFrame to 0 so that next check fails. <rdar://problem/6954295>
				newSamplesPerFrame.whole	= 0;
				newSamplesPerFrame.fraction	= 0;
		}

		// <rdar://problem/6954295>
		if (
				(newSamplesPerFrame.whole != 0)
		   &&
				(
					(newSamplesPerFrame.whole != oldSamplesPerFrame.whole)
				||
					(newSamplesPerFrame.fraction != oldSamplesPerFrame.fraction)
				)
		   )
		{
			// Need to make sure this sample rate isn't way out of the ballpark
			// i.e. each frame/microframe cannot vary by more than +/- one sample <rdar://problem/6954295>
			if (
					(newSamplesPerFrame.whole > (oldSamplesPerFrame.whole + 1))
			   ||
					(newSamplesPerFrame.whole < (oldSamplesPerFrame.whole - 1))
			   )
			{
				DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "::sampleRateHandler() - ignoring sample rate %u as out of bounds",
					self->getRateFromSamplesPerPacket(newSamplesPerFrame));
			}
			else
			{
				// The device has changed the sample rate that it needs, let's roll with the new sample rate <rdar://problem/6954295>
				self->mSamplesPerPacket = newSamplesPerFrame;

				DP(DBG_FEEDBACK, "? " LOGSTRAUDIOSTREAM "::sampleRateHandler() - Sample rate changed, requestedFrameRate: %u mSamplesPerPacket: %u %u\n",
					self->getRateFromSamplesPerPacket(self->mSamplesPerPacket), self->mSamplesPerPacket.whole, self->mSamplesPerPacket.fraction);
			}
		}

		DP(DBG_FEEDBACK, "? " LOGSTRAUDIOSTREAM "::sampleRateHandler() - currentFrameRate: %u mSamplesPerPacket: %u %u\n",
			self->getRateFromSamplesPerPacket(self->mSamplesPerPacket), self->mSamplesPerPacket.whole, self->mSamplesPerPacket.fraction);
	}
	else
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::sampleRateHandler() - ignoring isoc result due to error 0x%x", result);

		if (pFrames)
		{
			DP(DBG_INFO, "    pFrames->requestCount = %d", pFrames->requestCount);
			DP(DBG_INFO, "    pFrames->completeCount = %d", pFrames->completeCount);
			DP(DBG_INFO, "    pFrames->status = 0x%x", pFrames->status);
		}
		else
		{
			DP(DBG_INFO, "    pFrames = NULL");
		}
	}

	if (0 == self->mShouldStop)
	{
		// Have to reset these parameters otherwise the read doesn't happen.
		self->mSampleRateFrame.status		 = -1;
		self->mSampleRateFrame.requestCount	 = self->mFeedbackPacketSize;
		self->mSampleRateFrame.completeCount = 0;

		framesToAdvance = (1 << self->mRefreshInterval);

		// Due to limitations in IOUSBFamily, the earliest this isochronous transaction can be scheduled is kMinimumFrameOffset frames in the future.
		UInt32		minimumFrameOffset = kMinimumFrameOffset;

		while (framesToAdvance < minimumFrameOffset)
		{
			// Keep raising the power
			framesToAdvance *= 2;
		}

		if (NULL != self->mAssociatedPipe)
		{
			while ((kIOReturnSuccess != readStatus) && (framesToAdvance <= kMaxFeedbackPollingInterval))
			{
				self->mNextSyncReadFrame += framesToAdvance;

				readStatus = self->mAssociatedPipe->io(
														self->mAssociatedEndpointMemoryDescriptor,
														&(self->mSampleRateFrame),
														1,
														self->mNextSyncReadFrame,
														&(self->mSampleRateCompletion)
													);
				if (kIOReturnSuccess != readStatus)
				{
					DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::sampleRateHandler() - framesToAdvance = %u, mNextSyncReadFrame = %llu, readStatus = 0x%x",
						framesToAdvance, self->mNextSyncReadFrame, readStatus);

					self->mNextSyncReadFrame -= framesToAdvance;

					framesToAdvance *= 2;
				}
			}

			if (kIOReturnSuccess != readStatus)
			{
				DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::sampleRateHandler() - Could not queue feedback endpoint isoc request. Feedback request chain is halted!");
				DP(DBG_ERROR, "  mRefreshInterval = %u, framesToAdvance = %u, mNextSyncReadFrame = %llu, readStatus = 0x%x",
					self->mRefreshInterval, framesToAdvance, self->mNextSyncReadFrame, readStatus);
			}
		}
	}
	else
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "::sampleRateHandler() - Stopping feedback chain because stream is stopping.");
	}

	// self->release();

	DP(DBG_FEEDBACK, "- " LOGSTRAUDIOSTREAM "::sampleRateHandler()");
}

/****************************************************************************/
/**
 * @brief	タイムスタンプを生成．
 *			オーディオリングバッファが末尾に達し、先頭に折り返したときの時刻を計算する．
 *
 * @param[in]		transactionIndex	折り返しが発生したときの USB 転送番号．
 * @param[in]		preWrapBytes		折り返しが発生した転送のうち、バッファの末尾から転送したバイト数．
 * @param[in]		byteCount			折り返しが発生した転送の総バイト数．
 *
 * @return			生成したタイムスタンプ．
 */
// <rdar://problem/6354240> TimeStamp calculation is incorrect when there is more than one transaction per USB frame
// <rdar://problem/7378275> Improved timestamp generation accuracy
AbsoluteTime
PdjUsbAudioStream::generateTimeStamp(
	SInt32			transactionIndex,
	UInt32			preWrapBytes,
	UInt32			byteCount
)
{
	DP(DBG_TIMESTAMP, "+ " LOGSTRAUDIOSTREAM "::generateTimeStamp()");

	UInt64			raw_time_nanos				= 0ull;
	UInt64			filtered_time_nanos			= 0ull;
	UInt64			referenceWallTime_nanos		= 0ull;
	AbsoluteTime	anchorTime;
	AbsoluteTime	time;
	UInt64			usbCycleTime;
	SInt64			rawStampDifference			= 0;
	SInt64			filteredStampDifference		= 0;
	UInt64			thisFrameNum;
	UInt32			divisor;
	UInt32			remainingFullTransactions	= 0ul;
	UInt32			partialFrame = 0;
	UInt32			numOutstandingTransactions	= 0;
	UInt32			numOutStandingUSBFrames		= 0;
#if DEBUGTIMESTAMPS
	UInt64			frameDifference				= 0;
	UInt64			sampleRate					= 0;
	SInt64			stampJitter					= 0;
#endif

	FailIf(NULL == mFrameQueuedForList, Exit);
	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(0 == mTransactionsPerUSBFrame, Exit);

	// In the future, we could remove the increment/decrement adjustments to the numOutstandingTransactions
	// if we fix PrepareWriteFrameList() (or the writeHandler()) to account for pre-wrap bytes, as it is currently done for input
	numOutstandingTransactions = transactionIndex + 1;

	if (0 != preWrapBytes)
	{
		numOutstandingTransactions--;	// <rdar://5192321>

		DP(DBG_TIMESTAMP, "? " LOGSTRAUDIOSTREAM "[%p]::generateTimeStamp() - prewrap not zero, value: %u", this, preWrapBytes);
	}

	numOutStandingUSBFrames = numOutstandingTransactions / mTransactionsPerUSBFrame;

	// <rdar://problem/6328817> Fixed the calculation of remainingFullTransactions that was broken in cases where there is more than one transaction per USB frame.
	remainingFullTransactions = numOutstandingTransactions - (numOutStandingUSBFrames * mTransactionsPerUSBFrame);

	thisFrameNum = mFrameQueuedForList[mCurrentFrameList] + numOutStandingUSBFrames;

	DP(DBG_TIMESTAMP, "? " LOGSTRAUDIOSTREAM "[%p]::generateTimeStamp() - thisFrameNum = %llu, transactionIndex = %d",
			this, thisFrameNum, transactionIndex);
	DP(DBG_TIMESTAMP, "? " LOGSTRAUDIOSTREAM "[%p]::generateTimeStamp() - numOutstandingTransactions = %u, numOutStandingUSBFrames = %u remainingFullTransactions = %u",
			this, numOutstandingTransactions, numOutStandingUSBFrames, remainingFullTransactions);

	FailIf(kIOReturnSuccess != this->copyAnchor(thisFrameNum, &anchorTime, &usbCycleTime), Exit);	// always use this frame as the anchor frame

	// The following code seeks to implement the following equations (though in the code below the original equation is obfuscated by algebra used to defer
	// division for as long as possible to increase precision. Quotient terms involving byteCount and remainingFullTransactions are only included when these
	// respective variables are nonzero.
	//
	//	[ time = anchorTime + wallTimePerUSBCycle * ( remainingFullTransactions   +              preWrapBytes               ) ]
	//	[                                             --------------------------     ---------------------------------------  ]
	//	[                                              transactionsPerUSBFrame       ( transactionsPerUSBFrame * byteCount )  ]
	//

	partialFrame = ((mTransactionsPerUSBFrame != 1) && (byteCount)) ? (byteCount * remainingFullTransactions) : 0;

	divisor = byteCount ? (byteCount * mTransactionsPerUSBFrame) : mTransactionsPerUSBFrame;

	raw_time_nanos = partialFrame + preWrapBytes;

	DP(DBG_TIMESTAMP, "? " LOGSTRAUDIOSTREAM "[%p]::generateTimeStamp() - partialFrame: %u, preWrapBytes: %u usbCycleTime: %llu",
		this, partialFrame, preWrapBytes, usbCycleTime);

	// raw_time_nanos now represents the time at which the byte in question should have begun transfer.
	// In the case of input, we won't have access to this byte until one USB frame later.
	if (super::getDirection() == kIOAudioStreamDirectionInput)
	{
		raw_time_nanos += divisor;
	}

	// [rdar://5178614] Divide this into two operations to prevent roundoff error.
	raw_time_nanos *= usbCycleTime;

	raw_time_nanos /= (kWallTimeExtraPrecision * divisor);

	::absolutetime_to_nanoseconds(anchorTime, &referenceWallTime_nanos);

	raw_time_nanos += referenceWallTime_nanos;

	DP(DBG_TIMESTAMP, "? " LOGSTRAUDIOSTREAM "[%p]::generateTimeStamp() - time_nanos before filter: %llu", this, raw_time_nanos);

	filtered_time_nanos = raw_time_nanos;

	if (0ull != mLastRawTimeStamp_nanos)
	{
#if DEBUGTIMESTAMPS
		if (0ull != mLastWrapFrame)
		{
			frameDifference = thisFrameNum - mLastWrapFrame;
		}
#endif
		rawStampDifference = raw_time_nanos - mLastRawTimeStamp_nanos;

		filteredStampDifference = this->jitterFilter(rawStampDifference, mNumTimeStamp);

		mNumTimeStamp++;

		filtered_time_nanos = mLastFilteredTimeStamp_nanos + filteredStampDifference;

#if DEBUGTIMESTAMPS
		if (0ull != mLastFilteredStampDifference)
		{
			stampJitter = filteredStampDifference - mLastFilteredStampDifference;
		}

		sampleRate = (1000000000ull * (mSampleBufferSize / mSampleSize) * 1000) / filteredStampDifference;

		mStampDrift += stampJitter;

		DP(DBG_TIMESTAMP, "   transactionIndex = %d, remainingFullTransactions = %u, preWrapBytes = %u, byteCount = %u",
			transactionIndex, remainingFullTransactions, preWrapBytes, byteCount);
		DP(DBG_TIMESTAMP, "   frameDifference = %llu, referenceWallTime_nanos = %llu, mWallTimePerUSBCycle = %llu, time = %llu",
			frameDifference, referenceWallTime_nanos, mUSBAudioDevice->mWallTimePerUSBCycle, filtered_time_nanos);
		DP(DBG_TIMESTAMP, "    thisFrameNum = %llu, anchorTime = %llu", thisFrameNum, anchorTime);

		if (super::getDirection() == kIOAudioStreamDirectionInput)
		{
			DP(DBG_TIMESTAMP, "? " LOGSTRAUDIOSTREAM "[%p]::generateTimeStamp(I)   stampDifference = %lld, stampJitter = %lld, mStampDrift = %lld, sampleRate = %llu \n",
				this, filteredStampDifference, stampJitter, mStampDrift, sampleRate);
		}
		else
		{
			DP(DBG_TIMESTAMP, "? " LOGSTRAUDIOSTREAM "[%p]::generateTimeStamp(O)   stampDifference = %lld, stampJitter = %lld, mStampDrift = %lld, sampleRate = %llu \n",
				this, filteredStampDifference, stampJitter, mStampDrift, sampleRate);
		}

		if (MAGNITUDEOF(stampJitter) > 1000000)
		{
			DP(DBG_TIMESTAMP, "\nthisFrameNum = %llu, mFrameQueuedForList = %llu, remainingFullTransactions = %u",
				thisFrameNum, mFrameQueuedForList[mCurrentFrameList], remainingFullTransactions);
		}

		sampleRate = (1000000000ull * (mSampleBufferSize / mSampleSize) * 1000) / rawStampDifference;

		if (super::getDirection() == kIOAudioStreamDirectionInput)
		{
			DP(DBG_TIMESTAMP, "? " LOGSTRAUDIOSTREAM "[%p]::generateTimeStamp(RI)  stampDifference = %lld, sampleRate = %llu \n",
				this, rawStampDifference, sampleRate);
		}
		else
		{
			DP(DBG_TIMESTAMP, "? " LOGSTRAUDIOSTREAM "[%p]::generateTimeStamp(RO)  stampDifference = %lld, sampleRate = %llu \n",
				this, rawStampDifference, sampleRate);
		}
#endif
	}

	mLastRawTimeStamp_nanos		 = raw_time_nanos;
	mLastFilteredTimeStamp_nanos = filtered_time_nanos;

	if (0ull != filteredStampDifference)
	{
		mLastFilteredStampDifference = filteredStampDifference;
	}

	mLastWrapFrame = thisFrameNum;

Exit:

	::nanoseconds_to_absolutetime(filtered_time_nanos, &time);

	DP(DBG_TIMESTAMP, "- " LOGSTRAUDIOSTREAM "::generateTimeStamp()");
	return time;
}

/****************************************************************************/
/**
 * @brief	タイムスタンプのジッタ除去フィルタ．
 *			タップ数 32 の FIR フィルタである．
 *
 * @param[in]		curr			今回の入力値．
 * @param[in]		nIter			フィルタの呼び出し回数．
 *
 * @return			フィルタの出力値．
 */
// <rdar://problem/7378275> Improved timestamp generation accuracy
// 32-tap FIR
// nIter is the iteration number. It should begin at zero and continue increasing (up to the value of nFilterSize)
// If the timestamps are stopped and then restarted, the nIter value should reset to zero to ensure the filter starts up correctly.
UInt64
PdjUsbAudioStream::jitterFilter(
	UInt64			curr,
	UInt32			nIter
)
{
	DP(DBG_TIMESTAMP, "+ " LOGSTRAUDIOSTREAM "::jitterFilter()");

	const UInt64	filterCoefficients[] =
					{  1,  2,  4,  7, 10, 14, 19, 25, 31, 37, 43, 49, 54, 58, 62, 64,
					  64, 64, 62, 58, 54, 49, 43, 37, 31, 25, 19, 14, 10,  7,  4,  2,  1 };

	const UInt64	filterCoefficientsSmall[] =
					{ 256, 256, 256, 256 };

	UInt64			result = 0;

	// On the first iteration, initialise all the data with the first coefficient, otherwise, instert in the circular array
	if (0 == nIter)
	{
		for (UInt32 filterIndex = 0; filterIndex < kMaxFilterSize; filterIndex++)
		{
			mFilterData[filterIndex] = curr;
		}
	}
	else
	{
		mFilterData[mFilterWritePointer] = curr;
	}

	// Calculate filter output - if we are just starting up, use the smaller filter, otherwise use the larger filter with increased attenuation
	if (nIter < kMaxFilterSize)
	{
		for (UInt32 filterIndex = 0; filterIndex < (sizeof(filterCoefficientsSmall) / sizeof(filterCoefficientsSmall[0])); filterIndex++)
		{
			result += filterCoefficientsSmall[filterIndex] * mFilterData[(kMaxFilterSize + mFilterWritePointer - filterIndex) % kMaxFilterSize];
		}
	}
	else
	{
		for (UInt32 filterIndex = 0; filterIndex < kMaxFilterSize; filterIndex++)
		{
			result += filterCoefficients[filterIndex] * mFilterData[(kMaxFilterSize + mFilterWritePointer - filterIndex) % kMaxFilterSize];
		}
	}

	result += kFilterScale / 2;
	result /= kFilterScale;

	// Update the write pointer for the next iteration
	mFilterWritePointer = (kMaxFilterSize + mFilterWritePointer + 1) % kMaxFilterSize;

	DP(DBG_TIMESTAMP, "- " LOGSTRAUDIOSTREAM "::jitterFilter()");
	return result;
}

/****************************************************************************/
/**
 * @brief	基準時刻と基準 USB フレーム番号を取得．
 *			PdjUsbAudioDevice において定期的に作成している
 *			基準時刻と基準 USB フレーム番号の情報をコピーする．
 *
 * @param[in]		anchorFrame		基準 USB フレーム番号．
 * @param[out]		*anchorTime		基準時刻．
 * @param[out]		*usbCycleTime	USB フレーム間隔．
 *
 * @return			IOReturn コード．
 */
// <rdar://problem/7378275> Improved timestamp generation accuracy
IOReturn
PdjUsbAudioStream::copyAnchor(
	UInt64			anchorFrame,
	AbsoluteTime	*anchorTime,
	UInt64			*usbCycleTime
)
{
	DP(DBG_TIMESTAMP, "+ " LOGSTRAUDIOSTREAM "::copyAnchor()");

	IOReturn		result = kIOReturnError;
	UInt64			anchorTime_nanos;

	FailIf(NULL == mUSBAudioDevice, Exit);
	FailIf(NULL == mUSBAudioDevice->mTimeLock, Exit);

	IOLockLock(mUSBAudioDevice->mTimeLock);

	anchorTime_nanos = mUSBAudioDevice->getTimeForFrameNumber(anchorFrame);

	::nanoseconds_to_absolutetime(anchorTime_nanos, anchorTime);

	*usbCycleTime = mUSBAudioDevice->mWallTimePerUSBCycle;

	IOLockUnlock(mUSBAudioDevice->mTimeLock);

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TIMESTAMP, "- " LOGSTRAUDIOSTREAM "::copyAnchor() = 0x%x", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	１パケットあたりのサンプル数を計算．
 *
 * @param[in]		sampleRate					サンプリング周波数．
 * @param[out]		*averageFrameSamples		１パケットあたりのサンプル数．
 * @param[out]		*additionalSampleFrameFreq	端数がある場合の追加サンプル数．
 */
void
PdjUsbAudioStream::calculateSamplesPerPacket(
	UInt32			sampleRate,
	UInt16			*averageFrameSamples,
	UInt16			*additionalSampleFrameFreq
)
{
	DP(DBG_VERBOSE, "+ " LOGSTRAUDIOSTREAM "::calculateSamplesPerPacket()");

	UInt32			divisor;
	UInt32			modulus;

	// [rdar://4801012] For USB 2.0 audio, packets correspond to microframes.

	if (mTransactionsPerUSBFrame)
	{
		modulus = 1000 * mTransactionsPerUSBFrame;
	}
	else
	{
		modulus = 1000;
	}

	*averageFrameSamples = sampleRate / modulus;

	// [rdar://5600254] For UAC 2.0, the following calculation is not useful.
	// There should be no reason to use additionalSampleFrameFreq.
	divisor = sampleRate % modulus;

	if (divisor)
	{
		*additionalSampleFrameFreq = modulus / divisor;
	}
	else
	{
		*additionalSampleFrameFreq = 0;
	}

	// [rdar://5600254] We can log the data cadence here.
#if SHOWCADENCE
	UInt32			transferSampleRate;
	UInt32			currentSampleRate;
	UInt32			remainder;
	UInt32			accumulatedSamples = 0ul;
	UInt16			transfer = 0;
	UInt16			cycleLength = 0;
	UInt8			powerOfTwo = 0;
	UInt8			transactionsPerMS;

	transactionsPerMS = mTransactionsPerUSBFrame;

	while (transactionsPerMS > 1)
	{
		transactionsPerMS >>= 1;

		powerOfTwo++;
	}

	transferSampleRate = sampleRate << (16 - powerOfTwo);

	currentSampleRate  = sampleRate / modulus;
	currentSampleRate *= 1000;
	currentSampleRate <<= 16;

	remainder = transferSampleRate - currentSampleRate;

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::calculateSamplesPerPacket - transferSampleRate = 0x%8X, currentSampleRate = 0x%X",
		this, transferSampleRate, currentSampleRate);

	transfer	= 1;
	cycleLength	= 0;

	while (
				(remainder)
		  &&
				(
					(0 == accumulatedSamples)
				||
					(accumulatedSamples % (1000 << 16))
				)
		  )
	{
		accumulatedSamples += remainder;

		if (accumulatedSamples >= (1000 << 16))
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::calculateSamplesPerPacket - large packet on transfer %d (1-indexed)",
				this , transfer);

			accumulatedSamples -= 1000 << 16;
		}

		transfer++;
		cycleLength++;

		if (!accumulatedSamples)
		{
			break;
		}
	}

	DP(DBG_INFO, "? " LOGSTRAUDIOSTREAM "[%p]::calculateSamplesPerPacket - cycleLength = %d", this, cycleLength);
#endif

	DP(DBG_VERBOSE, "- " LOGSTRAUDIOSTREAM "::calculateSamplesPerPacket()");
}

/****************************************************************************/
/**
 * @brief	１パケットあたりのサンプル数からサンプリング周波数を計算．
 *
 * @param[in]		samplesPerPacket	１パケットあたりのサンプル数．
 *
 * @return			サンプリング周波数．
 */
// <rdar://problem/6954295> Store Async feedback in samples per frame/microframe as a 16.16 fixed point number
// <rdar://problem/7345441> PdjUsbAudio: getRateFromSamplesPerPacket() has a rounding error in sample rate calculation
UInt32
PdjUsbAudioStream::getRateFromSamplesPerPacket(
	IOAudioSamplesPerFrame	samplesPerPacket
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOSTREAM "::getRateFromSamplesPerPacket()");

	UInt64					sampleRate = 0;

	if (0 != samplesPerPacket.whole)
	{
		sampleRate   = samplesPerPacket.whole * 1000;	// multiply by 1000 to get Hz (stored per ms)
		sampleRate <<= 16;
		sampleRate  += samplesPerPacket.fraction;		// stored value already multiplied by 1000 in sampleRateHandler()
	}

	if (0 != mTransactionsPerUSBFrame)
	{
		sampleRate *= mTransactionsPerUSBFrame;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOSTREAM "::getRateFromSamplesPerPacket()");
	return (UInt32)((sampleRate >> 16) & 0x00000000FFFFFFFF);
}

#ifdef TARGET_PDJ
 /****************************************************************************/
 /**
  * @brief	出力データの処理．
  *			IOAudioStream::processOutputSamples() をオーバーライド．
  *
  * @param[in]		*clientBuffer		アプリのバッファ．
  * @param[in]		firstSampleFrame	処理開始位置（サンプルフレーム数）．
  * @param[in]		loopCount			処理開始位置（ループ回数）．
  * @param[in]		samplesAvailable	利用可能な出力サンプルの有無．
  *
  * @return			IOReturn コード．
  */
 IOReturn
 PdjUsbAudioStream::processOutputSamples(
 	IOAudioClientBuffer		*clientBuffer,
 	UInt32					firstSampleFrame,
 	UInt32					loopCount,
 	bool					samplesAvailable
 )
 {
 	if (!samplesAvailable)
 	{
 		mErrorCount++;
 
 		DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "::processOutputSamples() - samplesAvailable == false (%d)", mErrorCount);
 
 		if (mErrorCount >= kMaxErrorCount)
 		{
 			mShouldRestartStreaming = true;
 		}
 	}
 	else
 	{
 		if (mShouldRestartStreaming && mUSBAudioEngine)
 		{
 			mShouldRestartStreaming = false;
 
 			DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::processOutputSamples() - Too many glitches! Restarting...");
 
 			//mUSBAudioEngine->requestStreamingRestart();
 		}
 
 		mErrorCount = 0;
 	}
 
 	if (clientBuffer && (clientBuffer->numChannels == super::format.fNumChannels) && (clientBuffer->numSampleFrames > 0))
 	{
 		UInt32		frameSize = super::format.fNumChannels * (super::format.fBitWidth / 8);
 
 		mNumFramesInClientBuffer = (UInt32)clientBuffer->sourceBufferDescriptor->getLength();
 
 		if (frameSize > 0)
 		{
 			mNumFramesInClientBuffer /= frameSize;
 		}
 
 		return super::processOutputSamples(clientBuffer, firstSampleFrame, loopCount, samplesAvailable);
 	}
 	else
 	{
 		return kIOReturnSuccess;
 	}
 }
 
 /****************************************************************************/
 /**
  * @brief	出力データを浮動小数点型から整数型に変換．
  *			IOAudioStream::clipIfNecessary() をオーバーライド．
  */
 void
 PdjUsbAudioStream::clipIfNecessary(
 	void
 )
 {
 	DP(DBG_WRITE, "+ " LOGSTRAUDIOSTREAM "::clipIfNecessary()");
 
 	if (super::clientBufferListStart)
 	{
 		// Only try to clip if there is not an unmixed buffer
 		if (!IOAUDIOENGINEPOSITION_IS_ZERO(&(super::clientBufferListStart->mixedPosition)))
 		{
 			PdjUsbAudioEngine	*engine = OSDynamicCast(PdjUsbAudioEngine, super::audioEngine);
 			FailIf(NULL == engine, Exit);
 
 			// Check to see if we've clipped any samples yet
 			if (IOAUDIOENGINEPOSITION_IS_ZERO(&(super::clippedPosition)))
 			{
 				super::clippedPosition = super::startingPosition;
 			}
 
 			// Check to see if it is on the same loop as the starting position
 			// If not, adjust it to the same loop
 			if (
 					(
 						(super::clientBufferListStart->mixedPosition.fLoopCount == (super::clippedPosition.fLoopCount + 1))
 					&&
 						(super::clientBufferListStart->mixedPosition.fSampleFrame >= super::clippedPosition.fSampleFrame)
 					)
 				||
 					(super::clientBufferListStart->mixedPosition.fLoopCount > (super::clippedPosition.fLoopCount + 1))
 			)
 			{
 				//<rdar://problem/10305944>
 				DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::clipIfNecessary() - Error: attempting to clip to a position"
 					" more than one buffer ahead of last clip position (%u, %u)->(%u, %u).",
 					this,
 					super::clippedPosition.fLoopCount,
 					super::clippedPosition.fSampleFrame,
 					super::clientBufferListStart->mixedPosition.fLoopCount,
 					super::clientBufferListStart->mixedPosition.fSampleFrame);
 
 				if (super::clientBufferListStart->mixedPosition.fSampleFrame >= super::clippedPosition.fSampleFrame)
 				{
 					super::clippedPosition.fLoopCount = super::clientBufferListStart->mixedPosition.fLoopCount;
 				}
 				else
 				{
 					super::clippedPosition.fLoopCount = super::clientBufferListStart->mixedPosition.fLoopCount - 1;
 				}
 
 				//<rdar://problem/10305944>
 				DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::clipIfNecessary() - adjusting clipped position to (%u, %u)",
 					this,
 					super::clippedPosition.fLoopCount,
 					super::clippedPosition.fSampleFrame);
 			}
 
 			// Add a test to see if we'd be clipping more samples than delivered because the HAL might skip some samples around a loop increment
 			// If the HAL skipped samples around a loop increment, then just start from where it wants to
 			if (
 					(super::clientBufferListStart->mixedPosition.fLoopCount + 1 == super::clippedPosition.fLoopCount)
 				&&
 					(
 						super::clientBufferListStart->numSampleFrames
 					<
 						(engine->getNumSampleFramesPerBuffer() - super::clippedPosition.fSampleFrame)
 					)
 			)
 			{
 				super::clientBufferListStart->mixedPosition.fLoopCount = super::clippedPosition.fLoopCount;
 
 				DP(DBG_WARNING, "! " LOGSTRAUDIOSTREAM "[%p]::clipIfNecessary() - clip position is off %d < %d - %d",
 					this,
 					super::clientBufferListStart->numSampleFrames,
 					engine->getNumSampleFramesPerBuffer(),
 					super::clippedPosition.fSampleFrame);
 			}
 
 			UInt32	numSamplesToClip;	//<rdar://problem/5994776>
 
 			if (super::clientBufferListStart->mixedPosition.fLoopCount == super::clippedPosition.fLoopCount)
 			{
 				if (super::clientBufferListStart->mixedPosition.fSampleFrame > super::clippedPosition.fSampleFrame)
 				{
 					numSamplesToClip = super::clientBufferListStart->mixedPosition.fSampleFrame - super::clippedPosition.fSampleFrame;
 
 					if (   (!super::format.fIsMixable)
 						&& (numSamplesToClip > mNumFramesInClientBuffer))
 					{
 						DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::clipIfNecessary() - Error: mixBuffer overrun (%d > %d)",
 							numSamplesToClip, mNumFramesInClientBuffer);
 					}
 					else
 					{
 						this->clipOutputSamples(super::clippedPosition.fSampleFrame, numSamplesToClip);
 					}
 
 					super::clippedPosition.fSampleFrame = super::clientBufferListStart->mixedPosition.fSampleFrame;
 				}
 				else if (super::clientBufferListStart->mixedPosition.fSampleFrame < super::clippedPosition.fSampleFrame)
 				{
 					//<rdar://problem/10305944>
 					DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "[%p]::clipIfNecessary() - Error: already clipped to a position (%u, %u)"
 						" past data to be clipped (%u, %u) - data ignored.",
 						this,
 						super::clippedPosition.fLoopCount,
 						super::clippedPosition.fSampleFrame,
 						super::clientBufferListStart->mixedPosition.fLoopCount,
 						super::clientBufferListStart->mixedPosition.fSampleFrame);
 				}
 			}
 			else	// Clip wraps around
 			{
 				UInt32	numSampleFramesPerBuffer;
 
 				numSampleFramesPerBuffer = engine->getNumSampleFramesPerBuffer();
 
 				numSamplesToClip = numSampleFramesPerBuffer - super::clippedPosition.fSampleFrame;
 
 				if (   (!super::format.fIsMixable)
 					&& (numSamplesToClip > mNumFramesInClientBuffer))
 				{
 					DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::clipIfNecessary() - Error: mixBuffer overrun (%d > %d)",
 						numSamplesToClip, mNumFramesInClientBuffer);
 				}
 				else
 				{
 					this->clipOutputSamples(super::clippedPosition.fSampleFrame, numSamplesToClip);
 				}
 
 				if (!super::format.fIsMixable)
 				{
 					//<rdar://problem/5994776>
 					UInt32	remainingSamplesToClip = (numSampleFramesPerBuffer - super::clippedPosition.fSampleFrame);
 
 					// Move the mix buffer to where we left off because the clip routine always starts at the beginning of the source buffer,
 					// but that's not the right place when we don't have a source buffer and are using the mixbuffer as a pseduo-source buffer.
 					DP(DBG_WRITE, "! " LOGSTRAUDIOSTREAM "[%p]::clipIfNecessary() clipOutputSamples wrap:"
 						" mixBuffer = %p, remainingSamplesToClip = %u, clientBufferListStart->mixedPosition.fSampleFrame = %u",
 						this,
 						super::mixBuffer,
 						remainingSamplesToClip,
 						super::clientBufferListStart->mixedPosition.fSampleFrame);
 
 					super::mixBuffer = (char *)super::mixBuffer + (remainingSamplesToClip * super::format.fNumChannels * (super::format.fBitWidth / 8));
 
 					if ((remainingSamplesToClip + super::clientBufferListStart->mixedPosition.fSampleFrame) > mNumFramesInClientBuffer)
 					{
 						DP(DBG_ERROR, "! " LOGSTRAUDIOSTREAM "::clipIfNecessary() - Error: mixBuffer overrun (%d > %d)",
 							(remainingSamplesToClip + super::clientBufferListStart->mixedPosition.fSampleFrame), mNumFramesInClientBuffer);
 					}
 					else
 					{
 						this->clipOutputSamples(0, super::clientBufferListStart->mixedPosition.fSampleFrame);
 					}
 				}
 				else
 				{
 					this->clipOutputSamples(0, super::clientBufferListStart->mixedPosition.fSampleFrame);
 				}
 
 				super::clippedPosition = super::clientBufferListStart->mixedPosition;
 			}
 		}
 	}
 
 Exit:
 	DP(DBG_WRITE, "- " LOGSTRAUDIOSTREAM "::clipIfNecessary()");
 	return;
 }
#endif
