/*
 * Copyright (c) 1998-2009 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/*
	The classes contained herein abstract a USB audio class configuartion descriptor.

	A USB device is described by a section of length-formatted memory on the device.
	It is arranged as a list starting with the device descriptor and then followed by
	one or more configuration descriptors, one or more stream descriptors, and one or
	more HID descriptors. It is possible for the configuration and stream interface
	descriptors to come in any order, not necessarily all config descriptors *THEN*
	all stream interface descriptors.

	These objects only parse configuration and interface descriptors. Code calling
	ther parser must pass it configuration descriptor pointer obtained from the USB
	services API via GetFullConfigurationDescriptor().

	PdjUsbConfigurationDictionary is the entry point for accessing all of this stored data.
	Directly calling any of the other objects is unnecessary. First allocate a
	PdjUsbConfigurationDictionary and call its init function which will do the parsing of
	the configuration descriptor and create the other objects as necessary. Once the
	PdjUsbConfigurationDictionary has been init'ed, call getNumStreamInterfaces,
	GetNumAltStreamInterfaces, and getFirstStreamInterfaceNum so each interface can be
	queried for its properties. Once a suitable interface has been found, configuring
	the interface with the normal USB calls can be conducted with ease.

	PdjUsbConfigurationDictionary contains arrays of control and stream interfaces. As the
	configuration descriptor is parsed, new control or stream dictionaries are created
	and inserted into the appropriate arrays.
*/

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioDictionary.cpp
 *
 * @brief	OSDictionary クラスの派生クラス群．
 *			USB Audio Class のディスクリプタを表現する一連のクラス．
 *			PdjUsbAudioDevice::protectedInitHardware() においてインスタンス化される．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/

/*------------------------------------------*/
/*				インクルード				*/
/*------------------------------------------*/

#include "PdjUsbAudioCommon.h"

/*------------------------------------------*/
/*		PdjUsbAudioDictionary クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbAudioDictionary -

OSDefineMetaClassAndStructors(PdjUsbAudioDictionary, super)

/****************************************************************************/
/**
 * @brief	Dictionary から bool 値を取得．
 *
 * @param[in]		*key			キー．
 * @param[out]		*value			取得した bool 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDictionary::getDictionaryValue(
	const char		*key,
	bool			*value
)
{
	IOReturn		result = kIOReturnError;
	OSObject		*dictionaryObject	= NULL;
	OSBoolean		*dictionaryBoolean	= NULL;

	FailIf(NULL == value, Exit);
	FailIf(NULL == (dictionaryObject = super::getObject(key)), Exit);
	FailIf(NULL == (dictionaryBoolean = OSDynamicCast(OSBoolean, dictionaryObject)), Exit);

	*value = dictionaryBoolean->getValue();

	result = kIOReturnSuccess;

Exit:
	if ((value) && (kIOReturnSuccess != result))
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIODICTIONARY "[%p]::getDictionaryValue(%s, %p = %d) = 0x%x",
			this, key, value, *value, result);
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	Dictionary から UInt8 値を取得．
 *
 * @param[in]		*key			キー．
 * @param[out]		*value			取得した UInt8 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDictionary::getDictionaryValue(
	const char		*key,
	UInt8			*value
)
{
	IOReturn		result = kIOReturnError;
	OSObject		*dictionaryObject = NULL;
	OSNumber		*dictionaryNumber = NULL;

	FailIf(NULL == value, Exit);
	FailIf(NULL == (dictionaryObject = super::getObject(key)), Exit);
	FailIf(NULL == (dictionaryNumber = OSDynamicCast(OSNumber, dictionaryObject)), Exit);

	*value = dictionaryNumber->unsigned8BitValue();

	result = kIOReturnSuccess;

Exit:
	if ((value) && (kIOReturnSuccess != result))
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIODICTIONARY "[%p]::getDictionaryValue(%s, %p) = 0x%x",
			this, key, value, result);
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	Dictionary から UInt16 値を取得．
 *
 * @param[in]		*key			キー．
 * @param[out]		*value			取得した UInt16 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDictionary::getDictionaryValue(
	const char		*key,
	UInt16			*value
)
{
	IOReturn		result = kIOReturnError;
	OSObject		*dictionaryObject = NULL;
	OSNumber		*dictionaryNumber = NULL;

	FailIf(NULL == value, Exit);
	FailIf(NULL == (dictionaryObject = super::getObject(key)), Exit);
	FailIf(NULL == (dictionaryNumber = OSDynamicCast(OSNumber, dictionaryObject)), Exit);

	*value = dictionaryNumber->unsigned16BitValue();

	result = kIOReturnSuccess;

Exit:
	if ((value) && (kIOReturnSuccess != result))
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIODICTIONARY "[%p]::getDictionaryValue(%s, %p = %d) = 0x%x",
			this, key, value, *value, result);
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	Dictionary から UInt32 値を取得．
 *
 * @param[in]		*key			キー．
 * @param[out]		*value			取得した UInt32 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDictionary::getDictionaryValue(
	const char		*key,
	UInt32			*value
)
{
	IOReturn		result = kIOReturnError;
	OSObject		*dictionaryObject = NULL;
	OSNumber		*dictionaryNumber = NULL;

	FailIf(NULL == value, Exit);
	FailIf(NULL == (dictionaryObject = super::getObject(key)), Exit);
	FailIf(NULL == (dictionaryNumber = OSDynamicCast(OSNumber, dictionaryObject)), Exit);

	*value = dictionaryNumber->unsigned32BitValue();

	result = kIOReturnSuccess;

Exit:
	if ((value) && (kIOReturnSuccess != result))
	{
		DP(DBG_ERROR, "! " LOGSTRAUDIODICTIONARY "[%p]::getDictionaryValue(%s, %p = %u) = 0x%x",
			this, key, value, *value, result);
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	Dictionary から OSArray を取得．
 *
 * @param[in]		*key			キー．
 *
 * @return			取得した OSArray．
 */
OSArray *
PdjUsbAudioDictionary::getDictionaryArray(
	const char		*key
)
{
	OSObject		*dictionaryObject	= NULL;
	OSArray			*dictionaryArray	= NULL;

	dictionaryObject = super::getObject(key);

	if (dictionaryObject)
	{
		dictionaryArray = OSDynamicCast(OSArray, dictionaryObject);
	}

	// DP(DBG_INFO, "? " LOGSTRAUDIODICTIONARY "[%p]::getDictionaryArray(%s) = %p", this, key, dictionaryArray);
	return dictionaryArray;
}

/****************************************************************************/
/**
 * @brief	Dictionary に OSObject を設定し、OSObject をリリース．
 *
 * @param[in]		*key			キー．
 * @param[in]		*object			設定する OSObject．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDictionary::setDictionaryObjectAndRelease(
	const char		*key,
	OSObject		*object
)
{
	IOReturn		result = kIOReturnError;

	FailIf(NULL == object, Exit);
	FailIf(!super::setObject(key, object), Exit);

	result = kIOReturnSuccess;

Exit:
	if (object)
	{
		object->release();
	}

	DP(DBG_INFO, "? " LOGSTRAUDIODICTIONARY "[%p]::setDictionaryObjectAndRelease(%s, %p) = 0x%x",
		this, key, object, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	Dictionary に bool 値を設定．
 *
 * @param[in]		*key			キー．
 * @param[in]		value			bool 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDictionary::setDictionaryValue(
	const char		*key,
	bool			value
)
{
	IOReturn		result = kIOReturnError;
	OSBoolean		*dictionaryBoolean = NULL;

	FailIf(NULL == (dictionaryBoolean = OSBoolean::withBoolean(value)), Exit);
	FailIf(!super::setObject(key, dictionaryBoolean), Exit);

	dictionaryBoolean->release();
	dictionaryBoolean = NULL;

	result = kIOReturnSuccess;

Exit:
	DP(DBG_INFO, "? " LOGSTRAUDIODICTIONARY "[%p]::setDictionaryValue(%s, %d) = 0x%x",
		this, key, value, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	Dictionary に UInt8 値を設定．
 *
 * @param[in]		*key			キー．
 * @param[in]		value			UInt8 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDictionary::setDictionaryValue(
	const char		*key,
	UInt8			value
)
{
	IOReturn		result = kIOReturnError;
	OSNumber		*dictionaryNumber = NULL;

	FailIf(NULL == (dictionaryNumber = OSNumber::withNumber(value, SIZEINBITS(UInt8))), Exit);
	FailIf(!super::setObject(key, dictionaryNumber), Exit);

	dictionaryNumber->release();
	dictionaryNumber = NULL;

	result = kIOReturnSuccess;

Exit:
	DP(DBG_INFO, "? " LOGSTRAUDIODICTIONARY "[%p]::setDictionaryValue(%s, %d) = 0x%x",
		this, key, value, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	Dictionary に UInt16 値を設定．
 *
 * @param[in]		*key			キー．
 * @param[in]		value			UInt16 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDictionary::setDictionaryValue(
	const char		*key,
	UInt16			value
)
{
	IOReturn		result = kIOReturnError;
	OSNumber		*dictionaryNumber = NULL;

	FailIf(NULL == (dictionaryNumber = OSNumber::withNumber(value, SIZEINBITS(UInt16))), Exit);
	FailIf(!super::setObject(key, dictionaryNumber), Exit);

	dictionaryNumber->release();
	dictionaryNumber = NULL;

	result = kIOReturnSuccess;

Exit:
	DP(DBG_INFO, "? " LOGSTRAUDIODICTIONARY "[%p]::setDictionaryValue(%s, %d) = 0x%x",
		this, key, value, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	Dictionary に UInt32 値を設定．
 *
 * @param[in]		*key			キー．
 * @param[in]		value			UInt32 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioDictionary::setDictionaryValue(
	const char		*key,
	UInt32			value
)
{
	IOReturn		result = kIOReturnError;
	OSNumber		*dictionaryNumber = NULL;

	FailIf(NULL == (dictionaryNumber = OSNumber::withNumber(value, SIZEINBITS(UInt32))), Exit);
	FailIf(!super::setObject(key, dictionaryNumber), Exit);

	dictionaryNumber->release();
	dictionaryNumber = NULL;

	result = kIOReturnSuccess;

Exit:
	DP(DBG_INFO, "? " LOGSTRAUDIODICTIONARY "[%p]::setDictionaryValue(%s, %u) = 0x%x",
		this, key, value, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	USB ディスクリプタのバイト列をデバッグログに出力．
 *
 * @param[in]		*descriptor		USB ディスクリプタのバイト列．
 * @param[in]		length			ディスクリプタ長（bLength フィールドと一致している必要がある）．
 */
void
PdjUsbAudioDictionary::logDescriptor(
	UInt8			*descriptor,
	UInt8			length
)
{
	char			*descriptorString = NULL;
	char			byteAndSpace[4];
	UInt8			stringSize = 0;

	FailIf(NULL == descriptor, Exit);
	FailIf(descriptor[0] != length, Exit);

	stringSize = ((length * 3) + 1) * sizeof(char);

	FailIf(NULL == (descriptorString = (char *)IOMalloc(stringSize)), Exit);

	descriptorString[0] = '\0';

	for (UInt8 offset = 0; offset < length; offset++)
	{
		::snprintf(byteAndSpace, 4, "%02x ", descriptor[offset]);

		::strlcat(descriptorString, byteAndSpace, stringSize);
	}

	DP(DBG_INFO, "  Current Descriptor: %s", descriptorString);

Exit:
	if (NULL != descriptorString)
	{
		IOFree(descriptorString, stringSize);
		descriptorString = NULL;
	}
}

/*------------------------------------------*/
/*	PdjUsbConfigurationDictionary クラス	*/
/*------------------------------------------*/

#pragma mark - class PdjUsbConfigurationDictionary -

OSDefineMetaClassAndStructors(PdjUsbConfigurationDictionary, super)

/****************************************************************************/
/**
 * @brief	コンフィグレーション dictionary を生成．
 *
 * @param[in]		*newConfigurationDescriptor		USB コンフィグレーションディスクリプタ．
 * @param[in]		controlInterfaceNum				AudioControl Interface 番号．
 *
 * @return			生成されたコンフィグレーション dictionary．
 */
PdjUsbConfigurationDictionary *
PdjUsbConfigurationDictionary::create(
	const StandardUSB::ConfigurationDescriptor	*newConfigurationDescriptor,
	UInt8										controlInterfaceNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONFIGURATIONDICTIONARY "::create(%p, %d)", newConfigurationDescriptor, controlInterfaceNum);

	PdjUsbConfigurationDictionary		*configDictionary = NULL;

	FailIf(NULL == newConfigurationDescriptor, Exit);
	FailIf(NULL == (configDictionary = new PdjUsbConfigurationDictionary), Exit);

	RELEASE_IF_FALSE(configDictionary, configDictionary->init(newConfigurationDescriptor, controlInterfaceNum));

Exit:
	DP(DBG_TRACE, "- " LOGSTRCONFIGURATIONDICTIONARY "::create() = %p", configDictionary);
	return configDictionary;
}

/****************************************************************************/
/**
 * @brief	ストリーム dictionary にサンプリング周波数を追加．
 *
 * @param[in]		*sampleRates		設定可能なサンプリング周波数の OSArray．
 * @param[in]		streamInterface		interface 番号．
 * @param[in]		altSetting			alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
// [rdar://4867779]
IOReturn
PdjUsbConfigurationDictionary::addSampleRatesToStreamDictionary(
	OSArray			*sampleRates,
	UInt8			streamInterface,
	UInt8			altSetting
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	FailIf(NULL == (thisStream = this->getStreamDictionary(streamInterface, altSetting)), Exit);

	result = thisStream->addSampleRatesToStreamDictionary(sampleRates);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Alternate Setting 番号 0 でストリーミングが可能か判定する．
 *
 * @param[in]		interfaceNum	interface 番号．
 *
 * @retval			true			alternate setting 0 でストリーミングができる．
 * @retval			false			alternate setting 0 ではストリーミングができない．
 */
bool
PdjUsbConfigurationDictionary::alternateSettingZeroCanStream(
	UInt8			interfaceNum
)
{
#ifdef TARGET_PDJ
 	return false;
#else
	OSObject					*dictionaryValue	= NULL;
	OSArray						*streamInterfaces	= NULL;
	OSArray						*endpoints			= NULL;
	PdjUsbStreamDictionary		*thisStream			= NULL;
	PdjUsbEndpointDictionary	*thisEndpoint		= NULL;
	UInt16						maxPacketSize;
	UInt8						streamInterfaceNum	= 0;
	UInt8						streamAltSettingID	= 0;
	UInt8						streamNumEndpoints	= 0;
	bool						result				= false;

	dictionaryValue = OSDictionary::getObject(kStreamDictionaries);

	if (dictionaryValue)
	{
		streamInterfaces = OSDynamicCast(OSArray, dictionaryValue);
	}

	if (NULL != streamInterfaces)
	{
		for (UInt8 streamIndex = 0; streamIndex < streamInterfaces->getCount(); streamIndex++)
		{
			thisStream = OSDynamicCast(PdjUsbStreamDictionary, streamInterfaces->getObject(streamIndex));

			if (thisStream)
			{
				FailIf(kIOReturnSuccess != thisStream->getInterfaceNumber(&streamInterfaceNum), Exit);
				FailIf(kIOReturnSuccess != thisStream->getAlternateSetting(&streamAltSettingID), Exit);
				FailIf(kIOReturnSuccess != thisStream->getNumEndpoints(&streamNumEndpoints), Exit);
			}

			if (
					(thisStream)
			   &&
					(streamInterfaceNum == interfaceNum)
			   &&
					(streamAltSettingID == 0)
			   &&
					(streamNumEndpoints > 0)			// [rdar://4340026] We don't care about this unless it has a streaming endpoint.
			   )
			{
				// [rdar://4793892] - There are streaming endpoints @ alt interface 0. Check to see if any have a nonzero maximum packet size.
				FailIf(NULL == (endpoints = thisStream->getEndpoints()), Exit);

				for (UInt8 endpointIndex = 0; endpointIndex < streamNumEndpoints; endpointIndex++)
				{
					FailIf(NULL == (thisEndpoint = OSDynamicCast(PdjUsbEndpointDictionary, endpoints->getObject(endpointIndex))), Exit);
					FailIf(kIOReturnSuccess != thisEndpoint->getMaxPacketSize(&maxPacketSize), Exit);

					if (0 < maxPacketSize)
					{
						DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::alternateSettingZeroCanStream() = true", this);

						result = true;

						break;
					}
				}
			}
		}
	}

Exit:

	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	エンドポイントディスクリプタから LockDelay 値を取得．
 *
 * @param[out]		*lockDelay			取得した LockDelay 値．
 * @param[in]		interfaceNum		interface 番号．
 * @param[in]		altSettingID		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::asEndpointGetLockDelay(
	UInt8			*lockDelay,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
#ifdef TARGET_PDJ
 	*lockDelay = 0;
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*lockDelay = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->asEndpointGetLockDelay(lockDelay)), Exit);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	エンドポイントディスクリプタから LockDelayUnits 値を取得．
 *
 * @param[out]		*lockDelayUnits		取得した LockDelayUnits 値．
 * @param[in]		interfaceNum		interface 番号．
 * @param[in]		altSettingID		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::asEndpointGetLockDelayUnits(
	UInt8			*lockDelayUnits,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
#ifdef TARGET_PDJ
 	*lockDelayUnits = 0;
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*lockDelayUnits = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->asEndpointGetLockDelayUnits(lockDelayUnits)), Exit);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	エンドポイントがサンプリング周波数設定に対応しているかどうかを調べる．
 *			エンドポイントディスクリプタの bmAttributes フィールドで
 *			Sampling Frequency Control がサポートされているかどうか判定する．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @retval			true			サンプリング周波数設定に対応している．
 * @retval			false			サンプリング周波数設定に対応していない．
 */
bool
PdjUsbConfigurationDictionary::asEndpointHasSampleFreqControl(
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
#ifdef TARGET_PDJ
 	return true;
#else
	bool						result = false;
	PdjUsbStreamDictionary		*thisStream = NULL;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);

	result = thisStream->asEndpointHasSampleFreqControl();

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	チャンネルがミュート設定に対応しているかどうかを調べる．
 *			Feature Unit ディスクリプタの bmaControls フィールドで
 *			Mute Control がサポートされているかどうか判定する．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		featureUnitID	チェック対象の Feature Unit ID．
 * @param[in]		channelNum		チャンネル番号．
 *
 * @retval			true			ミュート設定に対応している．
 * @retval			false			ミュート設定に対応していない．
 */
bool
PdjUsbConfigurationDictionary::channelHasMuteControl(
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			featureUnitID,
	UInt8			channelNum
)
{
	bool						returnBool = false;
	PdjUsbControlDictionary		*thisControl;

	thisControl = this->getControlDictionary(interfaceNum, altSettingID);

	if (thisControl)
	{
		returnBool = thisControl->channelHasMuteControl(featureUnitID, channelNum);
	}

	return returnBool;
}

/****************************************************************************/
/**
 * @brief	チャンネルがボリューム設定に対応しているかどうかを調べる．
 *			Feature Unit ディスクリプタの bmaControls フィールドで
 *			Volume Control がサポートされているかどうか判定する．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		featureUnitID	チェック対象の Feature Unit ID．
 * @param[in]		channelNum		チャンネル番号．
 *
 * @retval			true			ボリューム設定に対応している．
 * @retval			false			ボリューム設定に対応していない．
 */
bool
PdjUsbConfigurationDictionary::channelHasVolumeControl(
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			featureUnitID,
	UInt8			channelNum
)
{
	bool						returnBool = false;
	PdjUsbControlDictionary		*thisControl;

	thisControl = this->getControlDictionary(interfaceNum, altSettingID);

	if (thisControl)
	{
		returnBool = thisControl->channelHasVolumeControl(featureUnitID, channelNum);
	}
	else
	{
		DP(DBG_ERROR, "! " LOGSTRCONFIGURATIONDICTIONARY "::channelHasVolumeControl() - no control for interface %d alt setting %d",
			interfaceNum, altSettingID);
	}

	return returnBool;
}

/****************************************************************************/
/**
 * @brief	クロックソースが周波数設定に対応しているかどうか調べる．
 *			Clock Source ディスクリプタの bmControls フィールドで
 *			Clock Frequency Control がサポートされているかどうか判定する．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSetting		alternate setting 番号．
 * @param[in]		clockSourceID	クロックソース ID．
 * @param[in]		isProgrammable	true なら、programmable な Frequency Control のみを探す．
 *									false なら、read-only な Frequency Control も含めて探す．
 *
 * @retval			true			周波数設定に対応している．
 * @retval			false			周波数設定に対応していない．
 */
// [rdar://4867779]
bool
PdjUsbConfigurationDictionary::clockSourceHasFrequencyControl(
	UInt8			interfaceNum,
	UInt8			altSetting,
	UInt8			clockSourceID,
	bool			isProgrammable
)
{
	bool						returnBool = false;
	PdjUsbControlDictionary		*thisControl;

	thisControl = this->getControlDictionary(interfaceNum, altSetting);

	if (thisControl)
	{
		returnBool = thisControl->clockSourceHasFrequencyControl(clockSourceID, isProgrammable);
	}
	else
	{
		DP(DBG_ERROR, "! " LOGSTRCONFIGURATIONDICTIONARY "::clockSourceHasFrequencyControl() - no control for interface %d alt setting %d programmable %d",
			interfaceNum, altSetting, isProgrammable);
	}

	return returnBool;
}

/****************************************************************************/
/**
 * @brief	クロックソースが有効性確認に対応しているかどうか調べる．
 *			Clock Source ディスクリプタの bmControls フィールドで
 *			Clock Validity Control がサポートされているかどうか判定する．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSetting		alternate setting 番号．
 * @param[in]		clockSourceID	クロックソース ID．
 *
 * @retval			true			有効性確認に対応している．
 * @retval			false			有効性確認に対応していない．
 */
// <rdar://7446555>
bool
PdjUsbConfigurationDictionary::clockSourceHasValidityControl(
	UInt8			interfaceNum,
	UInt8			altSetting,
	UInt8			clockSourceID
)
{
	bool						returnBool = false;
	PdjUsbControlDictionary		*thisControl;

	thisControl = this->getControlDictionary(interfaceNum, altSetting);

	if (thisControl)
	{
		returnBool = thisControl->clockSourceHasValidityControl(clockSourceID);
	}
	else
	{
		DP(DBG_ERROR, "! " LOGSTRCONFIGURATIONDICTIONARY "::clockSourceHasValidityControl() - no control for interface %d alt setting %d",
			interfaceNum, altSetting);
	}

	return returnBool;
}

/****************************************************************************/
/**
 * @brief	Audio Device Class のバージョンを取得．
 *
 * @param[out]		*adcVersion		インターフェイスディスクリプタの bcdADC フィールド．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getADCVersion(
	UInt16			*adcVersion
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;
	UInt8						controlInterfaceNum = 0;

	*adcVersion = 0;
	FailIf(kIOReturnSuccess != (result = this->getControlInterfaceNum(&controlInterfaceNum)), Exit);

	result = kIOReturnError;

	FailIf(NULL == (thisControl = this->getControlDictionary(controlInterfaceNum, 0)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getADCVersion(adcVersion)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	指定された条件に合う Alternate Setting を探す．
 *
 * @param[out]		*altSettingID	見つかった alternate setting 番号．
 * @param[in]		interfaceNum	探索対象の interface 番号．
 * @param[in]		numChannels		探索条件：チャンネル数．
 * @param[in]		sampleSize		探索条件：サンプルサイズ．
 * @param[in]		sampleRate		探索条件：サンプリング周波数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getAltSettingWithSettings(
	UInt8			*altSettingID,
	UInt8			interfaceNum,
	UInt8			numChannels,
	UInt8			sampleSize,
	UInt32			sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getAltSettingWithSettings(%p, %d, %d, %d, %u)",
		this, altSettingID, interfaceNum, numChannels, sampleSize, sampleRate);

#ifdef TARGET_PDJ
 	IOReturn		result = kIOReturnSuccess;
 
 	*altSettingID = 1;
#else
	IOReturn		result = kIOReturnError;
	IOReturn		interimResult;
	UInt8			potentialAltSetting;
	UInt8			thisAltSetting;
	UInt16			format;

	// [rdar://4228470] Only consider alternate interface 0 if there are streaming endpoints. /thw
	if (this->alternateSettingZeroCanStream(interfaceNum))
	{
		potentialAltSetting = 0;
	}
	else
	{
		potentialAltSetting = 1;
	}

	while ((kIOReturnSuccess != result) && (potentialAltSetting != 255))
	{
		if (
				(kIOReturnSuccess == this->getFormat(&format, interfaceNum, potentialAltSetting))
		   &&
				(0 != (format & 0x0FFF))
		   )
		{
			// Make sure it's not an undefined format
			if (
					kIOReturnSuccess != this->getNextAltSettingWithNumChannels(
													&thisAltSetting,
													interfaceNum,
													potentialAltSetting,
													numChannels
											  )
			   )
			{
				DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getAltSettingWithSettings() - No alternate setting ID for interface %d, %d channels. Terminating loop.",
					this, interfaceNum, numChannels);

				break;		// Done, so break out of while loop
			}
			else
			{
				DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getAltSettingWithSettings() - Examining alt interface %d.", this, thisAltSetting);

				potentialAltSetting = thisAltSetting;
			}

			// Any errors here are non-fatal.
			interimResult = this->getNextAltSettingWithSampleSize(
										&thisAltSetting,
										interfaceNum,
										potentialAltSetting,
										sampleSize
								  );

			if ((kIOReturnSuccess == interimResult) && (potentialAltSetting == thisAltSetting))
			{
				if (0 != sampleRate)
				{
					// Must find a specific sample rate here.
					DP(DBG_INFO, "? ConfigurationDictionary[%p]::getAltSettingWithSettings() - Attempting to match sample rate %u.", this, sampleRate);

					interimResult = this->getNextAltSettingWithSampleRate(
												&thisAltSetting,
												interfaceNum,
												potentialAltSetting,
												sampleRate
										  );

					if ((kIOReturnSuccess == interimResult) && (potentialAltSetting == thisAltSetting))
					{
						*altSettingID = potentialAltSetting;

						result = kIOReturnSuccess;
					}
					else
					{
						potentialAltSetting++;
					}
				}
				else
				{
					DP(DBG_INFO, "? ConfigurationDictionary[%p]::getAltSettingWithSettings() - Not attempting to match sample rate.", this);
					// Sample rate matching is unnecessary.

					*altSettingID = potentialAltSetting;

					result = kIOReturnSuccess;
				}
			}
			else
			{
				potentialAltSetting++;
			}
		}
		else
		{
			potentialAltSetting++;
		}
	} // while
#endif

	DP(DBG_TRACE, "- " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getAltSettingWithSettings(%p = %d, %d, %d, %d, %u) = 0x%x",
		this, altSettingID, *altSettingID, interfaceNum, numChannels, sampleSize, sampleRate, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	AC-3 デコーダの BSID モードを取得．
 *
 * @param[out]		*ac3BSID		AC-3 デコーダの bmBSID フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getAC3BSID(
	UInt32			*ac3BSID,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream;

	*ac3BSID = 0;

	thisStream = this->getStreamDictionary(interfaceNum, altSettingID);

	if (thisStream)
	{
		result = thisStream->getAC3BSID(ac3BSID);
	}
	else
	{
		DP(DBG_ERROR, "! " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getAC3BSID(%p, %d, %d) - Could not get stream interface dictionary.",
			this, ac3BSID, interfaceNum, altSettingID);
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	AudioControl Interface の配下にある AudioStreaming Interface を列挙．
 *
 * @param[out]		**controlledStreams		配下にある AudioStreaming Interface の OSArray．
 * @param[out]		*numControlledStreams	配下にある AudioStreaming Interface の数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getControlledStreamNumbers(
	OSArray			**controlledStreams,
	UInt8			*numControlledStreams
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;
	OSArray						*controlDictionaries = NULL;
	OSObject					*dictionaryValue = NULL;

	FailIf(NULL == controlledStreams, Exit);

	*controlledStreams		= NULL;
	*numControlledStreams	= NULL;

	FailIf(NULL == (dictionaryValue = OSDictionary::getObject(kControlDictionaries)), Exit);
	FailIf(NULL == (controlDictionaries = OSDynamicCast(OSArray, dictionaryValue)), Exit);
	FailIf(NULL == (thisControl = OSDynamicCast(PdjUsbControlDictionary, controlDictionaries->getObject(0))), Exit);
	FailIf(NULL == (*controlledStreams = thisControl->getStreamInterfaceNumbers()), Exit);
	FailIf(kIOReturnSuccess != (result = (thisControl->getNumStreamInterfaces(numControlledStreams))), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface のオーディオフォーマットを取得．
 *
 * @param[out]		*format			インターフェイスディスクリプタの wFormatTag フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getFormat(
	UInt16			*format,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
#ifdef TARGET_PDJ
 	*format = PCM;
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*format = TYPE_I_UNDEFINED;

	thisStream = this->getStreamDictionary(interfaceNum, altSettingID);

	if (thisStream)
	{
		FailIf(kIOReturnSuccess != (result = thisStream->getFormatTag(format)), Exit);
	}

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	設定可能なサンプリング周波数の最大値を求める．
 *
 * @param[out]		*sampleRate		サンプリング周波数の最大値．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getHighestSampleRate(
	UInt32			*sampleRate,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getHighestSampleRate(%p, %d, %d)",
		this, sampleRate, interfaceNum, altSettingID);

	IOReturn		result = kIOReturnError;
	OSArray			*sampleRates = NULL;
	OSObject		*arrayValue;
	OSNumber		*sampleRateNumber;
	UInt8			numSampleRates;

	*sampleRate = 0;

	FailIf(NULL == (sampleRates = this->getSampleRates(interfaceNum, altSettingID)), Exit);

	// [rdar://5313648] The number of sample rates could be variable, so check the size of the array.
	numSampleRates = sampleRates->getCount();

	for (UInt32 sampleRateIndex = 0; sampleRateIndex < numSampleRates; sampleRateIndex++)
	{
		FailIf(NULL == (arrayValue = sampleRates->getObject(sampleRateIndex)), Exit);
		FailIf(NULL == (sampleRateNumber = OSDynamicCast(OSNumber, arrayValue)), Exit);

		DP(DBG_INFO, "   %u: %u", sampleRateIndex, sampleRateNumber->unsigned32BitValue());

		// Keep the higher sample rate.
		*sampleRate = (sampleRateNumber->unsigned32BitValue() > (*sampleRate))
						? (sampleRateNumber->unsigned32BitValue())
						: (*sampleRate);
	}

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getHighestSampleRate(%p = %u, %u, %u) = 0x%x",
		this, sampleRate, *sampleRate, interfaceNum, altSettingID, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントのアドレスを取得．
 *
 * @param[out]		*address		取得したエンドポイントアドレス．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		direction		転送方向．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIsocEndpointAddress(
	UInt8			*address,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			direction
)
{
#ifdef TARGET_PDJ
 	*address = 0x82;
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*address = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);

	result = thisStream->getIsocEndpointAddress(address, direction);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの転送方向を取得．
 *
 * @param[out]		*direction		取得した転送方向．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
// Use getTerminalLink to get the unit number of the input or output terminal with which the endpoint is associated.
// From that terminal the direction can be determined, which in turn clarifies the endpoint direction.
IOReturn
PdjUsbConfigurationDictionary::getIsocEndpointDirection(
	UInt8			*direction,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getIsocEndpointDirection(%p, %d, %d)",
		this, direction, interfaceNum, altSettingID);

#ifdef TARGET_PDJ
 	IOReturn	result = kIOReturnSuccess;
 
 	switch (interfaceNum)
 	{
 	case 1:		*direction = kUSBOut;	break;
 	case 2:		*direction = kUSBIn;	break;
 	default:	*direction = 0xFF;		break;
 	}
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;
	PdjUsbControlDictionary		*thisControl = NULL;
	UInt8						endpointDirection;
	UInt8						terminalLink;
	UInt8						numOutputs;
	UInt8						numInputs;
	UInt8						numEndpoints = 0;
	UInt8						index = 0;
	UInt8						terminalID;
	UInt8						controlInterfaceNum;
	bool						done = false;

	FailIf(NULL == direction, Exit);

	*direction = 0xFF;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != this->getControlInterfaceNum(&controlInterfaceNum), Exit);
	FailIf(NULL == (thisControl = this->getControlDictionary(controlInterfaceNum, 0)), Exit);

	// The following returns the unitID of the terminal with which the endpoint is associated
	FailIf(kIOReturnSuccess != thisStream->getTerminalLink(&terminalLink), Exit);

	DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getIsocEndpointDirection() - terminalLink = %d", this, terminalLink);

	if (terminalLink)
	{
		FailIf(kIOReturnSuccess != thisControl->getNumOutputTerminals(&numOutputs), Exit);

		DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getIsocEndpointDirection() - numOutputs = %d", this, numOutputs);

		while (index < numOutputs)
		{
			FailIf(kIOReturnSuccess != thisControl->getIndexedOutputTerminalID(&terminalID, index), Exit);

			if (terminalID == terminalLink)
			{
				*direction = kUSBIn;

				numEndpoints = numOutputs;

				DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getIsocEndpointDirection() - Found an output terminal (%d) at index %d",
					this, terminalID, index);

				break;
			}
			else
			{
				index++;
			}
		}

		if (0xFF == *direction)
		{
			FailIf(kIOReturnSuccess != (result = thisControl->getNumInputTerminals(&numInputs)), Exit);

			DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getIsocEndpointDirection() - numInputs = %d", this, numInputs);

			index = 0;

			while (index < numInputs)
			{
				FailIf(kIOReturnSuccess != (result = thisControl->getIndexedInputTerminalID(&terminalID, index)), Exit);

				if (terminalID == terminalLink)
				{
					*direction = kUSBOut;

					numEndpoints = numInputs;

					DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getIsocEndpointDirection() - Found an input terminal (%d) at index %d",
						this, terminalID, index);

					break;
				}
				else
				{
					index++;
				}
			}
		}
	} // (if terminalLink)
	else
	{
		result = kIOReturnError;
	}

	if (0xFF != *direction)
	{
		for (index = 0; index < numEndpoints; index++)
		{
			FailIf(kIOReturnSuccess != (result = thisStream->getIsocEndpointDirection(&endpointDirection, index)), Exit);

			if (endpointDirection == *direction)
			{
				done = true;

				break;		// found the right endpoint, get out of for loop
			}
		}

		result = (done ? kIOReturnSuccess : kIOReturnError);
	}

Exit:
#endif

	DP(DBG_TRACE, "- " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getIsocEndpointDirection(%d, %d, %d) = 0x%x",
		this, (direction ? *direction : 0), interfaceNum, altSettingID, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの MaxPacketSize を取得．
 *
 * @param[out]		*maxPacketSize	取得した MaxPacketSize．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		direction		転送方向．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIsocEndpointMaxPacketSize(
	UInt16			*maxPacketSize,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			direction
)
{
#ifdef TARGET_PDJ
 	*maxPacketSize = 1024;
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*maxPacketSize = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);

	result = thisStream->getIsocEndpointMaxPacketSize(maxPacketSize, direction);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの同期タイプを取得．
 *
 * @param[out]		*syncType		エンドポイントディスクリプタの bmAttributes
 *									フィールドで指定されている同期タイプ．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		address			エンドポイントアドレス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIsocEndpointSyncType(
	UInt8			*syncType,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			address
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*syncType = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->getIsocEndpointSyncType(syncType, address)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントのインターバルを取得．
 *
 * @param[out]		*interval		エンドポイントディスクリプタの bInterval フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		direction		転送方向．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIsocEndpointInterval(
	UInt8			*interval,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			direction
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*interval = 0;

	thisStream = this->getStreamDictionary(interfaceNum, altSettingID);

	if (thisStream)
	{
		FailIf(kIOReturnSuccess != (result = thisStream->getIsocEndpointInterval(interval, direction)), Exit);
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの付随エンドポイントのアドレスを取得．
 *
 * @param[out]		*assocEndpointAddress	取得した付随エンドポイントのアドレス．
 * @param[in]		interfaceNum			interface 番号．
 * @param[in]		altSettingID			alternate setting 番号．
 * @param[in]		address					エンドポイントアドレス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIsocAssociatedEndpointAddress(
	UInt8			*assocEndpointAddress,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			address
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*assocEndpointAddress = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);

	result = thisStream->getIsocAssociatedEndpointAddress(assocEndpointAddress, address);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの付随エンドポイントの MaxPacketSize を取得．
 *
 * @param[out]		*maxPacketSize	取得した MaxPacketSize．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		address			エンドポイントアドレス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIsocAssociatedEndpointMaxPacketSize(
	UInt16			*maxPacketSize,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			address
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*maxPacketSize = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);

	result = thisStream->getIsocAssociatedEndpointMaxPacketSize(maxPacketSize, address);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの付随エンドポイントのリフレッシュ間隔を取得．
 *
 * @param[out]		*refreshInt		エンドポイントディスクリプタの bRefresh フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		address			エンドポイントアドレス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIsocAssociatedEndpointRefreshInt(
	UInt8			*refreshInt,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			address
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*refreshInt = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);

	result = thisStream->getIsocAssociatedEndpointRefreshInt(refreshInt, address);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	入力ターミナルのタイプを取得（ターミナル ID から）．
 *
 * @param[out]		*terminalType	Input Terminal ディスクリプタの wTerminalType フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		terminalID		ターミナル ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getInputTerminalType(
	UInt16			*terminalType,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			terminalID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*terminalType = USBToHostWord(INPUT_UNDEFINED);

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getInputTerminalType(terminalType, terminalID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	出力ターミナルのタイプを取得（ターミナル ID から）．
 *
 * @param[out]		*terminalType	Output Terminal ディスクリプタの wTerminalType フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		terminalID		ターミナル ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getOutputTerminalType(
	UInt16			*terminalType,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			terminalID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*terminalType = USBToHostWord(OUTPUT_UNDEFINED);

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getOutputTerminalType(terminalType, terminalID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	入力ターミナルのタイプを取得（OSArray のインデックスから）．
 *
 * @param[out]		*terminalType	Input Terminal ディスクリプタの wTerminalType フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		index			OSArray のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIndexedInputTerminalType(
	UInt16			*terminalType,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			index
)
{
#ifdef TARGET_PDJ
 	*terminalType = USBToHostWord(EXTERNAL_LINE_CONNECTOR);
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*terminalType = USBToHostWord(INPUT_UNDEFINED);

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getIndexedInputTerminalType(terminalType, index)), Exit);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	入力ターミナルの ID を取得（OSArray のインデックスから）．
 *
 * @param[out]		*terminalID		取得したターミナル ID．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		index			OSArray のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIndexedInputTerminalID(
	UInt8			*terminalID,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			index
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*terminalID = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getIndexedInputTerminalID(terminalID, index)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	出力ターミナルのタイプを取得（OSArray のインデックスから）．
 *
 * @param[out]		*terminalType	Output Terminal ディスクリプタの wTerminalType フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		index			OSArray のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIndexedOutputTerminalType(
	UInt16			*terminalType,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			index
)
{
#ifdef TARGET_PDJ
 	*terminalType = USBToHostWord(EXTERNAL_LINE_CONNECTOR);
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*terminalType = USBToHostWord(OUTPUT_UNDEFINED);

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getIndexedOutputTerminalType(terminalType, index)), Exit);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	出力ターミナルの ID を取得（OSArray のインデックスから）．
 *
 * @param[out]		*terminalID		取得したターミナル ID．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		index			OSArray のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getIndexedOutputTerminalID(
	UInt8			*terminalID,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			index
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*terminalID = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getIndexedOutputTerminalID(terminalID, index)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Interface Class を取得．
 *
 * @param[out]		*interfaceClass		インターフェイスディスクリプタの
 *										bInterfaceClass フィールド．
 * @param[in]		interfaceNum		interface 番号．
 * @param[in]		altSettingID		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getInterfaceClass(
	UInt8			*interfaceClass,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream;

	*interfaceClass = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->getInterfaceClass(interfaceClass)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Interface SubClass を取得．
 *
 * @param[out]		*interfaceSubClass	インターフェイスディスクリプタの
 *										bInterfaceSubClass フィールド．
 * @param[in]		interfaceNum		interface 番号．
 * @param[in]		altSettingID		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getInterfaceSubClass(
	UInt8			*interfaceSubClass,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*interfaceSubClass = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->getInterfaceSubClass(interfaceSubClass)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Alternate Setting 数を数える．
 *
 * @param[out]		*numAltSettings		インターフェイスに含まれる alternate setting の数．
 * @param[in]		interfaceNum		interface 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNumAltSettings(
	UInt8			*numAltSettings,
	UInt8			interfaceNum
)
{
#ifdef TARGET_PDJ
 	*numAltSettings = 2;
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;
	OSObject					*dictionaryValue = NULL;
	OSArray						*streamDictionaries = NULL;
	UInt8						streamInterfaceNum;

	*numAltSettings = 0;

	FailIf(NULL == (dictionaryValue = OSDictionary::getObject(kStreamDictionaries)), Exit);
	FailIf(NULL == (streamDictionaries = OSDynamicCast(OSArray, dictionaryValue)), Exit);

	for (UInt16 streamIndex = 0; streamIndex < streamDictionaries->getCount(); streamIndex++)
	{
		FailIf(NULL == (thisStream = OSDynamicCast(PdjUsbStreamDictionary, streamDictionaries->getObject(streamIndex))), Exit);
		FailIf(kIOReturnSuccess != (result = thisStream->getInterfaceNumber(&streamInterfaceNum)), Exit);

		if (streamInterfaceNum == interfaceNum)
		{
			(*numAltSettings)++;
		}
	}

	result = kIOReturnSuccess;

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	チャンネル数を取得．
 *
 * @param[out]		*numChannels	取得したチャンネル数．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNumChannels(
	UInt8			*numChannels,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
#ifdef TARGET_PDJ
 	switch (interfaceNum)
 	{
 	case 1:		*numChannels = 12;	break;
 	case 2:		*numChannels = 12;	break;
 	default:	*numChannels = 0;	break;
 	}
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*numChannels = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->getNumChannels(numChannels)), Exit);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	Feature Unit のコントロール数を取得．
 *
 * @param[out]		*numControls	取得したコントロール数．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		featureUnitID	Feature Unit ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNumControls(
	UInt8			*numControls,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			featureUnitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*numControls = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getNumControls(numControls, featureUnitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	設定可能なサンプリング周波数の数を取得．
 *
 * @param[out]		*numSampleRates		設定可能なサンプリング周波数の数．
 * @param[in]		interfaceNum		interface 番号．
 * @param[in]		altSettingID		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNumSampleRates(
	UInt8			*numSampleRates,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
#ifdef TARGET_PDJ
 	*numSampleRates = 3;
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*numSampleRates = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->getNumSampleRates(numSampleRates)), Exit);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	Input Terminal の数を取得．
 *
 * @param[out]		*numInputTerminals	Input Terminal の数．
 * @param[in]		interfaceNum		interface 番号．
 * @param[in]		altSettingID		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNumInputTerminals(
	UInt8			*numInputTerminals,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*numInputTerminals = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getNumInputTerminals(numInputTerminals)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Output Terminal の数を取得．
 *
 * @param[out]		*numOutputTerminals	Output Terminal の数．
 * @param[in]		interfaceNum		interface 番号．
 * @param[in]		altSettingID		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNumOutputTerminals(
	UInt8			*numOutputTerminals,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*numOutputTerminals = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getNumOutputTerminals(numOutputTerminals)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Selector Unit の数を取得．
 *
 * @param[out]		*numSelectorUnits	Selector Unit の数．
 * @param[in]		interfaceNum		interface 番号．
 * @param[in]		altSettingID		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNumSelectorUnits(
	UInt8			*numSelectorUnits,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*numSelectorUnits = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getNumSelectorUnits(numSelectorUnits)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットの入力ピンの数を取得．
 *
 * @param[out]		*numSources		ユニットの入力ピンの数．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNumSources(
	UInt8			*numSources,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*numSources = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getNumSources(numSources, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface の総数を数える．
 *
 * @param[out]		*numStreamInterfaces	AudioStreaming Interface の総数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNumStreamInterfaces(
	UInt8			*numStreamInterfaces
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;
	OSObject					*dictionaryValue = NULL;
	OSArray						*streamDictionaries = NULL;
	UInt8						streamInterfaceNum;
	UInt8						thisInterfaceNum = 0;

	*numStreamInterfaces = 0;

	FailIf(NULL == (dictionaryValue = OSDictionary::getObject(kStreamDictionaries)), Exit);
	FailIf(NULL == (streamDictionaries = OSDynamicCast(OSArray, dictionaryValue)), Exit);

	// Each time the stream interface changes, increment the number of stream interfaces. Begin with 0 since it is impossible to have
	// a class-complaint device with a USB audio streaming interface at interface 0 since it must be preceded by at least one control
	// interface.

	for (UInt16 streamIndex = 0; streamIndex < streamDictionaries->getCount(); streamIndex++)
	{
		FailIf(NULL == (thisStream = OSDynamicCast(PdjUsbStreamDictionary, streamDictionaries->getObject(streamIndex))), Exit);
		FailIf(kIOReturnSuccess != (result = thisStream->getInterfaceNumber(&streamInterfaceNum)), Exit);

		if (streamInterfaceNum != thisInterfaceNum)
		{
			thisInterfaceNum = streamInterfaceNum;

			(*numStreamInterfaces)++;
		}
	}

	result = kIOReturnSuccess;

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	設定可能なサンプリング周波数の OSArray を取得．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			設定可能なサンプリング周波数の OSArray．
 */
OSArray *
PdjUsbConfigurationDictionary::getSampleRates(
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
#ifdef TARGET_PDJ
 	if (NULL == mSampleRates)
 	{
 		OSNumber	*rate;
 
 		mSampleRates = OSArray::withCapacity(1);
 
 		rate = OSNumber::withNumber(96000, 24);
 		mSampleRates->setObject(rate);
 		rate->release();
 
 		rate = OSNumber::withNumber(48000, 24);
 		mSampleRates->setObject(rate);
 		rate->release();
 
 		rate = OSNumber::withNumber(44100, 24);
 		mSampleRates->setObject(rate);
 		rate->release();
 	}
 
 	return mSampleRates;
#else
	PdjUsbStreamDictionary		*thisStream = NULL;
	OSArray						*sampleRates = NULL;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);

	sampleRates = thisStream->getSampleRates();

Exit:
	return sampleRates;
#endif
}

/****************************************************************************/
/**
 * @brief	量子化ビット数を取得．
 *
 * @param[out]		*sampleSize		フォーマットタイプディスクリプタの bBitResolution フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getBitResolution(
	UInt8			*sampleSize,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
#ifdef TARGET_PDJ
 	*sampleSize = 24;
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*sampleSize = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->getBitResolution(sampleSize)), Exit);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	指定のクロックセレクタに属するクロックソースの OSArray を取得．
 *
 * @param[out]		**clockSelectorSources	クロックセレクタディスクリプタに列挙されている
 *											クロックソースの OSArray．
 * @param[in]		interfaceNum			interface 番号．
 * @param[in]		altSettingID			alternate setting 番号．
 * @param[in]		unitID					クロックセレクタのユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getClockSelectorSources(
	OSArray			**clockSelectorSources,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*clockSelectorSources = NULL;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getClockSelectorSources(clockSelectorSources, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースのタイプを取得．
 *
 * @param[out]		*clockType		クロックソースディスクリプタの bmAttributes フィールド
 *									で指定されたクロックタイプ．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		unitID			クロックソース ID．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbConfigurationDictionary::getClockSourceClockType(
	UInt8			*clockType,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*clockType = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getClockSourceClockType(clockType, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースの関連ターミナルを取得．
 *
 * @param[out]		*assocTerminal	クロックソースディスクリプタの bAssocTerminal フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		unitID			クロックソース ID．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbConfigurationDictionary::getClockSourceAssocTerminal(
	UInt8			*assocTerminal,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*assocTerminal = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getClockSourceAssocTerminal(assocTerminal, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	サブフレームサイズを取得．
 *
 * @param[out]		*subframeSize	フォーマットタイプディスクリプタの bSubFrameSize フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getSubframeSize(
	UInt8			*subframeSize,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
#ifdef TARGET_PDJ
 	*subframeSize = 3;
 
 	return kIOReturnSuccess;
#else
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*subframeSize = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->getSubframeSize(subframeSize)), Exit);

Exit:
	return result;
#endif
}

/****************************************************************************/
/**
 * @brief	ユニットの種類を取得．
 *
 * @param[out]		*subType		ディスクリプタの bDescriptorSubtype フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getSubType(
	UInt8			*subType,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*subType = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getSubType(subType, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットのソース ID を取得．
 *
 * @param[out]		*sourceID		ディスクリプタの bSourceID フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getSourceID(
	UInt8			*sourceID,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*sourceID = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getSourceID(sourceID, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットのソース ID の OSArray を取得．
 *
 * @param[out]		**sourceIDs		ディスクリプタの baSourceID フィールドに
 *									列挙されたソース ID の OSArray．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getSourceIDs(
	OSArray			**sourceIDs,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*sourceIDs = NULL;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getSourceIDs(sourceIDs, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Input Terminal および Output Terminal に接続されたクロックエンティティを列挙．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			Input Terminal ディスクリプタおよび Output Terminal ディスクリプタの
 *					bCSourceID フィールドから生成した OSArray．
 */
OSArray *
PdjUsbConfigurationDictionary::getTerminalClockEntities(
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	PdjUsbControlDictionary		*thisControl = NULL;
	OSArray						*clockEntities = NULL;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);

	clockEntities = thisControl->getTerminalClockEntities();

Exit:
	return clockEntities;
}

/****************************************************************************/
/**
 * @brief	ユニットのクロックソース ID を取得．
 *
 * @param[out]		*clockSourceID	ディスクリプタの bCSourceID フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getClockSourceID(
	UInt8			*clockSourceID,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*clockSourceID = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getClockSourceID(clockSourceID, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットのストリングディスクリプタのインデックスを取得．
 *
 * @param[out]		*stringIndex	ストリングディスクリプタのインデックス．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
// <rdar://6394629>
IOReturn
PdjUsbConfigurationDictionary::getStringIndex(
	UInt8			*stringIndex,
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt8			unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*stringIndex = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getStringIndex(stringIndex, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットのディスクリプタから Audio Channel Cluster Descriptor を作成．
 *
 * @param[out]		*clusterDescriptor	ユニットのディスクリプタから作成した
 *										Audio Channel Cluster Descriptor．
 *										メモリは呼び出し側で確保しておくこと．
 * @param[in]		interfaceNum		interface 番号．
 * @param[in]		altSettingID		alternate setting 番号．
 * @param[in]		unitID				ユニット ID．
 *
 * @return			IOReturn コード．
 */
// <rdar://6430836>
IOReturn
PdjUsbConfigurationDictionary::getAudioClusterDescriptor(
	AudioClusterDescriptor	*clusterDescriptor,
	UInt8					interfaceNum,
	UInt8					altSettingID,
	UInt8					unitID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	clusterDescriptor->bNrChannels		= 0;
	clusterDescriptor->bmChannelConfig	= 0;
	clusterDescriptor->iChannelNames	= 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisControl->getAudioClusterDescriptor(clusterDescriptor, unitID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface のターミナルリンクを取得．
 *
 * @param[out]		*terminalLink	インターフェイスディスクリプタの bTerminalLink フィールド．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getTerminalLink(
	UInt8					*terminalLink,
	UInt8					interfaceNum,
	UInt8					altSettingID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbStreamDictionary		*thisStream = NULL;

	*terminalLink = 0;

	FailIf(NULL == (thisStream = this->getStreamDictionary(interfaceNum, altSettingID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisStream->getTerminalLink(terminalLink)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Interrupt エンドポイントを検索．
 *
 * @param[out]		*address		見つかった Interrupt エンドポイントのアドレス．
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
IOReturn
PdjUsbConfigurationDictionary::getInterruptEndpointAddress(
	UInt8			*address,
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;

	*address = 0;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);

	result = thisControl->getInterruptEndpointAddress(address);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Interrupt エンドポイントの有無を調べる．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @retval			true			Interrupt エンドポイントがある．
 * @retval			false			Interrupt エンドポイントはない．
 */
// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
bool
PdjUsbConfigurationDictionary::hasInterruptEndpoint(
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	bool						result = false;

#ifdef TARGET_PDJ
#else
	PdjUsbControlDictionary		*thisControl = NULL;

	FailIf(NULL == (thisControl = this->getControlDictionary(interfaceNum, altSettingID)), Exit);

	result = thisControl->hasInterruptEndpoint();

Exit:
#endif

	return result;
}

/****************************************************************************/
/**
 * @brief	サンプリング周波数がサポートされているかどうか調べる．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 * @param[in]		verifyRate		チェックするサンプリング周波数．
 *
 * @retval			true			そのサンプリング周波数はサポートされている．
 * @retval			false			そのサンプリング周波数はサポートされていない．
 */
bool
PdjUsbConfigurationDictionary::verifySampleRateIsSupported(
	UInt8			interfaceNum,
	UInt8			altSettingID,
	UInt32			verifyRate
)
{
	bool			result = false;
	OSObject		*arrayValue				= NULL;
	OSArray			*sampleRates			= NULL;
	OSNumber		*sampleRateNumber		= NULL;
	OSNumber		*sampleRateNumberHigh	= NULL;
	UInt8			numSampleRates;

	FailIf(kIOReturnSuccess != this->getNumSampleRates(&numSampleRates, interfaceNum, altSettingID), Exit);
	FailIf(NULL == (sampleRates = this->getSampleRates(interfaceNum, altSettingID)), Exit);

	if (numSampleRates)
	{
		// There are a discrete number of sample rates supported, so check for the desired sample rate.
		for (UInt8 sampleRateIndex = 0; (sampleRateIndex < numSampleRates) && (!result); sampleRateIndex++)
		{
			FailIf(NULL == (arrayValue = sampleRates->getObject(sampleRateIndex)), Exit);
			FailIf(NULL == (sampleRateNumber = OSDynamicCast(OSNumber, arrayValue)), Exit);

			if (sampleRateNumber->unsigned32BitValue() == verifyRate)
			{
				result = true;
			}
		}
	}
	else
	{
		// There is a range of sample rates supported, so check for the desired sample rate within that range.
		FailIf(NULL == (arrayValue = sampleRates->getObject(0)), Exit);
		FailIf(NULL == (sampleRateNumber = OSDynamicCast(OSNumber, arrayValue)), Exit);

		FailIf(NULL == (arrayValue = sampleRates->getObject(1)), Exit);
		FailIf(NULL == (sampleRateNumberHigh = OSDynamicCast(OSNumber, arrayValue)), Exit);

		if (
				(sampleRateNumber->unsigned32BitValue() <= verifyRate)
		   &&
				(sampleRateNumberHigh->unsigned32BitValue() >= verifyRate)
		   )
		{
			result = true;
		}
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface が１つ以上存在するか調べる．
 *
 * @retval			true			AudioStreaming Interface が１つ以上存在する．
 * @retval			false			AudioStreaming Interface が１つも存在しない．
 */
bool
PdjUsbConfigurationDictionary::hasAudioStreamingInterfaces(
	void
)
{
	bool			result = false;

	OSArray			*controlDictionaries = this->getControlDictionaries();

	if ((NULL != controlDictionaries) && (0 != controlDictionaries->getCount()))
	{
		result = true;
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	インスタンス初期化．
 *
 * @param[in]		*newConfigurationDescriptor		USB コンフィグレーションディスクリプタ．
 * @param[in]		controlInterfaceNum				AudioControl Interface 番号．
 *
 * @retval			true			成功．
 * @retval			false			失敗．
 */
bool
PdjUsbConfigurationDictionary::init(
	const StandardUSB::ConfigurationDescriptor	*newConfigurationDescriptor,
	UInt8										controlInterfaceNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONFIGURATIONDICTIONARY "[%p]::init(%p, %d)", this, newConfigurationDescriptor, controlInterfaceNum);

	StandardUSB::ConfigurationDescriptor	*mutableDescriptor = NULL;
	bool									result = false;

	FailIf(!super::initDictionaryForUse(), Exit);

	FailIf(NULL == newConfigurationDescriptor, Exit);

	FailIf(kIOReturnSuccess != super::setDictionaryValue(kControlInterfaceNumber, controlInterfaceNum), Exit);

	DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::init() - Allocating %d bytes for mutable config descriptor.",
		this, USBToHostWord(newConfigurationDescriptor->wTotalLength));

	mutableDescriptor = (StandardUSB::ConfigurationDescriptor *)IOMalloc(USBToHostWord(newConfigurationDescriptor->wTotalLength) + 1);
	FailIf(NULL == mutableDescriptor, Exit);

	::memcpy(mutableDescriptor, newConfigurationDescriptor, USBToHostWord(newConfigurationDescriptor->wTotalLength));

	((UInt8 *)mutableDescriptor)[USBToHostWord(newConfigurationDescriptor->wTotalLength)] = 0;

#ifdef DEBUGLOGGING
	this->dumpConfigMemoryToIOLog(mutableDescriptor);
#endif

	FailIf(kIOReturnSuccess != this->parseConfigurationDescriptor(mutableDescriptor), Exit);

	result = true;

Exit:
	if (NULL != mutableDescriptor)
	{
		IOFree(mutableDescriptor, USBToHostWord(newConfigurationDescriptor->wTotalLength) + 1);
		mutableDescriptor = NULL;
	}

	DP(DBG_TRACE, "- " LOGSTRCONFIGURATIONDICTIONARY "[%p]::init() = 0x%x", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	コンフィグレーションディスクリプタを解析．
 *
 * @param[in]		*configurationDescriptor	コンフィグレーションディスクリプタ．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::parseConfigurationDescriptor(
	StandardUSB::ConfigurationDescriptor	*configurationDescriptor
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor(%p)", this, configurationDescriptor);

	IOReturn								result = kIOReturnError;
	USBInterfaceDescriptorPtr				theInterfacePtr = NULL;
	USBInterfaceAssociationDescriptorPtr	theInterfaceAssociationPtr = NULL;
	PdjUsbControlDictionary					*controlDictionary = NULL;
	PdjUsbStreamDictionary					*streamDictionary = NULL;
	OSArray									*streamInterfaceNumbers = NULL;
	OSArray									*controlDictionaries = NULL;
	OSArray									*streamDictionaries = NULL;
	OSObject								*arrayObject = NULL;
	OSNumber								*streamInterfaceNumber = NULL;
	UInt8									controlInterfaceNum;
	UInt8									thisInterfaceNumber;
	UInt8									lastInterfaceNumber = 0;
	UInt8									numStreamInterfaces = 0;
	UInt8									interfaceClass;
	UInt8									interfaceSubClass = 0;
	UInt8									interfaceProtocol = 0;
	UInt8									numParsedInterfaces = 0;
	// [rdar://5346021] Keep strack of the descriptor length to guard against malformed descriptors.
	UInt16									totalLength;
	UInt16									parsedLength;
	bool									haveControlInterface = false;
	bool									foundStreamInterface = false;

	FailIf(NULL == configurationDescriptor, Exit);
	FailIf(0 == configurationDescriptor->bLength, Exit);
	FailIf(CONFIGURATION != configurationDescriptor->bDescriptorType, Exit);
	FailIf(kIOReturnSuccess != (result = this->getControlInterfaceNum(&controlInterfaceNum)), Exit);
	FailIf(0 == (totalLength = USBToHostWord(configurationDescriptor->wTotalLength)), Exit);

	theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)configurationDescriptor + configurationDescriptor->bLength);

	// [rdar://5346021] In keeping track of the parsed length, we add the length of the descriptor before actually parsing it. Then we check
	// to ensure that the total length is not exceeded before attempting to parse.
	parsedLength = (theInterfacePtr ? (configurationDescriptor->bLength + theInterfacePtr->bLength) : 0);

	while (
				(theInterfacePtr)
		  &&
				(0 != theInterfacePtr->bLength)
		  &&
				(parsedLength <= totalLength)
		  )
	{
		this->logDescriptor((UInt8 *)theInterfacePtr, theInterfacePtr->bLength);

		if (INTERFACE_ASSOCIATION == theInterfacePtr->bDescriptorType)
		{
			DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - @ INTERFACE_ASSOCIATION (4.6)", this);

			if (
					(((USBInterfaceAssociationDescriptorPtr)theInterfacePtr)->bFunctionClass == USBAUDIO_0200::AUDIO_FUNCTION)
			   &&
					(((USBInterfaceAssociationDescriptorPtr)theInterfacePtr)->bFunctionSubClass == USBAUDIO_0200::FUNCTION_SUBCLASS_UNDEFINED)
			   &&
					(((USBInterfaceAssociationDescriptorPtr)theInterfacePtr)->bFunctionProtocol == USBAUDIO_0200::AF_VERSION_02_00)
			   )
			{
				DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - interfaceClass = first interface = %d, interface count = %d",
					this,
					((USBInterfaceAssociationDescriptorPtr)theInterfacePtr)->bFirstInterface,
					((USBInterfaceAssociationDescriptorPtr)theInterfacePtr)->bInterfaceCount);

				theInterfaceAssociationPtr = (USBInterfaceAssociationDescriptorPtr)theInterfacePtr;
			}

			theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

			parsedLength = (theInterfacePtr ? (parsedLength + theInterfacePtr->bLength) : totalLength);
		}
		else if (INTERFACE == ((ACInterfaceDescriptorPtr)theInterfacePtr)->bDescriptorType)
		{
			DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - @ INTERFACE (4.3.1/4.5.1)", this);

			thisInterfaceNumber	 = ((ACInterfaceDescriptorPtr)theInterfacePtr)->bInterfaceNumber;

			if (AUDIO == ((ACInterfaceDescriptorPtr)theInterfacePtr)->bInterfaceClass)
			{
				theInterfacePtr = this->parseInterfaceDescriptor(theInterfacePtr, &interfaceClass, &interfaceSubClass, &interfaceProtocol);

				parsedLength = (theInterfacePtr ? (parsedLength + theInterfacePtr->bLength) : totalLength);

				DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - controlInterfaceNum = %d, thisInterfaceNumber = %d",
					this, controlInterfaceNum, thisInterfaceNumber);

				if ((AUDIOCONTROL == interfaceSubClass) && (controlInterfaceNum == thisInterfaceNumber))
				{
					DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Found an AUDIOCONTROL CS_INTERFACE (4.3.2)", this);

					if (NULL != (controlDictionaries = this->getControlDictionaries()))
					{
						controlDictionary = OSDynamicCast(PdjUsbControlDictionary, controlDictionaries->getLastObject());
					}

					FailIf(NULL == controlDictionary, Exit);

					// Safeguard against USB-Audio 2.0 descriptors for rdar://4798933
					if (INTERFACE_PROTOCOL_UNDEFINED == interfaceProtocol)
					{
						UInt8		numEndpoints;	// <rdar://problem/6021475>

						FailIf(NULL == theInterfacePtr, Exit);

						theInterfacePtr = controlDictionary->parseACInterfaceDescriptor(
																	theInterfacePtr,
																	((ACInterfaceDescriptorPtr)theInterfacePtr)->bInterfaceNumber,
																	&parsedLength,
																	totalLength
															 );

						FailIf(kIOReturnSuccess != (result = this->getControlledStreamNumbers(&streamInterfaceNumbers, &numStreamInterfaces)), Exit);

						haveControlInterface = true;

						// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
						DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Finished parsing AC Interface Descriptor", this);

						if (kIOReturnSuccess == controlDictionary->getNumEndpoints(&numEndpoints))
						{
							if (1 == numEndpoints)
							{
								this->logDescriptor((UInt8 *)theInterfacePtr, theInterfacePtr->bLength);

								theInterfacePtr = controlDictionary->parseACInterruptEndpointDescriptor(theInterfacePtr);
							}
							else
							{
								DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - numEndpoints: %d", this, numEndpoints);
							}
						}
					}
					else if (IP_VERSION_02_00 == interfaceProtocol)
					{
						UInt8		numEndpoints;	// <rdar://problem/6021475>

						FailIf(NULL == theInterfacePtr, Exit);

						// Parse USB-Audio 2.0 descriptors.
						theInterfacePtr = controlDictionary->parseACInterfaceDescriptor_0200(
																	theInterfacePtr,
																	((ACInterfaceDescriptorPtr)theInterfacePtr)->bInterfaceNumber,
																	&parsedLength,
																	totalLength
															 );

						// Parse the interface association descriptor to determine the interfaces used by this function.
						controlDictionary->parseInterfaceAssociationDescriptor(theInterfaceAssociationPtr);

						FailIf(kIOReturnSuccess != (result = this->getControlledStreamNumbers(&streamInterfaceNumbers, &numStreamInterfaces)), Exit);

						haveControlInterface = true;

						// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
						// <rdar://problem/5811247> Clock support
						DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Finished parsing AC Interface Descriptor", this);

						if (kIOReturnSuccess == controlDictionary->getNumEndpoints(&numEndpoints))
						{
							if (1 == numEndpoints)
							{
								this->logDescriptor((UInt8 *)theInterfacePtr, theInterfacePtr->bLength);

								theInterfacePtr = controlDictionary->parseACInterruptEndpointDescriptor(theInterfacePtr);
							}
							else
							{
								DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - numEndpoints: %d", this, numEndpoints);
							}
						}
					}
					else
					{
						// currently don't support this interface protocol, so abort out of the while loop
						break;
					}
				}
				else if (haveControlInterface && (AUDIOSTREAMING == interfaceSubClass))
				{
					DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - @ AUDIOSTREAMING CS_INTERFACE (4.5.2)", this);

					FailIf(kIOReturnSuccess != (result = this->getControlledStreamNumbers(&streamInterfaceNumbers, &numStreamInterfaces)), Exit);

					for (UInt8 streamInterfaceIndex = 0; streamInterfaceIndex < numStreamInterfaces; streamInterfaceIndex++)
					{
						FailIf(NULL == (arrayObject = streamInterfaceNumbers->getObject(streamInterfaceIndex)), Exit);
						FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

						DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Comparing thisInterfaceNum = %d with %d",
							this, thisInterfaceNumber, streamInterfaceNumber->unsigned8BitValue());

						if (thisInterfaceNumber == streamInterfaceNumber->unsigned8BitValue())
						{
							DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Found a AUDIOSTREAMING CS_INTERFACE", this);

							if (NULL != (streamDictionaries = this->getStreamDictionaries()))
							{
								streamDictionary = OSDynamicCast(PdjUsbStreamDictionary, streamDictionaries->getLastObject());
							}

							FailIf(NULL == streamDictionary, Exit);

							if (INTERFACE_PROTOCOL_UNDEFINED == interfaceProtocol)
							{
								FailIf(NULL == theInterfacePtr, Exit);

								theInterfacePtr = streamDictionary->parseASInterfaceDescriptor(
																		theInterfacePtr,
																		((ACInterfaceDescriptorPtr)theInterfacePtr)->bInterfaceNumber,
																		&parsedLength,
																		totalLength
																	);
							}
							else if (IP_VERSION_02_00 == interfaceProtocol)
							{
								FailIf(NULL == theInterfacePtr, Exit);

								theInterfacePtr = streamDictionary->parseASInterfaceDescriptor_0200(
																		theInterfacePtr,
																		((ACInterfaceDescriptorPtr)theInterfacePtr)->bInterfaceNumber,
																		&parsedLength,
																		totalLength
																	);
							}

							foundStreamInterface = true;

							break;		// Get out of for loop
						}
					}

					if (thisInterfaceNumber != lastInterfaceNumber)
					{
						lastInterfaceNumber = thisInterfaceNumber;

						numParsedInterfaces++;

						if (numParsedInterfaces > numStreamInterfaces)
						{
							break;		// Get out of while loop, we've parsed everything associated with this control interface
						}
					}
				}
				else if (MIDISTREAMING == interfaceSubClass)
				{
					FailIf(NULL == theInterfacePtr, Exit);

					DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - MIDI, jumping forward %d bytes",
						this, theInterfacePtr->bLength);

					// We should prune the stream interfaces since this driver does not handle MIDI.
					for (UInt8 streamInterfaceIndex = 0; streamInterfaceIndex < numStreamInterfaces; streamInterfaceIndex++)
					{
						FailIf(NULL == (arrayObject = streamInterfaceNumbers->getObject(streamInterfaceIndex)), Exit);
						FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

						if (thisInterfaceNumber == streamInterfaceNumber->unsigned8BitValue())
						{
							streamInterfaceNumbers->removeObject(streamInterfaceIndex);
						}
					}

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					parsedLength = (theInterfacePtr ? (parsedLength + theInterfacePtr->bLength) : totalLength);
				}
				else if (AUDIOCONTROL == interfaceSubClass)
				{
					if (INTERFACE_PROTOCOL_UNDEFINED == interfaceProtocol)
					{
						FailIf(NULL == theInterfacePtr, Exit);

						DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Found a control interface that we don't care about. Skipping %d bytes ...",
							this, ((((ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->wTotalLength[1] << 8) | (((ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->wTotalLength[0])));

						parsedLength = (
											parsedLength
										-
											theInterfacePtr->bLength
										+
											(
												(((ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->wTotalLength[1] << 8)
											|
												(((ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->wTotalLength[0] << 0)
											)
									   );

						theInterfacePtr = (USBInterfaceDescriptorPtr)(
												(UInt8 *)theInterfacePtr
										  +
												(
													(((ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->wTotalLength[1] << 8)
												|
													(((ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->wTotalLength[0] << 0)
												)
										  );

						parsedLength = (theInterfacePtr ? (parsedLength + theInterfacePtr->bLength) : totalLength);
					}
					else if (IP_VERSION_02_00 == interfaceProtocol)
					{
						FailIf(NULL == theInterfacePtr, Exit);

						DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Found a control interface that we don't care about. Skipping %d bytes ...",
							this, USBToHostWord(((USBAUDIO_0200::ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->wTotalLength));

						parsedLength = (
											parsedLength
										-
											theInterfacePtr->bLength
										+
											USBToHostWord(((USBAUDIO_0200::ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->wTotalLength)
									   );

						theInterfacePtr = (USBInterfaceDescriptorPtr)(
												(UInt8 *)theInterfacePtr
										  +
												USBToHostWord(((USBAUDIO_0200::ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->wTotalLength)
										  );

						parsedLength = (theInterfacePtr ? (parsedLength + theInterfacePtr->bLength) : totalLength);
					}
					else
					{
						// currently don't support this interface protocol, so abort out of the while loop
						break;
					}
				}
				else
				{
					FailIf(NULL == theInterfacePtr, Exit);

					DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Unknown, skipping %d bytes",
						this, theInterfacePtr->bLength);

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					parsedLength = (theInterfacePtr ? (parsedLength + theInterfacePtr->bLength) : totalLength);
				}
			} // if (AUDIO == ((ACInterfaceDescriptorPtr)theInterfacePtr)->bInterfaceClass)
			else
			{
				DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Not an audio interface, skipping %d bytes",
					this, theInterfacePtr->bLength);

				theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

				parsedLength = (theInterfacePtr ? (parsedLength + theInterfacePtr->bLength) : totalLength);
			}
		} // if (INTERFACE == ((ACInterfaceDescriptorPtr)theInterfacePtr)->bDescriptorType)
		else
		{
			DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseConfigurationDescriptor() - Default, skipping %d bytes",
				this, theInterfacePtr->bLength);

			theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

			parsedLength = (theInterfacePtr ? (parsedLength + theInterfacePtr->bLength) : totalLength);
		}
	} // while (theInterfacePtr && (0 != theInterfacePtr->bLength))

	if (parsedLength > totalLength)
	{
		IOLog("DJM-V10 USBAudio encountered an invalid descriptor on an attached USB audio device. The device may not function properly.\n");

		DP(DBG_ERROR, "! " LOGSTRCONFIGURATIONDICTIONARY "::parseConfigurationDescriptor() - Encountered a bad descriptor. Halting the parser ...");
	}

	if (controlDictionary && controlDictionaries && !foundStreamInterface)
	{
		controlDictionaries->removeObject(controlDictionaries->getCount() - 1);
	}

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRCONFIGURATIONDICTIONARY "::parseConfigurationDescriptor() = 0x%x", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	インターフェイスディスクリプタを解析．
 *
 * @param[in]		theInterfacePtr		インターフェイスディスクリプタ．
 * @param[out]		*interfaceClass		インターフェイスディスクリプタの bInterfaceClass フィールド．
 * @param[out]		*interfaceSubClass	インターフェイスディスクリプタの bInterfaceSubClass フィールド．
 * @param[out]		*interfaceProtocol	インターフェイスディスクリプタの bInterfaceProtocol フィールド．
 *
 * @return			次のディスクリプタへのポインタ．
 */
USBInterfaceDescriptorPtr
PdjUsbConfigurationDictionary::parseInterfaceDescriptor(
	USBInterfaceDescriptorPtr	theInterfacePtr,
	UInt8						*interfaceClass,
	UInt8						*interfaceSubClass,
	UInt8						*interfaceProtocol
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseInterfaceDescriptor(%p, %p, %p)",
		this, theInterfacePtr, interfaceClass, interfaceSubClass);

	PdjUsbControlDictionary		*controlDictionary		= NULL;
	PdjUsbStreamDictionary		*streamDictionary		= NULL;
	OSArray						*controlDictionaries	= NULL;
	OSArray						*streamDictionaries		= NULL;
	UInt8						controlInterfaceNum;

	FailIf(NULL == theInterfacePtr, Exit);
	FailIf(0 == theInterfacePtr->bLength, Exit);
	FailIf(kIOReturnSuccess != this->getControlInterfaceNum(&controlInterfaceNum), Exit);

	if (NULL != interfaceClass)
	{
		*interfaceClass = theInterfacePtr->bInterfaceClass;
	}

	if (NULL != interfaceSubClass)
	{
		*interfaceSubClass = theInterfacePtr->bInterfaceSubClass;
	}

	if (NULL != interfaceProtocol)
	{
		*interfaceProtocol = theInterfacePtr->bInterfaceProtocol;
	}

	if (AUDIOCONTROL == theInterfacePtr->bInterfaceSubClass)
	{
		DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseInterfaceDescriptor() - Found an AUDIOCONTROL interface", this);

		if (controlInterfaceNum == theInterfacePtr->bDescriptorSubtype)
		{
			controlDictionary = PdjUsbControlDictionary::create();

			FailIf(NULL == controlDictionary, Exit);

			controlDictionary->setInterfaceNumber(theInterfacePtr->bDescriptorSubtype);
			controlDictionary->setAlternateSetting(theInterfacePtr->bAlternateSetting);
			controlDictionary->setNumEndpoints(theInterfacePtr->bNumEndpoints);
			controlDictionary->setInterfaceClass(theInterfacePtr->bInterfaceClass);
			controlDictionary->setInterfaceSubClass(theInterfacePtr->bInterfaceSubClass);
			controlDictionary->setInterfaceProtocol(theInterfacePtr->bInterfaceProtocol);

			controlDictionaries = this->getControlDictionaries();

			if (NULL == controlDictionaries)
			{
				controlDictionaries = OSArray::withObjects((const OSObject **)&controlDictionary, 1);
				FailIf(NULL == controlDictionaries, Exit);

				FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kControlDictionaries, controlDictionaries), Exit);
				controlDictionaries = NULL;
			}
			else
			{
				controlDictionaries->setObject(controlDictionary);
			}

			controlDictionary->release();
			controlDictionary = NULL;
		}
	}
	else if (AUDIOSTREAMING == theInterfacePtr->bInterfaceSubClass)
	{
		DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseInterfaceDescriptor() - Found an AUDIOSTREAMING interface", this);

		streamDictionary = PdjUsbStreamDictionary::create();

		FailIf(NULL == streamDictionary, Exit);

		streamDictionary->setInterfaceNumber(theInterfacePtr->bDescriptorSubtype);
		streamDictionary->setAlternateSetting(theInterfacePtr->bAlternateSetting);
		streamDictionary->setNumEndpoints(theInterfacePtr->bNumEndpoints);
		streamDictionary->setInterfaceClass(theInterfacePtr->bInterfaceClass);
		streamDictionary->setInterfaceSubClass(theInterfacePtr->bInterfaceSubClass);
		streamDictionary->setInterfaceProtocol(theInterfacePtr->bInterfaceProtocol);

		streamDictionaries = this->getStreamDictionaries();

		if (NULL == streamDictionaries)
		{
			streamDictionaries = OSArray::withObjects((const OSObject **)&streamDictionary, 1);
			FailIf(NULL == streamDictionaries, Exit);

			FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kStreamDictionaries, streamDictionaries), Exit);
			streamDictionaries = NULL;
		}
		else
		{
			streamDictionaries->setObject(streamDictionary);
		}

		streamDictionary->release();
		streamDictionary = NULL;
	}

	theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

Exit:
	DP(DBG_TRACE, "- " LOGSTRCONFIGURATIONDICTIONARY "[%p]::parseInterfaceDescriptor() = %p", this, theInterfacePtr);
	return theInterfacePtr;
}

/****************************************************************************/
/**
 * @brief	チャンネル数を指定して Alternate Setting を探索．
 *
 * @param[out]		*altSettingID			見つかった alternate setting 番号．
 * @param[in]		interfaceNum			探索対象の interface 番号．
 * @param[in]		startingAltSettingID	探索を開始する alternate setting 番号．
 * @param[in]		numChannelsRequested	探索条件：チャンネル数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNextAltSettingWithNumChannels(
	UInt8			*altSettingID,
	UInt8			interfaceNum,
	UInt8			startingAltSettingID,
	UInt8			numChannelsRequested
)
{
	IOReturn		result = kIOReturnError;
	UInt8			numAltSettings;
	UInt8			numChannels;

	FailIf(NULL == altSettingID, Exit);

	*altSettingID = 0;

	FailIf(kIOReturnSuccess != this->getNumAltSettings(&numAltSettings, interfaceNum), Exit);

	for (
			UInt8 altSettingIndex = startingAltSettingID;
			(altSettingIndex < numAltSettings) && (kIOReturnSuccess != result);
			altSettingIndex++
		)
	{
		FailIf(kIOReturnSuccess != this->getNumChannels(&numChannels, interfaceNum, altSettingIndex), Exit);

		if (numChannelsRequested == numChannels)
		{
			*altSettingID = altSettingIndex;

			result = kIOReturnSuccess;
		}
	}

Exit:
	if ((altSettingID) && (kIOReturnSuccess == result))
	{
		DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getNextAltSettingWithNumChannels(%p, %d, %d, %d) = 0x%x, choosing altSetting %d",
			this, altSettingID, interfaceNum, startingAltSettingID, numChannelsRequested, result, *altSettingID);
	}
	else
	{
		DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getNextAltSettingWithNumChannels(%p, %d, %d, %d) = 0x%x, not found",
			this, altSettingID, interfaceNum, startingAltSettingID, numChannelsRequested, result);
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	サンプルサイズを指定して Alternate Setting を探索．
 *
 * @param[out]		*altSettingID			見つかった alternate setting 番号．
 * @param[in]		interfaceNum			探索対象の interface 番号．
 * @param[in]		startingAltSettingID	探索を開始する alternate setting 番号．
 * @param[in]		sampleSizeRequested		探索条件：サンプルサイズ．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNextAltSettingWithSampleSize(
	UInt8			*altSettingID,
	UInt8			interfaceNum,
	UInt8			startingAltSettingID,
	UInt8			sampleSizeRequested
)
{
	IOReturn		result = kIOReturnError;
	UInt8			numAltSettings;
	UInt8			sampleSize;

	FailIf(NULL == altSettingID, Exit);
	FailIf(kIOReturnSuccess != this->getNumAltSettings(&numAltSettings, interfaceNum), Exit);

	for (
			UInt8 altSettingIndex = startingAltSettingID;
			(altSettingIndex < numAltSettings) && (kIOReturnSuccess != result);
			altSettingIndex++
		)
	{
		FailIf(kIOReturnSuccess != this->getBitResolution(&sampleSize, interfaceNum, altSettingIndex), Exit);

		if (sampleSize == sampleSizeRequested)
		{
			*altSettingID = altSettingIndex;

			result = kIOReturnSuccess;
		}
	}

Exit:
	if ((altSettingID) && (kIOReturnSuccess == result))
	{
		DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getNextAltSettingWithSampleSize(%p, %d, %d, %d) = 0x%x, choosing altSetting %d",
			this, altSettingID, interfaceNum, startingAltSettingID, sampleSizeRequested, result, *altSettingID);
	}
	else
	{
		DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getNextAltSettingWithSampleSize(%p, %d, %d, %d) = 0x%x, not found",
			this, altSettingID, interfaceNum, startingAltSettingID, sampleSizeRequested, result);
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	サンプリング周波数を指定して Alternate Setting を探索．
 *
 * @param[out]		*altSettingID			見つかった alternate setting 番号．
 * @param[in]		interfaceNum			探索対象の interface 番号．
 * @param[in]		startingAltSettingID	探索を開始する alternate setting 番号．
 * @param[in]		sampleRateRequested		探索条件：サンプリング周波数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getNextAltSettingWithSampleRate(
	UInt8			*altSettingID,
	UInt8			interfaceNum,
	UInt8			startingAltSettingID,
	UInt32			sampleRateRequested
)
{
	IOReturn		result = kIOReturnError;
	UInt8			numAltSettings;

	FailIf(NULL == altSettingID, Exit);
	FailIf(kIOReturnSuccess != this->getNumAltSettings(&numAltSettings, interfaceNum), Exit);

	for (
			UInt8 altSettingIndex = startingAltSettingID;
			(altSettingIndex < numAltSettings) && (kIOReturnSuccess != result);
			altSettingIndex++
		)
	{
		if (this->verifySampleRateIsSupported(interfaceNum, altSettingIndex, sampleRateRequested))
		{
			*altSettingID = altSettingIndex;

			result = kIOReturnSuccess;
		}
	}

Exit:
	if ((altSettingID) && (kIOReturnSuccess == result))
	{
		DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getNextAltSettingWithSampleRate(%p, %d, %d, %u) = 0x%x, choosing altSetting %d",
			this, altSettingID, interfaceNum, startingAltSettingID, sampleRateRequested, result, *altSettingID);
	}
	else
	{
		DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "[%p]::getNextAltSettingWithSampleRate(%p, %d, %d, %u) = 0x%x, not found",
			this, altSettingID, interfaceNum, startingAltSettingID, sampleRateRequested, result);
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	AudioControl Interface の interface 番号を取得．
 *
 * @param[out]		*interfaceNum	AudioControl Interface の interface 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbConfigurationDictionary::getControlInterfaceNum(
	UInt8			*interfaceNum
)
{
	return super::getDictionaryValue(kControlInterfaceNumber, interfaceNum);
}

/*
IOReturn
PdjUsbConfigurationDictionary::getControlInterfaceNum(
	UInt8			*interfaceNum
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbControlDictionary		*thisControl = NULL;
	OSArray						*controlDictionaries = NULL;
	OSObject					*dictionaryValue = NULL;

	*interfaceNum = 0;

	FailIf(NULL == (dictionaryValue = OSDictionary::getObject(kControlDictionaries)), Exit);
	FailIf(NULL == (controlDictionaries = OSDynamicCast(OSArray, dictionaryValue)), Exit);
	FailIf(NULL == (thisControl = OSDynamicCast(PdjUsbControlDictionary, controlDictionaries->getObject(0))), Exit);
	FailIf(kIOReturnSuccess != (result = (thisControl->getInterfaceNumber(interfaceNum))), Exit);

Exit:
	return result;
}
*/

/****************************************************************************/
/**
 * @brief	コントロール dictionary を取得．
 *			コントロール dictionary は AudioCotrol Interface に対応する．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			コントロール dictionary．
 */
PdjUsbControlDictionary *
PdjUsbConfigurationDictionary::getControlDictionary(
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	PdjUsbControlDictionary		*thisControl		 = NULL;
	OSObject					*dictionaryValue	 = NULL;
	OSArray						*controlDictionaries = NULL;
	UInt8						controlIndex = 0;
	UInt8						controlInterfaceNum;
	UInt8						controlAltSettingID;
	bool						found = false;

	FailIf(NULL == (dictionaryValue = OSDictionary::getObject(kControlDictionaries)), Exit);
	FailIf(NULL == (controlDictionaries = OSDynamicCast(OSArray, dictionaryValue)), Exit);

	while ((!found) && (controlIndex < controlDictionaries->getCount()))
	{
		FailIf(NULL == (thisControl = OSDynamicCast(PdjUsbControlDictionary, controlDictionaries->getObject(controlIndex))), Exit);

		FailIf(kIOReturnSuccess != thisControl->getInterfaceNumber(&controlInterfaceNum), Exit);
		FailIf(kIOReturnSuccess != thisControl->getAlternateSetting(&controlAltSettingID), Exit);

		if ((interfaceNum == controlInterfaceNum) && (altSettingID == controlAltSettingID))
		{
			found = true;
		}

		controlIndex++;
	}

Exit:
	if (!found)
	{
		thisControl = NULL;
	}

	return thisControl;
}

/****************************************************************************/
/**
 * @brief	ストリーム dictionary を取得．
 *			ストリーム dictionary は AudioStreaming Interface に対応する．
 *
 * @param[in]		interfaceNum	interface 番号．
 * @param[in]		altSettingID	alternate setting 番号．
 *
 * @return			ストリーム dictionary．
 */
PdjUsbStreamDictionary *
PdjUsbConfigurationDictionary::getStreamDictionary(
	UInt8			interfaceNum,
	UInt8			altSettingID
)
{
	PdjUsbStreamDictionary		*thisStream			= NULL;
	OSObject					*dictionaryValue	= NULL;
	OSArray						*streamDictionaries	= NULL;
	UInt8						streamIndex = 0;
	UInt8						streamInterfaceNum;
	UInt8						streamAltSettingID;
	bool						found = false;

	FailIf(NULL == (dictionaryValue = OSDictionary::getObject(kStreamDictionaries)), Exit);
	FailIf(NULL == (streamDictionaries = OSDynamicCast(OSArray, dictionaryValue)), Exit);

	while ((!found) && (streamIndex < streamDictionaries->getCount()))
	{
		FailIf(NULL == (thisStream = OSDynamicCast(PdjUsbStreamDictionary, streamDictionaries->getObject(streamIndex))), Exit);

		FailIf(kIOReturnSuccess != thisStream->getInterfaceNumber(&streamInterfaceNum), Exit);
		FailIf(kIOReturnSuccess != thisStream->getAlternateSetting(&streamAltSettingID), Exit);

		if ((interfaceNum == streamInterfaceNum) && (altSettingID == streamAltSettingID))
		{
			found = true;
		}
//		else
//		{
//			DP(DBG_INFO, "? " LOGSTRCONFIGURATIONDICTIONARY "::getStreamDictionary() - (%d, %d) != (%d, %d)",
//				interfaceNum, altSettingID, streamInterfaceNum, streamAltSettingID);
//		}

		streamIndex++;
	}

Exit:
	if (!found)
	{
		thisStream = NULL;
	}

	return thisStream;
}

/****************************************************************************/
/**
 * @brief	USB ディスクリプタのバイト列をデバッグログに出力．
 *
 * @param[in]		*configurationDescriptor	コンフィグレーションディスクリプタ．
 */
#ifdef DEBUGLOGGING
void
PdjUsbConfigurationDictionary::dumpConfigMemoryToIOLog(
	StandardUSB::ConfigurationDescriptor	*configurationDescriptor
)
{
	UInt8			*descriptorPtr;
	UInt8			*finalPtr;
	UInt8			length;
	char			descriptor[256];
	char			num[4];

	descriptorPtr = (UInt8 *)configurationDescriptor;

	finalPtr = (UInt8 *)((uintptr_t)configurationDescriptor + USBToHostWord(configurationDescriptor->wTotalLength));

	while (
				(0 != *descriptorPtr)
		  &&
				(((uintptr_t)descriptorPtr + *descriptorPtr) <= (uintptr_t)finalPtr)
		  )
	{
		length = *descriptorPtr;

		descriptor[0] = '\0';

		for (UInt8 descriptorIndex = 0; descriptorIndex < length; descriptorIndex++)
		{
			::snprintf(num, 4, "%02X ", *descriptorPtr++);

			::strlcat(descriptor, num, sizeof(descriptor));
		}

		DP(DBG_INFO, "%s", descriptor);
	}
}
#endif

/*------------------------------------------*/
/*		PdjUsbControlDictionary クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbControlDictionary -

OSDefineMetaClassAndStructors(PdjUsbControlDictionary, super)

/****************************************************************************/
/**
 * @brief	コントロール dictionary を生成．
 *			コントロール dictionary は AudioCotrol Interface に対応する．
 *
 * @return			生成されたコントロール dictionary．
 */
PdjUsbControlDictionary *
PdjUsbControlDictionary::create(
	void
)
{
	PdjUsbControlDictionary		*controlDictionary;

	FailIf(NULL == (controlDictionary = new PdjUsbControlDictionary), Exit);

	RELEASE_IF_FALSE(controlDictionary, controlDictionary->initDictionaryForUse());

Exit:
	return controlDictionary;
}

/****************************************************************************/
/**
 * @brief	Interface Association Descriptor を解析．
 *
 * @param[in]		theInterfaceAssociationPtr	Interface Association Descriptor．
 */
void
PdjUsbControlDictionary::parseInterfaceAssociationDescriptor(
	USBInterfaceAssociationDescriptorPtr	theInterfaceAssociationPtr
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONTROLDICTIONARY "[%p]::parseInterfaceAssociationDescriptor(%p)", this, theInterfaceAssociationPtr);

	OSNumber		*dictionaryNumber		= NULL;
	OSArray			*streamInterfaceNumbers	= NULL;
	UInt8			numStreamInterfaces;

	FailIf(NULL == theInterfaceAssociationPtr, Exit);
	FailIf(0 == theInterfaceAssociationPtr->bLength, Exit);
	FailIf(INTERFACE_ASSOCIATION != theInterfaceAssociationPtr->bDescriptorType, Exit);
	FailIf(USBAUDIO_0200::AUDIO_FUNCTION != theInterfaceAssociationPtr->bFunctionClass, Exit);
	FailIf(USBAUDIO_0200::FUNCTION_SUBCLASS_UNDEFINED != theInterfaceAssociationPtr->bFunctionSubClass, Exit);
	FailIf(USBAUDIO_0200::AF_VERSION_02_00 != theInterfaceAssociationPtr->bFunctionProtocol, Exit);

	// Store the number of stream interfaces
	numStreamInterfaces = theInterfaceAssociationPtr->bInterfaceCount;
	FailIf(0 == numStreamInterfaces, Exit);

	numStreamInterfaces -= 1;	// excluding the AudioControl interface.
	DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseInterfaceAssociationDescriptor() - numStreamInterfaces = %d",
		this, numStreamInterfaces);

	FailIf(kIOReturnSuccess != super::setDictionaryValue(kNumStreamInterfaces, numStreamInterfaces), Exit);

	// Create and store stream interface number array
	FailIf(NULL == (streamInterfaceNumbers = OSArray::withCapacity(numStreamInterfaces)), Exit);
	// FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kStreamInterfaceNumbers, streamInterfaceNumbers), Exit);
	// FailIf(NULL == (streamInterfaceNumbers = this->getStreamInterfaceNumbers()), Exit);

	DP(DBG_INFO, "    they are: ");

	for (UInt8 index = 1; index <= numStreamInterfaces; index++)
	{
		DP(DBG_INFO, "      %d ", theInterfaceAssociationPtr->bFirstInterface + index);

		// Add this streamInterfaceNumber to the OSArray.
		FailIf(
					NULL == (
								dictionaryNumber = OSNumber::withNumber(
																	theInterfaceAssociationPtr->bFirstInterface + index,
																	SIZEINBITS(UInt8)
															 )
							),
					Exit
			  );

		FailIf(!(streamInterfaceNumbers->setObject(dictionaryNumber)), Exit);

		dictionaryNumber->release();
		dictionaryNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kStreamInterfaceNumbers, streamInterfaceNumbers), Exit);

	DP(DBG_INFO, "");

Exit:
	DP(DBG_TRACE, "- " LOGSTRCONTROLDICTIONARY "[%p]::parseInterfaceAssociationDescriptor() = %p", this, theInterfaceAssociationPtr);
	return;
}

/****************************************************************************/
/**
 * @brief	AudioControl Interface ディスクリプタを解析（UAC 1.0 用）．
 *
 * @param[in]		theInterfacePtr		AudioControl Interface ディスクリプタ．
 * @param[in]		currentInterface	interface 番号．
 * @param[in, out]	*parsedLength		コンフィグレーションディスクリプタのうち、解析済みのバイト数．
 * @param[in]		totalLength			コンフィグレーションディスクリプタの総バイト数．
 *
 * @return			次のディスクリプタへのポインタ．
 */
USBInterfaceDescriptorPtr
PdjUsbControlDictionary::parseACInterfaceDescriptor(
	USBInterfaceDescriptorPtr	theInterfacePtr,
	UInt8 const					currentInterface,
	UInt16						*parsedLength,
	UInt16						totalLength
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor(%p, %d)",
		this, theInterfacePtr, currentInterface);

	OSArray								*inputTerminals			= NULL;
	OSArray								*outputTerminals		= NULL;
	OSArray								*featureUnits			= NULL;
	OSArray								*mixerUnits				= NULL;
	OSArray								*selectorUnits			= NULL;
	OSArray								*processingUnits		= NULL;
	OSArray								*extensionUnits			= NULL;
	PdjUsbInputTerminalDictionary		*inputTerminal			= NULL;
	PdjUsbOutputTerminalDictionary		*outputTerminal			= NULL;
	PdjUsbFeatureUnitDictionary			*featureUnit			= NULL;
	PdjUsbMixerUnitDictionary			*mixerUnit				= NULL;
	PdjUsbSelectorUnitDictionary		*selectorUnit			= NULL;
	PdjUsbProcessingUnitDictionary		*processingUnit			= NULL;
	PdjUsbExtensionUnitDictionary		*extensionUnit			= NULL;
	OSNumber							*dictionaryNumber		= NULL;
	OSArray								*streamInterfaceNumbers	= NULL;
	UInt8								*streamInterfacesPtr	= NULL;
	UInt16								adcVersion;
	UInt8								numStreamInterfaces;

	FailIf(NULL == theInterfacePtr, Exit);
	FailIf(0 == theInterfacePtr->bLength, Exit);
	FailIf(CS_INTERFACE != theInterfacePtr->bDescriptorType, Exit);

	while (
				(theInterfacePtr)
		  &&
				(theInterfacePtr->bLength > 0)
		  &&
				(CS_INTERFACE == theInterfacePtr->bDescriptorType)
		  &&
				(*parsedLength <= totalLength)
		  )
	{
		this->logDescriptor((UInt8 *)theInterfacePtr, theInterfacePtr->bLength);

		switch (theInterfacePtr->bDescriptorSubtype)
		{
			case HEADER:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - @ HEADER (4.3.2)", this);

				adcVersion = USBToHostWord(*(UInt16 *)(&(((ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->bcdADC[0])));

				if (kPdjUsbUsbSpec1_0 != adcVersion)
				{
					DP(DBG_ERROR, "! " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - Encountered a descriptor version (0x%x) that UsbAudio currently can't handle.",
						this, adcVersion);
				}

				// store adcVersion
				FailIf(kIOReturnSuccess != super::setDictionaryValue(kADCVersion, adcVersion), Exit);

				// Store the number of stream interfaces
				numStreamInterfaces = ((ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->bInCollection;
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - numStreamInterfaces = %d", this, numStreamInterfaces);
				FailIf(kIOReturnSuccess != super::setDictionaryValue(kNumStreamInterfaces, numStreamInterfaces), Exit);

				// Create and store stream interface number array
				FailIf(NULL == (streamInterfaceNumbers = OSArray::withCapacity(numStreamInterfaces)), Exit);
				FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kStreamInterfaceNumbers, streamInterfaceNumbers), Exit);
				FailIf(NULL == (streamInterfaceNumbers = this->getStreamInterfaceNumbers()), Exit);

				streamInterfacesPtr = ((ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->baInterfaceNr;

				DP(DBG_INFO, "    they are: ");

				for (UInt8 index = 0; index < numStreamInterfaces; index++)
				{
					DP(DBG_INFO, "      %d ", streamInterfacesPtr[index]);

					// Add this streamInterfaceNumber to the OSArray.
					FailIf(NULL == (dictionaryNumber = OSNumber::withNumber(streamInterfacesPtr[index], SIZEINBITS(UInt8))), Exit);

					streamInterfaceNumbers->setObject(dictionaryNumber);

					dictionaryNumber->release();
					dictionaryNumber = NULL;
				}

				DP(DBG_INFO, "");
				break;
			}
			case INPUT_TERMINAL:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - @ INPUT_TERMINAL (4.3.2.1)", this);

				FailIf(NULL == (inputTerminal = new PdjUsbInputTerminalDictionary), Exit);
				RELEASE_IF_FALSE(inputTerminal, inputTerminal->initDictionaryForUse());
				FailIf(NULL == inputTerminal, Exit);

				inputTerminal->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);
				inputTerminal->setUnitID(((ACInputTerminalDescriptorPtr)theInterfacePtr)->bTerminalID);
				inputTerminal->setTerminalType(USBToHostWord(((ACInputTerminalDescriptorPtr)theInterfacePtr)->wTerminalType));
				inputTerminal->setAssocTerminal(((ACInputTerminalDescriptorPtr)theInterfacePtr)->bAssocTerminal);
				inputTerminal->setNumChannels(((ACInputTerminalDescriptorPtr)theInterfacePtr)->bNrChannels);
				inputTerminal->setChannelConfig(USBToHostWord(((ACInputTerminalDescriptorPtr)theInterfacePtr)->wChannelConfig));
				inputTerminal->setStringIndex(((ACInputTerminalDescriptorPtr)theInterfacePtr)->iTerminal);		// <rdar://6394629>
				inputTerminal->setChannelNames(((ACInputTerminalDescriptorPtr)theInterfacePtr)->iChannelNames);	// <rdar://6430836>

				inputTerminals = this->getInputTerminals();

				if (NULL == inputTerminals)
				{
					FailIf(NULL == (inputTerminals = OSArray::withObjects((const OSObject **)&inputTerminal, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kInputTerminals, inputTerminals), Exit);
					inputTerminals = NULL;
				}
				else
				{
					inputTerminals->setObject(inputTerminal);
				}

				inputTerminal->release();
				inputTerminal = NULL;

				break;
			}
			case OUTPUT_TERMINAL:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - @ OUTPUT_TERMINAL (4.3.2.2)", this);

				FailIf(NULL == (outputTerminal = new PdjUsbOutputTerminalDictionary), Exit);
				RELEASE_IF_FALSE(outputTerminal, outputTerminal->initDictionaryForUse());
				FailIf(NULL == outputTerminal, Exit);

				outputTerminal->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);
				outputTerminal->setUnitID(((ACOutputTerminalDescriptorPtr)theInterfacePtr)->bTerminalID);
				outputTerminal->setTerminalType(USBToHostWord(((ACOutputTerminalDescriptorPtr)theInterfacePtr)->wTerminalType));
				outputTerminal->setAssocTerminal(((ACOutputTerminalDescriptorPtr)theInterfacePtr)->bAssocTerminal);
				outputTerminal->setSourceID(((ACOutputTerminalDescriptorPtr)theInterfacePtr)->bSourceID);
				outputTerminal->setStringIndex(((ACOutputTerminalDescriptorPtr)theInterfacePtr)->iTerminal);	// <rdar://6394629>

				outputTerminals = this->getOutputTerminals();

				if (NULL == outputTerminals)
				{
					FailIf(NULL == (outputTerminals = OSArray::withObjects((const OSObject **)&outputTerminal, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kOutputTerminals, outputTerminals), Exit);
					outputTerminals = NULL;
				}
				else
				{
					outputTerminals->setObject(outputTerminal);
				}

				outputTerminal->release();
				outputTerminal = NULL;

				break;
			}
			case FEATURE_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - @ FEATURE_UNIT (4.3.2.5)", this);

				UInt8		numControls;

				FailIf(NULL == (featureUnit = new PdjUsbFeatureUnitDictionary), Exit);
				RELEASE_IF_FALSE(featureUnit, featureUnit->initDictionaryForUse());
				FailIf(NULL == featureUnit, Exit);

				featureUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);
				featureUnit->setUnitID(((ACFeatureUnitDescriptorPtr)theInterfacePtr)->bUnitID);
				featureUnit->setSourceID(((ACFeatureUnitDescriptorPtr)theInterfacePtr)->bSourceID);
				featureUnit->setControlSize(((ACFeatureUnitDescriptorPtr)theInterfacePtr)->bControlSize);

				FailIf(0 == ((ACFeatureUnitDescriptorPtr)theInterfacePtr)->bControlSize, Exit);

				// subtract 7 because that's how many fields are guaranteed to be in the struct
				numControls = (
									(((ACFeatureUnitDescriptorPtr)theInterfacePtr)->bLength - 7)
							  /
									((ACFeatureUnitDescriptorPtr)theInterfacePtr)->bControlSize
							  );
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - There are %d controls on this feature unit",
					this, numControls);

				featureUnit->initControlsArray(&((ACFeatureUnitDescriptorPtr)theInterfacePtr)->bmaControls[0], numControls);
				featureUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));	// <rdar://6394629>

				featureUnits = this->getFeatureUnits();

				if (NULL == featureUnits)
				{
					FailIf(NULL == (featureUnits = OSArray::withObjects((const OSObject **)&featureUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kFeatureUnits, featureUnits), Exit);
					featureUnits = NULL;
				}
				else
				{
					featureUnits->setObject(featureUnit);
				}

				featureUnit->release();
				featureUnit = NULL;

				break;
			}
			case MIXER_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - @ MIXER_UNIT (4.3.2.3)", this);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - descriptor length = %d", this, theInterfacePtr->bLength);

				UInt32		controlSize;
				UInt16		*channelConfig;
				UInt8		nrChannels;

				FailIf(NULL == (mixerUnit = new PdjUsbMixerUnitDictionary), Exit);
				RELEASE_IF_FALSE(mixerUnit, mixerUnit->initDictionaryForUse());
				FailIf(NULL == mixerUnit, Exit);

				mixerUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				mixerUnit->setUnitID(((ACMixerUnitDescriptorPtr)theInterfacePtr)->bUnitID);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - unit ID = %d",
					this, ((ACMixerUnitDescriptorPtr)theInterfacePtr)->bUnitID);

				mixerUnit->setNumInPins(((ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - numInPins = %d",
					this, ((ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				mixerUnit->initSourceIDs(
								&((ACMixerUnitDescriptorPtr)theInterfacePtr)->baSourceID[0],
								((ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins
						   );

				nrChannels = ((ACMixerUnitDescriptorPtr)theInterfacePtr)->baSourceID[((ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins];
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - nrChannels = %d", this, nrChannels);
				mixerUnit->setNumChannels(nrChannels);

				channelConfig = (UInt16 *)&((ACMixerUnitDescriptorPtr)theInterfacePtr)->baSourceID[((ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 1];
				*channelConfig = USBToHostWord(*channelConfig);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - channelConfig = %d", this, *channelConfig);
				mixerUnit->setChannelConfig(*channelConfig);

				controlSize = ((ACMixerUnitDescriptorPtr)theInterfacePtr)->bLength - 10 - ((ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins;
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - controlSize = %u", this, controlSize);

				mixerUnit->initControlsArray(
								&((ACMixerUnitDescriptorPtr)theInterfacePtr)->baSourceID[((ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 4],
								controlSize
						   );

				mixerUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));		// <rdar://6394629>

				mixerUnit->setChannelNames(
								*(((UInt8 *)theInterfacePtr) + (((ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 8))
						   );	// <rdar://problem/6706026>

				mixerUnits = this->getMixerUnits();

				if (NULL == mixerUnits)
				{
					FailIf(NULL == (mixerUnits = OSArray::withObjects((const OSObject **)&mixerUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kMixerUnits, mixerUnits), Exit);
					mixerUnits = NULL;
				}
				else
				{
					mixerUnits->setObject(mixerUnit);
				}

				mixerUnit->release();
				mixerUnit = NULL;

				break;
			}
			case SELECTOR_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - @ SELECTOR_UNIT (4.3.2.4)", this);

				FailIf(NULL == (selectorUnit = new PdjUsbSelectorUnitDictionary), Exit);
				RELEASE_IF_FALSE(selectorUnit, selectorUnit->initDictionaryForUse());
				FailIf(NULL == selectorUnit, Exit);

				selectorUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				selectorUnit->setNumInPins(((ACSelectorUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				selectorUnit->setUnitID(((ACSelectorUnitDescriptorPtr)theInterfacePtr)->bUnitID);

				selectorUnit->initSourceIDs(
									&((ACSelectorUnitDescriptorPtr)theInterfacePtr)->baSourceID[0],
									((ACSelectorUnitDescriptorPtr)theInterfacePtr)->bNrInPins
							  );
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - numInPins on selector = %d",
					this, ((ACSelectorUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				selectorUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));	// <rdar://6394629>

				selectorUnits = this->getSelectorUnits();

				if (NULL == selectorUnits)
				{
					FailIf(NULL == (selectorUnits = OSArray::withObjects((const OSObject **)&selectorUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kSelectorUnits, selectorUnits), Exit);
					selectorUnits = NULL;
				}
				else
				{
					selectorUnits->setObject(selectorUnit);
				}

				selectorUnit->release();
				selectorUnit = NULL;

				break;
			}
			case PROCESSING_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - @ PROCESSING_UNIT (4.3.2.6)", this);

				UInt16		*channelConfig;
				UInt8		controlSize;
				UInt8		nrChannels;

				FailIf(NULL == (processingUnit = new PdjUsbProcessingUnitDictionary), Exit);
				RELEASE_IF_FALSE(processingUnit, processingUnit->initDictionaryForUse());
				FailIf(NULL == processingUnit, Exit);

				processingUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				processingUnit->setUnitID(((ACProcessingUnitDescriptorPtr)theInterfacePtr)->bUnitID);

				processingUnit->setProcessType(((ACProcessingUnitDescriptorPtr)theInterfacePtr)->wProcessType);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - processing unit type = 0x%x",
					this, ((ACProcessingUnitDescriptorPtr)theInterfacePtr)->wProcessType);

				processingUnit->setNumInPins(((ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				processingUnit->initSourceIDs(
									&((ACProcessingUnitDescriptorPtr)theInterfacePtr)->baSourceID[0],
									((ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins
								);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - numInPins = %d",
					this, ((ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				nrChannels = ((ACProcessingUnitDescriptorPtr)theInterfacePtr)->baSourceID[((ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins];
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - nrChannels = %d", this, nrChannels);
				processingUnit->setNumChannels(nrChannels);

				channelConfig = (UInt16 *)&((ACProcessingUnitDescriptorPtr)theInterfacePtr)->baSourceID[((ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 1];
				*channelConfig = USBToHostWord(*channelConfig);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - channelConfig = %d", this, *channelConfig);
				processingUnit->setChannelConfig(*channelConfig);

				controlSize = ((ACProcessingUnitDescriptorPtr)theInterfacePtr)->baSourceID[((ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 4];
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - controlSize = %d", this, controlSize);

				processingUnit->initControlsArray(
									&((ACProcessingUnitDescriptorPtr)theInterfacePtr)->baSourceID[((ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 5],
									controlSize
								);

				processingUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));	// <rdar://6394629>

				processingUnit->setChannelNames(
									*(((UInt8 *)theInterfacePtr) + (((ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 10))
								);	// <rdar://problem/6706026>

				processingUnits = this->getProcessingUnits();

				if (NULL == processingUnits)
				{
					FailIf(NULL == (processingUnits = OSArray::withObjects((const OSObject **)&processingUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kProcessingUnits, processingUnits), Exit);
					processingUnits = NULL;
				}
				else
				{
					processingUnits->setObject(processingUnit);
				}

				processingUnit->release();
				processingUnit = NULL;

				break;
			}
			case EXTENSION_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - @ EXTENSION_UNIT (4.3.2.7)", this);

				FailIf(NULL == (extensionUnit = new PdjUsbExtensionUnitDictionary), Exit);
				RELEASE_IF_FALSE(extensionUnit, extensionUnit->initDictionaryForUse());
				FailIf(NULL == extensionUnit, Exit);

				extensionUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				extensionUnit->setUnitID(((ACExtensionUnitDescriptorPtr)theInterfacePtr)->bUnitID);

				extensionUnit->setNumInPins(((ACExtensionUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				extensionUnit->initSourceIDs(
									&((ACExtensionUnitDescriptorPtr)theInterfacePtr)->baSourceID[0],
									((ACExtensionUnitDescriptorPtr)theInterfacePtr)->bNrInPins
							   );

				extensionUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));	// <rdar://6394629>

				extensionUnit->setChannelNames(
									*(((UInt8 *)theInterfacePtr) + (((ACExtensionUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 10))
							   );	// <rdar://problem/6706026>

				extensionUnits = this->getExtensionUnits();

				if (NULL == extensionUnits)
				{
					FailIf(NULL == (extensionUnits = OSArray::withObjects((const OSObject **)&extensionUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kExtensionUnits, extensionUnits), Exit);
					extensionUnits = NULL;
				}
				else
				{
					extensionUnits->setObject(extensionUnit);
				}

				extensionUnit->release();
				extensionUnit = NULL;

				break;
			}
			default:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() - @ default. Nothing to do here.", this);
				break;
			}
		}

		theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

		*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor() = %p", this, theInterfacePtr);
	return theInterfacePtr;
}

/****************************************************************************/
/**
 * @brief	AudioControl Interface ディスクリプタを解析（UAC 2.0 用）．
 *
 * @param[in]		theInterfacePtr		AudioControl Interface ディスクリプタ．
 * @param[in]		currentInterface	interface 番号．
 * @param[in, out]	*parsedLength		コンフィグレーションディスクリプタのうち、解析済みのバイト数．
 * @param[in]		totalLength			コンフィグレーションディスクリプタの総バイト数．
 *
 * @return			次のディスクリプタへのポインタ．
 */
USBInterfaceDescriptorPtr
PdjUsbControlDictionary::parseACInterfaceDescriptor_0200(
	USBInterfaceDescriptorPtr	theInterfacePtr,
	UInt8 const					currentInterface,
	UInt16						*parsedLength,
	UInt16						totalLength
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200(%p, %d)",
		this, theInterfacePtr, currentInterface);

	OSArray								*inputTerminals		= NULL;
	OSArray								*outputTerminals	= NULL;
	OSArray								*featureUnits		= NULL;
	OSArray								*mixerUnits			= NULL;
	OSArray								*selectorUnits		= NULL;
	OSArray								*effectUnits		= NULL;
	OSArray								*processingUnits	= NULL;
	OSArray								*extensionUnits		= NULL;
	OSArray								*clockSources		= NULL;
	OSArray								*clockSelectors		= NULL;
	OSArray								*clockMultipliers	= NULL;
	PdjUsbInputTerminalDictionary		*inputTerminal		= NULL;
	PdjUsbOutputTerminalDictionary		*outputTerminal		= NULL;
	PdjUsbFeatureUnitDictionary			*featureUnit		= NULL;
	PdjUsbMixerUnitDictionary			*mixerUnit			= NULL;
	PdjUsbSelectorUnitDictionary		*selectorUnit		= NULL;
	PdjUsbEffectUnitDictionary			*effectUnit			= NULL;
	PdjUsbProcessingUnitDictionary		*processingUnit		= NULL;
	PdjUsbExtensionUnitDictionary		*extensionUnit		= NULL;
	PdjUsbClockSourceDictionary			*clockSource		= NULL;
	PdjUsbClockSelectorDictionary		*clockSelector		= NULL;
	PdjUsbClockMultiplierDictionary		*clockMultiplier	= NULL;
	UInt16								adcVersion;

	FailIf(NULL == theInterfacePtr, Exit);
	FailIf(0 == theInterfacePtr->bLength, Exit);
	FailIf(CS_INTERFACE != theInterfacePtr->bDescriptorType, Exit);

	while (
				(theInterfacePtr)
		  &&
				(theInterfacePtr->bLength > 0)
		  &&
				(CS_INTERFACE == theInterfacePtr->bDescriptorType)
		  &&
				(*parsedLength <= totalLength)
		  )
	{
		this->logDescriptor((UInt8 *)theInterfacePtr, theInterfacePtr->bLength);

		switch (theInterfacePtr->bDescriptorSubtype)
		{
			case USBAUDIO_0200::HEADER:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ HEADER (4.7.2)", this);

				adcVersion = USBToHostWord(*(UInt16 *)(&(((USBAUDIO_0200::ACInterfaceHeaderDescriptorPtr)theInterfacePtr)->bcdADC[0])));

				if (kPdjUsbUsbSpec2_0 != adcVersion)
				{
					DP(DBG_ERROR, "! " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - Encountered a descriptor version (0x%x) that UsbAudio currently can't handle.",
						this, adcVersion);
				}

				// store adcVersion
				FailIf(kIOReturnSuccess != super::setDictionaryValue(kADCVersion, adcVersion), Exit);

				break;
			}
			case USBAUDIO_0200::INPUT_TERMINAL:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ INPUT_TERMINAL (4.7.2.4)", this);

				FailIf(NULL == (inputTerminal = new PdjUsbInputTerminalDictionary), Exit);
				RELEASE_IF_FALSE(inputTerminal, inputTerminal->initDictionaryForUse());
				FailIf(NULL == inputTerminal, Exit);

				inputTerminal->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);
				inputTerminal->setUnitID(((USBAUDIO_0200::ACInputTerminalDescriptorPtr)theInterfacePtr)->bTerminalID);
				inputTerminal->setTerminalType(USBToHostWord(((ACInputTerminalDescriptorPtr)theInterfacePtr)->wTerminalType));
				inputTerminal->setAssocTerminal(((USBAUDIO_0200::ACInputTerminalDescriptorPtr)theInterfacePtr)->bAssocTerminal);
				inputTerminal->setNumChannels(((USBAUDIO_0200::ACInputTerminalDescriptorPtr)theInterfacePtr)->bNrChannels);
				inputTerminal->setChannelConfig(USBToHostLong(*((UInt32 *)(&((USBAUDIO_0200::ACInputTerminalDescriptorPtr)theInterfacePtr)->bmChannelConfig[0]))));
				inputTerminal->setClockSourceID(((USBAUDIO_0200::ACInputTerminalDescriptorPtr)theInterfacePtr)->bCSourceID);
				inputTerminal->setStringIndex(((USBAUDIO_0200::ACInputTerminalDescriptorPtr)theInterfacePtr)->iTerminal);		// <rdar://6394629>
				inputTerminal->setChannelNames(((USBAUDIO_0200::ACInputTerminalDescriptorPtr)theInterfacePtr)->iChannelNames);	// <rdar://6430836>

				inputTerminals = this->getInputTerminals();

				if (NULL == inputTerminals)
				{
					FailIf(NULL == (inputTerminals = OSArray::withObjects((const OSObject **)&inputTerminal, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kInputTerminals, inputTerminals), Exit);
					inputTerminals = NULL;
				}
				else
				{
					inputTerminals->setObject(inputTerminal);
				}

				inputTerminal->release();
				inputTerminal = NULL;

				break;
			}
			case USBAUDIO_0200::OUTPUT_TERMINAL:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ OUTPUT_TERMINAL (4.7.2.5)", this);

				FailIf(NULL == (outputTerminal = new PdjUsbOutputTerminalDictionary), Exit);
				RELEASE_IF_FALSE(outputTerminal, outputTerminal->initDictionaryForUse());
				FailIf(NULL == outputTerminal, Exit);

				outputTerminal->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);
				outputTerminal->setUnitID(((USBAUDIO_0200::ACOutputTerminalDescriptorPtr)theInterfacePtr)->bTerminalID);
				outputTerminal->setTerminalType(USBToHostWord(((USBAUDIO_0200::ACOutputTerminalDescriptorPtr)theInterfacePtr)->wTerminalType));
				outputTerminal->setAssocTerminal(((USBAUDIO_0200::ACOutputTerminalDescriptorPtr)theInterfacePtr)->bAssocTerminal);
				outputTerminal->setSourceID(((USBAUDIO_0200::ACOutputTerminalDescriptorPtr)theInterfacePtr)->bSourceID);
				outputTerminal->setClockSourceID(((USBAUDIO_0200::ACOutputTerminalDescriptorPtr)theInterfacePtr)->bCSourceID);
				outputTerminal->setStringIndex(((USBAUDIO_0200::ACOutputTerminalDescriptorPtr)theInterfacePtr)->iTerminal);	// <rdar://6394629>

				outputTerminals = this->getOutputTerminals();

				if (NULL == outputTerminals)
				{
					FailIf(NULL == (outputTerminals = OSArray::withObjects((const OSObject **)&outputTerminal, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kOutputTerminals, outputTerminals), Exit);
					outputTerminals = NULL;
				}
				else
				{
					outputTerminals->setObject(outputTerminal);
				}

				outputTerminal->release();
				outputTerminal = NULL;

				break;
			}
			case USBAUDIO_0200::FEATURE_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ FEATURE_UNIT (4.7.2.8)", this);

				UInt8		numControls;

				FailIf(NULL == (featureUnit = new PdjUsbFeatureUnitDictionary), Exit);
				RELEASE_IF_FALSE(featureUnit, featureUnit->initDictionaryForUse());
				FailIf(NULL == featureUnit, Exit);

				featureUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				featureUnit->setUnitID(((USBAUDIO_0200::ACFeatureUnitDescriptorPtr)theInterfacePtr)->bUnitID);

				featureUnit->setSourceID(((USBAUDIO_0200::ACFeatureUnitDescriptorPtr)theInterfacePtr)->bSourceID);

				featureUnit->setControlSize(4);

				// subtract 6 because that's how many fields are guaranteed to be in the struct
				numControls = (((USBAUDIO_0200::ACFeatureUnitDescriptorPtr)theInterfacePtr)->bLength - 6) / 4;
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - There are %d controls on this feature unit",
					this, numControls);

				featureUnit->initControlsArray(
									&((USBAUDIO_0200::ACFeatureUnitDescriptorPtr)theInterfacePtr)->bmaControls[0],
									numControls
							 );

				featureUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));	// <rdar://6394629>

				featureUnits = this->getFeatureUnits();

				if (NULL == featureUnits)
				{
					FailIf(NULL == (featureUnits = OSArray::withObjects((const OSObject **)&featureUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kFeatureUnits, featureUnits), Exit);
					featureUnits = NULL;
				}
				else
				{
					featureUnits->setObject(featureUnit);
				}

				featureUnit->release();
				featureUnit = NULL;

				break;
			}
			case USBAUDIO_0200::MIXER_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ MIXER_UNIT (4.7.2.6)", this);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - descriptor length = %d",
					this, theInterfacePtr->bLength);

				UInt32		controlSize;
				UInt32		*channelConfig;
				UInt8		nrChannels;

				FailIf(NULL == (mixerUnit = new PdjUsbMixerUnitDictionary), Exit);
				RELEASE_IF_FALSE(mixerUnit, mixerUnit->initDictionaryForUse());
				FailIf(NULL == mixerUnit, Exit);

				mixerUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				mixerUnit->setUnitID(((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bUnitID);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - unit ID = %d",
					this, ((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bUnitID);

				mixerUnit->setNumInPins(((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - numInPins = %d",
					this, ((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				mixerUnit->initSourceIDs(
								&((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->baSourceID[0],
								((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins
						   );

				nrChannels = ((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->baSourceID[
																								((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins
																						 ];
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - nrChannels = %d", this, nrChannels);
				mixerUnit->setNumChannels(nrChannels);

				channelConfig = (UInt32 *)&((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->baSourceID[
																											((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 1
																									   ];
				*channelConfig = USBToHostLong(*channelConfig);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - channelConfig = %u", this, *channelConfig);
				mixerUnit->setChannelConfig(*channelConfig);

				controlSize = ((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bLength - 13 - ((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins;
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - controlSize = %u", this, controlSize);

				mixerUnit->initControlsArray(
								&((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->baSourceID[((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 6],
								controlSize
						   );

				mixerUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));		// <rdar://6394629>

				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - setChannelNames = 0x%X",
					this, *(((UInt8 *)theInterfacePtr) + (((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 10)));
				mixerUnit->setChannelNames(
								*(((UInt8 *)theInterfacePtr) + (((USBAUDIO_0200::ACMixerUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 10))
						   );		// <rdar://problem/6706026>

				mixerUnits = this->getMixerUnits();

				if (NULL == mixerUnits)
				{
					FailIf(NULL == (mixerUnits = OSArray::withObjects((const OSObject **)&mixerUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kMixerUnits, mixerUnits), Exit);
					mixerUnits = NULL;
				}
				else
				{
					mixerUnits->setObject(mixerUnit);
				}

				mixerUnit->release();
				mixerUnit = NULL;

				break;
			}
			case USBAUDIO_0200::SELECTOR_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ SELECTOR_UNIT (4.7.2.7)", this);

				FailIf(NULL == (selectorUnit = new PdjUsbSelectorUnitDictionary), Exit);
				RELEASE_IF_FALSE(selectorUnit, selectorUnit->initDictionaryForUse());
				FailIf(NULL == selectorUnit, Exit);

				selectorUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				selectorUnit->setNumInPins(((USBAUDIO_0200::ACSelectorUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				selectorUnit->setUnitID(((USBAUDIO_0200::ACSelectorUnitDescriptorPtr)theInterfacePtr)->bUnitID);

				selectorUnit->initSourceIDs(
									&((USBAUDIO_0200::ACSelectorUnitDescriptorPtr)theInterfacePtr)->baSourceID[0],
									((USBAUDIO_0200::ACSelectorUnitDescriptorPtr)theInterfacePtr)->bNrInPins
							  );
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - numInPins on selector = %d",
					this, ((USBAUDIO_0200::ACSelectorUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				selectorUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));	// <rdar://6394629>

				selectorUnits = this->getSelectorUnits();

				if (NULL == selectorUnits)
				{
					FailIf(NULL == (selectorUnits = OSArray::withObjects((const OSObject **)&selectorUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kSelectorUnits, selectorUnits), Exit);
					selectorUnits = NULL;
				}
				else
				{
					selectorUnits->setObject(selectorUnit);
				}

				selectorUnit->release();
				selectorUnit = NULL;

				break;
			}
			case USBAUDIO_0200::EFFECT_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ EFFECT_UNIT (4.7.2.10)", this);

				UInt8		numControls;

				FailIf(NULL == (effectUnit = new PdjUsbEffectUnitDictionary), Exit);
				RELEASE_IF_FALSE(effectUnit, effectUnit->initDictionaryForUse());
				FailIf(NULL == effectUnit, Exit);

				effectUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);
				effectUnit->setUnitID(((USBAUDIO_0200::ACEffectUnitDescriptorPtr)theInterfacePtr)->bUnitID);
				effectUnit->setEffectType(((USBAUDIO_0200::ACEffectUnitDescriptorPtr)theInterfacePtr)->wEffectType);
				effectUnit->setSourceID(((USBAUDIO_0200::ACEffectUnitDescriptorPtr)theInterfacePtr)->bSourceID);
				effectUnit->setControlSize(4);

				// subtract 8 because that's how many fields are guaranteed to be in the struct
				numControls = (((USBAUDIO_0200::ACEffectUnitDescriptorPtr)theInterfacePtr)->bLength - 8) / 4;
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - There are %d controls on this effect unit",
					this, numControls);

				effectUnit->initControlsArray(&((USBAUDIO_0200::ACEffectUnitDescriptorPtr)theInterfacePtr)->bmaControls[0][0], numControls);
				effectUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));	// <rdar://6394629>

				effectUnits = this->getEffectUnits();

				if (NULL == effectUnits)
				{
					FailIf(NULL == (effectUnits = OSArray::withObjects((const OSObject **)&effectUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kEffectUnits, effectUnits), Exit);
					effectUnits = NULL;
				}
				else
				{
					effectUnits->setObject(effectUnit);
				}

				effectUnit->release();
				effectUnit = NULL;

				break;
			}
			case USBAUDIO_0200::PROCESSING_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ PROCESSING_UNIT (4.7.2.11)", this);

				UInt32		*channelConfig;
				UInt8		controlSize;
				UInt8		nrChannels;

				FailIf(NULL == (processingUnit = new PdjUsbProcessingUnitDictionary), Exit);
				RELEASE_IF_FALSE(processingUnit, processingUnit->initDictionaryForUse());
				FailIf(NULL == processingUnit, Exit);

				processingUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				processingUnit->setUnitID(((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->bUnitID);

				processingUnit->setProcessType(((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->wProcessType);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - processing unit type = 0x%x",
					this, ((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->wProcessType);

				processingUnit->setNumInPins(((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				processingUnit->initSourceIDs(
									&((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->baSourceID[0],
									((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins
								);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - numInPins = %d",
					this, ((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				nrChannels = ((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->baSourceID[
																									((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins
																							  ];
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - nrChannels = %d", this, nrChannels);
				processingUnit->setNumChannels(nrChannels);

				channelConfig = (UInt32 *)&((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->baSourceID[
																												((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 1
																											];
				*channelConfig = USBToHostLong(*channelConfig);
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - channelConfig = %u", this, *channelConfig);
				processingUnit->setChannelConfig(*channelConfig);

				controlSize = 2;
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - controlSize = %d", this, controlSize);

				processingUnit->initControlsArray(
									&((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->baSourceID[
																											((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 6
																									  ],
									controlSize
								);

				processingUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));	// <rdar://6394629>

				processingUnit->setChannelNames(
									*(((UInt8 *)theInterfacePtr) + (((USBAUDIO_0200::ACProcessingUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 12))
								);	// <rdar://problem/6706026>

				processingUnits = this->getProcessingUnits();

				if (NULL == processingUnits)
				{
					FailIf(NULL == (processingUnits = OSArray::withObjects((const OSObject **)&processingUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kProcessingUnits, processingUnits), Exit);
					processingUnits = NULL;
				}
				else
				{
					processingUnits->setObject(processingUnit);
				}

				processingUnit->release();
				processingUnit = NULL;

				break;
			}
			case USBAUDIO_0200::EXTENSION_UNIT:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ EXTENSION_UNIT (4.7.2.12)", this);

				FailIf(NULL == (extensionUnit = new PdjUsbExtensionUnitDictionary), Exit);
				RELEASE_IF_FALSE(extensionUnit, extensionUnit->initDictionaryForUse());
				FailIf(NULL == extensionUnit, Exit);

				extensionUnit->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				extensionUnit->setUnitID(((USBAUDIO_0200::ACExtensionUnitDescriptorPtr)theInterfacePtr)->bUnitID);

				extensionUnit->setNumInPins(((USBAUDIO_0200::ACExtensionUnitDescriptorPtr)theInterfacePtr)->bNrInPins);

				extensionUnit->initSourceIDs(
									&((USBAUDIO_0200::ACExtensionUnitDescriptorPtr)theInterfacePtr)->baSourceID[0],
									((USBAUDIO_0200::ACExtensionUnitDescriptorPtr)theInterfacePtr)->bNrInPins
							   );

				extensionUnit->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));		// <rdar://6394629>

				extensionUnit->setChannelNames(
									*(((UInt8 *)theInterfacePtr) + (((USBAUDIO_0200::ACExtensionUnitDescriptorPtr)theInterfacePtr)->bNrInPins + 13))
							   );	// <rdar://problem/6706026>

				extensionUnits = this->getExtensionUnits();

				if (NULL == extensionUnits)
				{
					FailIf(NULL == (extensionUnits = OSArray::withObjects((const OSObject **)&extensionUnit, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kExtensionUnits, extensionUnits), Exit);
					extensionUnits = NULL;
				}
				else
				{
					extensionUnits->setObject(extensionUnit);
				}

				extensionUnit->release();
				extensionUnit = NULL;

				break;
			}
			case USBAUDIO_0200::CLOCK_SOURCE:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ CLOCK_SOURCE (4.7.2.1)", this);

				FailIf(NULL == (clockSource = new PdjUsbClockSourceDictionary), Exit);
				RELEASE_IF_FALSE(clockSource, clockSource->initDictionaryForUse());
				FailIf(NULL == clockSource, Exit);

				clockSource->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);
				clockSource->setUnitID(((USBAUDIO_0200::ACClockSourceDescriptorPtr)theInterfacePtr)->bClockID);
				clockSource->setAttributes(((USBAUDIO_0200::ACClockSourceDescriptorPtr)theInterfacePtr)->bmAttributes);
				clockSource->setAssocTerminal(((USBAUDIO_0200::ACClockSourceDescriptorPtr)theInterfacePtr)->bAssocTerminal);
				clockSource->initControlsArray(&((USBAUDIO_0200::ACClockSourceDescriptorPtr)theInterfacePtr)->bmControls, 1);
				clockSource->setStringIndex(((USBAUDIO_0200::ACClockSourceDescriptorPtr)theInterfacePtr)->iClockSource);	// <rdar://6394629>

				clockSources = this->getClockSources();

				if (NULL == clockSources)
				{
					FailIf(NULL == (clockSources = OSArray::withObjects((const OSObject **)&clockSource, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kClockSources, clockSources), Exit);
					clockSources = NULL;
				}
				else
				{
					clockSources->setObject(clockSource);
				}

				clockSource->release();
				clockSource = NULL;

				break;
			}
			case USBAUDIO_0200::CLOCK_SELECTOR:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ CLOCK_SELECTOR (4.7.2.2)", this);

				FailIf(NULL == (clockSelector = new PdjUsbClockSelectorDictionary), Exit);
				RELEASE_IF_FALSE(clockSelector, clockSelector->initDictionaryForUse());
				FailIf(NULL == clockSelector, Exit);

				clockSelector->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);

				clockSelector->setUnitID(((USBAUDIO_0200::ACClockSelectorDescriptorPtr)theInterfacePtr)->bClockID);

				clockSelector->setNumInPins(((USBAUDIO_0200::ACClockSelectorDescriptorPtr)theInterfacePtr)->bNrInPins);

				clockSelector->initClockSourceIDs(
									&((USBAUDIO_0200::ACClockSelectorDescriptorPtr)theInterfacePtr)->baCSourceID[0],
									((USBAUDIO_0200::ACClockSelectorDescriptorPtr)theInterfacePtr)->bNrInPins
							   );

				clockSelector->initControlsArray(
									&((USBAUDIO_0200::ACClockSelectorDescriptorPtr)theInterfacePtr)->baCSourceID[
																											((USBAUDIO_0200::ACClockSelectorDescriptorPtr)theInterfacePtr)->bNrInPins
																									 ],
									1
							   );

				clockSelector->setStringIndex(*(((UInt8 *)theInterfacePtr) + theInterfacePtr->bLength - 1));		// <rdar://6394629>

				clockSelectors = this->getClockSelectors();

				if (NULL == clockSelectors)
				{
					FailIf(NULL == (clockSelectors = OSArray::withObjects((const OSObject **)&clockSelector, 1)), Exit);
					FailIf(kIOReturnSuccess != setDictionaryObjectAndRelease(kClockSelectors, clockSelectors), Exit);
					clockSelectors = NULL;
				}
				else
				{
					clockSelectors->setObject(clockSelector);
				}

				clockSelector->release();
				clockSelector = NULL;

				break;
			}
			case USBAUDIO_0200::CLOCK_MULTIPLIER:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ CLOCK_MULTIPLIER (4.7.2.3)", this);

				FailIf(NULL == (clockMultiplier = new PdjUsbClockMultiplierDictionary), Exit);
				RELEASE_IF_FALSE(clockMultiplier, clockMultiplier->initDictionaryForUse());
				FailIf(NULL == clockMultiplier, Exit);

				clockMultiplier->setDescriptorSubType(theInterfacePtr->bDescriptorSubtype);
				clockMultiplier->setUnitID(((USBAUDIO_0200::ACClockMultiplierDescriptorPtr)theInterfacePtr)->bClockID);
				clockMultiplier->setClockSourceID(((USBAUDIO_0200::ACClockMultiplierDescriptorPtr)theInterfacePtr)->bCSourceID);
				clockMultiplier->initControlsArray(&((USBAUDIO_0200::ACClockMultiplierDescriptorPtr)theInterfacePtr)->bmControls, 1);
				clockMultiplier->setStringIndex(((USBAUDIO_0200::ACClockMultiplierDescriptorPtr)theInterfacePtr)->iClockMultiplier);	// <rdar://6394629>

				clockMultipliers = this->getClockMultipliers();

				if (NULL == clockMultipliers)
				{
					FailIf(NULL == (clockMultipliers = OSArray::withObjects((const OSObject **)&clockMultiplier, 1)), Exit);
					FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kClockMultipliers, clockMultipliers), Exit);
					clockMultipliers = NULL;
				}
				else
				{
					clockMultipliers->setObject(clockMultiplier);
				}

				clockMultiplier->release();
				clockMultiplier = NULL;

				break;
			}
			default:
			{
				DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() - @ default. Nothing to do here.", this);
			}
		}

		theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);
		*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterfaceDescriptor_0200() = %p", this, theInterfacePtr);
	return theInterfacePtr;
}

/****************************************************************************/
/**
 * @brief	Interrupt エンドポイントディスクリプタを解析．
 *
 * @param[in]		theInterfacePtr		エンドポイントディスクリプタ．
 *
 * @return			次のディスクリプタへのポインタ．
 */
// <rdar://problem/6021475> PdjUsbAudio: Status Interrupt Endpoint support
USBInterfaceDescriptorPtr
PdjUsbControlDictionary::parseACInterruptEndpointDescriptor(
	USBInterfaceDescriptorPtr	theInterfacePtr
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterruptEndpointDescriptor(%p)", this, theInterfacePtr);

	PdjUsbEndpointDictionary	*thisEndpoint = NULL;
	OSArray						*endpoints = 0;

	FailIf(NULL == theInterfacePtr, Exit);
	FailIf(0 == theInterfacePtr->bLength, Exit);

	if (ENDPOINT == theInterfacePtr->bDescriptorType)
	{
		if (kInterruptType == (((USBEndpointDescriptorPtr)theInterfacePtr)->bmAttributes & kInterruptType))
		{
			DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterruptEndpointDescriptor() - @ ENDPOINT (4.6.1.1)", this);

			FailIf(NULL == (thisEndpoint = PdjUsbEndpointDictionary::create()), Exit);

			thisEndpoint->setAddress(((USBEndpointDescriptorPtr)theInterfacePtr)->bEndpointAddress);
			thisEndpoint->setAttributes(((USBEndpointDescriptorPtr)theInterfacePtr)->bmAttributes);
			thisEndpoint->setMaxPacketSize(USBToHostWord(((USBEndpointDescriptorPtr)theInterfacePtr)->wMaxPacketSize));
			thisEndpoint->setInterval(((USBEndpointDescriptorPtr)theInterfacePtr)->bInterval);

			endpoints = this->getEndpoints();

			if (NULL == endpoints)
			{
				FailIf(NULL == (endpoints = OSArray::withObjects((const OSObject **)&thisEndpoint, 1)), Exit);
				FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kEndpoints, endpoints), Exit);
				endpoints = NULL;
			}
			else
			{
				endpoints->setObject(thisEndpoint);
			}

			thisEndpoint->release();
			thisEndpoint = NULL;

			theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRCONTROLDICTIONARY "[%p]::parseACInterruptEndpointDescriptor() = %p", this, theInterfacePtr);
	return theInterfacePtr;
}

/****************************************************************************/
/**
 * @brief	チャンネルがミュート設定に対応しているかどうかを調べる．
 *			Feature Unit ディスクリプタの bmaControls フィールドで
 *			Mute Control がサポートされているかどうか判定する．
 *
 * @param[in]		featureUnitID	Feature Unit ID．
 * @param[in]		channelNum		チャンネル番号．
 *
 * @retval			true			ミュート設定に対応している．
 * @retval			false			ミュート設定に対応していない．
 */
// Channel #1 is front left channel, #2 is front right channel
bool
PdjUsbControlDictionary::channelHasMuteControl(
	UInt8			featureUnitID,
	UInt8			channelNum
)
{
	bool							result = false;
	PdjUsbFeatureUnitDictionary		*featureUnitDictionary = NULL;

	FailIf(NULL == (featureUnitDictionary = this->getFeatureUnitDictionary(featureUnitID)), Exit);

	result = featureUnitDictionary->channelHasMuteControl(channelNum);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	チャンネルがボリューム設定に対応しているかどうかを調べる．
 *			Feature Unit ディスクリプタの bmaControls フィールドで
 *			Volume Control がサポートされているかどうか判定する．
 *
 * @param[in]		featureUnitID	Feature Unit ID．
 * @param[in]		channelNum		チャンネル番号．
 *
 * @retval			true			ボリューム設定に対応している．
 * @retval			false			ボリューム設定に対応していない．
 */
bool
PdjUsbControlDictionary::channelHasVolumeControl(
	UInt8			featureUnitID,
	UInt8			channelNum
)
{
	bool							result = false;
	PdjUsbFeatureUnitDictionary		*featureUnitDictionary;

	FailIf(NULL == (featureUnitDictionary = this->getFeatureUnitDictionary(featureUnitID)), Exit);

	result = featureUnitDictionary->channelHasVolumeControl(channelNum);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースが周波数設定に対応しているかどうか調べる．
 *			Clock Source ディスクリプタの bmControls フィールドで
 *			Clock Frequency Control がサポートされているかどうか判定する．
 *
 * @param[in]		clockSourceID	クロックソース ID．
 * @param[in]		isProgrammable	true なら、programmable な Frequency Control のみを探す．
 *									false なら、read-only な Frequency Control も含めて探す．
 *
 * @retval			true			周波数設定に対応している．
 * @retval			false			周波数設定に対応していない．
 */
bool
PdjUsbControlDictionary::clockSourceHasFrequencyControl(
	UInt8			clockSourceID,
	bool			isProgrammable
)
{
	bool							result = false;
	PdjUsbClockSourceDictionary		*clockSourceDictionary;
	UInt8							controls;
	bool							isControlPresent = false;
	bool							isControlProgrammable = false;

	FailIf(NULL == (clockSourceDictionary = this->getClockSourceDictionary(clockSourceID)), Exit);
	FailIf(kIOReturnSuccess != clockSourceDictionary->getControlBitmap(&controls), Exit);

	// D1..0:	Clock Frequency Control
	// D3..2:	Clock Validity Control
	// D7..4:	Reserved. Must be set to 0.

	isControlPresent		= (controls & 0x1) ? true : false;
	isControlProgrammable	= (controls & 0x2) ? true : false;

	if (isProgrammable)
	{
		result = isControlPresent && isControlProgrammable;
	}
	else
	{
		result = isControlPresent;
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースが有効性確認に対応しているかどうか調べる．
 *			Clock Source ディスクリプタの bmControls フィールドで
 *			Clock Validity Control がサポートされているかどうか判定する．
 *
 * @param[in]		clockSourceID	クロックソース ID．
 *
 * @retval			true			有効性確認に対応している．
 * @retval			false			有効性確認に対応していない．
 */
// <rdar://7446555>
bool
PdjUsbControlDictionary::clockSourceHasValidityControl(
	UInt8			clockSourceID
)
{
	bool							result = false;
	PdjUsbClockSourceDictionary		*clockSourceDictionary;
	UInt8							controls;
	bool							isControlPresent = false;

	FailIf(NULL == (clockSourceDictionary = this->getClockSourceDictionary(clockSourceID)), Exit);
	FailIf(kIOReturnSuccess != clockSourceDictionary->getControlBitmap(&controls), Exit);

	// D1..0:	Clock Frequency Control
	// D3..2:	Clock Validity Control
	// D7..4:	Reserved. Must be set to 0.

	isControlPresent = (controls & 0x4) ? true : false;

	result = isControlPresent;

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	入力ターミナルのタイプを取得（ターミナル ID から）．
 *
 * @param[out]		*terminalType	Input Terminal ディスクリプタの wTerminalType フィールド．
 * @param[in]		terminalID		ターミナル ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getInputTerminalType(
	UInt16			*terminalType,
	UInt8			terminalID
)
{
	IOReturn							result = kIOReturnError;
	PdjUsbInputTerminalDictionary		*inputTerminalDictionary;

	*terminalType = USBToHostWord(INPUT_UNDEFINED);

	FailIf(NULL == (inputTerminalDictionary = this->getInputTerminalDictionary(terminalID)), Exit);
	FailIf(kIOReturnSuccess != (result = inputTerminalDictionary->getDictionaryValue(kTerminalType, terminalType)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	出力ターミナルのタイプを取得（ターミナル ID から）．
 *
 * @param[out]		*terminalType	Output Terminal ディスクリプタの wTerminalType フィールド．
 * @param[in]		terminalID		ターミナル ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getOutputTerminalType(
	UInt16			*terminalType,
	UInt8			terminalID
)
{
	IOReturn							result = kIOReturnError;
	PdjUsbOutputTerminalDictionary		*outputTerminalDictionary = NULL;

	*terminalType = USBToHostWord(OUTPUT_UNDEFINED);

	FailIf(NULL == (outputTerminalDictionary = this->getOutputTerminalDictionary(terminalID)), Exit);
	FailIf(kIOReturnSuccess != (result = outputTerminalDictionary->getDictionaryValue(kTerminalType, terminalType)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	入力ターミナルのタイプを取得（OSArray のインデックスから）．
 *
 * @param[out]		*terminalType	Input Terminal ディスクリプタの wTerminalType フィールド．
 * @param[in]		index			OSArray のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getIndexedInputTerminalType(
	UInt16			*terminalType,
	UInt8			index
)
{
	IOReturn							result = kIOReturnError;
	PdjUsbInputTerminalDictionary		*inputTerminalDictionary = NULL;
	OSArray								*inputTerminals = NULL;

	*terminalType = USBToHostWord(INPUT_UNDEFINED);

	if (NULL != (inputTerminals = this->getInputTerminals()))
	{
		FailIf(NULL == (inputTerminalDictionary = OSDynamicCast(PdjUsbInputTerminalDictionary, inputTerminals->getObject(index))), Exit);
		FailIf(kIOReturnSuccess != (result = inputTerminalDictionary->getDictionaryValue(kTerminalType, terminalType)), Exit);
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	入力ターミナルの ID を取得（OSArray のインデックスから）．
 *
 * @param[out]		*inputTerminalID	取得したターミナル ID．
 * @param[in]		index				OSArray のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getIndexedInputTerminalID(
	UInt8			*inputTerminalID,
	UInt8			index
)
{
	IOReturn							result = kIOReturnError;
	PdjUsbInputTerminalDictionary		*inputTerminalDictionary = NULL;
	OSArray								*inputTerminals = NULL;

	*inputTerminalID = 0;

	if (NULL != (inputTerminals = this->getInputTerminals()))
	{
		FailIf(NULL == (inputTerminalDictionary = OSDynamicCast(PdjUsbInputTerminalDictionary, inputTerminals->getObject(index))), Exit);
		FailIf(kIOReturnSuccess != (result = inputTerminalDictionary->getUnitID(inputTerminalID)), Exit);
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	出力ターミナルのタイプを取得（OSArray のインデックスから）．
 *
 * @param[out]		*terminalType	Output Terminal ディスクリプタの wTerminalType フィールド．
 * @param[in]		index			OSArray のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getIndexedOutputTerminalType(
	UInt16			*terminalType,
	UInt8			index
)
{
	IOReturn							result = kIOReturnError;
	PdjUsbOutputTerminalDictionary		*outputTerminalDictionary = NULL;
	OSArray								*outputTerminals = NULL;

	*terminalType = USBToHostWord(OUTPUT_UNDEFINED);

	if (NULL != (outputTerminals = this->getOutputTerminals()))
	{
		FailIf(NULL == (outputTerminalDictionary = OSDynamicCast(PdjUsbOutputTerminalDictionary, outputTerminals->getObject(index))), Exit);
		FailIf(kIOReturnSuccess != (result = outputTerminalDictionary->getDictionaryValue(kTerminalType, terminalType)), Exit);
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	出力ターミナルの ID を取得（OSArray のインデックスから）．
 *
 * @param[out]		*outputTerminalID	取得したターミナル ID．
 * @param[in]		index				OSArray のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getIndexedOutputTerminalID(
	UInt8			*outputTerminalID,
	UInt8			index
)
{
	IOReturn							result = kIOReturnError;
	PdjUsbOutputTerminalDictionary		*outputTerminalDictionary = NULL;
	OSArray								*outputTerminals = NULL;

	*outputTerminalID = 0;

	if (NULL != (outputTerminals = this->getOutputTerminals()))
	{
		FailIf(NULL == (outputTerminalDictionary = OSDynamicCast(PdjUsbOutputTerminalDictionary, outputTerminals->getObject(index))), Exit);
		FailIf(kIOReturnSuccess != (result = outputTerminalDictionary->getUnitID(outputTerminalID)), Exit);
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Feature Unit のコントロール数を取得．
 *
 * @param[out]		*numControls	取得したコントロール数．
 * @param[in]		featureUnitID	Feature Unit ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getNumControls(
	UInt8			*numControls,
	UInt8			featureUnitID
)
{
	IOReturn						result = kIOReturnError;
	PdjUsbFeatureUnitDictionary		*featureUnitDictionary = NULL;

	*numControls = 0;

	FailIf(NULL == (featureUnitDictionary = this->getFeatureUnitDictionary(featureUnitID)), Exit);

	FailIf(kIOReturnSuccess != (result = featureUnitDictionary->getNumControls(numControls)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Input Terminal の数を取得．
 *
 * @param[out]		*numInputTerminals	Input Terminal の数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getNumInputTerminals(
	UInt8			*numInputTerminals
)
{
	IOReturn		result = kIOReturnError;
	OSArray			*inputTerminals = NULL;

	*numInputTerminals = 0;

	FailIf(NULL == (inputTerminals = this->getInputTerminals()), Exit);

	*numInputTerminals = inputTerminals->getCount();

	result = kIOReturnSuccess;

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Output Terminal の数を取得．
 *
 * @param[out]		*numOutputTerminals	Output Terminal の数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getNumOutputTerminals(
	UInt8			*numOutputTerminals
)
{
	IOReturn		result = kIOReturnError;
	OSArray			*outputTerminals = NULL;

	*numOutputTerminals = 0;

	FailIf(NULL == (outputTerminals = this->getOutputTerminals()), Exit);

	*numOutputTerminals = outputTerminals->getCount();

	result = kIOReturnSuccess;

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Selector Unit の数を取得．
 *
 * @param[out]		*numSelectorUnits	Selector Unit の数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getNumSelectorUnits(
	UInt8			*numSelectorUnits
)
{
	OSArray			*selectorUnits = NULL;

	*numSelectorUnits = 0;

	selectorUnits = this->getSelectorUnits();

	if (selectorUnits)
	{
		*numSelectorUnits = selectorUnits->getCount();
	}

	return kIOReturnSuccess;
}

/****************************************************************************/
/**
 * @brief	ユニットの入力ピンの数を取得．
 *
 * @param[out]		*numSources		ユニットの入力ピンの数．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getNumSources(
	UInt8			*numSources,
	UInt8			unitID
)
{
	IOReturn				result = kIOReturnError;
	PdjUsbUnitDictionary	*thisUnit = NULL;

	*numSources = 0;

	FailIf(NULL == (thisUnit = this->getUnitDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = thisUnit->getNumInPins(numSources)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	指定のクロックセレクタに属するクロックソースの OSArray を取得．
 *
 * @param[out]		**clockSelectorSources	クロックセレクタディスクリプタに列挙されている
 *											クロックソースの OSArray．
 * @param[in]		unitID					クロックセレクタのユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getClockSelectorSources(
	OSArray			**clockSelectorSources,
	UInt8			unitID
)
{
	IOReturn							result = kIOReturnError;
	PdjUsbClockSelectorDictionary		*thisClockSelector = NULL;

	*clockSelectorSources = NULL;

	FailIf(NULL == (thisClockSelector = this->getClockSelectorDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = (thisClockSelector->getClockSources(clockSelectorSources))), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースのタイプを取得．
 *
 * @param[out]		*clockType		クロックソースディスクリプタの bmAttributes フィールド
 *									で指定されたクロックタイプ．
 * @param[in]		unitID			クロックソース ID．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbControlDictionary::getClockSourceClockType(
	UInt8			*clockType,
	UInt8			unitID
)
{
	PdjUsbClockSourceDictionary		*thisClockSource = NULL;
	IOReturn						result = kIOReturnError;

	*clockType = 0;

	FailIf(NULL == (thisClockSource = this->getClockSourceDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = (thisClockSource->getClockType(clockType))), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースの関連ターミナルを取得．
 *
 * @param[out]		*assocTerminal	クロックソースディスクリプタの bAssocTerminal フィールド．
 * @param[in]		unitID			クロックソース ID．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbControlDictionary::getClockSourceAssocTerminal(
	UInt8			*assocTerminal,
	UInt8			unitID
)
{
	IOReturn						result = kIOReturnError;
	PdjUsbClockSourceDictionary		*thisClockSource = NULL;

	*assocTerminal = 0;

	FailIf(NULL == (thisClockSource = this->getClockSourceDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = (thisClockSource->getAssocTerminal(assocTerminal))), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットの種類を取得．
 *
 * @param[out]		*subType		ディスクリプタの bDescriptorSubtype フィールド．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getSubType(
	UInt8			*subType,
	UInt8			unitID
)
{
	IOReturn				result = kIOReturnError;
	PdjUsbUnitDictionary	*unitDictionary = NULL;

	*subType = 0;

	FailIf(NULL == (unitDictionary = this->getUnitDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = unitDictionary->getDictionaryValue(kSubType, subType)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットのソース ID を取得．
 *
 * @param[out]		*sourceID		ディスクリプタの bSourceID フィールド．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getSourceID(
	UInt8			*sourceID,
	UInt8			unitID
)
{
	IOReturn				result = kIOReturnError;
	PdjUsbUnitDictionary	*unitDictionary = NULL;

	*sourceID = 0;

	FailIf(NULL == (unitDictionary = this->getUnitDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = unitDictionary->getSourceID(sourceID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットのソース ID の OSArray を取得．
 *
 * @param[out]		**sourceIDs		ディスクリプタの baSourceID フィールドに
 *									列挙されたソース ID の OSArray．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getSourceIDs(
	OSArray			**sourceIDs,
	UInt8			unitID
)
{
	IOReturn				result = kIOReturnError;
	PdjUsbUnitDictionary	*unitDictionary = NULL;
	UInt8					descriptorSubType;
	UInt16					adcVersion = 0;

	*sourceIDs = NULL;

	FailIf(NULL == (unitDictionary = this->getUnitDictionary(unitID)), Exit);

	FailIf(kIOReturnSuccess != (result = this->getADCVersion(&adcVersion)), Exit);
	FailIf(kIOReturnSuccess != (result = unitDictionary->getDescriptorSubType(&descriptorSubType)), Exit);

	if (kPdjUsbUsbSpec1_0 == adcVersion)
	{
		switch (descriptorSubType)
		{
			case MIXER_UNIT:
				FailIf(kIOReturnSuccess != (result = this->getMixerSources(sourceIDs, unitID)), Exit);
				break;

			case SELECTOR_UNIT:
				FailIf(kIOReturnSuccess != (result = this->getSelectorSources(sourceIDs, unitID)), Exit);
				break;

			case PROCESSING_UNIT:
				FailIf(kIOReturnSuccess != (result = this->getProcessingUnitSources(sourceIDs, unitID)), Exit);
				break;

			case EXTENSION_UNIT:
				FailIf(kIOReturnSuccess != (result = this->getExtensionUnitSources(sourceIDs, unitID)), Exit);
				break;

			default:
				result = kIOReturnError;
				break;
		}
	}
	else if (kPdjUsbUsbSpec2_0 == adcVersion)
	{
		switch (descriptorSubType)
		{
			case USBAUDIO_0200::MIXER_UNIT:
				FailIf(kIOReturnSuccess != (result = this->getMixerSources(sourceIDs, unitID)), Exit);
				break;

			case USBAUDIO_0200::SELECTOR_UNIT:
				FailIf(kIOReturnSuccess != (result = this->getSelectorSources(sourceIDs, unitID)), Exit);
				break;

			case USBAUDIO_0200::PROCESSING_UNIT:
				FailIf(kIOReturnSuccess != (result = this->getProcessingUnitSources(sourceIDs, unitID)), Exit);
				break;

			case USBAUDIO_0200::EXTENSION_UNIT:
				FailIf(kIOReturnSuccess != (result = this->getExtensionUnitSources(sourceIDs, unitID)), Exit);
				break;

			case USBAUDIO_0200::CLOCK_SELECTOR:
				FailIf(kIOReturnSuccess != (result = this->getClockSelectorSources(sourceIDs, unitID)), Exit);
				break;

			default:
				result = kIOReturnError;
				break;
		}
	}
	else
	{
		// not supported
		result = kIOReturnUnsupported;
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Input Terminal および Output Terminal に接続されたクロックエンティティを列挙．
 *
 * @return			Input Terminal ディスクリプタおよび Output Terminal ディスクリプタの
 *					bCSourceID フィールドから生成した OSArray．
 */
OSArray *
PdjUsbControlDictionary::getTerminalClockEntities(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRCONTROLDICTIONARY "::getTerminalClockEntities()");

	OSArray								*clockEntities		= NULL;
	OSArray								*inputTerminals		= NULL;
	OSArray								*outputTerminals	= NULL;
	OSNumber							*clockIDNum			= NULL;
	PdjUsbInputTerminalDictionary		*inputTerminal		= NULL;
	PdjUsbOutputTerminalDictionary		*outputTerminal		= NULL;
	UInt8								clockID;
	bool								isDuplicate = false;

	FailIf(NULL == (inputTerminals = this->getInputTerminals()), Exit);
	FailIf(NULL == (outputTerminals = this->getOutputTerminals()), Exit);

	// Comb input terminals for unique clock entities
	for (UInt8 terminalIndex = 0; terminalIndex < inputTerminals->getCount(); terminalIndex++)
	{
		isDuplicate = false;

		FailIf(NULL == (inputTerminal = OSDynamicCast(PdjUsbInputTerminalDictionary, inputTerminals->getObject(terminalIndex))), Exit);
		FailIf(kIOReturnSuccess != inputTerminal->getClockSourceID(&clockID), Exit);

		for (UInt8 entityIndex = 0; (clockEntities) && (entityIndex < clockEntities->getCount()); entityIndex++)
		{
			FailIf(NULL == (clockIDNum = OSDynamicCast(OSNumber, clockEntities->getObject(entityIndex))), Exit);

			if (clockIDNum->unsigned8BitValue() == clockID)
			{
				isDuplicate = true;

				break;
			}
		}

		if (!isDuplicate)
		{
			DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "::getTerminalClockEntities() - Adding clockID %d ...", clockID);

			FailIf(NULL == (clockIDNum = OSNumber::withNumber(clockID, SIZEINBITS(UInt8))), Exit);

			if (NULL == clockEntities)
			{
				FailIf(NULL == (clockEntities = OSArray::withObjects((const OSObject **)&clockIDNum, 1)), Exit);
			}
			else
			{
				FailIf(!clockEntities->setObject(clockIDNum), Exit);
			}

			clockIDNum->release();
			clockIDNum = NULL;
		}
	}

	// Comb output terminals for unique clock entities
	for (UInt8 terminalIndex = 0; terminalIndex < outputTerminals->getCount(); terminalIndex++)
	{
		isDuplicate = false;

		FailIf(NULL == (outputTerminal = OSDynamicCast(PdjUsbOutputTerminalDictionary, outputTerminals->getObject(terminalIndex))), Exit);
		FailIf(kIOReturnSuccess != outputTerminal->getClockSourceID(&clockID), Exit);

		for (UInt8 entityIndex = 0; (clockEntities) && (entityIndex < clockEntities->getCount()); entityIndex++)
		{
			FailIf(NULL == (clockIDNum = OSDynamicCast(OSNumber, clockEntities->getObject(entityIndex))), Exit);

			if (clockIDNum->unsigned8BitValue() == clockID)
			{
				isDuplicate = true;

				break;
			}
		}

		if (!isDuplicate)
		{
			DP(DBG_INFO, "? " LOGSTRCONTROLDICTIONARY "::getTerminalClockEntities() - Adding clockID %d ...", clockID);

			FailIf(NULL == (clockIDNum = OSNumber::withNumber(clockID, SIZEINBITS(UInt8))), Exit);

			if (NULL == clockEntities)
			{
				FailIf(NULL == (clockEntities = OSArray::withObjects((const OSObject **)&clockIDNum, 1)), Exit);
			}
			else
			{
#ifdef BUG_FIX
				FailIf(!clockEntities->setObject(clockIDNum), Exit);
#else
				FailIf(kIOReturnSuccess != clockEntities->setObject(clockIDNum), Exit);
#endif
			}

			clockIDNum->release();
			clockIDNum = NULL;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRCONTROLDICTIONARY "::getTerminalClockEntities() = %p", clockEntities);
	return clockEntities;
}

/****************************************************************************/
/**
 * @brief	ユニットのクロックソース ID を取得．
 *
 * @param[out]		*clockSourceID	ディスクリプタの bCSourceID フィールド．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getClockSourceID(
	UInt8			*clockSourceID,
	UInt8			unitID
)
{
	IOReturn				result = kIOReturnError;
	PdjUsbUnitDictionary	*unitDictionary = NULL;

	*clockSourceID = 0;

	FailIf(NULL == (unitDictionary = this->getUnitDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = unitDictionary->getClockSourceID(clockSourceID)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットのストリングディスクリプタのインデックスを取得．
 *
 * @param[out]		*stringIndex	ストリングディスクリプタのインデックス．
 * @param[in]		unitID			ユニット ID．
 *
 * @return			IOReturn コード．
 */
// <rdar://6394629>
IOReturn
PdjUsbControlDictionary::getStringIndex(
	UInt8			*stringIndex,
	UInt8			unitID
)
{
	IOReturn				result = kIOReturnError;
	PdjUsbUnitDictionary	*unitDictionary = NULL;

	*stringIndex = 0;

	FailIf(NULL == (unitDictionary = this->getUnitDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = unitDictionary->getStringIndex(stringIndex)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニットのディスクリプタから Audio Channel Cluster Descriptor を作成．
 *
 * @param[out]		*clusterDescriptor	ユニットのディスクリプタから作成した
 *										Audio Channel Cluster Descriptor．
 *										メモリは呼び出し側で確保しておくこと．
 * @param[in]		unitID				ユニット ID．
 *
 * @return			IOReturn コード．
 */
// <rdar://6430836>
IOReturn
PdjUsbControlDictionary::getAudioClusterDescriptor(
	AudioClusterDescriptor	*clusterDescriptor,
	UInt8					unitID
)
{
	IOReturn				result = kIOReturnError;
	PdjUsbUnitDictionary	*unitDictionary = NULL;

	clusterDescriptor->bNrChannels		= 0;
	clusterDescriptor->bmChannelConfig	= 0;
	clusterDescriptor->iChannelNames	= 0;

	FailIf(NULL == (unitDictionary = this->getUnitDictionary(unitID)), Exit);

	FailIf(kIOReturnSuccess != (result = unitDictionary->getNumChannels(&clusterDescriptor->bNrChannels)), Exit);
	FailIf(kIOReturnSuccess != (result = unitDictionary->getChannelConfig(&clusterDescriptor->bmChannelConfig)), Exit);
	FailIf(kIOReturnSuccess != (result = unitDictionary->getChannelNames(&clusterDescriptor->iChannelNames)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Interrupt エンドポイントを検索．
 *
 * @param[out]		*address		見つかった Interrupt エンドポイントのアドレス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getInterruptEndpointAddress(
	UInt8			*address
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbEndpointDictionary	*thisEndpoint	= NULL;
	OSArray						*endpoints		= NULL;
	UInt8						endpointIndex = 0;
	UInt8						attributes;

	FailIf(NULL == address, Exit);

	*address = 0;

	FailIf(NULL == (endpoints = this->getEndpoints()), Exit);

	while ((0 == *address) && (endpointIndex < endpoints->getCount()))
	{
		FailIf(NULL == (thisEndpoint = this->getIndexedEndpointDictionary(endpointIndex)), Exit);
		FailIf(kIOReturnSuccess != (result = thisEndpoint->getAttributes(&attributes)), Exit);

		if (kInterruptType == (attributes & kInterruptType))
		{
			FailIf(kIOReturnSuccess != (result = thisEndpoint->getAddress(address)), Exit);
		}

		endpointIndex++;
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Interrupt エンドポイントの有無を調べる．
 *
 * @retval			true			Interrupt エンドポイントがある．
 * @retval			false			Interrupt エンドポイントはない．
 */
bool
PdjUsbControlDictionary::hasInterruptEndpoint(
	void
)
{
	bool						result			= false;
	PdjUsbEndpointDictionary	*thisEndpoint	= NULL;
	OSArray						*endpoints		= NULL;
	UInt8						endpointIndex	= 0;
	UInt8						attributes;

	endpoints = this->getEndpoints();

	if (NULL != endpoints)
	{
		while (endpointIndex < endpoints->getCount())
		{
			FailIf(NULL == (thisEndpoint = this->getIndexedEndpointDictionary(endpointIndex)), Exit);
			FailIf(kIOReturnSuccess != (result = thisEndpoint->getAttributes(&attributes)), Exit);

			if (kInterruptType == (attributes & kInterruptType))
			{
				result = true;

				break;
			}

			endpointIndex++;
		}
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ユニット dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			ユニット dictionary．
 */
PdjUsbUnitDictionary *
PdjUsbControlDictionary::getUnitDictionary(
	UInt8			unitID
)
{
	PdjUsbUnitDictionary	*unitDictionary = NULL;

	if (!unitDictionary)
	{
		unitDictionary = this->getInputTerminalDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getOutputTerminalDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getMixerUnitDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getSelectorUnitDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getFeatureUnitDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getEffectUnitDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getProcessingUnitDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getExtensionUnitDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getClockSourceDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getClockSelectorDictionary(unitID);
	}

	if (!unitDictionary)
	{
		unitDictionary = this->getClockMultiplierDictionary(unitID);
	}

	return unitDictionary;
}

/****************************************************************************/
/**
 * @brief	Input Terminal dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Input Terminal dictionary．
 */
PdjUsbInputTerminalDictionary *
PdjUsbControlDictionary::getInputTerminalDictionary(
	UInt8			unitID
)
{
	PdjUsbInputTerminalDictionary		*inputTerminalDictionary = NULL;
	OSArray								*inputTerminals = NULL;
	UInt8								inputTerminalIndex = 0;
	UInt8								thisUnitID;
	bool								found = false;

	if (NULL != (inputTerminals = this->getInputTerminals()))
	{
		while ((!found) && (inputTerminalIndex < inputTerminals->getCount()))
		{
			FailIf(
						NULL == (
									inputTerminalDictionary = OSDynamicCast(
																	PdjUsbInputTerminalDictionary,
																	inputTerminals->getObject(inputTerminalIndex)
															  )
								),
						Exit
				  );

			FailIf(kIOReturnSuccess != inputTerminalDictionary->getUnitID(&thisUnitID), Exit);

			if (unitID == thisUnitID)
			{
				found = true;
			}

			inputTerminalIndex++;
		}
	}

Exit:
	if (!found)
	{
		inputTerminalDictionary = NULL;
	}

	return inputTerminalDictionary;
}

/****************************************************************************/
/**
 * @brief	Output Terminal dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Output Terminal dictionary．
 */
PdjUsbOutputTerminalDictionary *
PdjUsbControlDictionary::getOutputTerminalDictionary(
	UInt8			unitID
)
{
	PdjUsbOutputTerminalDictionary		*outputTerminalDictionary = NULL;
	OSArray								*outputTerminals = NULL;
	UInt8								outputTerminalIndex = 0;
	UInt8								thisUnitID;
	bool								found = false;

	if (NULL != (outputTerminals = this->getOutputTerminals()))
	{
		while ((!found) && (outputTerminalIndex < outputTerminals->getCount()))
		{
			FailIf(
						NULL == (
									outputTerminalDictionary = OSDynamicCast(
																	PdjUsbOutputTerminalDictionary,
																	outputTerminals->getObject(outputTerminalIndex)
															   )
								),
						Exit
				  );

			FailIf(kIOReturnSuccess != outputTerminalDictionary->getUnitID(&thisUnitID), Exit);

			if (unitID == thisUnitID)
			{
				found = true;
			}

			outputTerminalIndex++;
		}
	}

Exit:
	if (!found)
	{
		outputTerminalDictionary = NULL;
	}

	return outputTerminalDictionary;
}

/****************************************************************************/
/**
 * @brief	Mixer Unit dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Mixer Unit dictionary．
 */
PdjUsbMixerUnitDictionary *
PdjUsbControlDictionary::getMixerUnitDictionary(
	UInt8			unitID
)
{
	PdjUsbMixerUnitDictionary		*thisMixerUnit = NULL;
	OSArray							*mixerUnits;
	UInt8							mixerUnitIndex;
	UInt8							mixerUnitID;
	bool							found = false;

	mixerUnitIndex = 0;

	mixerUnits = this->getMixerUnits();

	while ((mixerUnits) && (!found) && (mixerUnitIndex < mixerUnits->getCount()))
	{
		thisMixerUnit = OSDynamicCast(PdjUsbMixerUnitDictionary, mixerUnits->getObject(mixerUnitIndex));

		FailIf(kIOReturnSuccess != thisMixerUnit->getUnitID(&mixerUnitID), Exit);

		if (unitID == mixerUnitID)
		{
			found = true;
		}

		mixerUnitIndex++;
	}

Exit:
	if (!found)
	{
		thisMixerUnit = NULL;
	}

	return thisMixerUnit;
}

/****************************************************************************/
/**
 * @brief	Selector Unit dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Selector Unit dictionary．
 */
PdjUsbSelectorUnitDictionary *
PdjUsbControlDictionary::getSelectorUnitDictionary(
	UInt8			unitID
)
{
	PdjUsbSelectorUnitDictionary	*thisSelectorUnit = NULL;
	OSArray							*selectorUnits = NULL;
	UInt8							selectorUnitIndex;
	UInt8							selectorUnitID;
	bool							found = false;

	selectorUnitIndex = 0;

	selectorUnits = this->getSelectorUnits();

	while ((selectorUnits) && (!found) && (selectorUnitIndex < selectorUnits->getCount()))
	{
		FailIf(
					NULL == (
								thisSelectorUnit = OSDynamicCast(
														PdjUsbSelectorUnitDictionary,
														selectorUnits->getObject(selectorUnitIndex)
												   )
							),
					Exit
			  );

		FailIf(kIOReturnSuccess != thisSelectorUnit->getUnitID(&selectorUnitID), Exit);

		if (unitID == selectorUnitID)
		{
			found = true;
		}

		selectorUnitIndex++;
	}

Exit:
	if (!found)
	{
		thisSelectorUnit = NULL;
	}

	return thisSelectorUnit;
}

/****************************************************************************/
/**
 * @brief	Feature Unit dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Feature Unit dictionary．
 */
PdjUsbFeatureUnitDictionary *
PdjUsbControlDictionary::getFeatureUnitDictionary(
	UInt8			unitID
)
{
	PdjUsbFeatureUnitDictionary		*featureUnitDictionary = NULL;
	OSArray							*featureUnits = NULL;
	UInt8							featureUnitIndex = 0;
	UInt8							thisUnitID;
	bool							found = false;

	if (NULL != (featureUnits = this->getFeatureUnits()))
	{
		while ((!found) && (featureUnitIndex < featureUnits->getCount()))
		{
			featureUnitDictionary = OSDynamicCast(PdjUsbFeatureUnitDictionary, featureUnits->getObject(featureUnitIndex));

			FailIf(kIOReturnSuccess != featureUnitDictionary->getUnitID(&thisUnitID), Exit);

			if ((featureUnitDictionary) && (unitID == thisUnitID))
			{
				found = true;
			}

			featureUnitIndex++;
		}
	}

Exit:
	if (!found)
	{
		featureUnitDictionary = NULL;
	}

	return featureUnitDictionary;
}

/****************************************************************************/
/**
 * @brief	Effect Unit dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Effect Unit dictionary．
 */
PdjUsbEffectUnitDictionary *
PdjUsbControlDictionary::getEffectUnitDictionary(
	UInt8			unitID
)
{
	PdjUsbEffectUnitDictionary		*thisEffectUnit = NULL;
	OSArray							*effectUnits;
	UInt8							effectUnitIndex;
	UInt8							effectUnitID;
	bool							found = false;

	effectUnitIndex = 0;
	effectUnits = this->getEffectUnits();

	while ((effectUnits) && (!found) && (effectUnitIndex < effectUnits->getCount()))
	{
		FailIf(NULL == (thisEffectUnit = OSDynamicCast(PdjUsbEffectUnitDictionary, effectUnits->getObject(effectUnitIndex))), Exit);
		FailIf(kIOReturnSuccess != thisEffectUnit->getUnitID(&effectUnitID), Exit);

		if (unitID == effectUnitID)
		{
			found = true;
		}

		effectUnitIndex++;
	}

Exit:
	if (found != true)
	{
		thisEffectUnit = NULL;
	}

	return thisEffectUnit;
}

/****************************************************************************/
/**
 * @brief	Processing Unit dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Processing Unit dictionary．
 */
PdjUsbProcessingUnitDictionary *
PdjUsbControlDictionary::getProcessingUnitDictionary(
	UInt8			unitID
)
{
	PdjUsbProcessingUnitDictionary		*thisProcessingUnit = NULL;
	OSArray								*processingUnits;
	UInt8								processingUnitIndex;
	UInt8								processingUnitID;
	bool								found = false;

	processingUnitIndex = 0;

	processingUnits = this->getProcessingUnits();

	while ((processingUnits) && (!found) && (processingUnitIndex < processingUnits->getCount()))
	{
		FailIf(
					NULL == (
								thisProcessingUnit = OSDynamicCast(
															PdjUsbProcessingUnitDictionary,
															processingUnits->getObject(processingUnitIndex)
													 )
							),
					Exit
			  );

		FailIf(kIOReturnSuccess != thisProcessingUnit->getUnitID(&processingUnitID), Exit);

		if (unitID == processingUnitID)
		{
			found = true;
		}

		processingUnitIndex++;
	}

Exit:
	if (!found)
	{
		thisProcessingUnit = NULL;
	}

	return thisProcessingUnit;
}

/****************************************************************************/
/**
 * @brief	Extension Unit dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Extension Unit dictionary．
 */
PdjUsbExtensionUnitDictionary *
PdjUsbControlDictionary::getExtensionUnitDictionary(
	UInt8			unitID
)
{
	PdjUsbExtensionUnitDictionary		*thisExtensionUnit = NULL;
	OSArray								*extensionUnits;
	UInt8								extensionUnitIndex;
	UInt8								extensionUnitID;
	bool								found = false;

	extensionUnitIndex = 0;

	extensionUnits = this->getExtensionUnits();

	while ((extensionUnits) && (!found) && (extensionUnitIndex < extensionUnits->getCount()))
	{
		thisExtensionUnit = OSDynamicCast(PdjUsbExtensionUnitDictionary, extensionUnits->getObject(extensionUnitIndex));

		FailIf(kIOReturnSuccess != thisExtensionUnit->getUnitID(&extensionUnitID), Exit);

		if (unitID == extensionUnitID)
		{
			found = true;
		}

		extensionUnitIndex++;
	}

Exit:
	if (!found)
	{
		thisExtensionUnit = NULL;
	}

	return thisExtensionUnit;
}

/****************************************************************************/
/**
 * @brief	Clock Source dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Clock Source dictionary．
 */
PdjUsbClockSourceDictionary *
PdjUsbControlDictionary::getClockSourceDictionary(
	UInt8			unitID
)
{
	PdjUsbClockSourceDictionary		*thisClockSource = NULL;
	OSArray							*clockSources;
	UInt8							clockSourceIndex;
	UInt8							clockSourceID;
	bool							found = false;

	clockSourceIndex = 0;

	clockSources = this->getClockSources();

	while ((clockSources) && (!found) && (clockSourceIndex < clockSources->getCount()))
	{
		FailIf(
					NULL == (
								thisClockSource = OSDynamicCast(
														PdjUsbClockSourceDictionary,
														clockSources->getObject(clockSourceIndex)
												  )
							),
					Exit
			  );

		FailIf(kIOReturnSuccess != thisClockSource->getUnitID(&clockSourceID), Exit);

		if (unitID == clockSourceID)
		{
			found = true;
		}

		clockSourceIndex++;
	}

Exit:
	if (!found)
	{
		thisClockSource = NULL;
	}

	return thisClockSource;
}

/****************************************************************************/
/**
 * @brief	Clock Selector dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Clock Selector dictionary．
 */
PdjUsbClockSelectorDictionary *
PdjUsbControlDictionary::getClockSelectorDictionary(
	UInt8			unitID
)
{
	PdjUsbClockSelectorDictionary		*thisClockSelector = NULL;
	OSArray								*clockSelectors;
	UInt8								clockSelectorIndex;
	UInt8								clockSelectorID;
	bool								found = false;

	clockSelectorIndex = 0;

	clockSelectors = this->getClockSelectors();

	while ((clockSelectors) && (!found) && (clockSelectorIndex < clockSelectors->getCount()))
	{
		FailIf(
					NULL == (
								thisClockSelector = OSDynamicCast(
														PdjUsbClockSelectorDictionary,
														clockSelectors->getObject(clockSelectorIndex)
													)
							),
					Exit
			  );

		FailIf(kIOReturnSuccess != thisClockSelector->getUnitID(&clockSelectorID), Exit);

		if (unitID == clockSelectorID)
		{
			found = true;
		}

		clockSelectorIndex++;
	}

Exit:
	if (found != true)
	{
		thisClockSelector = NULL;
	}

	return thisClockSelector;
}

/****************************************************************************/
/**
 * @brief	Clock Multiplier dictionary を取得．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			Clock Multiplier dictionary．
 */
PdjUsbClockMultiplierDictionary *
PdjUsbControlDictionary::getClockMultiplierDictionary(
	UInt8			unitID
)
{
	PdjUsbClockMultiplierDictionary		*thisClockMultiplier = NULL;
	OSArray								*clockMultipliers;
	UInt8								clockMultiplierIndex;
	UInt8								clockMultiplierID;
	bool								found = false;

	clockMultiplierIndex = 0;

	clockMultipliers = this->getClockMultipliers();

	while ((clockMultipliers) && (!found) && (clockMultiplierIndex < clockMultipliers->getCount()))
	{
		FailIf(
					NULL == (
								thisClockMultiplier = OSDynamicCast(
															PdjUsbClockMultiplierDictionary,
															clockMultipliers->getObject(clockMultiplierIndex)
													  )
							),
					Exit
			  );

		FailIf(kIOReturnSuccess != thisClockMultiplier->getUnitID(&clockMultiplierID), Exit);

		if (unitID == clockMultiplierID)
		{
			found = true;
		}

		clockMultiplierIndex++;
	}

Exit:
	if (!found)
	{
		thisClockMultiplier = NULL;
	}

	return thisClockMultiplier;
}

/****************************************************************************/
/**
 * @brief	エンドポイント dictionary を取得．
 *
 * @param[in]		index			OSArray のインデックス．
 *
 * @return			エンドポイント dictionary．
 */
PdjUsbEndpointDictionary *
PdjUsbControlDictionary::getIndexedEndpointDictionary(
	UInt8			index
)
{
	PdjUsbEndpointDictionary	*thisEndpoint	= NULL;
	OSArray						*endpoints		= NULL;

	FailIf(NULL == (endpoints = this->getEndpoints()), Exit);
	FailIf(NULL == (thisEndpoint = OSDynamicCast(PdjUsbEndpointDictionary, endpoints->getObject(index))), Exit);

Exit:
	return thisEndpoint;
}

/****************************************************************************/
/**
 * @brief	エンドポイント dictionary の OSArray を取得．
 *
 * @return			エンドポイント dictionary の OSArray．
 */
OSArray *
PdjUsbControlDictionary::getEndpoints(
	void
)
{
	OSArray			*endpoints		 = NULL;
	OSObject		*dictionaryValue = NULL;

	dictionaryValue = OSDictionary::getObject(kEndpoints);

	if (dictionaryValue)
	{
		endpoints = OSDynamicCast(OSArray, dictionaryValue);
	}

	return endpoints;
}

/****************************************************************************/
/**
 * @brief	セレクタユニットのソース ID の OSArray を取得．
 *
 * @param[out]		**selectorSources	セレクタユニットディスクリプタの baSourceID フィールド
 *										に列挙されたソース ID の OSArray．
 * @param[in]		unitID				ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getSelectorSources(
	OSArray			**selectorSources,
	UInt8			unitID
)
{
	IOReturn						result = kIOReturnError;
	PdjUsbSelectorUnitDictionary	*thisSelectorUnit = NULL;

	*selectorSources = NULL;

	FailIf(NULL == (thisSelectorUnit = this->getSelectorUnitDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = (thisSelectorUnit->getSources(selectorSources))), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ミキサーユニットのソース ID の OSArray を取得．
 *
 * @param[out]		**mixerSources		ミキサーユニットディスクリプタの baSourceID フィールド
 *										に列挙されたソース ID の OSArray．
 * @param[in]		unitID				ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getMixerSources(
	OSArray			**mixerSources,
	UInt8			unitID
)
{
	PdjUsbMixerUnitDictionary		*thisMixerUnit = NULL;
	IOReturn						result = kIOReturnError;

	*mixerSources = NULL;

	FailIf(NULL == (thisMixerUnit = this->getMixerUnitDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = (thisMixerUnit->getSources(mixerSources))), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	プロセッシングユニットのソース ID の OSArray を取得．
 *
 * @param[out]		**processingUnitSources		プロセッシングユニットディスクリプタの baSourceID フィールド
 *												に列挙されたソース ID の OSArray．
 * @param[in]		unitID						ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getProcessingUnitSources(
	OSArray			**processingUnitSources,
	UInt8			unitID
)
{
	IOReturn							result = kIOReturnError;
	PdjUsbProcessingUnitDictionary		*thisProcessingUnit = NULL;

	*processingUnitSources = NULL;

	FailIf(NULL == (thisProcessingUnit = this->getProcessingUnitDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = (thisProcessingUnit->getSources(processingUnitSources))), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	エクステンションユニットのソース ID の OSArray を取得．
 *
 * @param[out]		**extensionUnitSources		エクステンションユニットディスクリプタの baSourceID フィールド
 *												に列挙されたソース ID の OSArray．
 * @param[in]		unitID						ユニット ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbControlDictionary::getExtensionUnitSources(
	OSArray			**extensionUnitSources,
	UInt8			unitID
)
{
	IOReturn							result = kIOReturnError;
	PdjUsbExtensionUnitDictionary		*thisExtensionUnit = NULL;

	*extensionUnitSources = NULL;

	FailIf(NULL == (thisExtensionUnit = this->getExtensionUnitDictionary(unitID)), Exit);
	FailIf(kIOReturnSuccess != (result = (thisExtensionUnit->getSources(extensionUnitSources))), Exit);

Exit:
	return result;
}

/*------------------------------------------*/
/*		PdjUsbStreamDictionary クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbStreamDictionary -

OSDefineMetaClassAndStructors(PdjUsbStreamDictionary, super)

/****************************************************************************/
/**
 * @brief	ストリーム dictionary を生成．
 *			ストリーム dictionary は AudioStreaming Interface に対応する．
 *
 * @return			生成されたストリーミング dictionary．
 */
PdjUsbStreamDictionary *
PdjUsbStreamDictionary::create(
	void
)
{
	PdjUsbStreamDictionary			*streamDictionary = NULL;

	FailIf(NULL == (streamDictionary = new PdjUsbStreamDictionary), Exit);
	RELEASE_IF_FALSE(streamDictionary, streamDictionary->initDictionaryForUse());

Exit:
	return streamDictionary;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface ディスクリプタを解析（UAC 1.0 用）．
 *
 * @param[in]		theInterfacePtr		AudioStreaming Interface ディスクリプタ．
 * @param[in]		currentInterface	interface 番号．
 * @param[in, out]	*parsedLength		コンフィグレーションディスクリプタのうち、解析済みのバイト数．
 * @param[in]		totalLength			コンフィグレーションディスクリプタの総バイト数．
 *
 * @return			次のディスクリプタへのポインタ．
 */
USBInterfaceDescriptorPtr
PdjUsbStreamDictionary::parseASInterfaceDescriptor(
	USBInterfaceDescriptorPtr	theInterfacePtr,
	UInt8 const					currentInterface,
	UInt16						*parsedLength,
	UInt16						totalLength
)
{
	DP(DBG_TRACE, "+ " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor(%p, %d)",
		this, theInterfacePtr, currentInterface);

	PdjUsbEndpointDictionary		*thisEndpoint	= NULL;
	PdjUsbASEndpointDictionary		*asIsocEndpoint	= NULL;
	UInt32							sampleRate;
	UInt16							formatTag;
	UInt8							numSampleFreqs;
	UInt8							initialBLength;
	bool							done = false;

	FailIf(NULL == theInterfacePtr, Exit);
	FailIf(0 == theInterfacePtr->bLength, Exit);

	initialBLength = theInterfacePtr->bLength;

	while (
				(theInterfacePtr)
		  &&
				(theInterfacePtr->bLength > 0)
		  &&
				(!done)
		  &&
				(*parsedLength <= totalLength)
		  )
	{
		if (theInterfacePtr->bLength < initialBLength)
		{
			this->logDescriptor((UInt8 *)theInterfacePtr, theInterfacePtr->bLength);
		}

		if (CS_INTERFACE == theInterfacePtr->bDescriptorType)
		{
			switch (theInterfacePtr->bDescriptorSubtype)
			{
				case AS_GENERAL:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ AS_GENERAL (4.5.2)", this);

					FailIf(
								kIOReturnSuccess != super::setDictionaryValue(
																kTerminalLink,
																((ASInterfaceDescriptorPtr)theInterfacePtr)->bTerminalLink
														   ),
								Exit
						  );

					FailIf(
								kIOReturnSuccess != super::setDictionaryValue(
																kDelay,
																((ASInterfaceDescriptorPtr)theInterfacePtr)->bDelay
														   ),
								Exit
						  );

					// formatTag = USBToHostWord(
					//					(((ASInterfaceDescriptorPtr)theInterfacePtr)->wFormatTag[1] << 8)
					//			   |
					//					(((ASInterfaceDescriptorPtr)theInterfacePtr)->wFormatTag[0] << 0)
					//			   );
					formatTag = (
									(((ASInterfaceDescriptorPtr)theInterfacePtr)->wFormatTag[1] << 8)
								|
									(((ASInterfaceDescriptorPtr)theInterfacePtr)->wFormatTag[0] << 0)
								);
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - formatTag = 0x%x", this, formatTag);

					FailIf(kIOReturnSuccess != super::setDictionaryValue(kFormatTag, formatTag), Exit);

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				case FORMAT_TYPE:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ FORMAT_TYPE", this);

					switch (((ASFormatTypeIDescriptorPtr)theInterfacePtr)->bFormatType)
					{
						case FORMAT_TYPE_I:
						case FORMAT_TYPE_III:
						{
							DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ FORMAT_TYPE_I/FORMAT_TYPE_III (Format 2.2.5/2.4.1)", this);

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kNumChannels,
																		((ASFormatTypeIDescriptorPtr)theInterfacePtr)->bNrChannels
																   ),
										Exit
								  );

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kSubframeSize,
																		((ASFormatTypeIDescriptorPtr)theInterfacePtr)->bSubframeSize
																   ),
										Exit
								  );

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kBitResolution,
																		((ASFormatTypeIDescriptorPtr)theInterfacePtr)->bBitResolution
																   ),
										Exit
								  );

							numSampleFreqs = ((ASFormatTypeIDescriptorPtr)theInterfacePtr)->bSamFreqType;

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kNumSampleRates,
																		numSampleFreqs
																   ),
										Exit
								  );

							if (0 != numSampleFreqs)
							{
								DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - Interface has a discrete number (%d) of sample rates ",
									this, numSampleFreqs);

								for (UInt8 sampleRateIndex = 0; sampleRateIndex < numSampleFreqs; sampleRateIndex++)
								{
									sampleRate = convertSampleFreq(
														&((ASFormatTypeIDescriptorPtr)theInterfacePtr)->sampleFreq[
																											sampleRateIndex * kBytesPerSampleFrequency
																										]
												 );

									FailIf(kIOReturnSuccess != this->addSampleRate(sampleRate), Exit);
								}
							}
							else
							{
								DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - Device has a variable number of sample rates", this);

								for (UInt8 sampleRateIndex = 0; sampleRateIndex < 2; sampleRateIndex++)
								{
									sampleRate = convertSampleFreq(
														&((ASFormatTypeIDescriptorPtr)theInterfacePtr)->sampleFreq[sampleRateIndex * 3]
												 );

									FailIf(kIOReturnSuccess != this->addSampleRate(sampleRate), Exit);
								}
							}

							break;
						}
						case FORMAT_TYPE_II:
						{
							DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ FORMAT_TYPE_II (Format 2.3.6)", this);

							UInt16		maxBitRate;
							UInt16		samplesPerFrame;

							maxBitRate = USBToHostWord(((ASFormatTypeIIDescriptorPtr)theInterfacePtr)->wMaxBitRate);
							FailIf(kIOReturnSuccess != super::setDictionaryValue(kMaxBitRate, maxBitRate), Exit);

							samplesPerFrame = USBToHostWord(((ASFormatTypeIIDescriptorPtr)theInterfacePtr)->wSamplesPerFrame);
							FailIf(kIOReturnSuccess != super::setDictionaryValue(kSamplesPerFrame, samplesPerFrame), Exit);

							numSampleFreqs = ((ASFormatTypeIIDescriptorPtr)theInterfacePtr)->bSamFreqType;
							FailIf(kIOReturnSuccess != super::setDictionaryValue(kNumSampleRates, numSampleFreqs), Exit);

							if (0 != numSampleFreqs)
							{
								DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - Interface has a discrete number (%d) of sample rates",
									this, numSampleFreqs);

								for (UInt8 sampleRateIndex = 0; sampleRateIndex < numSampleFreqs; sampleRateIndex++)
								{
									sampleRate = convertSampleFreq(
														&((ASFormatTypeIIDescriptorPtr)theInterfacePtr)->sampleFreq[
																												sampleRateIndex * kBytesPerSampleFrequency
																										 ]
												 );

									FailIf(kIOReturnSuccess != this->addSampleRate(sampleRate), Exit);
								}
							}
							else
							{
								DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - Device has a variable number of sample rates", this);

								for (UInt8 sampleRateIndex = 0; sampleRateIndex < 2; sampleRateIndex++)
								{
									sampleRate = convertSampleFreq(
														&((ASFormatTypeIIDescriptorPtr)theInterfacePtr)->sampleFreq[sampleRateIndex * 3]
												 );

									FailIf(kIOReturnSuccess != this->addSampleRate(sampleRate), Exit);
								}
							}

							break;
						}
						default:
						{
							DP(DBG_ERROR, "! " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ Unknown Format Type!", this);
						}
					}

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				case FORMAT_SPECIFIC:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ FORMAT_SPECIFIC", this);

					UInt32		bmAC3BSID;
					UInt16		bmMPEGCapabilities;

					formatTag = USBToHostWord(
									(((ASFormatSpecificDescriptorHeaderPtr)theInterfacePtr)->wFormatTag[1] << 8)
								|
									(((ASFormatSpecificDescriptorHeaderPtr)theInterfacePtr)->wFormatTag[0] << 0)
								);

					switch (formatTag)
					{
						case MPEG:
						{
							DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ MPEG (2.3.8.1.1)", this);

							bmMPEGCapabilities = USBToHostWord(
														(((ASMPEGFormatSpecificDescriptorPtr)theInterfacePtr)->bmMPEGCapabilities[1] << 8)
												 |
														(((ASMPEGFormatSpecificDescriptorPtr)theInterfacePtr)->bmMPEGCapabilities[0] << 0)
												 );

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kMPEGCapabilities,
																		bmMPEGCapabilities
																   ),
										Exit
								  );

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kMPEGFeatures,
																		((ASMPEGFormatSpecificDescriptorPtr)theInterfacePtr)->bmMPEGFeatures
																   ),
										Exit
								  );

							break;
						}
						case AC3:
						{
							DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ AC3 (Format 2.3.8.2.1)", this);

							bmAC3BSID = USBToHostLong(
											(((ASAC3FormatSpecificDescriptorPtr)theInterfacePtr)->bmBSID[3] << 24)
										|
											(((ASAC3FormatSpecificDescriptorPtr)theInterfacePtr)->bmBSID[2] << 16)
										|
											(((ASAC3FormatSpecificDescriptorPtr)theInterfacePtr)->bmBSID[1] <<  8)
										|
											(((ASAC3FormatSpecificDescriptorPtr)theInterfacePtr)->bmBSID[0] <<  0)
										);

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kAC3BSID,
																		bmAC3BSID
																   ),
										Exit
								  );

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kAC3Features,
																		((ASAC3FormatSpecificDescriptorPtr)theInterfacePtr)->bmAC3Features
																   ),
										Exit
								  );

							break;
						}
						default:
						{
							DP(DBG_ERROR, "! " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ Unknown format type 0x%x",
								this, formatTag);
							break;
						}
					}

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				default:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ Default", this);

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);
				}
			}
		}
		else // if (CS_INTERFACE == theInterfacePtr->bDescriptorType)
		{
			OSArray		*endpoints = 0;

			switch (theInterfacePtr->bDescriptorType)
			{
				case INTERFACE:
				{
					// Need to make a new interface dictionary for this new interface or new alternate setting
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ INTERFACE", this);

					done = true;

					break;
				}
				case ENDPOINT:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ ENDPOINT (4.6.1.1)", this);

					FailIf(NULL == (thisEndpoint = PdjUsbEndpointDictionary::create()), Exit);

					thisEndpoint->setAddress(((USBEndpointDescriptorPtr)theInterfacePtr)->bEndpointAddress);
					thisEndpoint->setAttributes(((USBEndpointDescriptorPtr)theInterfacePtr)->bmAttributes);
					thisEndpoint->setMaxPacketSize(USBToHostWord(((USBEndpointDescriptorPtr)theInterfacePtr)->wMaxPacketSize));
					thisEndpoint->setRefreshInt(((USBEndpointDescriptorPtr)theInterfacePtr)->bRefresh);
					thisEndpoint->setSynchAddress(((USBEndpointDescriptorPtr)theInterfacePtr)->bSynchAddress);

					endpoints = this->getEndpoints();

					if (NULL == endpoints)
					{
						FailIf(NULL == (endpoints = OSArray::withObjects((const OSObject **)&thisEndpoint, 1)), Exit);
						FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kEndpoints, endpoints), Exit);
						endpoints = NULL;
					}
					else
					{
						endpoints->setObject(thisEndpoint);
					}

					thisEndpoint->release();
					thisEndpoint = NULL;

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				case CS_ENDPOINT:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ CS_ENDPOINT (4.6.1.2)", this);

					if (EP_GENERAL == ((ASEndpointDescriptorPtr)theInterfacePtr)->bDescriptorSubtype)
					{
						DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ EP_GENERAL", this);

						asIsocEndpoint = new PdjUsbASEndpointDictionary(
													((ASEndpointDescriptorPtr)theInterfacePtr)->bmAttributes & (1 << sampleFreqControlBit	),
													((ASEndpointDescriptorPtr)theInterfacePtr)->bmAttributes & (1 << pitchControlBit		),
													((ASEndpointDescriptorPtr)theInterfacePtr)->bmAttributes & (1 << maxPacketsOnlyBit		),
													((ASEndpointDescriptorPtr)theInterfacePtr)->bLockDelayUnits,
													USBToHostWord(((UInt16 *)((ASEndpointDescriptorPtr)theInterfacePtr)->wLockDelay)[0])
											 );
						FailIf(NULL == asIsocEndpoint, Exit);

						FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kASIsocEndpoint, asIsocEndpoint), Exit);
						asIsocEndpoint = NULL;
					}

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				default:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() - @ Default (else)", this);

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);
				}
			}
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor() = %p", this, theInterfacePtr);
	return theInterfacePtr;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface ディスクリプタを解析（UAC 2.0 用）．
 *
 * @param[in]		theInterfacePtr		AudioStreaming Interface ディスクリプタ．
 * @param[in]		currentInterface	interface 番号．
 * @param[in, out]	*parsedLength		コンフィグレーションディスクリプタのうち、解析済みのバイト数．
 * @param[in]		totalLength			コンフィグレーションディスクリプタの総バイト数．
 *
 * @return			次のディスクリプタへのポインタ．
 */
USBInterfaceDescriptorPtr
PdjUsbStreamDictionary::parseASInterfaceDescriptor_0200(
	USBInterfaceDescriptorPtr	theInterfacePtr,
	UInt8 const					currentInterface,
	UInt16						*parsedLength,
	UInt16						totalLength
)
{
	DP(DBG_TRACE, "+ " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200(%p, %d)", this, theInterfacePtr, currentInterface);

	PdjUsbEndpointDictionary		*thisEndpoint		= NULL;
	PdjUsbASEndpointDictionary		*asIsocEndpoint		= NULL;
	PdjUsbEndpointDictionary		*dataEndpoint		= NULL;
	PdjUsbEndpointDictionary		*feedbackEndpoint	= NULL;
	UInt8							endpointAddress;
	UInt32							formats;
	UInt16							pseudoFormatTag;
	UInt8							initialBLength;
	bool							done = false;

	FailIf(NULL == theInterfacePtr, Exit);
	FailIf(0 == theInterfacePtr->bLength, Exit);

	initialBLength = theInterfacePtr->bLength;

	while (
				(theInterfacePtr)
		  &&
				(theInterfacePtr->bLength > 0)
		  &&
				(!done)
		  &&
				(*parsedLength <= totalLength)
		  )
	{
		if (theInterfacePtr->bLength < initialBLength)
		{
			this->logDescriptor((UInt8 *)theInterfacePtr, theInterfacePtr->bLength);
		}

		if (CS_INTERFACE == theInterfacePtr->bDescriptorType)
		{
			switch (theInterfacePtr->bDescriptorSubtype)
			{
				case USBAUDIO_0200::AS_GENERAL:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ AS_GENERAL (4.9.2)", this);

					FailIf(
								kIOReturnSuccess != super::setDictionaryValue(
																kTerminalLink,
																((USBAUDIO_0200::ASInterfaceDescriptorPtr)theInterfacePtr)->bTerminalLink
														   ),
								Exit
						  );

					formats = (
									(((USBAUDIO_0200::ASInterfaceDescriptorPtr)theInterfacePtr)->bmFormats[3] << 24)
							  |
									(((USBAUDIO_0200::ASInterfaceDescriptorPtr)theInterfacePtr)->bmFormats[2] << 16)
							  |
									(((USBAUDIO_0200::ASInterfaceDescriptorPtr)theInterfacePtr)->bmFormats[1] <<  8)
							  |
									(((USBAUDIO_0200::ASInterfaceDescriptorPtr)theInterfacePtr)->bmFormats[0] <<  0)
							  );

					FailIf(kIOReturnSuccess != super::setDictionaryValue(kFormats, formats), Exit);

					// [rdar://5600018] TODO: Should more fully support Type III formats
					switch (((USBAUDIO_0200::ASInterfaceDescriptorPtr)theInterfacePtr)->bFormatType)
					{
						case USBAUDIO_0200::FORMAT_TYPE_I:
						{
							if (formats & USBAUDIO_0200::PCM)
							{
								pseudoFormatTag = PCM;
							}
							else if (formats & USBAUDIO_0200::PCM8)
							{
								pseudoFormatTag = PCM8;
							}
							else if (formats & USBAUDIO_0200::IEEE_FLOAT)
							{
								pseudoFormatTag = IEEE_FLOAT;
							}
							else if (formats & USBAUDIO_0200::ALAW)
							{
								pseudoFormatTag = ALAW;
							}
							else
							{
								pseudoFormatTag = TYPE_I_UNDEFINED;
							}

							break;
						}
						case USBAUDIO_0200::FORMAT_TYPE_II:
						{
							if (formats & USBAUDIO_0200::MPEG)
							{
								pseudoFormatTag = MPEG;
							}
							else if (formats & USBAUDIO_0200::AC3)
							{
								pseudoFormatTag = AC3;
							}
							else
							{
								pseudoFormatTag = TYPE_II_UNDEFINED;
							}

							break;
						}
						case USBAUDIO_0200::FORMAT_TYPE_III:
						{
							if (formats & USBAUDIO_0200::IEC61937_AC3)
							{
								pseudoFormatTag = IEC1937_AC3;
							}
							else if (formats & USBAUDIO_0200::IEC1937_MPEG1_Layer1)
							{
								pseudoFormatTag = IEC1937_MPEG1_Layer1;
							}
							else if (formats & USBAUDIO_0200::IEC1937_MPEG1_Layer2or3)
							{
								pseudoFormatTag = IEC1937_MPEG1_Layer2or3;
							}
							else if (formats & USBAUDIO_0200::IEC61937_MPEG2_EXT)
							{
								pseudoFormatTag = IEC1937_MPEG2_EXT;
							}
							else if (formats & USBAUDIO_0200::IEC1937_MPEG2_Layer1_LS)
							{
								pseudoFormatTag = IEC1937_MPEG2_Layer1_LS;
							}
							else if (formats & USBAUDIO_0200::IEC1937_MPEG2_Layer2or3_LS)
							{
								pseudoFormatTag = IEC1937_MPEG2_Layer2or3_LS;
							}
							else
							{
								pseudoFormatTag = TYPE_III_UNDEFINED;
							}

							break;
						}
						default:
						{
							pseudoFormatTag = 0;

							break;
						}
					}

					FailIf(
								kIOReturnSuccess != super::setDictionaryValue(
																kFormatTag,
																pseudoFormatTag
														   ),
								Exit
						  );

					FailIf(
								kIOReturnSuccess != super::setDictionaryValue(
																kNumChannels,
																((USBAUDIO_0200::ASInterfaceDescriptorPtr)theInterfacePtr)->bNrChannels
														   ),
								Exit
						  );

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				case USBAUDIO_0200::FORMAT_TYPE:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ FORMAT_TYPE", this);

					switch (((USBAUDIO_0200::ASFormatTypeIDescriptorPtr)theInterfacePtr)->bFormatType)
					{
						case USBAUDIO_0200::FORMAT_TYPE_I:
						case USBAUDIO_0200::FORMAT_TYPE_III:
						{
							DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ FORMAT_TYPE_I/FORMAT_TYPE_III (Format 2.3.1.6/2.3.3.1)", this);

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kSubframeSize,
																		((USBAUDIO_0200::ASFormatTypeIDescriptorPtr)theInterfacePtr)->bSubslotSize
																   ),
										Exit
								  );

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kBitResolution,
																		((USBAUDIO_0200::ASFormatTypeIDescriptorPtr)theInterfacePtr)->bBitResolution
																   ),
										Exit
								  );

							break;
						}
						case USBAUDIO_0200::FORMAT_TYPE_II:
						{
							DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ FORMAT_TYPE_II (Format 2.3.2.6)", this);

							UInt16		maxBitRate;
							UInt16		samplesPerFrame;

							maxBitRate = USBToHostWord(((USBAUDIO_0200::ASFormatTypeIIDescriptorPtr)theInterfacePtr)->wMaxBitRate);
							FailIf(kIOReturnSuccess != super::setDictionaryValue(kMaxBitRate, maxBitRate), Exit);

							samplesPerFrame = USBToHostWord(((USBAUDIO_0200::ASFormatTypeIIDescriptorPtr)theInterfacePtr)->wSlotsPerFrame);
							FailIf(kIOReturnSuccess != super::setDictionaryValue(kSamplesPerFrame, samplesPerFrame), Exit);

							break;
						}
						default:
						{
							DP(DBG_ERROR, "! " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ Unknown/Unsupported Format Type!", this);
						}
					}

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				case USBAUDIO_0200::ENCODER:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ ENCODER", this);

					UInt8		bEncoder;

					bEncoder = ((USBAUDIO_0200::ASEncoderDescriptorPtr)theInterfacePtr)->bEncoder;

					switch (bEncoder)
					{
						default:
						{
							DP(DBG_ERROR, "! " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ Unknown/unsupported encoder type 0x%x",
								this, bEncoder);
							break;
						}
					}

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				case USBAUDIO_0200::DECODER:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ DECODER", this);

					UInt32		bmAC3BSID;
					UInt16		bmMPEGCapabilities;
					UInt8		bDecoder;

					bDecoder = ((USBAUDIO_0200::ASDecoderDescriptorPtr)theInterfacePtr)->bDecoder;

					switch (bDecoder)
					{
						case USBAUDIO_0200::MPEG_DECODER:
						{
							DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ MPEG (4.9.5.1)", this);

							bmMPEGCapabilities = USBToHostWord(
														(((USBAUDIO_0200::ASMPEGDecoderDescriptorPtr)theInterfacePtr)->bmMPEGCapabilities[1] << 8)
												 |
														(((USBAUDIO_0200::ASMPEGDecoderDescriptorPtr)theInterfacePtr)->bmMPEGCapabilities[0] << 0)
												 );

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kMPEGCapabilities,
																		bmMPEGCapabilities
																   ),
										Exit
								  );

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kMPEGFeatures,
																		((USBAUDIO_0200::ASMPEGDecoderDescriptorPtr)theInterfacePtr)->bmMPEGFeatures
																   ),
										Exit
								  );

							break;
						}
						case USBAUDIO_0200::AC3_DECODER:
						{
							DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ AC3 (4.9.5.2)", this);

							bmAC3BSID = USBToHostLong(
											(((USBAUDIO_0200::ASAC3DecoderDescriptorPtr)theInterfacePtr)->bmBSID[3] << 24)
										|
											(((USBAUDIO_0200::ASAC3DecoderDescriptorPtr)theInterfacePtr)->bmBSID[2] << 16)
										|
											(((USBAUDIO_0200::ASAC3DecoderDescriptorPtr)theInterfacePtr)->bmBSID[1] <<  8)
										|
											(((USBAUDIO_0200::ASAC3DecoderDescriptorPtr)theInterfacePtr)->bmBSID[0] <<  0)
										);

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kAC3BSID,
																		bmAC3BSID
																   ),
										Exit
								  );

							FailIf(
										kIOReturnSuccess != super::setDictionaryValue(
																		kAC3Features,
																		((ASAC3FormatSpecificDescriptorPtr)theInterfacePtr)->bmAC3Features
																   ),
										Exit
								  );

							break;
						}
						default:
						{
							DP(DBG_WARNING, "! " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ Unknown/unsupported decoder type 0x%x",
								this, bDecoder);
							break;
						}
					}

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				default:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ Default", this);

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);
				}
			}
		}
		else
		{
			OSArray		*endpoints = 0;

			switch (theInterfacePtr->bDescriptorType)
			{
				case INTERFACE:
				{
					// Need to make a new interface dictionary for this new interface or new alternate setting
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ INTERFACE", this);

					done = true;

					break;
				}
				case ENDPOINT:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ ENDPOINT (4.10.1.1)", this);

					UInt16		maxPacketSize;		// <rdar://problem/6789044>
					UInt16		numTransactions;	// <rdar://problem/6789044>

					FailIf(NULL == (thisEndpoint = PdjUsbEndpointDictionary::create()), Exit);

					thisEndpoint->setAddress(((USBAUDIO_0200::USBEndpointDescriptorPtr)theInterfacePtr)->bEndpointAddress);

					thisEndpoint->setAttributes(((USBAUDIO_0200::USBEndpointDescriptorPtr)theInterfacePtr)->bmAttributes);

					// <rdar://problem/6789044>
					maxPacketSize = ((USBAUDIO_0200::USBEndpointDescriptorPtr)theInterfacePtr)->wMaxPacketSize;

					numTransactions = ((maxPacketSize & kTransactionsPerMicroframe_Mask) >> 11) + 1;

					// (Bits 10..0 of wMaxPacketSize) * (number of transactions per microframe)
					thisEndpoint->setMaxPacketSize(USBToHostWord((maxPacketSize & kMaxPacketSize_Mask) * numTransactions));

					thisEndpoint->setInterval(((USBAUDIO_0200::USBEndpointDescriptorPtr)theInterfacePtr)->bInterval);

					endpoints = this->getEndpoints();

					if (NULL == endpoints)
					{
						FailIf(NULL == (endpoints = OSArray::withObjects((const OSObject **)&thisEndpoint, 1)), Exit);
						FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kEndpoints, endpoints), Exit);
						endpoints = NULL;
					}
					else
					{
						endpoints->setObject(thisEndpoint);
					}

					if (thisEndpoint->isIsocStreaming())
					{
						dataEndpoint = thisEndpoint;
					}
					else if (thisEndpoint->isIsocFeedback())
					{
						thisEndpoint->setRefreshInt(((USBAUDIO_0200::USBEndpointDescriptorPtr)theInterfacePtr)->bInterval - 1 - 3);

						feedbackEndpoint = thisEndpoint;
					}

					if ((NULL != feedbackEndpoint) && (NULL != dataEndpoint))
					{
						if (kIOReturnSuccess == feedbackEndpoint->getAddress(&endpointAddress))
						{
							dataEndpoint->setSynchAddress(endpointAddress);
						}
					}

					thisEndpoint->release();
					thisEndpoint = NULL;

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				case CS_ENDPOINT:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ CS_ENDPOINT (4.10.1.2)", this);

					if (EP_GENERAL == ((ASEndpointDescriptorPtr)theInterfacePtr)->bDescriptorSubtype)
					{
						DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ EP_GENERAL", this);

						asIsocEndpoint = new PdjUsbASEndpointDictionary(
													0,
													(((USBAUDIO_0200::ASEndpointDescriptorPtr)theInterfacePtr)->bmControls & 0x3) == 0x3,
													((USBAUDIO_0200::ASEndpointDescriptorPtr)theInterfacePtr)->bmAttributes & (1 << maxPacketsOnlyBit),
													((USBAUDIO_0200::ASEndpointDescriptorPtr)theInterfacePtr)->bLockDelayUnits,
													((USBAUDIO_0200::ASEndpointDescriptorPtr)theInterfacePtr)->wLockDelay
											 );
						FailIf(NULL == asIsocEndpoint, Exit);

						FailIf(kIOReturnSuccess != super::setDictionaryObjectAndRelease(kASIsocEndpoint, asIsocEndpoint), Exit);
						asIsocEndpoint = NULL;
					}

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);

					break;
				}
				default:
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() - @ Default (else)", this);

					theInterfacePtr = (USBInterfaceDescriptorPtr)((UInt8 *)theInterfacePtr + theInterfacePtr->bLength);

					*parsedLength = (theInterfacePtr ? (*parsedLength + theInterfacePtr->bLength) : totalLength);
				}
			}
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRSTREAMDICTIONARY "[%p]::parseASInterfaceDescriptor_0200() = %p", this, theInterfacePtr);
	return theInterfacePtr;
}

/****************************************************************************/
/**
 * @brief	ストリーム dictionary にサンプリング周波数を追加．
 *
 * @param[in]		*sampleRates	設定可能なサンプリング周波数の OSArray．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::addSampleRatesToStreamDictionary(
	OSArray			*sampleRates
)
{
	DP(DBG_TRACE, "+ " LOGSTRSTREAMDICTIONARY "::addSampleRatesToStreamDictionary()");

	IOReturn		result = kIOReturnError;
	OSArray			*existingSampleRates		= NULL;
	OSNumber		*sampleRateNumber			= NULL;
	OSNumber		*existingSampleRateNumber	= NULL;
	UInt8			numSampleFreqs = 0;
	bool			isDuplicate;

	FailIf(NULL == sampleRates, Exit);

	existingSampleRates = this->getSampleRates();

	for (UInt16 sampleRateIndex = 0; sampleRateIndex < sampleRates->getCount(); sampleRateIndex++)
	{
		FailIf(NULL == (sampleRateNumber = OSDynamicCast(OSNumber, sampleRates->getObject(sampleRateIndex))), Exit);

		isDuplicate = false;

		if (existingSampleRates)
		{
			// Check to make sure we aren't adding a duplicate
			for (UInt16 rateIndex= 0; rateIndex < existingSampleRates->getCount(); rateIndex++)
			{
				FailIf(NULL == (existingSampleRateNumber = OSDynamicCast(OSNumber, existingSampleRates->getObject(rateIndex))), Exit);

				if (existingSampleRateNumber->isEqualTo(sampleRateNumber))
				{
					isDuplicate = true;

					break;
				}
			}
		}

		if (!isDuplicate)
		{
			if (sampleRateNumber->unsigned32BitValue())
			{
				OSArray							*endpoints	= NULL;
				PdjUsbEndpointDictionary		*endpoint	= NULL;
				UInt8							direction	= 0;
				bool							found		= false;

				//[rdar://4801012] Only add the sample rate if the average frame size doesn't exceed the max packet size
				// We can cheat to get the isoc endpoint in UAC 2.0
				FailIf(NULL == (endpoints = this->getEndpoints()), Exit);

				for (UInt8 endpointIndex = 0; endpointIndex < endpoints->getCount(); endpointIndex++)
				{
					FailIf(NULL == (endpoint = OSDynamicCast(PdjUsbEndpointDictionary, endpoints->getObject(endpointIndex))), Exit);

					if (endpoint->isIsocStreaming())
					{
						found = true;

						FailIf(kIOReturnSuccess != endpoint->getDirection(&direction), Exit);

						break;
					}
				}

				FailIf(!found, Exit);

				UInt32		averageFrameSamples;
				UInt16		averageFrameSize;
				UInt16		maxPacketSize;
				UInt8		numChannels;
				UInt8		bitResolution;
				UInt8		interval;
				UInt8		transactionsPerUSBFrame;

				FailIf(kIOReturnSuccess != this->getIsocEndpointMaxPacketSize(&maxPacketSize, direction), Exit);
				FailIf(kIOReturnSuccess != this->getNumChannels(&numChannels), Exit);
				FailIf(kIOReturnSuccess != this->getBitResolution(&bitResolution), Exit);

				// Must determine the number of transfer opportunities per millisecond.
				FailIf(kIOReturnSuccess != this->getIsocEndpointInterval(&interval, direction), Exit);

				if (0 == interval)
				{
					DP(DBG_ERROR, "! " LOGSTRSTREAMDICTIONARY "[%p]::addSampleRatesToStreamDictionary() - ERROR! Isoc endpoint has a refresh interval of 0! Treating as 4 ...", this);

					transactionsPerUSBFrame = 1;
				}
				else
				{
					FailIf(interval > 4, Exit);

					transactionsPerUSBFrame = 8 >> (interval - 1);
				}

				averageFrameSamples = sampleRateNumber->unsigned32BitValue() / (1000 * transactionsPerUSBFrame);
				averageFrameSize = (averageFrameSamples * numChannels * (bitResolution / 8));

				// Continue the loop if this alternate setting can't add this sample rate
				if (averageFrameSize > maxPacketSize)
				{
					DP(DBG_ERROR, "! " LOGSTRSTREAMDICTIONARY "::addSampleRatesToStreamDictionary() - cannot add sample rate %u due to packet size constraints!",
						sampleRateNumber->unsigned32BitValue());
					continue;
				}
				else
				{
					DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "::addSampleRatesToStreamDictionary() - adding sample rate %u",
						sampleRateNumber->unsigned32BitValue());

					FailIf(kIOReturnSuccess != this->addSampleRate(sampleRateNumber->unsigned32BitValue()), Exit);
				}
			}
		}
	}

	existingSampleRates = this->getSampleRates();

	if (existingSampleRates)
	{
		numSampleFreqs = existingSampleRates->getCount();
	}

	FailIf(kIOReturnSuccess != super::setDictionaryValue(kNumSampleRates, numSampleFreqs), Exit);

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRSTREAMDICTIONARY "::addSampleRatesToStreamDictionary() = 0x%x", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	エンドポイントディスクリプタから LockDelay 値を取得．
 *
 * @param[out]		*lockDelay		取得した LockDelay 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::asEndpointGetLockDelay(
	UInt8			*lockDelay
)
{
	IOReturn						result = kIOReturnError;
	PdjUsbASEndpointDictionary		*asEndpoint = NULL;

	FailIf(NULL == (asEndpoint = this->getASEndpointDictionary()), Exit);
	FailIf(kIOReturnSuccess != (result = asEndpoint->getLockDelay(lockDelay)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	エンドポイントディスクリプタから LockDelayUnits 値を取得．
 *
 * @param[out]		*lockDelayUnits		取得した LockDelayUnits 値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::asEndpointGetLockDelayUnits(
	UInt8			*lockDelayUnits
)
{
	IOReturn						result = kIOReturnError;
	PdjUsbASEndpointDictionary		*asEndpoint = NULL;

	FailIf(NULL == (asEndpoint = this->getASEndpointDictionary()), Exit);
	FailIf(kIOReturnSuccess != (result = asEndpoint->getLockDelayUnits(lockDelayUnits)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	エンドポイントがサンプリング周波数設定に対応しているかどうかを調べる．
 *			エンドポイントディスクリプタの bmAttributes フィールドで
 *			Sampling Frequency Control がサポートされているかどうか判定する．
 *
 * @retval			true			サンプリング周波数設定に対応している．
 * @retval			false			サンプリング周波数設定に対応していない．
 */
bool
PdjUsbStreamDictionary::asEndpointHasSampleFreqControl(
	void
)
{
	bool							result = false;
	PdjUsbASEndpointDictionary		*asEndpoint = NULL;

	FailIf(NULL == (asEndpoint = this->getASEndpointDictionary()), Exit);
	FailIf(kIOReturnSuccess != asEndpoint->hasSampleFreqControl(&result), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	エンドポイント dictionary の OSArray を取得．
 *
 * @return			エンドポイント dictionary の OSArray．
 */
OSArray *
PdjUsbStreamDictionary::getEndpoints(
	void
)
{
	OSArray			*endpoints		 = NULL;
	OSObject		*dictionaryValue = NULL;

	dictionaryValue = OSDictionary::getObject(kEndpoints);

	if (dictionaryValue)
	{
		endpoints = OSDynamicCast(OSArray, dictionaryValue);
	}

	return endpoints;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントのアドレスを取得．
 *
 * @param[out]		*address		取得したエンドポイントアドレス．
 * @param[in]		direction		転送方向．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::getIsocEndpointAddress(
	UInt8			*address,
	UInt8			direction
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbEndpointDictionary	*thisEndpoint	= NULL;
	OSArray						*endpoints		= NULL;
	UInt8						endpointIndex	= 0;
	UInt8						thisDirection;

	*address = 0;

	FailIf(NULL == address, Exit);
	FailIf(NULL == (endpoints = this->getEndpoints()), Exit);

	while ((0 == *address) && (endpointIndex < endpoints->getCount()))
	{
		FailIf(NULL == (thisEndpoint = this->getIndexedEndpointDictionary(endpointIndex)), Exit);
		FailIf(kIOReturnSuccess != (result = thisEndpoint->getDirection(&thisDirection)), Exit);

		if (direction == thisDirection)
		{
			FailIf(kIOReturnSuccess != (result = thisEndpoint->getAddress(address)), Exit);
		}

		endpointIndex++;
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの転送方向を取得．
 *
 * @param[out]		*direction		取得した転送方向．
 * @param[in]		index			OSArray のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::getIsocEndpointDirection(
	UInt8			*direction,
	UInt8			index
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbEndpointDictionary	*thisEndpoint	= NULL;
	OSArray						*endpoints		= NULL;

	*direction = 0xFF;

	FailIf(NULL == (endpoints = this->getEndpoints()), Exit);
	FailIf(NULL == (thisEndpoint = this->getIndexedEndpointDictionary(index)), Exit);
	FailIf(kIOReturnSuccess != (result = thisEndpoint->getDirection(direction)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの MaxPacketSize を取得．
 *
 * @param[out]		*maxPacketSize	取得した MaxPacketSize．
 * @param[in]		direction		転送方向．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::getIsocEndpointMaxPacketSize(
	UInt16			*maxPacketSize,
	UInt8			direction
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbEndpointDictionary	*thisEndpoint	= NULL;
	OSArray						*endpoints		= NULL;
	UInt8						endpointIndex = 0;
	UInt8						thisDirection;

	FailIf(NULL == maxPacketSize, Exit);
	FailIf(NULL == (endpoints = this->getEndpoints()), Exit);

	*maxPacketSize = 0;

	while ((0 == *maxPacketSize) && (endpointIndex < endpoints->getCount()))
	{
		FailIf(NULL == (thisEndpoint = this->getIndexedEndpointDictionary(endpointIndex)), Exit);
		FailIf(kIOReturnSuccess != (result = thisEndpoint->getDirection(&thisDirection)), Exit);

		if (direction == thisDirection)
		{
			// This is the isoc endpoint for which we are looking.
			FailIf(kIOReturnSuccess != (result = thisEndpoint->getMaxPacketSize(maxPacketSize)), Exit);
		}

		endpointIndex++;
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの同期タイプを取得．
 *
 * @param[out]		*syncType		エンドポイントディスクリプタの bmAttributes
 *									フィールドで指定されている同期タイプ．
 * @param[in]		address			エンドポイントアドレス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::getIsocEndpointSyncType(
	UInt8			*syncType,
	UInt8			address
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbEndpointDictionary	*thisEndpoint = NULL;

	*syncType = 0;

	FailIf(NULL == (thisEndpoint = this->getEndpointByAddress(address)), Exit);
	FailIf(kIOReturnSuccess != (result = thisEndpoint->getSyncType(syncType)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントのインターバルを取得．
 *
 * @param[out]		*interval		エンドポイントディスクリプタの bInterval フィールド．
 * @param[in]		direction		転送方向．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::getIsocEndpointInterval(
	UInt8			*interval,
	UInt8			direction
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbEndpointDictionary	*thisEndpoint	= NULL;
	OSArray						*endpoints		= NULL;
	UInt8						endpointIndex = 0;
	UInt8						thisDirection;

	FailIf(NULL == interval, Exit);
	FailIf(NULL == (endpoints = this->getEndpoints()), Exit);

	*interval = 0;

	while ((0 == *interval) && (endpointIndex < endpoints->getCount()))
	{
		FailIf(NULL == (thisEndpoint = this->getIndexedEndpointDictionary(endpointIndex)), Exit);
		FailIf(kIOReturnSuccess != (result = thisEndpoint->getDirection(&thisDirection)), Exit);

		if (direction == thisDirection)
		{
			// This is the isoc endpoint for which we are looking.
			FailIf(!thisEndpoint->isIsocStreaming(), Exit);
			FailIf(kIOReturnSuccess != (result = thisEndpoint->getInterval(interval)), Exit);

			DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "::getIsocEndpointInterval(%p, %d) = %d", interval, direction, *interval);
		}
		else
		{
			DP(DBG_INFO, "? " LOGSTRSTREAMDICTIONARY "::getIsocEndpointInterval() - Endpoint at index %d has wrong direction %d (!= %d)",
				endpointIndex, thisDirection, direction);
		}

		endpointIndex++;
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの付随エンドポイントのアドレスを取得．
 *
 * @param[out]		*assocEndpointAddress	取得した付随エンドポイントのアドレス．
 * @param[in]		address					エンドポイントアドレス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::getIsocAssociatedEndpointAddress(
	UInt8			*assocEndpointAddress,
	UInt8			address
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbEndpointDictionary	*thisEndpoint = NULL;

	*assocEndpointAddress = 0;

	FailIf(NULL == (thisEndpoint = this->getEndpointByAddress(address)), Exit);
	FailIf(kIOReturnSuccess != (result = thisEndpoint->getSynchAddress(assocEndpointAddress)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの付随エンドポイントの MaxPacketSize を取得．
 *
 * @param[out]		*assocEndpointMaxPacketSize		取得した MaxPacketSize．
 * @param[in]		address							エンドポイントアドレス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::getIsocAssociatedEndpointMaxPacketSize(
	UInt16			*assocEndpointMaxPacketSize,
	UInt8			address
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbEndpointDictionary	*thisEndpoint = NULL;

	*assocEndpointMaxPacketSize = 0;

	FailIf(NULL == (thisEndpoint = this->getEndpointByAddress(address)), Exit);
	FailIf(kIOReturnSuccess != (result = thisEndpoint->getMaxPacketSize(assocEndpointMaxPacketSize)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Isochronous エンドポイントの付随エンドポイントのリフレッシュ間隔を取得．
 *
 * @param[out]		*assocEndpointRefreshInt	エンドポイントディスクリプタの bRefresh フィールド．
 * @param[in]		address						エンドポイントアドレス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::getIsocAssociatedEndpointRefreshInt(
	UInt8			*assocEndpointRefreshInt,
	UInt8			address
)
{
	IOReturn					result = kIOReturnError;
	PdjUsbEndpointDictionary	*thisEndpoint = NULL;

	*assocEndpointRefreshInt = 0;

	FailIf(NULL == (thisEndpoint = this->getEndpointByAddress(address)), Exit);
	FailIf(kIOReturnSuccess != (result = thisEndpoint->getRefreshInt(assocEndpointRefreshInt)), Exit);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	サンプリング周波数を追加．
 *
 * @param[in]		sampleRate		追加するサンプリング周波数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbStreamDictionary::addSampleRate(
	UInt32			sampleRate
)
{
	IOReturn		result = kIOReturnError;
	OSArray			*sampleRates		= NULL;
	OSNumber		*sampleRateNumber	= NULL;

	FailIf(NULL == (sampleRateNumber = OSNumber::withNumber(sampleRate, SIZEINBITS(UInt32))), Exit);

	sampleRates = this->getSampleRates();

	if (NULL == sampleRates)
	{
		FailIf(NULL == (sampleRates = OSArray::withObjects((const OSObject **)&sampleRateNumber, 1)), Exit);
		FailIf(kIOReturnSuccess != (result = super::setDictionaryObjectAndRelease(kSampleRates, sampleRates)), Exit);
		sampleRates = NULL;
	}
	else
	{
		FailIf(!sampleRates->setObject(sampleRateNumber), Exit);
	}

	result = kIOReturnSuccess;

Exit:
	if (NULL != sampleRateNumber)
	{
		sampleRateNumber->release();
		sampleRateNumber = NULL;
	}

	return result;
}

/****************************************************************************/
/**
 * @brief	エンドポイント dictionary を取得（エンドポイントアドレスから）．
 *
 * @param[in]		address			エンドポイントアドレス．
 *
 * @return			エンドポイント dictionary．
 */
PdjUsbEndpointDictionary *
PdjUsbStreamDictionary::getEndpointByAddress(
	UInt8			address
)
{
	PdjUsbEndpointDictionary	*thisEndpoint	= NULL;
	OSArray						*endpoints		= NULL;
	UInt8						endpointIndex = 0;
	UInt8						thisAddress;
	bool						found = false;

	FailIf(NULL == (endpoints = this->getEndpoints()), Exit);

	while ((!found) && (endpointIndex < endpoints->getCount()))
	{
		FailIf(NULL == (thisEndpoint = this->getIndexedEndpointDictionary(endpointIndex)), Exit);
		FailIf(kIOReturnSuccess != thisEndpoint->getAddress(&thisAddress), Exit);

		if (thisAddress == address)
		{
			found = true;
		}

		endpointIndex++;
	}

Exit:
	if (!found)
	{
		thisEndpoint = NULL;
	}

	return thisEndpoint;
}

/****************************************************************************/
/**
 * @brief	エンドポイント dictionary を取得（OSArray のインデックスから）．
 *
 * @param[in]		index			OSArray のインデックス．
 *
 * @return			エンドポイント dictionary．
 */
PdjUsbEndpointDictionary *
PdjUsbStreamDictionary::getIndexedEndpointDictionary(
	UInt8			index
)
{
	PdjUsbEndpointDictionary	*thisEndpoint	= NULL;
	OSArray						*endpoints		= NULL;

	FailIf(NULL == (endpoints = this->getEndpoints()), Exit);
	FailIf(NULL == (thisEndpoint = OSDynamicCast(PdjUsbEndpointDictionary, endpoints->getObject(index))), Exit);

Exit:
	return thisEndpoint;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming エンドポイント dictionary を取得．
 *
 * @return			AudioStreaming エンドポイント dictionary．
 */
PdjUsbASEndpointDictionary *
PdjUsbStreamDictionary::getASEndpointDictionary(
	void
)
{
	OSObject						*dictionaryValue = NULL;
	PdjUsbASEndpointDictionary		*asEndpoint		 = NULL;

	FailIf(NULL == (dictionaryValue = OSDictionary::getObject(kASIsocEndpoint)), Exit);
	FailIf(NULL == (asEndpoint = OSDynamicCast(PdjUsbASEndpointDictionary, dictionaryValue)), Exit);

Exit:
	return asEndpoint;
}

/*------------------------------------------*/
/*		PdjUsbEndpointDictionary クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbEndpointDictionary -

OSDefineMetaClassAndStructors(PdjUsbEndpointDictionary, super)

/****************************************************************************/
/**
 * @brief	エンドポイント dictionary を生成．
 *
 * @return			生成されたエンドポイント dictionary．
 */
PdjUsbEndpointDictionary *
PdjUsbEndpointDictionary::create(
	void
)
{
	PdjUsbEndpointDictionary	*endpointDictionary = NULL;

	FailIf(NULL == (endpointDictionary = new PdjUsbEndpointDictionary), Exit);
	RELEASE_IF_FALSE(endpointDictionary, endpointDictionary->initDictionaryForUse());

Exit:
	return endpointDictionary;
}

/****************************************************************************/
/**
 * @brief	転送方向を取得．
 *
 * @param[out]		*direction		エンドポイントアドレスから判別した転送方向．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbEndpointDictionary::getDirection(
	UInt8			*direction
)
{
	IOReturn		result = kIOReturnError;
	UInt8			address;

	FailIf(NULL == direction, Exit);
	FailIf(kIOReturnSuccess != (result = this->getAddress(&address)), Exit);

	*direction = (address & 0x80) >> 7;

	result = kIOReturnSuccess;

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	同期タイプを取得．
 *
 * @param[out]		*syncType		ディスクリプタの bmAttributes フィールドで
 *									指定されている同期タイプ．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbEndpointDictionary::getSyncType(
	UInt8			*syncType
)
{
	IOReturn		result = kIOReturnError;
	UInt8			attributes;

	FailIf(NULL == syncType, Exit);
	FailIf(kIOReturnSuccess != (result = this->getAttributes(&attributes)), Exit);

	*syncType = (attributes & 0xC) >> 2;

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Streaming 用 Isochronous エンドポイントであるかどうか判定．
 *				転送タイプ ＝ Isochronous
 *				同期タイプ ≠ No Synchronization
 *				使用タイプ ≠ Feedback Endpoint
 *
 * @retval			true			Streaming 用 Isochronous エンドポイントである．
 * @retval			false			Streaming 用 Isochronous エンドポイントではない．
 */
bool
PdjUsbEndpointDictionary::isIsocStreaming(
	void
)
{
	bool			result = false;
	UInt8			attributes;

	FailIf(kIOReturnSuccess != this->getAttributes(&attributes), Exit);

#ifdef BUG_FIX
	result = (
					(0x1 == ((attributes >> 0) & 0x3))
			 &&
					(0x0 != ((attributes >> 2) & 0x3))
			 &&
					(0x1 != ((attributes >> 4) & 0x3))
			 );
#else
	result = (
					(0x1 == ((attributes >> 0) & 0x1))
			 &&
					(0x0 != ((attributes >> 2) & 0x3))
			 &&
					(0x1 != ((attributes >> 4) & 0x3))
			 );
#endif

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	Feedback 用 Isochronous エンドポイントであるかどうか判定．
 *				転送タイプ ＝ Isochronous
 *				同期タイプ ＝ No Synchronization
 *				使用タイプ ＝ Feedback Endpoint
 *
 * @retval			true			Feedback 用 Isochronous エンドポイントである．
 * @retval			false			Feedback 用 Isochronous エンドポイントではない．
 */
bool
PdjUsbEndpointDictionary::isIsocFeedback(
	void
)
{
	bool			result = false;
	UInt8			attributes;

	FailIf(kIOReturnSuccess != this->getAttributes(&attributes), Exit);

#ifdef BUG_FIX
	result = (
					(0x1 == ((attributes >> 0) & 0x3))
			 &&
					(0x0 == ((attributes >> 2) & 0x3))
			 &&
					(0x1 == ((attributes >> 4) & 0x3))
			 );
#else
	result = (
					(0x1 == ((attributes >> 0) & 0x1))
			 &&
					(0x0 == ((attributes >> 2) & 0x3))
			 &&
					(0x1 == ((attributes >> 4) & 0x3))
			 );
#endif

Exit:
	return result;
}

/*------------------------------------------*/
/*	PdjUsbASEndpointDictionary クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbASEndpointDictionary -

/****************************************************************************/
/**
 * @brief	AudioStreaming エンドポイント dictionary のコンストラクタ．
 *
 * @param[in]		theSampleFreqControl	bmAttributes フィールドで指定された周波数コントロールの有無．
 * @param[in]		thePitchControl			bmAttributes フィールドで指定されたピッチコントロールの有無．
 * @param[in]		theMaxPacketsOnly		bmAttributes フィールドで指定されたショートパケット受け入れ可否．
 * @param[in]		theLockDelayUnits		ディスクリプタの bLockDelayUnits フィールド．
 * @param[in]		theLockDelay			ディスクリプタの wLockDelay フィールド．
 */
PdjUsbASEndpointDictionary::PdjUsbASEndpointDictionary(
	bool			theSampleFreqControl,
	bool			thePitchControl,
	bool			theMaxPacketsOnly,
	UInt8			theLockDelayUnits,
	UInt16			theLockDelay
)
{
	bool			result = false;

	FailIf(!(result = super::super::initDictionaryForUse()), Exit);

	super::super::setDictionaryValue(kHasSampleFreqControl,	theSampleFreqControl);
	super::super::setDictionaryValue(kHasPitchControl,		thePitchControl);
	super::super::setDictionaryValue(kHasMaxPacketsOnly,	theMaxPacketsOnly);
	super::super::setDictionaryValue(kLockDelayUnits,		theLockDelayUnits);
	super::super::setDictionaryValue(kLockDelay,			theLockDelay);

Exit:
	DP(DBG_INFO, "? " LOGSTRASENDPOINTDICTIONARY "::UsbASEndpointDictionary(%d, %d, %d, %d, %d) = %d",
		theSampleFreqControl, thePitchControl, theMaxPacketsOnly, theLockDelayUnits, theLockDelay, result);
}

/*------------------------------------------*/
/*		PdjUsbUnitDictionary クラス			*/
/*------------------------------------------*/

#pragma mark - class PdjUsbUnitDictionary -

OSDefineMetaClassAndStructors(PdjUsbUnitDictionary, super)

/*------------------------------------------*/
/*	PdjUsbInputTerminalDictionary クラス	*/
/*------------------------------------------*/

#pragma mark - class PdjUsbInputTerminalDictionary -

/*------------------------------------------*/
/*	PdjUsbOutputTerminalDictionary クラス	*/
/*------------------------------------------*/

#pragma mark - class PdjUsbOutputTerminalDictionary -

/*------------------------------------------*/
/*		PdjUsbMixerUnitDictionary クラス	*/
/*------------------------------------------*/

#pragma mark - class PdjUsbMixerUnitDictionary -

/****************************************************************************/
/**
 * @brief	ミキサーユニットディスクリプタの bmControls フィールドを設定．
 *
 * @param[in]		*bmCntrls		ディスクリプタの bmControls フィールド．
 * @param[in]		bmControlSize	bmControls フィールドのバイト数．
 */
void
PdjUsbMixerUnitDictionary::initControlsArray(
	UInt8			*bmCntrls,
	UInt8			bmControlSize
)
{
	OSArray			*controlsArray	= NULL;
	OSNumber		*arrayNumber	= NULL;

	FailIf(NULL == (controlsArray = OSArray::withCapacity(bmControlSize)), Exit);

	for (UInt8 controlNumber = 0; controlNumber < bmControlSize; controlNumber++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(bmCntrls[controlNumber], SIZEINBITS(UInt8))), Exit);
		FailIf(!(controlsArray->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kControlsArray, controlsArray), Exit);
	controlsArray = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != controlsArray)
	{
		controlsArray->release();
		controlsArray = NULL;
	}
}

/****************************************************************************/
/**
 * @brief	ミキサーユニットディスクリプタの baSourceID フィールドを設定．
 *
 * @param[in]		*baSrcIDs		ディスクリプタの baSourceID フィールド．
 * @param[in]		nrInPins		ディスクリプタの bNrInPins フィールド．
 */
void
PdjUsbMixerUnitDictionary::initSourceIDs(
	UInt8			*baSrcIDs,
	UInt8			nrInPins
)
{
	OSArray			*sourceIDs	 = NULL;
	OSNumber		*arrayNumber = NULL;

	FailIf(NULL == (sourceIDs = OSArray::withCapacity(nrInPins)), Exit);

	for (UInt8 sourceIndex = 0; sourceIndex < nrInPins; sourceIndex++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(baSrcIDs[sourceIndex], SIZEINBITS(UInt8))), Exit);
		FailIf(!(sourceIDs->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kSourceIDs, sourceIDs), Exit);
	sourceIDs = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != sourceIDs)
	{
		sourceIDs->release();
		sourceIDs = NULL;
	}
}

/*------------------------------------------*/
/*	PdjUsbSelectorUnitDictionary クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbSelectorUnitDictionary -

/****************************************************************************/
/**
 * @brief	セレクタユニットディスクリプタの baSourceID フィールドを設定．
 *
 * @param[in]		*baSrcIDs		ディスクリプタの baSourceID フィールド．
 * @param[in]		nrInPins		ディスクリプタの bNrInPins フィールド．
 */
void
PdjUsbSelectorUnitDictionary::initSourceIDs(
	UInt8			*baSrcIDs,
	UInt8			nrInPins
)
{
	OSArray			*sourceIDs	 = NULL;
	OSNumber		*arrayNumber = NULL;

	FailIf(NULL == (sourceIDs = OSArray::withCapacity(nrInPins)), Exit);

	for (UInt8 sourceIndex = 0; sourceIndex < nrInPins; sourceIndex++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(baSrcIDs[sourceIndex], SIZEINBITS(UInt8))), Exit);
		FailIf(!(sourceIDs->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kSourceIDs, sourceIDs), Exit);
	sourceIDs = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != sourceIDs)
	{
		sourceIDs->release();
		sourceIDs = NULL;
	}
}

/*------------------------------------------*/
/*		PdjUsbEffectUnitDictionary クラス	*/
/*------------------------------------------*/

#pragma mark - class PdjUsbEffectUnitDictionary -

/****************************************************************************/
/**
 * @brief	エフェクトユニットディスクリプタの bmaControls フィールドを設定．
 *
 * @param[in]		*bmaControlsArray	ディスクリプタの bmaControls フィールド．
 * @param[in]		numCntrls			エフェクトユニットのチャンネル数．
 */
void
PdjUsbEffectUnitDictionary::initControlsArray(
	UInt8			*bmaControlsArray,
	UInt8			numCntrls
)
{
	DP(DBG_TRACE, "+ " LOGSTREFFECTUNITDICTIONARY "[%p]::initControlsArray(%p, %d)", this, bmaControlsArray, numCntrls);

	OSArray			*controlsArray		 = NULL;
	OSNumber		*arrayNumber		 = NULL;
	UInt32			*bmaControlsArray_32 = NULL;
	UInt8			controlSize;

	FailIf(kIOReturnSuccess != super::super::setDictionaryValue(kNumControls, numCntrls), Exit);
	FailIf(kIOReturnSuccess != this->getControlSize(&controlSize), Exit);
	FailIf(NULL == (controlsArray = OSArray::withCapacity(numCntrls)), Exit);

	bmaControlsArray_32 = (UInt32 *)bmaControlsArray;

	for (UInt8 controlIndex = 0; controlIndex < numCntrls; controlIndex++)
	{
		if (1 == controlSize)
		{
			FailIf(NULL == (arrayNumber = OSNumber::withNumber(bmaControlsArray[controlIndex], SIZEINBITS(UInt8))), Exit);
		}
		else
		{
			FailIf(
						NULL == (
									arrayNumber = OSNumber::withNumber(
																USBToHostWord(bmaControlsArray_32[controlIndex]),
																controlSize * SIZEINBITS(UInt8)
															)
								),
						Exit
				  );
		}

		FailIf(!(controlsArray->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kControlsArray, controlsArray), Exit);
	controlsArray = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != controlsArray)
	{
		controlsArray->release();
		controlsArray = NULL;
	}

	DP(DBG_TRACE, "- " LOGSTREFFECTUNITDICTIONARY "[%p]::initControlsArray(%p, %d)", this, bmaControlsArray, numCntrls);
}

/*------------------------------------------*/
/*	PdjUsbProcessingUnitDictionary クラス	*/
/*------------------------------------------*/

#pragma mark - class PdjUsbProcessingUnitDictionary -

/****************************************************************************/
/**
 * @brief	プロセッシングユニットディスクリプタの bmControls フィールドを設定．
 *
 * @param[in]		*bmCntrls		ディスクリプタの bmControls フィールド．
 * @param[in]		bmControlSize	ディスクリプタの bControlSize フィールド．
 */
void
PdjUsbProcessingUnitDictionary::initControlsArray(
	UInt8			*bmCntrls,
	UInt8			bmControlSize
)
{
	OSArray			*controlsArray	= NULL;
	OSNumber		*arrayNumber	= NULL;

	FailIf(NULL == (controlsArray = OSArray::withCapacity(bmControlSize)), Exit);

	for (UInt8 controlNumber = 0; controlNumber < bmControlSize; controlNumber++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(bmCntrls[controlNumber], SIZEINBITS(UInt8))), Exit);
		FailIf(!(controlsArray->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kControlsArray, controlsArray), Exit);
	controlsArray = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != controlsArray)
	{
		controlsArray->release();
		controlsArray = NULL;
	}
}

/****************************************************************************/
/**
 * @brief	プロセッシングユニットディスクリプタの baSourceID フィールドを設定．
 *
 * @param[in]		*baSrcIDs		ディスクリプタの baSourceID フィールド．
 * @param[in]		nrInPins		ディスクリプタの bNrInPins フィールド．
 */
void
PdjUsbProcessingUnitDictionary::initSourceIDs(
	UInt8			*baSrcIDs,
	UInt8			nrInPins
)
{
	OSArray			*sourceIDs	 = NULL;
	OSNumber		*arrayNumber = NULL;

	FailIf(NULL == (sourceIDs = OSArray::withCapacity(nrInPins)), Exit);

	for (UInt8 sourceIndex = 0; sourceIndex < nrInPins; sourceIndex++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(baSrcIDs[sourceIndex], SIZEINBITS(UInt8))), Exit);
		FailIf(!(sourceIDs->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kSourceIDs, sourceIDs), Exit);
	sourceIDs = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != sourceIDs)
	{
		sourceIDs->release();
		sourceIDs = NULL;
	}
}

/*------------------------------------------*/
/*	PdjUsbExtensionUnitDictionary クラス	*/
/*------------------------------------------*/

#pragma mark - class PdjUsbExtensionUnitDictionary -

/****************************************************************************/
/**
 * @brief	エクステンションユニットディスクリプタの bmControls フィールドを設定．
 *
 * @param[in]		*bmCntrls		ディスクリプタの bmControls フィールド．
 * @param[in]		bmControlSize	ディスクリプタの bControlSize フィールド．
 */
void
PdjUsbExtensionUnitDictionary::initControlsArray(
	UInt8			*bmCntrls,
	UInt8			bmControlSize
)
{
	OSArray			*controlsArray	= NULL;
	OSNumber		*arrayNumber	= NULL;

	FailIf(NULL == (controlsArray = OSArray::withCapacity(bmControlSize)), Exit);

	for (UInt8 controlNumber = 0; controlNumber < bmControlSize; controlNumber++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(bmCntrls[controlNumber], SIZEINBITS(UInt8))), Exit);
		FailIf(!(controlsArray->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kControlsArray, controlsArray), Exit);
	controlsArray = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != controlsArray)
	{
		controlsArray->release();
		controlsArray = NULL;
	}
}

/****************************************************************************/
/**
 * @brief	エクステンションユニットディスクリプタの baSourceID フィールドを設定．
 *
 * @param[in]		*baSrcIDs		ディスクリプタの baSourceID フィールド．
 * @param[in]		nrInPins		ディスクリプタの bNrInPins フィールド．
 */
void
PdjUsbExtensionUnitDictionary::initSourceIDs(
	UInt8			*baSrcIDs,
	UInt8			nrInPins
)
{
	OSArray			*sourceIDs = NULL;
	OSNumber		*arrayNumber = NULL;

	FailIf(NULL == (sourceIDs = OSArray::withCapacity(nrInPins)), Exit);

	for (UInt8 sourceIndex = 0; sourceIndex < nrInPins; sourceIndex++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(baSrcIDs[sourceIndex], SIZEINBITS(UInt8))), Exit);
		FailIf(!(sourceIDs->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kSourceIDs, sourceIDs), Exit);
	sourceIDs = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != sourceIDs)
	{
		sourceIDs->release();
		sourceIDs = NULL;
	}
}

/*------------------------------------------*/
/*	PdjUsbFeatureUnitDictionary クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbFeatureUnitDictionary -

/****************************************************************************/
/**
 * @brief	フィーチャーユニットディスクリプタの bmaControls フィールドを設定．
 *
 * @param[in]		*bmaControlsArray	ディスクリプタの bmaControls フィールド．
 * @param[in]		numCntrls			フィーチャーユニットのチャンネル数．
 */
void
PdjUsbFeatureUnitDictionary::initControlsArray(
	UInt8			*bmaControlsArray,
	UInt8			numCntrls
)
{
	DP(DBG_TRACE, "+ " LOGSTRFEATUREUNITDICTIONARY "[%p]::initControlsArray(%p, %d)", this, bmaControlsArray, numCntrls);

	OSArray			*controlsArray		 = NULL;
	OSNumber		*arrayNumber		 = NULL;
	UInt16			*bmaControlsArray_16 = NULL;
	UInt32			*bmaControlsArray_32 = NULL;
	UInt8			controlSize;

	FailIf(kIOReturnSuccess != super::super::setDictionaryValue(kNumControls, numCntrls), Exit);
	FailIf(kIOReturnSuccess != this->getControlSize(&controlSize), Exit);
	FailIf(NULL == (controlsArray = OSArray::withCapacity(numCntrls)), Exit);

	bmaControlsArray_16 = (UInt16 *)bmaControlsArray;
	bmaControlsArray_32 = (UInt32 *)bmaControlsArray;

	for (UInt8 controlIndex = 0; controlIndex < numCntrls; controlIndex++)
	{
		if (1 == controlSize)
		{
			FailIf(
						NULL == (
									arrayNumber = OSNumber::withNumber(
																bmaControlsArray[controlIndex],
																SIZEINBITS(UInt8)
															)
								),
						Exit
				  );
		}
		else if (2 == controlSize)
		{
			FailIf(
						NULL == (
									arrayNumber = OSNumber::withNumber(
																USBToHostWord(bmaControlsArray_16[controlIndex]),
																controlSize * SIZEINBITS(UInt8)
															)
								),
						Exit
				  );
		}
		else
		{
			FailIf(
						NULL == (
									arrayNumber = OSNumber::withNumber(
																USBToHostLong(bmaControlsArray_32[controlIndex]),
																controlSize * SIZEINBITS(UInt8)
															)
								),
						Exit
				  );
		}

		FailIf(!(controlsArray->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kControlsArray, controlsArray), Exit);
	controlsArray = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != controlsArray)
	{
		controlsArray->release();
		controlsArray = NULL;
	}

	DP(DBG_TRACE, "- " LOGSTRFEATUREUNITDICTIONARY "[%p]::initControlsArray(%p, %d)", this, bmaControlsArray, numCntrls);
}

/****************************************************************************/
/**
 * @brief	チャンネルがミュート設定に対応しているかどうかを調べる．
 *			Feature Unit ディスクリプタの bmaControls フィールドで
 *			Mute Control がサポートされているかどうか判定する．
 *
 * @param[in]		channelNum		チャンネル番号．
 *
 * @retval			true			ミュート設定に対応している．
 * @retval			false			ミュート設定に対応していない．
 */
// Channel #1 is left channel, #2 is right channel
bool
PdjUsbFeatureUnitDictionary::channelHasMuteControl(
	UInt8			channelNum
)
{
	bool			result = false;
	OSArray			*controls		= NULL;
	OSObject		*arrayValue		= NULL;
	OSNumber		*controlNumber	= NULL;
	UInt8			numControls;
	UInt8			controlSize;

	FailIf(NULL == (controls = this->getControls()), Exit);
	FailIf(kIOReturnSuccess != this->getNumControls(&numControls), Exit);

	if (numControls >= channelNum + 1)
	{
		FailIf(kIOReturnSuccess != this->getControlSize(&controlSize), Exit);
		FailIf(NULL == (arrayValue = controls->getObject(channelNum)), Exit);
		FailIf(NULL == (controlNumber = OSDynamicCast(OSNumber, arrayValue)), Exit);

		if (1 == controlSize)
		{
			result = controlNumber->unsigned8BitValue() & (1 << kMuteBit);
		}
		else if (2 == controlSize)
		{
			result = controlNumber->unsigned16BitValue() & (1 << kMuteBit);
		}
		else // if (4 == controlSize)
		{
			// must be present & programmable
			result = (controlNumber->unsigned32BitValue() & (0x3 << (kMuteBit * 2))) == (0x3 << (kMuteBit * 2));
		}
	}

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	チャンネルがボリューム設定に対応しているかどうかを調べる．
 *			Feature Unit ディスクリプタの bmaControls フィールドで
 *			Volume Control がサポートされているかどうか判定する．
 *
 * @param[in]		channelNum		チャンネル番号．
 *
 * @retval			true			ボリューム設定に対応している．
 * @retval			false			ボリューム設定に対応していない．
 */
bool
PdjUsbFeatureUnitDictionary::channelHasVolumeControl(
	UInt8			channelNum
)
{
	bool			result = false;
	OSArray			*controls		= NULL;
	OSObject		*arrayValue		= NULL;
	OSNumber		*controlNumber	= NULL;
	UInt8			numControls;
	UInt8			controlSize;

	FailIf(NULL == (controls = this->getControls()), Exit);
	FailIf(kIOReturnSuccess != this->getNumControls(&numControls), Exit);

	if (numControls >= channelNum + 1)
	{
		FailIf(kIOReturnSuccess != this->getControlSize(&controlSize), Exit);
		FailIf(NULL == (arrayValue = controls->getObject(channelNum)), Exit);
		FailIf(NULL == (controlNumber = OSDynamicCast(OSNumber, arrayValue)), Exit);

		if (1 == controlSize)
		{
			result = controlNumber->unsigned8BitValue() & (1 << kVolumeBit);
		}
		else if (2 == controlSize)
		{
			result = controlNumber->unsigned16BitValue() & (1 << kVolumeBit);
		}
		else // if (4 == controlSize)
		{
			// must be present & programmable
			result = (controlNumber->unsigned32BitValue() & (0x3 << (kVolumeBit * 2))) == (0x3 << (kVolumeBit * 2));
		}
	}

Exit:
	DP(DBG_INFO, "? " LOGSTRFEATUREUNITDICTIONARY "::channelHasVolumeControl(%d) = %s", channelNum, (result ? "true" : "false"));
	return result;
}

/*------------------------------------------*/
/*	PdjUsbClockSourceDictionary クラス		*/
/*------------------------------------------*/

#pragma mark - class PdjUsbClockSourceDictionary -

/****************************************************************************/
/**
 * @brief	クロックソースディスクリプタの bmControls フィールドを設定．
 *
 * @param[in]		*bmCntrls		ディスクリプタの bmControls フィールド．
 * @param[in]		bmControlSize	bmControls フィールドのバイト数．常に 1 である．
 */
void
PdjUsbClockSourceDictionary::initControlsArray(
	UInt8			*bmCntrls,
	UInt8			bmControlSize
)
{
	OSArray			*controlsArray	= NULL;
	OSNumber		*arrayNumber	= NULL;

	FailIf(NULL == (controlsArray = OSArray::withCapacity(bmControlSize)), Exit);

	for (UInt8 controlNumber = 0; controlNumber < bmControlSize; controlNumber++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(bmCntrls[controlNumber], SIZEINBITS(UInt8))), Exit);
		FailIf(!(controlsArray->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kControlsArray, controlsArray), Exit);
	controlsArray = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != controlsArray)
	{
		controlsArray->release();
		controlsArray = NULL;
	}
}

/****************************************************************************/
/**
 * @brief	クロックタイプを取得．
 *
 * @param[out]		*clockType		ディスクリプタの bmAttributes フィールドで指定されたクロックタイプ．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbClockSourceDictionary::getClockType(
	UInt8			*clockType
)
{
	IOReturn		result = kIOReturnError;
	UInt8			attributes;

	FailIf(NULL == clockType, Exit);
	FailIf(kIOReturnSuccess != (result = this->getAttributes(&attributes)), Exit);

	*clockType = attributes & 0x3;

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	ディスクリプタの bmControls フィールドを取得．
 *
 * @param[out]		*bmControls		ディスクリプタの bmControls フィールド．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbClockSourceDictionary::getControlBitmap(
	UInt8			*bmControls
)
{
	IOReturn		result = kIOReturnError;
	OSArray			*controls		= NULL;
	OSNumber		*bitmapNumber	= NULL;

	FailIf(NULL == bmControls, Exit);
	FailIf(NULL == (controls = this->getControls()), Exit);
	FailIf(NULL == (bitmapNumber = OSDynamicCast(OSNumber, controls->getObject(0))), Exit);

	*bmControls = bitmapNumber->unsigned8BitValue();

	result = kIOReturnSuccess;

Exit:
	return result;
}

/*------------------------------------------*/
/*	PdjUsbClockSelectorDictionary クラス	*/
/*------------------------------------------*/

#pragma mark - class PdjUsbClockSelectorDictionary -

/****************************************************************************/
/**
 * @brief	クロックセレクタディスクリプタの bmControls フィールドを設定．
 *
 * @param[in]		*bmCntrls		ディスクリプタの bmControls フィールド．
 * @param[in]		bmControlSize	bmControls フィールドのバイト数．常に 1 である．
 */
void
PdjUsbClockSelectorDictionary::initControlsArray(
	UInt8			*bmCntrls,
	UInt8			bmControlSize
)
{
	OSArray			*controlsArray	= NULL;
	OSNumber		*arrayNumber	= NULL;

	FailIf(NULL == (controlsArray = OSArray::withCapacity(bmControlSize)), Exit);

	for (UInt8 controlNumber = 0; controlNumber < bmControlSize; controlNumber++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(bmCntrls[controlNumber], SIZEINBITS(UInt8))), Exit);
		FailIf(!(controlsArray->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kControlsArray, controlsArray), Exit);
	controlsArray = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != controlsArray)
	{
		controlsArray->release();
		controlsArray = NULL;
	}
}

/****************************************************************************/
/**
 * @brief	クロックセレクタディスクリプタの baCSourceID フィールドを設定．
 *
 * @param[in]		*baCSrcIDs		ディスクリプタの baCSourceID フィールド．
 * @param[in]		nrInPins		ディスクリプタの bNrInPins フィールド．
 */
void
PdjUsbClockSelectorDictionary::initClockSourceIDs(
	UInt8			*baCSrcIDs,
	UInt8			nrInPins
)
{
	OSArray			*cSourceIDs	 = NULL;
	OSNumber		*arrayNumber = NULL;

	FailIf(NULL == (cSourceIDs = OSArray::withCapacity(nrInPins)), Exit);

	for (UInt8 cSourceIndex = 0; cSourceIndex < nrInPins; cSourceIndex++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(baCSrcIDs[cSourceIndex], SIZEINBITS(UInt8))), Exit);
		FailIf(!(cSourceIDs->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kCSourceIDs, cSourceIDs), Exit);
	cSourceIDs = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != cSourceIDs)
	{
		cSourceIDs->release();
		cSourceIDs = NULL;
	}
}

/*------------------------------------------*/
/*	PdjUsbClockMultiplierDictionary クラス	*/
/*------------------------------------------*/

#pragma mark - class PdjUsbClockMultiplierDictionary -

/****************************************************************************/
/**
 * @brief	クロック逓倍／分周器ディスクリプタの bmControls フィールドを設定．
 *
 * @param[in]		*bmCntrls		ディスクリプタの bmControls フィールド．
 * @param[in]		bmControlSize	bmControls フィールドのバイト数．常に 1 である．
 */
void
PdjUsbClockMultiplierDictionary::initControlsArray(
	UInt8			*bmCntrls,
	UInt8			bmControlSize
)
{
	OSArray			*controlsArray	= NULL;
	OSNumber		*arrayNumber	= NULL;

	FailIf(NULL == (controlsArray = OSArray::withCapacity(bmControlSize)), Exit);

	for (UInt8 controlNumber = 0; controlNumber < bmControlSize; controlNumber++)
	{
		FailIf(NULL == (arrayNumber = OSNumber::withNumber(bmCntrls[controlNumber], SIZEINBITS(UInt8))), Exit);
		FailIf(!(controlsArray->setObject(arrayNumber)), Exit);

		arrayNumber->release();
		arrayNumber = NULL;
	}

	FailIf(kIOReturnSuccess != super::super::setDictionaryObjectAndRelease(kControlsArray, controlsArray), Exit);
	controlsArray = NULL;

Exit:
	if (NULL != arrayNumber)
	{
		arrayNumber->release();
		arrayNumber = NULL;
	}

	if (NULL != controlsArray)
	{
		controlsArray->release();
		controlsArray = NULL;
	}
}
