/*
 * Copyright (c) 1998-2009 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

/************************************************************************/
/**
 * PDJ USB Audio Driver for Mac OS X
 *
 * @file	PdjUsbAudioTopology.cpp
 *
 * @brief	OSObject クラスの派生クラス．
 *			USB Audio Class の Audio Function Topology に関わる処理．
 *			AppleUSBAudioDevice クラスを２つに分離して
 *			PdjUsbAudioDevice クラスおよび PdjUsbAudioTopology クラスとした．
 *
 * This code is based on AppleUSBAudio-273.4.1 for Mac OS 10.6.4.
 */
/************************************************************************/

/*------------------------------------------*/
/*				インクルード				*/
/*------------------------------------------*/

#include "PdjUsbAudioCommon.h"

/*------------------------------------------*/
/*			PdjUsbAudioTopology クラス		*/
/*------------------------------------------*/

OSDefineMetaClassAndStructors(PdjUsbAudioTopology, super)

#pragma mark - Public Methods -

/****************************************************************************/
/**
 * @brief	インスタンス生成．
 *			PdjUsbAudioDevice::protectedInitHardware() から呼び出される．
 *
 * @param[in]		*audioDevice	PdjUsbAudioDevice クラスのインスタンス．
 *
 * @return			生成された PdjUsbAudioTopology のインスタンス．
 */
PdjUsbAudioTopology *
PdjUsbAudioTopology::create(
	PdjUsbAudioDevice		*audioDevice
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::create()");

	PdjUsbAudioTopology		*topology = NULL;

	FailIf(NULL == audioDevice, Exit);

	topology = new PdjUsbAudioTopology;
	FailIf(NULL == topology, Exit);

	if (!(topology->init(audioDevice)))
	{
		topology->release();
		topology = NULL;
	}

	FailIf(NULL == topology, Exit);

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::create() = %p", topology);
	return topology;
}

/****************************************************************************/
/**
 * @brief	インスタンス初期化．
 *			OSObject::init() とはシグネチャが異なる．
 *
 * @param[in]		*audioDevice	PdjUsbAudioDevice クラスのインスタンス．
 *
 * @retval			true			初期化成功．
 * @retval			false			初期化失敗．
 */
bool
PdjUsbAudioTopology::init(
	PdjUsbAudioDevice		*audioDevice
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::init()", this);

	bool						result = false;

	FailIf(!super::init(), Exit);

	FailIf(NULL == audioDevice, Exit);
	mAudioDevice = audioDevice;

	mConfigDictionary = mAudioDevice->getConfigDictionary();
	FailIf(NULL == mConfigDictionary, Exit);
	mConfigDictionary->retain();

	mControlInterface = mAudioDevice->mControlInterface;
	FailIf(NULL == mControlInterface, Exit);
	mControlInterface->retain();

	result = true;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::init() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	インスタンス解放．
 *			OSObject::free() をオーバーライド．
 *			参照カウントがゼロになった時呼び出される．
 */
void
PdjUsbAudioTopology::free(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::free()", this);

	if (mMonoControlsArray)
	{
		mMonoControlsArray->release();
		mMonoControlsArray = NULL;
	}

	if (mClockGraph)
	{
		mClockGraph->release();
		mClockGraph = NULL;
	}

	if (mControlGraph)
	{
		mControlGraph->release();
		mControlGraph = NULL;
	}

	if (mControlInterface)
	{
		mControlInterface->release();
		mControlInterface = NULL;
	}

	if (mConfigDictionary)
	{
		mConfigDictionary->release();
		mConfigDictionary = NULL;
	}

	super::free();

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::free()", this);
}

/****************************************************************************/
/**
 * @brief	オーディオファンクションの接続グラフを構築．
 *			PdjUsbAudioDevice::protectedInitHardware() から呼び出される．
 *
 * @param[in]		controlInterfaceNum		USB interface 番号．
 *
 * @return			オーディオファンクションの接続グラフ．
 */
OSArray *
PdjUsbAudioTopology::buildConnectionGraph(
	UInt8			controlInterfaceNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::buildConnectionGraph(%d)", this, controlInterfaceNum);

	OSArray			*allOutputTerminalPaths = NULL;
	OSArray			*pathsFromOutputTerminalN = NULL;
#ifdef TARGET_PDJ
#else
	OSArray			*thisPath = NULL;
	UInt8			numTerminals;
	UInt8			terminalID;
#endif

	allOutputTerminalPaths = OSArray::withCapacity(1);
	FailIf(NULL == allOutputTerminalPaths, Exit);

	pathsFromOutputTerminalN = OSArray::withCapacity(1);
	FailIf(NULL == pathsFromOutputTerminalN, Exit);

#ifdef TARGET_PDJ
#else
	FailIf(kIOReturnSuccess != mConfigDictionary->getNumOutputTerminals(&numTerminals, controlInterfaceNum, 0), Exit);

	for (UInt8 terminalIndex = 0; terminalIndex < numTerminals; terminalIndex++)
	{
		FailIf(kIOReturnSuccess != mConfigDictionary->getIndexedOutputTerminalID(&terminalID, controlInterfaceNum, 0, terminalIndex), Exit);

#ifdef BUG_FIX
		OSArray		*path = this->buildPath(controlInterfaceNum, terminalID, pathsFromOutputTerminalN, thisPath);

		if (path)
		{
			path->release();
			path = NULL;
		}
#else
		this->buildPath(controlInterfaceNum, terminalID, pathsFromOutputTerminalN, thisPath);
#endif

		allOutputTerminalPaths->setObject(pathsFromOutputTerminalN);
		pathsFromOutputTerminalN->release();

		pathsFromOutputTerminalN = OSArray::withCapacity(1);
		FailIf(NULL == pathsFromOutputTerminalN, Exit);
	}
#endif

Exit:
	if (NULL != pathsFromOutputTerminalN)
	{
		pathsFromOutputTerminalN->release();
	}

	mControlGraph = allOutputTerminalPaths;

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::buildConnectionGraph(%d) = %p", this, controlInterfaceNum, mControlGraph);
	return mControlGraph;
}

/****************************************************************************/
/**
 * @brief	クロックグラフを構築．
 *			PdjUsbAudioDevice::protectedInitHardware() から呼び出される．
 *
 * @param[in]		controlInterfaceNum		USB interface 番号．
 *
 * @return			クロックグラフ．
 */
OSArray *
PdjUsbAudioTopology::buildClockGraph(
	UInt8			controlInterfaceNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::buildClockGraph(%d)", this, controlInterfaceNum);

	OSArray			*allClockPaths = NULL;
	OSArray			*terminalClockEntities = NULL;
	OSArray			*pathsFromClockEntityN = NULL;
	OSArray			*thisPath = NULL;
	OSNumber		*clockIDNum = NULL;

	allClockPaths = OSArray::withCapacity(1);
	FailIf(NULL == allClockPaths, Exit);

	pathsFromClockEntityN = OSArray::withCapacity(1);
	FailIf(NULL == pathsFromClockEntityN, Exit);

	FailIf(NULL == (terminalClockEntities = mConfigDictionary->getTerminalClockEntities(controlInterfaceNum, 0)), Exit);

	for (UInt8 clockIndex = 0; clockIndex < terminalClockEntities->getCount(); clockIndex++)
	{
		FailIf(NULL == (clockIDNum = OSDynamicCast(OSNumber, terminalClockEntities->getObject(clockIndex))), Exit);

		UInt8	clockID = clockIDNum->unsigned8BitValue();

		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::buildClockGraph() - Building clock paths from ID %d", this, clockID);
		this->buildClockPath(controlInterfaceNum, clockID, pathsFromClockEntityN, thisPath);

		allClockPaths->setObject(pathsFromClockEntityN);
		pathsFromClockEntityN->release();

		pathsFromClockEntityN = OSArray::withCapacity(1);
		FailIf(NULL == pathsFromClockEntityN, Exit);
	}

	// Print clock graph of interest

	DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::buildClockGraph(%d) - Displaying graph (%d) ...", this, controlInterfaceNum, allClockPaths->getCount());

	for (UInt8 groupIndex = 0; (allClockPaths && (groupIndex < allClockPaths->getCount())); groupIndex++)
	{
		DP(DBG_INFO, "   Path Group # %d", groupIndex);

		OSArray		*thisGroup = NULL;
		FailIf(NULL == (thisGroup = OSDynamicCast(OSArray, allClockPaths->getObject(groupIndex))), Exit);

		for (UInt8 pathIndex = 0; pathIndex < thisGroup->getCount(); pathIndex++)
		{
			char	pathLine[256];
			char	tempString[10];

			*pathLine = '\0';

			::snprintf(tempString, sizeof(tempString), "%2d: ", pathIndex);
			::strlcat(pathLine, tempString, sizeof(pathLine));

			FailIf(NULL == (thisPath = OSDynamicCast(OSArray, thisGroup->getObject(pathIndex))), Exit);

			for (UInt8 pathItem = 0; pathItem < thisPath->getCount(); pathItem++)
			{
				FailIf(NULL == (clockIDNum = (OSNumber *)thisPath->getObject(pathItem)), Exit);

				::snprintf(tempString, sizeof(tempString), "%d ", clockIDNum->unsigned8BitValue());
				::strlcat(pathLine, tempString, sizeof(pathLine));
			}

			DP(DBG_INFO, "  %s", pathLine);
		}
	}

Exit:
	if (NULL != pathsFromClockEntityN)
	{
		pathsFromClockEntityN->release();
	}

#ifdef BUG_FIX
	if (NULL != terminalClockEntities)
	{
		terminalClockEntities->release();
	}
#endif

	mClockGraph = allClockPaths;

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::buildClockGraph(%d) = %p", this, controlInterfaceNum, mClockGraph);
	return mClockGraph;
}

/****************************************************************************/
/**
 * @brief	最適なクロックパスを検索．
 *
 * @param[in]		*thisEngine							PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		streamInterface						AudioStreaming Interface の interface 番号．
 * @param[in]		altSetting							alternate setting 番号．
 * @param[in]		sampleRate							サンプリング周波数．
 * @param[out]		*otherEngineNeedSampleRateChange	このパスが他のオーディオエンジンで使用中の場合、true が返される．
 * @param[out]		*clockPathGroupIndex				最適なクロックパスを含むクロックパスグループのインデックス．
 *
 * @return			最適なクロックパス．
 */
// [rdar://4867843]
// <rdar://5811247>
OSArray *
PdjUsbAudioTopology::getOptimalClockPath(
	PdjUsbAudioEngine		*thisEngine,
	UInt8					streamInterface,
	UInt8					altSetting,
	UInt32					sampleRate,
	bool					*otherEngineNeedSampleRateChange,
	UInt8					*clockPathGroupIndex
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getOptimalClockPath(%u, %u, %u)",
		this, streamInterface, altSetting, sampleRate);

	UInt8					terminalID;
	UInt8					clockSourceID;
	UInt32					clockPathUnitUsage	= 0xFFFFFFFF;
	OSArray					*pathGroupArray		= NULL;
	OSArray					*pathArray			= NULL;
	OSArray					*optimalPathArray	= NULL;

	FailIf(NULL == mConfigDictionary, Exit);
	FailIf(NULL == mControlInterface, Exit);

	FailIf(
			kIOReturnSuccess != mConfigDictionary->getTerminalLink(
														&terminalID,
														streamInterface,
														altSetting
												   ),
			Exit
	);

	FailIf(
			kIOReturnSuccess != mConfigDictionary->getClockSourceID(
														&clockSourceID,
														mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
														0,
														terminalID
												   ),
			Exit
	);

	DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::getOptimalClockPath() - interface %d, alt setting %d, clock source ID %d",
		streamInterface, altSetting, clockSourceID);

	// Find the path group that begins with the clockSourceID.
	for (UInt8 pathGroupIndex = 0; pathGroupIndex < mClockGraph->getCount(); pathGroupIndex++)
	{
		OSNumber		*clockSourceIDNumber = NULL;

		FailIf(NULL == (pathGroupArray		= OSDynamicCast(OSArray,  mClockGraph->getObject(pathGroupIndex))),	Exit);
		FailIf(NULL == (pathArray			= OSDynamicCast(OSArray,  pathGroupArray->getObject(0))),			Exit);
		FailIf(NULL == (clockSourceIDNumber	= OSDynamicCast(OSNumber, pathArray->getObject(0))),				Exit);

		if (clockSourceID == clockSourceIDNumber->unsigned8BitValue())
		{
			// We have found the path group in which we are interested.
			// <rdar://5811247>
			if (NULL != clockPathGroupIndex)
			{
				*clockPathGroupIndex = pathGroupIndex;
			}

			break;
		}
		else
		{
			pathGroupArray = NULL;
		}
	}

	FailIf(NULL == pathGroupArray, Exit);

	// For each path in the path group, determine if it supported the requested sample rate.
	for (UInt8 pathIndex = 0; pathIndex < pathGroupArray->getCount(); pathIndex++)
	{
		FailIf(NULL == (pathArray = OSDynamicCast(OSArray, pathGroupArray->getObject(pathIndex))), Exit);

		if (this->supportSampleRateInClockPath(pathArray, sampleRate))
		{
			// Need to select the path with the clock units that are not used (or has the smallest usage count) by other paths
			// in other engines (if possible).
			UInt32		usageCount = this->determineClockPathUnitUsage(thisEngine, pathArray);

			if (usageCount < clockPathUnitUsage)
			{
				clockPathUnitUsage	= usageCount;
				optimalPathArray	= pathArray;
			}

			if (0 == clockPathUnitUsage)
			{
				// Found a path that is independent of the paths in other engines.
				break;
			}
		}
	}

Exit:
	if (NULL != otherEngineNeedSampleRateChange)
	{
		*otherEngineNeedSampleRateChange = (clockPathUnitUsage > 0);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getOptimalClockPath(%u, %u, %u) = %p",
		this, streamInterface, altSetting, sampleRate, optimalPathArray);
	return optimalPathArray;
}

/****************************************************************************/
/**
 * @brief	クロックパスグループを取得（interface 番号から）．
 *
 * @param[in]		streamInterface			AudioStreaming Interface の interface 番号．
 * @param[in]		altSetting				alternate setting 番号．
 * @param[out]		*clockPathGroupIndex	クロックパスグループのインデックス．
 *
 * @return			クロックパスグループ．
 */
// <rdar://5811247>
OSArray *
PdjUsbAudioTopology::getClockPathGroup(
	UInt8			streamInterface,
	UInt8			altSetting,
	UInt8			*clockPathGroupIndex
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getClockPathGroup(%d, %d)", this, streamInterface, altSetting);

	OSArray			*pathGroupArray = NULL;
	UInt8			terminalID;
	UInt8			clockSourceID;

	FailIf(NULL == mConfigDictionary, Exit);
	FailIf(NULL == mControlInterface, Exit);

	FailIf(kIOReturnSuccess != mConfigDictionary->getTerminalLink(&terminalID, streamInterface, altSetting), Exit);
	FailIf(kIOReturnSuccess != mConfigDictionary->getClockSourceID(&clockSourceID, mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, terminalID), Exit);

	DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::getClockPathGroup() - interface %d, alt setting %d, clock source ID %d",
		streamInterface, altSetting, clockSourceID);

	// Find the path group that begins with the clockSourceID.
	for (UInt8 pathGroupIndex = 0; pathGroupIndex < mClockGraph->getCount(); pathGroupIndex++)
	{
		OSNumber	*clockSourceIDNumber = NULL;
		OSArray		*pathArray = NULL;

		FailIf(NULL == (pathGroupArray = OSDynamicCast(OSArray, mClockGraph->getObject(pathGroupIndex))), Exit);
		FailIf(NULL == (pathArray = OSDynamicCast(OSArray, pathGroupArray->getObject(0))), Exit);
		FailIf(NULL == (clockSourceIDNumber = OSDynamicCast(OSNumber, pathArray->getObject(0))), Exit);

		if (clockSourceID == clockSourceIDNumber->unsigned8BitValue())
		{
			// We have found the path group in which we are interested.
			// <rdar://5811247>
			if (NULL != clockPathGroupIndex)
			{
				*clockPathGroupIndex = pathGroupIndex;
			}

			break;
		}
		else
		{
			pathGroupArray = NULL;
		}
	}

	FailIf(NULL == pathGroupArray, Exit);

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getClockPathGroup(%d, %d) = %p", this, streamInterface, altSetting, pathGroupArray);
	return pathGroupArray;
}

/****************************************************************************/
/**
 * @brief	クロックパスグループを取得（インデックスから）．
 *
 * @param[in]		pathGroupIndex		クロックパスグループのインデックス．
 *
 * @return			クロックパスグループ．
 */
// <rdar://5811247>
OSArray *
PdjUsbAudioTopology::getClockPathGroup(
	UInt8			pathGroupIndex
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getClockPathGroup(%d)", this, pathGroupIndex);

	OSArray			*pathGroupArray = NULL;

	// Find the path group that begins with the clockSourceID.
	FailIf(pathGroupIndex >= mClockGraph->getCount(), Exit);
	FailIf(NULL == (pathGroupArray = OSDynamicCast(OSArray, mClockGraph->getObject(pathGroupIndex))), Exit);

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getClockPathGroup(%d) = %p", this, pathGroupIndex, pathGroupArray);
	return pathGroupArray;
}

/****************************************************************************/
/**
 * @brief	共通のクロックを持つ AudioStreaming Interface を検索．
 *			PdjUsbAudioDevice::createAudioEngines() から呼び出される．
 *			１つのオーディオエンジンに属するオーディオストリームは、
 *				すべてクロックが共通でなければならない．
 *			クロックが異なる場合は別のオーディオエンジンが必要．
 *
 * @param[in, out]	*availableStreamList	検索対象となる AudioStreaming Interface の interface 番号の OSArray．
 *											検索条件に合致した要素は削除される．
 *
 * @return			共通のクロックを持つ AudioStreaming Interface の interface 番号の OSArray．
 */
// <rdar://5131786> Find streams that has common clocks, and return an array of the streams.
// The caller is responsible to release the returned stream array.
OSArray *
PdjUsbAudioTopology::findStreamsWithCommonClocks(
	OSArray			*availableStreamList
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithCommonClocks(%p)", this, availableStreamList);

	OSNumber		*streamInterfaceNumber;
	OSArray			*compatibleStreamList = NULL;
	OSArray			*inCompatibleStreamList = NULL;
	OSArray			*result = NULL;

	FailIf(NULL == availableStreamList, Exit);

	compatibleStreamList = OSArray::withCapacity(1);
	FailIf(NULL == compatibleStreamList, Exit);

	inCompatibleStreamList = OSArray::withCapacity(1);
	FailIf(NULL == inCompatibleStreamList, Exit);

	// Go thru all the available streams, and find all streams that have common sample rates and put them on
	// the compatible list. Otherwise, put them on the incompatible list.
	while (0 != availableStreamList->getCount())
	{
		FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, availableStreamList->getObject(0))), Exit);
		streamInterfaceNumber->retain();

		availableStreamList->removeObject(0);

		if (0 == compatibleStreamList->getCount())
		{
			// Determine if the stream has common clocks as one of the stream in the available list.
			if (this->isClockCommonWithAtLeastOneStreamsInList(streamInterfaceNumber, availableStreamList))
			{
				// Keep it in the compatible list.
				compatibleStreamList->setObject(streamInterfaceNumber);
			}
			else
			{
				// Keep it in the incompatible list.
				inCompatibleStreamList->setObject(streamInterfaceNumber);
			}
		}
		else
		{
			// Determine if the stream has common clocks with all the streams in the compatible list.
			if (this->isClockCommonWithAllStreamsInList(streamInterfaceNumber, compatibleStreamList))
			{
				// Keep it in the compatible list.
				compatibleStreamList->setObject(streamInterfaceNumber);
			}
			else
			{
				// Keep it in the incompatible list.
				inCompatibleStreamList->setObject(streamInterfaceNumber);
			}
		}

		streamInterfaceNumber->release();
	}

	// The available stream list should be empty here.
	FailIf(0 != availableStreamList->getCount(), Exit);

	if (0 != compatibleStreamList->getCount())
	{
		// There is some streams with common clocks in the compatible list.
		result = compatibleStreamList;
		result->retain();
	}
	else
	{
		// Nothing in the compatible stream list. Get one from the incompatible list.
		if (0 != inCompatibleStreamList->getCount())
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, inCompatibleStreamList->getObject(0))), Exit);
			streamInterfaceNumber->retain();

			inCompatibleStreamList->removeObject(0);

			compatibleStreamList->setObject(streamInterfaceNumber);
			result = compatibleStreamList;
			result->retain();

			streamInterfaceNumber->release();
		}
	}

#ifdef DEBUGLOGGING
	if (NULL != result)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithCommonClocks(%p) - Found streams:", this, availableStreamList);

		for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < result->getCount(); streamInterfaceIndex++)
		{
			streamInterfaceNumber = OSDynamicCast(OSNumber, result->getObject(streamInterfaceIndex));
			FailMessage(NULL == streamInterfaceNumber);

			if (NULL != streamInterfaceNumber)
			{
				DP(DBG_INFO, "--> #%u", streamInterfaceNumber->unsigned8BitValue());
			}
		}
	}
#endif

Exit:
	if (NULL != inCompatibleStreamList)
	{
		if (0 != inCompatibleStreamList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(inCompatibleStreamList);
		}
		inCompatibleStreamList->release();
	}

	if (NULL != compatibleStreamList)
	{
		// If it failed for some reason, put back the compatible streams into the available stream list.
		if ((NULL == result) && (0 != compatibleStreamList->getCount()))
		{
			availableStreamList->merge(compatibleStreamList);
		}
		compatibleStreamList->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithCommonClocks(%p) - result = %p", this, availableStreamList, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	対応サンプリング周波数が共通な AudioStreaming Interface を検索．
 *			PdjUsbAudioDevice::createAudioEngines() から呼び出される．
 *			１つのオーディオエンジンに属するオーディオストリームは、
 *				すべて対応サンプリング周波数が共通でなければならない．
 *			対応サンプリング周波数が共通でない場合は、
 *				別のオーディオエンジンが必要．
 *
 * @param[in, out]	*availableStreamList	検索対象となる AudioStreaming Interface の interface 番号の OSArray．
 *											検索条件に合致した要素は削除される．
 *
 * @return			対応サンプリング周波数が共通な AudioStreaming Interface の interface 番号の OSArray．
 */
// <rdar://5131786> Find streams that has common sample rates, and return an array of the streams.
// The caller is responsible to release the returned stream array.
OSArray *
PdjUsbAudioTopology::findStreamsWithCommonSampleRates(
	OSArray			*availableStreamList
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithCommonSampleRates(%p)", this, availableStreamList);

	OSNumber		*streamInterfaceNumber;
	OSArray			*compatibleStreamList = NULL;
	OSArray			*inCompatibleStreamList = NULL;
	OSArray			*result = NULL;

	FailIf(NULL == availableStreamList, Exit);

	compatibleStreamList = OSArray::withCapacity(1);
	FailIf(NULL == compatibleStreamList, Exit);

	inCompatibleStreamList = OSArray::withCapacity(1);
	FailIf(NULL == inCompatibleStreamList, Exit);

	// Go thru all the available streams, and find all streams that have common sample rates and put them on
	// the compatible list. Otherwise, put them on the incompatible list.
	while (0 != availableStreamList->getCount())
	{
		FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, availableStreamList->getObject(0))), Exit);
		streamInterfaceNumber->retain();

		availableStreamList->removeObject(0);

		if (0 == compatibleStreamList->getCount())
		{
			// Determine if the stream has common sample rates as one of the stream in the available list.
			if (this->isSampleRateCommonWithAtLeastOneStreamsInList(streamInterfaceNumber, availableStreamList))
			{
				// Keep it in the compatible list.
				compatibleStreamList->setObject(streamInterfaceNumber);
			}
			else
			{
				// Keep it in the incompatible list.
				inCompatibleStreamList->setObject(streamInterfaceNumber);
			}
		}
		else
		{
			// Determine if the stream has common sample rates with all the streams in the compatible list.
			if (this->isSampleRateCommonWithAllStreamsInList(streamInterfaceNumber, compatibleStreamList))
			{
				// Keep it in the compatible list.
				compatibleStreamList->setObject(streamInterfaceNumber);
			}
			else
			{
				// Keep it in the incompatible list.
				inCompatibleStreamList->setObject(streamInterfaceNumber);
			}
		}

		streamInterfaceNumber->release();
	}

	// The available stream list should be empty here.
	FailIf(0 != availableStreamList->getCount(), Exit);

	if (0 != compatibleStreamList->getCount())
	{
		// There is some streams with common sample rates in the compatible list.
		result = compatibleStreamList;
		result->retain();
	}
	else
	{
		// Nothing in the compatible stream list. Get one from the incompatible list.
		if (0 != inCompatibleStreamList->getCount())
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, inCompatibleStreamList->getObject(0))), Exit);
			streamInterfaceNumber->retain();

			inCompatibleStreamList->removeObject(0);

			compatibleStreamList->setObject(streamInterfaceNumber);
			result = compatibleStreamList;
			result->retain();

			streamInterfaceNumber->release();
		}
	}

#ifdef DEBUGLOGGING
	if (NULL != result)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithCommonSampleRates(%p) - Found streams:", this, availableStreamList);

		for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < result->getCount(); streamInterfaceIndex++)
		{
			streamInterfaceNumber = OSDynamicCast(OSNumber, result->getObject(streamInterfaceIndex));
			FailMessage(NULL == streamInterfaceNumber);

			if (NULL != streamInterfaceNumber)
			{
				DP(DBG_INFO, "--> #%u", streamInterfaceNumber->unsigned8BitValue());
			}
		}
	}
#endif

Exit:
	if (NULL != inCompatibleStreamList)
	{
		if (0 != inCompatibleStreamList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(inCompatibleStreamList);
		}
		inCompatibleStreamList->release();
	}

	if (NULL != compatibleStreamList)
	{
		// If it failed for some reason, put back the compatible streams into the available stream list.
		if ((NULL == result) && (0 != compatibleStreamList->getCount()))
		{
			availableStreamList->merge(compatibleStreamList);
		}
		compatibleStreamList->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithCommonSampleRates(%p) = %p", this, availableStreamList, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	共存可能なエンドポイントを持つ AudioStreaming Interface を検索．
 *			PdjUsbAudioDevice::createAudioEngines() から呼び出される．
 *			共存できないエンドポイントがある場合は、別のオーディオエンジンが必要．
 *
 * @param[in, out]	*availableStreamList	検索対象となる AudioStreaming Interface の interface 番号の OSArray．
 *											検索条件に合致した要素は削除される．
 *
 * @return			共存可能なエンドポイントを持つ AudioStreaming Interface の interface 番号の OSArray．
 */
// <rdar://5131786> Find streams that has compatible endpoints, and return an array of the streams.
// The caller is responsible to release the returned stream array.
OSArray *
PdjUsbAudioTopology::findStreamsWithCompatibleEndpoints(
	OSArray			*availableStreamList
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithCompatibleEndpoints(%p)", this, availableStreamList);

	OSArray			*inputAsynchronousList	= NULL;
	OSArray			*inputAdaptiveList		= NULL;
	OSArray			*inputSynchronousList	= NULL;
	OSArray			*inputUnknownList		= NULL;
	OSArray			*outputAsynchronousList	= NULL;
	OSArray			*outputAdaptiveList		= NULL;
	OSArray			*outputSynchronousList	= NULL;
	OSArray			*outputUnknownList		= NULL;
	OSArray			*compatibleStreamList	= NULL;
	OSArray			*result					= NULL;

	FailIf(NULL == availableStreamList, Exit);

	compatibleStreamList = OSArray::withCapacity(1);
	FailIf(NULL == compatibleStreamList, Exit);

	// Find the input & output streams for direction & sync type.
	inputAsynchronousList	= this->findStreamsWithDirectionAndSyncType(availableStreamList, kUSBIn,  kAsynchSyncType     );
	inputAdaptiveList		= this->findStreamsWithDirectionAndSyncType(availableStreamList, kUSBIn,  kAdaptiveSyncType   );
	inputSynchronousList	= this->findStreamsWithDirectionAndSyncType(availableStreamList, kUSBIn,  kSynchronousSyncType);
	inputUnknownList		= this->findStreamsWithDirectionAndSyncType(availableStreamList, kUSBIn,  kUnknownSyncType    );

	outputAsynchronousList	= this->findStreamsWithDirectionAndSyncType(availableStreamList, kUSBOut, kAsynchSyncType     );
	outputAdaptiveList		= this->findStreamsWithDirectionAndSyncType(availableStreamList, kUSBOut, kAdaptiveSyncType   );
	outputSynchronousList	= this->findStreamsWithDirectionAndSyncType(availableStreamList, kUSBOut, kSynchronousSyncType);
	outputUnknownList		= this->findStreamsWithDirectionAndSyncType(availableStreamList, kUSBOut, kUnknownSyncType    );

	// The available stream list should be empty here.
	FailIf(0 != availableStreamList->getCount(), Exit);

	// Check if there is an asynchronous input stream.
	if ((NULL != inputAsynchronousList) && (0 != inputAsynchronousList->getCount()))
	{
		// There is at least an asynchronous input stream. Put the streams on the compatible list.
		compatibleStreamList->merge(inputAsynchronousList);
		inputAsynchronousList->release();
		inputAsynchronousList = 0;

		// If there is at least an asynchronous output stream, put the streams on the compatible list.
		if ((NULL != outputAsynchronousList) && (0 != outputAsynchronousList->getCount()))
		{
			compatibleStreamList->merge(outputAsynchronousList);
			outputAsynchronousList->release();
			outputAsynchronousList = 0;
		}

		// If there is at least an adaptive output stream, put the streams on the compatible list.
		if ((NULL != outputAdaptiveList) && (0 != outputAdaptiveList->getCount()))
		{
			compatibleStreamList->merge(outputAdaptiveList);
			outputAdaptiveList->release();
			outputAdaptiveList = 0;
		}

		// If there is at least a synchronous output stream, put the streams on the compatible list.
		if ((NULL != outputSynchronousList) && (0 != outputSynchronousList->getCount()))
		{
			compatibleStreamList->merge(outputSynchronousList);
			outputSynchronousList->release();
			outputSynchronousList = 0;
		}
	}
	else
	{
		// Check if there is a synchronous or adaptive input stream.
		if ((NULL != inputSynchronousList) || (NULL != inputAdaptiveList))
		{
			// If there is at least a synchronous input stream, put the streams on the compatible list.
			if ((NULL != inputSynchronousList) && (0 != inputSynchronousList->getCount()))
			{
				compatibleStreamList->merge(inputSynchronousList);
				inputSynchronousList->release();
				inputSynchronousList = 0;
			}

			// If there is at least an adaptive input stream, put the streams on the compatible list.
			if ((NULL != inputAdaptiveList) && (0 != inputAdaptiveList->getCount()))
			{
				compatibleStreamList->merge(inputAdaptiveList);
				inputAdaptiveList->release();
				inputAdaptiveList = 0;
			}

			// If there is at least a synchronous output stream, put the streams on the compatible list.
			if ((NULL != outputSynchronousList) && (0 != outputSynchronousList->getCount()))
			{
				compatibleStreamList->merge(outputSynchronousList);
				outputSynchronousList->release();
				outputSynchronousList = 0;
			}

			// If there is at least an adaptive output stream, put the streams on the compatible list.
			if ((NULL != outputAdaptiveList) && (0 != outputAdaptiveList->getCount()))
			{
				compatibleStreamList->merge(outputAdaptiveList);
				outputAdaptiveList->release();
				outputAdaptiveList = 0;
			}
		}
		else
		{
			// No input stream present, so check if there is a synchronous or adaptive output stream present.
			if ((NULL != outputSynchronousList) || (NULL != outputAdaptiveList))
			{
				// If there is at least a synchronous output stream, put the streams on the compatible list.
				if ((NULL != outputSynchronousList) && (0 != outputSynchronousList->getCount()))
				{
					compatibleStreamList->merge(outputSynchronousList);
					outputSynchronousList->release();
					outputSynchronousList = 0;
				}

				// If there is at least an adaptive output stream, put the streams on the compatible list.
				if ((NULL != outputAdaptiveList) && (0 != outputAdaptiveList->getCount()))
				{
					compatibleStreamList->merge(outputAdaptiveList);
					outputAdaptiveList->release();
					outputAdaptiveList = 0;
				}
			}
			else
			{
				// No synchronous or adaptive output streams, so check if there is an asynchronous stream present.
				// If there is at least an asynchronous output stream, put the streams on the compatible list.
				if ((NULL != outputAsynchronousList) && (0 != outputAsynchronousList->getCount()))
				{
					compatibleStreamList->merge(outputAsynchronousList);
					outputAsynchronousList->release();
					outputAsynchronousList = 0;
				}
			}
		}
	}

	if (0 != compatibleStreamList->getCount())
	{
		// There is some streams with compatible endpoints in the compatible list.
		result = compatibleStreamList;
		result->retain();
	}
	else
	{
		// Nothing in the compatible stream list. Get one from the other lists.
		OSNumber	*streamInterfaceNumber = NULL;

		if		((NULL != inputAsynchronousList) && (0 != inputAsynchronousList->getCount()))
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, inputAsynchronousList->getObject(0))), Exit);
			streamInterfaceNumber->retain();
			inputAsynchronousList->removeObject(0);
		}
		else if ((NULL != inputAdaptiveList) && (0 != inputAdaptiveList->getCount()))
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, inputAdaptiveList->getObject(0))), Exit);
			streamInterfaceNumber->retain();
			inputAdaptiveList->removeObject(0);
		}
		else if ((NULL != inputSynchronousList) && (0 != inputSynchronousList->getCount()))
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, inputSynchronousList->getObject(0))), Exit);
			streamInterfaceNumber->retain();
			inputSynchronousList->removeObject(0);
		}
		else if ((NULL != outputAsynchronousList) && (0 != outputAsynchronousList->getCount()))
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, outputAsynchronousList->getObject(0))), Exit);
			streamInterfaceNumber->retain();
			outputAsynchronousList->removeObject(0);
		}
		else if ((NULL != outputAdaptiveList) && (0 != outputAdaptiveList->getCount()))
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, outputAdaptiveList->getObject(0))), Exit);
			streamInterfaceNumber->retain();
			outputAdaptiveList->removeObject(0);
		}
		else if ((NULL != outputSynchronousList) && (0 != outputSynchronousList->getCount()))
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, outputSynchronousList->getObject(0))), Exit);
			streamInterfaceNumber->retain();
			outputSynchronousList->removeObject(0);
		}
		else if ((NULL != inputUnknownList) && (0 != inputUnknownList->getCount()))
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, inputUnknownList->getObject(0))), Exit);
			streamInterfaceNumber->retain();
			inputUnknownList->removeObject(0);
		}
		else if ((NULL != outputUnknownList) && (0 != outputUnknownList->getCount()))
		{
			FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, outputUnknownList->getObject(0))), Exit);
			streamInterfaceNumber->retain();
			outputUnknownList->removeObject(0);
		}

		if (NULL != streamInterfaceNumber)
		{
			compatibleStreamList->setObject(streamInterfaceNumber);
			result = compatibleStreamList;
			result->retain();

			streamInterfaceNumber->release();
		}
	}

#ifdef DEBUGLOGGING
	if (NULL != result)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithCompatibleEndpoints(%p) - Found streams:", this, availableStreamList);

		for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < result->getCount(); streamInterfaceIndex++)
		{
			OSNumber	*streamInterfaceNumber = OSDynamicCast(OSNumber, result->getObject(streamInterfaceIndex));
			FailMessage(NULL == streamInterfaceNumber);

			if (NULL != streamInterfaceNumber)
			{
				DP(DBG_INFO, "--> #%u", streamInterfaceNumber->unsigned8BitValue());
			}
		}
	}
#endif

Exit:
	// Put back unused streams back on the available list.
	if (NULL != inputAsynchronousList)
	{
		if (0 != inputAsynchronousList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(inputAsynchronousList);
		}
		inputAsynchronousList->release();
	}

	if (NULL != inputAdaptiveList)
	{
		if (0 != inputAdaptiveList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(inputAdaptiveList);
		}
		inputAdaptiveList->release();
	}

	if (NULL != inputSynchronousList)
	{
		if (0 != inputSynchronousList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(inputSynchronousList);
		}
		inputSynchronousList->release();
	}

	if (NULL != outputAsynchronousList)
	{
		if (0 != outputAsynchronousList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(outputAsynchronousList);
		}
		outputAsynchronousList->release();
	}

	if (NULL != outputAdaptiveList)
	{
		if (0 != outputAdaptiveList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(outputAdaptiveList);
		}
		outputAdaptiveList->release();
	}

	if (NULL != outputSynchronousList)
	{
		if (0 != outputSynchronousList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(outputSynchronousList);
		}
		outputSynchronousList->release();
	}

	if (NULL != inputUnknownList)
	{
		if (0 != inputUnknownList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(inputUnknownList);
		}
		inputUnknownList->release();
	}

	if (NULL != outputUnknownList)
	{
		if (0 != outputUnknownList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(outputUnknownList);
		}
		outputUnknownList->release();
	}

	if (NULL != compatibleStreamList)
	{
		// If it failed for some reason, put back the compatible streams into the available stream list.
		if ((NULL == result) && (0 != compatibleStreamList->getCount()))
		{
			availableStreamList->merge(compatibleStreamList);
		}
		compatibleStreamList->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithCompatibleEndpoints(%p) - result = %p", this, availableStreamList, result);

	return result;
}

/****************************************************************************/
/**
 * @brief	クロック空間を走査し、すべての設定可能なサンプリング周波数を列挙．
 *			PdjUsbAudioDevice::protectedInitHardware() から呼び出される．
 *
 * @return			IOReturn コード．
 */
// [rdar://4867779]
IOReturn
PdjUsbAudioTopology::addSampleRatesFromClockSpace(
	void
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::addSampleRatesFromClockSpace()", this);

	IOReturn		result = kIOReturnError;
	OSArray			*streamInterfaceNumbers = NULL;
	UInt8			numStreamInterfaces;

	FailIf(NULL == mControlInterface, Exit);	// <rdar://7085810>
	FailIf(NULL == mConfigDictionary, Exit);
	FailIf(NULL == mClockGraph, Exit);

	// Discover sample rates for each stream interface
	FailIf(kIOReturnSuccess != mConfigDictionary->getControlledStreamNumbers(&streamInterfaceNumbers, &numStreamInterfaces), Exit);

	for (UInt8 streamInterfaceIndex = 0; streamInterfaceIndex < streamInterfaceNumbers->getCount(); streamInterfaceIndex++)
	{
		OSNumber	*streamInterfaceNumber = NULL;
		FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, streamInterfaceNumbers->getObject(streamInterfaceIndex))), Exit);

		UInt8		streamInterface = streamInterfaceNumber->unsigned8BitValue();

		bool		startAtZero = mConfigDictionary->alternateSettingZeroCanStream(streamInterface);

		// Discover sample rates for each alternate setting
		UInt8		numAltSettings;
		FailIf(kIOReturnSuccess != mConfigDictionary->getNumAltSettings(&numAltSettings, streamInterface), Exit);

		for (
				UInt8 altSettingIndex = (startAtZero ? 0 : 1);
				altSettingIndex < numAltSettings;
				altSettingIndex++
			)
		{
			OSArray		*pathGroupArray = NULL;
			OSArray		*pathArray = NULL;
			UInt8		terminalID;
			UInt8		clockSourceID;

			FailIf(kIOReturnSuccess != mConfigDictionary->getTerminalLink(&terminalID, streamInterface, altSettingIndex), Exit);
			FailIf(kIOReturnSuccess != mConfigDictionary->getClockSourceID(&clockSourceID, mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, terminalID), Exit);

			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::addSampleRatesFromClockSpace() - interface %d, alt setting %d, clock source ID %d",
				streamInterface, altSettingIndex, clockSourceID);

			// Find the path group that begins with the clockSourceID.
			for (UInt8 pathGroupIndex = 0; pathGroupIndex < mClockGraph->getCount(); pathGroupIndex++)
			{
				OSNumber	*clockSourceIDNumber = NULL;

				FailIf(NULL == (pathGroupArray = OSDynamicCast(OSArray, mClockGraph->getObject(pathGroupIndex))), Exit);
				FailIf(NULL == (pathArray = OSDynamicCast(OSArray, pathGroupArray->getObject(0))), Exit);
				FailIf(NULL == (clockSourceIDNumber = OSDynamicCast(OSNumber, pathArray->getObject(0))), Exit);

				if (clockSourceID == clockSourceIDNumber->unsigned8BitValue())
				{
					// We have found the path group in which we are interested.
					break;
				}
				else
				{
					pathGroupArray = NULL;
				}
			}

			FailIf(NULL == pathGroupArray, Exit);

			// For each path in the path group, add the available sample rates.
			for (UInt8 pathIndex = 0; pathIndex < pathGroupArray->getCount(); pathIndex++)
			{
				FailIf(NULL == (pathArray = OSDynamicCast(OSArray, pathGroupArray->getObject(pathIndex))), Exit);
				FailIf(kIOReturnSuccess != this->addSampleRatesFromClockPath(pathArray, streamInterface, altSettingIndex), Exit);
			}
		}
	}

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::addSampleRatesFromClockSpace() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックパスにサンプリング周波数を設定．
 *
 * @param[in]		sampleRate		設定するサンプリング周波数．
 * @param[in]		*clockPath		クロックパス．
 * @param[in]		failIfReadOnly	true なら、クロックソースが read-only で、
 *										且つ所望の設定値と異なる場合、失敗とする．
 *									false なら、クロックソースが read-only の場合、
 *										所望の設定値と異なっていても成功とする．
 *
 * @return			IOReturn コード．
 */
// <rdar://6945472>
IOReturn
PdjUsbAudioTopology::setClockPathCurSampleRate(
	UInt32			sampleRate,
	OSArray			*clockPath,
	bool			failIfReadOnly
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::setClockPathCurSampleRate(%u)", this, sampleRate);

	OSObject		*arrayObject = NULL;
	OSNumber		*arrayNumber = NULL;
	UInt8			clockID;
	UInt8			subType;
	UInt16			numerator;
	UInt16			denominator;
	OSArray			*clockSourceIDs = NULL;
	UInt8			nextClockID;
	IOReturn		result = kIOReturnError;

	FailIf(NULL == mControlInterface, Exit);

	for (UInt32 clockIndex = 0; clockIndex < clockPath->getCount(); clockIndex++)
	{
		FailIf(NULL == (arrayObject = clockPath->getObject(clockIndex)), Exit);
		FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);
		clockID = arrayNumber->unsigned8BitValue();

		FailIf(
					kIOReturnSuccess != (
											result = mConfigDictionary->getSubType(
																			&subType,
																			mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
																			0,
																			clockID
																		)
										),
					Exit
			  );

		if (USBAUDIO_0200::CLOCK_SOURCE == subType)
		{
			if (mConfigDictionary->clockSourceHasFrequencyControl(mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, clockID, true))
			{
				FailIf(kIOReturnSuccess != (result = this->setCurClockSourceSamplingFrequency(clockID, sampleRate)), Exit);
			}
			else
			{
				if (failIfReadOnly)		// <rdar://6945472> If clock is source read only, check if the current sample rate matches the new sample rate.
				{
					UInt32		clockRate;

					result = this->getCurClockSourceSamplingFrequency(clockID, &clockRate, NULL);

					if (kIOReturnSuccess == result)
					{
						result = (clockRate == sampleRate) ? kIOReturnSuccess : kIOReturnUnsupported;
					}
				}
				else
				{
					result = kIOReturnSuccess;
				}
			}

			break;
		}
		else if (USBAUDIO_0200::CLOCK_SELECTOR == subType)
		{
			FailIf(
						kIOReturnSuccess != (
												result = mConfigDictionary->getClockSelectorSources(
																				&clockSourceIDs,
																				mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
																				0,
																				clockID
																			)
											),
						Exit
			);

			FailIf(NULL == (arrayObject = clockPath->getObject(clockIndex + 1)), Exit);
			FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);
			nextClockID = arrayNumber->unsigned8BitValue();

			for (UInt32 clockSourceIndex = 0; clockSourceIndex < clockSourceIDs->getCount(); clockSourceIndex++)
			{
				FailIf(NULL == (arrayObject = clockSourceIDs->getObject(clockSourceIndex)), Exit);
				FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

				if (arrayNumber->unsigned8BitValue() == nextClockID)
				{
					this->setCurClockSelector(clockID, clockSourceIndex + 1);
					break;
				}
			}
		}
		else if (USBAUDIO_0200::CLOCK_MULTIPLIER == subType)
		{
			FailIf(kIOReturnSuccess != (result = this->getCurClockMultiplier(clockID, &numerator, &denominator)), Exit);

			sampleRate = sampleRate * denominator / numerator;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::setClockPathCurSampleRate(%u) = %d", this, sampleRate, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックパスのサンプリング周波数を取得．
 *
 * @param[out]		*sampleRate		取得したサンプリング周波数．
 * @param[out]		*validity		クロックソースの有効性．
 * @param[out]		*isReadOnly		クロックソースは programmable か read-only か．
 * @param[in]		*clockPath		クロックパス．
 *
 * @return			IOReturn コード．
 */
// <rdar://6945472>
IOReturn
PdjUsbAudioTopology::getClockPathCurSampleRate(
	UInt32			*sampleRate,
	bool			*validity,
	bool			*isReadOnly,
	OSArray			*clockPath
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getClockPathCurSampleRate()", this);

	IOReturn		result = kIOReturnError;
	OSObject		*arrayObject = NULL;
	OSNumber		*arrayNumber = NULL;
	UInt8			clockID;
	UInt8			subType;
	UInt32			clockRate = 0;
	bool			clockValidity	= false;
	bool			clockIsReadOnly	= false;	// <rdar://6945472>
	UInt16			numerator;
	UInt16			denominator;

	FailIf(NULL == mControlInterface, Exit);

	for (UInt32 clockIndex = clockPath->getCount(); clockIndex > 0; clockIndex--)
	{
		FailIf(NULL == (arrayObject = clockPath->getObject(clockIndex - 1)), Exit);
		FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

		clockID = arrayNumber->unsigned8BitValue();

		FailIf(
					kIOReturnSuccess != (
											result = mConfigDictionary->getSubType(
																			&subType,
																			mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
																			0,
																			clockID
																		)
										),
					Exit
			  );

		if (USBAUDIO_0200::CLOCK_SOURCE == subType)
		{
			// <rdar://6945472>
			if (!mConfigDictionary->clockSourceHasFrequencyControl(mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, clockID, true))
			{
				clockIsReadOnly = true;
			}

			FailIf(
					kIOReturnSuccess != (
											result = this->getCurClockSourceSamplingFrequency(
																clockID,
																&clockRate,
																&clockValidity
														   )
										),
					Exit
			);
		}
		else if (USBAUDIO_0200::CLOCK_MULTIPLIER == subType)
		{
			FailIf(kIOReturnSuccess != (result = this->getCurClockMultiplier(clockID, &numerator, &denominator)), Exit);

			clockRate = clockRate * numerator / denominator;
		}
	}

Exit:
	if (NULL != sampleRate)
	{
		*sampleRate = clockRate;
	}

	if (NULL != validity)
	{
		*validity = clockValidity;
	}

	if (NULL != isReadOnly)		// <rdar://6945472>
	{
		*isReadOnly = clockIsReadOnly;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getClockPathCurSampleRate(%u, %d) = %d",
		this, clockRate, clockValidity, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースにサンプリング周波数を設定．
 *
 * @param[in]		unitID				クロックソース ID．
 * @param[in]		samplingFrequency	設定するサンプリング周波数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::setCurClockSourceSamplingFrequency(
	UInt8			unitID,
	UInt32			samplingFrequency
)
{
	UInt32			frequency = USBToHostLong(samplingFrequency);

	return this->setClockSetting(USBAUDIO_0200::CS_SAM_FREQ_CONTROL, unitID, USBAUDIO_0200::CUR, &frequency, 4);
}

/****************************************************************************/
/**
 * @brief	クロックソースのサンプリング周波数を取得．
 *
 * @param[in]		unitID				クロックソース ID．
 * @param[out]		*samplingFrequency	取得したサンプリング周波数．
 * @param[out]		*validity			クロックソースの有効性．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getCurClockSourceSamplingFrequency(
	UInt8			unitID,
	UInt32			*samplingFrequency,
	bool			*validity
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getCurClockSourceSamplingFrequency()", this);

	UInt32			clockFrequency = 0;
	bool			clockValidity = false;
	bool			hasValidityControl;			// <rdar://7446555>
	IOReturn		result = kIOReturnError;	// <rdar://7446555>

	FailIf(NULL == mControlInterface, Exit);	// <rdar://7446555>
	FailIf(NULL == mConfigDictionary, Exit);	// <rdar://7446555>

	result = this->getClockSetting(USBAUDIO_0200::CS_SAM_FREQ_CONTROL, unitID, USBAUDIO_0200::CUR, &clockFrequency, 4);
	FailIf(kIOReturnSuccess != result, Exit);

	hasValidityControl = mConfigDictionary->clockSourceHasValidityControl(mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, unitID);	// <rdar://7446555>

	// <rdar://7446555> If the clock source has validity control, ask it for the value, otherwise assume it is there.
	if (hasValidityControl)
	{
		result = this->getClockSetting(USBAUDIO_0200::CS_CLOCK_VALID_CONTROL, unitID, USBAUDIO_0200::CUR, &clockValidity, 1);
		FailIf(kIOReturnSuccess != result, Exit);
	}
	else
	{
		clockValidity = true;
	}

Exit:
	if (NULL != samplingFrequency)
	{
		*samplingFrequency = USBToHostLong(clockFrequency);
	}

	if (NULL != validity)
	{
		*validity = clockValidity;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getCurClockSourceSamplingFrequency() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックセレクタにクロックパスを設定．
 *
 * @param[in]		unitID			クロックセレクタのユニット ID．
 * @param[in]		selector		クロックセレクタに設定するクロックパスのインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::setCurClockSelector(
	UInt8			unitID,
	UInt8			selector
)
{
	return this->setClockSetting(
						USBAUDIO_0200::CX_CLOCK_SELECTOR_CONTROL,
						unitID,
						USBAUDIO_0200::CUR,
						&selector,
						1
				 );
}

/****************************************************************************/
/**
 * @brief	クロックセレクタの ID とクロックパスのインデックスを取得．
 *
 * @param[out]		*selectorID		クロックセレクタのユニット ID．
 * @param[out]		*pathIndex		クロックセレクタから検出したクロックパスのインデックス．
 * @param[in]		*clockPath		クロックセレクタの探索対象となるクロックパス．
 *
 * @return			IOReturn コード．
 */
// <rdar://5811247>
IOReturn
PdjUsbAudioTopology::getClockSelectorIDAndPathIndex(
	UInt8			*selectorID,
	UInt8			*pathIndex,
	OSArray			*clockPath
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getClockSelectorIDAndPathIndex()", this);

	IOReturn		result = kIOReturnNotFound;

	FailIf(NULL == mControlInterface, Exit);
	FailIf(NULL == selectorID, Exit);
	FailIf(NULL == pathIndex, Exit);

	*selectorID	= 0;
	*pathIndex	= 0;

	for (UInt32 clockIndex = 0; clockIndex < clockPath->getCount(); clockIndex++)
	{
		OSNumber	*clockIDNumber;
		UInt8		clockID;
		UInt8		subType;

		FailIf(NULL == (clockIDNumber = OSDynamicCast(OSNumber, clockPath->getObject(clockIndex))), Exit);

		clockID = clockIDNumber->unsigned8BitValue();

		FailIf(
				kIOReturnSuccess != (
										result = mConfigDictionary->getSubType(
																		&subType,
																		mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
																		0,
																		clockID
																	)
									),
				Exit
		);

		if (USBAUDIO_0200::CLOCK_SELECTOR == subType)
		{
			OSArray		*clockSourceIDs		 = NULL;
			OSNumber	*clockSourceIDNumber = NULL;
			OSNumber	*nextClockIDNumber	 = NULL;
			UInt8		nextClockID;

			FailIf(
					kIOReturnSuccess != (
											result = mConfigDictionary->getClockSelectorSources(
																			&clockSourceIDs,
																			mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
																			0,
																			clockID
																		)
										),
					Exit
			);

			FailIf(
					NULL == (
								nextClockIDNumber = OSDynamicCast(
														OSNumber,
														clockPath->getObject(clockIndex + 1)
													)
							),
					Exit
			);

			nextClockID = nextClockIDNumber->unsigned8BitValue();

			for (UInt32 index = 0; index < clockSourceIDs->getCount(); index++)
			{
				FailIf(NULL == (clockSourceIDNumber = OSDynamicCast(OSNumber, clockSourceIDs->getObject(index))), Exit);

				if (clockSourceIDNumber->unsigned8BitValue() == nextClockID)
				{
					*selectorID	= clockID;
					*pathIndex	= index + 1;

					result = kIOReturnSuccess;
					break;
				}
			}
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getClockSelectorIDAndPathIndex() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオコントロール生成．
 *			PdjUsbAudioDevice::createControlsForInterface() から呼び出される．
 *			interface 内の各種オーディオコントロールを生成する．
 *
 * @param[in]		*audioEngine		PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		interfaceNum		USB interface 番号．
 * @param[in]		altSettingNum		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::doControlStuff(
	IOAudioEngine	*audioEngine,
	UInt8			interfaceNum,
	UInt8			altSettingNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::doControlStuff(%p, %d, %d)", audioEngine, interfaceNum, altSettingNum);

	IOReturn				result = kIOReturnError;
	PdjUsbAudioEngine		*usbAudioEngine;
	SInt32					engineIndex;
	SInt32					streamIndex;			// <rdar://6420832>
#ifdef TARGET_PDJ
#else
	OSArray					*arrayOfPathsFromOutputTerminal;
	OSArray					*aPath;
	OSNumber				*theUnitIDNum;
	UInt32					numOutputTerminalArrays;
	UInt32					numPathsFromOutputTerminal;
	UInt16					terminalType;
	UInt8					numInputTerminals;		// <rdar://6413207>
	UInt8					numOutputTerminals;
	UInt8					selectorUnitID;
	UInt8					featureUnitID;
	UInt8					volFeatureUnitID;		// <rdar://5366067>
	UInt8					muteFeatureUnitID;		// <rdar://5366067>
	UInt8					controlInterfaceNum;
	UInt8					outputTerminalID;
	UInt8					inputTerminalID;		// <rdar://6430836>
	UInt8					direction;
	bool					finished = false;		// <rdar://problem/6656791>
#endif

	usbAudioEngine = OSDynamicCast(PdjUsbAudioEngine, audioEngine);
	FailIf(NULL == usbAudioEngine, Exit);

	FailIf(NULL == mControlInterface, Exit);

	DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - This usbAudioEngine = %p", usbAudioEngine);

	FailIf(kIOReturnSuccess != this->registerEngineInfo(usbAudioEngine), Exit);					// <rdar://6420832>
	FailIf(kIOReturnSuccess != this->registerStreamInfo(interfaceNum, altSettingNum), Exit);	// <rdar://6420832>

	engineIndex = this->getEngineInfoIndex(usbAudioEngine);										// <rdar://6413207>
	streamIndex = this->getStreamInfoIndex(interfaceNum);										// <rdar://6413207>
	FailIf((-1 == engineIndex) || (-1 == streamIndex), Exit);									// <rdar://6413207>

#ifdef TARGET_PDJ
#else
	volFeatureUnitID	= 0;	// <rdar://5366067>
	muteFeatureUnitID	= 0;	// <rdar://5366067>
	outputTerminalID	= 0;
	controlInterfaceNum	= mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	FailIf(kIOReturnSuccess != mConfigDictionary->getNumInputTerminals(&numInputTerminals, controlInterfaceNum, 0), Exit);	// <rdar://6413207>
	FailIf(kIOReturnSuccess != mConfigDictionary->getNumOutputTerminals(&numOutputTerminals, controlInterfaceNum, 0), Exit);
	FailIf(kIOReturnSuccess != mConfigDictionary->getIsocEndpointDirection(&direction, interfaceNum, altSettingNum), Exit);

	if (direction == kIOAudioStreamDirectionOutput)
	{
		IOAudioSelectorControl	*outputSelector				 = NULL;	// <rdar://6413207>
		UInt32					numConnectedOutputTerminals	 = 0;		// <rdar://6413207>
		UInt8					defaultOutputTerminalID		 = 0;		// <rdar://6413207>
		bool					hasInitializedOutputControls = false;	// <rdar://6413207>
		bool					done						 = false;

		// <rdar://6413207> Get the input terminal that is associated with this interface
		FailIf(kIOReturnSuccess != mConfigDictionary->getTerminalLink(&inputTerminalID, interfaceNum, altSettingNum), Exit);

		if (0 == inputTerminalID)
		{
			// if not found, use the first USB streaming input terminal
			for (UInt32 inputTerminalIndex = 0; inputTerminalIndex < numInputTerminals; inputTerminalIndex++)
			{
				FailIf(kIOReturnSuccess != mConfigDictionary->getIndexedInputTerminalType(&terminalType, controlInterfaceNum, 0, inputTerminalIndex), Exit);

				if (USB_STREAMING == terminalType)
				{
					FailIf(kIOReturnSuccess != mConfigDictionary->getIndexedInputTerminalID(&inputTerminalID, controlInterfaceNum, 0, inputTerminalIndex), Exit);
					break;		// Found the (hopefully only) streaming input terminal we're looking for
				}
			}
		}

		defaultOutputTerminalID = this->getDefaultOutputTerminalID(inputTerminalID);			// <rdar://6413207>
		numConnectedOutputTerminals = this->getNumConnectedOutputTerminals(inputTerminalID);	// <rdar://6413207>

		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - numConnectedOutputTerminals = %u inputTerminalID: %u", numConnectedOutputTerminals, inputTerminalID);

		for (UInt32 outputTerminalIndex = 0; (outputTerminalIndex < numOutputTerminals) && (!done); outputTerminalIndex++)
		{
			FailIf(kIOReturnSuccess != mConfigDictionary->getIndexedOutputTerminalType(&terminalType, controlInterfaceNum, 0, outputTerminalIndex), Exit);

			if (terminalType != USB_STREAMING)
			{
				UInt8	inputUnitID, outputUnitID;	// <rdar://6413207>

				FailIf(kIOReturnSuccess != mConfigDictionary->getIndexedOutputTerminalID(&outputTerminalID, controlInterfaceNum, 0, outputTerminalIndex), Exit);

				numOutputTerminalArrays = mControlGraph->getCount();

				finished = false;	// <rdar://problem/6656791>

				for (UInt32 pathsToOutputTerminalN = 0; (pathsToOutputTerminalN < numOutputTerminalArrays) && (!finished); pathsToOutputTerminalN++)
				{
					arrayOfPathsFromOutputTerminal = OSDynamicCast(OSArray, mControlGraph->getObject(pathsToOutputTerminalN));
					FailIf(NULL == arrayOfPathsFromOutputTerminal, Exit);

					// <rdar://problem/6656791> PdjUsbAudio: Output controls not set up properly for stream interfaces
					UInt32		numPathsFromOutputTerminals = arrayOfPathsFromOutputTerminal->getCount();

					for (
							UInt32 arrayOfPathsFromOutputTerminalIndex = 0;
							(arrayOfPathsFromOutputTerminalIndex < numPathsFromOutputTerminals) && (!finished);
							arrayOfPathsFromOutputTerminalIndex++
						)
					{
						aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(arrayOfPathsFromOutputTerminalIndex));
						FailIf(NULL == aPath, Exit);

						theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(0));
						FailIf(NULL == theUnitIDNum, Exit);

						outputUnitID = theUnitIDNum->unsigned8BitValue();	// <rdar://6413207>

						theUnitIDNum = OSDynamicCast(OSNumber, aPath->getLastObject());
						FailIf(NULL == theUnitIDNum, Exit);

						inputUnitID = theUnitIDNum->unsigned8BitValue();

						if ((inputUnitID == inputTerminalID) && (outputUnitID == outputTerminalID))	// <rdar://6413207>
						{
							if (!hasInitializedOutputControls)
							{
								if (numConnectedOutputTerminals > 1)	// <rdar://6413207>, <rdar://6500500>
								{
									// Create a virtual selector for output path selection.
									outputSelector = IOAudioSelectorControl::createOutputSelector(
																				defaultOutputTerminalID,
																				kIOAudioControlChannelIDAll,
																				0,
																				(streamIndex << 16) | (engineIndex << 8) | 0
																			 );	// <rdar://6523676>
									FailIf(NULL == outputSelector, Exit);

									outputSelector->setValueChangeHandler(controlChangedHandler, this);
									usbAudioEngine->addDefaultAudioControl(outputSelector);
								}

								featureUnitID = this->getBestFeatureUnitInPath(
															aPath,
															kIOAudioControlUsageOutput,
															interfaceNum,
															altSettingNum,
															kVolumeControl
													  );
								if (featureUnitID)
								{
									// Create the output gain controls
									volFeatureUnitID = featureUnitID;	// <rdar://5366067>

									DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Creating output gain controls");

									this->addVolumeControls(
												usbAudioEngine,
												volFeatureUnitID,
												outputTerminalID,
												interfaceNum,
												altSettingNum,
												kIOAudioControlUsageOutput
										  );	// <rdar://6413207>
								}

								featureUnitID = this->getBestFeatureUnitInPath(
															aPath,
															kIOAudioControlUsageOutput,
															interfaceNum,
															altSettingNum,
															kMuteControl
													  );
								if (featureUnitID)
								{
									DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Creating output mute controls");

									muteFeatureUnitID = featureUnitID;	// <rdar://5366067>

									this->addMuteControl(
												usbAudioEngine,
												muteFeatureUnitID,
												outputTerminalID,
												interfaceNum,
												altSettingNum,
												kIOAudioControlUsageOutput
										  );	// <rdar://6413207>

									// If there is only 1 output terminal, then we are done.
									if (numConnectedOutputTerminals <= 1)
									{
										done = true;
									}
								}

								// <rdar://5366067> Handle the case where the volume & mute controls are on different feature units.
								if (volFeatureUnitID != muteFeatureUnitID)
								{
									UInt8		numControls;	// <rdar://5366067>

									// Unmute the volume feature unit.
									if (volFeatureUnitID)
									{
										FailIf(
												kIOReturnSuccess != mConfigDictionary->getNumControls(
																							&numControls,
																							controlInterfaceNum,
																							0,
																							volFeatureUnitID
																					   ),
												Exit
										);

										for (UInt8 channelNum = 0; channelNum < numControls; channelNum++)	// <rdar://5366067>
										{
											if (mConfigDictionary->channelHasMuteControl(controlInterfaceNum, 0, volFeatureUnitID, channelNum))
											{
												this->setCurMute(volFeatureUnitID, channelNum, 0);
											}
										}
									}

									// Set the gain of the mute feature unit.
									if (muteFeatureUnitID)
									{
										FailIf(
												kIOReturnSuccess != mConfigDictionary->getNumControls(
																							&numControls,
																							controlInterfaceNum,
																							0,
																							muteFeatureUnitID
																					   ),
												Exit
										);

										for (UInt8 channelNum = 0; channelNum < numControls; channelNum++)	// <rdar://5366067>
										{
											if (mConfigDictionary->channelHasVolumeControl(controlInterfaceNum, 0, muteFeatureUnitID, channelNum))
											{
												SInt16		deviceMax;	// <rdar://5366067>

												if (kIOReturnSuccess == this->getMaxVolume(muteFeatureUnitID, channelNum, &deviceMax))
												{
													this->setCurVolume(muteFeatureUnitID, channelNum, ((deviceMax >= 0) ? 0 : deviceMax));
												}
											}
										}
									}
								}

								usbAudioEngine->updateChannelNames(aPath, interfaceNum, altSettingNum);	// <rdar://6430836> <rdar://problem/6706026>

								hasInitializedOutputControls = true;
							}

							if (NULL != outputSelector)	// <rdar://6413207>
							{
								// <rdar://6394629> Get the name from the terminal if it exists.
								OSString	*nameString = this->getNameForTerminal(outputTerminalID, kIOAudioStreamDirectionOutput);	// <rdar://6413207>
								FailIf(NULL == nameString, Exit);

								if (!outputSelector->valueExists(outputTerminalID))
								{
									outputSelector->addAvailableSelection(outputTerminalID, nameString);
								}

								nameString->release();

#ifdef BUG_FIX
								outputSelector->release();
								outputSelector = NULL;
#endif
							}

							finished = true;	// <rdar://problem/6656791>
						}  // if ((inputUnitID == inputTerminalID) && (outputUnitID == outputTerminalID))
					}  // for arrayOfPathsFromOutputTerminalIndex
				} // for pathsToOutputTerminalN
			} // if (terminalType != USB_STREAMING)
		} // for outputTerminalIndex
	}
	else
	{	// direction == kIOAudioStreamDirectionInput
		IOAudioSelectorControl	*inputSelector = NULL;

		// <rdar://problem/6385557>
		// Get the output terminal that is associated with this interface
		FailIf(kIOReturnSuccess != mConfigDictionary->getTerminalLink(&outputTerminalID, interfaceNum, altSettingNum), Exit);

		if (outputTerminalID == 0)
		{
			// if not found, use the first USB streaming output terminal
			for (UInt32 outputTerminalIndex = 0; outputTerminalIndex < numOutputTerminals; outputTerminalIndex++)
			{
				FailIf(
						kIOReturnSuccess != mConfigDictionary->getIndexedOutputTerminalType(
																	&terminalType,
																	controlInterfaceNum,
																	0,
																	outputTerminalIndex
															   ),
						Exit
				);
				if (USB_STREAMING == terminalType)
				{
					FailIf(
							kIOReturnSuccess != mConfigDictionary->getIndexedOutputTerminalID(
																		&outputTerminalID,
																		controlInterfaceNum,
																		0,
																		outputTerminalIndex
																   ),
							Exit
					);
					break;	// Found the (hopefully only) streaming output terminal we're looking for
				}
			}
		}

		numOutputTerminalArrays = mControlGraph->getCount();
		finished = false;	// <rdar://problem/6656791>

		for (UInt32 pathsToOutputTerminalN = 0; (pathsToOutputTerminalN < numOutputTerminalArrays) && (!finished); pathsToOutputTerminalN++)
		{
			arrayOfPathsFromOutputTerminal = OSDynamicCast(OSArray, mControlGraph->getObject(pathsToOutputTerminalN));
			FailIf(NULL == arrayOfPathsFromOutputTerminal, Exit);

			// <rdar://problem/6656791>
			UInt32		numPathsFromOutputTerminals = arrayOfPathsFromOutputTerminal->getCount();

			for (
					UInt32 arrayOfPathsFromOutputTerminalIndex = 0;
					(arrayOfPathsFromOutputTerminalIndex < numPathsFromOutputTerminals) && (!finished);
					arrayOfPathsFromOutputTerminalIndex++
				)
			{
				UInt8	unitID;

				aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(arrayOfPathsFromOutputTerminalIndex));
				FailIf(NULL == aPath, Exit);

				theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(0));
				FailIf(NULL == theUnitIDNum, Exit);

				unitID = theUnitIDNum->unsigned8BitValue();

				theUnitIDNum = OSDynamicCast(OSNumber, aPath->getLastObject());	// <rdar://6430836>
				FailIf(NULL == theUnitIDNum, Exit);

				inputTerminalID = theUnitIDNum->unsigned8BitValue();

				if (unitID == outputTerminalID)
				{
					numPathsFromOutputTerminal = arrayOfPathsFromOutputTerminal->getCount();

					UInt8	numSelectorUnits;
					FailIf(kIOReturnSuccess != mConfigDictionary->getNumSelectorUnits(&numSelectorUnits, controlInterfaceNum, 0), Exit);

					if ((numPathsFromOutputTerminal > 1) && (numSelectorUnits))
					{
						// Found the array of paths that lead to our streaming output terminal
						UInt32		unitIndexInPath;
						UInt32		numUnitsInPath = aPath->getCount();

						for (unitIndexInPath = 1; unitIndexInPath < numUnitsInPath; unitIndexInPath++)
						{
							theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(unitIndexInPath));
							FailIf(NULL == theUnitIDNum, Exit);

							unitID = theUnitIDNum->unsigned8BitValue();

							UInt8	subType;
							FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

							if (SELECTOR_UNIT == subType)
							{
								// <rdar://6523676> Get the current selector setting
								UInt8		defaultSelectorSetting = this->getSelectorSetting(unitID);

								if (0 == defaultSelectorSetting)
								{
									DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Unable to get the selector setting. Defaulting to 1");
									defaultSelectorSetting = 1;
								}

								DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Default selector setting: %d", defaultSelectorSetting);

								// <rdar://6523676> Update the aPath and inputTerminalID if it is not first path.
								UInt32		pathIndex = 0;

								if (1 != defaultSelectorSetting)
								{
									pathIndex = this->getPathIndexForSelectorSetting(
															arrayOfPathsFromOutputTerminal,
															pathsToOutputTerminalN,
															arrayOfPathsFromOutputTerminalIndex,
															unitIndexInPath,
															defaultSelectorSetting
													  );	// <rdar://problem/6656791>

									aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(pathIndex));

									if (NULL == aPath)	// Make sure that there is a path associated with the selector setting.
									{					// The device may be returning an invalid value.
										pathIndex = 0;
										defaultSelectorSetting = 1;

										aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(pathIndex));
										FailIf(NULL == aPath, Exit);
									}

									theUnitIDNum = OSDynamicCast(OSNumber, aPath->getLastObject());
									FailIf(NULL == theUnitIDNum, Exit);

									inputTerminalID = theUnitIDNum->unsigned8BitValue();
								}

								DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Selected aPath = %p, pathIndex = %u, inputTerminalID = %u",
									aPath, pathIndex, inputTerminalID);

								if (kIOReturnSuccess == this->setSelectorSetting(unitID, defaultSelectorSetting))	// <rdar://6523676>
								{
									selectorUnitID	= unitID;
									engineIndex		= this->getEngineInfoIndex(usbAudioEngine);
									streamIndex		= this->getStreamInfoIndex(interfaceNum);	// <rdar://6420832>

									if ((-1 != engineIndex) && (-1 != streamIndex))		// <rdar://6420832>
									{
										UInt32	selection =   (0xFF000000 & (pathsToOutputTerminalN	<< 24))
															| (0x00FF0000 & (pathIndex				<< 16))
															| (0x0000FF00 & (selectorUnitID			<<  8))
															| (0x000000FF & (defaultSelectorSetting	<<  0));	// <rdar://6523676>

										inputSelector = IOAudioSelectorControl::createInputSelector(
																					selection,
																					kIOAudioControlChannelIDAll,
																					0,
																					(streamIndex << 16) | ( engineIndex << 8 ) | selectorUnitID
																				);	// <rdar://6420832>, <rdar://6523676>
										FailIf(NULL == inputSelector, Exit);

										inputSelector->setValueChangeHandler(controlChangedHandler, this);
										usbAudioEngine->addDefaultAudioControl(inputSelector);

										// <rdar://5366067> For the inputs, use the feature unit closest to the selector unit. The same feature unit is
										// used for volume & mute. Look for volume controls first. If it is present, then check if it has mute.
										// If volume controls are not present, then look for mute controls.
										featureUnitID = this->getBestFeatureUnitInPath(
																	aPath,
																	kIOAudioControlUsageInput,
																	interfaceNum,
																	altSettingNum,
																	kVolumeControl
															  );
										if (featureUnitID)
										{
											// Create the input gain controls
											DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Creating input gain controls");
											this->addVolumeControls(
														usbAudioEngine,
														featureUnitID,
														inputTerminalID,
														interfaceNum,
														altSettingNum,
														kIOAudioControlUsageInput
												  );	// <rdar://6413207>

											DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Creating input mute controls");
											this->addMuteControl(
														usbAudioEngine,
														featureUnitID,
														inputTerminalID,
														interfaceNum,
														altSettingNum,
														kIOAudioControlUsageInput
												  );	// <rdar://6413207>
										}
										else
										{
											featureUnitID = this->getBestFeatureUnitInPath(
																		aPath,
																		kIOAudioControlUsageInput,
																		interfaceNum,
																		altSettingNum,
																		kMuteControl
																  );
											if (featureUnitID)
											{
												DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Creating input mute controls");
												this->addMuteControl(
															usbAudioEngine,
															featureUnitID,
															inputTerminalID,
															interfaceNum,
															altSettingNum,
															kIOAudioControlUsageInput
													  );	// <rdar://6413207>
											}
										}
									} // if ((-1 != engineIndex) && (-1 != streamIndex))	// <rdar://6420832>
								} // if (kIOReturnSuccess == this->setSelectorSetting(unitID, defaultSelectorSetting))

								break;	// Get out of unitIndexInPath for loop
							} // if (SELECTOR_UNIT == subType)
						} // for unitIndexInPath

						if (NULL != inputSelector)
						{
							this->addSelectorSourcesToSelectorControl(
										inputSelector,
										arrayOfPathsFromOutputTerminal,
										pathsToOutputTerminalN,
										arrayOfPathsFromOutputTerminalIndex,
										unitIndexInPath
								  );	// <rdar://problem/6656791>

							inputSelector->release();
#ifdef BUG_FIX
							inputSelector = NULL;
#endif
						}
						else
						{
							// There are no programmable selectors, so just find the one feature unit, if it exists.
							featureUnitID = this->getBestFeatureUnitInPath(
														aPath,
														kIOAudioControlUsageInput,
														interfaceNum,
														altSettingNum,
														kVolumeControl
												  );
							if (featureUnitID)
							{
								// Create the playthrough volume controls
								DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Creating playthrough volume controls");
								this->addVolumeControls(
											usbAudioEngine,
											featureUnitID,
											inputTerminalID,
											interfaceNum,
											altSettingNum,
											kIOAudioControlUsageInput
									  );	// <rdar://6413207>
							}

							// <rdar://6523676> Add the mute controls
							featureUnitID = this->getBestFeatureUnitInPath(
														aPath,
														kIOAudioControlUsageInput,
														interfaceNum,
														altSettingNum,
														kMuteControl
												  );
							if (featureUnitID)
							{
								// Create the mute controls
								DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Creating input mute control (no selectors)");
								this->addMuteControl(
											usbAudioEngine,
											featureUnitID,
											inputTerminalID,
											interfaceNum,
											altSettingNum,
											kIOAudioControlUsageInput
									  );	// <rdar://6413207>
							}
						}
					} // if ((numPathsFromOutputTerminal > 1) && (numSelectorUnits))
					else
					{
						// There are no selectors, so just find the one feature unit, if it exists.
						featureUnitID = this->getBestFeatureUnitInPath(
													aPath,
													kIOAudioControlUsageInput,
													interfaceNum,
													altSettingNum,
													kVolumeControl
											  );
						if (featureUnitID)
						{
							// Create the playthrough volume controls
							DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - Creating input volume control (no selectors)");
							this->addVolumeControls(
										usbAudioEngine,
										featureUnitID,
										inputTerminalID,
										interfaceNum,
										altSettingNum,
										kIOAudioControlUsageInput
								  );	// <rdar://6413207>
						}

						featureUnitID = this->getBestFeatureUnitInPath(
													aPath,
													kIOAudioControlUsageInput,
													interfaceNum,
													altSettingNum,
													kMuteControl
											  );
						if (featureUnitID)
						{
							// Create the mute controls
							DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff () - Creating input mute control (no selectors)");
							this->addMuteControl(
										usbAudioEngine,
										featureUnitID,
										inputTerminalID,
										interfaceNum,
										altSettingNum,
										kIOAudioControlUsageInput
								  );		// <rdar://6413207>
						}
					}

					usbAudioEngine->updateChannelNames(aPath, interfaceNum, altSettingNum);	// <rdar://6430836> <rdar://problem/6706026>

					// <rdar://6523676> Moved the play thru control initialization after the input controls as the selected input terminal may
					// not be the first one. The code in the input part will set the inputTerminalID to the selected value if necessary.
					// Check for a playthrough path that would require a playthrough control
					OSArray		*playThroughPaths = this->getPlaythroughPaths(inputTerminalID);	// <rdar://5366067>

					if (playThroughPaths)
					{
						DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doControlStuff() - performing playthrough setup");
						this->doPlaythroughSetup(
									usbAudioEngine,
									playThroughPaths,
									interfaceNum,
									altSettingNum,
									inputTerminalID
							  );	// <rdar://5366067>

						playThroughPaths->release();
					}

					finished = true;	// <rdar://problem/6656791>
				} // if (unitID == outputTerminalID)
			} // for (arrayOfPathsFromOutputTerminalIndex)
		} // for (pathsToOutputTerminalN)
	} // else (input direction)
#endif // TARGET_PDJ

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::doControlStuff(%p, %d, %d) = %d", audioEngine, interfaceNum, altSettingNum, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	デバイス側でオーディオコントロールの設定が変更されたときの処理．
 *			PdjUsbAudioDevice::handleStatusInterrupt() から呼び出される．
 *
 * @param[in]		controlID				変更されたユニットの ID．
 * @param[in]		*defaultAudioControls	既定オーディオコントロールの集合．
 */
void
PdjUsbAudioTopology::controlHasChangedOnDevice(
	UInt8			controlID,
	OSSet			*defaultAudioControls
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::controlHasChangedOnDevice()", this);

	UInt32					controlSubType;
	UInt8					numControls;
	IOAudioControl			*controlObject		= NULL ;
	OSCollectionIterator	*controlsIterator	= NULL;

	// Find all volume/mute controls associated with this controlID
	controlsIterator = OSCollectionIterator::withCollection(defaultAudioControls);

	if (controlsIterator)
	{
		while (NULL != (controlObject = OSDynamicCast(IOAudioControl, controlsIterator->getNextObject())))
		{
			if (controlID == ((UInt8)controlObject->getControlID() & 0xFF))	// <rdar://6413207>
			{
				controlSubType = controlObject->getSubType();

				// only adjust volume and mute subtypes
				switch (controlSubType)
				{
					case kIOAudioLevelControlSubTypeVolume:

						FailIf(NULL == mControlInterface, Exit);

						FailIf(
									kIOReturnSuccess != mConfigDictionary->getNumControls(
																				&numControls,
																				mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
																				0,
																				controlID
																		   ),
									Exit
							  );

						for (UInt8 channelNum = 0; channelNum <= numControls; channelNum++)
						{
							// <rdar://6497818> Update the control with the values in the corresponding channel.
							if (controlObject->getChannelID() == channelNum)
							{
								// <rdar://6497818> Only if there is volume control present on the channel.
								if (mConfigDictionary->channelHasVolumeControl(mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, controlID, channelNum))
								{
									SInt16		deviceCur;
									SInt16		deviceMin;
									SInt16		deviceMax;
									UInt16		volRes = 0;
									SInt32		controlCur;
									OSNumber	*settingNumber = NULL;

									// Read settings on device
									FailIf(kIOReturnSuccess != this->getCurVolume(controlID, channelNum, &deviceCur), Exit);
									FailIf(kIOReturnSuccess != this->getMinVolume(controlID, channelNum, &deviceMin), Exit);
									FailIf(kIOReturnSuccess != this->getMaxVolume(controlID, channelNum, &deviceMax), Exit);

									this->getVolumeResolution(controlID, channelNum, &volRes);
									// [rdar://4511427] Need to check volRes for class compliance (Audio Spec 5.2.2.4.3.2).
									FailIf(0 == volRes, Exit);

									// The current control value is a bit of a special case because the device may violate the spec.
									// The rules are as follows:
									// * If the current value is negative infinity, use a control value of 0.
									// * If the current value is the minimum value, use a control value of 0.
									// * Otherwise calculate the current value as normal.
									if (((SInt16)kNegativeInfinity == deviceCur) || (deviceCur == deviceMin))
									{
										controlCur = 0;
									}
									else
									{
										controlCur = (deviceCur - deviceMin) / volRes;
									}

									// Call hardwareValueChanged() for each with settings read from device
									FailIf(NULL == (settingNumber = OSNumber::withNumber(controlCur, SIZEINBITS(SInt32))), Exit);

									controlObject->hardwareValueChanged(settingNumber);

									settingNumber->release();
									settingNumber = NULL;
								}
								break;	// <rdar://6497818>
							}
						}
						break;

					case kIOAudioToggleControlSubTypeMute:

						FailIf(NULL == mControlInterface, Exit);

						FailIf(
									kIOReturnSuccess != mConfigDictionary->getNumControls(
																				&numControls,
																				mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
																				0,
																				controlID
																		   ),
									Exit
							  );

						for (UInt8 channelNum = 0; channelNum <= numControls; channelNum++)
						{
							// <rdar://6497818> Update the control with the values in the corresponding channel.
							if (controlObject->getChannelID() == channelNum)
							{
								// <rdar://6497818> Only if there is mute control present on the channel.
								if (mConfigDictionary->channelHasMuteControl(mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, controlID, channelNum))
								{
									SInt16		deviceCur;
									OSNumber	*settingNumber = NULL;

									FailIf(kIOReturnSuccess != this->getCurMute(controlID, controlObject->getChannelID(), &deviceCur), Exit);

									FailIf(NULL == (settingNumber = OSNumber::withNumber(deviceCur, SIZEINBITS(SInt16))), Exit);

									controlObject->hardwareValueChanged(settingNumber);

									settingNumber->release();
									settingNumber = NULL;
								}
								break;	// <rdar://6497818>
							}
						}
						break;

					// <rdar://6523676> Handle changes in the input source selector
					case kIOAudioSelectorControlSubTypeInput:

						if (kIOAudioControlTypeSelector == controlObject->getType())
						{
							UInt8		oldSelectorPosition;
							UInt8		newSelectorPosition;

							FailIf(NULL == mControlInterface, Exit);

							oldSelectorPosition = controlObject->getIntValue() & 0x000000FF;
							newSelectorPosition = this->getSelectorSetting(controlID);

							if (oldSelectorPosition != newSelectorPosition)
							{
								// The current selected input is no longer valid. Switch over to the new input source.
								OSArray		*availableSelections = OSDynamicCast(
																		OSArray,
																		controlObject->getProperty(
																							kIOAudioSelectorControlAvailableSelectionsKey
																					   )
																   );

								if (NULL != availableSelections)
								{
									for (UInt32 index = 0; index < availableSelections->getCount(); index++)
									{
										OSDictionary	*selectionDictionary = OSDynamicCast(
																					OSDictionary,
																					availableSelections->getObject(index)
																			   );
										if (NULL != selectionDictionary)
										{
											OSNumber	*selectionNumber = OSDynamicCast(
																				OSNumber,
																				selectionDictionary->getObject(
																										kIOAudioSelectorControlSelectionValueKey
																									 )
																		   );
											if (NULL != selectionNumber)
											{
												SInt32		selection = (SInt32)selectionNumber->unsigned32BitValue();

												if ((selection & 0x000000FF) == newSelectorPosition)
												{
													// Found the input source. Switch over to it.
													DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::controlHasChangedOnDevice() - Switch input selector over to selection = 0x%x",
														this, selection);
													controlObject->setValue(selection);
													break;
												}
											}
										}
									}
								}
							}
						}
						break;

				}	// end switch
			}	// end if
		}	// end while
	}

Exit:
	if (controlsIterator)
	{
		controlsIterator->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::controlHasChangedOnDevice()", this);
}

/****************************************************************************/
/**
 * @brief	registerEngineInfo() で登録された２つのオーディオエンジンを取得する．
 *
 * @param[out]		**firstEngine	１つめのオーディオエンジン．
 * @param[out]		**secondEngine	２つめのオーディオエンジン．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getBothEngines(
	PdjUsbAudioEngine		**firstEngine,
	PdjUsbAudioEngine		**secondEngine
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getBothEngines()");

	IOReturn		result = kIOReturnError;
	SInt32			engineIndex = 0;
	OSDictionary	*firstAudioEngineInfo = NULL;

	FailIf(NULL == mAudioDevice->mRegisteredEngines, Exit);

	firstAudioEngineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(engineIndex));
	// Must stop here if we didn't find this engine's dictionary.
	FailIf(NULL == firstAudioEngineInfo, Exit);

	*firstEngine = OSDynamicCast(PdjUsbAudioEngine, firstAudioEngineInfo->getObject(kEngine));
	FailIf(NULL == *firstEngine, Exit);

	// Now that we have the firstEngine, get the second.
	*secondEngine = this->otherEngine(*firstEngine);
	FailIf(NULL == secondEngine, Exit);

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getBothEngines() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	registerEngineInfo() で登録された２つのオーディオエンジンのうち、
 *			指定されたエンジンでない方を取得する．
 *
 * @param[in]		*thisEngine		既に取得済みのオーディオエンジン．
 *
 * @return			指定されたエンジンでない方のオーディオエンジン．
 */
PdjUsbAudioEngine *
PdjUsbAudioTopology::otherEngine(
	PdjUsbAudioEngine		*thisEngine
)
{
	SInt32					engineIndex;
	SInt32					otherEngineIndex;
	OSDictionary			*otherAudioEngineInfo = NULL;
	PdjUsbAudioEngine		*otherAudioEngine = NULL;

	FailIf(NULL == thisEngine, Exit);

	FailIf(NULL == mAudioDevice->mRegisteredEngines, Exit);

	engineIndex = this->getEngineInfoIndex(thisEngine);
	// Must stop here if we didn't find this engine.
	FailIf(-1 == engineIndex, Exit);

	otherEngineIndex = ((1 == engineIndex) ? 0 : 1);
	// The two engine indeces should be 0 and 1, so we'll NOT this engine to get the other one.

	otherAudioEngineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(otherEngineIndex));
	FailIf(NULL == otherAudioEngineInfo, Exit);

	otherAudioEngine = OSDynamicCast(PdjUsbAudioEngine, otherAudioEngineInfo->getObject(kEngine));
	FailIf(NULL == otherAudioEngine, Exit);

Exit:
	return otherAudioEngine;
}

#pragma mark - Protected Methods -

/****************************************************************************/
/**
 * @brief	オーディオファンクションの接続パスを構築．
 *
 * @param[in]		controlInterfaceNum		USB interface 番号．
 * @param[in]		startingUnitID			パスの起点となるユニット ID．
 * @param[out]		*allPaths				指定した起点からはじまるすべてのパスの OSArray．
 * @param[in]		*startingPath			起点までのパス．
 *
 * @return			オーディオファンクションの接続パス．
 */
OSArray *
PdjUsbAudioTopology::buildPath(
	UInt8			controlInterfaceNum,
	UInt8			startingUnitID,
	OSArray			*allPaths,
	OSArray			*startingPath
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::buildPath()", this);

	OSArray			*curPath = NULL;
	OSNumber		*thisUnitIDNum;
	UInt8			thisUnitID;
	UInt8			subType;
	UInt16			adcVersion;

	FailIf(kIOReturnSuccess != mConfigDictionary->getADCVersion(&adcVersion), Exit);

	thisUnitID = startingUnitID;
	FailIf(NULL == (thisUnitIDNum = OSNumber::withNumber(thisUnitID, 8)), Exit);

	if (NULL != startingPath)
	{
		curPath = OSArray::withArray(startingPath);
	}

	if (NULL == curPath)
	{
		curPath = OSArray::withObjects((const OSObject **)&thisUnitIDNum, 1);

#ifdef BUG_FIX
		FailIf(NULL == curPath, Exit);
#endif
	}
	else
	{
		curPath->setObject(thisUnitIDNum);
	}

	thisUnitIDNum->release();
	thisUnitIDNum = NULL;

	FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, thisUnitID), Exit);

	while ((INPUT_TERMINAL != subType) && (subType != 0))
	{
		if (
				(
					(kPdjUsbUsbSpec1_0 == adcVersion)
				&&
					(
						(MIXER_UNIT == subType)
					||
						(SELECTOR_UNIT == subType)
					||
						(EXTENSION_UNIT == subType)
					||
						(PROCESSING_UNIT == subType)
					)
				)
		   ||
				(
					(kPdjUsbUsbSpec2_0 == adcVersion)
				&&
					(
						(USBAUDIO_0200::MIXER_UNIT == subType)
					||
						(USBAUDIO_0200::SELECTOR_UNIT == subType)
					||
						(USBAUDIO_0200::EXTENSION_UNIT == subType)
					||
						(USBAUDIO_0200::PROCESSING_UNIT == subType)
					)
				)
		   )
		{
			OSArray		*sourceArray = NULL;
			UInt8		numSources;

			FailIf(kIOReturnSuccess != mConfigDictionary->getNumSources(&numSources, controlInterfaceNum, 0, thisUnitID), Exit);
			FailIf(kIOReturnSuccess != mConfigDictionary->getSourceIDs(&sourceArray, controlInterfaceNum, 0, thisUnitID), Exit);

			OSArray		*tempPath = OSArray::withArray(curPath);
#ifdef BUG_FIX
			curPath->release();
			curPath = NULL;
#endif

			for (UInt32 i = 0; i < numSources; i++)
			{
				OSObject	*arrayObject = NULL;
				OSNumber	*arrayNumber = NULL;

#ifdef BUG_FIX
#else
				if (NULL == curPath)
				{
					curPath = OSArray::withCapacity(1);
				}

				FailIf(NULL == curPath, Exit);
#endif
				FailIf(NULL == (arrayObject = sourceArray->getObject(i)), Exit);
				FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);

				curPath = this->buildPath(controlInterfaceNum, arrayNumber->unsigned8BitValue(), allPaths, tempPath);

				if (curPath && curPath->getCount())
				{
					thisUnitIDNum = OSDynamicCast(OSNumber, curPath->getLastObject());
					FailIf(NULL == thisUnitIDNum, Exit);

					UInt8	unitID = thisUnitIDNum->unsigned8BitValue();
					FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

					if (unitID && (subType == INPUT_TERMINAL))
					{
						allPaths->setObject(curPath);
					}
				}

				if (curPath)
				{
					curPath->release();
					curPath = NULL;
				}
			}

			tempPath->release();

			subType = 0;	// while ループを抜ける
		}
		else
		{
			UInt8		sourceID;
			UInt8		startingSubType;

			// OUTPUT_TERMINAL, FEATURE_UNIT, EFFECT_UNIT:
			FailIf(kIOReturnSuccess != mConfigDictionary->getSourceID(&sourceID, controlInterfaceNum, 0, thisUnitID), Exit);

			thisUnitID		= sourceID;
			thisUnitIDNum	= OSNumber::withNumber(thisUnitID, 8);

			if (NULL != thisUnitIDNum)
			{
				curPath->setObject(thisUnitIDNum);
				thisUnitIDNum->release();
				thisUnitIDNum = NULL;
			}

			FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, thisUnitID), Exit);
			FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&startingSubType, controlInterfaceNum, 0, startingUnitID), Exit);

			if ((subType == INPUT_TERMINAL) && (startingSubType == OUTPUT_TERMINAL))
			{
				allPaths->setObject(curPath);
			}
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::buildPath()", this);
	return curPath;
}

/****************************************************************************/
/**
 * @brief	クロックパスを構築．
 *
 * @param[in]		controlInterfaceNum		USB interface 番号．
 * @param[in]		startingUnitID			パスの起点となるユニット ID．
 * @param[out]		*allPaths				指定した起点からはじまるすべてのパスの OSArray．
 * @param[in]		*startingPath			起点までのパス．
 */
void
PdjUsbAudioTopology::buildClockPath(
	UInt8		controlInterfaceNum,
	UInt8		startingUnitID,
	OSArray		*allPaths,
	OSArray		*startingPath
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::buildClockPath(%d, %d, %p, %p)",
		this, controlInterfaceNum, startingUnitID, allPaths, startingPath);

	OSArray		*curPath = NULL;
	OSNumber	*thisUnitIDNum;
	UInt8		thisUnitID;
	UInt8		subType;

	FailIf(NULL == mConfigDictionary, Exit);

	thisUnitID = startingUnitID;
	FailIf(NULL == (thisUnitIDNum = OSNumber::withNumber(thisUnitID, 8)), Exit);

	if (NULL != startingPath)
	{
		curPath = OSArray::withArray(startingPath);
	}

	if (NULL == curPath)
	{
		curPath = OSArray::withObjects((const OSObject **)&thisUnitIDNum, 1);
	}
	else
	{
		curPath->setObject(thisUnitIDNum);
	}

	thisUnitIDNum->release();
	thisUnitIDNum = NULL;

	FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, thisUnitID), Exit);

	while (
			   (0 != subType)
			&& (curPath)
			&& (USBAUDIO_0200::CLOCK_SOURCE != subType)
		  )
	{
		if (USBAUDIO_0200::CLOCK_SELECTOR == subType)
		{
			OSArray		*sourceArray = NULL;
			OSNumber	*arrayNumber = NULL;
			UInt8		numSources;

			DP(DBG_INFO, "    found clock selector @ ID %d", thisUnitID);

			FailIf(kIOReturnSuccess != mConfigDictionary->getNumSources(&numSources, controlInterfaceNum, 0, thisUnitID), Exit);
			DP(DBG_INFO, "    found clock selector %d has %d sources", thisUnitID, numSources);

			FailIf(kIOReturnSuccess != mConfigDictionary->getSourceIDs(&sourceArray, controlInterfaceNum, 0, thisUnitID), Exit);

			for (UInt8 sourceIndex = 0; sourceIndex < numSources; sourceIndex++)
			{
				FailIf(NULL == sourceArray, Exit);
				FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, sourceArray->getObject(sourceIndex))), Exit);

				this->buildClockPath(controlInterfaceNum, arrayNumber->unsigned8BitValue(), allPaths, curPath);
			}

			subType = 0;
		}
		else	// USBAUDIO_0200::CLOCK_MULTIPLIER:
		{
			UInt8		sourceID;

			DP(DBG_INFO, "    found clock multiplier @ ID %d", thisUnitID);
			if (1 != curPath->getCount())
			{
				// We haven't added this yet. We should do so.
				thisUnitIDNum = OSNumber::withNumber(thisUnitID, 8);
				if (NULL != thisUnitIDNum)
				{
					curPath->setObject(thisUnitIDNum);

					thisUnitIDNum->release();
					thisUnitIDNum = NULL;
				}
			}

			// Continue down the path.
			FailIf(kIOReturnSuccess != mConfigDictionary->getSourceID(&sourceID, controlInterfaceNum, 0, thisUnitID), Exit);
			thisUnitID = sourceID;

			FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, thisUnitID), Exit);
		}
	}

	if (USBAUDIO_0200::CLOCK_SOURCE == subType)
	{
		DP(DBG_INFO, "    found clock source @ ID %d", thisUnitID);

		// [rdar://4952145] This is the end of a clock path. We should set it now.
		DP(DBG_INFO, "    adding path...");
		allPaths->setObject(curPath);
	}

Exit:
	if (curPath)
	{
		curPath->release();
		curPath = NULL;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::buildClockPath()", this);
}

/****************************************************************************/
/**
 * @brief	クロックパスが他のオーディオエンジンで使われているクロックパスと重なっていないか調べる．
 *
 * @param[in]		*thisEngine		現在設定中のオーディオエンジン．
 * @param[in]		*thisClockPath	現在設定中のクロックパス．
 *
 * @return			他のオーディオエンジンで使われているクロックパスのうち
 *					このパスと交差しているものの数．
 */
UInt32
PdjUsbAudioTopology::determineClockPathUnitUsage(
	PdjUsbAudioEngine		*thisEngine,
	OSArray					*thisClockPath
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::determineClockPathUnitUsage(%p, %p)", this, thisEngine, thisClockPath);

	UInt32		usageCount = 0;

	if (NULL != mAudioDevice->mRegisteredEngines)
	{
		for (UInt32 engineIndex = 0; ; engineIndex++)
		{
			OSDictionary			*otherAudioEngineInfo;
			PdjUsbAudioEngine		*otherAudioEngine;

			if (NULL == (otherAudioEngineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(engineIndex))))
			{
				break;
			}

			if (NULL == (otherAudioEngine = OSDynamicCast(PdjUsbAudioEngine, otherAudioEngineInfo->getObject(kEngine))))
			{
				break;
			}

			if (thisEngine != otherAudioEngine)
			{
				if (NULL != otherAudioEngine->mIOAudioStreamArray)
				{
					for (UInt32 index = 0; index < (otherAudioEngine->mIOAudioStreamArray->getCount()); index++)
					{
						PdjUsbAudioStream	*pdjUsbAudioStream = OSDynamicCast(
																	PdjUsbAudioStream,
																	otherAudioEngine->mIOAudioStreamArray->getObject(index)
																 );

						if ((NULL != pdjUsbAudioStream) && (NULL != pdjUsbAudioStream->mActiveClockPath))
						{
							if (this->clockPathCrossed(thisClockPath, pdjUsbAudioStream->mActiveClockPath))
							{
								usageCount++;
							}
						}
					}
				}
			}
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::determineClockPathUnitUsage(%p, %p) = %u",
		this, thisEngine, thisClockPath, usageCount);
	return usageCount;
}

/****************************************************************************/
/**
 * @brief	クロックパスが交差しているか調べる．
 *
 * @param[in]		*clockPathA		クロックパス（Ａ）．
 * @param[in]		*clockPathB		クロックパス（Ｂ）．
 *
 * @retval			true			クロックパスは交差している．
 * @retval			false			クロックパスは交差していない．
 */
bool
PdjUsbAudioTopology::clockPathCrossed(
	OSArray			*clockPathA,
	OSArray			*clockPathB
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::clockPathCrossed()", this);

	OSNumber		*clockIDNumberA = NULL;
	OSNumber		*clockIDNumberB = NULL;
	bool			pathCrossed		= false;

	// Determine it the 2 paths crossed.
	for (UInt8 pathItemA = 0; pathItemA < clockPathA->getCount(); pathItemA++)
	{
		FailIf(NULL == (clockIDNumberA = OSDynamicCast(OSNumber, clockPathA->getObject(pathItemA))), Exit);

		for (UInt8 pathItemB = 0; pathItemB < clockPathB->getCount(); pathItemB++)
		{
			FailIf(NULL == (clockIDNumberB = OSDynamicCast(OSNumber, clockPathB->getObject(pathItemB))), Exit);

			if (clockIDNumberA->unsigned8BitValue() == clockIDNumberB->unsigned8BitValue())
			{
				pathCrossed = true;
				break;
			}
		}

		if (pathCrossed)
		{
			break;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::clockPathCrossed() = %d", this, pathCrossed);
	return pathCrossed;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface のリスト中にクロックが共通の要素があるか調べる．
 *
 * @param[in]		*refStreamInterfaceNumber	基準とする AudioStreaming Interface の interface 番号．
 * @param[in]		*streamInterfaceNumberList	検索対象となる AudioStreaming Interface の OSArray．
 *
 * @retval			true						リスト中にクロックが共通の要素が１つ以上存在する．
 * @retval			false						リスト中にクロックが共通の要素が１つも存在しない．
 */
// <rdar://5131786> Determine if the stream's clocks are common with at least one other stream in the stream list.
bool
PdjUsbAudioTopology::isClockCommonWithAtLeastOneStreamsInList(
	OSNumber		*refStreamInterfaceNumber,
	OSArray			*streamInterfaceNumberList
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::isClockCommonWithAtLeastOneStreamsInList()", this);

	bool	result = false;

	for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < streamInterfaceNumberList->getCount(); streamInterfaceIndex++)
	{
		OSNumber	*streamInterfaceNumber = OSDynamicCast(OSNumber, streamInterfaceNumberList->getObject(streamInterfaceIndex));
		FailIf(NULL == streamInterfaceNumber, Exit);

		if (this->streamsHaveCommonClocks(refStreamInterfaceNumber, streamInterfaceNumber))
		{
			result = true;
			break;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::isClockCommonWithAtLeastOneStreamsInList() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface のリスト中にクロックが共通でない要素があるか調べる．
 *
 * @param[in]		*refStreamInterfaceNumber	基準とする AudioStreaming Interface の interface 番号．
 * @param[in]		*streamInterfaceNumberList	検索対象となる AudioStreaming Interface の OSArray．
 *
 * @retval			true						リスト中の全要素とクロックが共通である．
 * @retval			false						リスト中にクロックが共通でない要素が１つ以上存在する．
 */
// <rdar://5131786> Determine if the stream's clocks are common with all other streams in the stream list.
bool
PdjUsbAudioTopology::isClockCommonWithAllStreamsInList(
	OSNumber		*refStreamInterfaceNumber,
	OSArray			*streamInterfaceNumberList
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::isClockCommonWithAllStreamsInList()", this);

	bool	result = true;

	for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < streamInterfaceNumberList->getCount(); streamInterfaceIndex++)
	{
		OSNumber	*streamInterfaceNumber = OSDynamicCast(OSNumber, streamInterfaceNumberList->getObject(streamInterfaceIndex));
		FailIf(NULL == streamInterfaceNumber, Exit);

		if (!this->streamsHaveCommonClocks(refStreamInterfaceNumber, streamInterfaceNumber))
		{
			result = false;
			break;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::isClockCommonWithAllStreamsInList() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	２つの AudioStreaming Interface のクロックが共通であるか調べる．
 *
 * @param[in]		*streamInterfaceNumberA		比較対象となる AudioStreaming Interface の interface 番号（Ａ）．
 * @param[in]		*streamInterfaceNumberB		比較対象となる AudioStreaming Interface の interface 番号（Ｂ）．
 *
 * @retval			true						クロックは共通である．
 * @retval			false						クロックは共通でない．
 */
// <rdar://5131786> Determine if the stream's clocks are common with each other.
bool
PdjUsbAudioTopology::streamsHaveCommonClocks(
	OSNumber		*streamInterfaceNumberA,
	OSNumber		*streamInterfaceNumberB
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::streamsHaveCommonClocks(%p, %p)", this, streamInterfaceNumberA, streamInterfaceNumberB);

	UInt8			interfaceNum;
	UInt32			numClockPathSupported = 0;
	UInt32			numClockPathCrossed = 0;
	UInt8			numAltSettings = 0;
	bool			result = false;

	FailIf(NULL == streamInterfaceNumberA, Exit);
	FailIf(NULL == streamInterfaceNumberB, Exit);

	DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::streamsHaveCommonClocks(%d, %d)",
		this, streamInterfaceNumberA->unsigned8BitValue(), streamInterfaceNumberB->unsigned8BitValue());

	interfaceNum = streamInterfaceNumberB->unsigned8BitValue();

	if (kIOReturnSuccess == mConfigDictionary->getNumAltSettings(&numAltSettings, interfaceNum))
	{
		bool	startAtZero = mConfigDictionary->alternateSettingZeroCanStream(interfaceNum);

		for (
				UInt8 altSetting = (startAtZero ? 0 : 1);
				altSetting < numAltSettings;
				altSetting++
			)
		{
			OSArray		*rates = mConfigDictionary->getSampleRates(interfaceNum, altSetting);

			if (NULL != rates)
			{
				OSArray		*clockPathGroup = this->getClockPathGroup(interfaceNum, altSetting);
				FailIf(NULL == clockPathGroup, Exit);

				for (UInt8 rateIndex = 0; rateIndex < rates->getCount(); rateIndex++)
				{
					OSNumber	*rate = OSDynamicCast(OSNumber, rates->getObject(rateIndex));

					if (NULL != rate)
					{
						// For each path in the path group, determine if it crosses path with the other stream interface.
						for (UInt8 pathIndex = 0; pathIndex < clockPathGroup->getCount(); pathIndex++)
						{
							OSArray		*clockPath;
							FailIf(NULL == (clockPath = OSDynamicCast(OSArray, clockPathGroup->getObject(pathIndex))), Exit);

							if (this->supportSampleRateInClockPath(clockPath, rate->unsigned32BitValue()))
							{
								numClockPathSupported ++;

								if (this->isClockPathCrossed(streamInterfaceNumberA, clockPath, rate->unsigned32BitValue()))
								{
									numClockPathCrossed ++;
								}
								else
								{
									break;
								}
							}
						}
					}

					if (numClockPathSupported != numClockPathCrossed)
					{
						break;
					}
				}
			}

			if (numClockPathSupported != numClockPathCrossed)
			{
				break;
			}
		}
	}

	result = (0 != numClockPathSupported) && (0 != numClockPathCrossed) && (numClockPathSupported == numClockPathCrossed);

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::streamsHaveCommonClocks(%p, %p) = %d", this, streamInterfaceNumberA, streamInterfaceNumberB, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックパスが交差しているか調べる．
 *
 * @param[in]		*streamInterfaceNumber		AudioStreaming Interface の interface 番号．
 * @param[in]		*otherClockPath				比較対象となるクロックパス．
 * @param[in]		sampleRate					比較条件となるサンプリング周波数．
 *
 * @retval			true						クロックパスは交差している．
 * @retval			false						クロックパスは交差していない．
 */
// <rdar://5131786> Check if the specified clock path crosses with clock path in the first interface.
bool
PdjUsbAudioTopology::isClockPathCrossed(
	OSNumber		*streamInterfaceNumber,
	OSArray			*otherClockPath,
	UInt32			sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::isClockPathCrossed(%p, %p, %u)", this, streamInterfaceNumber, otherClockPath, sampleRate);

	UInt8			interfaceNum;
	UInt8			numAltSettings = 0;
	UInt32			numClockPathSupported = 0;
	UInt32			numClockPathCrossed = 0;
	bool			result = false;

	// Check if the clock path against the first stream interface to determine if they have a common clock source.
	interfaceNum = streamInterfaceNumber->unsigned8BitValue();

	if (kIOReturnSuccess == mConfigDictionary->getNumAltSettings(&numAltSettings, interfaceNum))
	{
		bool	startAtZero = mConfigDictionary->alternateSettingZeroCanStream(interfaceNum);

		for (UInt8 altSetting = (startAtZero ? 0 : 1); altSetting < numAltSettings; altSetting++)
		{
			OSArray		*clockPathGroup = this->getClockPathGroup(interfaceNum, altSetting);

			if (NULL != clockPathGroup)
			{
				// For each path in the path group, determine if it supported the requested sample rate.
				for (UInt8 pathIndex = 0; pathIndex < clockPathGroup->getCount(); pathIndex++)
				{
					OSArray		*thisClockPath;

					FailIf(NULL == (thisClockPath = OSDynamicCast(OSArray, clockPathGroup->getObject(pathIndex))), Exit);

					if (this->supportSampleRateInClockPath(thisClockPath, sampleRate))
					{
						numClockPathSupported++;

						if (this->clockPathCrossed(thisClockPath, otherClockPath))
						{
							numClockPathCrossed++;
						}
						else
						{
							break;
						}
					}
				}

				if (numClockPathSupported != numClockPathCrossed)
				{
					break;
				}
			}
		}
	}

	result = (0 != numClockPathSupported) && (0 != numClockPathCrossed) && (numClockPathSupported == numClockPathCrossed);

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::isClockPathCrossed(%p, %p, %u) = %d", this, streamInterfaceNumber, otherClockPath, sampleRate, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	指定のクロックパスにおいて指定のサンプリング周波数がサポートされているか調べる．
 *
 * @param[in]		*pathArray					クロックパス．
 * @param[in]		sampleRate					サンプリング周波数．
 *
 * @retval			true						サポートされている．
 * @retval			false						サポートされていない．
 */
bool
PdjUsbAudioTopology::supportSampleRateInClockPath(
	OSArray			*pathArray,
	UInt32			sampleRate
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::supportSampleRateInClockPath()", this);

	bool			sampleRateSupported = false;
	UInt8			numRange;

	FailIf(kIOReturnSuccess != this->getNumSampleRatesForClockPath(&numRange, pathArray), Exit);

	for (UInt8 rangeIndex = 0; rangeIndex < numRange; rangeIndex++)
	{
		SubRange32		subRange;

		if (kIOReturnSuccess == this->getIndexedSampleRatesForClockPath(&subRange, pathArray, rangeIndex))
		{
			// If dRES is not zero, there should be a total of 1 + ( ( dMAX - dMIN ) / dRES ) sample rates including dMIN and dMAX.
			if (0 != subRange.dRES)
			{
				// [rdar://5614104] Correct the boundary condition below.
				// sampleRateIndex should iterate from 0 to one less than the total number of sample rates.
				for (
						UInt16 sampleRateIndex = 0;
						sampleRateIndex <= ((subRange.dMAX - subRange.dMIN) / subRange.dRES);
						sampleRateIndex++
					)	// <rdar://7446555>
				{
					// [rdar://5614104] Correct the conditional below based on the above change.
					if (sampleRate == (subRange.dMIN + (sampleRateIndex * subRange.dRES)))
					{
						sampleRateSupported = true;
						break;
					}
				}
			}
			else
			{
				if ((sampleRate == subRange.dMIN) || (sampleRate == subRange.dMAX))
				{
					sampleRateSupported = true;
				}
			}
		}

		if (sampleRateSupported)
		{
			break;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::supportSampleRateInClockPath() = %d", this, sampleRateSupported);
	return sampleRateSupported;
}

/****************************************************************************/
/**
 * @brief	指定のクロックパスにおいて設定可能なサンプリング周波数の個数を取得．
 *
 * @param[out]		*numSampleRates				設定可能なサンプリング周波数のサブレンジ数．
 * @param[in]		*clockPath					クロックパス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getNumSampleRatesForClockPath(
	UInt8			*numSampleRates,
	OSArray			*clockPath
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getNumSampleRatesForClockPath()", this);

	OSObject		*arrayObject = NULL;
	OSNumber		*arrayNumber = NULL;
	UInt8			clockID;
	UInt8			subType;
	UInt16			numSubRanges = 0;
	IOReturn		result		 = kIOReturnError;

	FailIf(NULL == numSampleRates, Exit);
	FailIf(NULL == mControlInterface, Exit);

	// The last object in the clock path is the clock source.
	FailIf(NULL == (arrayObject = clockPath->getLastObject()), Exit);
	FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);
	clockID = arrayNumber->unsigned8BitValue();

	FailIf(kIOReturnSuccess != (result = mConfigDictionary->getSubType(&subType, mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, clockID)), Exit);

	if (USBAUDIO_0200::CLOCK_SOURCE == subType)
	{
		FailIf(kIOReturnSuccess != (result = this->getNumClockSourceSamplingFrequencySubRanges(clockID, &numSubRanges)), Exit);
	}

Exit:
	if (NULL != numSampleRates)
	{
		*numSampleRates = numSubRanges;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getNumSampleRatesForClockPath() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックパスからサンプリング周波数を取り出す．
 *
 * @param[out]		*sampleRates	取り出したサンプリング周波数のサブレンジ．
 * @param[in]		*clockPath		クロックパス．
 * @param[in]		rangeIndex		クロックソースに設定可能なサンプリング周波数のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getIndexedSampleRatesForClockPath(
	SubRange32		*sampleRates,
	OSArray			*clockPath,
	UInt32			rangeIndex
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getIndexedSampleRatesForClockPath()", this);

	IOReturn		result = kIOReturnError;
	SubRange32		subRange;

	FailIf(NULL == mControlInterface, Exit);		// <rdar://7085810>
	FailIf(NULL == sampleRates, Exit);

	subRange.dMIN = subRange.dMAX = subRange.dRES = 0;

	for (UInt32 clockIndex = clockPath->getCount(); clockIndex > 0; clockIndex--)
	{
		OSObject	*arrayObject = NULL;
		OSNumber	*arrayNumber = NULL;
		UInt8		clockID;
		UInt8		subType;
		UInt16		numerator;
		UInt16		denominator;

		FailIf(NULL == (arrayObject = clockPath->getObject(clockIndex - 1)), Exit);
		FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);
		clockID = arrayNumber->unsigned8BitValue();

		FailIf(
			kIOReturnSuccess != (
									result = mConfigDictionary->getSubType(
																	&subType,
																	mControlInterface->getInterfaceDescriptor()->bInterfaceNumber,
																	0,
																	clockID
																)
								),
			Exit
		);

		if (USBAUDIO_0200::CLOCK_SOURCE == subType)
		{
			FailIf(
				kIOReturnSuccess != (
										result = this->getIndexedClockSourceSamplingFrequencySubRange(
															clockID,
															&subRange,
															rangeIndex
													   )
									),
				Exit
			);
		}
		else if (USBAUDIO_0200::CLOCK_MULTIPLIER == subType)
		{
			FailIf(
				kIOReturnSuccess != (
										result = this->getCurClockMultiplier(
															clockID,
															&numerator,
															&denominator
													   )
									),
				Exit
			);

			subRange.dMIN = subRange.dMIN * numerator / denominator;
			subRange.dMAX = subRange.dMAX * numerator / denominator;
			subRange.dRES = subRange.dRES * numerator / denominator;
		}
	}

Exit:
	if (NULL != sampleRates)
	{
		*sampleRates = subRange;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getIndexedSampleRatesForClockPath() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックパスを走査し、設定可能なサンプリング周波数を列挙．
 *
 * @param[in]		*path				クロックパス．
 * @param[in]		streamInterface		AudioStreaming Interface の interface 番号．
 * @param[in]		altSetting			alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::addSampleRatesFromClockPath(
	OSArray			*path,
	UInt8			streamInterface,
	UInt8			altSetting
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::addSampleRatesFromClockPath(%p, %d, %d)", this, path, streamInterface, altSetting);

	IOReturn		result = kIOReturnError;
	OSNumber		*clockSourceNumber;
	OSArray			*sampleRates = NULL;

	FailIf(NULL == path, Exit);
	FailIf(NULL == mConfigDictionary, Exit);

	// First get the samples rates on the clock source, which will be the last element of the path.
	FailIf(NULL == (clockSourceNumber = OSDynamicCast(OSNumber, path->getLastObject())), Exit);
	FailIf(kIOReturnSuccess != this->getClockSourceSampleRates(&sampleRates, clockSourceNumber->unsigned8BitValue()), Exit);

	// Clock multipliers will modify the available sample rates, so search for them in the path.
	// [rdar://5600081] Add sample rates via clock multipliers
	for (UInt32 clockIndex = path->getCount(); clockIndex > 0; clockIndex--)
	{
		OSObject	*arrayObject = NULL;
		OSNumber	*arrayNumber = NULL;
		UInt8		clockID;
		UInt8		subType;

		FailIf(NULL == (arrayObject = path->getObject(clockIndex - 1)), Exit);
		FailIf(NULL == (arrayNumber = OSDynamicCast(OSNumber, arrayObject)), Exit);
		clockID = arrayNumber->unsigned8BitValue();

		FailIf(kIOReturnSuccess != (result = mConfigDictionary->getSubType(&subType, mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, clockID)), Exit);

		if (USBAUDIO_0200::CLOCK_MULTIPLIER == subType)
		{
			OSArray		*rates;
			UInt16		numerator;
			UInt16		denominator;

			FailIf(NULL == sampleRates, Exit);
			FailIf(kIOReturnSuccess != (result = this->getCurClockMultiplier(clockID, &numerator, &denominator)), Exit);

			rates = OSArray::withCapacity(sampleRates->getCount());
			FailIf(NULL == rates, Exit);

			for (UInt32 rateIndex = 0; rateIndex < sampleRates->getCount(); rateIndex++)
			{
				OSNumber	*sampleRateNumber;
				OSNumber	*rateNumber;

				sampleRateNumber = OSDynamicCast(OSNumber, sampleRates->getObject(rateIndex));
				FailIf(NULL == sampleRateNumber, Exit);

				rateNumber = OSNumber::withNumber(sampleRateNumber->unsigned32BitValue() * numerator / denominator, 32);
				FailIf(NULL == rateNumber, Exit);

				rates->setObject(rateNumber);
#ifdef BUG_FIX
				rateNumber->release();
#else
				rates->release();
#endif
			}

			sampleRates->release();
			sampleRates = rates;
		}
	}

	// Add the list of discrete sample rates to the stream dictionary
	FailIf(kIOReturnSuccess != (result = mConfigDictionary->addSampleRatesToStreamDictionary(sampleRates, streamInterface, altSetting)), Exit);

Exit:
	if (NULL != sampleRates)
	{
		sampleRates->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::addSampleRatesFromClockPath(%p, %d, %d) = %d", this, path, streamInterface, altSetting, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースに設定可能なサンプリング周波数の OSArray を取得．
 *
 * @param[out]		**sampleRates		クロックソースに設定可能なサンプリング周波数の OSArray．
 * @param[in]		clockSource			クロックソース ID．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getClockSourceSampleRates(
	OSArray			**sampleRates,
	UInt8			clockSource
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getClockSourceSampleRates(%p, %d)", this, sampleRates, clockSource);

	IOReturn		result = kIOReturnError;
	OSNumber		*sampleRateNumber = NULL;
	UInt32			sampleRate;

	FailIf(NULL == mControlInterface, Exit);
	FailIf(NULL == mConfigDictionary, Exit);
	FailIf(NULL == sampleRates, Exit);
	FailIf(0 == clockSource, Exit);

	*sampleRates = NULL;

	if (
			mConfigDictionary->clockSourceHasFrequencyControl(mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, clockSource, true)
	   ||
			mConfigDictionary->clockSourceHasFrequencyControl(mControlInterface->getInterfaceDescriptor()->bInterfaceNumber, 0, clockSource, false)
	   )
	{
		UInt16		numSampleRanges;

		// Get the number of sample ranges.
		FailIf(kIOReturnSuccess != (result = this->getNumClockSourceSamplingFrequencySubRanges(clockSource, &numSampleRanges)), Exit);

		for (UInt8 sampleRangeIndex = 0; sampleRangeIndex < numSampleRanges; sampleRangeIndex++)
		{
			SubRange32		subRange;
			bool			subRangeIsValid;

			FailIf(kIOReturnSuccess != (result = this->getIndexedClockSourceSamplingFrequencySubRange(clockSource, &subRange, sampleRangeIndex)), Exit);

			// [rdar://4952486] Make sure that the device isn't just spewing garbage before trying to add these values.
			subRangeIsValid = (
									(subRange.dMIN <= subRange.dMAX)
							  &&
									(
										(0 == subRange.dRES)
									||
										(0 == (subRange.dMAX - subRange.dMIN) % subRange.dRES)
									)
							  );

			if (!subRangeIsValid)
			{
				DP(DBG_WARNING, "! " LOGSTRAUDIOTOPOLOGY "[%p]::getClockSourceSampleRates() - invalid subrange, skipping ...", this);
				DP(DBG_WARNING, "    subRange.dMIN = %u", subRange.dMIN);
				DP(DBG_WARNING, "    subRange.dMAX = %u", subRange.dMAX);
				DP(DBG_WARNING, "    subRange.dRES = %u", subRange.dRES);
				continue;
			}

			// Add the minimum sample rate no matter what.
			FailWithAction(NULL == (sampleRateNumber = OSNumber::withNumber(subRange.dMIN, SIZEINBITS(UInt32))), result = kIOReturnError, Exit);
			if (*sampleRates)
			{
				(*sampleRates)->setObject(sampleRateNumber);
			}
			else
			{
				*sampleRates = OSArray::withObjects((const OSObject **)&sampleRateNumber, 1);
			}

			sampleRateNumber->release();
			sampleRateNumber = NULL;

			// Add the "middle" sample rates if necessary.
			// If dRES is not zero, there should be a total of 1 + ( ( dMAX - dMIN ) / dRES ) sample rates including dMIN and dMAX.
			if (0 != subRange.dRES)
			{
				for (UInt16 sampleRateIndex = 0; sampleRateIndex < ((subRange.dMAX - subRange.dMIN) / subRange.dRES) - 1; sampleRateIndex++)
				{
					sampleRate = subRange.dMIN + (1 + sampleRateIndex) * subRange.dRES;

					FailWithAction(
						NULL == (sampleRateNumber = OSNumber::withNumber(sampleRate, SIZEINBITS(UInt32))),
						result = kIOReturnError,
						Exit
					);

					(*sampleRates)->setObject(sampleRateNumber);
					sampleRateNumber->release();
					sampleRateNumber = NULL;
				}
			}

			// Add the maximum sample rate if necessary.
			if (subRange.dMAX != subRange.dMIN)
			{
				FailWithAction(
					NULL == (sampleRateNumber = OSNumber::withNumber(subRange.dMAX, SIZEINBITS(UInt32))),
					result = kIOReturnError,
					Exit
				);

				(*sampleRates)->setObject(sampleRateNumber);
				sampleRateNumber->release();
				sampleRateNumber = NULL;
			}
		}
	}
	else
	{
		bool	clockIsValid;

		// There is no frequency control. Get the current sample rate only.
		FailIf(kIOReturnSuccess != (result = this->getCurClockSourceSamplingFrequency(clockSource, &sampleRate, &clockIsValid)), Exit);

		FailWithAction(NULL == (sampleRateNumber = OSNumber::withNumber(sampleRate, SIZEINBITS(UInt32))), result = kIOReturnError, Exit);

		*sampleRates = OSArray::withObjects((const OSObject **)&sampleRateNumber, 1);

		sampleRateNumber->release();
		sampleRateNumber = NULL;
	}

	// All sample rates have been added to the array. We are done.
	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getClockSourceSampleRates(%p, %d) = %d", this, sampleRates, clockSource, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースに設定可能なサンプリング周波数のサブレンジ数を取得．
 *
 * @param[in]		unitID			クロックソース ID．
 * @param[out]		*numSubRanges	設定可能なサンプリング周波数のサブレンジ数．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getNumClockSourceSamplingFrequencySubRanges(
	UInt8			unitID,
	UInt16			*numSubRanges
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getNumClockSourceSamplingFrequencySubRanges()", this);

	IOReturn		result = kIOReturnError;

	struct {
		UInt16		wNumSubRanges;
		SubRange32	SubRanges[1];
	}				rangeParameterBlock;

	rangeParameterBlock.wNumSubRanges = 0;

	FailIf(NULL == numSubRanges, Exit);

	result = this->getClockSetting(
						USBAUDIO_0200::CS_SAM_FREQ_CONTROL,
						unitID,
						USBAUDIO_0200::RANGE,
						&rangeParameterBlock,
						sizeof(rangeParameterBlock)
				   );
	FailIf(kIOReturnSuccess != result, Exit);

Exit:
	if (NULL != numSubRanges)
	{
		*numSubRanges = USBToHostWord(rangeParameterBlock.wNumSubRanges);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getNumClockSourceSamplingFrequencySubRanges() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソースからサンプリング周波数を取り出す．
 *
 * @param[in]		unitID			クロックソース ID．
 * @param[out]		*subRange		取り出したサンプリング周波数のサブレンジ．
 * @param[in]		subRangeIndex	設定可能なサンプリング周波数のインデックス．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getIndexedClockSourceSamplingFrequencySubRange(
	UInt8			unitID,
	SubRange32		*subRange,
	UInt16			subRangeIndex
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getIndexedClockSourceSamplingFrequencySubRange()", this);

	struct {
		UInt16		wNumSubRanges;
		SubRange32	SubRanges[1];
	}				rangeParameterBlock;

	void			*theRangeParameterBlock = NULL;
	UInt32			theRangeParameterBlockLength = 0;
	SubRange32		*theSubRanges;
	IOReturn		result = kIOReturnError;

	FailIf(NULL == subRange, Exit);

	rangeParameterBlock.wNumSubRanges = 0;

	FailIf(
		kIOReturnSuccess != (
								this->getClockSetting(
											USBAUDIO_0200::CS_SAM_FREQ_CONTROL,
											unitID,
											USBAUDIO_0200::RANGE,
											&rangeParameterBlock,
											sizeof(rangeParameterBlock)
									  )
							),
		Exit
	);

	rangeParameterBlock.wNumSubRanges = USBToHostWord(rangeParameterBlock.wNumSubRanges);

	FailIf(subRangeIndex >= rangeParameterBlock.wNumSubRanges, Exit);

	theRangeParameterBlockLength = 2 + (rangeParameterBlock.wNumSubRanges * sizeof(SubRange32));

	theRangeParameterBlock = IOMalloc(theRangeParameterBlockLength);
	FailIf(NULL == theRangeParameterBlock, Exit);

	result = this->getClockSetting(
						USBAUDIO_0200::CS_SAM_FREQ_CONTROL,
						unitID,
						USBAUDIO_0200::RANGE,
						theRangeParameterBlock,
						theRangeParameterBlockLength
				   );
	FailIf(kIOReturnSuccess != result, Exit);

	theSubRanges = (SubRange32 *)(((UInt8 *)theRangeParameterBlock) + 2);

	subRange->dMIN = USBToHostLong(theSubRanges[subRangeIndex].dMIN);
	subRange->dMAX = USBToHostLong(theSubRanges[subRangeIndex].dMAX);
	subRange->dRES = USBToHostLong(theSubRanges[subRangeIndex].dRES);

Exit:
	if (NULL != theRangeParameterBlock)
	{
		IOFree(theRangeParameterBlock, theRangeParameterBlockLength);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getIndexedClockSourceSamplingFrequencySubRange() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックの逓倍率・分周率を取得．
 *
 * @param[in]		unitID			クロック逓倍／分周器のユニット ID．
 * @param[out]		*numerator		取得した逓倍率．
 * @param[out]		*denominator	取得した分周率．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getCurClockMultiplier(
	UInt8			unitID,
	UInt16			*numerator,
	UInt16			*denominator
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getCurClockMultiplier()", this);

	UInt16			clockNumerator = 0;
	UInt16			clockDenominator = 0;
	IOReturn		result;

	result = this->getClockSetting(
						USBAUDIO_0200::CM_NUMERATOR_CONTROL,
						unitID,
						USBAUDIO_0200::CUR,
						&clockNumerator,
						2
				   );
	FailIf(kIOReturnSuccess != result, Exit);

	result = this->getClockSetting(
						USBAUDIO_0200::CM_DENOMINATOR_CONTROL,
						unitID,
						USBAUDIO_0200::CUR,
						&clockDenominator,
						2
				   );
	FailIf(kIOReturnSuccess != result, Exit);

Exit:
	if (NULL != numerator)
	{
		*numerator = USBToHostWord(clockNumerator);
	}

	if (NULL != denominator)
	{
		*denominator = USBToHostWord(clockDenominator);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getCurClockMultiplier() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	クロックソース設定．
 *			クロックソースに SET 系デバイスリクエストを発行．
 *
 * @param[in]		controlSelector		コントロールセレクタ．
 * @param[in]		unitID				クロックソース ID．
 * @param[in]		requestType			bRequest フィールド．
 * @param[in]		*target				設定する値．
 * @param[in]		length				バッファサイズ．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::setClockSetting(
	UInt8			controlSelector,
	UInt8			unitID,
	UInt8			requestType,
	void			*target,
	UInt16			length
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::setClockSetting()");

	IOReturn					result = kIOReturnError;
	StandardUSB::DeviceRequest	devReq;
	IOBufferMemoryDescriptor	*theSettingDesc = NULL;

	FailIf(NULL == mControlInterface, Exit);

	theSettingDesc = IOBufferMemoryDescriptor::withBytes(target, length, kIODirectionOut);
	FailIf(NULL == theSettingDesc, Exit);

	devReq.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBClass, kUSBInterface);
	devReq.bRequest		 = requestType;
	devReq.wValue		 = (controlSelector << 8);
	devReq.wIndex		 = (0xFF00 & (unitID << 8)) | (0x00FF & mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
	devReq.wLength		 = length;

	FailIf(mAudioDevice->isInactive(), DeviceInactive);	// In case we've been unplugged during sleep

	result = mAudioDevice->deviceRequest(&devReq, theSettingDesc);

Exit:
	if (NULL != theSettingDesc)
	{
		theSettingDesc->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::setClockSetting() = %d", result);
	return result;

DeviceInactive:
	DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "::setClockSetting() - ERROR attempt to send a device request to an inactive device");
	goto Exit;
}

/****************************************************************************/
/**
 * @brief	クロックソース設定取得．
 *			クロックソースに GET 系デバイスリクエストを発行．
 *
 * @param[in]		controlSelector		コントロールセレクタ．
 * @param[in]		unitID				クロックソース ID．
 * @param[in]		requestType			bRequest フィールド．
 * @param[out]		*target				取得した設定値．
 * @param[in]		length				バッファサイズ．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getClockSetting(
	UInt8			controlSelector,
	UInt8			unitID,
	UInt8			requestType,
	void			*target,
	UInt16			length
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getClockSetting()", this);

	IOReturn					result = kIOReturnError;
	StandardUSB::DeviceRequest	devReq;
	IOBufferMemoryDescriptor	*theSettingDesc = NULL;

	FailIf(NULL == target, Exit);
	FailIf(NULL == mControlInterface, Exit);

	theSettingDesc = IOBufferMemoryDescriptor::withOptions(kIODirectionIn, length);
	FailIf(NULL == theSettingDesc, Exit);

	devReq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBClass, kUSBInterface);
	devReq.bRequest		 = requestType;
	devReq.wValue		 = (controlSelector << 8);
	devReq.wIndex		 = (0xFF00 & (unitID << 8)) | (0x00FF & mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
	devReq.wLength		 = length;

	result = mAudioDevice->deviceRequest(&devReq, theSettingDesc);
	FailIf(kIOReturnSuccess != result, Exit);

	if (NULL != target)
	{
		::memcpy(target, theSettingDesc->getBytesNoCopy(), length);
	}

Exit:
	if (NULL != theSettingDesc)
	{
		theSettingDesc->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getClockSetting() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ミュート状態設定．
 *
 * @param[in]		unitID				Feature Unit のユニット ID．
 * @param[in]		channelNumber		対象チャンネル番号．
 * @param[in]		mute				設定する値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::setCurMute(
	UInt8			unitID,
	UInt8			channelNumber,
	SInt16			mute
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::setCurMute()");

	IOReturn		result;

	result = this->setFeatureUnitSetting(MUTE_CONTROL, unitID, channelNumber, SET_CUR, mute, 1);

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::setCurMute() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ミュート状態取得．
 *
 * @param[in]		unitID				Feature Unit のユニット ID．
 * @param[in]		channelNumber		対象チャンネル番号．
 * @param[out]		*target				取得した値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getCurMute(
	UInt8			unitID,
	UInt8			channelNumber,
	SInt16			*target
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getCurMute()");

	IOReturn		result = kIOReturnError;

	FailIf(NULL == mControlInterface, Exit);

	if (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	{
		result = this->getFeatureUnitSetting(MUTE_CONTROL, unitID, channelNumber, USBAUDIO_0200::CUR, target);
	}
	else
	{
		result = this->getFeatureUnitSetting(MUTE_CONTROL, unitID, channelNumber, GET_CUR, target);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getCurMute() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ボリュームの現在値設定．
 *
 * @param[in]		unitID				Feature Unit のユニット ID．
 * @param[in]		channelNumber		対象チャンネル番号．
 * @param[in]		volume				設定する値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::setCurVolume(
	UInt8			unitID,
	UInt8			channelNumber,
	SInt16			volume
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::setCurVolume()");

	IOReturn		result;

	result = this->setFeatureUnitSetting(VOLUME_CONTROL, unitID, channelNumber, SET_CUR, volume, 2);

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::setCurVolume() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ボリュームの現在値取得．
 *
 * @param[in]		unitID				Feature Unit のユニット ID．
 * @param[in]		channelNumber		対象チャンネル番号．
 * @param[out]		*target				取得した現在値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getCurVolume(
	UInt8			unitID,
	UInt8			channelNumber,
	SInt16			*target
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getCurVolume()");

	IOReturn		result = kIOReturnError;

	FailIf(NULL == mControlInterface, Exit);

	if (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	{
		result = this->getFeatureUnitSetting(VOLUME_CONTROL, unitID, channelNumber, USBAUDIO_0200::CUR, target);
	}
	else
	{
		result = this->getFeatureUnitSetting(VOLUME_CONTROL, unitID, channelNumber, GET_CUR, target);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getCurVolume() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ボリュームの最小値取得．
 *
 * @param[in]		unitID				Feature Unit のユニット ID．
 * @param[in]		channelNumber		対象チャンネル番号．
 * @param[out]		*target				取得した最小値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getMinVolume(
	UInt8			unitID,
	UInt8			channelNumber,
	SInt16			*target
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getMinVolume()");

	IOReturn		result = kIOReturnError;

	FailIf(NULL == mControlInterface, Exit);

	if (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	{
		SubRange16	subRange;

		result = this->getFeatureUnitRange(VOLUME_CONTROL, unitID, channelNumber, USBAUDIO_0200::RANGE, &subRange);

		if (target)
		{
			::memcpy(target, &subRange.wMIN, 2);
		}
	}
	else
	{
		result = this->getFeatureUnitSetting(VOLUME_CONTROL, unitID, channelNumber, GET_MIN, target);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getMinVolume() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ボリュームの最大値取得．
 *
 * @param[in]		unitID				Feature Unit のユニット ID．
 * @param[in]		channelNumber		対象チャンネル番号．
 * @param[out]		*target				取得した最大値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getMaxVolume(
	UInt8			unitID,
	UInt8			channelNumber,
	SInt16			*target
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getMaxVolume()");

	IOReturn		result = kIOReturnError;

	FailIf(NULL == mControlInterface, Exit);

	if (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	{
		SubRange16		subRange;

		result = this->getFeatureUnitRange(VOLUME_CONTROL, unitID, channelNumber, USBAUDIO_0200::RANGE, &subRange);

		if (target)
		{
			::memcpy(target, &subRange.wMAX, 2);
		}
	}
	else
	{
		result = this->getFeatureUnitSetting(VOLUME_CONTROL, unitID, channelNumber, GET_MAX, target);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getMaxVolume() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ボリュームの分解能取得．
 *
 * @param[in]		unitID				Feature Unit のユニット ID．
 * @param[in]		channelNumber		対象チャンネル番号．
 * @param[out]		*target				取得した分解能．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getVolumeResolution(
	UInt8			unitID,
	UInt8			channelNumber,
	UInt16			*target
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getVolumeResolution()");

	IOReturn		result = kIOReturnError;

	FailIf(NULL == mControlInterface, Exit);

	if (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
	{
		SubRange16		subRange;

		result = this->getFeatureUnitRange(VOLUME_CONTROL, unitID, channelNumber, USBAUDIO_0200::RANGE, &subRange);

		if (target)
		{
			::memcpy(target, &subRange.wRES, 2);
		}
	}
	else
	{
		result = this->getFeatureUnitSetting(VOLUME_CONTROL, unitID, channelNumber, GET_RES, (SInt16 *)target);
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getVolumeResolution() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	フィーチャーユニット設定．
 *			フィーチャーユニットに SET 系デバイスリクエストを発行．
 *
 * @param[in]		controlSelector		コントロールセレクタ．
 * @param[in]		unitID				フィーチャーユニットのユニット ID．
 * @param[in]		channelNumber		チャンネル番号．
 * @param[in]		requestType			bRequest フィールド．
 * @param[in]		newValue			設定する値．
 * @param[in]		newValueLen			バッファサイズ．２バイトまで．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::setFeatureUnitSetting(
	UInt8			controlSelector,
	UInt8			unitID,
	UInt8			channelNumber,
	UInt8			requestType,
	UInt16			newValue,
	UInt16			newValueLen
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::setFeatureUnitSetting()");

	IOReturn					result = kIOReturnError;
	IOBufferMemoryDescriptor	*theSettingDesc = NULL;
	StandardUSB::DeviceRequest	devReq;

	FailIf(NULL == mControlInterface, Exit);

	theSettingDesc = IOBufferMemoryDescriptor::withBytes(&newValue, newValueLen, kIODirectionOut);
	FailIf(NULL == theSettingDesc, Exit);

	devReq.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBClass, kUSBInterface);
	devReq.bRequest		 = requestType;
	devReq.wValue		 = (controlSelector << 8) | channelNumber;
	devReq.wIndex		 = (0xFF00 & (unitID << 8)) | (0x00FF & mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
	devReq.wLength		 = newValueLen;

	FailIf(mAudioDevice->isInactive(), DeviceInactive);	// In case we've been unplugged during sleep

	result = mAudioDevice->deviceRequest(&devReq, theSettingDesc);

Exit:
	if (NULL != theSettingDesc)
	{
		theSettingDesc->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::setFeatureUnitSetting() = %d", result);
	return result;

DeviceInactive:
	DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "::setFeatureUnitSetting() - ERROR attempt to send a device request to an inactive device.");
	goto Exit;
}

/****************************************************************************/
/**
 * @brief	フィーチャーユニット設定取得．
 *			フィーチャーユニットに GET 系デバイスリクエストを発行．
 *
 * @param[in]		controlSelector		コントロールセレクタ．
 * @param[in]		unitID				フィーチャーユニットのユニット ID．
 * @param[in]		channelNumber		チャンネル番号．
 * @param[in]		requestType			bRequest フィールド．
 * @param[out]		*target				取得した設定値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getFeatureUnitSetting(
	UInt8			controlSelector,
	UInt8			unitID,
	UInt8			channelNumber,
	UInt8			requestType,
	SInt16			*target
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getFeatureUnitSetting()");

	IOReturn					result = kIOReturnError;
	IOBufferMemoryDescriptor	*theSettingDesc = NULL;
	StandardUSB::DeviceRequest	devReq;
	UInt8						length;
	UInt16						theSetting = 0;

	FailIf(NULL == target, Exit);
	FailIf(NULL == mControlInterface, Exit);

	switch (controlSelector)
	{
		case MUTE_CONTROL:		length = 1;		break;
		case VOLUME_CONTROL:	length = 2;		break;
		default:				length = 0;		break;
	}

	theSettingDesc = IOBufferMemoryDescriptor::withOptions(kIODirectionIn, length);
	FailIf(NULL == theSettingDesc, Exit);

	devReq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBClass, kUSBInterface);
	devReq.bRequest		 = requestType;
	devReq.wValue		 = (controlSelector << 8) | channelNumber;
	devReq.wIndex		 = (0xFF00 & (unitID << 8)) | (0x00FF & mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
	devReq.wLength		 = length;

	result = mAudioDevice->deviceRequest(&devReq, theSettingDesc);
	FailIf(kIOReturnSuccess != result, Exit);

	::memcpy(&theSetting, theSettingDesc->getBytesNoCopy(), length);

Exit:
	if (NULL != theSettingDesc)
	{
		theSettingDesc->release();
	}

	if (NULL != target)
	{
		*target = USBToHostWord(theSetting);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getFeatureUnitSetting() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	フィーチャーユニットからレンジを取得．
 *			フィーチャーユニットに GET RANGE デバイスリクエストを発行．
 *
 * @param[in]		controlSelector		コントロールセレクタ．
 * @param[in]		unitID				フィーチャーユニットのユニット ID．
 * @param[in]		channelNumber		チャンネル番号．
 * @param[in]		requestType			bRequest フィールド．
 * @param[out]		*target				取得したレンジ設定．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::getFeatureUnitRange(
	UInt8			controlSelector,
	UInt8			unitID,
	UInt8			channelNumber,
	UInt8			requestType,
	SubRange16		*target
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getFeatureUnitRange()");

	struct {
		UInt16					wNumSubRanges;
		SubRange16				subRanges[1];
	}							theSetting;

	IOReturn					result = kIOReturnError;
	IOBufferMemoryDescriptor	*theSettingDesc = NULL;
	StandardUSB::DeviceRequest	devReq;
	UInt8						length = sizeof(theSetting);

	theSetting.subRanges[0].wMIN = 0;
	theSetting.subRanges[0].wMAX = 0;
	theSetting.subRanges[0].wRES = 0;

	FailIf(NULL == target, Exit);
	FailIf(NULL == mControlInterface, Exit);

	theSettingDesc = IOBufferMemoryDescriptor::withOptions(kIODirectionIn, length);
	FailIf(NULL == theSettingDesc, Exit);

	devReq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBClass, kUSBInterface);
	devReq.bRequest		 = requestType;
	devReq.wValue		 = (controlSelector << 8) | channelNumber;
	devReq.wIndex		 = (0xFF00 & (unitID << 8)) | (0x00FF & mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
	devReq.wLength		 = length;

	result = mAudioDevice->deviceRequest(&devReq, theSettingDesc);
	FailIf(kIOReturnSuccess != result, Exit);

	::memcpy(&theSetting, theSettingDesc->getBytesNoCopy(), length);

Exit:
	if (NULL != theSettingDesc)
	{
		theSettingDesc->release();
	}
	if (NULL != target)
	{
		if (USBToHostWord(theSetting.wNumSubRanges) > 0)
		{
			target->wMIN = USBToHostWord(theSetting.subRanges[0].wMIN);
			target->wMAX = USBToHostWord(theSetting.subRanges[0].wMAX);
			target->wRES = USBToHostWord(theSetting.subRanges[0].wRES);
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getFeatureUnitRange() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	セレクタユニット設定．
 *			セレクタユニットに SET_CUR デバイスリクエストを発行．
 *
 * @param[in]		selectorID			セレクタユニットのユニット ID．
 * @param[in]		setting				設定する値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::setSelectorSetting(
	UInt8			selectorID,
	UInt8			setting
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::setSelectorSetting()");

	IOReturn					result = kIOReturnError;
	IOBufferMemoryDescriptor	*settingDesc = NULL;
	StandardUSB::DeviceRequest	devReq;

	FailIf(NULL == mControlInterface, Exit);

#ifdef BUG_FIX
	settingDesc = IOBufferMemoryDescriptor::withBytes(&setting, 1, kIODirectionOut);
#else
	settingDesc = IOBufferMemoryDescriptor::withBytes(&setting, 1, kIODirectionIn);
#endif
	FailIf(NULL == settingDesc, Exit);

	devReq.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBClass, kUSBInterface);
	devReq.bRequest		 = SET_CUR;
	devReq.wValue		 = (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
							? (((UInt16)USBAUDIO_0200::SU_SELECTOR_CONTROL) << 8)
							: 0;															// <rdar://problem/6001162>
	devReq.wIndex		 = (0xFF00 & (selectorID << 8)) | (0x00FF & mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
	devReq.wLength		 = 1;

	result = mAudioDevice->deviceRequest(&devReq, settingDesc);

Exit:
	if (NULL != settingDesc)
	{
		settingDesc->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::setSelectorSetting() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	セレクタユニット設定取得．
 *			セレクタユニットに GET_CUR デバイスリクエストを発行．
 *
 * @param[in]		selectorID			セレクタユニットのユニット ID．
 *
 * @return			取得した設定値．
 */
UInt8
PdjUsbAudioTopology::getSelectorSetting(
	UInt8			selectorID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getSelectorSetting()");

	IOBufferMemoryDescriptor	*settingDesc = NULL;
	StandardUSB::DeviceRequest	devReq;
	UInt8						setting = 0;

	FailIf(NULL == mControlInterface, Exit);

	settingDesc = IOBufferMemoryDescriptor::withOptions(kIODirectionIn, 1);
	FailIf(NULL == settingDesc, Exit);

	devReq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBClass, kUSBInterface);
	devReq.bRequest		 = (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
							? ((UInt8)USBAUDIO_0200::CUR)
							: ((UInt8)GET_CUR);
	devReq.wValue		 = (IP_VERSION_02_00 == mControlInterface->getInterfaceDescriptor()->bInterfaceProtocol)
							? (((UInt16)USBAUDIO_0200::SU_SELECTOR_CONTROL) << 8)
							: 0;															// <rdar://problem/6001162>
	devReq.wIndex		 = (0xFF00 & (selectorID << 8)) | (0x00FF & mControlInterface->getInterfaceDescriptor()->bInterfaceNumber);
	devReq.wLength		 = 1;

	FailIf(kIOReturnSuccess != mAudioDevice->deviceRequest(&devReq, settingDesc), Exit);

	::memcpy(&setting, settingDesc->getBytesNoCopy(), 1);

Exit:
	if (NULL != settingDesc)
	{
		settingDesc->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getSelectorSetting() = %d", setting);
	return setting;
}

/****************************************************************************/
/**
 * @brief	ボリュームコントロールの wVolume 値を dB 値に変換．
 *
 * @param[in]		volume			ボリュームコントロールの wVolume 値．
 *
 * @return			固定小数点形式で表した dB 値．
 *					IOFixed 型は符号 1bit，整数部 15bit，小数部 16bit から成る固定小数点型である．
 */
// This is how the thing is defined in the USB Audio spec (section 5.2.2.4.3.2 for the curious).
// The volume setting of a device is described in 1/256 dB increments using a number that goes from
// a max of 0x7fff (127.9961 dB) down to 0x8001 (-127.9961 dB) using standard signed math, but 0x8000
// is actually negative infinity (not -128 dB), so I have to special case it.
IOFixed
PdjUsbAudioTopology::convertUSBVolumeTodB(
	SInt16			volume
)
{
	IOFixed			dBVolumeFixed;

	if (volume == (SInt16)0x8000)
	{
		dBVolumeFixed = (UInt32)((SInt16)0x8000 * 256) << 8;	// really is negative infinity
	}
	else
	{
		dBVolumeFixed = volume * 256;
	}

	DP(DBG_INFO, "volume = %d, dBVolumeFixed = 0x%x", volume, dBVolumeFixed);
	return dBVolumeFixed;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface のリスト中に対応サンプリング周波数が共通の要素があるか調べる．
 *
 * @param[in]		*refStreamInterfaceNumber	基準とする AudioStreaming Interface の interface 番号．
 * @param[in]		*streamInterfaceNumberList	検索対象となる AudioStreaming Interface の OSArray．
 *
 * @retval			true						リスト中に対応サンプリング周波数が共通の要素が１つ以上存在する．
 * @retval			false						リスト中に対応サンプリング周波数が共通の要素が１つも存在しない．
 */
// <rdar://5131786> Determine if the stream's sample rates are common with at least one other stream in the stream list.
bool
PdjUsbAudioTopology::isSampleRateCommonWithAtLeastOneStreamsInList(
	OSNumber		*refStreamInterfaceNumber,
	OSArray			*streamInterfaceNumberList
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::isSampleRateCommonWithAtLeastOneStreamsInList()", this);

	bool	result = false;

	for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < streamInterfaceNumberList->getCount(); streamInterfaceIndex++)
	{
		OSNumber	*streamInterfaceNumber = OSDynamicCast(OSNumber, streamInterfaceNumberList->getObject(streamInterfaceIndex));
		FailIf(NULL == streamInterfaceNumber, Exit);

		if (this->streamsHaveCommonSampleRates(refStreamInterfaceNumber, streamInterfaceNumber))
		{
			result = true;
			break;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::isSampleRateCommonWithAtLeastOneStreamsInList() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface のリスト中に対応サンプリング周波数が共通でない要素があるか調べる．
 *
 * @param[in]		*refStreamInterfaceNumber	基準とする AudioStreaming Interface の interface 番号．
 * @param[in]		*streamInterfaceNumberList	検索対象となる AudioStreaming Interface の OSArray．
 *
 * @retval			true						リスト中の全要素と対応サンプリング周波数が共通である．
 * @retval			false						リスト中に対応サンプリング周波数が共通でない要素が１つ以上存在する．
 */
// <rdar://5131786> Determine if the stream's sample rates are common with all other streams in the stream list.
bool
PdjUsbAudioTopology::isSampleRateCommonWithAllStreamsInList(
	OSNumber		*refStreamInterfaceNumber,
	OSArray			*streamInterfaceNumberList
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::isSampleRateCommonWithAllStreamsInList()", this);

	bool	result = true;

	for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < streamInterfaceNumberList->getCount(); streamInterfaceIndex++)
	{
		OSNumber	*streamInterfaceNumber = OSDynamicCast(OSNumber, streamInterfaceNumberList->getObject(streamInterfaceIndex));
		FailIf(NULL == streamInterfaceNumber, Exit);

		if (!(this->streamsHaveCommonSampleRates(refStreamInterfaceNumber, streamInterfaceNumber)))
		{
			result = false;
			break;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::isSampleRateCommonWithAllStreamsInList() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	２つの AudioStreaming Interface の対応サンプリング周波数が共通であるか調べる．
 *
 * @param[in]		*streamInterfaceNumberA		比較対象となる AudioStreaming Interface の interface 番号（Ａ）．
 * @param[in]		*streamInterfaceNumberB		比較対象となる AudioStreaming Interface の interface 番号（Ｂ）．
 *
 * @retval			true						対応サンプリング周波数は共通である．
 * @retval			false						対応サンプリング周波数は共通でない．
 */
// <rdar://5131786> Determine if the stream's sample rates are common with each other.
bool
PdjUsbAudioTopology::streamsHaveCommonSampleRates(
	OSNumber		*streamInterfaceNumberA,
	OSNumber		*streamInterfaceNumberB
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::streamsHaveCommonSampleRates(%p, %p)",
		this, streamInterfaceNumberA, streamInterfaceNumberB);

	bool		result			= false;
	OSArray		*sampleRatesA	= NULL;
	OSArray		*sampleRatesB	= NULL;

	FailIf(NULL == streamInterfaceNumberA, Exit);
	FailIf(NULL == streamInterfaceNumberB, Exit);

	DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::streamsHaveCommonSampleRates(%d, %d)",
		this, streamInterfaceNumberA->unsigned8BitValue(), streamInterfaceNumberB->unsigned8BitValue());

	sampleRatesA = this->getSampleRatesFromStreamInterface(streamInterfaceNumberA);
	sampleRatesB = this->getSampleRatesFromStreamInterface(streamInterfaceNumberB);

#ifdef DEBUGLOGGING
	DP(DBG_INFO, "Sample rates for interface %d:", streamInterfaceNumberA->unsigned8BitValue());

	if (NULL != sampleRatesA)
	{
		for (UInt32 rateIndex = 0; rateIndex < sampleRatesA->getCount(); rateIndex++)
		{
			OSNumber	*rate = OSDynamicCast(OSNumber, sampleRatesA->getObject(rateIndex));

			FailMessage(NULL == rate);

			if (NULL != rate)
			{
				DP(DBG_INFO, "--> %u", rate->unsigned32BitValue());
			}
		}
	}

	DP(DBG_INFO, "Sample rates for interface %d:", streamInterfaceNumberB->unsigned8BitValue());

	if (NULL != sampleRatesB)
	{
		for (UInt32 rateIndex = 0; rateIndex < sampleRatesB->getCount(); rateIndex++)
		{
			OSNumber	*rate = OSDynamicCast(OSNumber, sampleRatesB->getObject(rateIndex));

			FailMessage(NULL == rate);

			if (NULL != rate)
			{
				DP(DBG_INFO, "--> %u", rate->unsigned32BitValue());
			}
		}
	}
#endif

	if ((NULL != sampleRatesA) && (NULL != sampleRatesB))
	{
		if (this->compareSampleRates(sampleRatesA, sampleRatesB) && this->compareSampleRates(sampleRatesB, sampleRatesA))
		{
			result = true;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::streamsHaveCommonSampleRates(%p, %p) = %d",
		this, streamInterfaceNumberA, streamInterfaceNumberB, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface の対応サンプリング周波数の OSArray を取得．
 *
 * @param[in]		*streamInterfaceNumber		AudioStreaming Interface の interface 番号．
 *
 * @return			対応サンプリング周波数の OSArray．
 */
// <rdar://5131786> Get the sample rates supported by a stream interface.
OSArray *
PdjUsbAudioTopology::getSampleRatesFromStreamInterface(
	OSNumber		*streamInterfaceNumber
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getSampleRatesFromStreamInterface()", this);

	UInt8			interfaceNum;
	UInt8			numAltSettings = 0;
	OSArray			*sampleRates = NULL;

	FailIf(NULL == streamInterfaceNumber, Exit);

	interfaceNum = streamInterfaceNumber->unsigned8BitValue();

	if (kIOReturnSuccess == mConfigDictionary->getNumAltSettings(&numAltSettings, interfaceNum))
	{
		for (UInt8 altSetting = 0; altSetting < numAltSettings; altSetting++)
		{
			OSArray		*rates = mConfigDictionary->getSampleRates(interfaceNum, altSetting);

			if (NULL != rates)
			{
				if (NULL == sampleRates)
				{
					sampleRates = OSArray::withCapacity(rates->getCount());
				}

				if (NULL != sampleRates)
				{
					this->mergeSampleRates(sampleRates, rates);
				}
			}
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getSampleRatesFromStreamInterface()", this);
	return sampleRates;
}

/****************************************************************************/
/**
 * @brief	サンプリング周波数の OSArray を比較．
 *
 * @param[in]		*sampleRatesA	比較対象となる OSArray（Ａ）．
 * @param[in]		*sampleRatesB	比較対象となる OSArray（Ｂ）．
 *
 * @retval			true			Ａの要素はすべてＢにも含まれる．Ａ⊆Ｂ．
 * @retval			false			Ａの要素のうち、Ｂに含まれないものがある．
 */
// <rdar://5131786> Compare the sample rates in both array to determine if all rates are supported in both array.
bool
PdjUsbAudioTopology::compareSampleRates(
	OSArray			*sampleRatesA,
	OSArray			*sampleRatesB
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::compareSampleRates()", this);

	bool	sampleRatesSupported = true;

	for (UInt32 indexA = 0; indexA < sampleRatesA->getCount(); indexA++)
	{
		OSNumber	*rateA = OSDynamicCast(OSNumber, sampleRatesA->getObject(indexA));

		if (NULL != rateA)
		{
			bool	rateFound = false;

			for (UInt32 indexB = 0; indexB < sampleRatesB->getCount(); indexB++)
			{
				OSNumber	*rateB = OSDynamicCast(OSNumber, sampleRatesB->getObject(indexB));

				if (NULL != rateB)
				{
					if (rateA->unsigned32BitValue() == rateB->unsigned32BitValue())
					{
						rateFound = true;
						break;
					}
				}
			}

			if (!rateFound)
			{
				sampleRatesSupported = false;
				break;
			}
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::compareSampleRates() = %d", this, sampleRatesSupported);
	return sampleRatesSupported;
}

/****************************************************************************/
/**
 * @brief	サンプリング周波数の OSArray をマージ．
 *
 * @param[in, out]	*thisArray		ベースとなる OSArray．
 * @param[in]		*otherArray		マージする OSArray．
 */
// <rdar://5131786> Merge sample rates array, taking in only new sample rates.
void
PdjUsbAudioTopology::mergeSampleRates(
	OSArray			*thisArray,
	OSArray			*otherArray
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::mergeSampleRates()", this);

	for (UInt32 otherIndex = 0; otherIndex < otherArray->getCount(); otherIndex++)
	{
		OSNumber	*otherRate = OSDynamicCast(OSNumber, otherArray->getObject(otherIndex));

		if (NULL != otherRate)
		{
			bool	rateFound = false;

			// Check if the sample rate above is already in our array.
			for (UInt32 thisIndex = 0; thisIndex < thisArray->getCount(); thisIndex++)
			{
				OSNumber	*thisRate = OSDynamicCast(OSNumber, thisArray->getObject(thisIndex));

				if (NULL != thisRate)
				{
					if (thisRate->unsigned32BitValue() == otherRate->unsigned32BitValue())
					{
						rateFound = true;
						break;
					}
				}
			}

			if (!rateFound)
			{
				thisArray->setObject(otherRate);
			}
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::mergeSampleRates()", this);
}

/****************************************************************************/
/**
 * @brief	指定の転送方向と同期タイプを持つ AudioStreaming Interface を検索．
 *
 * @param[in, out]	*availableStreamList	検索対象となる AudioStreaming Interface の interface 番号の OSArray．
 *											検索条件に合致した要素は削除される．
 * @param[in]		direction				検索条件：転送方向．
 * @param[in]		syncType				検索条件：同期タイプ．
 *
 * @return			指定の転送方向と同期タイプを持つ AudioStreaming Interface の interface 番号の OSArray．
 */
// <rdar://5131786> Find streams with the specified direction & sync type in the available list.
OSArray *
PdjUsbAudioTopology::findStreamsWithDirectionAndSyncType(
	OSArray			*availableStreamList,
	UInt8			direction,
	UInt8			syncType
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithDirectionAndSyncType(%p, %d, %d)",
		this, availableStreamList, direction, syncType);

	OSNumber		*streamInterfaceNumber;
	OSArray			*compatibleStreamList	= NULL;
	OSArray			*inCompatibleStreamList	= NULL;
	OSArray			*result = NULL;

	FailIf(NULL == availableStreamList, Exit);

	compatibleStreamList = OSArray::withCapacity(1);
	FailIf(0 == compatibleStreamList, Exit);

	inCompatibleStreamList = OSArray::withCapacity(1);
	FailIf(0 == inCompatibleStreamList, Exit);

	// Go thru all the available streams, and find all streams that have compatible endpoints and put them on
	// the compatible list. Otherwise, put them on the incompatible list.
	while (0 != availableStreamList->getCount())
	{
		FailIf(NULL == (streamInterfaceNumber = OSDynamicCast(OSNumber, availableStreamList->getObject(0))), Exit);
		streamInterfaceNumber->retain();

		availableStreamList->removeObject(0);

		if (this->streamEndpointsHaveSpecifiedDirectionAndSyncType(streamInterfaceNumber, direction, syncType))
		{
			// Keep it in the compatible list.
			compatibleStreamList->setObject(streamInterfaceNumber);
		}
		else
		{
			// Keep it in the incompatible list.
			inCompatibleStreamList->setObject(streamInterfaceNumber);
		}

		streamInterfaceNumber->release();
	}

	// The available stream list should be empty here.
	FailIf(0 != availableStreamList->getCount(), Exit);

	if (0 != compatibleStreamList->getCount())
	{
		// There is some streams with compatible endpoints in the compatible list.
		result = compatibleStreamList;
		result->retain();
	}

#ifdef DEBUGLOGGING
	if (NULL != result)
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithDirectionAndSyncType(%p, %d, %d) - Found streams:",
			this, availableStreamList, direction, syncType);

		for (UInt32 streamInterfaceIndex = 0; streamInterfaceIndex < result->getCount(); streamInterfaceIndex++)
		{
			streamInterfaceNumber = OSDynamicCast(OSNumber, result->getObject(streamInterfaceIndex));
			FailMessage(NULL == streamInterfaceNumber);

			if (NULL != streamInterfaceNumber)
			{
				DP(DBG_INFO, "--> #%u", streamInterfaceNumber->unsigned8BitValue());
			}
		}
	}
#endif

Exit:
	if (NULL != inCompatibleStreamList)
	{
		if (0 != inCompatibleStreamList->getCount())
		{
			// Put back the incompatible streams into the available stream list.
			availableStreamList->merge(inCompatibleStreamList);
		}

		inCompatibleStreamList->release();
	}

	if (NULL != compatibleStreamList)
	{
		// If it failed for some reason, put back the compatible streams into the available stream list.
		if ((NULL == result) && (0 != compatibleStreamList->getCount()))
		{
			availableStreamList->merge(compatibleStreamList);
		}

		compatibleStreamList->release();
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::findStreamsWithDirectionAndSyncType(%p, %d, %d) - result = %p",
		this, availableStreamList, direction, syncType, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	AudioStreaming Interface のエンドポイントが指定の転送方向と同期タイプを持つか調べる．
 *
 * @param[in]		*streamInterfaceNumber	AudioStreaming Interface の interface 番号．
 * @param[in]		endpointDirection		検索条件：転送方向．
 * @param[in]		endpointSyncType		検索条件：同期タイプ．
 *
 * @retval			true					エンドポイントは指定の転送方向と同期タイプを持つ．
 * @retval			false					エンドポイントは指定の転送方向と同期タイプを持たない．
 */
// <rdar://5131786> Determine if the stream's endpoints are compatible with the specified direction & sync type.
bool
PdjUsbAudioTopology::streamEndpointsHaveSpecifiedDirectionAndSyncType(
	OSNumber		*streamInterfaceNumber,
	UInt8			endpointDirection,
	UInt8			endpointSyncType
)
{
	UInt8			interfaceNum;
	UInt8			numAltSettings = 0;
	bool			hasSpecifiedDirection	= true;
	bool			hasSpecifiedSyncType	= true;
	bool			result = false;

	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::streamEndpointsHaveSpecifiedDirectionAndSyncType(%p, %d, %d)",
		this, streamInterfaceNumber, endpointDirection, endpointSyncType);

	FailIf(NULL == streamInterfaceNumber, Exit);

	interfaceNum = streamInterfaceNumber->unsigned8BitValue();

	DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::streamEndpointsHaveSpecifiedDirectionAndSyncType(%d, %d, %d)",
		this, interfaceNum, endpointDirection, endpointSyncType);

	if (kIOReturnSuccess == mConfigDictionary->getNumAltSettings(&numAltSettings, interfaceNum))
	{
		bool	startAtZero = mConfigDictionary->alternateSettingZeroCanStream(interfaceNum);

		if (kUnknownSyncType == endpointSyncType)
		{
			UInt32	compareSyncType = kUnknownSyncType;

			hasSpecifiedSyncType = false;

			// Make sure that all alternate settings has the same sync type.
			for (UInt8 altSetting = (startAtZero ? 0 : 1); altSetting < numAltSettings; altSetting++)
			{
				UInt8	direction	= 0;
				UInt8	address		= 0;
				UInt8	syncType	= 0;

				// <rdar://6378666> Don't exit if there are AS interfaces with missing class-specific interface descriptor.
				if (
						(kIOReturnSuccess == mConfigDictionary->getIsocEndpointDirection(&direction, interfaceNum, altSetting))
				   &&
						(kIOReturnSuccess == mConfigDictionary->getIsocEndpointAddress(&address, interfaceNum, altSetting, direction))
				   &&
						(kIOReturnSuccess == mConfigDictionary->getIsocEndpointSyncType(&syncType, interfaceNum, altSetting, address))
				   )
				{
					if (endpointDirection != direction)
					{
						hasSpecifiedDirection = false;
						break;
					}

					if (kUnknownSyncType == compareSyncType)
					{
						compareSyncType = syncType;
					}
					else
					{
						if (compareSyncType != syncType)
						{
							hasSpecifiedSyncType = true;
						}
					}
				}
			}
		}
		else
		{
			// Make sure that all the input & output endpoints has the compatible sync type.
			for (UInt8 altSetting = (startAtZero ? 0 : 1); altSetting < numAltSettings; altSetting++)
			{
				UInt8	direction	= 0;
				UInt8	address		= 0;
				UInt8	syncType	= 0;

				// <rdar://6378666> Don't exit if there are AS interfaces with missing class-specific interface descriptor.
				if (
						(kIOReturnSuccess == mConfigDictionary->getIsocEndpointDirection(&direction, interfaceNum, altSetting))
				   &&
						(kIOReturnSuccess == mConfigDictionary->getIsocEndpointAddress(&address, interfaceNum, altSetting, direction))
				   &&
						(kIOReturnSuccess == mConfigDictionary->getIsocEndpointSyncType(&syncType, interfaceNum, altSetting, address))
				   )
				{
					if (endpointDirection != direction)
					{
						hasSpecifiedDirection = false;
						break;
					}

					switch (endpointSyncType)
					{
						case kNoneSyncType:
						case kSynchronousSyncType:
							if ((kNoneSyncType != syncType) && (kSynchronousSyncType != syncType))
							{
								hasSpecifiedSyncType = false;
							}
							break;

						default:
							if (endpointSyncType != syncType)
							{
								hasSpecifiedSyncType = false;
							}
							break;
					}
				}

				if (!hasSpecifiedSyncType)
				{
					break;
				}
			}
		}
	}

	if (hasSpecifiedDirection && hasSpecifiedSyncType)
	{
		result = true;
	}

Exit:

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::streamEndpointsHaveSpecifiedDirectionAndSyncType(%p, %d, %d) - result = %d",
		this, streamInterfaceNumber, endpointDirection, endpointSyncType, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオエンジン管理用の OSDictionary を作成．
 *			作成した OSDictionary はメンバ変数 mRegisteredEngines に登録される．
 *
 * @param[in]		*usbAudioEngine		PdjUsbAudioEngine クラスのインスタンス．
 *
 * @return			IOReturn コード．
 */
// <rdar://6420832>
IOReturn
PdjUsbAudioTopology::registerEngineInfo(
	PdjUsbAudioEngine		*usbAudioEngine
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::registerEngineInfo()");

	IOReturn		result = kIOReturnError;
	OSDictionary	*engineInfo;
	SInt32			oldEngineIndex;

	// <rdar://problem/6369110> Occassional kernel panic when multiple threads access mRegisteredEngines array simultaneously
	if (NULL == mAudioDevice->mRegisteredEngines)
	{
		mAudioDevice->mRegisteredEngines = OSArray::withCapacity(2);
		FailIf(NULL == mAudioDevice->mRegisteredEngines, Exit);
	}

	engineInfo = OSDictionary::withCapacity(1);
	FailIf(NULL == engineInfo, Exit);

	engineInfo->setObject(kEngine, usbAudioEngine);

	// <rdar://problem/6369110>
	if (mAudioDevice->mRegisteredEnginesMutex)
	{
		IORecursiveLockLock(mAudioDevice->mRegisteredEnginesMutex);
	}

	// [rdar://4102789] Be sure to preserve the integrity of mRegisteredEngines. It is vital to perform emergency format changes.
	oldEngineIndex = this->getEngineInfoIndex(usbAudioEngine);

	if (-1 != oldEngineIndex)
	{
		// This engine already has information stored. We should replace it.
		mAudioDevice->mRegisteredEngines->replaceObject(oldEngineIndex, engineInfo);
	}
	else
	{
		mAudioDevice->mRegisteredEngines->setObject(engineInfo);
	}

	if (mAudioDevice->mRegisteredEnginesMutex)
	{
		IORecursiveLockUnlock(mAudioDevice->mRegisteredEnginesMutex);
	}

	engineInfo->release();

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::registerEngineInfo() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオエンジン管理用 OSDictionary のインデックスを取得．
 *
 * @param[in]		*inAudioEngine		検索するオーディオエンジン．
 *
 * @return			該当する OSDictionary のインデックス．該当なしなら -1．
 */
SInt32
PdjUsbAudioTopology::getEngineInfoIndex(
	PdjUsbAudioEngine		*inAudioEngine
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getEngineInfoIndex()");

	SInt32			returnIndex = -1;

	// <rdar://problem/6369110> Occassional kernel panic when multiple threads access mRegisteredEngines array simultaneously
	if (mAudioDevice->mRegisteredEnginesMutex)
	{
		IORecursiveLockLock(mAudioDevice->mRegisteredEnginesMutex);
	}

	if (mAudioDevice->mRegisteredEngines)
	{
		for (UInt16 engineIndex = 0; engineIndex < mAudioDevice->mRegisteredEngines->getCount(); engineIndex++)
		{
			OSDictionary	*engineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(engineIndex));

			if (engineInfo)
			{
				PdjUsbAudioEngine	*usbAudioEngine = OSDynamicCast(PdjUsbAudioEngine, engineInfo->getObject(kEngine));

				if (inAudioEngine == usbAudioEngine)
				{
					returnIndex = engineIndex;
					break;		// Get out of for loop with correct index
				}
			}
		}
	}

	if (mAudioDevice->mRegisteredEnginesMutex)
	{
		IORecursiveLockUnlock(mAudioDevice->mRegisteredEnginesMutex);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getEngineInfoIndex() = %d", returnIndex);
	return returnIndex;
}

/****************************************************************************/
/**
 * @brief	オーディオストリーム管理用の OSDictionary を作成．
 *			作成した OSDictionary はメンバ変数 mRegisteredStreams に登録される．
 *			オーディオストリーム配下の各種オーディオコントロールの管理のために
 *			使用される．
 *
 * @param[in]		interfaceNum		USB interface 番号．
 * @param[in]		altSettingNum		alternate setting 番号．
 *
 * @return			IOReturn コード．
 */
// <rdar://6420832>
IOReturn
PdjUsbAudioTopology::registerStreamInfo(
	UInt8			interfaceNum,
	UInt8			altSettingNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::registerStreamInfo()");

	IOReturn		result = kIOReturnError;
	OSDictionary	*streamInfo;
	SInt32			oldStreamIndex;
	OSNumber		*number;

	if (NULL == mAudioDevice->mRegisteredStreams)
	{
		mAudioDevice->mRegisteredStreams = OSArray::withCapacity(4);
		FailIf(NULL == mAudioDevice->mRegisteredStreams, Exit);
	}

	streamInfo = OSDictionary::withCapacity(2);
	FailIf(NULL == streamInfo, Exit);

	number = OSNumber::withNumber(interfaceNum, 8);
	streamInfo->setObject(kInterface, number);
	number->release();

	number = OSNumber::withNumber(altSettingNum, 8);
	streamInfo->setObject(kAltSetting, number);
	number->release();

	if (mAudioDevice->mRegisteredStreamsMutex)
	{
		IORecursiveLockLock(mAudioDevice->mRegisteredStreamsMutex);
	}

	oldStreamIndex = this->getStreamInfoIndex(interfaceNum);

	if (-1 != oldStreamIndex)
	{
		// This stream already has information stored. We should replace it.
		mAudioDevice->mRegisteredStreams->replaceObject(oldStreamIndex, streamInfo);
	}
	else
	{
		mAudioDevice->mRegisteredStreams->setObject(streamInfo);
	}

	if (mAudioDevice->mRegisteredStreamsMutex)
	{
		IORecursiveLockUnlock(mAudioDevice->mRegisteredStreamsMutex);
	}

	streamInfo->release();

	result = kIOReturnSuccess;

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::registerStreamInfo() = %d", result);
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオストリーム管理用 OSDictionary のインデックスを取得．
 *
 * @param[in]		interfaceNum		USB interface 番号．
 *
 * @return			該当する OSDictionary のインデックス．該当なしなら -1．
 */
// <rdar://6420832>
SInt32
PdjUsbAudioTopology::getStreamInfoIndex(
	UInt8			interfaceNum
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getStreamInfoIndex()");

	SInt32			returnIndex = -1;

	if (mAudioDevice->mRegisteredStreamsMutex)
	{
		IORecursiveLockLock(mAudioDevice->mRegisteredStreamsMutex);
	}

	if (mAudioDevice->mRegisteredStreams)
	{
		for (UInt16 streamIndex = 0; streamIndex < mAudioDevice->mRegisteredStreams->getCount(); streamIndex++)
		{
			OSDictionary	*streamInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredStreams->getObject(streamIndex));

			if (streamInfo)
			{
				OSNumber	*streamInterfaceNumber = OSDynamicCast(OSNumber, streamInfo->getObject(kInterface));

				if ((NULL != streamInterfaceNumber) && (streamInterfaceNumber->unsigned8BitValue() == interfaceNum))
				{
					returnIndex = streamIndex;
					break;		// Get out of for loop with correct index
				}
			}
		}
	}

	if (mAudioDevice->mRegisteredStreamsMutex)
	{
		IORecursiveLockUnlock(mAudioDevice->mRegisteredStreamsMutex);
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getStreamInfoIndex() = %d", returnIndex);
	return returnIndex;
}

/****************************************************************************/
/**
 * @brief	指定の入力ターミナルに対するデフォルトの出力ターミナルを検索．
 *			指定の入力ターミナルから到達可能な出力ターミナルをすべて走査し、
 *			優先順位（Speaker → Headphones → Line Out の順）の高いものを選ぶ．
 *
 * @param[in]		inputTerminalID		入力ターミナル ID．
 *
 * @return			出力ターミナル ID．
 */
// <rdar://6413207>
UInt8
PdjUsbAudioTopology::getDefaultOutputTerminalID(
	UInt8			inputTerminalID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getDefaultOutputTerminalID()");

	UInt8			controlInterfaceNum;
	UInt8			numOutputTerminals;
	UInt8			defaultOutputTerminalID = 0;
	UInt16			defaultOutputTerminalType = 0;

	FailIf(NULL == mControlInterface, Exit);

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	FailIf(kIOReturnSuccess != mConfigDictionary->getNumOutputTerminals(&numOutputTerminals, controlInterfaceNum, 0), Exit);

	for (UInt32 outputTerminalIndex = 0; outputTerminalIndex < numOutputTerminals; outputTerminalIndex++)
	{
		UInt16		terminalType;

		FailIf(
				kIOReturnSuccess != mConfigDictionary->getIndexedOutputTerminalType(
															&terminalType,
															controlInterfaceNum,
															0,
															outputTerminalIndex
													   ),
				Exit
		);

		if (terminalType != USB_STREAMING)
		{
			UInt8		outputTerminalID;
			UInt32		numOutputTerminalArrays;

			FailIf(
					kIOReturnSuccess != mConfigDictionary->getIndexedOutputTerminalID(
																&outputTerminalID,
																controlInterfaceNum,
																0,
																outputTerminalIndex
														   ),
					Exit
			);

			numOutputTerminalArrays = mControlGraph->getCount();

			bool	finished = false;	// <rdar://problem/6656791>

			for (UInt32 pathsToOutputTerminalN = 0; (pathsToOutputTerminalN < numOutputTerminalArrays) && (!finished); pathsToOutputTerminalN++)
			{
				OSArray		*arrayOfPathsFromOutputTerminal = OSDynamicCast(OSArray, mControlGraph->getObject(pathsToOutputTerminalN));
				FailIf(NULL == arrayOfPathsFromOutputTerminal, Exit);

				// <rdar://problem/6656791>
				UInt32		numPathsFromOutputTerminal = arrayOfPathsFromOutputTerminal->getCount();

				for (
						UInt32 arrayOfPathsFromOutputTerminalIndex = 0;
						(arrayOfPathsFromOutputTerminalIndex < numPathsFromOutputTerminal) && (!finished);
						arrayOfPathsFromOutputTerminalIndex++
					)
				{
					OSNumber	*theUnitIDNum;
					UInt8		inputUnitID;
					UInt8		outputUnitID;

					OSArray		*aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(0));
					FailIf(NULL == aPath, Exit);

					theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(0));
					FailIf(NULL == theUnitIDNum, Exit);

					outputUnitID = theUnitIDNum->unsigned8BitValue();

					theUnitIDNum = OSDynamicCast(OSNumber, aPath->getLastObject());
					FailIf(NULL == theUnitIDNum, Exit);

					inputUnitID = theUnitIDNum->unsigned8BitValue();

					if ((inputUnitID == inputTerminalID) && (outputUnitID == outputTerminalID))
					{
						// Select the default output terminal according to the following rule.
						// Speaker has priority over all the terminal types. Headphones has the priority over
						// all terminal types except Speaker. Line Out has the 3rd highest priority.
						switch (defaultOutputTerminalType)
						{
							case 0:
								defaultOutputTerminalType	= terminalType;
								defaultOutputTerminalID		= outputTerminalID;
								break;

							case 0x0301: // Speaker
								break;

							case 0x0302: // Headphones
								if (0x0301 == terminalType)
								{
									defaultOutputTerminalType	= terminalType;
									defaultOutputTerminalID		= outputTerminalID;
								}
								break;

							case 0x0603: // Line Out
								if ((0x0301 == terminalType) || (0x0302 == terminalType))
								{
									defaultOutputTerminalType	= terminalType;
									defaultOutputTerminalID		= outputTerminalID;
								}
								break;
						}

						finished = true;	// <rdar://problem/6656791>
					} // if ((inputUnitID == inputTerminalID) && (outputUnitID == outputTerminalID))
				} // for arrayOfPathsFromOutputTerminalIndex
			} // for pathsToOutputTerminalN
		} // if (terminalType != USB_STREAMING)
	} // for outputTerminalIndex

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getDefaultOutputTerminalID() = %d", defaultOutputTerminalID);
	return defaultOutputTerminalID;
}

/****************************************************************************/
/**
 * @brief	指定の入力ターミナルから到達可能な出力ターミナルの総数を数える．
 *
 * @param[in]		inputTerminalID		入力ターミナル ID．
 *
 * @return			到達可能な出力ターミナルの総数．
 */
// <rdar://6413207>
UInt32
PdjUsbAudioTopology::getNumConnectedOutputTerminals(
	UInt8			inputTerminalID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getNumConnectedOutputTerminals()");

	UInt8			controlInterfaceNum;
	UInt8			numOutputTerminals;
	UInt32			numConnectedOutputTerminals = 0;

	FailIf(NULL == mControlInterface, Exit);

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	FailIf(kIOReturnSuccess != mConfigDictionary->getNumOutputTerminals(&numOutputTerminals, controlInterfaceNum, 0), Exit);

	for (UInt32 outputTerminalIndex = 0; outputTerminalIndex < numOutputTerminals; outputTerminalIndex++)
	{
		UInt16		terminalType;

		FailIf(
				kIOReturnSuccess != mConfigDictionary->getIndexedOutputTerminalType(
															&terminalType,
															controlInterfaceNum,
															0,
															outputTerminalIndex
													   ),
				Exit
		);

		if (terminalType != USB_STREAMING)
		{
			UInt8		outputTerminalID;

			// <rdar://problem/6656791>
			bool		done = false;

			FailIf(
					kIOReturnSuccess != mConfigDictionary->getIndexedOutputTerminalID(
																&outputTerminalID,
																controlInterfaceNum,
																0,
																outputTerminalIndex
														   ),
					Exit
			);

			UInt32		numOutputTerminalArrays = mControlGraph->getCount();

			for (UInt32 pathsToOutputTerminalN = 0; (pathsToOutputTerminalN < numOutputTerminalArrays) && (!done); pathsToOutputTerminalN++)
			{
				OSArray		*arrayOfPathsFromOutputTerminal = OSDynamicCast(OSArray, mControlGraph->getObject(pathsToOutputTerminalN));

				FailIf(NULL == arrayOfPathsFromOutputTerminal, Exit);

				for (
						UInt32 arrayOfPathsFromOutputTerminalIndex = 0;
						(arrayOfPathsFromOutputTerminalIndex < arrayOfPathsFromOutputTerminal->getCount()) && (!done);
						arrayOfPathsFromOutputTerminalIndex++
					)
				{
					OSNumber	*theUnitIDNum;
					UInt8		inputUnitID;
					UInt8		outputUnitID;

					OSArray		*aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(arrayOfPathsFromOutputTerminalIndex));
					FailIf(NULL == aPath, Exit);

					theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(0));
					FailIf(NULL == theUnitIDNum, Exit);

					outputUnitID = theUnitIDNum->unsigned8BitValue();

					theUnitIDNum = OSDynamicCast(OSNumber, aPath->getLastObject());
					FailIf(NULL == theUnitIDNum, Exit);

					inputUnitID = theUnitIDNum->unsigned8BitValue();

					if ((inputUnitID == inputTerminalID) && (outputUnitID == outputTerminalID))
					{
						numConnectedOutputTerminals++;
						done = true;	// <rdar://problem/6656791>
					}
				} // for arrayOfPathsFromOutputTerminalIndex
			} // for pathsToOutputTerminalN
		} // if (terminalType != USB_STREAMING)
	} // for outputTerminalIndex

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getNumConnectedOutputTerminals() = %u", numConnectedOutputTerminals);
	return numConnectedOutputTerminals;
}

/****************************************************************************/
/**
 * @brief	経路上の Feature Unit を探す．
 *
 * @param[in]		*thePath			探索するオーディオパス．
 * @param[in]		direction			オーディオパスの方向．
 * @param[in]		interfaceNum		USB interface 番号．
 * @param[in]		altSettingNum		alternate setting 番号．
 * @param[in]		controlTypeWanted	探索するコントロールタイプ．
 *										kVolumeControl または kMuteControl．
 *
 * @return			見つかった Feature Unit のユニット ID．
 */
// This finds the feature unit closest to the input terminal.
UInt8
PdjUsbAudioTopology::getBestFeatureUnitInPath(
	OSArray			*thePath,
	UInt32			direction,
	UInt8			interfaceNum,
	UInt8			altSettingNum,
	UInt32			controlTypeWanted
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::getBestFeatureUnitInPath(%p, %u, %d, %d, %u)",
		this, thePath, direction, interfaceNum, altSettingNum, controlTypeWanted);

	bool			foundSelectorUnit;			// <rdar://5366067>
	bool			foundMixerUnit;				// <rdar://5366067>
	bool			foundFeatureUnit;			// <rdar://5366067>
	UInt8			controlInterfaceNum;
	UInt32			numUnitsInPath;
	UInt8			mixerUnitID;				// <rdar://5366067>
	OSNumber		*theUnitIDNum;
	UInt8			unitID;
	UInt8			subType;
	UInt8			numChannels;
	UInt8			featureUnitID		= 0;
	UInt8			endingUnitIndex		= 0;	// <rdar://5366067>
	UInt8			selectorUnitIndex	= 0;	// <rdar://5366067>
	UInt8			mixerUnitIndex		= 0;	// <rdar://5366067>

	FailIf(NULL == mControlInterface, Exit);
	FailIf(NULL == thePath, Exit);

	controlInterfaceNum	= mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;
	numUnitsInPath		= thePath->getCount();
	foundFeatureUnit	= false;	// <rdar://5366067>

	switch (direction)
	{
		case kIOAudioControlUsagePassThru:
			// [rdar://4586274] Require playthrough control feature units to be path unique.
			// <rdar://5366067> Find the mixer unit and look for the FU between the input terminal & mixer unit.
			foundMixerUnit = false;
			mixerUnitID = 0;

			for (UInt8 unitIndex = numUnitsInPath - 2; unitIndex > 0; unitIndex--)
			{
				theUnitIDNum = OSDynamicCast(OSNumber, thePath->getObject(unitIndex));

				if (NULL != theUnitIDNum)
				{
					unitID = theUnitIDNum->unsigned8BitValue();

					FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

					if (MIXER_UNIT == subType)
					{
						foundMixerUnit	= true;
						mixerUnitIndex	= unitIndex;
						mixerUnitID		= unitID;
						break;
					}
				}
			}

			endingUnitIndex = foundMixerUnit ? mixerUnitIndex : 0;

			// Find the feature unit closest to the input terminal. If there is a mixer unit, also check to see if
			// the FU is unique between the mixer unit & the input terminal.
			for (UInt8 unitIndex = numUnitsInPath - 2; (unitIndex > endingUnitIndex) && (!foundFeatureUnit); unitIndex--)
			{
				theUnitIDNum = OSDynamicCast(OSNumber, thePath->getObject(unitIndex));

				if (NULL != theUnitIDNum)
				{
					unitID = theUnitIDNum->unsigned8BitValue();

					FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

					if (
							(FEATURE_UNIT == subType)
					   &&
							(
								(1 == this->pathsContaining(unitID))
							||
								(
									(foundMixerUnit)
								&&
									(0 == this->pathsContainingFeatureUnitButNotMixerUnit(unitID, mixerUnitID))
								)
							)
					   )
					{
						FailIf(kIOReturnSuccess != mConfigDictionary->getNumChannels(&numChannels, interfaceNum, altSettingNum), Exit);

						for (UInt8 channelNum = 0; channelNum <= numChannels; channelNum++)
						{
							switch (controlTypeWanted)
							{
								case kVolumeControl:
									if (mConfigDictionary->channelHasVolumeControl(controlInterfaceNum, 0, unitID, channelNum))
									{
										featureUnitID	 = unitID;
										foundFeatureUnit = true;
									}
									break;

								case kMuteControl:
									if (mConfigDictionary->channelHasMuteControl(controlInterfaceNum, 0, unitID, channelNum))
									{
										featureUnitID	 = unitID;
										foundFeatureUnit = true;
									}
									break;
							}
						}
					}
				}
				else
				{
					DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "::getBestFeatureUnitInPath() - something is wrong here!!!");
				}
			} // for unitIndex

			break;

		case kIOAudioControlUsageInput:
			// <rdar://5366067> See if a selector unit is present in the path.
			foundSelectorUnit = false;

			for (UInt8 unitIndex = numUnitsInPath - 2; unitIndex > 0; unitIndex--)
			{
				theUnitIDNum = OSDynamicCast(OSNumber, thePath->getObject(unitIndex));

				if (NULL != theUnitIDNum)
				{
					unitID = theUnitIDNum->unsigned8BitValue();

					FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

					if (SELECTOR_UNIT == subType)
					{
						foundSelectorUnit = true;
						selectorUnitIndex = unitIndex;
						break;
					}
				}
			}

			foundFeatureUnit = false;

			if (foundSelectorUnit)
			{
				// <rdar://5366067> Find the feature unit closest to the selector.
				for (UInt8 unitIndex = numUnitsInPath - 2; unitIndex > selectorUnitIndex; unitIndex--)
				{
					theUnitIDNum = OSDynamicCast(OSNumber, thePath->getObject(unitIndex));

					if (NULL != theUnitIDNum)
					{
						unitID = theUnitIDNum->unsigned8BitValue();

						FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

						if (FEATURE_UNIT == subType)
						{
							FailIf(kIOReturnSuccess != mConfigDictionary->getNumChannels(&numChannels, interfaceNum, altSettingNum), Exit);

							for (UInt8 channelNum = 0; channelNum <= numChannels; channelNum++)
							{
								switch (controlTypeWanted)
								{
									case kVolumeControl:
										if (mConfigDictionary->channelHasVolumeControl(controlInterfaceNum, 0, unitID, channelNum))
										{
											featureUnitID	 = unitID;
											foundFeatureUnit = true;
										}
										break;

									case kMuteControl:
										if (mConfigDictionary->channelHasMuteControl(controlInterfaceNum, 0, unitID, channelNum))
										{
											featureUnitID	 = unitID;
											foundFeatureUnit = true;
										}
										break;
								}
							}
						}
					}
					else
					{
						DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "::getBestFeatureUnitInPath() - something is wrong here!!!");
					}
				} // for unitIndex
			}

			if (!foundFeatureUnit)
			{
				// <rdar://5366067> If selector is not present, this will find the feature unit closest to the input terminal.
				// Otherwise, it continues on from selector to the output terminal.
				UInt8		startingUnitIndex = foundSelectorUnit ? (selectorUnitIndex - 1) : (numUnitsInPath - 2);

				for (UInt8 unitIndex = startingUnitIndex; (unitIndex > 0) && (!foundFeatureUnit); unitIndex--)
				{
					theUnitIDNum = OSDynamicCast(OSNumber, thePath->getObject(unitIndex));

					if (NULL != theUnitIDNum)
					{
						unitID = theUnitIDNum->unsigned8BitValue();

						FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

						if (FEATURE_UNIT == subType)
						{
							FailIf(kIOReturnSuccess != mConfigDictionary->getNumChannels(&numChannels, interfaceNum, altSettingNum), Exit);

							for (UInt8 channelNum = 0; channelNum <= numChannels; channelNum++)
							{
								switch (controlTypeWanted)
								{
									case kVolumeControl:
										if (mConfigDictionary->channelHasVolumeControl(controlInterfaceNum, 0, unitID, channelNum))
										{
											featureUnitID	 = unitID;
											foundFeatureUnit = true;
										}
										break;

									case kMuteControl:
										if (mConfigDictionary->channelHasMuteControl(controlInterfaceNum, 0, unitID, channelNum))
										{
											featureUnitID	 = unitID;
											foundFeatureUnit = true;
										}
										break;
								}
							}
						}
					}
					else
					{
						DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "::getBestFeatureUnitInPath() - something is wrong here!!!");
					}
				} // for unitIndex
			}
			break;

		case kIOAudioControlUsageOutput:
		default:
			// Find the feature unit closest to the output terminal.
			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::getBestFeatureUnitInPath() - kIOAudioControlUsageOutput");
			for (UInt8 unitIndex = 1; (unitIndex < numUnitsInPath) && (!foundFeatureUnit); unitIndex++)
			{
				theUnitIDNum = OSDynamicCast(OSNumber, thePath->getObject(unitIndex));

				if (NULL != theUnitIDNum)
				{
					unitID = theUnitIDNum->unsigned8BitValue();

					FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

					if (FEATURE_UNIT == subType)
					{
						DP(DBG_INFO, "  examining feature unit %d ...", unitID);

						FailIf(kIOReturnSuccess != mConfigDictionary->getNumChannels(&numChannels, interfaceNum, altSettingNum), Exit);

						for (UInt8 channelNum = 0; channelNum <= numChannels; channelNum++)
						{
							switch (controlTypeWanted)
							{
								case kVolumeControl:
									if (mConfigDictionary->channelHasVolumeControl(controlInterfaceNum, 0, unitID, channelNum))
									{
										featureUnitID	 = unitID;
										foundFeatureUnit = true;
									}
									break;

								case kMuteControl:
									if (mConfigDictionary->channelHasMuteControl(controlInterfaceNum, 0, unitID, channelNum))
									{
										featureUnitID	 = unitID;
										foundFeatureUnit = true;
									}
									break;
							}
						}
					}
				}
				else
				{
					DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "::getBestFeatureUnitInPath() - something is wrong here!!!");
				}
			} // for unitIndex
			break;
	} // switch (direction)

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::getBestFeatureUnitInPath() = %d", this, featureUnitID);
	return featureUnitID;
}

/****************************************************************************/
/**
 * @brief	ボリュームコントロール生成．
 *			PdjUsbAudioTopology::doControlStuff() から呼び出される．
 *
 * @param[in]		*usbAudioEngine		PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		featureUnitID		Feature Unit のユニット ID．
 * @param[in]		terminalID			関連付けるターミナル ID．
 * @param[in]		interfaceNum		USB interface 番号．
 * @param[in]		altSettingNum		alternate setting 番号．
 * @param[in]		usage				入出力方向．
 */
// <rdar://6413207>
void
PdjUsbAudioTopology::addVolumeControls(
	PdjUsbAudioEngine		*usbAudioEngine,
	UInt8					featureUnitID,
	UInt8					terminalID,
	UInt8					interfaceNum,
	UInt8					altSettingNum,
	UInt32					usage
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::addVolumeControls(%p, %d, %d, %d, %d, %u)",
		usbAudioEngine, featureUnitID, terminalID, interfaceNum, altSettingNum, usage);

	OSArray					*inputGainControlsArray;
	OSArray					*passThruVolControlsArray;
	OSArray					*outputVolControlsArray;
	OSDictionary			*engineInfo;
	OSDictionary			*streamInfo;			// <rdar://6420832>
	SInt32					engineInfoIndex;
	SInt32					streamInfoIndex;	// <rdar://6420832>
	UInt8					controlInterfaceNum;
	UInt8					numControls;

	FailIf(NULL == mControlInterface, Exit);

	engineInfoIndex = this->getEngineInfoIndex(usbAudioEngine);
	FailIf(-1 == engineInfoIndex, Exit);

	// <rdar://6420832>
	streamInfoIndex = this->getStreamInfoIndex(interfaceNum);
	FailIf(-1 == streamInfoIndex, Exit);

	engineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(engineInfoIndex));
	FailIf(NULL == engineInfo, Exit);

	// <rdar://6420832>
	streamInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredStreams->getObject(streamInfoIndex));
	FailIf(NULL == streamInfo, Exit);

	inputGainControlsArray	 = NULL;
	passThruVolControlsArray = NULL;
	outputVolControlsArray	 = NULL;

	// remove mono controls array if adding volume controls for output
	if ((kIOAudioControlUsageOutput == usage) && (NULL != mMonoControlsArray))
	{
		mMonoControlsArray->release();
		mMonoControlsArray = NULL;
	}

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	FailIf(kIOReturnSuccess != mConfigDictionary->getNumControls(&numControls, controlInterfaceNum, 0, featureUnitID), Exit);

	for (UInt8 channelNum = 0; channelNum <= numControls; channelNum++)
	{
		bool	extraStep = false;

		if (mConfigDictionary->channelHasVolumeControl(controlInterfaceNum, 0, featureUnitID, channelNum))
		{
			SInt16		deviceCur;
			SInt16		deviceMin;
			SInt16		deviceMax;
			UInt16		volRes = 0;
			SInt32		controlCur;
			SInt32		controlMax;
			SInt32		controlMin;

			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Creating volume controls for channel %d", this, channelNum);

			FailIf(kIOReturnSuccess != this->getCurVolume(featureUnitID, channelNum, &deviceCur), Error);
			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - deviceCur = 0x%04x", this, deviceCur);

			FailIf(kIOReturnSuccess != this->getMinVolume(featureUnitID, channelNum, &deviceMin), Error);
			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - deviceMin = 0x%04x", this, deviceMin);

			FailIf(kIOReturnSuccess != this->getMaxVolume(featureUnitID, channelNum, &deviceMax), Error);
			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - deviceMax = 0x%04x", this, deviceMax);

			this->getVolumeResolution(featureUnitID, channelNum, &volRes);
			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - volRes = 0x%04x", this, volRes);

			// [rdar://4511427] Need to check volRes for class compliance (Audio Spec 5.2.2.4.3.2).
			FailIf(0 == volRes, Error);

			// [rdar://4228556] Unless the current volume is negative infinity, we should try to flush the volume out to the device.
			if ((SInt16)kNegativeInfinity != deviceCur)
			{
				// [rdar://5292769] If deviceCur lies outside the accepted range, flush out the deviceMin value.
				if ((deviceCur < deviceMin) || (deviceCur > deviceMax))
				{
					DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "::addVolumeControls() - deviceCur is not in volume range! Setting to deviceMin ...");
					deviceCur = deviceMin;
				}

				DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Attempting to set volume to current volume ...", this);

				FailIf(kIOReturnSuccess != this->setCurVolume(featureUnitID, channelNum, HostToUSBWord(deviceCur)), Error);

				DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Volume was set successfully.", this);
			}

			if ((SInt16)kNegativeInfinity == deviceMin)
			{
				DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - device violates USB 2.0 audio spec Section 5.2.5.7.2! Setting deviceMin to 0x8001 ...", this);
				deviceMin = (SInt16)0x8001;
				// Even though the firmware on the device is incorrect, we need to account for this loss of a control step by adding to the maximum later.
				extraStep = true;
			}

			IOFixed		deviceMinDB = this->convertUSBVolumeTodB(deviceMin);
			IOFixed		deviceMaxDB = this->convertUSBVolumeTodB(deviceMax);

			// [rdar://5118683] Create volume controls with correct ranges.

			controlMin = 0;
			// Also need to add an extra step here if the device reported negative infinity as the min value.
			controlMax = ((deviceMax - deviceMin) / volRes) + (extraStep ? 1 : 0);
			// The current control value is a bit of a special case because the device may violate the spec. The rules are as follows:
			// * If the current value is negative infinity, use a control value of 0.
			// * If the current value is the minimum value, use a control value of 0.
			// * If the current value is the maximum value, use the maximum control value (may be altered by an extra step if the device breaks spec).
			// * Otherwise calculate the current value as normal.
			if (((SInt16)kNegativeInfinity == deviceCur) || (deviceCur == deviceMin))
			{
				controlCur = 0;
			}
			else if (deviceCur == deviceMax)
			{
				controlCur = controlMax;
			}
			else
			{
				// If this device breaks spec, we don't care if this value matches with what the control published.
				controlCur = ((deviceCur - deviceMin) / volRes);
			}

			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Creating control... [%d : %d : %d]", this, controlMin, controlCur, controlMax);

			// <rdar://6413207> Add the terminal ID to the control ID so that it is unique.
			IOAudioLevelControl		*theLevelControl = IOAudioLevelControl::createVolumeControl(
																				controlCur,
																				controlMin,
																				controlMax,
																				deviceMinDB,
																				deviceMaxDB,
																				channelNum,
																				0,
																				(terminalID << 8) | featureUnitID,
																				usage
																			);
			FailIf(NULL == theLevelControl, Exit);

			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Created control %p", this, theLevelControl);

			theLevelControl->setValueChangeHandler(controlChangedHandler, this);
			usbAudioEngine->addDefaultAudioControl(theLevelControl);

			switch (usage)
			{
				case kIOAudioControlUsageInput:
					if (NULL == inputGainControlsArray)
					{
						inputGainControlsArray = OSArray::withObjects((const OSObject **)&theLevelControl, 1);
					}
					else
					{
						inputGainControlsArray->setObject(theLevelControl);
					}
					break;

				case kIOAudioControlUsagePassThru:
					if (NULL == passThruVolControlsArray)
					{
						passThruVolControlsArray = OSArray::withObjects((const OSObject **)&theLevelControl, 1);
					}
					else
					{
						passThruVolControlsArray->setObject(theLevelControl);
					}
					break;

				case kIOAudioControlUsageOutput:
					if (NULL == outputVolControlsArray)
					{
						outputVolControlsArray = OSArray::withObjects((const OSObject **)&theLevelControl, 1);
					}
					else
					{
						outputVolControlsArray->setObject(theLevelControl);
					}

					// add channel number to mono output controls array if necessary
					if (mDeviceIsInMonoMode)
					{
						OSNumber	*number = OSNumber::withNumber(channelNum, 8);

						if (NULL == mMonoControlsArray)
						{
							mMonoControlsArray = OSArray::withObjects((const OSObject **)&number, 1);
						}
						else
						{
							mMonoControlsArray->setObject(number);
						}

						DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Added channel %d to mono controls array", this, channelNum);

						number->release();
					}
					break;

				default:
					DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Control %p has an unknown usage!", this, theLevelControl);
			}

			theLevelControl->release();
		}
		else
		{
			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Channel %d has no volume controls; skipping ...", this, channelNum);
		}
		goto NoError;
Error:
		DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Error creating controls for channel %d!", this, channelNum);
NoError:
		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::addVolumeControls() - Done with channel %d", this, channelNum);
	}

	if (NULL != inputGainControlsArray)
	{
		streamInfo->setObject(kInputGainControls, inputGainControlsArray);		// <rdar://6420832>
		inputGainControlsArray->release();
	}

	if (NULL != passThruVolControlsArray)
	{
		streamInfo->setObject(kPassThruVolControls, passThruVolControlsArray);	// <rdar://6420832>
		passThruVolControlsArray->release();
	}

	if (NULL != outputVolControlsArray)
	{
		streamInfo->setObject(kOutputVolControls, outputVolControlsArray);		// <rdar://6420832>
		outputVolControlsArray->release();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::addVolumeControls(%p, %d, %d, %d, %d, %u)",
		usbAudioEngine, featureUnitID, terminalID, interfaceNum, altSettingNum, usage);
	return;
}

/****************************************************************************/
/**
 * @brief	ミュートコントロール生成．
 *			PdjUsbAudioTopology::doControlStuff() から呼び出される．
 *
 * @param[in]		*usbAudioEngine		PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		featureUnitID		Feature Unit のユニット ID．
 * @param[in]		terminalID			関連付けるターミナル ID．
 * @param[in]		interfaceNum		USB interface 番号．
 * @param[in]		altSettingNum		alternate setting 番号．
 * @param[in]		usage				入出力方向．
 */
// <rdar://6413207>
void
PdjUsbAudioTopology::addMuteControl(
	PdjUsbAudioEngine		*usbAudioEngine,
	UInt8					featureUnitID,
	UInt8					terminalID,
	UInt8					interfaceNum,
	UInt8					altSettingNum,
	UInt32					usage
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::addMuteControl(%p, %d, %d, %d, %d, 0%x)",
		this, usbAudioEngine, featureUnitID, terminalID, interfaceNum, altSettingNum, usage);

	OSArray					*inputMuteControlsArray;
	OSArray					*outputMuteControlsArray;
	OSArray					*passThruToggleControlsArray;	// <rdar://5366067>
	OSDictionary			*engineInfo;
	OSDictionary			*streamInfo;					// <rdar://6420832>
	SInt32					engineInfoIndex;
	SInt32					streamInfoIndex;				// <rdar://6420832>
	UInt8					controlInterfaceNum;
	UInt8					numControls;

	FailIf(NULL == mControlInterface, Exit);

	engineInfoIndex = this->getEngineInfoIndex(usbAudioEngine);
	FailIf(-1 == engineInfoIndex, Exit);

	// <rdar://6420832>
	streamInfoIndex = this->getStreamInfoIndex(interfaceNum);
	FailIf(-1 == streamInfoIndex, Exit);

	engineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(engineInfoIndex));
	FailIf(NULL == engineInfo, Exit);

	// <rdar://6420832>
	streamInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredStreams->getObject(streamInfoIndex));
	FailIf(NULL == streamInfo, Exit);

	inputMuteControlsArray		= NULL;
	outputMuteControlsArray		= NULL;		// <rdar://6413207>
	passThruToggleControlsArray	= NULL;		// <rdar://5366067>

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	FailIf(kIOReturnSuccess != mConfigDictionary->getNumControls(&numControls, controlInterfaceNum, 0, featureUnitID), Exit);

	for (UInt8 channelNum = 0; channelNum <= numControls; channelNum++)
	{
		if (mConfigDictionary->channelHasMuteControl(controlInterfaceNum, 0, featureUnitID, channelNum))
		{
			SInt16					deviceCur = 0;
			IOAudioToggleControl	*theMuteControl;

			this->getCurMute(featureUnitID, channelNum, &deviceCur);
			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::addMuteControl() - channel %d, feature unit %d has current mute value 0x%x",
				channelNum, featureUnitID, deviceCur);

			// [rdar://5292769] Force unmute the device since the only easy way to unmute it afterwards is Audio MIDI Setup
			if ((0 != deviceCur) && (kIOAudioControlUsageInput == usage))
			{
				DP(DBG_ERROR, "! " LOGSTRAUDIOTOPOLOGY "::addMuteControl() - forcing channel %d of this device to unmute in hardware.", channelNum);
				deviceCur = 0x0;
			}

			// <rdar://6413207> Add the terminal ID to the control ID so that it is unique.
			theMuteControl = IOAudioToggleControl::createMuteControl(deviceCur, channelNum, 0, (terminalID << 8) | featureUnitID, usage);
			FailIf(NULL == theMuteControl, Exit);

			this->setCurMute(featureUnitID, channelNum, HostToUSBWord(deviceCur));

			theMuteControl->setValueChangeHandler(controlChangedHandler, this);

			usbAudioEngine->addDefaultAudioControl(theMuteControl);

			switch (usage)
			{
				case kIOAudioControlUsageInput:
					if (NULL == inputMuteControlsArray)
					{
						inputMuteControlsArray = OSArray::withObjects((const OSObject **)&theMuteControl, 1);
					}
					else
					{
						inputMuteControlsArray->setObject(theMuteControl);
					}
					break;

				case kIOAudioControlUsagePassThru:
					// <rdar://5366067>
					if (NULL == passThruToggleControlsArray)
					{
						passThruToggleControlsArray = OSArray::withObjects((const OSObject **)&theMuteControl, 1);
					}
					else
					{
						passThruToggleControlsArray->setObject(theMuteControl);
					}
					break;

				case kIOAudioControlUsageOutput:
					// <rdar://6413207>
					if (NULL == outputMuteControlsArray)
					{
						outputMuteControlsArray = OSArray::withObjects((const OSObject **)&theMuteControl, 1);
					}
					else
					{
						outputMuteControlsArray->setObject(theMuteControl);
					}
					break;
			}

			theMuteControl->release();
		}
	}

	if (NULL != inputMuteControlsArray)
	{
		streamInfo->setObject(kInputMuteControls, inputMuteControlsArray);	// <rdar://6420832>
		inputMuteControlsArray->release();
	}

	// <rdar://5366067>
	if (NULL != passThruToggleControlsArray)
	{
		streamInfo->setObject(kPassThruToggleControls, passThruToggleControlsArray);	// <rdar://6420832>
		passThruToggleControlsArray->release();
	}

	// <rdar://6413207>
	if (NULL != outputMuteControlsArray)
	{
		streamInfo->setObject(kOutputMuteControls, outputMuteControlsArray);	// <rdar://6420832>
		outputMuteControlsArray->release();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::addMuteControl(%p, %d, %d, %d, %d, 0x%x)",
		this, usbAudioEngine, featureUnitID, terminalID, interfaceNum, altSettingNum, usage);
	return;
}

/****************************************************************************/
/**
 * @brief	指定のセレクタユニット設定に対応するオーディオパスのインデックスを取得．
 *
 * @param[in]		*arrayOfPathsFromOutputTerminal		セレクタユニットを含むオーディオパスの OSArray．
 * @param[in]		pathsToOutputTerminalN				mControlGraph 内での arrayOfPathsFromOutputTerminal のインデックス．
 * @param[in]		graphPathIndex						arrayOfPathsFromOutputTerminal 内でのオーディオパスのインデックス．
 * @param[in]		selectorUnitIndex					オーディオパス内でのセレクタユニットのインデックス．
 * @param[in]		selectorSetting						セレクタユニットの設定値．
 *
 * @return			指定のセレクタユニット設定に対応するオーディオパスのインデックス．
 */
// <rdar://6523676> Get the path index corresponding to the selector setting
// <rdar://problem/6656791>
UInt32
PdjUsbAudioTopology::getPathIndexForSelectorSetting(
	OSArray			*arrayOfPathsFromOutputTerminal,
	UInt32			pathsToOutputTerminalN,
	UInt32			graphPathIndex,
	UInt8			selectorUnitIndex,
	UInt8			selectorSetting
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getPathIndexForSelectorSetting()");

	UInt32			pathIndex;
	UInt8			controlInterfaceNum;
	OSArray			*aPath;
	OSNumber		*theUnitIDNum;
	UInt8			selectorID;
	UInt8			numSelectorSources;

	pathIndex = graphPathIndex;	// <rdar://problem/6656791>

	FailIf(NULL == mControlInterface, Exit);

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(graphPathIndex));	// <rdar://problem/6656791>
	FailIf(NULL == aPath, Exit);

	theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(selectorUnitIndex));
	FailIf(NULL == theUnitIDNum, Exit);

	selectorID = theUnitIDNum->unsigned8BitValue();

	FailIf(kIOReturnSuccess != mConfigDictionary->getNumSources(&numSelectorSources, controlInterfaceNum, 0, selectorID), Exit);

	for (UInt32 selectorSourceIndex = 0; selectorSourceIndex < numSelectorSources; selectorSourceIndex++)
	{
		OSString	*nameString;

		nameString = this->getNameForPath(arrayOfPathsFromOutputTerminal, &pathIndex, selectorUnitIndex + 1);

		if (NULL != nameString)
		{
			nameString->release();
		}

		if ((selectorSourceIndex + 1) == selectorSetting)
		{
			// Found the path index that correspond to the selector setting
			pathIndex--;
			break;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getPathIndexForSelectorSetting() = %u", pathIndex);
	return pathIndex;
}

/****************************************************************************/
/**
 * @brief	セレクタコントロールへの入力を登録．
 *			PdjUsbAudioTopology::doControlStuff() から呼び出される．
 *
 * @param[in]		*theSelectorControl					セレクタコントロール．
 * @param[in]		*arrayOfPathsFromOutputTerminal		セレクタユニットを含むオーディオパスの OSArray．
 * @param[in]		pathsToOutputTerminalN				mControlGraph 内での arrayOfPathsFromOutputTerminal のインデックス．
 * @param[in]		graphPathIndex						arrayOfPathsFromOutputTerminal 内でのオーディオパスのインデックス．
 * @param[in]		selectorIndex						オーディオパス内でのセレクタユニットのインデックス．
 *
 * @return			IOReturn コード．
 */
// <rdar://problem/6656791>
IOReturn
PdjUsbAudioTopology::addSelectorSourcesToSelectorControl(
	IOAudioSelectorControl	*theSelectorControl,
	OSArray					*arrayOfPathsFromOutputTerminal,
	UInt32					pathsToOutputTerminalN,
	UInt32					graphPathIndex,
	UInt8					selectorIndex
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::addSelectorSourcesToSelectorControl()");

	UInt8					controlInterfaceNum;
	OSArray					*aPath;
	OSNumber				*theUnitIDNum;
	UInt8					selectorID;
	UInt32					pathIndex;
	UInt8					numSelectorSources;

	FailIf(NULL == mControlInterface, Exit);

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(graphPathIndex));	// <rdar://problem/6656791>
	FailIf(NULL == aPath, Exit);

	theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(selectorIndex));
	FailIf(NULL == theUnitIDNum, Exit);

	selectorID = theUnitIDNum->unsigned8BitValue();

	pathIndex = graphPathIndex;	// <rdar://problem/6656791>

	FailIf(kIOReturnSuccess != mConfigDictionary->getNumSources(&numSelectorSources, controlInterfaceNum, 0, selectorID), Exit);

	for (UInt32 selectorSourceIndex = 0; selectorSourceIndex < numSelectorSources; selectorSourceIndex++)
	{
		OSString	*nameString = this->getNameForPath(arrayOfPathsFromOutputTerminal, &pathIndex, selectorIndex + 1);

		if (NULL != nameString)
		{
			UInt32		selection =   (0xFF000000 & (pathsToOutputTerminalN		<< 24))
									| (0x00FF0000 & ((pathIndex - 1)			<< 16))
									| (0x0000FF00 & (selectorID					<<  8))
									| (0x000000FF & ((selectorSourceIndex + 1)	<<  0));

			theSelectorControl->addAvailableSelection(selection, nameString);

			nameString->release();
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::addSelectorSourcesToSelectorControl()");
	return kIOReturnSuccess;
}

/****************************************************************************/
/**
 * @brief	入出力ターミナルがいずれも USB streaming でないパスを列挙．
 *
 * @param[in]		inputTerminalID		入力ターミナル ID．
 *
 * @return			条件に合致するオーディオパスの OSArray．
 */
// This should detect a playthrough path; which is a non-streaming input terminal connected to a non-streaming output terminal.
// <rdar://5366067>
OSArray *
PdjUsbAudioTopology::getPlaythroughPaths(
	UInt8			inputTerminalID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getPlaythroughPaths()");

	OSArray			*playThroughPaths = NULL;
	OSArray			*aPath;
	OSNumber		*theUnitIDNum;
	UInt32			numOutputTerminalArrays;
	UInt16			terminalType;
	UInt8			controlInterfaceNum;
	UInt8			unitID;

	FailIf(NULL == mControlInterface, Exit);

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	numOutputTerminalArrays = mControlGraph->getCount();

	for (UInt32 pathsToOutputTerminalN = 0; pathsToOutputTerminalN < numOutputTerminalArrays; pathsToOutputTerminalN++)
	{
		OSArray		*arrayOfPathsFromOutputTerminal = OSDynamicCast(OSArray, mControlGraph->getObject(pathsToOutputTerminalN));
		FailIf(NULL == arrayOfPathsFromOutputTerminal, Exit);

		aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(0));
		FailIf(NULL == aPath, Exit);

		theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(0));
		FailIf(NULL == theUnitIDNum, Exit);

		unitID = theUnitIDNum->unsigned8BitValue();

		FailIf(kIOReturnSuccess != mConfigDictionary->getOutputTerminalType(&terminalType, controlInterfaceNum, 0, unitID), Exit);

		// only looking for non-streaming outputs
		if (USB_STREAMING == terminalType)
		{
			continue;
		}

		UInt32		numPathsFromOutputTerminal = arrayOfPathsFromOutputTerminal->getCount();

		for (UInt32 pathNumber = 0; pathNumber < numPathsFromOutputTerminal; pathNumber++)
		{
			aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(pathNumber));
			FailIf(NULL == aPath, Exit);

			theUnitIDNum = OSDynamicCast(OSNumber, aPath->getLastObject());
			FailIf(NULL == theUnitIDNum, Exit);

			unitID = theUnitIDNum->unsigned8BitValue();

			// <rdar://5366067> only looking for outputs attached to the current input.
			if (unitID != inputTerminalID)
			{
				continue;
			}

			FailIf(kIOReturnSuccess != mConfigDictionary->getInputTerminalType(&terminalType, controlInterfaceNum, 0, unitID), Exit);

			if (terminalType != USB_STREAMING)
			{
				if (NULL == playThroughPaths)
				{
					playThroughPaths = OSArray::withObjects((const OSObject **)&aPath, 1);
				}
				else
				{
					playThroughPaths->setObject(aPath);
				}
			}
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getPlaythroughPaths()");
	return playThroughPaths;
}

/****************************************************************************/
/**
 * @brief	入出力ターミナルがいずれも USB streaming でないパスをセットアップ．
 *
 * @param[in]		*usbAudioEngine		PdjUsbAudioEngine クラスのインスタンス．
 * @param[in]		*playThroughPaths	入出力ターミナルがいずれも USB streaming でないパスの OSArray．
 * @param[in]		interfaceNum		USB interface 番号．
 * @param[in]		altSettingNum		alternate setting 番号．
 * @param[in]		inputTerminalID		入力ターミナル ID．
 *
 * @return			IOReturn コード．
 */
// <rdar://5366067>
IOReturn
PdjUsbAudioTopology::doPlaythroughSetup(
	PdjUsbAudioEngine		*usbAudioEngine,
	OSArray					*playThroughPaths,
	UInt8					interfaceNum,
	UInt8					altSettingNum,
	UInt8					inputTerminalID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::doPlaythroughSetup()");

	IOReturn				result = kIOReturnError;
	UInt8					controlInterfaceNum;
	SInt32					engineInfoIndex;
	SInt32					streamInfoIndex;					// <rdar://6420832>
	OSDictionary			*engineInfo;
	OSDictionary			*streamInfo;						// <rdar://6420832>
	UInt32					numPlayThroughPaths;
	OSArray					*aPath;
	OSNumber				*theUnitIDNum;
	UInt8					outputTerminalID;					// <rdar://5366067>
	UInt32					defaultPathIndex = 0;				// <rdar://5366067>
	UInt16					defaultPathTerminalType = 0;		// <rdar://5366067>
	UInt16					terminalType;
	UInt8					featureUnitID;

	FailIf(NULL == mControlInterface, Exit);

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	engineInfoIndex = this->getEngineInfoIndex(usbAudioEngine);
	FailIf(-1 == engineInfoIndex, Exit);

	// <rdar://6420832>
	streamInfoIndex = this->getStreamInfoIndex(interfaceNum);
	FailIf(-1 == streamInfoIndex, Exit);

	engineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(engineInfoIndex));
	FailIf(NULL == engineInfo, Exit);

	// <rdar://6420832>
	streamInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredStreams->getObject(streamInfoIndex));
	FailIf(NULL == streamInfo, Exit);

	streamInfo->setObject(kPassThruPathsArray, playThroughPaths);

	numPlayThroughPaths = playThroughPaths->getCount();

	if (numPlayThroughPaths > 0)	// <rdar://5366067>
	{
		// <rdar://5366067> Find the default path index according to the following priority order: Speaker, Headphone, Line Out, ...
		for (UInt32 pathIndex = 0; pathIndex < numPlayThroughPaths; pathIndex++)
		{
			aPath = OSDynamicCast(OSArray, playThroughPaths->getObject(pathIndex));
			FailIf(NULL == aPath, Exit);

			theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(0));
			FailIf(NULL == theUnitIDNum, Exit);

			outputTerminalID = theUnitIDNum->unsigned8BitValue();

			FailIf(
					kIOReturnSuccess != mConfigDictionary->getOutputTerminalType(
																&terminalType,
																controlInterfaceNum,
																0,
																outputTerminalID
														   ),
					Exit
			);

			switch (defaultPathTerminalType)
			{
				case 0:
					defaultPathTerminalType	= terminalType;
					defaultPathIndex		= pathIndex;
					break;

				case 0x0301: // Speaker
					break;

				case 0x0302: // Headphones
					if (0x0301 == terminalType)
					{
						defaultPathTerminalType	= terminalType;
						defaultPathIndex		= pathIndex;
					}
					break;

				case 0x0603: // Line Out
					if ((0x0301 == terminalType) || (0x0302 == terminalType))
					{
						defaultPathTerminalType	= terminalType;
						defaultPathIndex		= pathIndex;
					}
					break;
			}
		}

		if (numPlayThroughPaths > 1)	// <rdar://6500500>
		{
			IOAudioSelectorControl	*playThroughSelector;

			// Create a virtual selector to manipulate the mutes on the feature units to toggle through playthrough sources.
			playThroughSelector = IOAudioSelectorControl::create(
																defaultPathIndex,
																kIOAudioControlChannelIDAll,
																0,
																(inputTerminalID << 24) | (streamInfoIndex << 16) | ( engineInfoIndex << 8 ) | 0,
																kIOAudioSelectorControlSubTypeDestination,
																kIOAudioControlUsagePassThru
														  );	// <rdar://6420832>, <rdar://5366067>, <rdar://6523676>
			FailIf(NULL == playThroughSelector, Exit);

			playThroughSelector->setValueChangeHandler(controlChangedHandler, this);

			usbAudioEngine->addDefaultAudioControl(playThroughSelector);

			for (UInt32 pathIndex = 0; pathIndex < numPlayThroughPaths; pathIndex++)
			{
				aPath = OSDynamicCast(OSArray, playThroughPaths->getObject(pathIndex));
				FailIf(NULL == aPath, Exit);

				featureUnitID = this->getBestFeatureUnitInPath(aPath, kIOAudioControlUsagePassThru, interfaceNum, altSettingNum, kMuteControl);

				if (featureUnitID)
				{
					UInt8		stringIndex;										// <rdar://6394629>
					char		stringBuffer[kStringBufferSize];					// <rdar://6394629>
					OSString	*nameString;

					theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(0));	// <rdar://5366067>
					FailIf(NULL == theUnitIDNum, Exit);

					outputTerminalID = theUnitIDNum->unsigned8BitValue();			// <rdar://5366067>

					FailIf(
							kIOReturnSuccess != mConfigDictionary->getOutputTerminalType(
																		&terminalType,
																		controlInterfaceNum,
																		0,
																		outputTerminalID
																   ),
							Exit
					);

					// <rdar://6394629> Get the name from the terminal if it exists.
					stringIndex = 0;

					if (
							(
								kIOReturnSuccess == mConfigDictionary->getStringIndex(
																			&stringIndex,
																			controlInterfaceNum,
																			0,
																			outputTerminalID
																	   )
							)	// <rdar://5366067>
					   &&
							(0 != stringIndex)
					   &&
							(
								kIOReturnSuccess == mAudioDevice->getStringDescriptor(
																		stringIndex,
																		stringBuffer,
																		kStringBufferSize
																  )
							)
					   )
					{
						nameString = OSString::withCString(stringBuffer);
					}
					else
					{
						nameString = OSString::withCString(mAudioDevice->terminalTypeString(terminalType));
					}

					FailIf(NULL == nameString, Exit);

					playThroughSelector->addAvailableSelection(pathIndex, nameString);

					nameString->release();
				}
			}

			streamInfo->setObject(kPassThruSelectorControl, playThroughSelector);	// <rdar://5366067>
#ifdef BUG_FIX
			playThroughSelector->release();
			playThroughSelector = NULL;
#endif
		}

		aPath = OSDynamicCast(OSArray, playThroughPaths->getObject(defaultPathIndex));	// <rdar://5366067>
		FailIf(NULL == aPath, Exit);

		featureUnitID = this->getBestFeatureUnitInPath(
									aPath,
									kIOAudioControlUsagePassThru,
									interfaceNum,
									altSettingNum,
									kVolumeControl
							  );	// <rdar://5366067>
		if (featureUnitID)
		{
			// Create the playthrough volume controls
			this->addVolumeControls(
						usbAudioEngine,
						featureUnitID,
						inputTerminalID,
						interfaceNum,
						altSettingNum,
						kIOAudioControlUsagePassThru
				  );	// <rdar://6413207>
		}

		featureUnitID = this->getBestFeatureUnitInPath(
									aPath,
									kIOAudioControlUsagePassThru,
									interfaceNum,
									altSettingNum,
									kMuteControl
							  );	// <rdar://5366067>
		if (featureUnitID)
		{
			this->addMuteControl(
						usbAudioEngine,
						featureUnitID,
						inputTerminalID,
						interfaceNum,
						altSettingNum,
						kIOAudioControlUsagePassThru
				  );	// <rdar://6413207>
		}

		result = kIOReturnSuccess;
	} // if (numPlayThroughPaths > 0)
	else
	{
		// No playthrough path.
		result = kIOReturnSuccess;
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::doPlaythroughSetup()");
	return result;
}

/****************************************************************************/
/**
 * @brief	指定のユニットを通るオーディオパスの総数を数える．
 *
 * @param[in]		unitID			ユニット ID．
 *
 * @return			指定のユニットを通るオーディオパスの総数．
 */
UInt8
PdjUsbAudioTopology::pathsContaining(
	UInt8			unitID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::pathsContaining()");

	UInt8			numPaths = 0;

	if (mControlGraph)
	{
		for (UInt8 pathGroupIndex = 0; pathGroupIndex < mControlGraph->getCount(); pathGroupIndex++)
		{
			OSArray		*thisPathGroup = OSDynamicCast(OSArray, mControlGraph->getObject(pathGroupIndex));

			if (thisPathGroup)
			{
				for (UInt8 pathIndex = 0; pathIndex < thisPathGroup->getCount(); pathIndex++)
				{
					OSArray		*thisPath = OSDynamicCast(OSArray, thisPathGroup->getObject(pathIndex));

					if (thisPath)
					{
						for (UInt8 unitIndex = 0; unitIndex < thisPath->getCount(); unitIndex++)
						{
							OSNumber	*unitNumber = OSDynamicCast(OSNumber, thisPath->getObject(unitIndex));

							if (unitID == unitNumber->unsigned8BitValue())
							{
								numPaths++;
								// Circular topologies are not allowed.
								break;
							}
						}
					}
				}
			}
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::pathsContaining() = %d", numPaths);
	return numPaths;
}

/****************************************************************************/
/**
 * @brief	指定の Feature Unit を通り、且つ指定の Mixer Unit を通らないオーディオパスの総数を数える．
 *
 * @param[in]		featureUnitID	通るべき Feature Unit のユニット ID．
 * @param[in]		mixerUnitID		通ってはいけない Mixer Unit のユニット ID．
 *
 * @return			指定の Feature Unit を通り、且つ指定の Mixer Unit を通らないオーディオパスの総数．
 */
// <rdar://5366067>
UInt8
PdjUsbAudioTopology::pathsContainingFeatureUnitButNotMixerUnit(
	UInt8			featureUnitID,
	UInt8			mixerUnitID
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::pathsContainingFeatureUnitButNotMixerUnit()");

	UInt8			numPaths = 0;

	if (mControlGraph)
	{
		for (UInt8 pathGroupIndex = 0; pathGroupIndex < mControlGraph->getCount(); pathGroupIndex++)
		{
			OSArray		*thisPathGroup = OSDynamicCast(OSArray, mControlGraph->getObject(pathGroupIndex));

			if (thisPathGroup)
			{
				for (UInt8 pathIndex = 0; pathIndex < thisPathGroup->getCount(); pathIndex++)
				{
					OSArray		*thisPath = OSDynamicCast(OSArray, thisPathGroup->getObject(pathIndex));

					if (thisPath)
					{
						bool	foundFeatureUnit = false;
						bool	foundMixerUnit	 = false;

						for (UInt8 unitIndex = 0; unitIndex < thisPath->getCount(); unitIndex++)
						{
							OSNumber	*unitNumber = OSDynamicCast(OSNumber, thisPath->getObject(unitIndex));

							if (NULL != unitNumber)
							{
								UInt8		unitID = unitNumber->unsigned8BitValue();

								if (featureUnitID == unitID)
								{
									foundFeatureUnit = true;
								}
								else if (mixerUnitID == unitID)
								{
									foundMixerUnit = true;
								}
							}
						}

						if (foundFeatureUnit && !foundMixerUnit)
						{
							numPaths++;
						}
					}
				}
			}
		}
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::pathsContainingFeatureUnitButNotMixerUnit() = %d", numPaths);
	return numPaths;
}

/****************************************************************************/
/**
 * @brief	オーディオパスの名称を作成．
 *			パスを走査し、見つかったターミナル名またはその組み合わせによって名称を決定する．
 *
 * @param[in]		*arrayOfPathsFromOutputTerminal		名付けたいパスを含む OSArray．
 * @param[in, out]	*pathIndex							arrayOfPathsFromOutputTerminal 内での名付けたいパスのインデックス．
 *														return 時には、次に参照すべきインデックスに書き換えられる．
 * @param[in]		startingPoint						パス内での走査開始位置．
 *
 * @return			オーディオパスの名称．
 */
// Starting point is the array index of the element after the selector unit.
OSString *
PdjUsbAudioTopology::getNameForPath(
	OSArray			*arrayOfPathsFromOutputTerminal,
	UInt32			*pathIndex,
	UInt8			startingPoint
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getNameForPath()");

	OSString		*theString = NULL;
	OSArray			*aPath;
	UInt32			numElementsInPath;
	UInt8			controlInterfaceNum;
	UInt8			unitID;
	UInt16			terminalType;
	UInt8			subType;
	bool			done = false;

	FailIf(NULL == mControlInterface, Exit);

	aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(*pathIndex));
	FailIf(NULL == aPath, Exit);

	numElementsInPath = aPath->getCount();

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	for (UInt32 elementIndex = startingPoint; (elementIndex < numElementsInPath) && (!done); elementIndex++)
	{
		OSNumber	*theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(elementIndex));
		FailIf(NULL == theUnitIDNum, Exit);

		unitID = theUnitIDNum->unsigned8BitValue();

		FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

		switch (subType)
		{
			case INPUT_TERMINAL:
				FailIf(
						kIOReturnSuccess != mConfigDictionary->getInputTerminalType(
																	&terminalType,
																	controlInterfaceNum,
																	0,
																	unitID
															   ),
						Exit
				);

				// <rdar://6394629> Get the string from the terminal if it exists.
				if (USB_STREAMING != terminalType)
				{
					theString = this->getNameForTerminal(unitID, kIOAudioStreamDirectionInput);	// <rdar://6430836>, <rdar://6413207>
				}

				(*pathIndex)++;
				break;

			case MIXER_UNIT:
				theString = this->getNameForMixerPath(arrayOfPathsFromOutputTerminal, pathIndex, elementIndex);
				done = true;
				break;
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getNameForPath()");
	return theString;
}

/****************************************************************************/
/**
 * @brief	ミキサーユニットによって分岐したパスの名称を作成．
 *			パスを走査し、見つかったターミナル名の組み合わせによって名称を決定する．
 *
 * @param[in]		*arrayOfPathsFromOutputTerminal		名付けたいパスを含む OSArray．
 * @param[in, out]	*pathIndex							arrayOfPathsFromOutputTerminal 内での名付けたいパスのインデックス．
 *														return 時には、次に参照すべきインデックスに書き換えられる．
 * @param[in]		startingPoint						パス内での走査開始位置．
 *
 * @return			オーディオパスの名称．
 */
// Starting point is the array index of the mixer unit.
OSString *
PdjUsbAudioTopology::getNameForMixerPath(
	OSArray			*arrayOfPathsFromOutputTerminal,
	UInt32			*pathIndex,
	UInt8			startingPoint
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getNameForMixerPath()");

	char			string[255];
	OSArray			*aPath;
	OSNumber		*theUnitIDNum;
	UInt8			unitID;
	UInt32			numElementsInPath;
	UInt8			controlInterfaceNum;
	UInt32			mixerSourceIndex;
	UInt8			numMixerSources;
	UInt8			subType;
	OSString		*tempString;

	string[0] = '\0';

	FailIf(NULL == mControlInterface, Exit);

	aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(*pathIndex));
	FailIf(NULL == aPath, Exit);

	theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(startingPoint));
	FailIf(NULL == theUnitIDNum, Exit);

	unitID = theUnitIDNum->unsigned8BitValue();

	numElementsInPath = aPath->getCount();
	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	FailIf(kIOReturnSuccess != mConfigDictionary->getNumSources(&numMixerSources, controlInterfaceNum, 0, unitID), Exit);

	for (mixerSourceIndex = *pathIndex; mixerSourceIndex < (*pathIndex + numMixerSources); /* mixerSourceIndex incremented elsewhere */)
	{
		for (UInt32 elementIndex = startingPoint + 1; elementIndex < numElementsInPath; elementIndex++)
		{
			theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(elementIndex));
			FailIf(NULL == theUnitIDNum, Exit);

			unitID = theUnitIDNum->unsigned8BitValue();
			FailIf(kIOReturnSuccess != mConfigDictionary->getSubType(&subType, controlInterfaceNum, 0, unitID), Exit);

			switch (subType)
			{
				case INPUT_TERMINAL:
					tempString = this->getNameForPath(arrayOfPathsFromOutputTerminal, &mixerSourceIndex, elementIndex);
					if (NULL != tempString)
					{
						::strlcat(string, tempString->getCStringNoCopy(), sizeof(string));
						::strlcat(string, " & ", sizeof(string));
						tempString->release();
					}
					break;

				case MIXER_UNIT:
					tempString = this->getNameForMixerPath(arrayOfPathsFromOutputTerminal, &mixerSourceIndex, elementIndex);
					if (NULL != tempString)
					{
						::strlcat(string, tempString->getCStringNoCopy(), sizeof(string));
						tempString->release();
					}
					break;
			}
		}
	}

	*pathIndex = mixerSourceIndex;

	// 余分な " & " を削除
	if (::strlen(string) > 3)
	{
		string[::strlen(string) - 3] = '\0';
	}

Exit:
	OSString		*theString = OSString::withCString(string);

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getNameForMixerPath()");
	return theString;
}

/****************************************************************************/
/**
 * @brief	ターミナルの名称を作成．
 *			ストリングディスクリプタがあればそれを使用する．
 *			なければターミナルタイプに応じて標準的な名称を与える．
 *
 * @param[in]		terminalID			ターミナル ID．
 * @param[in]		direction			kIOAudioStreamDirectionOutput または kIOAudioStreamDirectionInput．
 *
 * @return			ターミナル名．
 */
// <rdar://6430836>
OSString *
PdjUsbAudioTopology::getNameForTerminal(
	UInt8			terminalID,
	UInt8			direction
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "::getNameForTerminal()");

	OSString		*theString = NULL;
	OSString		*tempString;
	char			stringBuffer[kStringBufferSize];	// <rdar://6394629>
	UInt8			stringIndex = 0;
	UInt8			controlInterfaceNum;
	UInt16			terminalType;

	FailIf(NULL == mControlInterface, Exit);

	controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

	// <rdar://6413207> Get the terminal type based on the direction of the terminal.
	if (kIOAudioStreamDirectionOutput == direction)
	{
		FailIf(kIOReturnSuccess != mConfigDictionary->getOutputTerminalType(&terminalType, controlInterfaceNum, 0, terminalID), Exit);
	}
	else
	{
		FailIf(kIOReturnSuccess != mConfigDictionary->getInputTerminalType(&terminalType, controlInterfaceNum, 0, terminalID), Exit);
	}

	if (
			(kIOReturnSuccess == mConfigDictionary->getStringIndex(&stringIndex, controlInterfaceNum, 0, terminalID))
	   &&
			(0 != stringIndex)
	   &&
			(kIOReturnSuccess == mAudioDevice->getStringDescriptor(stringIndex, stringBuffer, kStringBufferSize))
	   )
	{
		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::getNameForTerminal() - terminalID = %d, stringIndex = %d, stringBuffer = %s",
			terminalID, stringIndex, stringBuffer);

		tempString = OSString::withCString(stringBuffer);
	}
	else
	{
		tempString = OSString::withCString(mAudioDevice->terminalTypeString(terminalType));
	}

	FailIf(NULL == tempString, Exit);

	theString = OSString::withString(tempString);
	tempString->release();

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "::getNameForTerminal()");
	return theString;
}

/****************************************************************************/
/**
 * @brief	オーディオコントロールの設定値変更ハンドラ．
 *			IOAudioControl::setValueChangeHandler() によって登録される．
 *
 * @param[in]		*target			PdjUsbAudioTopology クラスのインスタンス．
 * @param[in]		*audioControl	設定値が変更されるオーディオコントロール．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::controlChangedHandler(
	OSObject		*target,
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	IOReturn		result = kIOReturnError;

	PdjUsbAudioTopology		*self = OSDynamicCast(PdjUsbAudioTopology, target);
	FailIf(NULL == self, Exit);

	result = self->protectedControlChangedHandler(audioControl, oldValue, newValue);

Exit:
	return result;
}

/****************************************************************************/
/**
 * @brief	オーディオコントロールの設定値変更．
 *
 * @param[in]		*audioControl	設定値が変更されるオーディオコントロール．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::protectedControlChangedHandler(
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::protectedControlChangedHandler()", this);

	IOReturn		result = kIOReturnError;

	switch (audioControl->getType())
	{
		case kIOAudioControlTypeLevel:
			result = this->doVolumeControlChange(audioControl, oldValue, newValue);
			break;
		case kIOAudioControlTypeToggle:
			result = this->doToggleControlChange(audioControl, oldValue, newValue);
			break;
		case kIOAudioControlTypeSelector:
			result = this->doSelectorControlChange(audioControl, oldValue, newValue);
			break;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::protectedControlChangedHandler() = %d", this, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ボリューム変更．
 *
 * @param[in]		*audioControl	設定値が変更されるボリュームコントロール．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::doVolumeControlChange(
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::doVolumeControlChange(%p, 0x%x, 0x%x)",
		this, audioControl, oldValue, newValue);

	IOReturn		result = kIOReturnError;
	UInt8			unitID;
	UInt8			channelNum;
	SInt16			deviceMin = 0;
	SInt16			offset;
	SInt16			newVolume;
	UInt16			volRes = 0;

	unitID		= audioControl->getControlID() & 0xFF;		// <rdar://6413207>
	channelNum	= audioControl->getChannelID();

	if (
			(kIOAudioControlUsageInput == audioControl->getUsage())
	   ||
			(!mDeviceIsInMonoMode)
	   )
	{
		this->getMinVolume(unitID, channelNum, &deviceMin);
		offset = -deviceMin;

		if (newValue < 0)
		{
			newVolume = 0x8000;
		}
		else
		{
			this->getVolumeResolution(unitID, channelNum, &volRes);
			newVolume = (newValue * volRes) - offset;	// <rdar://6377425>
		}

		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::doVolumeControlChange() - Setting channel %d (unit %d) volume to 0x%x",
			this, channelNum, unitID, newVolume);

		result = this->setCurVolume(unitID, channelNum, HostToUSBWord(newVolume));
	}
	else	// mono output case
	{
		FailIf(NULL == mMonoControlsArray, Exit);

		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::doVolumeControlChange() - Performing mono volume control change", this);

		for (UInt8 i = 0; i < mMonoControlsArray->getCount(); i++)
		{
			channelNum = ((OSNumber *)(mMonoControlsArray->getObject(i)))->unsigned8BitValue();

			this->getMinVolume(unitID, channelNum, &deviceMin);
			offset = -deviceMin;

			if (newValue < 0)
			{
				newVolume = 0x8000;
			}
			else
			{
				this->getVolumeResolution(unitID, channelNum, &volRes);
				newVolume = (newValue * volRes) - offset;	// <rdar://6377425>
			}

			result = this->setCurVolume(unitID, channelNum, HostToUSBWord(newVolume));

			DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::doVolumeControlChange() - Set volume for channel %d to 0x%x = %d",
				this, channelNum, newVolume, result);
		}
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::doVolumeControlChange(%p, 0x%x, 0x%x) = 0x%x",
		this, audioControl, oldValue, newValue, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	ミュート状態変更．
 *
 * @param[in]		*audioControl	設定値が変更されるミュートコントロール．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::doToggleControlChange(
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::doToggleControlChange(%p, 0x%x, 0x%x)",
		this, audioControl, oldValue, newValue);

	IOReturn		result;
	UInt8			unitID;
	UInt8			channelNum;

	unitID		= audioControl->getControlID() & 0xFF;	// <rdar://6413207>
	channelNum	= audioControl->getChannelID();

	DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::doToggleControlChange(%p, 0x%x, 0x%x) - unitID = %d, channelNum = %d",
		this, audioControl, oldValue, newValue, unitID, channelNum);

	result = this->setCurMute(unitID, channelNum, HostToUSBWord(newValue));

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::doToggleControlChange(%p, 0x%x, 0x%x) = 0x%x",
		this, audioControl, oldValue, newValue, kIOReturnSuccess);
	return result;
}

/****************************************************************************/
/**
 * @brief	セレクタ設定変更．
 *
 * @param[in]		*audioControl	設定値が変更されるセレクタコントロール．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::doSelectorControlChange(
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::doSelectorControlChange(%p, 0x%x, 0x%x)",
		this, audioControl, oldValue, newValue);

	IOReturn		result = kIOReturnError;

	switch (audioControl->getUsage())
	{
		case kIOAudioControlUsageInput:
			result = this->doInputSelectorChange(audioControl, oldValue, newValue);
			break;
		case kIOAudioControlUsageOutput:
			result = this->doOutputSelectorChange(audioControl, oldValue, newValue);	// <rdar://6413207>
			break;
		case kIOAudioControlUsagePassThru:
			result = this->doPassThruSelectorChange(audioControl, oldValue, newValue);
			break;
	}

	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::doSelectorControlChange(%p, 0x%x, 0x%x) = 0x%x",
		this, audioControl, oldValue, newValue, kIOReturnSuccess);
	return result;
}

/****************************************************************************/
/**
 * @brief	入力用セレクタ設定変更．
 *
 * @param[in]		*audioControl	設定値が変更されるセレクタコントロール．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::doInputSelectorChange(
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::doInputSelectorChange(%p, 0x%x, 0x%x)",
		this, audioControl, oldValue, newValue);

	IOReturn				result = kIOReturnError;
	UInt8					controlInterfaceNum;				// <rdar://5366067>
	UInt8					pathsToOutputTerminal;
	UInt8					pathIndex;
	UInt8					selectorUnitID;
	UInt8					selectorPosition;
	SInt32					engineInfoIndex;					// <rdar://6420832>
	SInt32					streamInfoIndex;					// <rdar://6420832>
	OSDictionary			*engineInfo;
	OSDictionary			*streamInfo;						// <rdar://6420832>
	PdjUsbAudioEngine		*usbAudioEngine;
	OSNumber				*number;
	UInt8					interfaceNum;
	UInt8					altSetting;
	OSArray					*inputGainControlsArray;
	OSArray					*inputMuteControlsArray;			// <rdar://5366067>
	OSArray					*passThruVolControlsArray;			// <rdar://5366067>
	OSArray					*passThruToggleControlsArray;		// <rdar://5366067>
	IOAudioSelectorControl	*playThroughSelector;				// <rdar://5366067>
	UInt32					numGainControls;
	UInt32					numMuteControls;					// <rdar://5366067>
	UInt32					numPassThruVolControls;				// <rdar://5366067>
	UInt32					numPassThruToggleControls;			// <rdar://5366067>
	OSArray					*arrayOfPathsFromOutputTerminal;
	OSArray					*thePath;
	UInt8					inputTerminalID;					// <rdar://6430836>
	UInt8					featureUnitID;
	bool					foundMixerUnit = false;				// <rdar://5366067>

	if (oldValue != newValue)	// <rdar://5366067>
	{
		// <rdar://5366067>
		FailIf(NULL == mControlInterface, Exit);

		controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

		pathsToOutputTerminal	= (newValue & 0xFF000000) >> 24;
		pathIndex				= (newValue & 0x00FF0000) >> 16;
		selectorUnitID			= (newValue & 0x0000FF00) >>  8;
		selectorPosition		= (newValue & 0x000000FF) >>  0;

		result = this->setSelectorSetting(selectorUnitID, selectorPosition);

		FailIf(kIOReturnSuccess != result, Exit);

		engineInfoIndex = (audioControl->getControlID() >>  8) & 0xFF;		// <rdar://6420832>, <rdar://6523676>
		streamInfoIndex = (audioControl->getControlID() >> 16) & 0xFF;		// <rdar://6420832>

		engineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(engineInfoIndex));	// <rdar://6420832>
		FailIf(NULL == engineInfo, Exit);

		usbAudioEngine = OSDynamicCast(PdjUsbAudioEngine, engineInfo->getObject(kEngine));
		FailIf(NULL == usbAudioEngine, Exit);

		streamInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredStreams->getObject(streamInfoIndex));	// <rdar://6420832>
		FailIf(NULL == streamInfo, Exit);

		number = OSDynamicCast(OSNumber, streamInfo->getObject(kInterface));						// <rdar://6420832>
		FailIf(NULL == number, Exit);

		interfaceNum = number->unsigned8BitValue();

		number = OSDynamicCast(OSNumber, streamInfo->getObject(kAltSetting));						// <rdar://6420832>
		FailIf(NULL == number, Exit);

		altSetting = number->unsigned8BitValue();

		inputGainControlsArray		= OSDynamicCast(OSArray, streamInfo->getObject(kInputGainControls));		// <rdar://6420832>
		inputMuteControlsArray		= OSDynamicCast(OSArray, streamInfo->getObject(kInputMuteControls));		// <rdar://5366067>
		passThruVolControlsArray	= OSDynamicCast(OSArray, streamInfo->getObject(kPassThruVolControls));		// <rdar://6420832>
		passThruToggleControlsArray	= OSDynamicCast(OSArray, streamInfo->getObject(kPassThruToggleControls));	// <rdar://5366067>

		playThroughSelector = OSDynamicCast(IOAudioSelectorControl, streamInfo->getObject(kPassThruSelectorControl));	// <rdar://5366067>

		usbAudioEngine->pauseAudioEngine();
		usbAudioEngine->beginConfigurationChange();

		// <rdar://5366067>
		if (NULL != inputGainControlsArray)
		{
			numGainControls = inputGainControlsArray->getCount();

			for (UInt32 i = 0; i < numGainControls; i++)
			{
				usbAudioEngine->removeDefaultAudioControl((IOAudioLevelControl *)inputGainControlsArray->getObject(i));
			}

			inputGainControlsArray->flushCollection();
			streamInfo->removeObject(kInputGainControls);
		}

		// <rdar://5366067>
		if (NULL != inputMuteControlsArray)
		{
			numMuteControls = inputMuteControlsArray->getCount();

			for (UInt32 i = 0; i < numMuteControls; i++)
			{
				usbAudioEngine->removeDefaultAudioControl((IOAudioLevelControl *)inputMuteControlsArray->getObject(i));
			}

			inputMuteControlsArray->flushCollection();
			streamInfo->removeObject(kInputMuteControls);
		}

		// <rdar://5366067>
		if (NULL != passThruVolControlsArray)
		{
			numPassThruVolControls = passThruVolControlsArray->getCount();

			for (UInt32 i = 0; i < numPassThruVolControls; i++)
			{
				usbAudioEngine->removeDefaultAudioControl((IOAudioLevelControl *)passThruVolControlsArray->getObject(i));
			}

			passThruVolControlsArray->flushCollection();
			streamInfo->removeObject(kPassThruVolControls);
		}

		// <rdar://5366067>
		if (NULL != passThruToggleControlsArray)
		{
			numPassThruToggleControls = passThruToggleControlsArray->getCount();

			for (UInt32 i = 0; i < numPassThruToggleControls; i++)
			{
				usbAudioEngine->removeDefaultAudioControl((IOAudioLevelControl *)passThruToggleControlsArray->getObject(i));
			}

			passThruToggleControlsArray->flushCollection();
			streamInfo->removeObject(kPassThruToggleControls);
		}

		// <rdar://5366067>
		if (NULL != playThroughSelector)
		{
			usbAudioEngine->removeDefaultAudioControl(playThroughSelector);
			streamInfo->removeObject(kPassThruSelectorControl);
		}

		arrayOfPathsFromOutputTerminal = OSDynamicCast(OSArray, mControlGraph->getObject(pathsToOutputTerminal));
		FailIf(NULL == arrayOfPathsFromOutputTerminal, Exit);

		thePath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(pathIndex));
		FailIf(NULL == thePath, Exit);

		number = OSDynamicCast(OSNumber, thePath->getLastObject());		// <rdar://6430836>
		FailIf(NULL == number, Exit);

		inputTerminalID = number->unsigned8BitValue();

		featureUnitID = this->getBestFeatureUnitInPath(
									thePath,
									kIOAudioControlUsageInput,
									interfaceNum,
									altSetting,
									kVolumeControl
							  );
		if (0 != featureUnitID)	// <rdar://5366067>
		{
			this->addVolumeControls(
						usbAudioEngine,
						featureUnitID,
						inputTerminalID,
						interfaceNum,
						altSetting,
						kIOAudioControlUsageInput
				  );	// <rdar://6413207>

			this->addMuteControl(
						usbAudioEngine,
						featureUnitID,
						inputTerminalID,
						interfaceNum,
						altSetting,
						kIOAudioControlUsageInput
				  );	// <rdar://6413207>
		}
		else
		{
			featureUnitID = this->getBestFeatureUnitInPath(
										thePath,
										kIOAudioControlUsageInput,
										interfaceNum,
										altSetting,
										kMuteControl
								  );
			if (0 != featureUnitID)
			{
				this->addMuteControl(
							usbAudioEngine,
							featureUnitID,
							inputTerminalID,
							interfaceNum,
							altSetting,
							kIOAudioControlUsageInput
					  );	// <rdar://6413207>
			}
		}

		// <rdar://5366067> If the input path has a mixer in it, then don't create playthru path.
		foundMixerUnit = false;

		for (UInt32 unitIndex = thePath->getCount() - 2; unitIndex > 0; unitIndex--)
		{
			UInt8		subType;	// <rdar://5366067>

			number = OSDynamicCast(OSNumber, thePath->getObject(unitIndex));

			if (
					(NULL != number)
			   &&
					(
						kIOReturnSuccess == mConfigDictionary->getSubType(
																	&subType,
																	controlInterfaceNum,
																	0,
																	number->unsigned8BitValue()
															   )
					)
			   )
			{
				if (MIXER_UNIT == subType)
				{
					foundMixerUnit = true;
					break;
				}
			}
		}

		if (!foundMixerUnit)
		{
			// <rdar://5366067>
			OSArray		*playThroughPaths = this->getPlaythroughPaths(inputTerminalID);

			if (playThroughPaths)
			{
				DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "::doInputSelectorChange() - performing playthrough setup");

				this->doPlaythroughSetup(usbAudioEngine, playThroughPaths, interfaceNum, altSetting, inputTerminalID);

				playThroughPaths->release();
			}
		}

		usbAudioEngine->updateChannelNames(thePath, interfaceNum, altSetting);	// <rdar://6430836> <rdar://problem/6706026>
		usbAudioEngine->completeConfigurationChange();
		usbAudioEngine->resumeAudioEngine();
	}
	else
	{
		result = kIOReturnSuccess;
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::doInputSelectorChange(%p, 0x%x, 0x%x) = 0x%x",
		this, audioControl, oldValue, newValue, result);
	return result;
}

/****************************************************************************/
/**
 * @brief	出力用セレクタ設定変更．
 *
 * @param[in]		*audioControl	設定値が変更されるセレクタコントロール．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
// <rdar://6413207>
IOReturn
PdjUsbAudioTopology::doOutputSelectorChange(
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::doOutputSelectorChange(%p, 0x%x, 0x%x)",
		this, audioControl, oldValue, newValue);

	UInt8					controlInterfaceNum;		// <rdar://5366067>
	UInt8					selectedOutputTerminalID;
	SInt32					engineInfoIndex;			// <rdar://6420832>
	SInt32					streamInfoIndex;			// <rdar://6420832>
	OSDictionary			*engineInfo;
	PdjUsbAudioEngine		*usbAudioEngine;
	OSDictionary			*streamInfo;				// <rdar://6420832>
	OSNumber				*number;
	UInt8					interfaceNum;
	UInt8					altSetting;
	OSArray					*outputVolControlsArray;
	OSArray					*outputMuteControlsArray;
	UInt32					numOutputTerminalArrays;
	OSArray					*thePath = NULL;
	UInt8					volFeatureUnitID;			// <rdar://5366067>
	UInt8					muteFeatureUnitID;			// <rdar://5366067>

	if (oldValue != newValue)
	{
		// <rdar://5366067>
		FailIf(NULL == mControlInterface, Exit);

		controlInterfaceNum = mControlInterface->getInterfaceDescriptor()->bInterfaceNumber;

		selectedOutputTerminalID = newValue & 0x000000FF;

		engineInfoIndex = (audioControl->getControlID() >>  8) & 0xFF;	// <rdar://6420832>, <rdar://6523676>
		streamInfoIndex = (audioControl->getControlID() >> 16) & 0xFF;	// <rdar://6420832>

		engineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(engineInfoIndex));	// <rdar://6420832>
		FailIf(NULL == engineInfo, Exit);

		usbAudioEngine = OSDynamicCast(PdjUsbAudioEngine, engineInfo->getObject(kEngine));
		FailIf(NULL == usbAudioEngine, Exit);

		streamInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredStreams->getObject(streamInfoIndex));	// <rdar://6420832>
		FailIf(NULL == streamInfo, Exit);

		number = OSDynamicCast(OSNumber, streamInfo->getObject(kInterface));						// <rdar://6420832>
		FailIf(NULL == number, Exit);

		interfaceNum = number->unsigned8BitValue();

		number = OSDynamicCast(OSNumber, streamInfo->getObject(kAltSetting));						// <rdar://6420832>
		FailIf(NULL == number, Exit);

		altSetting = number->unsigned8BitValue();

		outputVolControlsArray  = OSDynamicCast(OSArray, streamInfo->getObject(kOutputVolControls));	// <rdar://6420832>
		outputMuteControlsArray = OSDynamicCast(OSArray, streamInfo->getObject(kOutputMuteControls));

		usbAudioEngine->pauseAudioEngine();
		usbAudioEngine->beginConfigurationChange();

		if (NULL != outputVolControlsArray)
		{
			UInt32		numOutputVolControls = outputVolControlsArray->getCount();

			for (UInt32 i = 0; i < numOutputVolControls; i++)
			{
				usbAudioEngine->removeDefaultAudioControl((IOAudioLevelControl *)outputVolControlsArray->getObject(i));
			}

			outputVolControlsArray->flushCollection();
			streamInfo->removeObject(kOutputVolControls);
		}

		if (NULL != outputMuteControlsArray)
		{
			UInt32		numOutputMuteControls = outputMuteControlsArray->getCount();

			for (UInt32 i = 0; i < numOutputMuteControls; i++)
			{
				usbAudioEngine->removeDefaultAudioControl((IOAudioLevelControl *)outputMuteControlsArray->getObject(i));
			}

			outputMuteControlsArray->flushCollection();
			streamInfo->removeObject(kOutputMuteControls);
		}

		numOutputTerminalArrays = mControlGraph->getCount();

		for (UInt32 pathsToOutputTerminalN = 0; pathsToOutputTerminalN < numOutputTerminalArrays; pathsToOutputTerminalN++)
		{
			OSArray		*arrayOfPathsFromOutputTerminal;
			OSArray		*aPath;
			OSNumber	*theUnitIDNum;
			UInt8		outputUnitID;

			arrayOfPathsFromOutputTerminal = OSDynamicCast(OSArray, mControlGraph->getObject(pathsToOutputTerminalN));
			FailIf(NULL == arrayOfPathsFromOutputTerminal, Exit);

			aPath = OSDynamicCast(OSArray, arrayOfPathsFromOutputTerminal->getObject(0));
			FailIf(NULL == aPath, Exit);

			theUnitIDNum = OSDynamicCast(OSNumber, aPath->getObject(0));
			FailIf(NULL == theUnitIDNum, Exit);

			outputUnitID = theUnitIDNum->unsigned8BitValue();

			if (outputUnitID == selectedOutputTerminalID)
			{
				thePath = aPath;
				break;
			}
		}

		FailIf(NULL == thePath, Exit);

		number = OSDynamicCast(OSNumber, thePath->getLastObject());	// <rdar://6430836>
		FailIf(NULL == number, Exit);

		volFeatureUnitID = this->getBestFeatureUnitInPath(
										thePath,
										kIOAudioControlUsageOutput,
										interfaceNum,
										altSetting,
										kVolumeControl
								 );
		if (0 != volFeatureUnitID)
		{
			this->addVolumeControls(
						usbAudioEngine,
						volFeatureUnitID,
						selectedOutputTerminalID,
						interfaceNum,
						altSetting,
						kIOAudioControlUsageOutput
				  );	// <rdar://6413207>
		}

		muteFeatureUnitID = this->getBestFeatureUnitInPath(
										thePath,
										kIOAudioControlUsageOutput,
										interfaceNum,
										altSetting,
										kMuteControl
								  );
		if (0 != muteFeatureUnitID)
		{
			this->addMuteControl(
						usbAudioEngine,
						muteFeatureUnitID,
						selectedOutputTerminalID,
						interfaceNum,
						altSetting,
						kIOAudioControlUsageOutput
				  );	// <rdar://6413207>
		}

		// <rdar://5366067> Handle the case where the volume & mute controls are on different feature units.
		if (volFeatureUnitID != muteFeatureUnitID)
		{
			UInt8		numControls;	// <rdar://5366067>

			// Unmute the volume feature unit.
			if (volFeatureUnitID)
			{
				FailIf(
						kIOReturnSuccess != mConfigDictionary->getNumControls(
																	&numControls,
																	controlInterfaceNum,
																	0,
																	volFeatureUnitID
															   ),
						Exit
				);

				for (UInt8 channelNum = 0; channelNum < numControls; channelNum++)
				{
					if (mConfigDictionary->channelHasMuteControl(controlInterfaceNum, 0, volFeatureUnitID, channelNum))
					{
						this->setCurMute(volFeatureUnitID, channelNum, 0);
					}
				}
			}

			// Set the gain of the mute feature unit.
			if (muteFeatureUnitID)
			{
				FailIf(
						kIOReturnSuccess != mConfigDictionary->getNumControls(
																	&numControls,
																	controlInterfaceNum,
																	0,
																	muteFeatureUnitID
															   ),
						Exit
				);

				for (UInt8 channelNum = 0; channelNum < numControls; channelNum++)
				{
					if (mConfigDictionary->channelHasVolumeControl(controlInterfaceNum, 0, muteFeatureUnitID, channelNum))
					{
						SInt16		deviceMax;	// <rdar://5366067>

						if (kIOReturnSuccess == this->getMaxVolume(muteFeatureUnitID, channelNum, &deviceMax))
						{
							this->setCurVolume(muteFeatureUnitID, channelNum, ((deviceMax >= 0) ? 0 : deviceMax));
						}
					}
				}
			}
		}

		usbAudioEngine->updateChannelNames(thePath, interfaceNum, altSetting);	// <rdar://6430836> <rdar://problem/6706026>
		usbAudioEngine->completeConfigurationChange();
		usbAudioEngine->resumeAudioEngine();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::doOutputSelectorChange(%p, 0x%x, 0x%x) = 0x%x",
		this, audioControl, oldValue, newValue, kIOReturnSuccess);
	return kIOReturnSuccess;
}

/****************************************************************************/
/**
 * @brief	パススルー用セレクタ設定変更．
 *
 * @param[in]		*audioControl	設定値が変更されるセレクタコントロール．
 * @param[in]		oldValue		変更前の値．
 * @param[in]		newValue		変更後の値．
 *
 * @return			IOReturn コード．
 */
IOReturn
PdjUsbAudioTopology::doPassThruSelectorChange(
	IOAudioControl	*audioControl,
	SInt32			oldValue,
	SInt32			newValue
)
{
	DP(DBG_TRACE, "+ " LOGSTRAUDIOTOPOLOGY "[%p]::doPassThruSelectorChange(%p, 0x%x, 0x%x)",
		this, audioControl, oldValue, newValue);

	PdjUsbAudioEngine		*usbAudioEngine;
	OSArray					*playThroughPaths;
	OSArray					*passThruVolControlsArray;
	OSArray					*passThruToggleControlsArray;	// <rdar://5366067>
	OSArray					*thePath;
	OSNumber				*number;
	OSDictionary			*engineInfo;
	OSDictionary			*streamInfo;					// <rdar://6420832>
	UInt32					numPassThruVolControls;			// <rdar://5366067>
	UInt32					numPassThruToggleControls;		// <rdar://5366067>
	UInt8					interfaceNum;
	UInt8					altSetting;
	UInt8					featureUnitID;
	UInt8					inputTerminalID;				// <rdar://6413207>
	UInt8					pathIndex;
	SInt32					engineInfoIndex;				// <rdar://6420832>
	SInt32					streamInfoIndex;				// <rdar://6420832>

	if (oldValue != newValue)	// <rdar://5366067>
	{
		pathIndex = newValue & 0x000000FF;

		DP(DBG_INFO, "? " LOGSTRAUDIOTOPOLOGY "[%p]::doPassThruSelectorChange(%p, 0x%x, 0x%x) - controlID = 0x%x",
			this, audioControl, oldValue, newValue, audioControl->getControlID());

		engineInfoIndex = (audioControl->getControlID() >>  8) & 0xFF;	// <rdar://6420832>, <rdar://6523676>
		streamInfoIndex = (audioControl->getControlID() >> 16) & 0xFF;	// <rdar://6420832>

		engineInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredEngines->getObject(engineInfoIndex));	// <rdar://6420832>
		FailIf(NULL == engineInfo, Exit);

		usbAudioEngine = OSDynamicCast(PdjUsbAudioEngine, engineInfo->getObject(kEngine));
		FailIf(NULL == usbAudioEngine, Exit);

		streamInfo = OSDynamicCast(OSDictionary, mAudioDevice->mRegisteredStreams->getObject(streamInfoIndex));	// <rdar://6420832>
		FailIf(NULL == streamInfo, Exit);

		number = OSDynamicCast(OSNumber, streamInfo->getObject(kInterface));						// <rdar://6420832>
		FailIf(NULL == number, Exit);

		interfaceNum = number->unsigned8BitValue();

		number = OSDynamicCast(OSNumber, streamInfo->getObject(kAltSetting));						// <rdar://6420832>
		FailIf(NULL == number, Exit);

		altSetting = number->unsigned8BitValue();

		passThruVolControlsArray	= OSDynamicCast(OSArray, streamInfo->getObject(kPassThruVolControls));		// <rdar://6420832>
		passThruToggleControlsArray	= OSDynamicCast(OSArray, streamInfo->getObject(kPassThruToggleControls));	// <rdar://5366067>

		usbAudioEngine->pauseAudioEngine();
		usbAudioEngine->beginConfigurationChange();

		if (NULL != passThruVolControlsArray)
		{
			numPassThruVolControls = passThruVolControlsArray->getCount();	// <rdar://5366067>

			for (UInt32 i = 0; i < numPassThruVolControls; i++)
			{
				usbAudioEngine->removeDefaultAudioControl((IOAudioLevelControl *)(passThruVolControlsArray->getObject(i)));
			}

			passThruVolControlsArray->flushCollection();

			streamInfo->removeObject(kPassThruVolControls);
		}

		// <rdar://5366067>
		if (NULL != passThruToggleControlsArray)
		{
			numPassThruToggleControls = passThruToggleControlsArray->getCount();

			for (UInt32 i = 0; i < numPassThruToggleControls; i++)
			{
				usbAudioEngine->removeDefaultAudioControl((IOAudioLevelControl *)(passThruToggleControlsArray->getObject(i)));
			}

			passThruToggleControlsArray->flushCollection();

			streamInfo->removeObject(kPassThruToggleControls);
		}

		// <rdar://5366067>
		playThroughPaths = OSDynamicCast(OSArray, streamInfo->getObject(kPassThruPathsArray));	// <rdar://6420832>
		FailIf(NULL == playThroughPaths, Exit);

		thePath = OSDynamicCast(OSArray, playThroughPaths->getObject(pathIndex));
		FailIf(NULL == thePath, Exit);

		number = OSDynamicCast(OSNumber, thePath->getLastObject());	// <rdar://6413207>
		FailIf(NULL == number, Exit);

		inputTerminalID = number->unsigned8BitValue();

		featureUnitID = this->getBestFeatureUnitInPath(
									thePath,
									kIOAudioControlUsagePassThru,
									interfaceNum,
									altSetting,
									kVolumeControl
							  );
		if (0 != featureUnitID)
		{
			this->addVolumeControls(
						usbAudioEngine,
						featureUnitID,
						inputTerminalID,
						interfaceNum,
						altSetting,
						kIOAudioControlUsagePassThru
				  );	// <rdar://6413207>
		}

		featureUnitID = this->getBestFeatureUnitInPath(
									thePath,
									kIOAudioControlUsagePassThru,
									interfaceNum,
									altSetting,
									kMuteControl
							  );
		if (0 != featureUnitID)
		{
			this->addMuteControl(
						usbAudioEngine,
						featureUnitID,
						inputTerminalID,
						interfaceNum,
						altSetting,
						kIOAudioControlUsagePassThru
				  );	// <rdar://6413207>
		}

		usbAudioEngine->completeConfigurationChange();
		usbAudioEngine->resumeAudioEngine();
	}

Exit:
	DP(DBG_TRACE, "- " LOGSTRAUDIOTOPOLOGY "[%p]::doPassThruSelectorChange(%p, 0x%x, 0x%x) = 0x%x",
		this, audioControl, oldValue, newValue, kIOReturnSuccess);
	return kIOReturnSuccess;
}
